
// entities.c : central structure tree.

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "strings.h"
#include "lex.h"
#include "error.h"
#include "goptions.h"
#include "entities.h"

#include "typedecl.h"

/******************************************************************************/

typedef struct {
  PREGION  first, last;
} REGION_QUEUE;

typedef struct level_node {
  REGION_QUEUE       regions;
  struct level_node  *outer;
} LEVEL_NODE;

/******************************************************************************/

static LEVEL_NODE  *current;
static long        unique_entity_nr;
static ENTITY_LIST *hidden_entities;

PENTITY  type_bool, type_char, type_wchar, type_byte, type_int, type_uint, type_long;
PENTITY  type_float, type_double;
PENTITY  type_string, type_wstring, type_object, type_array_of_string, type_array_of_object;
PENTITY  type_int_literal, type_float_literal, type_null_literal, type_unsafe_ptr_to_byte;
PENTITY  type_void;
PENTITY  enum_false, enum_true;

PENTITY global_locked_struct_entity;  // locked during struct declaration

/******************************************************************************/

typedef struct entity_from_unit_list {
  PENTITY                      e;
  PENTITY                      unit_decl;
  struct entity_from_unit_list *next;
} ENTITY_FROM_UNIT_LIST;

/******************************************************************************/

PENTITY surrounding_entity ()
{
  LEVEL_NODE *outer;
  PREGION    r;

  outer = current->outer;
  r = outer->regions.last;
  return r->entities.last;
}

/******************************************************************************/

void insert_entity_in_list (PENTITY e, ENTITY_LIST **root)
{
  ENTITY_LIST *n;

  n = malloc (sizeof(ENTITY_LIST));
  if (n == null)
    fatal_out_of_memory_error ("insert_entity_in_list()");

  n->e = e;
  n->next = *root;

  *root = n;
}

/******************************************************************************/

void insert_entity_from_unit_in_list (PENTITY e, PENTITY unit_decl, ENTITY_FROM_UNIT_LIST **root)
{
  ENTITY_FROM_UNIT_LIST *n;

  n = malloc (sizeof(ENTITY_FROM_UNIT_LIST));
  if (n == null)
    fatal_out_of_memory_error ("insert_entity_from_unit_in_list()");

  n->e = e;
  n->unit_decl = unit_decl;
  n->next = *root;

  *root = n;
}

/******************************************************************************/

static void insert_each_entity_of_list_to_list (ENTITY_LIST *list,   // in
                                                ENTITY_LIST **root)  // ref
{
  ENTITY_LIST *l;

  l = list;
  while (l != null)
  {
    insert_entity_in_list (l->e, root);
    l = l->next;
  }
}

/******************************************************************************/

static void insert_each_entity_of_list_to_from_unit_list (ENTITY_LIST           *list,      // in
                                                          PENTITY               unit_decl,
                                                          ENTITY_FROM_UNIT_LIST **root)     // ref
{
  ENTITY_LIST *l;

  l = list;
  while (l != null)
  {
    insert_entity_from_unit_in_list (l->e, unit_decl, root);
    l = l->next;
  }
}

/******************************************************************************/

static bool entity_exists_in_list (PENTITY e, ENTITY_LIST **root)
{
  ENTITY_LIST *p;

  p = *root;

  while (p != null)
  {
    if (p->e == e)
      return true;
    p = p->next;
  }

  return false;
}

/******************************************************************************/

static void deallocate_entity_list (ENTITY_LIST **root)
{
  ENTITY_LIST *p, *temp;

  p = *root;
  *root = null;

  while (p != null)
  {
    temp = p;
    p = p->next;
    free (temp);
  }
}

/******************************************************************************/

static void deallocate_entity_from_unit_list (ENTITY_FROM_UNIT_LIST **root)
{
  ENTITY_FROM_UNIT_LIST *p, *temp;

  p = *root;
  *root = null;

  while (p != null)
  {
    temp = p;
    p = p->next;
    free (temp);
  }
}

/******************************************************************************/

void create_region_level ()
{
  LEVEL_NODE *p;

  p = malloc (sizeof(LEVEL_NODE));
  if (p == null)
    fatal_out_of_memory_error ("create_region_level()");

  memset (p, 0x00, sizeof(LEVEL_NODE));
  p->outer = current;

  current = p;
}

/******************************************************************************/

void unlink_region_level ()
{
  LEVEL_NODE  *p;

  if (current == null)
    fatal_compiler_error ("unlink_region_level()", &token.pos);

  p = current;
  current = current->outer;
  free (p);
}

/******************************************************************************/

PREGION new_region ()
{
  PREGION r;

  r = malloc (sizeof(REGION));
  if (r == NULL)
    fatal_out_of_memory_error ("append_new_region()");

  memset (r, 0x00, sizeof(REGION));

  return r;
}

/******************************************************************************/

void append_region (PREGION r)
{
  if (current == null)
    fatal_compiler_error ("append_new_region()", &token.pos);

  if (current->regions.first == null)
    current->regions.first = r;
  else
    current->regions.last->next = r;

  current->regions.last = r;

  r->next = null;
}

/******************************************************************************/

void cleanup_main_region_after_compilation ()
{
  deallocate_entity_list (&current->regions.last->packages);
  memset (current->regions.last, 0x00, sizeof(REGION));
}

/******************************************************************************/

// returns 0 if OK, -1 if symbol already exists

static int insert_tree (PENTITY  *tree,       // insertion point
                        PENTITY  new_entity,  // new entity to insert
                        bool     *overflow)   // out
{
  PENTITY  p, p1, p2;
  int      rc, cmp, ins0, ins1;
  tiny     bal1, bal2;

  if (*tree == null)    // tree is empty
  {
    *tree = new_entity;
    new_entity->ptr[0] = null;
    new_entity->ptr[1] = null;
    new_entity->bal = 0;

    *overflow = true;
    return 0;
  }

  p = *tree;

  cmp = wstrcmp (new_entity->identifier_or_null,
                 p->identifier_or_null);

  if (cmp == 0)   // identifier was found
    return -1;    // already exists

  if (cmp < 0)
  {
    ins0 = 0;     // insertion in left subtree
    ins1 = 1;
    bal1 = +1;
    bal2 = -1;
  }
  else    // cmp > 0
  {
    ins0 = 1;     // insertion in right subtree
    ins1 = 0;
    bal1 = -1;
    bal2 = +1;
  }

  rc = insert_tree (&p->ptr[ins0], new_entity, overflow);
  if (rc)
    return rc;    // already exists

  if (*overflow)
  {
    if (p->bal == bal1)
    {
      p->bal = 0;
      *overflow = false;
    }
    else if (p->bal == 0)
    {
      p->bal = bal2;
    }
    else  // p->bal == bal2
    {
      p1 = p->ptr[ins0];

      if (p1->bal == bal2)
      {
        p->ptr[ins0] = p1->ptr[ins1];
        p1->ptr[ins1] = p;
        p->bal = 0;

        p = p1;
        *tree = p1;
      }
      else
      {
        p2 = p1->ptr[ins1];

        p1->ptr[ins1] = p2->ptr[ins0];
        p2->ptr[ins0] = p1;
        p->ptr[ins0]  = p2->ptr[ins1];
        p2->ptr[ins1] = p;

        if (p2->bal == bal2)
          p->bal = bal1;
        else
          p->bal = 0;

        if (p2->bal == bal1)
          p1->bal = bal2;
        else
          p1->bal = 0;

        p = p2;
        *tree = p2;
      }

      p->bal = 0;
      *overflow = false;
    }
  }

  return 0;
}

/************************************************************************/

// returns null if not found

PENTITY find_tree (PENTITY tree, wstring identifier)
{
  PENTITY  p;
  int      cmp;

  p = tree;
  while (p != null)
  {
    cmp = wstrcmp (identifier, p->identifier_or_null);

    if (cmp == 0)   // identifier was found
      return p;

    if (cmp < 0)
      p = p->ptr[0];
    else
      p = p->ptr[1];
  }

  return null;
}

/************************************************************************/

static uint size_of_entity (ENTITY_KIND kind)
{
  static const uint size[LAST_ENTITY_DENOTING_A_STATEMENT+1] =
    {sizeof(AN_INTEGER_TYPE_ENTITY),
     sizeof(A_FLOAT_TYPE_ENTITY),
     sizeof(AN_ENUMERATION_TYPE_ENTITY),
     sizeof(AN_OPEN_ARRAY_TYPE_ENTITY),
     sizeof(AN_ARRAY_TYPE_ENTITY),
     sizeof(A_STRUCT_TYPE_ENTITY),
     sizeof(A_CONSTRAINED_STRUCT_TYPE_ENTITY),
     sizeof(A_UNION_TYPE_ENTITY),
     sizeof(A_POINTER_TYPE_ENTITY),
     sizeof(A_FUNCTION_POINTER_TYPE_ENTITY),
     sizeof(AN_UNSAFE_POINTER_TYPE_ENTITY),
     sizeof(A_RENAMED_TYPE_ENTITY),
     sizeof(AN_INCOMPLETE_TYPE_ENTITY),
     sizeof(AN_OPAQUE_TYPE_ENTITY),
     sizeof(A_GENERIC_TYPE_ENTITY),
     sizeof(A_NULL_POINTER_TYPE_ENTITY),
     sizeof(A_VOID_TYPE_ENTITY),
     sizeof(LAST_ENTITY_DENOTING_A_TYPE),
     sizeof(AN_INCOMPLETE_TYPE_COMPLETITION_ENTITY),
     sizeof(AN_OPAQUE_TYPE_COMPLETITION_ENTITY),
     sizeof(A_FIELD_ENTITY),
     sizeof(A_VARYING_FIELD_ENTITY),
     sizeof(AN_ENUMERATION_LITERAL_ENTITY),
     sizeof(A_CONSTANT_ENTITY),
     sizeof(A_GLOBAL_VARIABLE_ENTITY),
     sizeof(A_LOCAL_VARIABLE_ENTITY),
     sizeof(A_REFERENCE_ENTITY),
     sizeof(A_PARAMETER_ENTITY),
     sizeof(A_GENERIC_FUNCTION_ENTITY),
     sizeof(A_FUNCTION_DECLARATION_ENTITY),
     sizeof(A_FUNCTION_BODY_ENTITY),
     sizeof(A_PACKAGE_DECLARATION_ENTITY),
     sizeof(A_PACKAGE_BODY_ENTITY),
     sizeof(A_UNIT_INTERFACE_ENTITY),
     sizeof(A_UNIT_BODY_ENTITY),
     sizeof(FIRST_ENTITY_DENOTING_A_STATEMENT),
     sizeof(A_CLEAR_STATEMENT_ENTITY),
     sizeof(AN_ASSIGNMENT_STATEMENT_ENTITY),
     sizeof(A_PRE_OR_POSTFIX_STATEMENT_ENTITY),
     sizeof(A_FUNCTION_CALL_STATEMENT_ENTITY),
     sizeof(A_RETURN_STATEMENT_ENTITY),
     sizeof(A_BREAK_STATEMENT_ENTITY),
     sizeof(A_CONTINUE_STATEMENT_ENTITY),
     sizeof(A_FREE_STATEMENT_ENTITY),
     sizeof(AN_ABORT_STATEMENT_ENTITY),
     sizeof(AN_ASSERT_STATEMENT_ENTITY),
     sizeof(A_SLEEP_STATEMENT_ENTITY),
     sizeof(A_CODE_STATEMENT_ENTITY),
     sizeof(AN_UNUSED_STATEMENT_ENTITY),
     sizeof(A_BLOCK_STATEMENT_ENTITY),
     sizeof(AN_IF_STATEMENT_ENTITY),
     sizeof(A_SWITCH_STATEMENT_ENTITY),
     sizeof(A_WHILE_STATEMENT_ENTITY),
     sizeof(A_FOR_STATEMENT_ENTITY),
     sizeof(LAST_ENTITY_DENOTING_A_STATEMENT),
    };

  if (kind > LAST_ENTITY_DENOTING_A_STATEMENT)
    fatal_compiler_error ("size_of_entity", &token.pos);

  return size[kind];
}

/************************************************************************/

// give an identifier to an entity that was declared without identifier

void patch_entity_by_setting_an_identifier (PENTITY       e,
                                            wstring       identifier,   // can be empty string
                                            TEXT_POSITION *pos)
{
  bool     dummy, found;
  int      rc;
  PREGION  r;

  if (wstrlen(identifier) > 0)
  {
    e->identifier_or_null = malloc (2*(wstrlen(identifier)+1));
    if (e->identifier_or_null == null)
      fatal_out_of_memory_error ("append_entity(3)");
    wstrcpy (e->identifier_or_null, identifier);


    // check if the identifier exists in any region of this level,
    // EXCEPT LAST ONE (because this will be checked by insert_tree)

    found = false;

    r = current->regions.first;
    while (r != current->regions.last)
    {
      if (find_tree (r->identifier_btree, e->identifier_or_null) != null)
      {
        found = true;
        break;
      }

      r = r->next;
    }


    // if we're inside a compound statement, search it also in surrounding levels

    {
      LEVEL_NODE  *n;
      PREGION     r0;
      PENTITY     e0;
      ENTITY_KIND k;

      n = current->outer;     // surrounding level

      for (;;)
      {
        if (n == null)
          break;

        r0 = n->regions.last;
        if (r0 == null)
          break;

        e0 = r0->entities.last;
        if (e0 == null)
          break;

        k = e0->kind;

        if (k >= FIRST_ENTITY_DENOTING_A_STATEMENT && k <= LAST_ENTITY_DENOTING_A_STATEMENT)
        {
          // last entity of scope to be searched contains a statement

          r = n->regions.first;
          while (r != null)
          {
            if (find_tree (r->identifier_btree, e->identifier_or_null) != null)
            {
              found = true;
              break;
            }

            r = r->next;
          }
        }
        else
        {
          break;   // it's not a statement
        }

        n = n->outer;
      }
    }


    // insert it even if it's not unique

    rc = insert_tree (&current->regions.last->identifier_btree, e, &dummy);
    if (rc < 0)   // duplicate
      found = true;


    if (found)
      semantic_error ("identifier declared twice", pos);
  }
}

/******************************************************************************/

// create new entity and insert it in symbol tree,
// but does not append it in elaboration queue or assign it a unique nr.

PENTITY create_new_entity (ENTITY_KIND   kind,
                           void          *entity_info,
                           wstring       identifier,   // can be empty string
                           TEXT_POSITION *pos)
{
  PENTITY  e;
  uint     size;

  if (current == null || current->regions.last == null)
    fatal_compiler_error ("append_entity()", &token.pos);

  size = size_of_entity (kind);

  e = malloc (sizeof(ENTITY));
  if (e == NULL)
    fatal_out_of_memory_error ("append_entity(1)");
  memset (e, 0x00, sizeof(ENTITY));

  e->kind = kind;
  e->info = malloc (size);
  if (e->info == null)
    fatal_out_of_memory_error ("append_entity(2)");
  memcpy (e->info, entity_info, size);

  patch_entity_by_setting_an_identifier (e, identifier, pos);

  return e;
}

/******************************************************************************/

// assign unique entity nr and append newly created entity in elaboration queue
// (this function is separate so that forward references in entity links are avoided)

void append_entity_in_entity_queue (PENTITY e)
{
  e->nr = ++unique_entity_nr;

  if (current->regions.last->entities.first == null)
    current->regions.last->entities.first = e;
  else
    current->regions.last->entities.last->next = e;

  current->regions.last->entities.last = e;
}

/******************************************************************************/

// create new entity and insert it in symbol tree and append it in elaboration queue

PENTITY append_new_entity (ENTITY_KIND   kind,
                           void          *entity_info,
                           wstring       identifier,   // can be empty string
                           TEXT_POSITION *pos)
{
  PENTITY e;

  e = create_new_entity (kind, entity_info, identifier, pos);

  append_entity_in_entity_queue (e);

  return e;
}

/******************************************************************************/

PENTITY append_new_entity_for_instantiation
                          (ENTITY_KIND   kind,
                           void          *entity_info,
                           wstring       identifier,   // can be empty string
                           TEXT_POSITION *pos,
                           REGION        *region)
{
  PENTITY  e;
  uint     size;
  int      rc;
  bool     dummy;

  size = size_of_entity (kind);

  e = malloc (sizeof(ENTITY));
  if (e == NULL)
    fatal_out_of_memory_error ("append_entity(4)");
  memset (e, 0x00, sizeof(ENTITY));

  e->kind = kind;
  e->info = malloc (size);
  if (e->info == null)
    fatal_out_of_memory_error ("append_entity(5)");
  memcpy (e->info, entity_info, size);

  if (wstrlen(identifier) > 0)
  {
    e->identifier_or_null = malloc (2*(wstrlen(identifier)+1));
    if (e->identifier_or_null == null)
      fatal_out_of_memory_error ("append_entity(3)");
    wstrcpy (e->identifier_or_null, identifier);

    rc = insert_tree (&region->identifier_btree, e, &dummy);
    if (rc < 0)   // duplicate
      fatal_compiler_error ("app.e.inst", &token.pos);
  }

  e->nr = ++unique_entity_nr;

  if (region->entities.first == null)
    region->entities.first = e;
  else
    region->entities.last->next = e;

  region->entities.last = e;

  return e;
}

/******************************************************************************/

void append_new_unit_entity_for_import
                      (PENTITY       e,
                       wstring       identifier,   // can be empty string
                       TEXT_POSITION *pos,
                       bool          add_in_package_list)
{
  int     rc;
  bool    dummy;
  PENTITY q;


  e->next = null;

  // check if this entity already exists in the elaboration queue

  q = current->regions.last->entities.first;
  while (q != null)
  {
    if (q == e)
    {
      semantic_error ("this unit must not be imported twice", pos);
      return;
    }

    q = q->next;
  }


  free (e->identifier_or_null);
  e->identifier_or_null = null;

  if (wstrlen(identifier) > 0)
  {
    e->identifier_or_null = malloc (2*(wstrlen(identifier)+1));
    if (e->identifier_or_null == null)
      fatal_out_of_memory_error ("append_new_unit_entity(1)");
    wstrcpy (e->identifier_or_null, identifier);

    rc = insert_tree (&current->regions.last->identifier_btree, e, &dummy);
    if (rc < 0)  // duplicate
      semantic_error ("a unit of this name already exists", pos);
  }


  if (current->regions.last->entities.first == null)
    current->regions.last->entities.first = e;
  else
    current->regions.last->entities.last->next = e;

  current->regions.last->entities.last = e;


  if (add_in_package_list)
  {
    insert_entity_in_list (e, &current->regions.last->packages);
  }
}

/******************************************************************************/

typedef struct {
  TOKEN_KIND kind;
  PENTITY    e;
} TYPE_NAME_INFO;

#define MAX_TYPES  (24+1)

static TYPE_NAME_INFO type_name_info[MAX_TYPES];
static int            type_name_count;

/******************************************************************************/

INTEGER_INFO integer_info[MAX_INTEGER_TYPES] =
 {{a_int1,  L"int1",  TOKEN_int1,  TOKEN_tiny,   true,  1,   -128,   +127, 0xFF},
  {a_int2,  L"int2",  TOKEN_int2,  TOKEN_short,  true,  2, -32768, +32767, 0xFFFF},
  {a_int4,  L"int4",  TOKEN_int4,  TOKEN_int,    true,  4, -(int8)2147483648, (int8)2147483647, 0xFFFFFFFF},
  {a_int8,  L"int8",  TOKEN_int8,  TOKEN_long,   true,  8, -(int8)9223372036854775808, (int8)9223372036854775807, 0xFFFFFFFFFFFFFFFF},
  {a_uint1, L"uint1", TOKEN_uint1, TOKEN_byte,   false, 1, 0, 255, 0xFF},
  {a_uint2, L"uint2", TOKEN_uint2, TOKEN_ushort, false, 2, 0, 65535, 0xFFFF},
  {a_uint4, L"uint4", TOKEN_uint4, TOKEN_uint,   false, 4, 0, 4294967295, 0xFFFFFFFF}};

FLOAT_INFO float_info [MAX_FLOAT_TYPES] =
 {{a_float,  L"float",  TOKEN_float,  TOKEN_float4,  4},
  {a_double, L"double", TOKEN_double, TOKEN_float8,  8}};

/******************************************************************************/

// returns the type entity corresponding to the token,
// or null if the token does not denote a type.

PENTITY type_entity_of_token (TOKEN_KIND kind)
{
  int i;

  for (i=0; type_name_info[i].kind != LAST_TOKEN; i++)
  {
    if (type_name_info[i].kind == kind)
      return type_name_info[i].e;
  }

  return null;
}

/******************************************************************************/

static void add_type (TOKEN_KIND kind, PENTITY e)
{
  if (type_name_count == MAX_TYPES)
    fatal_compiler_error ("add_type", &token.pos);
  type_name_info[type_name_count].kind = kind;
  type_name_info[type_name_count].e    = e;
  type_name_count++;
}

/******************************************************************************/

void create_predefined_entities ()
{
  A_UNIT_INTERFACE_ENTITY    std;
  AN_INTEGER_TYPE_ENTITY     integer;
  A_FLOAT_TYPE_ENTITY        floating;
  AN_ENUMERATION_TYPE_ENTITY enumer;
  AN_OPEN_ARRAY_TYPE_ENTITY  array;
  int                        i;
  PENTITY                    e;
  A_NULL_POINTER_TYPE_ENTITY null_lit;
  AN_ENUMERATION_LITERAL_ENTITY lit;
  AN_UNSAFE_POINTER_TYPE_ENTITY ptr;
  A_VOID_TYPE_ENTITY            vtype;

  create_region_level ();
  append_region (new_region());

  // create empty predefined unit
  memset (&std, 0x00, sizeof(std));
  std.declarations = new_region();
  append_new_entity (A_UNIT_INTERFACE, &std, L"<>", &token.pos);

  create_region_level ();
  append_region (std.declarations);

  for (i=0; i<MAX_INTEGER_TYPES; i++)
  {
    memset (&integer, 0x00, sizeof(integer));
    integer.type = integer_info[i].typ;
    e = append_new_entity (AN_INTEGER_TYPE, &integer, integer_info[i].name, &token.pos);

    if (integer.type == a_uint1)
      type_byte = e;

    if (integer.type == a_int4)
      type_int = e;

    if (integer.type == a_uint4)
      type_uint = e;

    if (integer.type == a_int8)
      type_long = e;

    add_type (integer_info[i].token1, e);
    add_type (integer_info[i].token2, e);
  }

  for (i=0; i<MAX_FLOAT_TYPES; i++)
  {
    memset (&floating, 0x00, sizeof(floating));
    floating.type = float_info[i].typ;
    e = append_new_entity (A_FLOAT_TYPE, &floating, float_info[i].name, &token.pos);
    add_type (float_info[i].token1, e);
    add_type (float_info[i].token2, e);

    if (floating.type == a_float)
      type_float = e;

    if (floating.type == a_double)
      type_double = e;
  }


  memset (&enumer, 0x00, sizeof(enumer));
  enumer.base = a_uint1;
  enumer.last = 255;
  type_char = append_new_entity (AN_ENUMERATION_TYPE, &enumer, L"char", &token.pos);
  add_type (TOKEN_char, type_char);

  memset (&enumer, 0x00, sizeof(enumer));
  enumer.base = a_uint2;
  enumer.last = 65535;
  type_wchar = append_new_entity (AN_ENUMERATION_TYPE, &enumer, L"wchar", &token.pos);
  add_type (TOKEN_wchar, type_wchar);

  memset (&array, 0x00, sizeof(array));
  array.element = type_char;
  type_string = append_new_entity (AN_OPEN_ARRAY_TYPE, &array, L"string", &token.pos);
  add_type (TOKEN_string, type_string);

  memset (&array, 0x00, sizeof(array));
  array.element = type_wchar;
  type_wstring = append_new_entity (AN_OPEN_ARRAY_TYPE, &array, L"wstring", &token.pos);
  add_type (TOKEN_wstring, type_wstring);

  memset (&array, 0x00, sizeof(array));
  array.element = type_byte;
  type_object = append_new_entity (AN_OPEN_ARRAY_TYPE, &array, L"object", &token.pos);
  add_type (TOKEN_object, type_object);


  // array of string (for attribute 'string)

  memset (&array, 0x00, sizeof(array));
  array.element = type_string;
  type_array_of_string = append_new_entity (AN_OPEN_ARRAY_TYPE, &array, L"", &token.pos);


  // array of object (for varying parameter lists)

  memset (&array, 0x00, sizeof(array));
  array.element = type_object;
  type_array_of_object = append_new_entity (AN_OPEN_ARRAY_TYPE, &array, L"", &token.pos);


  // now array of string is defined, we can define type bool and its enumeration literals

  memset (&enumer, 0x00, sizeof(enumer));
  enumer.base = a_uint1;
  enumer.last = 1;
  type_bool = append_new_entity (AN_ENUMERATION_TYPE, &enumer, L"bool", &token.pos);
  add_type (TOKEN_bool, type_bool);

  memset (&lit, 0x00, sizeof(lit));
  lit.type = type_bool;
  lit.value = 0;
  enum_false = append_new_entity (AN_ENUMERATION_LITERAL, &lit, L"false", &token.pos);

  memset (&lit, 0x00, sizeof(lit));
  lit.type = type_bool;
  lit.value = 1;
  enum_true = append_new_entity (AN_ENUMERATION_LITERAL, &lit, L"true", &token.pos);

  create_enumeration_literal_string_table (type_bool);



  // add pseudo types

  memset (&integer, 0x00, sizeof(integer));
  integer.type = a_int8;
  type_int_literal = append_new_entity (AN_INTEGER_TYPE, &integer, L"int_literal", &token.pos);

  memset (&floating, 0x00, sizeof(floating));
  floating.type = a_double;
  type_float_literal = append_new_entity (A_FLOAT_TYPE, &floating, L"float_literal", &token.pos);

  memset (&null_lit, 0x00, sizeof(null_lit));
  type_null_literal = append_new_entity (A_NULL_POINTER_TYPE, &null_lit, L"null_literal", &token.pos);

  memset (&ptr, 0x00, sizeof(ptr));
  ptr.designated_type = type_byte;
  type_unsafe_ptr_to_byte = append_new_entity (AN_UNSAFE_POINTER_TYPE, &ptr, L"", &token.pos);


  // type void

  memset (&vtype, 0x00, sizeof(vtype));
  type_void = append_new_entity (A_VOID_TYPE, &vtype, L"", &token.pos);


  add_type (LAST_TOKEN, NULL);

  unlink_region_level ();
  cleanup_main_region_after_compilation ();
}

/******************************************************************************/

// produces error messages if not found or ambiguous
// returns null if not found

PENTITY find_entity (wstring identifier, TEXT_POSITION *pos)
{
  LEVEL_NODE  *n;
  PREGION     r;
  PENTITY     e, result;
  ENTITY_KIND kind;
  int         count;
  ENTITY_FROM_UNIT_LIST *list, *list2;

  n = current;
  while (n != null)
  {
    r = n->regions.first;

    while (r != null)
    {
      e = find_tree (r->identifier_btree, identifier);
      if (e != null)
        return e;

      r = r->next;
    }


    // search within all local unit interfaces or package declarations

    count = 0;
    list = null;
    r = n->regions.first;

    while (r != null)
    {
      ENTITY_LIST *l;

      l = r->packages;     // list of local unit-or-package declarations

      while (l != null)
      {
        kind = l->e->kind;

        if (kind == A_UNIT_INTERFACE)
        {
          e = find_tree (l->e->info->the_unit_interface.declarations->identifier_btree, identifier);
          if (e != null)
          {
            count++;
            result = e;
            l->e->info->the_unit_interface.unit_is_used = true;
          }

          insert_each_entity_of_list_to_from_unit_list
               (l->e->info->the_unit_interface.declarations->packages,
                l->e,
                &list);
        }
        else if (kind == A_PACKAGE_DECLARATION)
        {
          e = find_tree (l->e->info->the_package_declaration.declarations->identifier_btree, identifier);
          if (e != null)
          {
            count++;
            result = e;
          }
          insert_each_entity_of_list_to_from_unit_list
               (l->e->info->the_package_declaration.declarations->packages,
                null,
                &list);
        }

        l = l->next;
      }

      r = r->next;
    }

    if (count == 1)
    {
      deallocate_entity_from_unit_list (&list);
      return result;
    }

    if (count > 1)
    {
      semantic_error ("identifier is ambiguous (was declared twice)", pos);
      deallocate_entity_from_unit_list (&list);
      return null;
    }



    // search in inner nested packages

    while (list != null)
    {
      ENTITY_FROM_UNIT_LIST *l;

      list2 = null;

      l = list;

      while (l != null)
      {
        kind = l->e->kind;

        if (kind == A_PACKAGE_DECLARATION)
        {
          e = find_tree (l->e->info->the_package_declaration.declarations->identifier_btree, identifier);
          if (e != null)
          {
            count++;
            result = e;
            if (l->unit_decl != null)
              l->unit_decl->info->the_unit_interface.unit_is_used = true;
          }
          insert_each_entity_of_list_to_from_unit_list
               (l->e->info->the_package_declaration.declarations->packages,
                l->unit_decl,
                &list2);
        }

        l = l->next;
      }

      deallocate_entity_from_unit_list (&list);

      list = list2;

      if (count == 1)
      {
        deallocate_entity_from_unit_list (&list);
        return result;
      }

      if (count > 1)
      {
        semantic_error ("identifier is ambiguous (was declared twice)", pos);
        deallocate_entity_from_unit_list (&list);
        return null;
      }
    }

    n = n->outer;
  }

  semantic_error ("identifier was not declared", pos);
  return null;
}

/******************************************************************************/

// does not produce error messages
// returns null if not found

PENTITY search_entity_in_local_scope (wstring identifier)
{
  PREGION r;
  PENTITY e;

  r = current->regions.first;

  while (r != null)
  {
    e = find_tree (r->identifier_btree, identifier);
    if (e != null)
      return e;

    r = r->next;
  }

  return null;
}

/******************************************************************************/

// check that the current search point is within the package declaration or package body

bool are_we_in_package (PENTITY epackage)
{
  LEVEL_NODE *ln;
  PREGION    r;
  PENTITY    se;

  ln = current;
  for (;;)
  {
    ln = ln->outer;

    if (ln == null)
      return false;

    r = ln->regions.last;
    se = r->entities.last;

    if (se->kind == A_PACKAGE_BODY)
      se = se->info->the_package_body.to_package_declaration;

    if (se == epackage)    // found (we're inside)
      return true;
  }
}

/******************************************************************************/

// assert: current token is an identifier
// returns null if not found

PENTITY parse_expanded_name ()
{
  PENTITY p, e;

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("identifier expected", &token.pos);
    return null;
  }

  p = find_entity (token.info._identifier.value, &token.pos);

  if (p == null)     // error already given above
  {
    get_token ();    // skip identifier
    return null;
  }

  if (entity_exists_in_list (p, &hidden_entities))
  {
    semantic_error ("identifier is not fully declared", &token.pos);
    get_token ();    // skip identifier
    return null;
  }

  get_token ();    // skip identifier

  if (p->kind == A_UNIT_INTERFACE && token.kind == DOT)
  {
    p->info->the_unit_interface.unit_is_used = true;

    get_token ();   // skip DOT

    if (token.kind != IDENTIFIER)
    {
      syntax_error ("identifier expected", &token.pos);
      return p;
    }

    e = find_tree (p->info->the_unit_interface.declarations->identifier_btree, token.info._identifier.value);
    if (e == null)
    {
      syntax_error ("identifier was not declared in unit", &token.pos);
      get_token ();    // skip identifier
      return p;
    }

    p = e;

    if (entity_exists_in_list (p, &hidden_entities))
    {
      semantic_error ("identifier is not fully declared", &token.pos);
      get_token ();    // skip identifier
      return null;
    }

    get_token ();    // skip identifier
  }


  while (p->kind == A_PACKAGE_DECLARATION && token.kind == DOT)
  {
    if (p->info->the_package_declaration.is_generic)
    {
      if (!are_we_in_package (p))   // we're outside the generic package
        break;
    }

    get_token ();   // skip DOT

    if (token.kind != IDENTIFIER)
    {
      syntax_error ("identifier expected", &token.pos);
      return p;
    }

    e = find_tree (p->info->the_package_declaration.declarations->identifier_btree, token.info._identifier.value);
    if (e == null)
    {
      syntax_error ("identifier was not declared in package", &token.pos);
      get_token ();    // skip identifier
      return p;
    }

    p = e;

    if (entity_exists_in_list (p, &hidden_entities))
    {
      semantic_error ("identifier is not fully declared", &token.pos);
      get_token ();    // skip identifier
      return null;
    }

    get_token ();    // skip identifier
  }

  return p;
}

/******************************************************************************/

void hide_entity (PENTITY e)
{
  insert_entity_in_list (e, &hidden_entities);
}

/******************************************************************************/

void unhide_all_entities ()
{
  deallocate_entity_list (&hidden_entities);
}

/******************************************************************************/

// insert package declaration in package list of surrounding scope

void append_package_declaration_to_surrounding_scope_package_list (PENTITY e)
{
  LEVEL_NODE *outer;
  PREGION    r;

  outer = current->outer;
  r = outer->regions.last;

  insert_entity_in_list (e, &r->packages);
}

/******************************************************************************/

ENTITY_LIST *build_list_of_surrounding_generic_packages ()
{
  ENTITY_LIST *l;
  LEVEL_NODE  *lev;
  PREGION     r;
  PENTITY     e, edecl;

  l = null;

  lev = current->outer;
  while (lev != null)
  {
    r = lev->regions.last;
    e = r->entities.last;

    if (e->kind == A_PACKAGE_DECLARATION)
    {
      if (e->info->the_package_declaration.is_generic)
      {
        insert_entity_in_list (e, &l);
      }
    }
      
    if (e->kind == A_PACKAGE_BODY)
    {
      edecl = e->info->the_package_body.to_package_declaration;
      if (edecl->info->the_package_declaration.is_generic)
      {
        insert_entity_in_list (edecl, &l);
      }
    }

    lev = lev->outer;
  }

  return l;
}

/******************************************************************************/

ENTITY_LIST *duplicate_entity_list (ENTITY_LIST *list)
{
  ENTITY_LIST *a, *b, *n, *l;

  a = list;
  b = null;
  l = null;

  while (a != null)
  {
    n = malloc (sizeof(ENTITY_LIST));
    if (n == null)
      fatal_out_of_memory_error ("dup_entity_list()");
    n->e    = a->e;
    n->next = null;

    if (l == null)
      b = n;
    else
      l->next = n;
    l = n;

    a = a->next;
  }

  return b;
}

/******************************************************************************/

LIST_OF_EXPRESSIONS *duplicate_list_of_expressions (LIST_OF_EXPRESSIONS *list)
{
  LIST_OF_EXPRESSIONS *a, *b, *n, *l;

  a = list;
  b = null;
  l = null;

  while (a != null)
  {
    n = malloc (sizeof(LIST_OF_EXPRESSIONS));
    if (n == null)
      fatal_out_of_memory_error ("dup_exp_list()");

    n->prev = l;
    n->next = null;
    n->exp  = a->exp;
    n->type = a->type;
    n->e    = a->e;

    if (l == null)
      b = n;
    else
      l->next = n;
    l = n;

    a = a->next;
  }

  return b;
}

/******************************************************************************/

CASE_CONSTANT *duplicate_case_constant_list (CASE_CONSTANT *list)
{
  CASE_CONSTANT *a, *b, *n, *l;

  a = list;
  b = null;
  l = null;

  while (a != null)
  {
    n = malloc (sizeof(CASE_CONSTANT));
    if (n == null)
      fatal_out_of_memory_error ("dup_case_constant()");
    n->value = a->value;
    n->next  = null;

    if (l == null)
      b = n;
    else
      l->next = n;
    l = n;

    a = a->next;
  }

  return b;
}

/******************************************************************************/

void store_location (LOCATION *l)
{
  l->unit_key    = g_unit_key;
  l->source_line = token.pos.line;
}

/******************************************************************************/
