
// entities.h : central structure tree.

#include "pool.h"

typedef struct entity ENTITY, *PENTITY;
typedef struct region REGION, *PREGION;
typedef struct expression EXPRESSION, *PEXPRESSION;


typedef struct {
  int unit_key;   // 0 means undefined location
  int source_line;
} LOCATION;


// mode

typedef enum {MODE_IN, MODE_OUT, MODE_REF} MODE;


// constraint

typedef enum {
  DOES_NOT_APPLY,        // (must be first in enum list) neither open-array nor open-struct type
  UNCONSTRAINED,         // open-array or open-struct without constraint
  CONSTANT_CONSTRAINT,   // constant constraint (value is given)
  RUNTIME_CONSTRAINT,    // runtime constraint
} KIND_CONSTRAINT;

typedef struct {
  KIND_CONSTRAINT  kind;
  uint4            value;  // if kind == CONSTANT_CONSTRAINT : length or discriminant constraint
} CONSTRAINT;


// context - used for array, struct and pointer types as context of aggregates or allocators.

typedef struct {
  PENTITY     base_type_or_null;   // null if no context available
  CONSTRAINT  constraint;
  bool        address_of;          // true if we want to take the expression's address
} CONTEXT, *PCONTEXT;


// comparison

typedef enum {
  CMP_FILLER,
  CMP_SMALLER,           // 1
  CMP_EQUAL,             // 2
  CMP_SMALLER_OR_EQUAL,  // 3
  CMP_LARGER,            // 4
  CMP_NOT_EQUAL,         // 5
  CMP_LARGER_OR_EQUAL,   // 6
} COMPARISON_FLAG;


// expression

typedef enum {

  // 3 params
  OP_CONDITIONAL_TEST,  // "? :"
                        // arg 1 = bool
                        // arg 2 & 3 = pair of
                        //   { integer,
                        //     float,
                        //     enum,
                        //     pointer (or null_pointer),
                        //     function pointer(or null_pointer),
                        //     unsafe pointer(or null_pointer),
                        //     non-open struct,
                        //     union,
                        //     open array (see constant/runtime constraint),
                        //     open struct (see constant/runtime constraint),
                        //     opaque,
                        //     generic }
                        // result = same as arg2

  // 2 params
  OP_SHORT_CIRCUIT_AND, // bool && bool -> bool
  OP_SHORT_CIRCUIT_OR,  // bool || bool -> bool

  OP_AND,               // bool & bool -> bool
  OP_OR,                // bool | bool -> bool
  OP_XOR,               // bool ^ bool -> bool

  OP_BITAND,            // integer & integer -> integer
  OP_BITOR,             // integer | integer -> integer
  OP_BITXOR,            // integer ^ integer -> integer

  // (uses mask 'cmp'), result is bool.
  OP_COMPARE_SIGNED,     // args = pair of signed integer
  OP_COMPARE_UNSIGNED,   // args = pair of enum or unsigned integer
  OP_COMPARE_FLOAT,      // args = pair of floating-point
  OP_COMPARE_PTR,        // args = pair of pointer, function pointer, unsafe pointer, or null pointer

  OP_SHIFT_LEFT_SIGNED,    // signed integer   << signed integer   -> signed integer
  OP_SHIFT_LEFT_UNSIGNED,  // unsigned integer << unsigned integer -> unsigned integer
  OP_SHIFT_RIGHT_SIGNED,   // signed integer   >> signed integer   -> signed integer
  OP_SHIFT_RIGHT_UNSIGNED, // unsigned integer >> unsigned integer -> unsigned integer

  OP_ADD_INT,              // integer + integer -> integer,  enum + integer -> enum
  OP_SUB_INT,              // integer - integer -> integer,  enum - integer -> enum

  OP_MUL_INT_SIGNED,       // signed integer * signed integer  ->  signed integer (promoted integer)
  OP_DIV_INT_SIGNED,       // same
  OP_MOD_INT_SIGNED,       // same
 
  OP_MUL_INT_UNSIGNED,     // unsigned integer * unsigned integer  ->  unsigned integer (promoted integer)
  OP_DIV_INT_UNSIGNED,     // same
  OP_MOD_INT_UNSIGNED,     // same

  OP_ADD_FLOAT,     // floating-point + floating_point -> floating_point
  OP_SUB_FLOAT,     // same
  OP_MUL_FLOAT,     // same
  OP_DIV_FLOAT,     // same

  OP_ADD_PTR_INT,   // unsafe_ptr + integer    -> unsafe_ptr  ; add integer*sizeof(elem) to unsafe_ptr
  OP_SUB_PTR_INT,   // unsafe_ptr - integer    -> unsafe_ptr  ; sub integer*sizeof(elem) from unsafe_ptr
  OP_SUB_PTR_PTR,   // unsafe_ptr - unsafe_ptr -> uint4       ; (ptr-ptr)/sizeof(elem) (type uint4)

  // 1 param
  OP_CONVERT_INT_INT,     // convert between various enumeration and integer types
  OP_CONVERT_PTR_PTR,     // convert between various unsafe pointers
  OP_CONVERT_FLOAT_INT,   // convert from floating-point to integer
  OP_CONVERT_INT_FLOAT,   // convert from integer to floating-point
  OP_CONVERT_FLOAT_FLOAT, // convert from floating-point to floating-point

  OP_UNARY_PLUS,   // integer -> integer,  or floating-point -> floating_point  (no effect other than promotion)
  OP_UNARY_MINUS,  // integer -> integer,  or floating-point -> floating_point

  OP_BOOL_NOT,     // bool -> bool
  OP_INT_NOT,      // integer -> integer

  OP_PRE_DEC,      // arg = object of type enum, int or unsafe-ptr, returns = value of that type
  OP_PRE_INC,

  OP_POST_DEC,     // arg = object of type enum, int or unsafe-ptr, returns = value of that type
  OP_POST_INC,

  OP_ADDRESS_OF,   // arg = an object, returns = unsafe pointer.
  OP_LENGTH,       // arg = an object of type array and runtime constraint, returns = type int.
  OP_SIZE,         // arg = an object of open type having runtime constraint, returns = type uint.

  OP_NONE,         // used internally only
} KIND_OPERATOR;

typedef enum {

  // constant value or constant object
  A_CONST_ENUMERATION_VALUE,  // ex: false, color'first.  Must be object for prefix of OP_ADDRESS_OF.
  A_CONST_INTEGER_VALUE,
  A_CONST_FLOAT_VALUE,
  A_CONST_NULL_VALUE,         // compatible with pointer, function pointer, unsafe pointer
  A_POOL_CONSTANT,            // array/struct string literal, constant aggregate,
                              // declared constant or index/slice/field of it.
  // runtime value
  AN_OPERATOR_VALUE,          // runtime value
  A_RUN_CALL,
  A_FUNCTION_VALUE,           // value of type function pointer
  A_FUNCTION_CALL,            // return type can be type_void
  A_DISCRIMINANT_VALUE,
  AN_UNC_ARRAY_AGGREGATE,     // array (context-dependant size)
  AN_AGGREGATE_VALUE,         // array or struct (constant size)
  A_QUALIFIED_EXPRESSION,
  AN_ARRAY_QUALIFIED_EXPRESSION,
  A_STRUCT_QUALIFIED_EXPRESSION,
  AN_ALLOCATOR,

  // runtime object (readonly or readwrite)
  A_GLOBAL_VARIABLE_OBJECT,
  A_LOCAL_VARIABLE_OBJECT,
  A_REFERENCE_OBJECT,
  A_PARAMETER_OBJECT,
  AN_ARRAY_ELEMENT_OBJECT,
  AN_ARRAY_SLICE_OBJECT,
  A_STRUCT_FIELD_OBJECT,
  A_DEREFERENCED_OBJECT,
  AN_UNSAFE_DEREFERENCED_OBJECT,
  AN_ATTR_BYTE_OBJECT,      // has either constant or runtime constraint

  // special conversion objects (can only appear as actual parameters, possibly inlined)
  A_BOXED_OBJECT,       // convert from any packed type into byte[]
  AN_UNBOXED_OBJECT,    // convert from byte[] into any packed type
  A_BOXED_ARRAY_OBJECT, // an array of boxed objects

} KIND_EXPRESSION;

/******************************************/

// used for code generation:
// the following fields are provided on input after pre-analysis
// to avoid that two different offsets are used by "?:" operands

typedef enum {_NO, _YES, _YES_BUT_CAN_BE_NULL} USES_TOMBSTONE;

typedef struct {
  USES_TOMBSTONE anchor_provided;
  int4           anchor_offset;     // offset to safe-ptr to tombstone ptr in local frame
} TOMBSTONE;

/******************************************/

typedef struct list_of_expressions {
  struct list_of_expressions *prev, *next;
  PEXPRESSION                exp;
  PENTITY                    type;  // for aggregates : target type, for parameters : actual base type before boxing
  PENTITY                    e;     // for struct aggregate: field or varying_field
                                    // for function call: (formal) parameter
  TOMBSTONE                  tomb;  // for code generation; stores heap object anchor
                                    //    of parameters and array objects.
} LIST_OF_EXPRESSIONS;

typedef struct {
  uint4 value;
} CONST_ENUMERATION_VALUE_INFO;

typedef struct {
  int8 value;
} CONST_INTEGER_VALUE_INFO;

typedef struct {
  double value;
} CONST_FLOAT_VALUE_INFO;

typedef struct {
  bool dummy;
} CONST_NULL_VALUE_INFO;

typedef struct {
  POOL *pool_cte;
} POOL_CONSTANT_INFO;

typedef struct {
  KIND_OPERATOR   op;
  COMPARISON_FLAG cmp;
  PEXPRESSION     arg[3];
  PENTITY         type;   // promoted type (for OP_COMPARE_SIGNED/_UNSIGNED/_FLOAT)
} OPERATOR_VALUE_INFO;

typedef struct {
  PEXPRESSION function_call;
} RUN_CALL_INFO;

typedef struct {
  PENTITY to_function_declaration_or_generic_function;   // generic function or function declaration entity
} FUNCTION_VALUE_INFO;

typedef struct {
  PEXPRESSION         func;   // a value of type function pointer
  LIST_OF_EXPRESSIONS *param;
} FUNCTION_CALL_INFO;

typedef struct {
  PEXPRESSION   prefix;      // has runtime constraint
  PENTITY       field;
} DISCRIMINANT_VALUE_INFO;

typedef struct {
  PEXPRESSION  element;
} UNC_ARRAY_AGGREGATE_INFO;  // array (context-dependant size)

typedef struct {
  LIST_OF_EXPRESSIONS *list;
} AGGREGATE_VALUE_INFO;      // array or struct (constant size)

typedef struct {
  PEXPRESSION  value;        // can be null if used inside an allocator
} QUALIFIED_EXPRESSION_INFO; // expression can be UNCONSTRAINED.

typedef struct {
  PEXPRESSION  length;       // runtime array length
  PEXPRESSION  value;        // can be null if used inside an allocator
} ARRAY_QUALIFIED_EXPRESSION_INFO;

typedef struct {
  PEXPRESSION  discriminant; // runtime discriminant value
  PEXPRESSION  value;        // can be null if used inside an allocator
} STRUCT_QUALIFIED_EXPRESSION_INFO;

typedef struct {
  PEXPRESSION  value;       // a qualified expression/array_qualified_expression/struct/qualified_expression
  bool         has_header;  // true = aligned header in heap object for length/discriminant
} ALLOCATOR_INFO;

typedef struct {
  PENTITY  pobject;
} GLOBAL_VARIABLE_OBJECT_INFO;

typedef struct {
  PENTITY       pobject;
  TEXT_POSITION pos;
} LOCAL_VARIABLE_OBJECT_INFO;

typedef struct {
  PENTITY       pobject;
  TEXT_POSITION pos;
} REFERENCE_OBJECT_INFO;

typedef struct {
  PENTITY       pobject;
  TEXT_POSITION pos;
} PARAMETER_OBJECT_INFO;

typedef struct {
  PEXPRESSION   prefix;       // array object, or unsafe-ptr object-or-value
  PEXPRESSION   index;        // can have types int, uint, or enum (for attribute 'string).
} ARRAY_ELEMENT_OBJECT_INFO;

typedef struct {
  PEXPRESSION   prefix;       // array object, or unsafe-ptr object-or-value
  PEXPRESSION   index;
  PEXPRESSION   length;
} ARRAY_SLICE_OBJECT_INFO;

typedef struct {
  PEXPRESSION   prefix;       // has struct or union type
  PENTITY       field;        // field or varying_field
} STRUCT_FIELD_OBJECT_INFO;

typedef struct {
  PEXPRESSION   prefix;       // value of type pointer
} DEREFERENCED_OBJECT_INFO;

typedef struct {
  PEXPRESSION   unsafe_ptr_value;
} UNSAFE_DEREFERENCED_OBJECT_INFO;

typedef struct {
  PEXPRESSION   prefix;
} ATTR_BYTE_OBJECT_INFO;

typedef struct {
  PEXPRESSION   parameter;   // convert parameter of any packed type into byte[]
} BOXED_OBJECT_INFO;

typedef struct {
  PEXPRESSION   parameter;     // convert parameter of type byte[] into any packed type
  bool          to_open_array; // true = formal parameter is open array with constraint
} UNBOXED_OBJECT_INFO;

typedef struct {
  LIST_OF_EXPRESSIONS *list;
} BOXED_ARRAY_OBJECT_INFO;


typedef enum {
  AN_OBJECT,
  A_VALUE,
} FORM;

typedef enum {
  ACCESS_CONSTANT,    // constant value or constant object
  ACCESS_READONLY,    // runtime object or runtime value
  ACCESS_READWRITE,   // runtime object
} ACCESS;

struct expression
{
  PENTITY         base_type_or_null; // "null" after compilation error
  CONSTRAINT      constraint;
  FORM            form;         // AN_OBJECT / A_VALUE
  ACCESS          access;       // ACCESS_CONSTANT, ACCESS_READONLY, ACCESS_READWRITE
  KIND_EXPRESSION kind;

  union {
    CONST_ENUMERATION_VALUE_INFO     const_enumeration_value_info;
    CONST_INTEGER_VALUE_INFO         const_integer_value_info;
    CONST_FLOAT_VALUE_INFO           const_float_value_info;
    CONST_NULL_VALUE_INFO            const_null_value_info;
    POOL_CONSTANT_INFO               pool_constant_info;
    OPERATOR_VALUE_INFO              operator_value_info;
    RUN_CALL_INFO                    run_call_info;
    FUNCTION_VALUE_INFO              function_value_info;
    FUNCTION_CALL_INFO               function_call_info;
    DISCRIMINANT_VALUE_INFO          discriminant_value_info;
    UNC_ARRAY_AGGREGATE_INFO         unc_array_aggregate_info;
    AGGREGATE_VALUE_INFO             aggregate_value_info;
    QUALIFIED_EXPRESSION_INFO        qualified_expression_info;
    ARRAY_QUALIFIED_EXPRESSION_INFO  array_qualified_expression_info;
    STRUCT_QUALIFIED_EXPRESSION_INFO struct_qualified_expression_info;
    ALLOCATOR_INFO                   allocator_info;
    GLOBAL_VARIABLE_OBJECT_INFO      global_variable_object_info;
    LOCAL_VARIABLE_OBJECT_INFO       local_variable_object_info;
    REFERENCE_OBJECT_INFO            reference_object_info;
    PARAMETER_OBJECT_INFO            parameter_object_info;
    ARRAY_ELEMENT_OBJECT_INFO        array_element_object_info;
    ARRAY_SLICE_OBJECT_INFO          array_slice_object_info;
    STRUCT_FIELD_OBJECT_INFO         struct_field_object_info;
    DEREFERENCED_OBJECT_INFO         dereferenced_object_info;
    UNSAFE_DEREFERENCED_OBJECT_INFO  unsafe_dereferenced_object_info;
    ATTR_BYTE_OBJECT_INFO            attr_byte_object_info;
    BOXED_OBJECT_INFO                boxed_object_info;
    UNBOXED_OBJECT_INFO              unboxed_object_info;
    BOXED_ARRAY_OBJECT_INFO          boxed_array_object_info;
  } info;
};


// entities

typedef enum {

  //========================================================================
  // entities denoting a type must appear first
  //========================================================================

  AN_INTEGER_TYPE,
  A_FLOAT_TYPE,

  AN_ENUMERATION_TYPE,
  AN_OPEN_ARRAY_TYPE,
  AN_ARRAY_TYPE,
  A_STRUCT_TYPE,        // can be open, can be packed
  A_CONSTRAINED_STRUCT_TYPE,
  A_UNION_TYPE,
  A_POINTER_TYPE,
  A_FUNCTION_POINTER_TYPE,
  AN_UNSAFE_POINTER_TYPE,

  A_RENAMED_TYPE,
  AN_INCOMPLETE_TYPE,
  AN_OPAQUE_TYPE,
  A_GENERIC_TYPE,
  A_NULL_POINTER_TYPE,  // only used internally for 'null' literal expressions
  A_VOID_TYPE,          // only used internally for return values of function_calls

  //========================================================================
  LAST_ENTITY_DENOTING_A_TYPE,   // used as sentinel
  //========================================================================

  AN_INCOMPLETE_TYPE_COMPLETITION,    // used for generic instantiations
  AN_OPAQUE_TYPE_COMPLETITION,        // used for generic instantiations

  A_FIELD,              // of a struct or union type, can be a discriminant
  A_VARYING_FIELD,      // of an open struct, depends on a discriminant value

  AN_ENUMERATION_LITERAL,
  A_CONSTANT,
  A_GLOBAL_VARIABLE,
  A_LOCAL_VARIABLE,
  A_REFERENCE,
  A_PARAMETER,
  A_GENERIC_FUNCTION,
  A_FUNCTION_DECLARATION,
  A_FUNCTION_BODY,
  A_PACKAGE_DECLARATION,   // can be generic or not
  A_PACKAGE_BODY,
  A_UNIT_INTERFACE,  // .h
  A_UNIT_BODY,       // .c

  //========================================================================
  FIRST_ENTITY_DENOTING_A_STATEMENT,   // used as sentinel
  //========================================================================
  A_CLEAR_STATEMENT,
  AN_ASSIGNMENT_STATEMENT,
  A_PRE_OR_POSTFIX_STATEMENT,
  A_FUNCTION_CALL_STATEMENT,
  A_RETURN_STATEMENT,
  A_BREAK_STATEMENT,
  A_CONTINUE_STATEMENT,
  A_FREE_STATEMENT,
  AN_ABORT_STATEMENT,
  AN_ASSERT_STATEMENT,
  A_SLEEP_STATEMENT,
  A_CODE_STATEMENT,
  AN_UNUSED_STATEMENT,
  A_BLOCK_STATEMENT,
  AN_IF_STATEMENT,
  A_SWITCH_STATEMENT,
  A_WHILE_STATEMENT,
  A_FOR_STATEMENT,
  //========================================================================
  LAST_ENTITY_DENOTING_A_STATEMENT,   // used as sentinel
  //========================================================================

} ENTITY_KIND;


// ========================================================

typedef enum {a_int1, a_int2, a_int4, a_int8, a_uint1, a_uint2, a_uint4}
  INTEGER_TYPE;

typedef struct {
  INTEGER_TYPE typ;
  wstring      name;
  TOKEN_KIND   token1;
  TOKEN_KIND   token2;
  bool         is_signed;
  int          size;
  int8         min;
  int8         max;
  int8         mask;
} INTEGER_INFO;

#define MAX_INTEGER_TYPES  7
extern INTEGER_INFO integer_info[MAX_INTEGER_TYPES];

// ========================================================

typedef enum {a_float, a_double}
  FLOAT_TYPE;

typedef struct {
  FLOAT_TYPE typ;
  wstring    name;
  TOKEN_KIND token1;
  TOKEN_KIND token2;
  int        size;
} FLOAT_INFO;

#define MAX_FLOAT_TYPES  2
extern FLOAT_INFO float_info [MAX_FLOAT_TYPES];

// ========================================================

typedef struct {
  INTEGER_TYPE type;
} AN_INTEGER_TYPE_ENTITY;

typedef struct {
  FLOAT_TYPE type;
} A_FLOAT_TYPE_ENTITY;

typedef struct {
  INTEGER_TYPE base;           // must be an unsigned type
  uint4        last;           // must be <= base'last
  POOL         *string_table;  // string table of all enumeration literals
} AN_ENUMERATION_TYPE_ENTITY;

typedef struct {
  PENTITY  element;    // element type
} AN_OPEN_ARRAY_TYPE_ENTITY;

typedef struct {
  PENTITY  open_array;
  uint4    length;
} AN_ARRAY_TYPE_ENTITY;

typedef struct {
  bool    is_packed;
  bool    is_open_type;
  bool    is_unsafe;    // contains at least an unsafe pointer field
  PREGION discriminant; // if .is_open_type, region contains just one discriminant field
  PREGION fields;       // fields, varying fields, anonymous type entities
  POOL    *size_table;  // for open struct types : uint4 table of struct size per enum value;
                        // null means table was not created by code generator yet.
  int     header_size;  // access object header size for open struct (4 or 8 bytes),
                        // depends on largest alignment requirement of all fields.
} A_STRUCT_TYPE_ENTITY;

typedef struct {
  PENTITY   open_struct;
  uint4     discriminant_value;
} A_CONSTRAINED_STRUCT_TYPE_ENTITY;

typedef struct {
  bool    is_unsafe;    // contains at least an unsafe pointer field
  PREGION fields;       // fields, anonymous type entities
} A_UNION_TYPE_ENTITY;

typedef struct {
  PENTITY   designated_type;
} A_POINTER_TYPE_ENTITY;

typedef struct {
  bool    is_inline;
  bool    is_callback;          // -> means the function can be called by Windows
  bool    is_public;            // -> declared in .h or package decl.
  bool    is_thread_entry_point;// requires M16 stack alignment at entry
  wstring extern_dll_or_null;   // -> means there's no body
  PENTITY return_type;          // can be type_void
  PREGION parameters;           // contains parameter entities or anonymous types
} A_FUNCTION_POINTER_TYPE_ENTITY;

typedef struct {
  PENTITY   designated_type;
} AN_UNSAFE_POINTER_TYPE_ENTITY;

typedef struct {
  PENTITY   actual_type;
} A_RENAMED_TYPE_ENTITY;

typedef struct {
  PENTITY   full_type;              // forward reference
  bool      is_full_type_visible;
} AN_INCOMPLETE_TYPE_ENTITY;

typedef struct {
  PENTITY   full_type;              // forward reference
  bool      is_full_type_visible;
  bool      is_limited;
} AN_OPAQUE_TYPE_ENTITY;

typedef struct {
  PENTITY   actual_type;   // used during instantiation and in instantiated package
} A_GENERIC_TYPE_ENTITY;

typedef struct {
  bool dummy;
} A_NULL_POINTER_TYPE_ENTITY;

typedef struct {
  bool dummy;
} A_VOID_TYPE_ENTITY;

//========================================================================

typedef struct {
  PENTITY   incomplete_type;
  PENTITY   full_type;
} AN_INCOMPLETE_TYPE_COMPLETITION_ENTITY;

typedef struct {
  PENTITY   opaque_type;
  PENTITY   full_type;
} AN_OPAQUE_TYPE_COMPLETITION_ENTITY;

typedef struct {      // of a struct or union type, can be a discriminant
  PENTITY   type;
} A_FIELD_ENTITY;

typedef struct {
  PENTITY   type;
  uint4     discriminant_value;   // field existance depends on discriminant
} A_VARYING_FIELD_ENTITY;

typedef struct {
  PENTITY   type;
  uint4     value;
} AN_ENUMERATION_LITERAL_ENTITY;

typedef struct {
  PENTITY       type;
  PEXPRESSION   value;
  bool          is_used;   // is used in exp.c
  TEXT_POSITION pos;
} A_CONSTANT_ENTITY;

typedef struct {
  PENTITY       type;
  PEXPRESSION   initial_value_or_null;   // a constant initial value
  bool          is_volatile;
  int8          offset;
} A_GLOBAL_VARIABLE_ENTITY;

typedef struct {
  PENTITY       type;
  PEXPRESSION   initial_value_or_null;
  bool          is_read;
  bool          is_written;
  bool          is_used;      // in any primary or if it has initial value
  TEXT_POSITION pos;
  LOCATION      loc;
  int4          offset;       // always negative
  bool          allocated_on_heap;
} A_LOCAL_VARIABLE_ENTITY;

typedef struct {
  PENTITY       type;
  MODE          mode;         // MODE_IN or MODE_REF
  PEXPRESSION   name;
  bool          is_used;      // is read or used in attribute prefix
  bool          references_heap_object;   // requires a tombstone (code generator)
  TEXT_POSITION pos;
  LOCATION      loc;
  int4          offset;       // always negative
} A_REFERENCE_ENTITY;

typedef struct {
  PENTITY       type;
  MODE          mode;
  PEXPRESSION   default_value_or_null;   // an optional constant initial value
  bool          is_used;      // is read or used in attribute
  bool          init_error;   // used to avoid giving a non-initialized error twice
  TEXT_POSITION pos;          // position of parameter in function body
  int4          offset;       // address offset[EBP] (always positive)
} A_PARAMETER_ENTITY;

typedef struct {
  PENTITY   to_type;      // to A_FUNCTION_POINTER_TYPE
  PENTITY   actual_func;  // used during instantiation and in instantiated package (to generic function/function decl)
} A_GENERIC_FUNCTION_ENTITY;

typedef struct {
  PENTITY   to_type;                    // to A_FUNCTION_POINTER_TYPE
  PENTITY   to_function_body_or_null;   // forward reference
  bool      code_generated;             // true=code was generated for this function
  int4      frame_size;                 // always negative (to be subtracted from EBP)
  int4      func_label_nr;              // function label nr (0 = none)
} A_FUNCTION_DECLARATION_ENTITY;

typedef struct {
  PENTITY   to_function_declaration;
  PREGION   inner;          // contains declarations and statements
  LOCATION  begin_loc;
  LOCATION  end_loc;
  bool      contains_code_statements;
} A_FUNCTION_BODY_ENTITY;   // this entity has no identifier

typedef struct {
  PENTITY   to_package_body_or_null;// forward reference
  bool      is_generic;
  bool      is_body_required;
  PREGION   generic_part;   // contains generic type, generic function, anonymous types or constants
  PREGION   declarations;
} A_PACKAGE_DECLARATION_ENTITY;

typedef struct {
  PENTITY   to_package_declaration;
  PREGION   declarations;
} A_PACKAGE_BODY_ENTITY;    // this entity has no identifier

typedef struct {
  PENTITY   to_unit_body;   // forward reference
  bool      is_body_required;
  bool      unit_is_used;   // is referenced at least once
  PREGION   declarations;
} A_UNIT_INTERFACE_ENTITY;

typedef struct {
  PENTITY   to_unit_interface;
  PREGION   declarations;
} A_UNIT_BODY_ENTITY;      // this entity has no identifier

//========================================================================

typedef struct {
  PEXPRESSION name;    // object to clear
  LOCATION    loc;
} A_CLEAR_STATEMENT_ENTITY;

typedef enum {
  _ASSIGN,

  _ASSIGN_ADD_INT,            // for integer, enum
  _ASSIGN_SUB_INT,

  _ASSIGN_ADD_PTR_INT,        // for ptr + int -> ptr+int*size(elem)
  _ASSIGN_SUB_PTR_INT,        // for ptr - int -> ptr-int*size(elem) (type ptr)

  _ASSIGN_MULT_INT_SIGNED,    // for signed integer
  _ASSIGN_DIV_INT_SIGNED,
  _ASSIGN_MOD_INT_SIGNED,

  _ASSIGN_MULT_INT_UNSIGNED,  // for unsigned integer
  _ASSIGN_DIV_INT_UNSIGNED,
  _ASSIGN_MOD_INT_UNSIGNED,

  _ASSIGN_ADD_FLOAT,          // for floating-point
  _ASSIGN_SUB_FLOAT,
  _ASSIGN_MULT_FLOAT,
  _ASSIGN_DIV_FLOAT,

  _ASSIGN_SHIFT_LEFT_SIGNED,      // for integer
  _ASSIGN_SHIFT_LEFT_UNSIGNED,

  _ASSIGN_SHIFT_RIGHT_SIGNED,     // for integer
  _ASSIGN_SHIFT_RIGHT_UNSIGNED,

  _ASSIGN_AND,       // for bool
  _ASSIGN_OR,
  _ASSIGN_XOR,

  _ASSIGN_BITAND,    // for integer
  _ASSIGN_BITOR,
  _ASSIGN_BITXOR,

} ASSIGNMENT_OP;

typedef struct {
  PEXPRESSION    name;
  ASSIGNMENT_OP  op;
  PEXPRESSION    value;
  LOCATION       loc;
} AN_ASSIGNMENT_STATEMENT_ENTITY;

typedef struct {
  PEXPRESSION  name;
  LOCATION     loc;
} A_PRE_OR_POSTFIX_STATEMENT_ENTITY;

typedef struct {
  PEXPRESSION  name;
  LOCATION     loc;
} A_FUNCTION_CALL_STATEMENT_ENTITY;

typedef struct {
  PEXPRESSION   value;   // null if no return value
  PENTITY       type;
  PENTITY       outer;   // surrounding function body, or block/if/switch/while/for statement.
  TEXT_POSITION pos;
  LOCATION      loc;
} A_RETURN_STATEMENT_ENTITY;

typedef struct {
  PENTITY       outer;   // surrounding function body, or block/if/switch/while/for statement.
  TEXT_POSITION pos;
} A_BREAK_STATEMENT_ENTITY;

typedef struct {
  PENTITY       outer;   // surrounding function body, or block/if/switch/while/for statement.
  TEXT_POSITION pos;
} A_CONTINUE_STATEMENT_ENTITY;

typedef struct {
  PEXPRESSION  value;
  LOCATION     loc;
} A_FREE_STATEMENT_ENTITY;

typedef struct {
  LOCATION      loc;
  TEXT_POSITION pos;
} AN_ABORT_STATEMENT_ENTITY;

typedef struct {
  PEXPRESSION  value;
  LOCATION     loc;
} AN_ASSERT_STATEMENT_ENTITY;

typedef struct {
  PEXPRESSION  value;
  LOCATION     loc;
} A_SLEEP_STATEMENT_ENTITY;

typedef struct {
  LOCATION     loc;
  byte         *code;
  int          nb_bytes;
} A_CODE_STATEMENT_ENTITY;

typedef struct {
  TEXT_POSITION pos;
  PENTITY       item;
} AN_UNUSED_STATEMENT_ENTITY;

typedef struct {
  PREGION  inner;    // contain declarations and statements
  PENTITY  outer;    // surrounding function body, or block/if/switch/while/for statement.
} A_BLOCK_STATEMENT_ENTITY;

typedef struct {
  PEXPRESSION  condition;
  PREGION      true_branch;
  PREGION      false_branch;
  PENTITY      outer;    // surrounding function body, or block/if/switch/while/for statement.
  LOCATION     loc;
} AN_IF_STATEMENT_ENTITY;

typedef struct case_constant {
  int8                 value;
  struct case_constant *next;
} CASE_CONSTANT;

typedef struct flow_alternative {
  bool          is_default;    // true = last default case
  CASE_CONSTANT *cte_list;
  PREGION       inner;
  struct flow_alternative *next;
  int           label_nr_drop;      // for code generator
  int           label_nr_no_drop;
  bool          code_generated;
} FLOW_ALTERNATIVE;

typedef struct {
  PEXPRESSION      value;
  uint4            count;     // nb case constants, not counting the last default case.
  FLOW_ALTERNATIVE *first_alt;
  FLOW_ALTERNATIVE *last_alt;
  PENTITY          outer;     // surrounding function body, or block/if/switch/while/for statement.
  LOCATION         loc;
} A_SWITCH_STATEMENT_ENTITY;

typedef struct {
  PEXPRESSION   condition;
  PREGION       inner;
  PENTITY       outer;    // surrounding function body, or block/if/switch/while/for statement.
  TEXT_POSITION pos;      // while token
  LOCATION      loc;      // of condition
} A_WHILE_STATEMENT_ENTITY;

typedef struct {
  PREGION      pre;
  PREGION      exp_region;  // exp_region must be elaborated BEFORE evaluating condition !
  PEXPRESSION  condition;   // null means true
  PREGION      post;
  PREGION      inner;
  PENTITY      outer;    // surrounding function body, or block/if/switch/while/for statement.
  TEXT_POSITION pos;     // for token
  LOCATION     loc;      // of condition
} A_FOR_STATEMENT_ENTITY;

/******************************************************************************/

typedef union {
  AN_INTEGER_TYPE_ENTITY            the_integer_type;
  A_FLOAT_TYPE_ENTITY               the_float_type;
  AN_ENUMERATION_TYPE_ENTITY        the_enumeration_type;
  AN_OPEN_ARRAY_TYPE_ENTITY         the_open_array_type;
  AN_ARRAY_TYPE_ENTITY              the_array_type;
  A_STRUCT_TYPE_ENTITY              the_struct_type;
  A_CONSTRAINED_STRUCT_TYPE_ENTITY  the_constrained_struct_type;
  A_UNION_TYPE_ENTITY               the_union_type;
  A_POINTER_TYPE_ENTITY             the_pointer_type;
  A_FUNCTION_POINTER_TYPE_ENTITY    the_function_pointer_type;
  AN_UNSAFE_POINTER_TYPE_ENTITY     the_unsafe_pointer_type;
  A_RENAMED_TYPE_ENTITY             the_renamed_type;
  AN_INCOMPLETE_TYPE_ENTITY         the_incomplete_type;
  AN_OPAQUE_TYPE_ENTITY             the_opaque_type;
  A_GENERIC_TYPE_ENTITY             the_generic_type;
  A_NULL_POINTER_TYPE_ENTITY        the_null_pointer_type;
  A_VOID_TYPE_ENTITY                the_void_type;
  AN_INCOMPLETE_TYPE_COMPLETITION_ENTITY the_incomplete_type_completition;
  AN_OPAQUE_TYPE_COMPLETITION_ENTITY     the_opaque_type_completition;
  A_FIELD_ENTITY                    the_field;
  A_VARYING_FIELD_ENTITY            the_varying_field;
  AN_ENUMERATION_LITERAL_ENTITY     the_enumeration_literal;
  A_CONSTANT_ENTITY                 the_constant;
  A_GLOBAL_VARIABLE_ENTITY          the_global_variable;
  A_LOCAL_VARIABLE_ENTITY           the_local_variable;
  A_REFERENCE_ENTITY                the_reference;
  A_PARAMETER_ENTITY                the_parameter;
  A_GENERIC_FUNCTION_ENTITY         the_generic_function;
  A_FUNCTION_DECLARATION_ENTITY     the_function_declaration;
  A_FUNCTION_BODY_ENTITY            the_function_body;
  A_PACKAGE_DECLARATION_ENTITY      the_package_declaration;
  A_PACKAGE_BODY_ENTITY             the_package_body;
  A_UNIT_INTERFACE_ENTITY           the_unit_interface;
  A_UNIT_BODY_ENTITY                the_unit_body;
  A_CLEAR_STATEMENT_ENTITY          the_clear_statement;
  AN_ASSIGNMENT_STATEMENT_ENTITY    the_assignment_statement;
  A_PRE_OR_POSTFIX_STATEMENT_ENTITY the_pre_or_postfix_statement;
  A_FUNCTION_CALL_STATEMENT_ENTITY  the_function_call_statement;
  A_RETURN_STATEMENT_ENTITY         the_return_statement;
  A_BREAK_STATEMENT_ENTITY          the_break_statement;
  A_CONTINUE_STATEMENT_ENTITY       the_continue_statement;
  A_FREE_STATEMENT_ENTITY           the_free_statement;
  AN_ABORT_STATEMENT_ENTITY         the_abort_statement;
  AN_ASSERT_STATEMENT_ENTITY        the_assert_statement;
  A_SLEEP_STATEMENT_ENTITY          the_sleep_statement;
  A_CODE_STATEMENT_ENTITY           the_code_statement;
  AN_UNUSED_STATEMENT_ENTITY        the_unused_statement;
  A_BLOCK_STATEMENT_ENTITY          the_block_statement;
  AN_IF_STATEMENT_ENTITY            the_if_statement;
  A_SWITCH_STATEMENT_ENTITY         the_switch_statement;
  A_WHILE_STATEMENT_ENTITY          the_while_statement;
  A_FOR_STATEMENT_ENTITY            the_for_statement;
} ENTITY_INFO;

/******************************************************************************/

struct entity
{
  ENTITY_KIND   kind;
  ENTITY_INFO   *info;
  PENTITY       next;                  // next entity in declaration queue
  long          nr;                    // unique entity number (for import/export/pcodes)
  wstring       identifier_or_null;    // allocated on heap
  PENTITY       ptr[2];                // pointers to left/right identifier btrees
  tiny          bal;                   // -1, 0 or +1 (balancing coefficient for btree node)
};

/******************************************************************************/

typedef struct {
  PENTITY  first, last;
} DECLARATION_QUEUE;

/******************************************************************************/

// for hiding not fully declared entities (ex: object, reference, function, parameters)
// or for providing a list of inner packages.

typedef struct entity_list {
  PENTITY            e;
  struct entity_list *next;
} ENTITY_LIST;

void insert_entity_in_list (PENTITY e, ENTITY_LIST **root);

struct region {
  DECLARATION_QUEUE entities;           // entities in declaration order
  PENTITY           identifier_btree;
  ENTITY_LIST       *packages;
  PREGION           next;
};

/******************************************************************************/

void create_predefined_entities ();

extern PENTITY  type_bool, type_char, type_wchar, type_byte, type_int, type_uint, type_long;
extern PENTITY  type_float, type_double;
extern PENTITY  type_string, type_wstring, type_object, type_array_of_string, type_array_of_object;
extern PENTITY  type_int_literal, type_float_literal, type_null_literal, type_unsafe_ptr_to_byte;
extern PENTITY  type_void;
extern PENTITY  enum_false, enum_true;

extern PENTITY global_locked_struct_entity;  // locked during struct declaration

/******************************************************************************/

// returns the type entity corresponding to the token,
// or null if the token does not denote a type.

PENTITY type_entity_of_token (TOKEN_KIND kind);

/******************************************************************************/

void create_region_level ();
void unlink_region_level ();

PREGION new_region ();
void append_region (PREGION r);

void cleanup_main_region_after_compilation ();

PENTITY append_new_entity (ENTITY_KIND   kind,
                           void          *entity_info,
                           wstring       identifier,  // can be empty string
                           TEXT_POSITION *pos);       // for error if declared twice

/******************************************************************************/

// create new entity and insert it in symbol tree,
// but does not append it in elaboration queue.

PENTITY create_new_entity (ENTITY_KIND   kind,
                           void          *entity_info,
                           wstring       identifier,   // can be empty string
                           TEXT_POSITION *pos);

/******************************************************************************/

void append_entity_in_entity_queue (PENTITY e);

/******************************************************************************/

void patch_entity_by_setting_an_identifier (PENTITY       e,
                                            wstring       identifier,   // can be empty string
                                            TEXT_POSITION *pos);

/******************************************************************************/

PENTITY append_new_entity_for_instantiation
                          (ENTITY_KIND   kind,
                           void          *entity_info,
                           wstring       identifier,   // can be empty string
                           TEXT_POSITION *pos,
                           REGION        *region);

/******************************************************************************/

void append_new_unit_entity_for_import
                      (PENTITY       e,
                       wstring       identifier,   // can be empty string
                       TEXT_POSITION *pos,
                       bool          add_in_package_list);

/******************************************************************************/

// produces error messages if not found or ambiguous
// returns null if not found

PENTITY find_entity (wstring identifier, TEXT_POSITION *pos);

/******************************************************************************/

// does not produce error messages
// returns null if not found

PENTITY search_entity_in_local_scope (wstring identifier);

/******************************************************************************/

// returns null if not found

PENTITY find_tree (PENTITY tree, wstring identifier);

/******************************************************************************/

// returns entity surrounding current scope (unit, package, ..)

PENTITY surrounding_entity ();

/******************************************************************************/

// check that the current search point is within the package declaration or package body

bool are_we_in_package (PENTITY epackage);

/******************************************************************************/

// assert: current token is an identifier
// returns null if not found

PENTITY parse_expanded_name ();

/******************************************************************************/

void hide_entity (PENTITY e);
void unhide_all_entities ();

/******************************************************************************/

// insert package declaration in package list of surrounding scope
void append_package_declaration_to_surrounding_scope_package_list (PENTITY e);

/******************************************************************************/

ENTITY_LIST *build_list_of_surrounding_generic_packages ();

/******************************************************************************/

ENTITY_LIST         *duplicate_entity_list         (ENTITY_LIST *list);
LIST_OF_EXPRESSIONS *duplicate_list_of_expressions (LIST_OF_EXPRESSIONS *list);
CASE_CONSTANT       *duplicate_case_constant_list  (CASE_CONSTANT *list);

/******************************************************************************/

void store_location (LOCATION *l);

/******************************************************************************/
