
// heaptype.c

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "bintree.h"

#include "types.h"
#include "token.h"
#include "entities.h"
#include "type.h"
#include "error.h"

#include "heaptype.h"

/*****************************************************************************/

static BINARY_TREE heap_type_btree;
static int4        unique_type_serial_nr;

/*****************************************************************************/

// signature format :
//
// int8   entity nr
// uint4  AN_ARRAY_TYPE,
// uint4  A_CONSTRAINED_STRUCT_TYPE,
// [ (-1) AN_OPEN_ARRAY_TYPE,
// ^ (-2) A_POINTER_TYPE,
// * (-3) AN_UNSAFE_POINTER_TYPE,
//   (-4) A_FUNCTION_POINTER_TYPE
//        . callback(y/n)
//        . extern_dll(y/n),
//        . return type
//        . { (-5) A_PARAMETER
//                 . mode (in/out/ref)
//                 . type
//          }
//   (-6) END_FUNCTION

/*****************************************************************************/

static int signature_length (PENTITY subtype)
{
  PENTITY     e;
  ENTITY_KIND kind;
  int         count;

  e = complete_type_of (subtype);
  kind = e->kind;

  if (kind == AN_INTEGER_TYPE || kind == A_FLOAT_TYPE || kind == AN_ENUMERATION_TYPE ||
      kind == A_STRUCT_TYPE   || kind == A_UNION_TYPE || kind == A_VOID_TYPE)
  {
    return 1;
  }

  if (kind == AN_ARRAY_TYPE)
    return 1 + signature_length (e->info->the_array_type.open_array);

  if (kind == A_CONSTRAINED_STRUCT_TYPE)
    return 1 + signature_length (e->info->the_constrained_struct_type.open_struct);

  if (kind == AN_OPEN_ARRAY_TYPE)
    return 1 + signature_length (e->info->the_open_array_type.element);

  if (kind == A_POINTER_TYPE)
    return 1 + signature_length (e->info->the_pointer_type.designated_type);

  if (kind == AN_UNSAFE_POINTER_TYPE)
    return 1 + signature_length (e->info->the_unsafe_pointer_type.designated_type);

  if (kind != A_FUNCTION_POINTER_TYPE)
    fatal_compiler_error0 ("signature_length(1)");

  count = 1 + 1 + 1     // -4 + callback(y/n) + extern_dll(y/n) + return_type.
        + signature_length(e->info->the_function_pointer_type.return_type);

  e = e->info->the_function_pointer_type.parameters->entities.first;
  while (e != null)
  {
    if (e->kind == A_PARAMETER)
    {
      // -5(PAR), mode, type
      count += (1 + 1 + signature_length(e->info->the_parameter.type));
    }
    e = e->next;
  }

  count++;     // -6 (END_FUNC)

  return count;
}

/*****************************************************************************/

static int fill_signature (PENTITY subtype, int8 signature[])
{
  PENTITY     e;
  ENTITY_KIND kind;
  int         index;

  e = complete_type_of (subtype);
  kind = e->kind;

  if (kind == AN_INTEGER_TYPE || kind == A_FLOAT_TYPE || kind == AN_ENUMERATION_TYPE ||
      kind == A_STRUCT_TYPE   || kind == A_UNION_TYPE || kind == A_VOID_TYPE)
  {
    signature[0] = e->nr;
    return 1;
  }

  if (kind == AN_ARRAY_TYPE)
  {
    signature[0] = e->info->the_array_type.length;
    return 1 + fill_signature (e->info->the_array_type.open_array, &signature[1]);
  }

  if (kind == A_CONSTRAINED_STRUCT_TYPE)
  {
    signature[0] = e->info->the_constrained_struct_type.discriminant_value;
    return 1 + fill_signature (e->info->the_constrained_struct_type.open_struct, &signature[1]);
  }

  if (kind == AN_OPEN_ARRAY_TYPE)
  {
    signature[0] = -1;
    return 1 + fill_signature (e->info->the_open_array_type.element, &signature[1]);
  }

  if (kind == A_POINTER_TYPE)
  {
    signature[0] = -2;
    return 1 + fill_signature (e->info->the_pointer_type.designated_type, &signature[1]);
  }

  if (kind == AN_UNSAFE_POINTER_TYPE)
  {
    signature[0] = -3;
    return 1 + fill_signature (e->info->the_unsafe_pointer_type.designated_type, &signature[1]);
  }

  if (kind != A_FUNCTION_POINTER_TYPE)
    fatal_compiler_error0 ("fill_signature(1)");

  signature[0] = -4;
  signature[1] = (int8)e->info->the_function_pointer_type.is_callback;
  signature[2] = (int8)(e->info->the_function_pointer_type.extern_dll_or_null != null);

  index = 3;

  index += fill_signature (e->info->the_function_pointer_type.return_type, &signature[index]);

  e = e->info->the_function_pointer_type.parameters->entities.first;
  while (e != null)
  {
    if (e->kind == A_PARAMETER)
    {
      signature[index++] = -5;  // PAR
      signature[index++] = (int8)e->info->the_parameter.mode;

      index += fill_signature (e->info->the_parameter.type, &signature[index]);
    }
    e = e->next;
  }

  signature[index++] = -6;  // END_FUNC

  return index;
}

/*****************************************************************************/

typedef struct {
  int4 unique_type;
  int8 *signature;
  int4 signature_length;
} NODE;

/*****************************************************************************/

static int compare (void               *user,
                    const NODE         *a,
                    const unsigned int dummy1,
                    const NODE         *b,
                    const unsigned int dummy2)
{
  int rc;

  rc = memcmp ((char *)a->signature, (char *)b->signature,
               sizeof(int8) * min(a->signature_length, b->signature_length));
  if (rc != 0)
    return rc;

  if (a->signature_length > b->signature_length)
    return +1;
  if (a->signature_length < b->signature_length)
    return -1;
  return 0;
}

/**********************************************************************************/

void init_heap_type_btree (void)
{
  create_btree (&heap_type_btree, NULL, compare);
}

/**********************************************************************************/

int heap_object_type_unique_nr (PENTITY subtype)
{
  int  signature_len;
  int8 *signature;
  NODE n;
  int  rc;
  unsigned int dummy;

  signature_len = signature_length (subtype);

  signature = malloc (sizeof(int8) * signature_len);
  if (signature == null)
    fatal_out_of_memory_error ("heap_object_type_unique_nr(1)");

  if (fill_signature (subtype, signature) != signature_len)
    fatal_out_of_memory_error ("heap_object_type_unique_nr(2)");

if (false)
{
  int i;
  printf ("signature is : ");
  for (i=0; i<signature_len; i++)
    printf (" %I64d", signature[i]);
  printf ("\n");
}

  n.signature        = signature;
  n.signature_length = signature_len;

  rc = retrieve_btree (&heap_type_btree, &n, sizeof(n), sizeof(n), &dummy, BT_EQUAL);
  if (rc == 0)
  {
    free (signature);
    return n.unique_type;
  }

  if (rc != BT_KEY_NOT_FOUND)
    fatal_compiler_error0 ("heap_object_type_unique_nr(3)");

  n.signature        = signature;
  n.signature_length = signature_len;
  n.unique_type      = ++unique_type_serial_nr;

  rc = insert_btree (&heap_type_btree, &n, sizeof(n));
  if (rc < 0)
    fatal_compiler_error0 ("heap_object_type_unique_nr(2)");

  return unique_type_serial_nr;
}

/*****************************************************************************/
