
/* bintree.c */

/************************************************************************/

#include <string.h>
#include <malloc.h>
#include <assert.h>
#include "bintree.h"

/************************************************************************/

#ifdef unix
  char *alloca (unsigned);
#endif

/************************************************************************/

#define MAGIC_OPEN   0x7403A0B3
#define MAGIC_VALID  0x34A67246

/************************************************************************/

typedef struct {
  BLOCK_HANDLE ptr[2];   /* pointers to left/right nodes */
  BLOCK_HANDLE data;     /* pointer to user data         */
  short        bal;      /* -1, 0 or +1                  */
} NODE;

/************************************************************************/

#define handlecpy(target,source) memcpy(target,source,sizeof(BLOCK_HANDLE))

/************************************************************************/

void create_btree (BINARY_TREE     *btree,
                   void            *user,
                   COMPARE_ELEMENT *compare)
{
  assert (btree != (BINARY_TREE *)0);
  set_null_handle (&btree->root);
  btree->open    = MAGIC_OPEN;
  btree->valid   = MAGIC_VALID;
  btree->user    = user;
  btree->compare = compare;
}

/************************************************************************/

static int deallocate_tree (BINARY_TREE  *btree,
                            BLOCK_HANDLE *tree)
{
  NODE p;
  int  dummy;

  if (is_null_handle (tree))
    return 0;

  /* load node 'tree' into 'p' */
  if (retrieve_block (tree, (char *)&p, sizeof(NODE), &dummy) < 0)
  {
    btree->valid = 0;
    return BT_TREE_CORRUPTED;
  }

  if (deallocate_tree (btree, &p.ptr[0]) < 0 ||
      deallocate_tree (btree, &p.ptr[1]) < 0)
  {
    btree->valid = 0;
    return BT_TREE_CORRUPTED;
  }

  /* delete node data + the node itself */
  if (deallocate_block (&p.data) != 0 ||  /* will be set to null handle */
      deallocate_block (tree)    != 0)    /* will be set to null handle */
  {
    btree->valid = 0;
    return BT_TREE_CORRUPTED;
  }

  return 0;
}

/************************************************************************/

void close_btree (BINARY_TREE *btree)
{
  assert (btree != (BINARY_TREE *)0);
  assert (btree->open == MAGIC_OPEN);

  if (btree->valid == MAGIC_VALID)
    (void)deallocate_tree (btree, &btree->root);

  btree->open = 0;
}

/************************************************************************/

static int compare_data (BINARY_TREE  *btree,
                         void         *data,
                         unsigned int size,
                         BLOCK_HANDLE *handle_data,
                         int          *cmp)
{
  unsigned int size2;
  void         *p;
  int          dummy;

  /* get size of user data */
  size2 = (unsigned int)size_of_block (handle_data);

  /* reserve stack space to hold handle_data's data */
  p = (void *)alloca ((size_t)size2);
  if (p == (void *)0)
    return BT_OUT_OF_MEMORY;

  if (retrieve_block (handle_data, (char *)p, size2, &dummy) < 0)
    return BT_IO_ERROR;  /* cannot retrieve virtual memory block */

  *cmp = (*btree->compare)(btree->user, data, size, (char *)p, size2);

  return 0;
}

/************************************************************************/

static int insert_tree (BINARY_TREE  *btree,        /* main fields      */
                        BLOCK_HANDLE *tree,         /* insertion point  */
                        void         *data,         /* new data         */
                        unsigned int size,          /* new data size    */
                        short        *help,         /* 1=overflow, 0=ok */
                        short        *tree_changed) /* 1= *tree changed */
{
  NODE         p, p1, p2;
  int          rc, dummy, cmp, ins0, ins1;
  BLOCK_HANDLE handle_p1, handle_p2;
  short        bal1, bal2, local_tree_changed;

  if (is_null_handle (tree))
  {
    set_null_handle (&p.ptr[0]);
    set_null_handle (&p.ptr[1]);
    p.bal = 0;

    if (allocate_block (&p.data, data, size) < 0)  /* allocate user data */
      return BT_OUT_OF_MEMORY;

    if (allocate_block (tree, (char *)&p, sizeof(NODE)) < 0)   /* new node */
    {
      (void)deallocate_block (&p.data);
      return BT_OUT_OF_MEMORY;
    }

    *help = 1;
    *tree_changed = 1;
    return 0;
  }

  /* load node 'tree' into 'p' */
  if (retrieve_block (tree, (char *)&p, sizeof(NODE), &dummy) < 0)
    return BT_IO_ERROR;  /* cannot retrieve virtual memory block */

  rc = compare_data (btree, data, size, &p.data, &cmp);
  if (rc)
    return rc;

  if (cmp == 0)                   /* key was found */
    return BT_DUPLICATE_KEY;

  if (cmp < 0)
  {
    ins0 = 0;     /* insertion in left subtree */
    ins1 = 1;
    bal1 = +1;
    bal2 = -1;
  }
  else    /* cmp > 0 */
  {
    ins0 = 1;     /* insertion in right subtree */
    ins1 = 0;
    bal1 = -1;
    bal2 = +1;
  }

  rc = insert_tree (btree, &p.ptr[ins0], data, size, help, &local_tree_changed);
  if (rc)
    return rc;

  if (*help)
  {
    local_tree_changed = 1;   /* force update of node p */

    if (p.bal == bal1)
    {
      p.bal = 0;
      *help = 0;
      *tree_changed = 0;
    }
    else if (p.bal == 0)
    {
      p.bal = bal2;
      *tree_changed = 0;
    }
    else  /* p.bal == bal2 */
    {
      /* load node 'p1' */
      handlecpy (&handle_p1, &p.ptr[ins0]);
      if (retrieve_block (&handle_p1, (char *)&p1, sizeof(NODE), &dummy) < 0)
      {
        btree->valid = 0;
        return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
      }

      if (p1.bal == bal2)
      {
        handlecpy (&p.ptr[ins0], &p1.ptr[ins1]);
        handlecpy (&p1.ptr[ins1], tree);
        p.bal = 0;

        /* save 'p' */
        if (update_block (tree, (char *)&p, sizeof(NODE)) < 0)
        {
          btree->valid = 0;    /* cannot update virtual memory block */
          return BT_TREE_CORRUPTED;
        }

        memcpy (&p, &p1, sizeof(NODE));
        handlecpy (tree, &handle_p1);
      }
      else
      {
        handlecpy (&handle_p2, &p1.ptr[ins1]);
        if (retrieve_block (&handle_p2, (char *)&p2, sizeof(NODE), &dummy) < 0)
        {
          btree->valid = 0;
          return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
        }

        handlecpy (&p1.ptr[ins1], &p2.ptr[ins0]);
        handlecpy (&p2.ptr[ins0], &handle_p1);
        handlecpy (&p.ptr[ins0],  &p2.ptr[ins1]);
        handlecpy (&p2.ptr[ins1], tree);

        if (p2.bal == bal2)
          p.bal = bal1;
        else
          p.bal = 0;

        if (p2.bal == bal1)
          p1.bal = bal2;
        else
          p1.bal = 0;

        /* save 'p', 'p1' */
        if (update_block (tree, (char *)&p, sizeof(NODE)) < 0 ||
            update_block (&handle_p1, (char *)&p1, sizeof(NODE)) < 0)
        {
          btree->valid = 0;
          return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
        }

        memcpy (&p, &p2, sizeof(NODE));
        handlecpy (tree, &handle_p2);
      }

      p.bal = 0;
      *help = 0;
      *tree_changed = 1;
    }
  }

  if (local_tree_changed)
  {
    if (update_block (tree, (char *)&p, sizeof(NODE)) < 0)
    {
      btree->valid = 0;
      return BT_TREE_CORRUPTED;
    }
  }

  return 0;
}

/************************************************************************/

/* insert new data in binary tree.      */
/* returns 0 if OK or a negative error. */

int insert_btree (BINARY_TREE  *btree,
                  void         *data,
                  unsigned int size)
{
  short help, root_changed;

  if (btree->open != MAGIC_OPEN)
    return BT_NOT_OPEN;

  if (btree->valid != MAGIC_VALID)
    return BT_TREE_CORRUPTED;

  return insert_tree (btree, &btree->root, data, size, &help, &root_changed);
}

/************************************************************************/

/* this function is called to rebalance the tree.            */
/* node '*p' must always be rewritten to disk after the call */
/* because its .bal field was changed.                       */

static int rebalance (BINARY_TREE  *btree,
                      BLOCK_HANDLE *tree,
                      NODE         *p,
                      short        mode,     /* 1 = left, 2 = right */
                      short        *help)    /* always 1 before the call */
{
  BLOCK_HANDLE handle_p1, handle_p2, temp;
  NODE         p1, p2;
  short        bal1, bal2, b1, b2;
  int          dummy, del0, del1;

  if (mode == 1)
  {
    bal1 = -1;
    bal2 = +1;
    del0 = 0;
    del1 = 1;
  }
  else
  {
    bal1 = +1;
    bal2 = -1;
    del0 = 1;
    del1 = 0;
  }

  if (p->bal == bal1)   /* a subtree was taller : it is balanced now */
  {
    p->bal = 0;
    return 0;           /* rebalance the node above (*help still equals 1) */
  }

  if (p->bal == 0)    /* it was balanced : it is slighly unbalanced now */
  {
    p->bal = bal2;
    *help = 0;
    return 0;
  }

  /* subtree is already taller : we're out of balance ! */

  handlecpy (&handle_p1, &p->ptr[del1]);

  /* load node 'p1' */
  if (retrieve_block (&handle_p1, (char *)&p1, sizeof(NODE), &dummy) < 0)
  {
    btree->valid = 0;
    return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
  }

  b1 = p1.bal;
  if (b1 == bal2 || b1 == 0)
  {
    handlecpy (&p->ptr[del1], &p1.ptr[del1]);
    handlecpy (&p1.ptr[del1], &p1.ptr[del0]);
    handlecpy (&p1.ptr[del0], &p->ptr[del0]);
    handlecpy (&p->ptr[del0], &handle_p1);

    /* exchange data fields of p and p1 */
    handlecpy (&temp,    &p->data);
    handlecpy (&p->data, &p1.data);
    handlecpy (&p1.data, &temp);

    if (b1 == 0)
    {
      p1.bal = bal2;
      p->bal = bal1;
      *help = 0;
    }
    else
    {
      p->bal = p1.bal = 0;
    }
  }
  else   /* b1 == bal1 */
  {
    handlecpy (&handle_p2, &p1.ptr[del0]);

    /* load node 'p2' */
    if (retrieve_block (&handle_p2, (char *)&p2, sizeof(NODE), &dummy) < 0)
    {
      btree->valid = 0;
      return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
    }

    b2 = p2.bal;

    handlecpy (&p1.ptr[del0], &p2.ptr[del1]);
    handlecpy (&p2.ptr[del1], &p2.ptr[del0]);
    handlecpy (&p2.ptr[del0], &p->ptr[del0]);
    handlecpy (&p->ptr[del0], &handle_p2);
    handlecpy (&p->ptr[del1], &handle_p1);

    /* exchange data fields of p and p2 */
    handlecpy (&temp,    &p->data);
    handlecpy (&p->data, &p2.data);
    handlecpy (&p2.data, &temp);

    if (b2 == bal2)
      p2.bal = bal1;
    else
      p2.bal = 0;

    if (b2 == bal1)
      p1.bal = bal2;
    else
      p1.bal = 0;

    p->bal = 0;


    /* rewrite node 'p2' */

    if (update_block (&handle_p2, (char *)&p2, sizeof(NODE)) < 0)
    {
      btree->valid = 0;
      return BT_TREE_CORRUPTED;
    }
  }


  /* rewrite node 'p1' */

  if (update_block (&handle_p1, (char *)&p1, sizeof(NODE)) < 0)
  {
    btree->valid = 0;
    return BT_TREE_CORRUPTED;
  }

  return 0;
}

/************************************************************************/

/* node '*p' must always be rewritten to disk after the recursion ends */
/* even if *help == 0, because its data field was changed.             */

static int del (BINARY_TREE  *btree,
                BLOCK_HANDLE *handle_r,   /* node just smaller than p */
                NODE         *p,
                short        *help)
{
  NODE r;
  int  rc, dummy;

  /* load node 'handle_r' into 'r' */
  if (retrieve_block (handle_r, (char *)&r, sizeof(NODE), &dummy) < 0)
  {
    btree->valid = 0;
    return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
  }

  if (!is_null_handle (&r.ptr[1]))
  {
    rc = del (btree, &r.ptr[1], p, help);  /* r.ptr[1] changes if (*help) */
    if (rc)
      return rc;

    if (*help)    /* this means also that r.ptr[1] was changed above */
    {
      rc = rebalance (btree, handle_r, &r, 2, help);
      if (rc)
        return rc;

      if (update_block (handle_r, (char *)&r, sizeof(NODE)) < 0)
      {
        btree->valid = 0;
        return BT_TREE_CORRUPTED;
      }
    }

    return 0;
  }

  /* we arrive at the bottom of the tree (r.ptr[1] is a null handle) */

  /* fill node *p with data reference of node *r */
  handlecpy (&p->data, &r.data);

  /* delete node r */
  rc = deallocate_block (handle_r);    /* will be set to null handle */
  if (rc)
  {
    btree->valid = 0;
    return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
  }

  handlecpy (handle_r, &r.ptr[0]);
  *help = 1;       /* node above r must be rewritten */
  return 0;
}

/************************************************************************/

static int delete_tree (BINARY_TREE  *btree,    /* main fields       */
                        BLOCK_HANDLE *tree,     /* deletion point    */
                        void         *data,     /* data              */
                        unsigned int size,      /* data size         */
                        short        *help)     /* 1=underflow, 0=ok */
{
  NODE p;
  int  rc, dummy, cmp;

  if (is_null_handle (tree))
  {
    *help = 0;
    return BT_KEY_NOT_FOUND;
  }

  /* load node 'tree' into 'p' */
  if (retrieve_block (tree, (char *)&p, sizeof(NODE), &dummy) < 0)
    return BT_IO_ERROR;  /* cannot retrieve virtual memory block */

  rc = compare_data (btree, data, size, &p.data, &cmp);
  if (rc)
    return rc;

  if (cmp == 0)    /* found node to be deleted */
  {
    /* delete already the user data of the node */
    rc = deallocate_block (&p.data);    /* p.data becomes a null handle */
    if (rc)
    {
      btree->valid = 0;
      return BT_TREE_CORRUPTED;    /* cannot delete node */
    }

    if (is_null_handle (&p.ptr[1]))      /* right subtree is empty */
    {
      rc = deallocate_block (tree);        /* delete the node itself */
      if (rc)
      {
        btree->valid = 0;
        return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
      }

      handlecpy (tree, &p.ptr[0]);
      *help = 1;     /* node above must be updated */
    }
    else if (is_null_handle (&p.ptr[0])) /* left subtree is empty */
    {
      rc = deallocate_block (tree);        /* delete the node itself */
      if (rc)
      {
        btree->valid = 0;
        return BT_TREE_CORRUPTED;    /* cannot rebalance tree */
      }

      handlecpy (tree, &p.ptr[1]);
      *help = 1;     /* node above must be updated */
    }
    else     /* both subtrees exist */
    {
      rc = del (btree, &p.ptr[0], &p, help);   /* this always changes 'p' */
      if (rc)
        return rc;

      if (*help)  /* this means also that p.ptr[0] was changed */
      {
        rc = rebalance (btree, tree, &p, 1, help);
        if (rc)
          return rc;
      }

      if (update_block (tree, (char *)&p, sizeof(NODE)) < 0)
      {
        btree->valid = 0;
        return BT_TREE_CORRUPTED;
      }
    }

    return 0;
  }

  if (cmp < 0)    /* delete in left subtree */
  {
    rc = delete_tree (btree, &p.ptr[0], data, size, help);
    if (rc)
      return rc;

    if (*help)   /* this means also that p.ptr[0] was changed above */
    {
      rc = rebalance (btree, tree, &p, 1, help);
      if (rc)
        return rc;

      if (update_block (tree, (char *)&p, sizeof(NODE)) < 0)
      {
        btree->valid = 0;
        return BT_TREE_CORRUPTED;
      }
    }
  }
  else /* cmp > 0 */    /* delete in ptr[1] subtree */
  {
    rc = delete_tree (btree, &p.ptr[1], data, size, help);
    if (rc)
      return rc;

    if (*help)
    {
      rc = rebalance (btree, tree, &p, 2, help);
      if (rc)
        return rc;

      if (update_block (tree, (char *)&p, sizeof(NODE)) < 0)
      {
        btree->valid = 0;
        return BT_TREE_CORRUPTED;
      }
    }
  }

  return 0;
}

/************************************************************************/

int delete_btree (BINARY_TREE  *btree,
                  void         *data,
                  unsigned int size)
{
  short help;

  if (btree->open != MAGIC_OPEN)
    return BT_NOT_OPEN;

  if (btree->valid != MAGIC_VALID)
    return BT_TREE_CORRUPTED;

  return delete_tree (btree, &btree->root, data, size, &help);
}

/************************************************************************/

int update_btree (BINARY_TREE  *btree,
                  void         *data,
                  unsigned int size)
{
  BLOCK_HANDLE tree, temp;
  NODE         p;
  int          rc, dummy, cmp;

  if (btree->open != MAGIC_OPEN)
    return BT_NOT_OPEN;

  if (btree->valid != MAGIC_VALID)
    return BT_TREE_CORRUPTED;

  handlecpy (&tree, &btree->root);

  for (;;)
  {
    if (is_null_handle (&tree))
      return BT_KEY_NOT_FOUND;

    /* load node 'tree' into 'p' */
    if (retrieve_block (&tree, (char *)&p, sizeof(NODE), &dummy) < 0)
      return BT_IO_ERROR;  /* cannot retrieve virtual memory block */

    rc = compare_data (btree, data, size, &p.data, &cmp);
    if (rc)
      return rc;

    if (cmp == 0)                    /* found node to be modified */
      break;

    if (cmp < 0)                     /* modify in left subtree */
      handlecpy (&tree, &p.ptr[0]);
    else                             /* modify in right subtree */
      handlecpy (&tree, &p.ptr[1]);
  }

  /* modify the data of node p */

  if (size == (unsigned int)size_of_block (&p.data))
  {
    /* the easy case : sizes are the same : simply update the data */

    if (update_block (&p.data, (char *)data, size) < 0)
    {
      btree->valid = 0;
      return BT_TREE_CORRUPTED;
    }
  }
  else
  {
    /* the hard case : we must free the block and allocate a new one */

    if (allocate_block (&temp, data, size) < 0)  /* allocate user data */
      return BT_OUT_OF_MEMORY;

    rc = deallocate_block (&p.data);    /* p.data becomes a null handle */
    if (rc)
    {
      (void)deallocate_block (&temp);
      btree->valid = 0;
      return BT_TREE_CORRUPTED;
    }

    handlecpy (&p.data, &temp);

    if (update_block (&tree, (char *)&p, sizeof(NODE)) < 0)
    {
      btree->valid = 0;
      return BT_TREE_CORRUPTED;
    }
  }

  return 0;
}

/************************************************************************/

int retrieve_btree (BINARY_TREE  *btree,
                    void         *data,
                    unsigned int size,         /* size of search data */
                    unsigned int max_size,     /* maximum buffer size */
                    unsigned int *actual_size,      /* of result data */
                    short        retrieve_mode)
{
  BLOCK_HANDLE tree, result;
  NODE         p;
  int          rc, dummy, cmp;
  unsigned int return_size;

  if (btree->open != MAGIC_OPEN)
    return BT_NOT_OPEN;

  if (btree->valid != MAGIC_VALID)
    return BT_TREE_CORRUPTED;

  assert (retrieve_mode >= 0 && retrieve_mode <= 6);

  handlecpy (&tree, &btree->root);

  set_null_handle (&result);

  while (!is_null_handle (&tree))
  {
    /* load node 'tree' into 'p' */
    if (retrieve_block (&tree, (char *)&p, sizeof(NODE), &dummy) < 0)
      return BT_IO_ERROR;  /* cannot retrieve virtual memory block */

    if (retrieve_mode == BT_FIRST || retrieve_mode == BT_LAST)
    {
      if (retrieve_mode == BT_FIRST)
        handlecpy (&tree, &p.ptr[0]);
      else
        handlecpy (&tree, &p.ptr[1]);

      handlecpy (&result, &p.data);  /* save data in case it's a leaf */
      continue;   /* continue til reaching a leaf node */
    }

    rc = compare_data (btree, data, size, &p.data, &cmp);
    if (rc)
      return rc;

    if (cmp == 0)                    /* found equal node */
    {
      if (retrieve_mode == BT_EQUAL            ||
          retrieve_mode == BT_EQUAL_OR_SMALLER ||
          retrieve_mode == BT_EQUAL_OR_LARGER)
      {
        handlecpy (&result, &p.data);
        break;                            /* the search ends here */
      }
      else
      {
        if (retrieve_mode == BT_SMALLER)
          handlecpy (&tree, &p.ptr[0]);
        else  /* retrieve_mode == BT_LARGER */
          handlecpy (&tree, &p.ptr[1]);
      }
    }
    else
    {
      /* test if the node matches the criteria */

      switch (retrieve_mode)
      {
        case BT_EQUAL:   /* do nothing */
          break;

        case BT_EQUAL_OR_SMALLER:
        case BT_SMALLER:
          if (cmp > 0)     /* provided value is larger than node data */
            handlecpy (&result, &p.data);
          break;

        case BT_EQUAL_OR_LARGER:
        case BT_LARGER:
          if (cmp < 0)     /* provided value is smaller than node data */
            handlecpy (&result, &p.data);
          break;

        default:
          assert (0);
      }

      /* try to find a node with a closer value */
      if (cmp < 0)
        handlecpy (&tree, &p.ptr[0]);
      else
        handlecpy (&tree, &p.ptr[1]);
    }
  }

  if (is_null_handle (&result))
    return BT_KEY_NOT_FOUND;

  /* retrieve result data */

  return_size = (unsigned int)size_of_block (&result);

  if (return_size > max_size)
    return BT_BUFFER_TOO_SMALL;

  /* load node 'tree' into 'p' */
  if (retrieve_block (&result, (char *)data, return_size, &dummy) < 0)
    return BT_IO_ERROR;  /* cannot retrieve virtual memory block */

  *actual_size = return_size;

  return 0;
}

/************************************************************************/

static int operate_on_data (BINARY_TREE     *btree,
                            BLOCK_HANDLE    *data,
                            OPERATE_ELEMENT *operate)
{
  unsigned int size;
  void         *p;
  int          dummy;

  /* get size of user data */
  size = (unsigned int)size_of_block (data);

  /* reserve stack space to hold data */
  p = (void *)alloca ((size_t)size);
  if (p == (void *)0)
    return BT_OUT_OF_MEMORY;

  if (retrieve_block (data, (char *)p, size, &dummy) < 0)
    return BT_IO_ERROR;  /* cannot retrieve virtual memory block */

  if ((*operate)(btree->user, p, size))
    return BT_ABORT;

  return 0;
}

/************************************************************************/

static int intern_traverse_tree (BINARY_TREE     *btree,
                                 BLOCK_HANDLE    *tree,
                                 OPERATE_ELEMENT *operate,
                                 int             index)     /* 0 or 1 */
{
  NODE p;
  int  dummy, rc;

  if (is_null_handle (tree))
    return 0;

  /* load node 'tree' into 'p' */
  if (retrieve_block (tree, (char *)&p, sizeof(NODE), &dummy) < 0)
    return BT_IO_ERROR;  /* cannot retrieve virtual memory block */

  rc = intern_traverse_tree (btree, &p.ptr[index], operate, index);
  if (rc)
    return rc;

  rc = operate_on_data (btree, &p.data, operate);
  if (rc)
    return rc;

  rc = intern_traverse_tree (btree, &p.ptr[1-index], operate, index);
  if (rc)
    return rc;

  return 0;
}

/************************************************************************/

int traverse_btree (BINARY_TREE     *btree,
                    OPERATE_ELEMENT *operate,
                    short           order)     /* -1 or +1 */
{
  int index;

  if (btree->open != MAGIC_OPEN)
    return BT_NOT_OPEN;

  if (btree->valid != MAGIC_VALID)
    return BT_TREE_CORRUPTED;

  assert (order == -1 || order == +1);

  index = (order == +1) ? 0 : 1;

  return intern_traverse_tree (btree, &btree->root, operate, index);
}

/************************************************************************/
