#define BINTREE_H

/* bintree.h : balanced binary trees */

/************************************************************************/

#ifndef BLOCK_H
#include "block.h"
#endif

/************************************************************************/

/* extract and compare key part of user data.         */
/* returns -1 if (*data1, size1) < (*data2, size2),   */
/*          0 if equal,                               */
/*         +1 if (*data1, size1) > (*data2, size2),   */
/* warning: the user must not modify data1 or data2 ! */

typedef int COMPARE_ELEMENT (void               *user,
                             const void         *data1,
                             const unsigned int size1,
                             const void         *data2,
                             const unsigned int size2);

/************************************************************************/

/* this structure is private and must not be accessed by the user */

typedef struct {
  BLOCK_HANDLE    root;
  long            open;
  long            valid;
  void            *user;
  COMPARE_ELEMENT *compare;
} BINARY_TREE;

/************************************************************************/

/* initialize a binary tree structure.                     */
/* 'user' can be used freely to designate user data. The   */
/* pointer is made accessible in the comparison and        */
/* operate functions.                                      */
/* assertion: the btree structure must have been allocated */
/*            by the user before the call.                 */

void create_btree (BINARY_TREE     *btree,
                   void            *user,
                   COMPARE_ELEMENT *compare);

/************************************************************************/

/* deallocate the tree and invalidate the btree structure. */
/* assertion: btree was created earlier.                   */

void close_btree (BINARY_TREE *btree);

/************************************************************************/

/* insert new data in binary tree.       */
/* returns 0 if OK or a negative error.  */

int insert_btree (BINARY_TREE  *btree,
                  void         *data,
                  unsigned int size);

/************************************************************************/

/* delete data in binary tree.                                     */
/* 'data' must contain enough information to allow key comparison. */
/* returns 0 if OK or a negative error.                            */

int delete_btree (BINARY_TREE  *btree,
                  void         *data,
                  unsigned int size);

/************************************************************************/

/* modify data in binary tree.                                     */
/* the btree must already contain data with the same key part.     */
/* returns 0 if OK or a negative error.                            */

int update_btree (BINARY_TREE  *btree,
                  void         *data,
                  unsigned int size);

/************************************************************************/

/* absolute retrieve modes (do not use the data value) */
#define BT_FIRST             0
#define BT_LAST              5

/* relative retrieve modes (use the data value) */
#define BT_EQUAL             2
#define BT_SMALLER           1
#define BT_LARGER            4
#define BT_EQUAL_OR_SMALLER  3
#define BT_EQUAL_OR_LARGER   6

/************************************************************************/

/* retrieve data in binary tree.                                       */
/* 'data' must contain enough information to allow key comparison.     */
/* assertion: the retrieve_mode is one of the constants defined above. */
/* returns 0 if OK or a negative error.                                */

int retrieve_btree (BINARY_TREE  *btree,
                    void         *data,
                    unsigned int size,      /* size of search data */
                    unsigned int max_size,  /* maximum buffer size */
                    unsigned int *actual_size,   /* of result data */
                    short        retrieve_mode);

/************************************************************************/

/* this function is called once for each traversed node.  */
/* the user must return 0 to continue. Any other value    */
/* terminates the function 'traverse_tree' with BT_ABORT. */
/* warning: the user must not modify data !               */

typedef int OPERATE_ELEMENT (void               *user,
                             const void         *data,
                             const unsigned int size);

/************************************************************************/

/* traverse btree in ascending (+1) or descending (-1) order.  */
/* returns 0 if OK or a negative error.                        */

int traverse_btree (BINARY_TREE     *btree,
                    OPERATE_ELEMENT *operate,
                    short           order);     /* -1 or +1 */

/************************************************************************/

/* possible error codes: */
#define BT_KEY_NOT_FOUND    (-1)  /* data with this key was not found  */
#define BT_DUPLICATE_KEY    (-2)  /* data with this key already exists */
#define BT_OUT_OF_MEMORY    (-3)  /* out of virtual memory             */
#define BT_IO_ERROR         (-4)  /* cannot access virtual memory      */
#define BT_TREE_CORRUPTED   (-5)  /* btree unusable due to I/O error   */
#define BT_NOT_OPEN         (-6)  /* btree was not created             */
#define BT_BUFFER_TOO_SMALL (-7)  /* buffer is too small for result    */
#define BT_ABORT            (-8)  /* user aborted traverse_tree        */

/************************************************************************/
