
/* block.c */

#ifdef DOS
  #define WITH_HEAP             0   /* 0 or 1 */   /* malloc()          */
  #define WITH_VIRTUAL_MEMORY   1   /* 0 or 1 */   /* uses 128K of heap */
#else
  #define WITH_HEAP             1   /* 0 or 1 */   /* malloc()          */
  #define WITH_VIRTUAL_MEMORY   0   /* 0 or 1 */
#endif

#if WITH_HEAP
  #include <malloc.h>
#endif

#if WITH_VIRTUAL_MEMORY
  #include <vmemory.h>
  #include <dos.h>
#endif

#include <string.h>
#include <assert.h>
#include "block.h"

/* set to 1 when handling confidential data, to 0 otherwise */
#define ERASE_BEFORE_DEALLOCATING  1

/**********************************************************************/

#define KIND_NULL             0
#define KIND_VIRTUAL_MEMORY   1
#define KIND_MEMORY           2

/**********************************************************************/

void set_null_handle (BLOCK_HANDLE *handle)
{
  handle->kind = KIND_NULL;
}

/**********************************************************************/

int is_null_handle (BLOCK_HANDLE *handle)
{
  return (handle->kind == KIND_NULL);
}

/**********************************************************************/

#if WITH_VIRTUAL_MEMORY

static int          virtual_memory_manager_initialized = 0;
static int          virtual_memory_manager_available   = 0;
static unsigned int valid_ss, valid_sp, valid_bp;
static unsigned int irq_ss, irq_sp, irq_bp;
static void (_cdecl _interrupt _far *old_handler_int_21) ();

static void _cdecl _interrupt _far new_handler_int_21
 (unsigned _es, unsigned _ds, unsigned _di, unsigned _si,
  unsigned _bp, unsigned _sp, unsigned _bx, unsigned _dx,
  unsigned _cx, unsigned _ax, unsigned _ip, unsigned _cs, unsigned _flags)
{
  if (_ax == 0x2522)  /* this program terminates or calls another program */
  {
    /* save current context */
    _asm  mov ax,ss
    _asm  mov irq_ss,ax
    _asm  mov irq_sp,sp
    _asm  mov irq_bp,bp

    if (irq_ss < valid_ss)     /* program terminates */
    {
      /* set valid context */
      _asm  mov ax,valid_ss
      _asm  mov ss,ax
      _asm  mov sp,valid_sp
      _asm  mov bp,valid_bp

      _dos_setvect (0x21, old_handler_int_21);    /* restore old vector */
      virtual_memory_manager_initialized = 0;
      virtual_memory_manager_available   = 0;
      _vheapterm();

      /* reset irq context */
      _asm  mov ax,irq_ss
      _asm  mov ss,ax
      _asm  mov sp,irq_sp
      _asm  mov bp,irq_bp
    }
  }
  _chain_intr (old_handler_int_21);   /* continue with old int21 routine */
}

#endif

/**********************************************************************/

/* side effect: 'handle' is allowed to be a field of 'buffer', */
/* so don't modify 'handle' until 'buffer' is written !!!      */

int allocate_block (BLOCK_HANDLE *handle,
                    char         *buffer,
                    int          size)
{
#if WITH_VIRTUAL_MEMORY
  if (!virtual_memory_manager_initialized)
  {
    if (_vheapinit (0, 8192, _VM_ALLSWAP) != 0)   /* ok (max 128 K RAM) */
    {
      /* save valid stack area */
      _asm  mov ax,ss
      _asm  mov valid_ss,ax
      _asm  mov valid_sp,sp
      _asm  mov valid_bp,bp

      old_handler_int_21 = _dos_getvect(0x21);  /* save old vector */
      _dos_setvect (0x21, new_handler_int_21);  /* set new vector  */
      virtual_memory_manager_available = 1;
    }
    virtual_memory_manager_initialized = 1;
  }
#endif

  if (size < 0)
    return RC_ILLEGAL_SIZE;

#if WITH_VIRTUAL_MEMORY
  /* try first 'virtual' memory */
  if (virtual_memory_manager_available)
  {
    long hnd;
    char *p;

    hnd = _vmalloc ((unsigned long)size);
    if (hnd != 0)
    {
      p = (char *)_vload (hnd, _VM_DIRTY);
      if (p != (char *)0)
      {
        memcpy (p, buffer, (size_t)size);

        handle->kind = KIND_VIRTUAL_MEMORY;
        handle->o.hnd = hnd;
        handle->size  = size;
        return 0;
      }
    }
  }
#endif

#if WITH_HEAP
  /* try secondly 'normal' memory */
  {
    char *p;
    p = (char *)malloc ((size_t)size);
    if (p != (char *)0)
    {
      memcpy (p, buffer, (size_t)size);

      handle->kind      = KIND_MEMORY;
      handle->o.pointer = p;
      handle->size      = size;
      return 0;
    }
  }
#endif

  return RC_OUT_OF_SPACE;
}

/**********************************************************************/

int retrieve_block (BLOCK_HANDLE *handle,
                    char         *buffer,
                    unsigned int buffer_size,
                    int          *size)
{
  switch (handle->kind)
  {

#if WITH_VIRTUAL_MEMORY
    case KIND_VIRTUAL_MEMORY:
      {
        char *p;
        unsigned int siz;

        p = (char *)_vload (handle->o.hnd, _VM_DIRTY);
        if (p == (char *)0)
          return RC_DEVICE_ERROR;

        siz = (unsigned int)handle->size;
        if (siz > buffer_size)
          siz = buffer_size;
        memcpy (buffer, p, siz);
        *size = siz;

        return 0;
      }
#endif

#if WITH_HEAP
    case KIND_MEMORY:
      {
        unsigned int siz;
        siz = (unsigned int)handle->size;
        if (siz > buffer_size)
          siz = buffer_size;
        memcpy (buffer, handle->o.pointer, siz);
        *size = siz;
        return 0;
      }
#endif

    default:
      return RC_ILLEGAL_HANDLE;
  }
}

/**********************************************************************/

int update_block (BLOCK_HANDLE *handle,
                  char         *buffer,
                  int          size)
{
  if (size != handle->size)
    return RC_ILLEGAL_SIZE;

  switch (handle->kind)
  {

#if WITH_VIRTUAL_MEMORY
    case KIND_VIRTUAL_MEMORY:
      {
        char *p;
        p = (char *)_vload (handle->o.hnd, _VM_DIRTY);
        if (p == (char *)0)
          return RC_DEVICE_ERROR;
        memcpy (p, buffer, size);
        return 0;
      }
#endif

#if WITH_HEAP
    case KIND_MEMORY :
      memcpy (handle->o.pointer, buffer, (size_t)size);
      return 0;
#endif

    default:
      return RC_ILLEGAL_HANDLE;
  }
}

/**********************************************************************/

int deallocate_block (BLOCK_HANDLE *handle)
{
  switch (handle->kind)
  {

#if WITH_VIRTUAL_MEMORY
    case KIND_VIRTUAL_MEMORY:
      #if ERASE_BEFORE_DEALLOCATING
      {
        char *p;
        p = (char *)_vload (handle->o.hnd, _VM_DIRTY);
        if (p != (char *)0)
          memset (p, 0x00, (size_t)handle->size);
      }
      #endif
      _vfree (handle->o.hnd);
      handle->kind = KIND_NULL;
      return 0;
#endif

#if WITH_HEAP
    case KIND_MEMORY:
#if ERASE_BEFORE_DEALLOCATING
      memset (handle->o.pointer, 0x00, (size_t)handle->size);
#endif
      free ((void *)handle->o.pointer);
      handle->kind = KIND_NULL;
      return 0;
#endif

    default:
      return RC_ILLEGAL_HANDLE;
  }
}

/**********************************************************************/

int size_of_block (BLOCK_HANDLE *handle)
{
  assert (handle->kind != KIND_NULL);
  return handle->size;
}

/**********************************************************************/
