
/* calendar.c */

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include "calendar.h"

#ifdef _WIN32
 #include <windows.h>
#else
 #include <time.h>
#endif

/* ------------------------------------------------------------------- */

void get_datetime (DATE_TIME *datetime)
{
#ifdef _WIN32     /* Windows 32-bit */

  SYSTEMTIME info;

  GetLocalTime (&info);

  datetime->year  = (unsigned int)info.wYear;
  datetime->month = (unsigned char)info.wMonth;
  datetime->day   = (unsigned char)info.wDay;
  datetime->hour  = (unsigned char)info.wHour;
  datetime->min   = (unsigned char)info.wMinute;
  datetime->sec   = (unsigned char)info.wSecond;

#else    /* 16-bit windows and unix -- NOTE: big problem in 2038 !! */

  time_t    now;
  struct tm *p;

  now = time ((time_t *)0);
  p = localtime (&now);

  datetime->year  = p->tm_year + 1900;
  datetime->month = (unsigned char)p->tm_mon  + 1;
  datetime->day   = (unsigned char)p->tm_mday;
  datetime->hour  = (unsigned char)p->tm_hour;
  datetime->min   = (unsigned char)p->tm_min;
  datetime->sec   = (unsigned char)p->tm_sec;

#endif
}

/* ------------------------------------------------------------------- */

static int is_leap_year (int year)
{
  int is_leap = 0;

  if ((year % 4) == 0)
  {
    if ((year % 100) == 0)
    {
      if ((year % 400) == 0)
        is_leap = 1;
    }
    else
    {
      is_leap = 1;
    }
  }

  return is_leap;
}

/* ------------------------------------------------------------------- */

/* assertion: year is in range 1901 .. 9999 */

int max_days_in_month (int month, int year)
{
  int n;

  assert (month >= 1 && month <= 12 && year >= 1901 && year <= 9999);

  if (month == 2)
  {
    if (is_leap_year (year))
      return 29;
    else
      return 28;
  }
  else
  {
    n = month;

    if (n > 7)       /* add 1 for month >= August */
      n++;

    if ((n % 2) == 0)
      return 30;     /* even month */
    else
      return 31;     /* odd month */
  }
}

/* ------------------------------------------------------------------- */

/* note: only years in range 1901 .. 9999 are considered as valid */

int is_valid_date (int day, int month, int year)
{
  return (year  >= 1901 && year  <= 9999)
      && (month >= 1    && month <= 12)
      && (day   >= 1    && day   <= max_days_in_month (month, year));
}

/* ------------------------------------------------------------------- */

/* assertion: year is in range 1901 .. 9999 */

long nb_days_since_1901 (int day, int month, int year)
{
  int  yy;
  long nb_days;

  assert (is_valid_date (day, month, year));

  yy = year - 1;
  nb_days = yy * 365L + (yy >> 2) - (yy / 100) + (yy / 400) - 693960
          + (month-1) * 30L + (month >> 1);

  if (month >= 3)      /* February included */
  {
    if (is_leap_year (year))
      nb_days--;           /* leap year : 29 instead of 30 */
    else
      nb_days -= 2;        /* normal year : 28 instead of 30 */
  }

  if (month == 9 || month == 11)    /* July/August = twice 31 days */
    nb_days++;

  return nb_days + (day-1);
}

/* ------------------------------------------------------------------- */

/* returns: 1=Monday, ..., 7=Sunday */
/* assertion: year is in range 1901 .. 9999 */

int day_of_week (int day, int month, int year)
{
  long nb_days;

  nb_days = nb_days_since_1901 (day, month, year);

  return 1 + (int)((nb_days+1) % 7);
}

/* ------------------------------------------------------------------- */

/* add 'days' to date; 'days' can be negative. */

void add_days (DATE_TIME *date, long days)
{
  int count;

  while (days > 0L)
  {
    count = max_days_in_month (date->month, date->year) + 1 - date->day;
    if (count > days)
    {
      date->day += (unsigned char)days;
      return;
    }

    days -= count;
    date->day = 1;

    if (date->month < 12)
      date->month++;
    else
    {
      date->month = 1;
      date->year++;

      while (days >= 365 + is_leap_year (date->year))
      {
        days -= (365 + is_leap_year (date->year));
        date->year++;
      }
    }
  }

  while (days < 0L)
  {
    if (-(int)date->day < days)
    {
      date->day -= (unsigned char)(-days);
      return;
    }

    days += date->day;

    if (date->month > 1)
      date->month--;
    else
    {
      date->month = 12;
      date->year--;

      while (days <= -(365 + is_leap_year (date->year)))
      {
        days += (365 + is_leap_year (date->year));
        date->year--;
      }
    }

    date->day = max_days_in_month (date->month, date->year);
  }
}

/* ------------------------------------------------------------------- */

/* add 'secs' to date; 'secs' can be negative. */

void add_seconds (DATE_TIME *date, long secs)
{
  if (secs >= 0)
  {
    add_days (date, secs / 86400);
    secs %= 86400;
  }
  else
  {
    add_days (date, -((-secs) / 86400));
    secs = -((-secs) % 86400);
  }

  secs += (date->hour * 3600L + date->min * 60L + date->sec);

  while (secs < 0)
  {
    add_days (date, -1L);
    secs += 86400;
  }

  while (secs >= 86400)
  {
    add_days (date, +1L);
    secs -= 86400;
  }

  date->sec = secs % 60;
  secs /= 60;

  date->min = secs % 60;
  secs /= 60;

  date->hour = (int)secs;
}

/* ------------------------------------------------------------------- */

int get_time_zone (void)
{
  TIME_ZONE_INFORMATION info;
  info.Bias = 0;
  GetTimeZoneInformation (&info);
  return -(int)info.Bias;
}

/* ------------------------------------------------------------------- */
