
/* files.h : portable file access */

/**************************************************************************/

#define MAX_FILENAME_LENGTH  260

/* file type : */
#define TYPE_REGULAR_FILE 1
#define TYPE_DIRECTORY    2
#define TYPE_DEVICE       3     /* used for UNIX only */

typedef struct {
  unsigned int  year;     /* 1901 to 9999 */
  unsigned char month;    /*    1 to   12 */
  unsigned char day;      /*    1 to   31 */
  unsigned char hour;     /*    0 to   23 */
  unsigned char min;      /*    0 to   59 */
  unsigned char sec;      /*    0 to   59 */
} FILE_TIME;

typedef struct {
  char          name[MAX_FILENAME_LENGTH];
  short         type;
  FILE_TIME     time;                   /* date/time of last write access */
  unsigned long size;
  unsigned long size_high;     /* high word of filesize (for sizes > 4GB) */

/* reserved for intern use: */
#ifdef _WIN32
  int handle;
#endif
#ifdef unix
  char *dir;
#endif

} FILE_INFO;

/* return codes : */
#define RC_NO_MORE_FILES  (-1)
#define RC_SYSTEM_ERROR   (-9)

/**************************************************************************/

/* 'directory_path' : "c:\\temp\\*.dat" for MS-DOS/Win or "/temp/" for Unix.*/
/* returns 0 if OK or a negative error code.                                */

int open_directory (FILE_INFO *file_info, char *directory_path);

/**************************************************************************/

/* returns 0 if OK or a negative error code (RC_NO_MORE_FILES = end-of-list)*/

int read_directory (FILE_INFO *file_info);

/**************************************************************************/

int close_directory (FILE_INFO *file_info);

/**************************************************************************/

/* note: MS-DOS programs do no allow recursive searching (you must finish */
/*       a search before starting the next one).                          */

/**************************************************************************/

/* source file attributes (date/time/flags) are not copied to the target */
int copy_file (char *source, char *target);

/**************************************************************************/

/* compute a filename from :                                      */
/* - a current directory,                                         */
/* - an absolute or relative filename.                            */
/* normalizes the filename.                                       */
/* note: source and result filenames can denote the same buffer.  */
/* returns 0 if OK, -1 if the filenames are too large.            */

int expand_pathname (char *current_directory,
                     char *source_filename,
                     char *result_filename,
                     int  size_of_result_filename);

/**************************************************************************/

/* filename  ::= [device] [pathname]                        */
/* device    ::= drive-letter: | \\computer | (empty)       */
/* pathname  ::= {[/|\] [name]}                             */
/*                                                          */
/* replace "\" by "/" except first one after \\computer     */
/* replace "//" by "/".                                     */
/* normalize ".", "..", "...", .. path components.          */

void normalize_pathname (char *pathname);

/**************************************************************************/

/* returns 0 or -1 */
int get_file_size (int           fd,
                   unsigned long *low,
                   unsigned long *high);

/**************************************************************************/

/* returns 0 or -1 */
int get_file_pos (int           fd,
                  unsigned long *low,
                  unsigned long *high);

/**************************************************************************/

/* returns 0 or -1 */
int set_file_pos (int           fd,
                  unsigned long low,
                  unsigned long high);

/**************************************************************************/

/* add signed 'offset' to file position (low,high) */
void add_file_pos (unsigned long *low,
                   unsigned long *high,
                   long          offset);

/**************************************************************************/

/* add unsigned filesize to file position (low,high) */

void add_file_size (unsigned long *low,
                    unsigned long *high,
                    unsigned long size_low,
                    unsigned long size_high);

/**************************************************************************/

/* test if a "name" matches a specified "pattern".            */
/* the "name" can contain special characters :                */
/* "replace_one",  which stands for any single character,     */
/* "replace_many", which stands for zero or more characters.  */
/* All other characters must match exactly.                   */
/* the replace characters should be set to '\0' if not used.  */
/* returns 1 (TRUE) if there's a match, 0 (FALSE) otherwise.  */
/* warning: the function calls itself recursively n times,    */
/*          where n is the number of non-consecutive          */
/*          replace_many characters in the pattern.           */

int name_matches_pattern (char *name,
                          char *pattern,
                          char replace_one,     /* should be '?' */
                          char replace_many);   /* should be '*' */

/**************************************************************************/

int get_disk_free_space (char          *directory,
                         unsigned long *free_low,
                         unsigned long *free_high,
                         unsigned long *total_low,
                         unsigned long *total_high);

/**************************************************************************/

/* get date/time of last write access */

int get_file_datetime (int       fd,
                       FILE_TIME *datetime);

/**************************************************************************/
