
/* inifile.c : read .ini file */

#ifdef _WIN32
  #include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <malloc.h>
#include "tracing.h"
#include "inifile.h"

/************************************************************************/

typedef struct item {
  char        *key;
  char        *value;
  struct item *next;
} ITEM;

typedef struct section {
  char           *section;
  struct section *next;
  ITEM           *first;
  ITEM           *last;
} SECTION;

typedef struct {
  char    *filename;
  int     rc;           /* 0 if OK, or E_INI_FILE_NOT_FOUND */
  SECTION *first;
  SECTION *last;
} INI_DATA;

#define MAX_INI_FILES  10   /* max bufferized ini files */

typedef struct {
  INI_DATA ini[MAX_INI_FILES];
  int      count;
} INI_INFO;

static INI_INFO ini;

/************************************************************************/

static void append_section (INI_DATA *p, char *section)
{
  SECTION *node;
  char    *s;

  node = malloc (sizeof(SECTION));
  s    = malloc (strlen(section)+1);

  if (node == NULL || s == NULL)
  {
    free (node);
    free (s);
    trace ("error: inifile.c : malloc() failed\n");
    return;
  }

  memset (node, 0x00, sizeof(SECTION));
  strcpy (s, section);

  node->section = s;

  if (p->first == NULL)
    p->first = p->last = node;
  else
  {
    p->last->next = node;
    p->last = node;
  }
}

/************************************************************************/

static void append_item (INI_DATA *p, char *key, char *value)
{
  ITEM *item;
  char *k, *v;

  if (p->last == NULL)   /* no section exists yet : skip item */
    return;

  item = malloc (sizeof(ITEM));
  k    = malloc (strlen(key)+1);
  v    = malloc (strlen(value)+1);

  if (item == NULL || k == NULL || v == NULL)
  {
    free (item);
    free (k);
    free (v);
    trace ("error: inifile.c : malloc() failed\n");
    return;
  }

  memset (item, 0x00, sizeof(ITEM));
  strcpy (k, key);
  strcpy (v, value);

  item->key   = k;
  item->value = v;

  if (p->last->first == NULL)
    p->last->first = p->last->last = item;
  else
  {
    p->last->last->next = item;
    p->last->last = item;
  }
}

/************************************************************************/

static void load_file (char *filename, INI_DATA *p)
{
  FILE          *fp;
  unsigned char line[4097];
  unsigned int  size, i, j, k;

  memset (p, 0x00, sizeof(INI_DATA));

  /* save filename */
  p->filename = malloc (strlen(filename)+1);
  if (p->filename == NULL)
  {
    trace ("error: inifile.c : malloc() failed\n");
    return;
  }
  strcpy (p->filename, filename);

  /* open the file */
  fp = fopen (filename, "r");
  if (fp == (FILE *)0)
  {
    p->rc = E_INI_FILE_NOT_FOUND;
    return;
  }

  for (;;)
  {
    if (fgets (line, sizeof(line), fp) == NULL)      /* end-of-file */
      break;

    /* 'size' = the active line length */
    size = strlen(line);

    /* delete trailing '\n' character */
    if (size > 0 && line[size-1] == '\n')
      size--;

    /* compact doubled '#' or ';' characters and also find      */
    /* non-doubled '#' or ';' character, and cut the line then. */

    i = 0;    /* source index */
    j = 0;    /* target index */

    while (i < size)
    {
      if (line[i] == '#' || line[i] == ';')   /* special symbol */
      {
        if (line[i] == line[i+1])   /* doubled */
        {
          line[j] = line[i];
          i += 2;
          j += 1;
        }
        else     /* non-doubled -> end of line */
        {
          break;
        }
      }
      else   /* normal symbol */
      {
        line[j++] = line[i++];
      }
    }

    size = j;


    /* delete all trailing white space */
    while (size > 0 && line[size-1] <= ' ')
      size--;

    /* skip empty line */
    if (size == 0)
      continue;

    line[size] = '\0';


    /* test whether '[' or '=' comes first */

    for (i=0; i<size; i++)
    {
      if (line[i] == '[' || line[i] == '=')
        break;
    }

    /* skip line having incorrect format */
    if (i == size)
      continue;

    if (line[i] == '[')   /* it's a section */
    {
      i = 0;

      while (i<size && line[i]<=' ')
        i++;

      if (i==size || line[i] != '[')
        continue;

      i++;    /* skip '[' */

      while (i<size && line[i]<=' ')
        i++;

      j = i;

      while (j<size && line[j]!=']')
        j++;

      while (j>i && line[j-1]<=' ')
        j--;

      if (i == j)  /* empty */
        continue;

      line[j] = '\0';
      append_section (p, &line[i]);
    }
    else    /* it's an assignment "item = value" */
    {
      i = 0;

      while (i<size && line[i]<=' ')
        i++;

      j = i;

      while (j<size && line[j]!='=')
        j++;

      k = j;  /* points to '=' */

      while (j>i && line[j-1]<=' ')
        j--;

      if (i == j)  /* empty */
        continue;

      line[j] = '\0';

      k++;
      while (k<size && line[k]<=' ')
        k++;

      append_item (p, &line[i], &line[k]);
    }
  }

  fclose (fp);
}

/************************************************************************/

static void deallocate_tree (INI_DATA *p)
{
  SECTION *s, *s2;
  ITEM    *i, *i2;

  free (p->filename);

  s = p->first;
  while (s != NULL)
  {
    free (s->section);

    i = s->first;
    while (i != NULL)
    {
      i2 = i;
      i = i->next;
      free (i2->key);
      free (i2->value);
      free (i2);
    }

    s2 = s;
    s = s->next;
    free (s2);
  }

  memset (p, 0x00, sizeof(INI_DATA));
}

/************************************************************************/

/* make sure the file is loaded in-memory */

static INI_DATA *fetch_file (char *filename)
{
  int        i;
  static int next_i;

  for (i=0; i<ini.count; i++)
  {
    if (_stricmp (filename, ini.ini[i].filename) == 0)
      return &ini.ini[i];
  }

  /* find a victim slot 'i' to deallocate */
  if (i < MAX_INI_FILES)
  {
    ini.count = i + 1;
  }
  else
  {
    i = next_i;
    next_i = (next_i + 1) % MAX_INI_FILES;
  }

  deallocate_tree (&ini.ini[i]);
  load_file (filename, &ini.ini[i]);

  return &ini.ini[i];
}

/************************************************************************/

static int copy_value (char         *value,    /* can be NULL */
                       char         *buffer,
                       unsigned int buffer_size)
{
  unsigned int size;

  if (buffer_size == 0)
    return E_INI_BUFFER_TOO_SMALL;

  if (value == NULL)
  {
    memset (buffer, 0x00, buffer_size);
    return 0;
  }

  size = strlen (value);
  if (size+1 > buffer_size)
  {
    memcpy (buffer, value, buffer_size-1);
    buffer[buffer_size-1] = '\0';
    return E_INI_BUFFER_TOO_SMALL;
  }
  else
  {
    strncpy (buffer, value, buffer_size);
    return 0;
  }
}

/************************************************************************/

#ifdef _WIN32
 volatile static int              critical_section_initialized = 0;
 static LONG             critical_section_inc = -1L;
 static CRITICAL_SECTION critical_section;
#endif

/***********************************************************************/

static void private_begin_critical_section (void)
{
#ifdef _WIN32
  if (!critical_section_initialized)
  {
    if (InterlockedIncrement (&critical_section_inc) == 0L) /* first thread */
    {
      InitializeCriticalSection (&critical_section);
      critical_section_initialized = 1;
    }
    else  /* not first thread : wait until critical section is initialized */
    {
      while (!critical_section_initialized)
        Sleep (0);
    }
  }

  EnterCriticalSection (&critical_section);
#else
  assert (0);
#endif
}

/***********************************************************************/

static void private_end_critical_section (void)
{
#ifdef _WIN32
  if (critical_section_initialized)
    LeaveCriticalSection (&critical_section);
#else
  assert (0);
#endif
}

/***********************************************************************/

/* retrieve parameter from .ini file                     */
/* returns 0 or a negative error code.                   */
/* in case of error, the default value (if any) is used. */

int get_parameter (char         *filename,        /* param.ini     */
                   char         *section,         /* section name  */
                   char         *key,             /* key name      */
                   char         *parameter,       /* out buffer    */
                   unsigned int parameter_size,   /* sizeof buffer */
                   char         *default_value)   /* can be NULL   */
{
  INI_DATA *p;
  SECTION  *s;
  ITEM     *i;
  int      rc;

  private_begin_critical_section ();

  /* load file */
  p = fetch_file (filename);
  if (p->rc)
  {
    copy_value (default_value, parameter, parameter_size);
    rc = p->rc;
    private_end_critical_section ();
    return rc;
  }

  /* find matching section */
  s = p->first;
  while (s != NULL)
  {
    if (_stricmp (s->section, section) == 0)
    {
      i = s->first;
      while (i != NULL)
      {
        if (_stricmp (i->key, key) == 0)
        {
          rc = copy_value (i->value, parameter, parameter_size);
          private_end_critical_section ();
          return rc;
        }
        i = i->next;
      }

      copy_value (default_value, parameter, parameter_size);
      private_end_critical_section ();
      return E_INI_KEY_NOT_FOUND;
    }

    s = s->next;
  }

  copy_value (default_value, parameter, parameter_size);
  private_end_critical_section ();
  return E_INI_SECTION_NOT_FOUND;
}

/************************************************************************/

/* retrieve nth section of the file.   */
/* returns 0 or a negative error code. */

int get_nth_section (char         *filename,        /* ini filename  */
                     int          nth,              /* 0 .. ?        */
                     char         *section,         /* out buffer    */
                     unsigned int section_size)     /* sizeof buffer */
{
  INI_DATA *p;
  SECTION  *s;
  int      j;
  int      rc;

  private_begin_critical_section ();

  /* load file */
  p = fetch_file (filename);
  if (p->rc)
  {
    memset (section, 0x00, section_size);
    rc = p->rc;
    private_end_critical_section ();
    return rc;
  }

  if (nth < 0)
  {
    memset (section, 0x00, section_size);
    private_end_critical_section ();
    return E_INI_BAD_INDEX;
  }

  /* find nth section */
  s = p->first;
  for (j=0; j<nth; j++)
  {
    if (s == NULL)
    {
      memset (section, 0x00, section_size);
      private_end_critical_section ();
      return E_INI_BAD_INDEX;
    }
    s = s->next;
  }

  if (s == NULL)
  {
    memset (section, 0x00, section_size);
    private_end_critical_section ();
    return E_INI_BAD_INDEX;
  }

  rc = copy_value (s->section, section, section_size);
  private_end_critical_section ();
  return rc;
}

/************************************************************************/

/* retrieve nth key and parameter within given section of the file. */
/* returns 0 or a negative error code.                              */

static int get_nth_key_param (char         *filename,  /* ini filename  */
                              char         *section,   /* section name  */
                              int          nth,        /* 0 .. ?        */
                              char         *key,       /* out buffer */
                              unsigned int key_size,   /* sizeof buffer or 0 */
                              char         *parameter,     /* out buffer */
                              unsigned int parameter_size) /* sizeof buffer or 0 */
{
  INI_DATA *p;
  SECTION  *s;
  ITEM     *i;
  int      j;
  int      rc, rc2;

  private_begin_critical_section ();

  /* load file */
  p = fetch_file (filename);
  if (p->rc)
  {
    memset (key, 0x00, key_size);
    memset (parameter, 0x00, parameter_size);
    rc = p->rc;
    private_end_critical_section ();
    return rc;
  }

  if (nth < 0)
  {
    memset (key, 0x00, key_size);
    memset (parameter, 0x00, parameter_size);
    private_end_critical_section ();
    return E_INI_BAD_INDEX;
  }

  /* find matching section */
  s = p->first;
  while (s != NULL)
  {
    if (_stricmp (s->section, section) == 0)
    {
      i = s->first;

      for (j=0; j<nth; j++)
      {
        if (i == NULL)
        {
          memset (key, 0x00, key_size);
          memset (parameter, 0x00, parameter_size);
          private_end_critical_section ();
          return E_INI_BAD_INDEX;
        }
        i = i->next;
      }

      if (i == NULL)
      {
        memset (key, 0x00, key_size);
        memset (parameter, 0x00, parameter_size);
        private_end_critical_section ();
        return E_INI_BAD_INDEX;
      }

      if (key_size)
        rc = copy_value (i->key, key, key_size);
      else
        rc = 0;

      if (parameter_size)
        rc2 = copy_value (i->value, parameter, parameter_size);
      else
        rc2 = 0;

      private_end_critical_section ();
      return rc2 ? rc2 : rc;
    }

    s = s->next;
  }

  memset (key, 0x00, key_size);
  memset (parameter, 0x00, parameter_size);
  private_end_critical_section ();
  return E_INI_SECTION_NOT_FOUND;
}

/************************************************************************/

/* retrieve nth key within given section of the file. */
/* returns 0 or a negative error code.                */

int get_nth_key (char         *filename,  /* ini filename  */
                 char         *section,   /* section name  */
                 int          nth,        /* 0 .. ?        */
                 char         *key,       /* out buffer    */
                 unsigned int key_size)   /* sizeof buffer */
{
  return get_nth_key_param (filename, section, nth, key, key_size, key, 0);
}

/************************************************************************/

/* retrieve nth parameter within given section of the file. */
/* returns 0 or a negative error code.                      */

int get_nth_parameter (char         *filename,  /* ini filename  */
                       char         *section,   /* section name  */
                       int          nth,        /* 0 .. ?        */
                       char         *parameter,       /* out buffer    */
                       unsigned int parameter_size)   /* sizeof buffer */
{
  return get_nth_key_param (filename, section, nth, parameter, 0, parameter, parameter_size);
}

/************************************************************************/

/* signal that initialization file was changed and must be reloaded. */

void signal_ini_file_changed (void)
{
  int i;
  private_begin_critical_section ();
  for (i=0; i<ini.count; i++)
    deallocate_tree (&ini.ini[i]);
  ini.count = 0;
  private_end_critical_section ();
}

/************************************************************************/
