
/* pack0.c : main function and deflate algorithm (pass 1) */

/***********************************************************************/

#define DEBUG_DEFLATE  0    /* printf output after pass 1 */
                            /* (literals & references)    */

/***********************************************************************/

/* the maximum backward distance of a reference is : */
/* from -1 to -(CHUNK_SIZE - LOOK_AHEAD)             */

#define CHUNK_SIZE   32768

/* the minimum possible length of a reference at the worst point */
/* in the input buffer is LOOK_AHEAD.                            */

#define LOOK_AHEAD     256

/* constrained range of lengths accepted in a reference */
#define MIN_LENGTH      3      /* at least 3 */
#define MAX_LENGTH    258

/* max entries in hash table (2^N should be close to CHUNK_SIZE) */
#define MAX_HASH_BITS  15    /* must be >= 8 */

/* DO NOT CHANGE */
#define MAX_HASH_ENTRIES    (1 << MAX_HASH_BITS)

/* DO NOT CHANGE */
#define INPUT_BUFFER_SIZE   (2 * CHUNK_SIZE)

/* DO NOT CHANGE */
#define MAX_HNODES          (CHUNK_SIZE - LOOK_AHEAD)

/* this type must hold all values in the range 0 .. MAX_HNODES - 1 */
/* as well as the extra value NIL (see below).                     */
typedef short LINK;

/* note: use a value of LINK outside the range 0 .. MAX_HNODES - 1 */
#define NIL (-1)

/* the maximum chain length we want to travel per input byte */
#define MAX_HASH_CHAIN_LENGTH   20

/***********************************************************************/

/* memory allocation notes:                                           */
/* we allocate an input buffer of (2 * CHUNK_SIZE) bytes.             */
/* we allocate a hnode table of (CHUNK_SIZE - LOOK_AHEAD) * (6 or 8). */
/* we allocate a hash table of MAX_ENTRIES * (2 or 4 bytes).          */
/* the input file is loaded with chunks of CHUNK_SIZE bytes.          */

/***********************************************************************/

typedef struct {
  LINK next;   /* link to next hnode of the chain.                     */
  LINK *prev;  /* ptr to .next of previous hnode or hash table or NULL */
} HNODE;

/* produce a well-shaken hash value from a 3-byte character array */
/* (we take all bits into account in an evenly manner)            */

#define HASH_VALUE_OF(s) \
  ((((unsigned int)(s)[0])) \
 ^ (((unsigned int)(s)[1]) << ((MAX_HASH_BITS-8+1)) / 2) \
 ^ (((unsigned int)(s)[2]) << (MAX_HASH_BITS-8)))

#if (MAX_HASH_BITS < 8)
  ERROR : MAX_HASH_BITS must be >= 8
#endif

/***********************************************************************/

/* main packing function */

int pack (void *user_info, IO_FUNC *user_read, IO_FUNC *user_write)
{
  unsigned char input  [INPUT_BUFFER_SIZE+2];  /* input buffer + sentinel */
  LINK          hash   [MAX_HASH_ENTRIES];     /* hash table */
  HNODE         hnode  [MAX_HNODES];           /* hnode table */
  CODE          output [CHUNK_SIZE];
  int           output_i;
  int           i, rc, first, first9, last, size, eof, limit, hnode_i, first0;
  unsigned char *p;
  HNODE         *nd;
  OUTPUT_INFO   out;

  /* init hash table */
  for (i=0; i<MAX_HASH_ENTRIES; i++)
    hash[i] = NIL;                       /* set to -1 */

  /* init hnode table */
  memset (hnode, 0x00, sizeof(hnode));   /* set all .prev fields to NULL */

  /* init 'out' structure */
  init_output (&out, user_write, user_info);


  /* init various variables */

  first = last = CHUNK_SIZE;    /* portion of input[] buffer */

  eof     = 0;            /* end-of-file = FALSE */
  hnode_i = 0;            /* index of next free entry in hnode table */


  /* main loop : fill input buffer, deflate,               */
  /*             send to huffman/output, shift input data. */

  for (;;)
  {
    /* fill input buffer as much as possible, increasing 'last'. */
    /* position 'eof' in case we reached end-of-file.            */

    while (!eof)
    {
      size = INPUT_BUFFER_SIZE - last;         /* max size to fill */

      if (size == 0)                           /* the buffer is full */
        break;

      if (size > CHUNK_SIZE)                   /* limit reading size */
        size = CHUNK_SIZE;

      rc = (*user_read)(user_info, &input[last], size);
      if (rc < 0 || rc > size)
        return PK_READ_ERROR;

      if (rc == 0)      /* end-of-file */
        eof = 1;

      last += rc;
    }


    /* compute 'limit' : we will treat the range 'first' .. 'limit' */
    /* note however that we try to keep look-ahead bytes and that   */
    /* the treatment can use the entire range [first..last[.        */

    limit = INPUT_BUFFER_SIZE - LOOK_AHEAD;
    if (limit > last)   /* we don't have so many bytes */
      limit = last;


    /* save initial input buffer position for this block */
    first0 = first;


    /* setup index for output[] buffer */
    output_i = 0;


    /* setup pointer to current input position */
    p = &input[first];


    /* attention : 'first' can be set to a value larger than 'limit'    */
    /*             within the loop, in case a long reference is found   */
    /*             that crosses into the look-ahead part.               */

    while (first < limit)
    {
      int           h_value, index, idx, best_len, max_len, counter;
      int           offset, len, best_offset;
      unsigned char *q, save;


      /* compute maximum length of input buffer at current position */

      max_len = last - first;
      if (max_len > MAX_LENGTH)           /* algorithmic contraint */
        max_len = MAX_LENGTH;


      /* compute 'index' : the start hnode of the hash chain.         */
      /* note: p[1] and p[2] point possibly at input[MAX_BUFFER_SIZE] */
      /*       and input[MAX_BUFFER_SIZE+1], respectively.            */
      /* note: 'h_value' and 'idx' will be used later on.             */

      h_value = HASH_VALUE_OF (p);
      idx = index = hash [h_value];


      /* find longest possible common match */

      best_len = 0;

      counter = 0;              /* init chain length */
      for (;;)
      {
        if (index == NIL)       /* end of chain reached */
          break;

        /* compute pattern's offset in history of input buffer */
        offset = index - hnode_i;
        if (offset >= 0)
          offset -= MAX_HNODES;
        offset += first;

        /* compute pointer to fragment to compare */
        q = &input[offset];

        /* compute common length of fragments */
        if (p[best_len] == q[best_len])         /* could become a longer */
        {                                       /* matching pattern.     */
          /* store sentinel at max_len */
          save = p[max_len];              /* save sentinel value */
          p[max_len] = q[max_len] ^ 0x01; /* put mismatching character */

          len = 0;
          while (p[len] == q[len])
            len++;

          p[max_len] = save;              /* restore sentinel value */


          /* evaluate the result 'len' */

          if (len > best_len)       /* we found a better length */
          {
            best_offset = offset;
            best_len    = len;

            if (best_len == max_len)   /* we cannot find a better match */
              break;
          }
        }

        index = hnode[index].next;    /* continue with next hnode */

        /* check if we reached the maximum chain length */
        counter++;
        if (counter >= MAX_HASH_CHAIN_LENGTH)
          break;
      }


      /* if we found an interesting match, store its backward reference */

      if (best_len >= MIN_LENGTH)    /* a match of acceptable length */
      {
        output[output_i].value = best_offset - first;
        output[output_i++].length = best_len - 3;
      }
      else   /* insufficient common length : store as single literal */
      {
        output[output_i++].value = input[first];
        best_len = 1;                              /* force to length 1 */
      }


      /* advance 'best_len' bytes, updating the hnode table. */
      /* note: this might increment 'first' past 'limit'.   */

      first9 = first + best_len;     /* final value for 'first' */

      for (;;)
      {
        /* setup pointer to current hnode */
        nd = &hnode[hnode_i];

        /* remove hnode from old chain (if any) */
        if (nd->prev != NULL)
          *(nd->prev) = NIL;

        /* note: 'h_value' and 'idx' were initialized above */
        if (idx != NIL)                 /* at least 1 hnode in chain */
          hnode[idx].prev = &nd->next;

        nd->next = idx;
        nd->prev = &hash[h_value];

        hash[h_value] = hnode_i;

        /* increment hnode_i */
        hnode_i++;
        if (hnode_i == MAX_HNODES)
          hnode_i = 0;

        /* increment 'first' and 'p' */
        first++;
        p++;

        if (first == first9)    /* 'first' reached its final value */
          break;

        /* initialize 'h_value' and 'idx' for next loop */
        h_value = HASH_VALUE_OF (p);
        idx = hash[h_value];
      }
    }


#if DEBUG_DEFLATE
    /* printf output data (literals and references) */
    {
      int i, j, offset;

      offset = first0;      /* start index of input[] for this block */

      for (i=0; i<output_i; i++)
      {
        if (output[i].value >= 0)     /* a literal */
        {
          printf ("%4d : %4d (%c)\n", offset, output[i].value,
                  isprint(output[i].value) ? output[i].value : '.');
          offset++;
        }
        else                         /* a reference */
        {
          printf ("%4d : (%5d,%3d) ", offset, output[i].value,
                                              output[i].length+3);
          for (j=offset+output[i].value;
               j<offset+output[i].value+output[i].length+3;
               j++)
            printf ("%c", isprint(input[j]) ? input[j] : '.');
          printf ("\n");

          offset += output[i].length+3;
        }
      }
    }
#endif


    /* output code bit BFINAL (indicates if this is the final block) */

    rc = send_code (&out, limit == last, 1, LSB_FIRST);
    if (rc)
      return rc;


    /* encode output[] using Huffman (pass2) and send it to output */

    assert (output_i < sizeof(output));
    rc = output_block (output, output_i,
                       &input[first0], first - first0,
                       &out);
    if (rc)
      return rc;


    /* check if there is more data to treat */

    if (limit == last)   /* we have treated all data until the end */
      break;


    /* move the input buffer backwards CHUNK_SIZE bytes */

    memmove (input, &input[CHUNK_SIZE], CHUNK_SIZE);
    first -= CHUNK_SIZE;
    last  -= CHUNK_SIZE;
  }


  /* send remaining bytes in output buffer */

  rc = flush_output (&out);
  if (rc)
    return rc;

  return 0;
}

/***********************************************************************/
