
/* tracing.c */

#ifdef WIN32
  #include <windows.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <malloc.h>
#include "tracing.h"

#ifndef WIN32
  #include "calendar.h"
#endif

/***********************************************************************/

static FILE  *fp = (FILE *)0;
static char  *fn, *fn0;
static int   prefix_date, prefix_time;
static long  max_size;
static int   newline;

#ifdef WIN32
  static int prefix_msec;
  volatile static int     critical_section_initialized;
  static LONG             critical_section_inc;
  static CRITICAL_SECTION critical_section;
#endif

/***********************************************************************/

#ifdef WIN32

static void intern_begin_critical_section (void)
{
  if (!critical_section_initialized)
  {
    if (InterlockedExchange (&critical_section_inc, 7) != 7)  /* first thread to call it */
    {
      InitializeCriticalSection (&critical_section);
      critical_section_initialized = 1;
    }
    else  /* not first thread : wait until critical section is initialized */
    {
      while (!critical_section_initialized)
        Sleep (0);
    }
  }

  EnterCriticalSection (&critical_section);
}

#else

static void intern_begin_critical_section (void)
{
  $
}

#endif

/***********************************************************************/

#ifdef WIN32

static void intern_end_critical_section (void)
{
  if (critical_section_initialized)
    LeaveCriticalSection (&critical_section);
}

#else

static void intern_end_critical_section (void)
{
  $
}

#endif

/***********************************************************************/

void open_trace (char *filename,
                 int  prefix_datestamp,
                 int  prefix_timestamp,
                 long max_file_size)
{
  int i;

  intern_begin_critical_section ();

  if (fp != (FILE *)0)
  {
    intern_end_critical_section ();
    return;
  }

  if (fn != (char *)0)
    free (fn);
  if (fn0 != (char *)0)
    free (fn0);

  fn = malloc (strlen(filename)+1);
  assert (fn != (char *)0);

  fn0 = malloc (strlen(filename)+4+1);
  assert (fn0 != (char *)0);

  strcpy (fn,  filename);
  strcpy (fn0, filename);

  /* cut the extension of fn0, if any */
  for (i=(int)strlen(fn0)-1; i>=0; i--)
  {
    if (fn0[i] == '.')
    {
      fn0[i] = (char)0;
      break;
    }
  }

  /* append extension ".old" */
  strcat (fn0, ".old");

  fp = fopen (fn, "a");     /* mode 'append' */
  /* we don't care if this fails */

  newline     = 1;
  prefix_date = prefix_datestamp;
  prefix_time = prefix_timestamp;
  max_size    = max_file_size;

#ifdef WIN32
  prefix_msec = (prefix_timestamp & 2);
#endif

  intern_end_critical_section ();
}

/***********************************************************************/

void trace (char *format, ...)
{
  va_list argptr;

  intern_begin_critical_section ();

  if (fp == (FILE *)0)
  {
    intern_end_critical_section ();
    return;
  }

#ifdef _WIN32     /* Windows 32-bit */
  if (newline && (prefix_date || prefix_time || prefix_msec))
  {
    SYSTEMTIME info;

    GetLocalTime (&info);

    if (prefix_date)
      fprintf (fp, "%02d/%02d/%02d ", info.wYear % 100, info.wMonth, info.wDay);
    if (prefix_time)
      fprintf (fp, "%02d:%02d:%02d", info.wHour, info.wMinute, info.wSecond);
    if (prefix_msec)
      fprintf (fp, ".%03lu", info.wMilliseconds);

    fprintf (fp, " : ");
  }

#else   /* not WINDOWS */

  if (newline && (prefix_date || prefix_time))
  {
    DATE_TIME now;

    get_datetime (&now);

    if (prefix_date)
      fprintf (fp, "%02d/%02d/%02d ", now.year % 100, now.month, now.day);
    if (prefix_time)
      fprintf (fp, "%02d:%02d:%02d ", now.hour, now.min, now.sec);

    fprintf (fp, ": ");
  }
#endif

  va_start (argptr, format);
  vfprintf (fp, format, argptr);
  va_end (argptr);

  fflush (fp);

  newline = (format[strlen(format)-1] == '\n');

  if (ftell (fp) >= max_size)
  {
    /* we don't care if the following operations fail */
    fclose (fp);
    (void)_unlink (fn0);
    (void)rename (fn, fn0);
    fp = fopen (fn, "w");
  }

  intern_end_critical_section ();
}

/***********************************************************************/

void close_trace (void)
{
  intern_begin_critical_section ();

  if (fp != (FILE *)0)
  {
    fclose (fp);
    free (fn);
    free (fn0);
    fn = fn0 = (char *)0;
    fp = (FILE *)0;
  }

  intern_end_critical_section ();
}

/***********************************************************************/

void trace_block (unsigned char *block, unsigned int size)
{
  unsigned int  offset, i;
  unsigned char ch;
  unsigned char line[80], item[32];

  for (offset=0L; offset<size; offset+=16L)
  {
    sprintf (line, "%04X |", (unsigned int)(offset & 65535));  /* 6 chars */

    for (i=0; i<16; i++)        /* 16 x 3 = 48 chars */
    {
      if (offset + i < size)
        sprintf (item, " %02X", block[offset+i]);
      else
        strcpy (item, "   ");
      strcat (line, item);
    }

    strcat (line, " | ");    /* 3 chars */

    for (i=0; i<16; i++)       /* 16 chars */
    {
      if (offset + i < size)
      {
        ch = block[offset+i];
        if (ch >= 32 && ch != 127)
          sprintf (item, "%c", ch);
        else
          strcpy (item, ".");
        strcat (line, item);
      }
      else
      {
        strcat (line, " ");
      }
    }

    trace ("%s\n", line);    /* 6 + 48 + 3 + 16 = 73 chars */
  }
}

/***********************************************************************/
