
/* unpack3.c : input bit stream */

/**************************************************************************/

static void init_input (INPUT_INFO *info,
                        IO_FUNC    *user_read,
                        void       *user_info) /* passed to user_read() */
{
  info->first = info->count = info->power = 0;

  info->user_read = user_read;
  info->user_info = user_info;
}

/**************************************************************************/

/* returns a fixed-length code in LSB order, or a negative error. */
/* nb_bits must not exceed 15.                                    */

static int receive_code (INPUT_INFO *info, int nb_bits)
{
  int rc, code, power;

  assert (nb_bits >= 1 && nb_bits <= 15);

  code = 0;
  power = 1;      /* term to add to 'code' */

  while (nb_bits--)
  {
    if (info->first == info->count)    /* we need to read a new block */
    {
      rc = (*info->user_read) (info->user_info, info->buffer, INPUT_BUFFER_SIZE);
      if (rc <= 0)
        return UNPK_READ_ERROR;    /* we need to read at least 1 byte */

      info->first = 0;
      info->count = rc;
      info->power = 1;
    }

    if ((info->buffer[info->first] & info->power) != 0)
      code |= power;

    info->power <<= 1;
    if (info->power > 128)
    {
      info->power = 1;
      info->first++;
    }

    power <<= 1;
  }

  return code;
}

/**************************************************************************/

static int receive_block (INPUT_INFO    *info,
                          unsigned char *buf,
                          unsigned int  length)
{
  unsigned int i;
  int          rc;

  /* skip all remaining bits of the current byte */
  if (info->power > 1)
  {
    info->power = 1;
    info->first++;
  }

  /* read 'length' bytes */
  for (i=0; i<length; i++)
  {
    if (info->first == info->count)    /* we need to read a new block */
    {
      rc = (*info->user_read) (info->user_info, info->buffer, INPUT_BUFFER_SIZE);
      if (rc <= 0)
        return UNPK_READ_ERROR;    /* we need to read at least 1 byte */

      info->first = 0;
      info->count = rc;
      info->power = 1;
    }

    buf[i] = info->buffer[info->first++];
  }

  return 0;
}

/**************************************************************************/

/* analyzes a varying-length bit sequence in MSB order.     */
/* returns 0 and the corresponding huffman code in '*code'  */
/* or an error code if it does not exist.                   */

static int receive_huffman_code (INPUT_INFO *info,
                                 NODE       *node,      /* Huffman tree  */
                                 int        max_nodes,  /* nb leaf nodes */
                                 int        *code)      /* result code   */
{
  int n, bit, rc;

  n = 2 * max_nodes - 1;     /* root node */

  for (;;)
  {
    if (info->first == info->count)    /* we need to read a new block */
    {
      rc = (*info->user_read) (info->user_info, info->buffer, INPUT_BUFFER_SIZE);
      if (rc <= 0)
        return UNPK_READ_ERROR;    /* we need to read at least 1 byte */

      info->first = 0;
      info->count = rc;
      info->power = 1;
    }

    bit = ((info->buffer[info->first] & info->power) != 0);

    info->power <<= 1;
    if (info->power > 128)
    {
      info->power = 1;
      info->first++;
    }

    if (bit)
      n = node[n].index_one;
    else
      n = node[n].index_zero;

    if (n == -1)
      return UNPK_HUFF_BAD_CODE;

    if (n < max_nodes)
    {
      *code = n;
      return 0;
    }
  }
}

/**************************************************************************/

static void copy_to_read_ahead (INPUT_INFO *info, READ_AHEAD *extra)
{
  assert (INPUT_BUFFER_SIZE <= UNPACK_MAX_READ_AHEAD);

  /* skip all remaining bits of the current byte */
  if (info->power > 1)
  {
    info->power = 1;
    info->first++;
  }

  extra->size = info->count - info->first;
  memcpy (extra->buffer, &info->buffer[info->first], extra->size);
}

/**************************************************************************/
