
// libunit.c : stores a directory of all imported library unit

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <io.h>
#include <fcntl.h>

#include "bintree.h"
#include "pack.h"

#include "types.h"
#include "token.h"
#include "lex.h"
#include "error.h"

#include "makelib.h"
#include "glibpaths.h"
#include "libunit.h"

//=================================================================================

bool is_console_app;    // set to true if /std/console.h is loaded

//=================================================================================

static BINARY_TREE tree;    // ordered by library + source file name

typedef struct {
  char      *library;
  char      *source;
  NODE_INFO info;     // see makelib.h
  char      *library_filename;
} DATA;

//=================================================================================

static int sv_compare (void               *user,
                       const DATA         *data1,
                       const unsigned int size1,
                       const DATA         *data2,
                       const unsigned int size2)
{
  return 2 * strcmp (data1->library, data2->library)
           + strcmp (data1->source, data2->source);
}

//=================================================================================

void init_library_unit_tree ()
{
  create_btree (&tree, NULL, sv_compare);
}

//=================================================================================

static char *new_string (string str)
{
  char *s;
  s = malloc (strlen(str)+1);
  if (s == null)
    fatal_out_of_memory_error ("new_string() failed");
  strcpy (s, str);
  return s;
}

//=================================================================================

static bool library_unit_already_loaded (char *library_unit)
{
  DATA d, d0;
  int  rc, dummy;

  d.library = new_string (library_unit);
  d.source = new_string ("");

  memcpy (&d0, &d, sizeof(d));

  rc = retrieve_btree (&tree, &d, sizeof(d), sizeof(d), &dummy, BT_EQUAL_OR_LARGER);

  free (d0.library);
  free (d0.source);

  if (rc == BT_KEY_NOT_FOUND || strcmp (d.library, library_unit) != 0)
    return false;

  return true;
}

//=================================================================================

static void store_unit (string library, string source, NODE_INFO *info, string library_filename)
{
  DATA u;
  int  rc;

  u.library = new_string (library);
  u.source  = new_string (source);
  memcpy (&u.info, info, sizeof(NODE_INFO));
  u.library_filename = new_string (library_filename);

  rc = insert_btree (&tree, &u, sizeof(u));
  if (rc < 0)
    fatal_compiler_error ("store_unit(1)", &token.pos);
}

//=================================================================================

static void load_library_unit (char *library_unit)
{
  char   library_filename[2*260];
  int    fd;
  HEADER header;
  char   msg[2*260];
  long   size;
  byte   *b, *ptr;
  char   source[260];
  NODE_INFO *n;

  if (library_unit_already_loaded (library_unit))
    return;

  // search for library unit in paths

  find_library (library_unit,
                library_filename);   // out

  fd = open (library_filename, O_RDONLY|O_BINARY);
  if (fd < 0)
  {
    sprintf (msg, "cannot open library %s", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  if (read (fd, &header, sizeof(header)) != sizeof(header))
  {
    sprintf (msg, "cannot read library %s", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  if (header.magic != MAGIC)
  {
    sprintf (msg, "library %s has bad header", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  if (header.version != LIB_VERSION)
  {
    sprintf (msg, "library %s belongs to another compiler version", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  size = lseek (fd, 0L, SEEK_END);
  if (size < 0)
  {
    sprintf (msg, "cannot seek in library %s", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  if (lseek (fd, header.offset_dir, SEEK_SET) < 0)
  {
    sprintf (msg, "cannot seek in library %s", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  if (header.offset_dir > (unsigned long)size)
  {
    sprintf (msg, "library %s has bad offset information", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  size -= header.offset_dir;

  b = malloc (size);
  if (b == null)
    fatal_out_of_memory_error ("cannot load library");

  if (read (fd, b, size) != size)
  {
    free (b);
    close (fd);
    sprintf (msg, "cannot read library %s", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  close (fd);

  ptr = b;

  while (ptr < &b[size])
  {
    if (ptr + sizeof(NODE_INFO) > &b[size])
    {
      free (b);
      close (fd);
      sprintf (msg, "library %s is truncated", library_filename);
      fatal_compiler_error (msg, &token.pos);
    }

    n = (NODE_INFO *)ptr;
    ptr += sizeof(NODE_INFO);

    if (n->unit_name_length >= 260 || ptr + n->unit_name_length > &b[size])
    {
      free (b);
      close (fd);
      sprintf (msg, "library %s has bad data", library_filename);
      fatal_compiler_error (msg, &token.pos);
    }

    sprintf (source, "%.*s", n->unit_name_length, ptr);
    ptr += n->unit_name_length;

    store_unit (library_unit, source, n, library_filename);
  }

  if (ptr != &b[size])
  {
    free (b);
    close (fd);
    sprintf (msg, "library %s is truncated", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  free (b);
  close (fd);
}

//=================================================================================

static int retrieve_unit (string library, string source,
                          NODE_INFO *info, string library_filename)  // out
{
  DATA d, d0;
  int  rc, dummy;

  d.library = new_string (library);
  d.source  = new_string (source);

  memcpy (&d0, &d, sizeof(d));

  rc = retrieve_btree (&tree, &d, sizeof(d), sizeof(d), &dummy, BT_EQUAL);

  free (d0.library);
  free (d0.source);

  if (rc != 0 && rc != BT_KEY_NOT_FOUND)
    fatal_compiler_error ("retrieve_unit(1)", &token.pos);

  if (rc == BT_KEY_NOT_FOUND)
    return -1;

  memcpy (info, &d.info, sizeof(NODE_INFO));
  strcpy (library_filename, d.library_filename);

  return 0;
}

//=================================================================================

typedef struct {
  byte *in;
  byte *out;
  int  in_index;
  int  out_index;
  int  in_size;
  int  out_size;
} ZIP_INFO;

static int my_read (ZIP_INFO *u,
                    unsigned char *buffer,
                    unsigned int  size)
{
  int actual_size = min ((int)size, u->in_size - u->in_index);
  memcpy (buffer, &u->in[u->in_index], actual_size);
  u->in_index += actual_size;
  return actual_size;
}

static int my_write (ZIP_INFO *u,
                     unsigned char *buffer,
                     unsigned int  size)
{
  int actual_size = min ((int)size, u->out_size - u->out_index);
  memcpy (&u->out[u->out_index], buffer, actual_size);
  u->out_index += actual_size;
  return actual_size;
}

//=================================================================================

// returns -1 if unit was not found in library

int load_library_source (string library, string source_name,
                         byte **source, int *source_size)
{
  NODE_INFO info;
  char      library_filename[2*260];
  int       fd;
  byte      *a, *b;
  ZIP_INFO  z;
  char      msg[2*260];

  load_library_unit (library);

  if (retrieve_unit (library, source_name, &info, library_filename) < 0)
    return -1;   // source_name not found (note: can be optional)

  if (stricmp (library, "std") == 0 && stricmp (source_name, "/console.h") == 0)
    is_console_app = true;    // set to true if console.h is loaded from std

  fd = open (library_filename, O_RDONLY|O_BINARY);
  if (fd < 0)
  {
    sprintf (msg, "cannot open library '%s'", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  if (lseek (fd, info.offset_in_file, SEEK_SET) < 0)
  {
    sprintf (msg, "cannot seek in library '%s'", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  a = malloc (info.packed_size);
  b = malloc (info.unpacked_size);
  if (a == null || b == null)
  {
    sprintf (msg, "bad data in library '%s'", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  if (read (fd, a, info.packed_size) != info.packed_size)
  {
    sprintf (msg, "cannot read library '%s'", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  close (fd);

  z.in = a;
  z.out = b;
  z.in_index = 0;
  z.out_index = 0;
  z.in_size = info.packed_size;
  z.out_size = info.unpacked_size;

  if (unpack (&z, my_read, my_write, NULL) < 0)
  {
    sprintf (msg, "cannot retrieve data from library '%s'", library_filename);
    fatal_compiler_error (msg, &token.pos);
  }

  free (a);

  *source = b;
  *source_size = info.unpacked_size;

  return 0;
}

//=================================================================================
