
// makelib.c : create packed library file

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>

#include "files.h"
#include "pack.h"

#include "makelib.h"

/**************************************************************************/

static int out, in;
static int out_offset;

/**************************************************************************/

typedef struct {
  int in;
  int out;
  int count_in;
  int count_out;
} USER_INFO;

/**************************************************************************/

typedef struct node {
  NODE_INFO   i;
  char        *unit_name;
  struct node *next;
} NODE;

static NODE *first, *last;

/**************************************************************************/

static char *new_string (char *name)
{
  char *p;
  p = malloc (strlen(name)+1);
  if (p == NULL)
  {
    printf ("out of memory\n");
    exit (-1);
  }

  strcpy (p, name);
  return p;
}

/**************************************************************************/

static void insert_node (NODE_INFO *n, char *unit_name)
{
  NODE *p;
  p = malloc (sizeof(NODE));
  if (p == NULL)
  {
    printf ("out of memory\n");
    exit (-1);
  }
  memcpy (&p->i, n, sizeof(NODE_INFO));
  p->i.unit_name_length = strlen(unit_name);
  p->unit_name = new_string (unit_name);
  p->next = NULL;

  if (first == NULL)
    first = p;
  else
    last->next = p;
  last = p;
}

/**************************************************************************/

static int write_header (unsigned long offset_dir)
{
  HEADER h;
  memset (&h, 0x00, sizeof(h));
  h.magic = MAGIC;
  h.version = LIB_VERSION;
  h.offset_dir = offset_dir;
  if (write (out, (char *)&h, sizeof(h)) != (int)sizeof(h))
    return -1;
  return 0;
}

/**************************************************************************/

static int writeout_directory_table ()
{
  NODE *p = first;

  while (p != NULL)
  {
    if (write (out, &p->i, sizeof(NODE_INFO)) != sizeof(NODE_INFO))
      return -1;
    if (write (out, p->unit_name, p->i.unit_name_length) != p->i.unit_name_length)
      return -1;
    p = p->next;
  }

  return 0;  
}

/**************************************************************************/

static int my_read (USER_INFO *user_info, unsigned char *buffer, unsigned int size)
{
  int rc;
  rc = read (user_info->in, buffer, size);
  user_info->count_in += rc;
  return rc;
}

/***********************************************************************/

static int my_write (USER_INFO *user_info, unsigned char *buffer, unsigned int size)
{
  int rc;
  rc = write (user_info->out, buffer, size);
  user_info->count_out += rc;
  return rc;
}

/***********************************************************************/

static void import (char *filename, char *unit_name)
{
  int       rc;
  USER_INFO info;
  NODE_INFO n;

  printf ("%-30.30s   ->  %-30.30s\n", filename, unit_name);

  in = open (filename, O_RDONLY|O_BINARY);
  if (in < 0)
  {
    printf ("error: cannot open %s\n", filename);
    exit (-1);
  }

  info.in = in;
  info.out = out;
  info.count_in = 0;
  info.count_out = 0;

  rc = pack (&info, my_read, my_write);
  if (rc < 0)
  {
    printf ("error: cannot treat %s (error %d)\n", filename, rc);
    exit (-1);
  }

  close (in);

//  printf ("(%d -> %d bytes)\n", info.count_in, info.count_out);

  n.unit_name_length = 0;
  n.offset_in_file = out_offset;
  n.packed_size = info.count_out;
  n.compression = 0;
  n.unpacked_size = info.count_in;

  out_offset += info.count_out;

  insert_node (&n, unit_name);
}

/**************************************************************************/

static void extract (char *pathname, char *dir)
{
  FILE_INFO info;
  char      path[260];
  int       rc;

  sprintf (path, "%s*.*", pathname);
  rc = open_directory (&info, path);
  while (rc == 0)
  {
    if (info.type == TYPE_REGULAR_FILE)
    {
      char local_name[260], lib_name[260];
      sprintf (local_name, "%s%s", pathname, info.name);
      sprintf (lib_name,   "%s%s", dir, info.name);
      import (local_name, lib_name);
    }
    else if (info.type == TYPE_DIRECTORY &&
             strcmp (info.name, ".") != 0 && strcmp (info.name, "..") != 0)
    {
      char local_dir[260], lib_dir[260];
      sprintf (local_dir, "%s%s/", pathname, info.name);
      sprintf (lib_dir,   "%s%s/", dir, info.name);
      extract (local_dir, lib_dir);
    }

    rc = read_directory (&info);
  }

  close_directory (&info);
}

/**************************************************************************/

int main (int argc, char *argv[])
{
  char libfilename[260], pathname[260];

  if (argc != 3)
  {
    printf ("makelib  <newfile>.lib  <path/>\n");
    return -1;
  }

  sprintf (libfilename, "%.200s", argv[1]);
  strcat (libfilename, ".lib");

  sprintf (pathname, "%.200s", argv[2]);
  if (pathname[strlen(pathname)-1] != '\\' && pathname[strlen(pathname)-1] != '/')
  {
    printf ("error: pathname must end with \\ or /\n");
    return -1;
  }

  out = open (libfilename, O_CREAT|O_TRUNC|O_RDWR|O_BINARY, 0660);
  if (out < 0)
  {
    printf ("error: cannot create %s\n", libfilename);
    return -1;
  }

  if (write_header (0) < 0)
  {
    printf ("error: cannot write to %s\n", libfilename);
    return -1;
  }

  out_offset = sizeof(HEADER);
  extract (pathname, "/");

  if (lseek (out, 0, SEEK_SET) < 0)
  {
    printf ("error: cannot seek in %s\n", libfilename);
    return -1;
  }

  if (write_header (out_offset) < 0)
  {
    printf ("error: cannot write to %s\n", libfilename);
    return -1;
  }

  if (lseek (out, out_offset, SEEK_SET) < 0)
  {
    printf ("error: cannot seek in %s\n", libfilename);
    return -1;
  }

  if (writeout_directory_table() < 0)
  {
    printf ("error: cannot write to %s\n", libfilename);
    return -1;
  }

  if (close (out) < 0)
  {
    printf ("error: cannot close %s\n", libfilename);
    return -1;
  }

  printf ("ok (%s created)\n", libfilename);
  return 0;
}

/**************************************************************************/
