
// mk.c : Safe-C compiler, linker, and make utility.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <direct.h>
#include <io.h>
#include <fcntl.h>

#include "files.h"
#include "tracing.h"

#include "token.h"
#include "lex.h"
#include "goptions.h"
#include "gsymb.h"
#include "glibpaths.h"
#include "strings.h"

#include "entities.h"
#include "unit.h"
#include "libunit.h"

#include "error.h"
#include "package.h"
#include "codgen.h"

#include "pe.h"

#include "makelib.h"


#define DEFAULT_CONFIG_FILE   "mk.cfg"

static char config_filename[260];

//=================================================================================

static void get_current_dir (char *current_dir, int size)   // always ends with '/'
{
  char c;

  if (_getcwd (current_dir, size) == NULL)  // buffer is too small
  {
    printf ("fatal error: cannot read current directory\n");
    exit (-1);
  }

  if (current_dir[0] != '\0')  // make sure it ends with a '/'
  {
    c = current_dir[strlen(current_dir)-1];
    if (c != '\\' && c != '/')
      strcat (current_dir, "/");
  }

  normalize_pathname (current_dir);
}

//=================================================================================

static void display_errors (bool final)
{
  if (nb_of_compilation_errors() > 0)
  {
    printf ("%d error%s", nb_of_compilation_errors(),  nb_of_compilation_errors() == 1 ? "" : "s");

    if (nb_of_compilation_warnings() > 0)
      printf (", %d warning%s", nb_of_compilation_warnings(),  nb_of_compilation_warnings() == 1 ? "" : "s");

    printf ("\n");
    exit (-1);
  }

  if (final)
  {
    printf ("ok");
    if (nb_of_compilation_warnings() > 0)
      printf (" (%d warning%s)", nb_of_compilation_warnings(),  nb_of_compilation_warnings() == 1 ? "" : "s");
    printf ("\n");
  }
}

//=================================================================================

static int delete_file_and_confirm (char *filename)
{
  int fd;
  fd = open (filename, O_RDONLY|O_BINARY);
  if (fd < 0)
    return 0;
  close (fd);
  if (unlink (filename) < 0)
    return -1;
  return 0;
}

//=================================================================================

void print_usage ()
{
  printf ("\n");
  printf ("Safe-C Compiler v1.%d (%s)\n", LIB_VERSION, __DATE__);
  printf ("usage: mk source.c [cfg=%s] [listing] [crc]\n", DEFAULT_CONFIG_FILE);
  printf ("\n");
}

//=================================================================================

int main (int argc, char *argv[])
{
  char current_dir[260], filename[260], absolute_filename[260];
  char exec_filename[260], debug_filename[260], res_filename[260];
  int  nb_args = argc;
  int  i;
  bool option_crc;

  g_tracing = false;
  option_crc = 0;
  strcpy (config_filename, DEFAULT_CONFIG_FILE);

  for (i=2; i<nb_args; i++)
  {
    char *a = argv[i];

    if (strcmp (a, "listing") == 0)
      g_tracing = true;
    else if (strcmp (a, "crc") == 0)
      option_crc = true;
    else if (strnicmp (a, "cfg=", 4) == 0)
      sprintf (config_filename, "%.200s", &a[4]);
    else
    {
      print_usage ();
      return -1;
    }
  }

  if (g_tracing)
  {
    unlink ("listing.txt");
    open_trace ("listing.txt", 0, 0, 256*1024*1024);
  }

  if (nb_args < 2)  // no filename
  {
    print_usage ();
    return -1;
  }

  sprintf (filename, "%.255s", argv[1]);

  if (strlen(filename) < 2 || stricmp (&filename[strlen(filename)-2], ".c") != 0)
    strcat (filename, ".c");

  get_current_dir (current_dir, sizeof(current_dir));
  if (expand_pathname (current_dir, filename, absolute_filename, sizeof(absolute_filename)-5) < 0)
  {
    printf ("fatal error: cannot create pathname for '%s'\n", filename);
    return -1;
  }

#ifdef WIN32
  to_lower_case (absolute_filename);
#endif

  {
    int i, j;  // i included, j excluded

    i = strlen(absolute_filename);
    while (i-1 >= 0 && absolute_filename[i-1] != '/' && absolute_filename[i-1] != '\\')
      i--;
    // i points on character just after last slash

    j = strlen(absolute_filename);
    while (j-1 >= 0 && absolute_filename[j-1] != '/' && absolute_filename[j-1] != '\\' &&
           absolute_filename[j-1] != '.')
      j--;
    // j points on character just after last slash or last dot

    sprintf (exec_filename,  "%.*s%s", j-i, &absolute_filename[i], "exe");
    sprintf (res_filename,   "%.*s%s", j-i, &absolute_filename[i], "rc");
    sprintf (debug_filename, "%.*s%s", j-i, &absolute_filename[i], "dbg");
  }

  // deleting any previous .exe and .dbg
  if (delete_file_and_confirm (exec_filename) < 0)
  {
    printf ("error: cannot create %s\n", exec_filename);
    return -1;
  }
  if (delete_file_and_confirm (debug_filename) < 0)
  {
    printf ("error: cannot create %s\n", debug_filename);
    return -1;
  }

  load_global_options       (config_filename);  // MUST BE FIRST to load address_size
  load_global_symbols       (config_filename);  // MUST BE SECOND because uses address_size
  load_global_library_paths (config_filename);

  init_constant_pool ();          // do this before creating predefined entities
  create_predefined_entities ();  // this creates also the root region


  init_unit_tree ();
  init_library_unit_tree ();

  if (compile_main_unit (absolute_filename) != 0)
    return -1;

  display_errors (false);


  // instantiate all generic package bodies

  instantiate_all_pending_package_bodies ();

  display_errors (false);


  // perform function inlining

// $


  // parse resource file

  parse_resource_file (current_dir, res_filename);
  display_errors (false);


  // generate code

  generate_code (exec_filename, res_filename, debug_filename, option_crc);

  display_errors (true);

  return 0;
}

//=================================================================================
