
// overlap.c : check for overlapping memory zones

#include "types.h"
#include "token.h"
#include "entities.h"
#include "type.h"
#include "error.h"

#include "overlap.h"

/********************************************************************************************************/

typedef enum {
  OVL_CONST,
  OVL_GLOBAL,
  OVL_LOCAL,
  OVL_PARAM_VALUE,  // simple type of mode IN
  OVL_PARAM_ADDR,   // simple type of mode OUT/REF, or array/struct/union of any mode.
  OVL_ACCESS,
  OVL_UNSAFE,
} OVERLAP_CATEGORY;

typedef enum {
  OVL_NO,        // don't overlap
  OVL_YES,       // overlap
  OVL_IF_ID,     // overlap if entities are identical
  OVL_ERROR,     // should not occur (bad object form)
} OVL_TYP;

static const OVL_TYP overlap_table[1+(int)OVL_UNSAFE][1+(int)OVL_UNSAFE] =
//              constant   global     local      par-by-value  par-by-addr   access     unsafe
/* cte    */  {{OVL_ERROR, OVL_ERROR, OVL_ERROR, OVL_ERROR,    OVL_ERROR,    OVL_ERROR, OVL_ERROR},
/* global */   {OVL_NO,    OVL_IF_ID, OVL_NO,    OVL_NO,       OVL_YES,      OVL_NO,    OVL_YES},
/* local  */   {OVL_NO,    OVL_NO,    OVL_IF_ID, OVL_NO,       OVL_NO,       OVL_NO,    OVL_YES},
/* par-by-v */ {OVL_ERROR, OVL_ERROR, OVL_ERROR, OVL_ERROR,    OVL_ERROR,    OVL_ERROR, OVL_ERROR},
/* par-by-a */ {OVL_NO,    OVL_YES,   OVL_NO,    OVL_NO,       OVL_YES,      OVL_YES,   OVL_YES},
/* access */   {OVL_NO,    OVL_NO,    OVL_NO,    OVL_NO,       OVL_YES,      OVL_YES,   OVL_YES},
/* unsafe */   {OVL_NO,    OVL_YES,   OVL_YES,   OVL_YES,      OVL_YES,      OVL_YES,   OVL_YES}};

/********************************************************************************************************/

typedef struct {
  OVL_TYP typ;
  PENTITY e;
} OBJECT_OVL_INFO;

/********************************************************************************************************/

static void evaluate_object_ovl (PEXPRESSION obj, OBJECT_OVL_INFO *o)
{
  PEXPRESSION prefix;

  o->e = null;

  switch (obj->kind)
  {
    case A_CONST_ENUMERATION_VALUE:
    case A_CONST_INTEGER_VALUE:
    case A_CONST_FLOAT_VALUE:
    case A_CONST_NULL_VALUE:
    case A_POOL_CONSTANT:
      o->typ = OVL_CONST;
      o->e   = null;
      break;

    case A_GLOBAL_VARIABLE_OBJECT:
      o->typ = OVL_GLOBAL;
      o->e   = obj->info.global_variable_object_info.pobject;
      break;

    case A_LOCAL_VARIABLE_OBJECT:
      o->typ = OVL_LOCAL;
      o->e   = obj->info.local_variable_object_info.pobject;
      break;

    case A_REFERENCE_OBJECT:
      evaluate_object_ovl (obj->info.reference_object_info.pobject->info->the_reference.name, o);
      break;

    case A_PARAMETER_OBJECT:
      {
        PENTITY e, typ;

        e = obj->info.parameter_object_info.pobject;
        typ = complete_type_of (e->info->the_parameter.type);

        if (e->info->the_parameter.mode != MODE_IN ||
            typ->kind == AN_OPEN_ARRAY_TYPE || typ->kind == AN_ARRAY_TYPE ||
            typ->kind == A_STRUCT_TYPE      || typ->kind == A_CONSTRAINED_STRUCT_TYPE ||
            typ->kind == A_UNION_TYPE)
        {
          o->typ = OVL_PARAM_ADDR;
        }
        else
        {
          o->typ = OVL_PARAM_VALUE;
        }
      }
      break;

    case AN_ARRAY_ELEMENT_OBJECT:
      prefix = obj->info.array_element_object_info.prefix;
      if (prefix->base_type_or_null == null || prefix->base_type_or_null->kind == AN_UNSAFE_POINTER_TYPE)
        o->typ = OVL_UNSAFE;
      else
        evaluate_object_ovl (prefix, o);
      break;

    case AN_ARRAY_SLICE_OBJECT:
      prefix = obj->info.array_slice_object_info.prefix;
      if (prefix->base_type_or_null == null || prefix->base_type_or_null->kind == AN_UNSAFE_POINTER_TYPE)
        o->typ = OVL_UNSAFE;
      else
        evaluate_object_ovl (prefix, o);
      break;

    case A_STRUCT_FIELD_OBJECT:
      evaluate_object_ovl (obj->info.struct_field_object_info.prefix, o);
      break;

    case A_DEREFERENCED_OBJECT:
      o->typ = OVL_ACCESS;
      break;

    case AN_UNSAFE_DEREFERENCED_OBJECT:
      o->typ = OVL_UNSAFE;
      break;

    case AN_ATTR_BYTE_OBJECT:
      evaluate_object_ovl (obj->info.attr_byte_object_info.prefix, o);
      break;

    case A_BOXED_OBJECT:
      evaluate_object_ovl (obj->info.boxed_object_info.parameter, o);
      break;

    case AN_UNBOXED_OBJECT:
      evaluate_object_ovl (obj->info.unboxed_object_info.parameter, o);
      break;

    default:
      code_generator_error ("evaluate_object_ovl(9)");
      break;
  }
}

/********************************************************************************************************/

static bool object_overlaps_with_object_exp (OBJECT_OVL_INFO *o, PEXPRESSION exp)
{
  OBJECT_OVL_INFO o2;

  evaluate_object_ovl (exp, &o2);

  switch (overlap_table[o->typ][o2.typ])
  {
    case OVL_NO:
      return false;

    case OVL_YES:
      return true;

    case OVL_IF_ID:
      return o->e == o2.e;

    case OVL_ERROR:
      code_generator_error ("object_overlaps_with_object_exp(8)");
      return true;     // to avoid warning

    default:
      code_generator_error ("object_overlaps_with_object_exp(9)");
      return true;     // to avoid warning
  }
}

/********************************************************************************************************/

static bool object_overlaps_with_exp (OBJECT_OVL_INFO *o,
                                      PEXPRESSION     exp,
                                      bool            evaluate_simple_inner_expressions);

/********************************************************************************************************/

static bool object_overlaps_with_exp_list (OBJECT_OVL_INFO *o, LIST_OF_EXPRESSIONS *list)
{
  LIST_OF_EXPRESSIONS *l;

  l = list;

  while (l != null)
  {
    if (object_overlaps_with_exp (o, l->exp, true))
      return true;

    l = l->next;
  }

  return false;
}

/********************************************************************************************************/

static bool object_overlaps_with_exp (OBJECT_OVL_INFO *o,
                                      PEXPRESSION     exp,
                                      bool            evaluate_simple_inner_expressions)
{
  switch (exp->kind)
  {
    case A_CONST_ENUMERATION_VALUE:
    case A_CONST_INTEGER_VALUE:
    case A_CONST_FLOAT_VALUE:
    case A_CONST_NULL_VALUE:
    case A_POOL_CONSTANT:
      return false;

    case AN_OPERATOR_VALUE:
      {
        int i;

        if (evaluate_simple_inner_expressions || exp->info.operator_value_info.op == OP_CONDITIONAL_TEST)
        {
          for (i=0; i<3; i++)
          {
            if (exp->info.operator_value_info.arg[i] != null &&
                object_overlaps_with_exp (o, exp->info.operator_value_info.arg[i], evaluate_simple_inner_expressions))
              return true;
          }
        }
      }
      return false;

    case A_RUN_CALL:
      return evaluate_simple_inner_expressions &&
             object_overlaps_with_exp (o, exp->info.run_call_info.function_call, true);

    case A_FUNCTION_VALUE:
      return false;

    case A_FUNCTION_CALL:
      return evaluate_simple_inner_expressions &&
             object_overlaps_with_exp (o, exp->info.function_call_info.func, true) ||
             object_overlaps_with_exp_list (o, exp->info.function_call_info.param);

    case A_DISCRIMINANT_VALUE:
      return evaluate_simple_inner_expressions &&
             object_overlaps_with_exp (o, exp->info.discriminant_value_info.prefix, true);

    case AN_UNC_ARRAY_AGGREGATE:
      return false;  // assume this never overlaps

    case AN_AGGREGATE_VALUE:
      return evaluate_simple_inner_expressions &&
             object_overlaps_with_exp_list (o, exp->info.aggregate_value_info.list);

    case A_QUALIFIED_EXPRESSION:
      return exp->info.qualified_expression_info.value != null &&
             object_overlaps_with_exp (o, exp->info.qualified_expression_info.value, evaluate_simple_inner_expressions);

    case AN_ARRAY_QUALIFIED_EXPRESSION:
      return (evaluate_simple_inner_expressions && object_overlaps_with_exp (o, exp->info.array_qualified_expression_info.length, true)) ||
             (exp->info.array_qualified_expression_info.value != null &&
              object_overlaps_with_exp (o, exp->info.array_qualified_expression_info.value, evaluate_simple_inner_expressions));

    case A_STRUCT_QUALIFIED_EXPRESSION:
      return (evaluate_simple_inner_expressions && object_overlaps_with_exp (o, exp->info.struct_qualified_expression_info.discriminant, true)) ||
             (exp->info.struct_qualified_expression_info.value != null &&
              object_overlaps_with_exp (o, exp->info.struct_qualified_expression_info.value, evaluate_simple_inner_expressions));

    case AN_ALLOCATOR:
      return evaluate_simple_inner_expressions &&
             object_overlaps_with_exp (o, exp->info.allocator_info.value, true);

    case A_GLOBAL_VARIABLE_OBJECT:
    case A_LOCAL_VARIABLE_OBJECT:
    case A_REFERENCE_OBJECT:
    case A_PARAMETER_OBJECT:
      return object_overlaps_with_object_exp (o, exp);

    case AN_ARRAY_ELEMENT_OBJECT:
      return object_overlaps_with_object_exp (o, exp) ||
             object_overlaps_with_exp (o, exp->info.array_element_object_info.prefix, evaluate_simple_inner_expressions) ||
             (evaluate_simple_inner_expressions && object_overlaps_with_exp (o, exp->info.array_element_object_info.index, true));

    case AN_ARRAY_SLICE_OBJECT:
      return object_overlaps_with_object_exp (o, exp) ||
             object_overlaps_with_exp (o, exp->info.array_slice_object_info.prefix, evaluate_simple_inner_expressions) ||
             (evaluate_simple_inner_expressions && object_overlaps_with_exp (o, exp->info.array_slice_object_info.index, true)) ||
             (evaluate_simple_inner_expressions && object_overlaps_with_exp (o, exp->info.array_slice_object_info.length, true));

    case A_STRUCT_FIELD_OBJECT:
      return object_overlaps_with_exp (o, exp->info.struct_field_object_info.prefix, evaluate_simple_inner_expressions);

    case A_DEREFERENCED_OBJECT:
      return object_overlaps_with_object_exp (o, exp) ||
             object_overlaps_with_exp (o, exp->info.dereferenced_object_info.prefix, evaluate_simple_inner_expressions);

    case AN_UNSAFE_DEREFERENCED_OBJECT:
      return object_overlaps_with_object_exp (o, exp) ||
             object_overlaps_with_exp (o, exp->info.unsafe_dereferenced_object_info.unsafe_ptr_value, evaluate_simple_inner_expressions);

    case AN_ATTR_BYTE_OBJECT:
      return object_overlaps_with_exp (o, exp->info.attr_byte_object_info.prefix, evaluate_simple_inner_expressions);

    case A_BOXED_OBJECT:
      return object_overlaps_with_exp (o, exp->info.boxed_object_info.parameter, evaluate_simple_inner_expressions);

    case AN_UNBOXED_OBJECT:
      return object_overlaps_with_exp (o, exp->info.unboxed_object_info.parameter, evaluate_simple_inner_expressions);

    case A_BOXED_ARRAY_OBJECT:
      return object_overlaps_with_exp_list (o, exp->info.boxed_array_object_info.list);

    default:
      code_generator_error ("object_overlaps_with_exp(9)");
      return true;   // to avoid warning
  }
}

/********************************************************************************************************/

// use to test if an assignment can be done in any memory order

bool object_shares_memory_with_expression (PEXPRESSION obj, PEXPRESSION exp)
{
  OBJECT_OVL_INFO o;

  evaluate_object_ovl (obj, &o);

  return object_overlaps_with_exp (&o, exp, false);
}

/********************************************************************************************************/

// used to test if an aggregate can be evaluated "inline"

bool object_used_in_expression (PEXPRESSION obj, PEXPRESSION exp)
{
  OBJECT_OVL_INFO o;

  evaluate_object_ovl (obj, &o);

  return object_overlaps_with_exp (&o, exp, true);
}

/********************************************************************************************************/
