
// package.c : package declaration and body

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "strings.h"
#include "token.h"
#include "lex.h"
#include "entities.h"
#include "decl.h"
#include "error.h"
#include "type.h"
#include "typedecl.h"
#include "complete.h"

#include "package.h"

//=================================================================================

typedef struct {
  long    nr;
  PENTITY e;
} TRANSLATE_ENTRY;

typedef struct instantiation_data {      // data for a generic instantiation
  PENTITY        egen_package_decl;
  PENTITY        epackage_decl;
  ENTITY_LIST    *list;   // list of generic package declations surrounding this instantiation.
                          // Rule: if instantiation of A occurs within B's declaration or body, 
                          //       and instantiation of B occurs within A's declaration or body,
                          //       then this is disallowed (no circularity)
  BINARY_TREE    tree;
  struct instantiation_data *next;
} INSTANTIATION_DATA;

static INSTANTIATION_DATA *g_instantiation_data_first, *g_instantiation_data_last;

//=================================================================================

//  generic_association ::=  identifier  "=>"  type_definition
//                        |  identifier  "=>"  function_name

static int parse_generic_association (PENTITY epar)
{
  TEXT_POSITION pos;
  PENTITY       t, f;

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("generic parameter identifier expected", &token.pos);
    return -1;
  }

  if (wstrcmp (token.info._identifier.value, epar->identifier_or_null) != 0)
  {
    syntax_error ("bad identifier", &token.pos);
    return -1;
  }

  get_token();  // skip identifier

  if (token.kind != ARROW)
  {
    syntax_error ("'=>' expected", &token.pos);
    return -1;
  }

  get_token();  // skip arrow

  memcpy (&pos, &token.pos, sizeof(pos));

  if (epar->kind == A_GENERIC_TYPE)
  {
    t = complete_type_of (parse_type_definition (&token.pos, null));

    if (is_jagged_type (t))
      semantic_error ("a jagged type is not allowed", &pos);
    else if (is_open_type (t))
      semantic_error ("an open type is not allowed", &pos);
    else if (t->kind == AN_INCOMPLETE_TYPE)
      semantic_error ("an incomplete type is not allowed", &pos);
    else if (t->kind == AN_OPAQUE_TYPE)
      semantic_error ("an opaque type is not allowed", &pos);
    else if (is_limited_type (t))
      semantic_error ("a limited type is not allowed", &pos);

    epar->info->the_generic_type.actual_type = t;
  }
  else   // a generic function
  {
    f = parse_expanded_name ();
    if (f == null || (f->kind != A_GENERIC_FUNCTION && f->kind != A_FUNCTION_DECLARATION))
      semantic_error ("a function name is expected here", &pos);
    else
    {
      if (f->kind == A_GENERIC_FUNCTION)
      {
        t = f->info->the_generic_function.to_type;
      }
      else
      {
        t = f->info->the_function_declaration.to_type;
      }

      if (!types_are_equal (t, epar->info->the_generic_function.to_type))
        semantic_error ("function profiles do not match", &pos);
    }

    epar->info->the_generic_function.actual_func = f;
  }

  return 0;
}

//=================================================================================

static void unlink_all_generic_parameters (PENTITY epackage)
{
  PENTITY e;

  e = epackage->info->the_package_declaration.generic_part->entities.first;

  while (e != null)
  {
    if (e->kind == A_GENERIC_TYPE)
      e->info->the_generic_type.actual_type = null;

    if (e->kind == A_GENERIC_FUNCTION)
      e->info->the_generic_function.actual_func = null;

    e = e->next;
  }
}

//=================================================================================

static void skip_to_next_id (PENTITY *pe)
{
  while (*pe != null && (*pe)->identifier_or_null == null)
    *pe = (*pe)->next;
}

//=================================================================================

static int compare (void                  *user,
                    const TRANSLATE_ENTRY *data1,
                    const unsigned int    size1,
                    const TRANSLATE_ENTRY *data2,
                    const unsigned int    size2)
{
  if (data1->nr < data2->nr)
    return -1;
  if (data1->nr > data2->nr)
    return +1;
  return 0;
}

//=================================================================================

static void store_translation (INSTANTIATION_DATA *p, long nr, PENTITY e2)
{
  TRANSLATE_ENTRY rec;
  int             rc;

  rec.nr = nr;
  rec.e  = e2;

  rc = insert_btree (&p->tree, &rec, sizeof(rec));
  if (rc != 0)
    fatal_compiler_error ("store_trans()", &token.pos);
}

//=================================================================================

static void translate_instance_entity (INSTANTIATION_DATA *p, PENTITY *e)
{
  TRANSLATE_ENTRY rec;
  int             rc;
  unsigned int    dummy;

  if (*e == null)    // null entity
    return;

  rec.nr = (*e)->nr;

  rc = retrieve_btree (&p->tree, &rec, sizeof(rec), sizeof(rec), &dummy, BT_EQUAL);
  if (rc == 0)
  {
    (*e) = rec.e;
  }
  else if (rc == BT_KEY_NOT_FOUND)
    ;
  else
    fatal_compiler_error ("trans_inst_ent()", &token.pos);
}

//=================================================================================

static void instantiate_entity (INSTANTIATION_DATA *p, PENTITY e, REGION *region);
static void translate_instance_expression (INSTANTIATION_DATA *p, PEXPRESSION *exp);

//=================================================================================

static void translate_instance_region (INSTANTIATION_DATA *p, PREGION *rr)
{
  PENTITY     e = (*rr)->entities.first;
  REGION      *r;
  ENTITY_LIST *list;

  r = new_region ();

  while (e != null)
  {
    instantiate_entity (p, e, r);
    e = e->next;
  }


  // translate list of visible packages

  r->packages = duplicate_entity_list ((*rr)->packages);
  list = r->packages;
  while (list != null)
  {
    translate_instance_entity (p, &list->e);
    list = list->next;
  }


  *rr = r;
}

//=================================================================================

static void translate_flow_alternatives (INSTANTIATION_DATA *p, FLOW_ALTERNATIVE **first_alt,
                                                                FLOW_ALTERNATIVE **last_alt)
{
  FLOW_ALTERNATIVE *a, *b, *n, *l;

  a = *first_alt;
  b = null;
  l = null;

  while (a != null)
  {
    n = malloc (sizeof(FLOW_ALTERNATIVE));
    if (n == null)
      fatal_out_of_memory_error ("tr_flow_alt()");

    memcpy (n, a, sizeof(FLOW_ALTERNATIVE));
    n->cte_list = duplicate_case_constant_list (a->cte_list);
    translate_instance_region (p, &n->inner);
    n->next = null;

    if (l == null)
      b = n;
    else
      l->next = n;
    l = n;

    a = a->next;
  }

  *first_alt = b;
  *last_alt  = l;
}

//=================================================================================

static void translate_instance_list_of_expression (INSTANTIATION_DATA *p, LIST_OF_EXPRESSIONS **list)
{
  LIST_OF_EXPRESSIONS *l;

  l = duplicate_list_of_expressions (*list);

  (*list) = l;

  while (l != null)
  {
    translate_instance_expression (p, &l->exp);
    translate_instance_entity     (p, &l->type);
    translate_instance_entity     (p, &l->e);
    l = l->next;
  }
}

//=================================================================================

static void translate_instance_expression (INSTANTIATION_DATA *p, PEXPRESSION *exp)
{
  PEXPRESSION e0, e;

  if (*exp == null)
    return;

  e0 = *exp;
  e = malloc (sizeof(EXPRESSION));
  if (e == null)
    fatal_out_of_memory_error ("trans_inst_exp()");
  memcpy (e, e0, sizeof(EXPRESSION));

  translate_instance_entity (p, &e->base_type_or_null);

  // when instantiating an expression of a generic type and it becomes an expression of array or struct type,
  // we need to fill the expression's constraint field and recompute the base type.
  if (complete_type_of (e0->base_type_or_null)->kind == A_GENERIC_TYPE)
  {
    PENTITY  t = complete_type_of (e->base_type_or_null);
    uint4    value;

    if (t->kind == AN_ARRAY_TYPE || t->kind == A_CONSTRAINED_STRUCT_TYPE)
    {
      if (t->kind == AN_ARRAY_TYPE)
      {
        value = t->info->the_array_type.length;
      }
      else
      {
        value = t->info->the_constrained_struct_type.discriminant_value;
      }

      e->constraint.kind = CONSTANT_CONSTRAINT;
      e->constraint.value = value;
      e->base_type_or_null = base_type_of (t);
    }
  }

  switch (e->kind)
  {
    case A_CONST_ENUMERATION_VALUE:
    case A_CONST_INTEGER_VALUE:
    case A_CONST_FLOAT_VALUE:
    case A_CONST_NULL_VALUE:
    case A_POOL_CONSTANT:
      break;

    case AN_OPERATOR_VALUE:
      translate_instance_expression (p, &e->info.operator_value_info.arg[0]);
      translate_instance_expression (p, &e->info.operator_value_info.arg[1]);
      translate_instance_expression (p, &e->info.operator_value_info.arg[2]);
      break;

    case A_RUN_CALL:
      translate_instance_expression (p, &e->info.run_call_info.function_call);
      break;

    case A_FUNCTION_VALUE:
      translate_instance_entity     (p, &e->info.function_value_info.to_function_declaration_or_generic_function);
      break;

    case A_FUNCTION_CALL:
      translate_instance_expression (p, &e->info.function_call_info.func);
      translate_instance_list_of_expression (p, &e->info.function_call_info.param);
      break;

    case A_DISCRIMINANT_VALUE:
      translate_instance_expression (p, &e->info.discriminant_value_info.prefix);
      translate_instance_entity     (p, &e->info.discriminant_value_info.field);
      break;

    case AN_UNC_ARRAY_AGGREGATE:
      translate_instance_expression (p, &e->info.unc_array_aggregate_info.element);
      break;

    case AN_AGGREGATE_VALUE:
      translate_instance_list_of_expression (p, &e->info.aggregate_value_info.list);
      break;

    case A_QUALIFIED_EXPRESSION:
      translate_instance_expression (p, &e->info.qualified_expression_info.value);
      break;

    case AN_ARRAY_QUALIFIED_EXPRESSION:
      translate_instance_expression (p, &e->info.array_qualified_expression_info.length);
      translate_instance_expression (p, &e->info.array_qualified_expression_info.value);
      break;

    case A_STRUCT_QUALIFIED_EXPRESSION:
      translate_instance_expression (p, &e->info.struct_qualified_expression_info.discriminant);
      translate_instance_expression (p, &e->info.struct_qualified_expression_info.value);
      break;

    case AN_ALLOCATOR:
      translate_instance_expression (p, &e->info.allocator_info.value);
      break;

    case A_GLOBAL_VARIABLE_OBJECT:
      translate_instance_entity     (p, &e->info.global_variable_object_info.pobject);
      break;

    case A_LOCAL_VARIABLE_OBJECT:
      translate_instance_entity     (p, &e->info.local_variable_object_info.pobject);
      break;

    case A_REFERENCE_OBJECT:
      translate_instance_entity     (p, &e->info.reference_object_info.pobject);
      break;

    case A_PARAMETER_OBJECT:
      translate_instance_entity     (p, &e->info.parameter_object_info.pobject);
      break;

    case AN_ARRAY_ELEMENT_OBJECT:
      translate_instance_expression (p, &e->info.array_element_object_info.prefix);
      translate_instance_expression (p, &e->info.array_element_object_info.index);
      break;

    case AN_ARRAY_SLICE_OBJECT:
      translate_instance_expression (p, &e->info.array_slice_object_info.prefix);
      translate_instance_expression (p, &e->info.array_slice_object_info.index);
      translate_instance_expression (p, &e->info.array_slice_object_info.length);
      break;

    case A_STRUCT_FIELD_OBJECT:
      translate_instance_expression (p, &e->info.struct_field_object_info.prefix);
      translate_instance_entity     (p, &e->info.struct_field_object_info.field);
      break;

    case A_DEREFERENCED_OBJECT:
      translate_instance_expression (p, &e->info.dereferenced_object_info.prefix);
      break;

    case AN_UNSAFE_DEREFERENCED_OBJECT:
      translate_instance_expression (p, &e->info.unsafe_dereferenced_object_info.unsafe_ptr_value);
      break;

    case AN_ATTR_BYTE_OBJECT:
      translate_instance_expression (p, &e->info.attr_byte_object_info.prefix);
      break;

    case A_BOXED_OBJECT:
      translate_instance_expression (p, &e->info.boxed_object_info.parameter);
      break;

    case AN_UNBOXED_OBJECT:
      translate_instance_expression (p, &e->info.unboxed_object_info.parameter);
      break;

    case A_BOXED_ARRAY_OBJECT:
      translate_instance_list_of_expression (p, &e->info.boxed_array_object_info.list);
      break;
  }

  *exp = e;
}

//=================================================================================

static void instantiate_entity (INSTANTIATION_DATA *p, PENTITY e0, REGION *region)
{
  PENTITY e2;

  e2 = append_new_entity_for_instantiation
                (e0->kind,
                 e0->info,
                 e0->identifier_or_null != null ? e0->identifier_or_null : L"",
                 &token.pos,
                 region);

  store_translation (p, e0->nr, e2);


  // translate inner field : entity links in generic package range, regions, expressions

  switch (e0->kind)
  {
    case AN_INTEGER_TYPE:
    case A_FLOAT_TYPE:
    case AN_ENUMERATION_TYPE:
      break;

    case AN_OPEN_ARRAY_TYPE:
      translate_instance_entity (p, &e2->info->the_open_array_type.element);
      break;

    case AN_ARRAY_TYPE:
      translate_instance_entity (p, &e2->info->the_array_type.open_array);
      break;

    case A_STRUCT_TYPE:
      translate_instance_region (p, &e2->info->the_struct_type.discriminant);
      translate_instance_region (p, &e2->info->the_struct_type.fields);
      break;

    case A_CONSTRAINED_STRUCT_TYPE:
      translate_instance_entity (p, &e2->info->the_constrained_struct_type.open_struct);
      break;

    case A_UNION_TYPE:
      translate_instance_region (p, &e2->info->the_union_type.fields);
      break;

    case A_POINTER_TYPE:
      translate_instance_entity (p, &e2->info->the_pointer_type.designated_type);
      break;

    case A_FUNCTION_POINTER_TYPE:
      if (e2->info->the_function_pointer_type.extern_dll_or_null != null)
      {
        e2->info->the_function_pointer_type.extern_dll_or_null =
                malloc (2*wstrlen(e2->info->the_function_pointer_type.extern_dll_or_null) + 2);
        wstrcpy (e2->info->the_function_pointer_type.extern_dll_or_null,
                 e0->info->the_function_pointer_type.extern_dll_or_null);
      }

      translate_instance_entity (p, &e2->info->the_function_pointer_type.return_type);
      translate_instance_region (p, &e2->info->the_function_pointer_type.parameters);
      break;

    case AN_UNSAFE_POINTER_TYPE:
      translate_instance_entity (p, &e2->info->the_unsafe_pointer_type.designated_type);
      break;

    case A_RENAMED_TYPE:
      translate_instance_entity (p, &e2->info->the_renamed_type.actual_type);
      break;

    case AN_INCOMPLETE_TYPE:
      e2->info->the_incomplete_type.full_type = null;
      break;

    case AN_OPAQUE_TYPE:
      e2->info->the_opaque_type.full_type = null;
      break;

    case A_GENERIC_TYPE:
      translate_instance_entity (p, &e2->info->the_generic_type.actual_type);
      break;

    case A_NULL_POINTER_TYPE:
    case A_VOID_TYPE:
      break;

    case AN_INCOMPLETE_TYPE_COMPLETITION:
      translate_instance_entity (p, &e2->info->the_incomplete_type_completition.incomplete_type);
      translate_instance_entity (p, &e2->info->the_incomplete_type_completition.full_type);
      e2->info->the_incomplete_type_completition.incomplete_type
        ->info->the_incomplete_type.full_type =
              e2->info->the_incomplete_type_completition.full_type;
      break;

    case AN_OPAQUE_TYPE_COMPLETITION:
      translate_instance_entity (p, &e2->info->the_opaque_type_completition.opaque_type);
      translate_instance_entity (p, &e2->info->the_opaque_type_completition.full_type);
      e2->info->the_opaque_type_completition.opaque_type
        ->info->the_opaque_type.full_type =
              e2->info->the_opaque_type_completition.full_type;
      break;

    case A_FIELD:
      translate_instance_entity (p, &e2->info->the_field.type);
      break;

    case A_VARYING_FIELD:
      translate_instance_entity (p, &e2->info->the_varying_field.type);
      break;

    case AN_ENUMERATION_LITERAL:
      translate_instance_entity (p, &e2->info->the_enumeration_literal.type);
      break;

    case A_CONSTANT:
      translate_instance_entity     (p, &e2->info->the_constant.type);
      translate_instance_expression (p, &e2->info->the_constant.value);
      break;

    case A_GLOBAL_VARIABLE:
      translate_instance_entity     (p, &e2->info->the_global_variable.type);
      translate_instance_expression (p, &e2->info->the_global_variable.initial_value_or_null);
      break;

    case A_LOCAL_VARIABLE:
      translate_instance_entity     (p, &e2->info->the_local_variable.type);
      translate_instance_expression (p, &e2->info->the_local_variable.initial_value_or_null);
      break;

    case A_REFERENCE:
      translate_instance_entity     (p, &e2->info->the_reference.type);
      translate_instance_expression (p, &e2->info->the_reference.name);
      break;

    case A_PARAMETER:
      translate_instance_entity     (p, &e2->info->the_parameter.type);
      translate_instance_expression (p, &e2->info->the_parameter.default_value_or_null);
      break;

    case A_GENERIC_FUNCTION:
      translate_instance_entity (p, &e2->info->the_generic_function.to_type);
      translate_instance_entity (p, &e2->info->the_generic_function.actual_func);
      break;

    case A_FUNCTION_DECLARATION:
      translate_instance_entity (p, &e2->info->the_function_declaration.to_type);
      e2->info->the_function_declaration.to_function_body_or_null = null;
      break;

    case A_FUNCTION_BODY:
      translate_instance_entity (p, &e2->info->the_function_body.to_function_declaration);
      translate_instance_region (p, &e2->info->the_function_body.inner);
      e2->info->the_function_body.to_function_declaration
        ->info->the_function_declaration.to_function_body_or_null = e2;
      break;

    case A_PACKAGE_DECLARATION:
      e2->info->the_package_declaration.to_package_body_or_null = null;
      translate_instance_region (p, &e2->info->the_package_declaration.generic_part);
      translate_instance_region (p, &e2->info->the_package_declaration.declarations);
      break;

    case A_PACKAGE_BODY:
      translate_instance_entity (p, &e2->info->the_package_body.to_package_declaration);
      translate_instance_region (p, &e2->info->the_package_body.declarations);
      e2->info->the_package_body.to_package_declaration
        ->info->the_package_declaration.to_package_body_or_null = e2;
      break;

    case A_CLEAR_STATEMENT:
      translate_instance_expression (p, &e2->info->the_clear_statement.name);
      break;

    case AN_ASSIGNMENT_STATEMENT:
      translate_instance_expression (p, &e2->info->the_assignment_statement.name);
      translate_instance_expression (p, &e2->info->the_assignment_statement.value);
      break;

    case A_PRE_OR_POSTFIX_STATEMENT:
      translate_instance_expression (p, &e2->info->the_pre_or_postfix_statement.name);
      break;

    case A_FUNCTION_CALL_STATEMENT:
      translate_instance_expression (p, &e2->info->the_function_call_statement.name);
      break;

    case A_RETURN_STATEMENT:
      translate_instance_expression (p, &e2->info->the_return_statement.value);
      translate_instance_entity     (p, &e2->info->the_return_statement.type);
      translate_instance_entity     (p, &e2->info->the_return_statement.outer);
      break;

    case A_BREAK_STATEMENT:
      translate_instance_entity     (p, &e2->info->the_break_statement.outer);
      break;

    case A_CONTINUE_STATEMENT:
      translate_instance_entity     (p, &e2->info->the_continue_statement.outer);
      break;

    case A_FREE_STATEMENT:
      translate_instance_expression (p, &e2->info->the_free_statement.value);
      break;

    case AN_ABORT_STATEMENT:
      break;

    case AN_ASSERT_STATEMENT:
      translate_instance_expression (p, &e2->info->the_assert_statement.value);
      break;

    case A_SLEEP_STATEMENT:
      translate_instance_expression (p, &e2->info->the_sleep_statement.value);
      break;

    case A_CODE_STATEMENT:
      break;

    case AN_UNUSED_STATEMENT:
      translate_instance_entity     (p, &e2->info->the_unused_statement.item);
      break;

    case A_BLOCK_STATEMENT:
      translate_instance_region     (p, &e2->info->the_block_statement.inner);
      translate_instance_entity     (p, &e2->info->the_block_statement.outer);
      break;

    case AN_IF_STATEMENT:
      translate_instance_expression (p, &e2->info->the_if_statement.condition);
      translate_instance_region     (p, &e2->info->the_if_statement.true_branch);
      translate_instance_region     (p, &e2->info->the_if_statement.false_branch);
      translate_instance_entity     (p, &e2->info->the_if_statement.outer);
      break;

    case A_SWITCH_STATEMENT:
      translate_instance_expression (p, &e2->info->the_switch_statement.value);
      translate_flow_alternatives   (p, &e2->info->the_switch_statement.first_alt,
                                        &e2->info->the_switch_statement.last_alt);
      translate_instance_entity     (p, &e2->info->the_switch_statement.outer);
      break;

    case A_WHILE_STATEMENT:
      translate_instance_expression (p, &e2->info->the_while_statement.condition);
      translate_instance_region     (p, &e2->info->the_while_statement.inner);
      translate_instance_entity     (p, &e2->info->the_while_statement.outer);
      break;

    case A_FOR_STATEMENT:
      translate_instance_region     (p, &e2->info->the_for_statement.pre);
      translate_instance_region     (p, &e2->info->the_for_statement.exp_region);
      translate_instance_expression (p, &e2->info->the_for_statement.condition);
      translate_instance_region     (p, &e2->info->the_for_statement.post);
      translate_instance_region     (p, &e2->info->the_for_statement.inner);
      translate_instance_entity     (p, &e2->info->the_while_statement.outer);
      break;

    default:
      fatal_compiler_error ("inst.e", &token.pos);
      break;
  }
}

//=================================================================================

static void free_entity_list (ENTITY_LIST *list)
{
  ENTITY_LIST *p, *t;

  p = list;
  while (p != null)
  {
    t = p;
    p = p->next;
    free (t);
  }
}

//=================================================================================

static void free_inst_node (INSTANTIATION_DATA *p)
{
  free_entity_list (p->list);
  close_btree (&p->tree);
  free (p);
}

//=================================================================================

//             gen-unit  pkg-body-to-fill  surrounding-gen-package-bodies
// -> 1 add_body_inst (gen SWAP, place *1, tree, (SORT*A), (EXTRACT*B))
// -> 2 add_body_inst (gen SWAP, place *2, tree, (SORT*A))
// -> 3 add_body_inst (gen SORT, place *3, tree, *)
// ----> pick a line(3) to instantiate, but no other line must have this entity its
// surrounding-gen-package-bodies(1) !  otherwise pick this line(1) first;
// if nb of tries exceeds nb of table lines -> circular dependancies !!

void instantiate_all_pending_package_bodies ()
{
  INSTANTIATION_DATA *p, *l, *q;
  ENTITY_LIST        *el;

  // instantiate all remaining bodies

  while (g_instantiation_data_first != null)
  {
    l = null;
    p = g_instantiation_data_first;

    // find an instantiation where no other line has it as its dependancies

    while (p != null)         // loop on all p instantiations
    {
      q = g_instantiation_data_first;
      while (q != null)          // loop on all q instantiations
      {
        // check if any dependances equals p
        el = q->list;
        while (el != null)
        {
          if (el->e == p->egen_package_decl)
            break;    // dependance found
          el = el->next;
        }

        if (el != null)   // dependancy found
          break;

        q = q->next;
      }

      if (q == null)   // not dependancy found for p
        break;

// printf ("dependancy found between %S and %S\n",  p->egen_package_decl->identifier_or_null, q->egen_package_decl->identifier_or_null);

      // p has dependancies : try next p
      l = p;
      p = p->next;
    }

    if (p != null)  // found line p
    {
      PENTITY egbody, ebody;

      // instantiate package body of entry p

// printf ("instantiate %S\n", p->egen_package_decl->identifier_or_null);

      egbody = p->egen_package_decl->info->the_package_declaration.to_package_body_or_null;
      
      if (egbody != null)
      {
        ebody = p->epackage_decl->info->the_package_declaration.to_package_body_or_null;

        free (ebody->info->the_package_body.declarations);  // free old dummy region
        ebody->info->the_package_body.declarations =
              egbody->info->the_package_body.declarations;  // set copy of generic body

        translate_instance_region (p, &ebody->info->the_package_body.declarations);
      }


      // remove p from list and free it

      if (l == null)   // first item
      {
        g_instantiation_data_first = p->next;
        if (g_instantiation_data_first == null)
          g_instantiation_data_last = null;
      }
      else
      {
        if (g_instantiation_data_last == p)
          g_instantiation_data_last = l;
        l->next = p->next;
      }

      free_inst_node (p);
    }
    else
    {
      char buffer[240];
      sprintf (buffer, "circular references instantiating generic package body %S",
                       g_instantiation_data_first->egen_package_decl->identifier_or_null);
      semantic_error (buffer, &token.pos);
      return;
    }
  }
}

//=================================================================================

//  package_instantiation ::=
//     "package"  identifier  "="  "new"  generic_package_name
//    ["(" generic_association {"," generic_association} ")"]
//     ";"

static void parse_package_instantiation ()
{
  TEXT_POSITION new_pos;
  wchar         new_package_id[MAX_IDENTIFIER_LENGTH+1];
  PENTITY       e_gpackage, epar;

  // store new id
  memcpy (&new_pos, &token.pos, sizeof(new_pos));
  wstrcpy (new_package_id, token.info._identifier.value);

  get_token();  // skip id
  get_token();  // skip =

  if (token.kind != TOKEN_new)
  {
    syntax_error ("keyword 'new' expected", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }

  get_token();  // skip 'new'


  // parse a generic package name

  e_gpackage = parse_expanded_name ();
  if (e_gpackage == null ||
      e_gpackage->kind != A_PACKAGE_DECLARATION ||
      !e_gpackage->info->the_package_declaration.is_generic)
  {
    syntax_error ("generic package name expected", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }


  // check that we're NOT within the generic package or its body

  if (are_we_in_package (e_gpackage))   // we're inside the generic package or its body
  {
    syntax_error ("a generic instantiation is not allowed within the package itself", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }

  epar = e_gpackage->info->the_package_declaration.generic_part->entities.first;
  skip_to_next_id (&epar);

  if (token.kind == LEFT_PARENTHESIS)
  {
    get_token();  // skip '('

    for (;;)
    {
      if (epar == null)
      {
        semantic_error ("too many generic parameters", &token.pos);
        skip_until (SEMICOLON, true);
        unlink_all_generic_parameters (e_gpackage);
        return;
      }

      if (parse_generic_association (epar) < 0)
      {
        skip_until (SEMICOLON, true);
        unlink_all_generic_parameters (e_gpackage);
        return;
      }

      if (epar != null)
      {
        epar = epar->next;
        skip_to_next_id (&epar);
      }

      if (token.kind == RIGHT_PARENTHESIS)
        break;

      if (token.kind != COMMA)
      {
        syntax_error ("',' or ')' expected", &token.pos);
        skip_until (SEMICOLON, true);
        unlink_all_generic_parameters (e_gpackage);
        return;
      }

      get_token();   // skip comma
    }
    
    get_token();  // skip ')'
  }


  // check that all generic entities have received an association

  if (epar != null)
  {
    syntax_error ("missing generic parameters", &token.pos);
    skip_until (SEMICOLON, true);
    unlink_all_generic_parameters (e_gpackage);
    return;
  }


  parse_semicolon ();


  {
    A_PACKAGE_DECLARATION_ENTITY decl;
    A_PACKAGE_BODY_ENTITY        body;
    PENTITY                      edecl, e, ebody;
    INSTANTIATION_DATA           *p;

    memset (&decl, 0x00, sizeof(decl));
    decl.to_package_body_or_null = null;
    decl.is_generic              = false;
    decl.is_body_required        = false;
    decl.generic_part            = e_gpackage->info->the_package_declaration.generic_part;
    decl.declarations            = e_gpackage->info->the_package_declaration.declarations;
    edecl = append_new_entity (A_PACKAGE_DECLARATION, &decl, new_package_id, &new_pos);


    // compute instantiation data

    p = malloc (sizeof(INSTANTIATION_DATA));
    if (p == null)
      fatal_out_of_memory_error ("inst_data");
    memset (p, 0x00, sizeof(INSTANTIATION_DATA));

    p->egen_package_decl = e_gpackage;
    p->epackage_decl = edecl;
    p->list          = build_list_of_surrounding_generic_packages ();
    create_btree (&p->tree, NULL, compare);
    p->next = null;
    
    if (g_instantiation_data_last == null)  // empty
      g_instantiation_data_first = p;
    else
      g_instantiation_data_last->next = p;
    g_instantiation_data_last = p;

    // translate inner regions

    store_translation (p, e_gpackage->nr, edecl);

    translate_instance_region (p, &edecl->info->the_package_declaration.generic_part);
    translate_instance_region (p, &edecl->info->the_package_declaration.declarations);


    // insert package declaration in package list of surrounding scope

    create_region_level ();
    append_package_declaration_to_surrounding_scope_package_list (edecl);
    unlink_region_level ();


    // make opaque types limited

    e = edecl->info->the_package_declaration.declarations->entities.first;
    while (e != null)
    {
      if (e->kind == AN_OPAQUE_TYPE)
      {
        e->info->the_opaque_type.is_limited = true;
      }

      e = e->next;
    }




    // body

    memset (&body, 0x00, sizeof(body));
    body.to_package_declaration = edecl;
    body.declarations = new_region();
    ebody = append_new_entity (A_PACKAGE_BODY, &body, L"", &token.pos);

    edecl->info->the_package_declaration.to_package_body_or_null = ebody;
  }


  unlink_all_generic_parameters (e_gpackage);
}

//=================================================================================

static void parse_end_package (PENTITY edecl)
{
  if (token.kind == TOKEN_end)
  {
    get_token();

    if (token.kind == IDENTIFIER)
    {
      if (wstrcmp (token.info._identifier.value, edecl->identifier_or_null) != 0)
        syntax_error ("identifier does not match", &token.pos);

      get_token();

      parse_semicolon ();
    }
    else
    {
      syntax_error ("identifier is expected here", &token.pos);
      skip_until (SEMICOLON, true);
    }
  }
  else
  {
    syntax_error ("keyword 'end' is expected here", &token.pos);
    skip_until (SEMICOLON, true);
  }
}

//=================================================================================

static void parse_package_declaration ()
{
  A_PACKAGE_DECLARATION_ENTITY decl;
  PENTITY                      edecl, e;

  memset (&decl, 0x00, sizeof(decl));
  decl.to_package_body_or_null = null;
  decl.is_generic              = false;
  decl.is_body_required        = false;
  decl.generic_part            = new_region ();
  decl.declarations            = new_region ();

  edecl = append_new_entity (A_PACKAGE_DECLARATION, &decl, token.info._identifier.value, &token.pos);

  create_region_level ();
  append_region (decl.generic_part);
  append_region (decl.declarations);

  get_token();

  parse_global_declarations (/* is_body => */ false);


  // insert package declaration in package list of surrounding scope

  append_package_declaration_to_surrounding_scope_package_list (edecl);



  // make opaque types limited

  e = edecl->info->the_package_declaration.declarations->entities.first;
  while (e != null)
  {
    if (e->kind == AN_OPAQUE_TYPE)
    {
      e->info->the_opaque_type.is_limited = true;
    }

    e = e->next;
  }


  // tests if a package body is required

  edecl->info->the_package_declaration.is_body_required =
       is_body_required (edecl->info->the_package_declaration.declarations->entities.first);


  unlink_region_level ();


  parse_end_package (edecl);
}

//=================================================================================

static void parse_package_body (TEXT_POSITION *pos0)
{
  PENTITY               edecl, e;
  A_PACKAGE_BODY_ENTITY body;
  PENTITY               ebody;


  edecl = search_entity_in_local_scope (token.info._identifier.value);

  if (edecl == null ||  // no earlier package declaration of this name
      edecl->kind != A_PACKAGE_DECLARATION)
  {
    syntax_error ("a package declaration must occur before the body", pos0);
    return;
  }

  memset (&body, 0x00, sizeof(body));
  body.to_package_declaration = edecl;
  body.declarations = new_region ();

  ebody = append_new_entity (A_PACKAGE_BODY, &body, L"", &token.pos);

  edecl->info->the_package_declaration.to_package_body_or_null = ebody;

  create_region_level ();
  append_region (edecl->info->the_package_declaration.generic_part);
  append_region (edecl->info->the_package_declaration.declarations);
  append_region (body.declarations);

  get_token();   // skip identifier


  // when entering region (package body. or unit body),
  // scan package decl. or unit interface :
  // - for opaque type :
  //   . set non-limited

  e = edecl->info->the_package_declaration.declarations->entities.first;
  while (e != null)
  {
    if (e->kind == AN_OPAQUE_TYPE)
    {
      e->info->the_opaque_type.is_limited = false;
    }

    e = e->next;
  }



  parse_global_declarations (/* is_body => */ true);



  check_global_completion (edecl->info->the_package_declaration.declarations->entities.first);
  check_global_completion (ebody->info->the_package_body.declarations->entities.first);



  // when leaving region (package body. or unit body),
  // scan package decl. or unit interface :
  // - for incomplete type :
  //   . if full type's keynr is past the package body nr,
  //     it means they are in different parts ! -> full type no longer available.
  // - for opaque type :
  //   . full type no longer available.

  e = edecl->info->the_package_declaration.declarations->entities.first;
  while (e != null)
  {
    if (e->kind == AN_INCOMPLETE_TYPE &&
        e->info->the_incomplete_type.full_type != null &&
        e->info->the_incomplete_type.full_type->nr > ebody->nr)
    {
      e->info->the_incomplete_type.is_full_type_visible = false;
    }

    if (e->kind == AN_OPAQUE_TYPE)
    {
      e->info->the_opaque_type.is_full_type_visible = false;
      e->info->the_opaque_type.is_limited = true;
    }

    e = e->next;
  }


  unlink_region_level ();


  parse_end_package (edecl);
}

//=================================================================================

void parse_package_declaration_or_package_instantiation_or_package_body
                (bool body_allowed,
                 bool *body_was_parsed)
{
  TEXT_POSITION pos0;

  memcpy (&pos0, &token.pos, sizeof(pos0));

  get_token();   // skip 'package'

  if (token.kind == IDENTIFIER)
  {
    // check if look-ahead token is "=" in case it's a package instantiation

    get_look_ahead_token ();
    if (look_ahead_token.kind == ASSIGN)
    {
      parse_package_instantiation ();
    }
    else
    {
      parse_package_declaration ();
    }
  }
  else if (token.kind == TOKEN_body)
  {
    if (!body_allowed)
    {
      syntax_error ("a package body is not allowed here", &pos0);
      return;
    }

    get_token();   // skip 'body'

    parse_package_body (&pos0);

    *body_was_parsed = true;
  }
  else
  {
    syntax_error ("an identifier or 'body' is expected here", &token.pos);
  }
}

//=================================================================================

//  ["<" identifier  {","  identifier}  ">"]

static void parse_generic_type_declarations ()
{
  A_GENERIC_TYPE_ENTITY gt;

  if (token.kind != SMALLER)
    return;

  get_token();   // skip <

  for (;;)
  {
    if (token.kind != IDENTIFIER)
    {
      syntax_error ("identifier expected", &token.pos);
      skip_until (LARGER, true);
      return;
    }

    // create a generic type

    clear (gt);
    (void) append_new_entity (A_GENERIC_TYPE, &gt, token.info._identifier.value, &token.pos);


    get_token();   // skip identifier

    if (token.kind == LARGER)
      break;

    if (token.kind != COMMA)
    {
      syntax_error ("',' or '>' expected", &token.pos);
      skip_until (LARGER, true);
      return;
    }

    get_token();   // skip comma
  }

  get_token();   // skip >
}

//=================================================================================

static void parse_generic_function_specifications ()
{
  bool dummy;

  while (token.kind != TOKEN_package && token.kind != LAST_TOKEN)
  {
    parse_function_specification (&token.pos,
                                  null,
                                  /* body_allowed    = */ false,
                                  /* body_was_parsed = */ &dummy,
                                  /* in_typedef      = */ false,
                                  /* in_generic_part = */ true);
  }
}

//=================================================================================

//  generic_package_declaration ::= generic_formal_part
//                                  package_declaration
//
//  generic_formal_part ::=
//     "generic"  ["<" identifier  {","  identifier}  ">"]
//     {  function_specification ";"  }

void parse_generic_package_declaration ()
{
  A_PACKAGE_DECLARATION_ENTITY decl;
  PENTITY                      edecl, e;


  get_token();   // skip 'generic'


  clear (decl);
  decl.to_package_body_or_null = null;
  decl.is_generic              = true;
  decl.is_body_required        = false;
  decl.generic_part            = new_region ();
  decl.declarations            = new_region ();

  edecl = append_new_entity (A_PACKAGE_DECLARATION, &decl, L"", &token.pos);


  create_region_level ();

  append_region (decl.generic_part);

  parse_generic_type_declarations ();
  parse_generic_function_specifications ();

  unlink_region_level ();



  if (token.kind != TOKEN_package)
  {
    syntax_error ("'package' expected", &token.pos);
    return;
  }

  get_token();   // skip 'package'

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("package identifier expected", &token.pos);
    return;
  }

  patch_entity_by_setting_an_identifier (edecl, token.info._identifier.value, &token.pos);

  get_token();  // skip identifier


  create_region_level ();

  append_region (decl.generic_part);
  append_region (decl.declarations);

  parse_global_declarations (/* is_body => */ false);


  // make opaque types limited

  e = edecl->info->the_package_declaration.declarations->entities.first;
  while (e != null)
  {
    if (e->kind == AN_OPAQUE_TYPE)
    {
      e->info->the_opaque_type.is_limited = true;
    }

    e = e->next;
  }


  // tests if a package body is required

  edecl->info->the_package_declaration.is_body_required =
       is_body_required (edecl->info->the_package_declaration.declarations->entities.first);


  unlink_region_level ();

  parse_end_package (edecl);
}

//=================================================================================
