
// pe.c : generate PE structure

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <ctype.h>

#include "calendar.h"
#include "crc.h"
#include "bintree.h"
#include "files.h"
#include "tracing.h"

#include "error.h"
#include "types.h"
#include "goptions.h"
#include "pool.h"
#include "dbginfo.h"

#include "pe.h"

/************************************************************************/

#define PAGE      0x1000       // 4K
#define HEADER    PAGE         // mz_header + pe_header + fillers

/************************************************************************/

typedef struct {
  uint4 addr;      // virtual address
  uint4 size;
} IMAGE_DATA_DIR;

/************************************************************************/

typedef struct {
  char   name[8];
  uint4  size_in_memory;
  uint4  addr;
  uint4  SizeOnDisk;            // multiple of FileAlignment  (0 for .bss section)
  uint4  PointerToDisk;         // ptr into section data in file, relative to optional header (0 for .bss)
  uint4  PointerToRelocations;  // 0
  uint4  PointerToLinenumbers;  // 0
  uint2  NumberOfRelocations;   // 0
  uint2  NumberOfLinenumbers;   // 0
  uint4  Characteristics;       // contains executable code    = 0x00000020,
                                // contains initialized data   = 0x00000040,
                                // contains uninitialized data = 0x00000080,
                                // access rights exec  = 0x20000000,
                                // access rights read  = 0x40000000,
                                // access rights write = 0x80000000.
                                //
                                // .text -> code + exec
                                // .data -> initialized + read
                                // .bss  -> uninitialized + read + write
} SECTION;  // 40 bytes

/************************************************************************/

typedef struct {
  uint4 signature;        // 0x50 0x45 0x00 0x00
  uint2 machineType;      // 0x014C for i386, 0x8664 for x64.
  uint2 nb_sections;      // 3 (text, data, bss) (see at the end of this struct)
  uint4 created_dtf;      // file creation dtf (nb secs since 1970)
  uint4 symbols;          // 0
  uint4 nb_symbols;       // 0
  uint2 opt_header_size;  // 00E0  (= .code - .magic)
  uint2 characteristics;  // 0x0103  (0x0100=32-bit, 0x020 = large addr aware, 0x0002=file valid, 0x0001=no reloc)
} PE32_PART1;

typedef struct {
  uint2 magic;                   // 0x010B=PE32, 0x20B=PE32+)
  byte  MajorLinkerVersion;      // 0x06(32 bit) or 0x0C or 0x0E
  byte  MinorLinkerVersion;      // 0x00
  uint4 SizeOfCode;              // size of .text section
  uint4 SizeOfInitializedData;   // size of .data section
  uint4 SizeOfUninitializedData; // size of .bss section (often zero)
  uint4 EntryPoint;              // RVA of PC start address
  uint4 BaseOfCode;              // 0x00001000  (.text loaded at ImageBase+BaseOfCode = 0x00401000)
  uint4 BaseOfData;              // 0x00005000  (.data loaded at ImageBase+BaseOfData)

  // optional header - windows-specific fields
  uint4 ImageBase;               // 0x00400000  (MS-DOS header loaded at this address)
  uint4 SectionAlignment;        // 0x00001000
  uint4 FileAlignment;           // 0x00001000
  uint2 MajorOSVersion;          // 4 or 5
  uint2 MinorOSVersion;          // 0 or 2
  uint2 MajorImageVersion;       // 0
  uint2 MinorImageVersion;       // 0
  uint2 MajorSubsystemVersion;   // 4 (Win32 4.0) or 5
  uint2 MinorSubsystemVersion;   // 0 or 2
  uint4 Win32VersionValue;       // 0
  uint4 SizeOfImage;             // size of image to be loaded in memory, incl. headers (multiple of SectionAlignment)
  uint4 SizeOfHeaders;           // 0x1000
  uint4 CheckSum;                // 0
  uint2 Subsystem;               // 2=WIN_GUI, 3=DOS_CONSOLE
  uint2 DllCharacteristics;      // 0
  uint4 SizeOfStackReserve;      // 0x00100000 (1 MB stack)
  uint4 SizeOfStackCommit;       // 0x00001000 (4 KB pour dmarrer)
  uint4 SizeOfHeapReserve;       // 0x00100000 (1 MB heap)
  uint4 SizeOfHeapCommit;        // 0x00001000 (4 KB pour dmarrer)
  uint4 LoaderFlags;             // 0
  uint4 NumberOfDirectories;     // 0x10 (16)
} PE32_PART2_32BIT;

/************************************************************************/

typedef struct {
  uint2 magic;                   // 0x010B=PE32, 0x20B=PE32+)
  byte  MajorLinkerVersion;      // 0x06(32 bit) or 0x0C or 0x0E
  byte  MinorLinkerVersion;      // 0x00
  uint4 SizeOfCode;              // size of .text section
  uint4 SizeOfInitializedData;   // size of .data section
  uint4 SizeOfUninitializedData; // size of .bss section (often zero)
  uint4 EntryPoint;              // RVA of PC start address
  uint4 BaseOfCode;              // 0x00001000  (.text loaded at ImageBase+BaseOfCode = 0x00401000)

  // optional header - windows-specific fields
  int8  ImageBase64;             // 0x0140000000  (MS-DOS header loaded at this address)
  uint4 SectionAlignment;        // 0x00001000
  uint4 FileAlignment;           // 0x00001000
  uint2 MajorOSVersion;          // 4 or 5
  uint2 MinorOSVersion;          // 0 or 2
  uint2 MajorImageVersion;       // 0
  uint2 MinorImageVersion;       // 0
  uint2 MajorSubsystemVersion;   // 4 (Win32 4.0) or 5
  uint2 MinorSubsystemVersion;   // 0 or 2
  uint4 Win32VersionValue;       // 0
  uint4 SizeOfImage;             // size of image to be loaded in memory, incl. headers (multiple of SectionAlignment)
  uint4 SizeOfHeaders;           // 0x1000
  uint4 CheckSum;                // 0
  uint2 Subsystem;               // 2=WIN_GUI, 3=DOS_CONSOLE
  uint2 DllCharacteristics;      // 0
  int8 SizeOfStackReserve;       // 0x00100000 (1 MB stack)
  int8 SizeOfStackCommit;        // 0x00001000 (4 KB pour dmarrer)
  int8 SizeOfHeapReserve;        // 0x00100000 (1 MB heap)
  int8 SizeOfHeapCommit;         // 0x00001000 (4 KB pour dmarrer)
  uint4 LoaderFlags;             // 0
  uint4 NumberOfDirectories;     // 0x10 (16)
} PE32_PART2_64BIT;

/************************************************************************/

typedef struct {
  // optional header - data directories (16 x 8 bytes = 128 bytes)
  IMAGE_DATA_DIR  export;
  IMAGE_DATA_DIR  import;
  IMAGE_DATA_DIR  resource;
  IMAGE_DATA_DIR  exception;
  IMAGE_DATA_DIR  certificate;      // must be last section - 1
  IMAGE_DATA_DIR  base_relocation;
  IMAGE_DATA_DIR  debug;            // must be last section
  IMAGE_DATA_DIR  architecture;
  IMAGE_DATA_DIR  global_ptr;
  IMAGE_DATA_DIR  tls;
  IMAGE_DATA_DIR  load_config;
  IMAGE_DATA_DIR  bound_import;
  IMAGE_DATA_DIR  iat;
  IMAGE_DATA_DIR  delay_import_descriptor;
  IMAGE_DATA_DIR  clr_runtime;
  IMAGE_DATA_DIR  reserved;

} PE32_PART3;

/************************************************************************/

typedef struct {

  // section table
  SECTION  code;   // code
  SECTION  data;   // constants pool, switch jump tables, dll import tables.
  SECTION  bss;    // global variables.

} PE32_PART4;

/************************************************************************/

static byte *image;
static uint image_size;
static uint image_index;
static uint ofs_pe32_part1;
static uint ofs_pe32_part2;
static uint ofs_pe32_part3;
static uint ofs_pe32_part4;

static uint4 code_size;   // must be multiple of PAGE !
static uint4 data_size;   // must be multiple of PAGE !

static uint4 image_index_start_code, image_index_end_code;
static uint4 image_index_start_data, image_index_end_data;
static uint4 image_index_start_bss,  image_index_end_bss;

/************************************************************************/

// btree of dll_names, btree of function_names, list of relocations.

typedef struct reloc_info {
  uint4             addr;         // offset in image where offset32 should be stored
  bool              absolute;     // absolute or relative
  struct reloc_info *next_reloc;  // next reloc for this DLL function
  struct reloc_info *next_move;   // next reloc for this generated function (to fix when block of code is moved)
} RELOC_INFO;

typedef struct {
  char       *func_name;
  uint4      vector_addr;   // indirect call vector's address
  uint4      hint_name_addr_fill;
  uint4      func_name_addr_fill;
  RELOC_INFO *first;
  RELOC_INFO *last;
} FUNC_INFO;

typedef struct {
  char        *dll_name;
  uint4       iat_start_addr;
  uint4       lookup_addr_fill;
  uint4       dll_name_fill;
  BINARY_TREE funcs;
} DLL_INFO;

static bool        dll_tree_created;
static BINARY_TREE dll_tree;
static RELOC_INFO  *dll_move_chain;   // in reverse address order

/************************************************************************/

uint4 current_RIP (void)
{
  return LOAD_ADDRESS + image_index;
}

/************************************************************************/

static void probe (uint4 size)
{
  byte  *new_image;
  uint4 new_size;

  if (image_size - image_index < size)   // not enough memory left
  {
    new_size = image_size;
    while (new_size - image_index < size)   // double allocated size until enough space
    {
      if (new_size >= 0x80000000)  // 2GB
        fatal_out_of_memory_error ("probe1");
      new_size <<= 1;
    }

    new_image = malloc (new_size);
    if (new_image == null)
      fatal_out_of_memory_error ("probe2");

    memcpy (new_image, image, image_size);
    memset (&new_image[image_size], 0x00, new_size - image_size);

    free (image);    // free old image

    image      = new_image;
    image_size = new_size;
  }
}

/************************************************************************/

static void align_at (uint4 mod)  // ex: mod = PAGE
{
  probe (mod);
  while ((image_index & (mod-1)) > 0)
    image[image_index++] = 0;
}

/************************************************************************/

static void write_vector (int8 value)
{
  probe (address_size);
  if (address_size == 4)
    *((uint4 *)&image[image_index]) = (uint4)value;
  else
    *((int8 *)&image[image_index]) = value;
  image_index += address_size;
}

/************************************************************************/

void exe_write_byte (int value)
{
  probe (1);
  *((int1 *)&image[image_index]) = (int1)value;
  image_index++;
}

/************************************************************************/

void exe_write_int1 (int value)
{
  probe (1);
  *((int1 *)&image[image_index]) = (int1)value;
  image_index++;
}

/************************************************************************/

void exe_write_int2 (int value)
{
  probe (2);
  *((int2 *)&image[image_index]) = (int2)value;
  image_index += 2;
}

/************************************************************************/

void exe_write_int4 (int4 value)
{
  probe (4);
  *((int4 *)&image[image_index]) = value;
  image_index += 4;
}

/************************************************************************/

void exe_write_int8 (int8 value)
{
  probe (8);
  *((int8 *)&image[image_index]) = value;
  image_index += 8;
}

/************************************************************************/

// called when generating code for new function

void exe_reset_dll_move_chain ()
{
  dll_move_chain = null;
}

/************************************************************************/

// move all dll relocations addresses of a function starting at a given address
// (used when moving chunks of code for longer branch instructions)

static void exe_move_dll_chain (uint4 start_addr, int4 offset)
{
  RELOC_INFO *r = dll_move_chain;
  while (r != null && r->addr + LOAD_ADDRESS >= start_addr)
  {
    r->addr += (uint4)offset;      // move offset in image to fix
    r = r->next_move;
  }
}

/************************************************************************/

static void add_reloc_in_func_node (FUNC_INFO *func_info, bool absolute, uint4 reloc)
{
  RELOC_INFO *n;

  n = malloc (sizeof(RELOC_INFO));
  if (n == null)
    fatal_out_of_memory_error ("add_reloc_in_func_node");

  memset (n, 0x00, sizeof(n));
  n->addr       = reloc;             // offset in image to fix
  n->absolute   = absolute;
  n->next_reloc = null;
  n->next_move  = dll_move_chain;    // link in move chain
  dll_move_chain = n;

  if (func_info->last == null)
    func_info->first = n;
  else
    func_info->last->next_reloc = n;
  func_info->last = n;
}

/************************************************************************/

static int compare_dll (void           *user,
                        const DLL_INFO *a,
                        const uint      size1,
                        const DLL_INFO *b,
                        const uint      size2)
{
  return stricmp (a->dll_name, b->dll_name);  // case insensitive compare
}

/************************************************************************/

static int compare_func (void            *user,
                         const FUNC_INFO *a,
                         const uint      size1,
                         const FUNC_INFO *b,
                         const uint      size2)
{
  return strcmp (a->func_name, b->func_name);  // case sensitive compare
}

/************************************************************************/

static void add_func_in_dll_node (DLL_INFO *dll_info,
                                  char     *func,
                                  bool     absolute,
                                  uint4    reloc)   // offset in image to fix
{
  char         *func_name;
  FUNC_INFO    func_info;
  int          rc;
  unsigned int dummy;

  func_name = malloc (strlen(func)+1);
  if (func_name == null)
    fatal_out_of_memory_error ("add_func_in_dll_node1");
  strcpy (func_name, func);

  clear(func_info);
  func_info.func_name = func_name;

  rc = retrieve_btree (&dll_info->funcs, &func_info, sizeof(func_info), sizeof(func_info), &dummy, BT_EQUAL);

  if (rc != 0 && rc != BT_KEY_NOT_FOUND)
    fatal_compiler_error0 ("add_func_in_dll_node2");

  if (rc == BT_KEY_NOT_FOUND)
  {
    clear(func_info);
    func_info.func_name = func_name;
    func_info.first     = null;
    func_info.last      = null;

    add_reloc_in_func_node (&func_info, absolute, reloc);

    rc = insert_btree (&dll_info->funcs, &func_info, sizeof(func_info));
    if (rc)
      fatal_compiler_error0 ("add_func_in_dll_node3");
  }
  else    // func exists
  {
    free (func_name);

    add_reloc_in_func_node (&func_info, absolute, reloc);

    rc = update_btree (&dll_info->funcs, &func_info, sizeof(func_info));
    if (rc)
      fatal_compiler_error0 ("add_func_in_dll_node4");
  }
}

/************************************************************************/

void exe_dll_reference_written (char *dll, char *func, bool absolute)
{
  uint4        reloc = image_index - 4;
  char         *dll_name;
  DLL_INFO     dll_info;
  int          rc;
  unsigned int dummy;

  if (!dll_tree_created)
  {
    create_btree (&dll_tree, NULL, compare_dll);
    dll_tree_created = true;
  }

  dll_name = malloc (strlen(dll)+1);
  if (dll_name == null)
    fatal_out_of_memory_error ("exe_dll_reference_written(1)");
  strcpy (dll_name, dll);

  clear(dll_info);
  dll_info.dll_name = dll_name;

  rc = retrieve_btree (&dll_tree, &dll_info, sizeof(dll_info), sizeof(dll_info), &dummy, BT_EQUAL);

  if (rc != 0 && rc != BT_KEY_NOT_FOUND)
    fatal_compiler_error0 ("exe_dll_reference_written(2)");

  if (rc == BT_KEY_NOT_FOUND)
  {
    clear(dll_info);
    dll_info.dll_name = dll_name;
    create_btree (&dll_info.funcs, NULL, compare_func);

    add_func_in_dll_node (&dll_info, func, absolute, reloc);

    rc = insert_btree (&dll_tree, &dll_info, sizeof(dll_info));
    if (rc)
      fatal_compiler_error0 ("exe_dll_reference_written(3)");
  }
  else    // dll exists
  {
    free (dll_name);

    add_func_in_dll_node (&dll_info, func, absolute, reloc);

    rc = update_btree (&dll_tree, &dll_info, sizeof(dll_info));
    if (rc)
      fatal_compiler_error0 ("exe_dll_reference_written(4)");
  }
}

/************************************************************************/

// it's possible to add 72 extra bytes to the 128, to have a large header.

static unsigned char mz_header[128] =
{ 0x4D, 0x5A, 0x90, 0x00, 0x03, 0x00, 0x00, 0x00,
  0x04, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00,
  0xB8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, sizeof(mz_header), 0x00, 0x00, 0x00,
  0x0E, 0x1F, 0xBA, 0x0E, 0x00, 0xB4, 0x09, 0xCD,

/* original "This program cannot be run in DOS mode" */
/*
  0x21, 0xB8, 0x01, 0x4C, 0xCD, 0x21, 0x54, 0x68,
  0x69, 0x73, 0x20, 0x70, 0x72, 0x6F, 0x67, 0x72,
  0x61, 0x6D, 0x20, 0x63, 0x61, 0x6E, 0x6E, 0x6F,
  0x74, 0x20, 0x62, 0x65, 0x20, 0x72, 0x75, 0x6E,
  0x20, 0x69, 0x6E, 0x20, 0x44, 0x4F, 0x53, 0x20,
  0x6D, 0x6F, 0x64, 0x65, 0x2E, 0x0D, 0x0D, 0x0A,
*/

/* new : "generated by SAFE-C Compiler for Windows" */
  0x21, 0xB8, 0x01, 0x4C, 0xCD, 0x21, 0x67, 0x65,
  0x6E, 0x65, 0x72, 0x61, 0x74, 0x65, 0x64, 0x20,
  0x62, 0x79, 0x20, 0x53, 0x41, 0x46, 0x45, 0x2D,
  0x43, 0x20, 0x43, 0x6F, 0x6D, 0x70, 0x69, 0x6C,
  0x65, 0x72, 0x20, 0x66, 0x6F, 0x72, 0x20, 0x57,
  0x69, 0x6E, 0x64, 0x6F, 0x77, 0x73, 0x0D, 0x0A,

  0x24, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

/************************************************************************/

void exe_init_image ()
{
  image_size = 2*PAGE;   // at least 2 pages of 4K

  image = malloc (image_size);
  if (image == null)
    fatal_out_of_memory_error ("init_exec_image()");

  memset (image, 0x00, image_size);

  ofs_pe32_part1 = sizeof(mz_header);
  ofs_pe32_part2 = ofs_pe32_part1 + sizeof(PE32_PART1);

  if (address_size == 4)
    ofs_pe32_part3 = ofs_pe32_part2 + sizeof(PE32_PART2_32BIT);
  else
    ofs_pe32_part3 = ofs_pe32_part2 + sizeof(PE32_PART2_64BIT);

  ofs_pe32_part4 = ofs_pe32_part3 + sizeof(PE32_PART3);

  image_index = ofs_pe32_part4 + sizeof(PE32_PART4);

  align_at (PAGE);
  
  image_index_start_code = image_index;
}

/************************************************************************/

static void write_dll_tables_in_exec_image ()
{
  PE32_PART3  *p;
  short        mode1, mode2;
  unsigned int dummy;
  DLL_INFO     dll_info;
  FUNC_INFO    func_info;
  RELOC_INFO   *r;
  int          rc, i;

  if (!dll_tree_created)
    return;

  // 1) write dummy IAT table with all zeroes

  align_at (256);

  p = (PE32_PART3 *)&image[ofs_pe32_part3];
  p->iat.addr = image_index;

  mode1 = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&dll_tree, &dll_info, sizeof(dll_info), sizeof(dll_info), &dummy, mode1);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_dll_tables_in_exec_image1");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    dll_info.iat_start_addr = image_index;   // store start of IAT vectors

    mode2 = BT_FIRST;
    for (;;)
    {
      rc = retrieve_btree (&dll_info.funcs, &func_info, sizeof(func_info), sizeof(func_info), &dummy, mode2);
      if (rc != 0 && rc != BT_KEY_NOT_FOUND)
        fatal_compiler_error0 ("write_dll_tables_in_exec_image2");

      if (rc == BT_KEY_NOT_FOUND)
        break;

      func_info.vector_addr = image_index;

      rc = update_btree (&dll_info.funcs, &func_info, sizeof(func_info));
      if (rc)
        fatal_compiler_error0 ("write_dll_tables_in_exec_image3");

      r = func_info.first;
      while (r != null)
      {
        if (r->absolute)
        {
          *((uint4 *)&image[r->addr]) = LOAD_ADDRESS + image_index;
        }
        else
        {
          *((uint4 *)&image[r->addr]) = LOAD_ADDRESS + image_index - (r->addr + LOAD_ADDRESS + 4);
        }

        r = r->next_reloc;
      }

      write_vector (0);  // add a dummy zero vector

      mode2 = BT_LARGER;
    }

    rc = update_btree (&dll_tree, &dll_info, sizeof(dll_info));
    if (rc)
      fatal_compiler_error0 ("write_dll_tables_in_exec_image4");

    write_vector (0);  // add a trailing zero vector

    mode1 = BT_LARGER;
  }

  p = (PE32_PART3 *)&image[ofs_pe32_part3];
  p->iat.size = image_index - p->iat.addr;



  // 2) write IMPORT DIRECTORY ENTRIES

  p = (PE32_PART3 *)&image[ofs_pe32_part3];
  p->import.addr = image_index;


  mode1 = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&dll_tree, &dll_info, sizeof(dll_info), sizeof(dll_info), &dummy, mode1);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_dll_tables_in_exec_image5");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    dll_info.lookup_addr_fill = image_index;

    exe_write_int4 (0);  // lookup table
    exe_write_int4 (0);    // dtf
    exe_write_int4 (-1);   // forwarder

    dll_info.dll_name_fill = image_index;
    exe_write_int4 (0);  // dll name

    exe_write_int4 (dll_info.iat_start_addr);  // iat address

    rc = update_btree (&dll_tree, &dll_info, sizeof(dll_info));
    if (rc)
      fatal_compiler_error0 ("write_dll_tables_in_exec_image6");

    mode1 = BT_LARGER;
  }

  // write trailing entry with all zeroes

  exe_write_int4  (0);  // lookup table
  exe_write_int4  (0);  // dtf
  exe_write_int4  (0);  // forwarder
  exe_write_int4  (0);  // dll name
  exe_write_int4  (0);  // iat address




  // 3) write LOOKUP table

  align_at (address_size);

  mode1 = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&dll_tree, &dll_info, sizeof(dll_info), sizeof(dll_info), &dummy, mode1);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_dll_tables_in_exec_image7");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    *((uint4 *)&image[dll_info.lookup_addr_fill]) = image_index;

    mode2 = BT_FIRST;
    for (;;)
    {
      rc = retrieve_btree (&dll_info.funcs, &func_info, sizeof(func_info), sizeof(func_info), &dummy, mode2);
      if (rc != 0 && rc != BT_KEY_NOT_FOUND)
        fatal_compiler_error0 ("write_dll_tables_in_exec_image8");

      if (rc == BT_KEY_NOT_FOUND)
        break;

      func_info.hint_name_addr_fill = image_index;
      write_vector (0);  // lookup table

      rc = update_btree (&dll_info.funcs, &func_info, sizeof(func_info));
      if (rc)
        fatal_compiler_error0 ("write_dll_tables_in_exec_image9");

      mode2 = BT_LARGER;
    }

    rc = update_btree (&dll_tree, &dll_info, sizeof(dll_info));
    if (rc)
      fatal_compiler_error0 ("write_dll_tables_in_exec_image10");

    write_vector (0);  // add a trailing zero vector

    mode1 = BT_LARGER;
  }


  // 4) write dll names + hint/name tables

  mode1 = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&dll_tree, &dll_info, sizeof(dll_info), sizeof(dll_info), &dummy, mode1);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_dll_tables_in_exec_image11");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    align_at (2);
    *((uint4 *)&image[dll_info.dll_name_fill]) = image_index;

    for (i=0; i<(int)strlen(dll_info.dll_name); i++)
      exe_write_int1 (dll_info.dll_name[i]);
    exe_write_int1 (0);   // trailing zero

    align_at (2);

    mode2 = BT_FIRST;
    for (;;)
    {
      rc = retrieve_btree (&dll_info.funcs, &func_info, sizeof(func_info), sizeof(func_info), &dummy, mode2);
      if (rc != 0 && rc != BT_KEY_NOT_FOUND)
        fatal_compiler_error0 ("write_dll_tables_in_exec_image12");

      if (rc == BT_KEY_NOT_FOUND)
        break;

      align_at (2);

      *((uint4 *)&image[func_info.hint_name_addr_fill]) = image_index;
      *((uint4 *)&image[func_info.vector_addr]) = image_index;

      exe_write_int2 (0);  // ordinal always zero

      // ascii string of function name
      for (i=0; i<(int)strlen(func_info.func_name); i++)
        exe_write_int1 (func_info.func_name[i]);
      exe_write_int1 (0);   // trailing zero

      align_at (2);

      mode2 = BT_LARGER;
    }

    mode1 = BT_LARGER;
  }

  p = (PE32_PART3 *)&image[ofs_pe32_part3];
  p->import.size = image_index - p->import.addr;
}

/************************************************************************/

// func label relocation (filled at end-of-program)

/* a tree */

typedef struct {
  int        func_label_nr;
  uint4      func_addr;      // RIP of function
} RELOC_FUNC_NODE;

static bool         func_reloc_tree_created;
static BINARY_TREE  func_reloc_tree;


/* a list of addresses to backfill, in decreasing backfill order */

typedef struct reloc_func_info {
  uint4      backfill_addr;    // RIP of 4-byte address to fill
  int        func_label_nr;
  bool       absolute;         // absolute or relative
  struct reloc_func_info *next;
} RELOC_FUNC_INFO;

static RELOC_FUNC_INFO *func_reloc_list;   // in reverse backfill address order


/************************************************************************/

static int func_reloc_tree_compare (void                  *user,
                                    const RELOC_FUNC_NODE *a,
                                    const uint            size1,
                                    const RELOC_FUNC_NODE *b,
                                    const uint            size2)
{
  if (a->func_label_nr < b->func_label_nr)
    return -1;
  if (a->func_label_nr > b->func_label_nr)
    return +1;
  return 0;
}

/************************************************************************/

void exe_add_func_label (int func_label_nr)
{
  RELOC_FUNC_NODE n;
  int             rc;

  if (g_tracing)
    trace ("    declare entry func #%d\n", func_label_nr);
  
  if (!func_reloc_tree_created)
  {
    create_btree (&func_reloc_tree, NULL, func_reloc_tree_compare);
    func_reloc_tree_created = true;
  }

  clear(n);
  n.func_label_nr = func_label_nr;
  n.func_addr     = current_RIP();

  rc = insert_btree (&func_reloc_tree, &n, sizeof(n));
  if (rc)
    fatal_compiler_error0 ("exe_add_func_label(1)");
}

/************************************************************************/

void exe_func_backfill_addr4_written (int func_label_nr, bool absolute)
{
  RELOC_FUNC_INFO *n;

  n = malloc (sizeof(RELOC_FUNC_INFO));
  if (n == null)
    fatal_compiler_error0 ("exe_write_func_backfill_addr4(1)");
  n->backfill_addr = current_RIP() - 4;
  n->func_label_nr = func_label_nr;
  n->absolute      = absolute;
  n->next = func_reloc_list;

  func_reloc_list = n;
}

/************************************************************************/

static void move_all_func_backfills (uint4 start_addr, int4 offset)
{
  RELOC_FUNC_INFO *n = func_reloc_list;

  while (n != null && n->backfill_addr >= start_addr)
  {
    n->backfill_addr += offset;
    n = n->next;
  }
}

/************************************************************************/

static void fill_all_func_backfills ()
{
  RELOC_FUNC_INFO *n = func_reloc_list;
  int             rc;
  RELOC_FUNC_NODE nd;
  unsigned int    dummy;
  uint4           target_address;

  while (n != null)
  {
    nd.func_label_nr = n->func_label_nr;

    rc = retrieve_btree (&func_reloc_tree, &nd, sizeof(nd), sizeof(nd), &dummy, BT_EQUAL);
    if (rc != 0)
      fatal_compiler_error0 ("fill_all_func_backfills()");

    target_address = nd.func_addr;

    if (n->absolute)
    {
      *((uint4 *)&image[n->backfill_addr - LOAD_ADDRESS]) = target_address;
    }
    else
    {
      *((uint4 *)&image[n->backfill_addr - LOAD_ADDRESS]) = target_address - (n->backfill_addr + 4);
    }

    n = n->next;
  }
}

/************************************************************************/
/************************************************************************/
/************************************************************************/

/* a list of addresses to backfill, in decreasing backfill order */

typedef struct reloc_global_info {
  uint4      backfill_addr;    // RIP of 4-byte address to fill
  int        global_address;   // global variable's address, relative to start of bss segment
  bool       absolute;         // absolute or relative
  struct reloc_global_info *next;
} RELOC_GLOBAL_INFO;

static RELOC_GLOBAL_INFO *global_reloc_list;   // in reverse backfill address order

/************************************************************************/

void exe_global_backfill_addr4_written (int relative_global_addr, bool absolute)
{
  RELOC_GLOBAL_INFO *n;

  n = malloc (sizeof(RELOC_GLOBAL_INFO));
  if (n == null)
    fatal_compiler_error0 ("exe_global_backfill_addr4_written(1)");
  n->backfill_addr = current_RIP() - 4;
  n->global_address = relative_global_addr;
  n->absolute      = absolute;
  n->next = global_reloc_list;

  global_reloc_list = n;
}

/************************************************************************/

static void move_all_global_backfills (uint4 start_addr, int4 offset)
{
  RELOC_GLOBAL_INFO *n = global_reloc_list;

  while (n != null && n->backfill_addr >= start_addr)
  {
    n->backfill_addr += offset;
    n = n->next;
  }
}

/************************************************************************/

// to be called once bss segment's address is fixed

static void fill_all_global_backfills ()
{
  RELOC_GLOBAL_INFO *n = global_reloc_list;
  uint4             target;

  while (n != null)
  {
    target = LOAD_ADDRESS + image_index + n->global_address;
    if (n->absolute)
    {
      *((uint4 *)&image[n->backfill_addr - LOAD_ADDRESS]) = target;
    }
    else
    {
      *((uint4 *)&image[n->backfill_addr - LOAD_ADDRESS]) = target - (n->backfill_addr + 4);
    }

    n = n->next;
  }
}

/************************************************************************/
/************************************************************************/
/************************************************************************/

/* a list of jump table constants to fix (convert each label_nr -> code_address) */

typedef struct reloc_jumptable_info {
  int8                        serial_nr;
  struct reloc_jumptable_info *next;
} RELOC_JUMPTABLE_INFO;

static RELOC_JUMPTABLE_INFO *reloc_jumptable_list;

/************************************************************************/

// to be called for each jump table
void exe_mark_jump_table (int8 pool_nr)
{
  RELOC_JUMPTABLE_INFO *n;

  n = malloc (sizeof(RELOC_JUMPTABLE_INFO));
  if (n == null)
    fatal_compiler_error0 ("exe_mark_jump_table(1)");
  n->serial_nr = pool_nr;
  n->next      = reloc_jumptable_list;

  reloc_jumptable_list = n;
}

/************************************************************************/
/************************************************************************/
/************************************************************************/

static uint4 *pnear_labels;
static int   nb_near_labels;

/************************************************************************/

/* a list of near addresses to backfill, in decreasing backfill order */

typedef struct reloc_near_info {
  uint4      backfill_addr;    // RIP of 1-byte or 4-byte address to fill
  int        label_nr;         // target label nr
  bool       conditional;
  bool       was_expanded;
  struct reloc_near_info *next;
} RELOC_NEAR_INFO;

static RELOC_NEAR_INFO *reloc_near_list;   // in reverse backfill address order

/************************************************************************/

// to be called after p-code for a function was generated,
// before generating asm.

void exe_allocate_near_label_table (int nb_labels)
{
  free (pnear_labels);

  pnear_labels = malloc (sizeof(uint4) * nb_labels);
  if (pnear_labels == null)
    fatal_compiler_error0 ("exe_allocate_near_label_table(1)");

  memset (pnear_labels, 0x00, sizeof(uint4) * nb_labels);
  nb_near_labels = nb_labels;
}

void exe_declare_near_label (int label_nr)
{
  if (label_nr < 0 || label_nr >= nb_near_labels)
    fatal_compiler_error0 ("exe_declare_near_label(1)");
  pnear_labels[label_nr] = current_RIP();
  
  if (g_tracing)
    trace ("    dcl label #%d\n", label_nr);
}

static void move_near_labels (uint4 start_addr, int4 offset)
{
  int             i;
  RELOC_NEAR_INFO *p;

  for (i=0; i<nb_near_labels; i++)
  {
    if (pnear_labels[i] >= start_addr)
      pnear_labels[i] += offset;
  }

  p = reloc_near_list;
  while (p != null && p->backfill_addr >= start_addr)
  {
    p->backfill_addr += offset;
    p = p->next;
  }
}

// we just wrote a 1-byte branch
void exe_near_branch_written (int label_nr, bool conditional)
{
  RELOC_NEAR_INFO *n;

  n = malloc (sizeof(RELOC_NEAR_INFO));
  if (n == null)
    fatal_compiler_error0 ("exe_near_branch_written(1)");
  n->backfill_addr = current_RIP() - 1;
  n->label_nr      = label_nr;
  n->conditional   = conditional;
  n->was_expanded  = false;
  n->next          = reloc_near_list;

  reloc_near_list = n;
}


void exe_insert_code_sequence (uint4 backfill_addr, uint4 offset_size_increase)
{
  move_all_func_backfills   (backfill_addr, offset_size_increase);
  exe_move_dll_chain        (backfill_addr, offset_size_increase);
  move_all_pool_backfills   (backfill_addr, offset_size_increase);
  move_all_global_backfills (backfill_addr, offset_size_increase);
  move_near_labels          (backfill_addr, offset_size_increase);
  move_dbg_lines            (backfill_addr, offset_size_increase);

  probe (offset_size_increase);
  memmove (&image[(backfill_addr - LOAD_ADDRESS) + offset_size_increase],
           &image[(backfill_addr - LOAD_ADDRESS)],
           image_index - (backfill_addr - LOAD_ADDRESS));
  image_index += offset_size_increase;
}


void exe_update_code_sequence (uint4 backfill_addr, byte *seq, int size)
{
  memmove (&image[(backfill_addr - LOAD_ADDRESS)], seq, size);
}


// to be called after generating asm for each function

static void expand_branch_offsets ()
{
  bool changes_done;

  for (;;)
  {
    RELOC_NEAR_INFO *n;
    int             offset, offset_size_increase;
    uint4           backfill_addr;

    changes_done = false;

    for (n=reloc_near_list; n!=null; n=n->next)
    {
      if (n->was_expanded)
        continue;

      offset = pnear_labels[n->label_nr] - (n->backfill_addr + 1);
      if (offset >= -128 && offset <= 127)    // 1 byte is enough
        continue;

      // we need to increase the offset size

      offset_size_increase = (n->conditional) ? 4 : 3;
      backfill_addr = n->backfill_addr;
      n->was_expanded = true;

      // +1 to avoid that the backfill address itself be relocated
      exe_insert_code_sequence (backfill_addr+1, offset_size_increase);

      changes_done = true;
    }

    if (!changes_done)
      break;
  }
}


// to be called after generating asm for each function

static void backfill_all_near_branch_offsets ()
{
  RELOC_NEAR_INFO *n;
  int             offset;

  for (n=reloc_near_list; n!=null; n=n->next)
  {
    if (n->was_expanded)
    {
      if (n->conditional)
      {
        // c_jcond() generates either 2 or 6 bytes of code
        // before: 0x7n i1
        // after:  0x0F 0x8n i4)
        //               ^ (backfill_addr)

        image[(n->backfill_addr - LOAD_ADDRESS)] =
          (byte)(image[(n->backfill_addr - LOAD_ADDRESS - 1)] + 0x10);

        image[(n->backfill_addr - LOAD_ADDRESS - 1)] = 0x0F;

        *((uint4 *)&image[(n->backfill_addr - LOAD_ADDRESS + 1)]) =
          pnear_labels[n->label_nr] - (n->backfill_addr + 5);
      }
      else
      {
        // c_jump_relative() generates either 2 or 5 bytes of code
        // before: 0xEB i1
        // after:  0xE9 i4
        //              ^ (backfill_addr)

        image[(n->backfill_addr - LOAD_ADDRESS - 1)] = 0xE9;

        *((uint4 *)&image[(n->backfill_addr - LOAD_ADDRESS)]) =
          pnear_labels[n->label_nr] - (n->backfill_addr + 4);
      }
    }
    else   // 1-byte offset
    {
      offset = pnear_labels[n->label_nr] - (n->backfill_addr + 1);
      image[(n->backfill_addr - LOAD_ADDRESS)] = (byte)offset;
    }
  }
}


// to be called after generating asm for each function

static void free_near_list ()
{
  RELOC_NEAR_INFO *n, *prev;

  n = reloc_near_list;

  while (n != null)
  {
    prev = n;
    n = n->next;
    free (prev);
  }

  reloc_near_list = null;
}

/************************************************************************/

static void reloc_jumptable_data ()
{
  RELOC_JUMPTABLE_INFO *n;

  n = reloc_jumptable_list;

  while (n != null)
  {
    relocate_jumptable_pool_constant (n->serial_nr, pnear_labels, nb_near_labels);
    n = n->next;
  }
}

/************************************************************************/

static void free_jumptable ()
{
  RELOC_JUMPTABLE_INFO *n, *prev;

  n = reloc_jumptable_list;

  while (n != null)
  {
    prev = n;
    n = n->next;
    free (prev);
  }

  reloc_jumptable_list = null;
}

/************************************************************************/

// to be called after generating asm for each function

void relocate_all_near_labels ()
{
  expand_branch_offsets ();
  backfill_all_near_branch_offsets ();
  free_near_list ();

  reloc_jumptable_data ();
  free_jumptable ();
}

/************************************************************************/
/************************************************************************/
/************************************************************************/

void exe_finish_code ()
{
  fill_all_func_backfills ();

  image_index_end_code = image_index;

  align_at (PAGE);
  image_index_start_data = image_index;

  code_size = image_index - HEADER;    // aligned code size
}

/************************************************************************/

uint4 exe_store_pool_constant (byte *cte, uint4 size, uint4 align)
{
  uint4 ofs;

  align_at (align);
  ofs = image_index;

  probe (size);
  memcpy (&image[image_index], cte, size);
  image_index += size;

  return LOAD_ADDRESS + ofs;
}

/************************************************************************/

void exe_patch_code (uint4 backfill_addr, bool absolute, uint4 value)
{
  if (absolute)
    *((uint4 *)&image[backfill_addr - LOAD_ADDRESS]) = value;
  else
    *((uint4 *)&image[backfill_addr - LOAD_ADDRESS]) = value - (backfill_addr + 4);
}

/************************************************************************/
static void write_resources_in_exec_image (string res_filename);
/************************************************************************/

void exe_finish_data (string res_filename, int8 global_offset)
{
  write_resources_in_exec_image (res_filename);

  image_index_end_data = image_index;

  write_dll_tables_in_exec_image ();

  align_at (PAGE);
  image_index_start_bss = image_index;
  image_index_end_bss = image_index + (int)global_offset;
  
  data_size = image_index - (HEADER + code_size);   // aligned

  // to be called once bss segment's address is fixed
  fill_all_global_backfills ();
}

/************************************************************************/

void exe_terminate_image (uint4 bss_size,    // must be multiple of PAGE !
                          uint4 stack_size,  // must be multiple of PAGE !
                          bool  console_app)
{
  DATE_TIME  now;

  memcpy (image, mz_header, sizeof(mz_header));

  {
    PE32_PART1 *p1 = (PE32_PART1 *)&image[ofs_pe32_part1];

    p1->signature   = 0x4550;     // "PE" 0x00 0x00
    p1->machineType = (address_size == 4) ? 0x014C : 0x8664;     // 0x014C for i386, 0x8664 for x64.
    p1->nb_sections = 3;          // 3 (text, data, bss) (see at the end of this struct)

    // file creation dtf (nb secs since 1/1/1970)
    get_datetime (&now);
    p1->created_dtf = 86400 * (nb_days_since_1901 (now.day, now.month, now.year) - nb_days_since_1901 (1, 1, 1970))
                    + now.hour * 3600 + now.min * 60 + now.sec;
    p1->symbols     = 0;
    p1->nb_symbols  = 0;

    p1->opt_header_size = ofs_pe32_part4 - ofs_pe32_part2;     // 00E0 or 00F0
    p1->characteristics = (address_size == 4) ? 0x103 : 0x023;  // (0x0100=32-bit, 0x020 = large addr aware, 0x0002=file valid, 0x0001=no reloc)
  }

  if (address_size == 4)
  {
    PE32_PART2_32BIT *p2 = (PE32_PART2_32BIT *)&image[ofs_pe32_part2];

    // optional header - standard fields
    p2->magic                   = 0x010B;  // 0x010B=PE32, 0x020B=PE32+)
    p2->MajorLinkerVersion      = 0x06;
    p2->MinorLinkerVersion      = 0x00;
    p2->SizeOfCode              = code_size;
    p2->SizeOfInitializedData   = data_size;
    p2->SizeOfUninitializedData = bss_size;
    p2->EntryPoint              = HEADER;        // PC start address (skip MZ header)
    p2->BaseOfCode              = HEADER;        // (.text loaded at ImageBase+BaseOfCode = 0x00401000)
    p2->BaseOfData              = p2->BaseOfCode + code_size;   // warning: this field does not exist for PE32+

    // optional header - windows-specific fields
    p2->ImageBase               = LOAD_ADDRESS;    // (MZ_header loaded at this address)
    p2->SectionAlignment        = PAGE;
    p2->FileAlignment           = PAGE;
    p2->MajorOSVersion          = 4;
    p2->MinorOSVersion          = 0;
    p2->MajorImageVersion       = 0;
    p2->MinorImageVersion       = 0;
    p2->MajorSubsystemVersion   = 4;   // 4 (Win32 4.0)
    p2->MinorSubsystemVersion   = 0;
    p2->Win32VersionValue       = 0;
    p2->SizeOfImage             = HEADER + code_size + data_size + bss_size;  // size of image to be loaded in memory, incl. headers (multiple of SectionAlignment)
    p2->SizeOfHeaders           = HEADER;
    p2->CheckSum                = 0;
    p2->Subsystem               = console_app ? 3 : 2;     // 2=WINDOWS_APP, 3=CONSOLE_APP
    p2->DllCharacteristics      = 0;           // 0
    p2->SizeOfStackReserve      = stack_size;  // (1 MB stack : 0x02FFFF to 0x12FFFF)
                                               // (2 MB stack : 0x02FFFF to 0x22FFFF)
                                               // (3 MB stack : 0x02FFFF to 0x32FFFF)
                                               // (4 MB stack : 0x40FFFF to 0x80FFFF)
    p2->SizeOfStackCommit       = PAGE;        // (4 KB pour dmarrer)
    p2->SizeOfHeapReserve       = 0x00100000;  // (1 MB heap)
    p2->SizeOfHeapCommit        = PAGE;        // (4 KB pour dmarrer)
    p2->LoaderFlags             = 0;           // 0
    p2->NumberOfDirectories     = 16;
  }
  else   // 64 bit
  {
    PE32_PART2_64BIT *p2 = (PE32_PART2_64BIT *)&image[ofs_pe32_part2];

    // optional header - standard fields
    p2->magic                   = 0x020B;  // 0x010B=PE32, 0x020B=PE32+)
    p2->MajorLinkerVersion      = 0x06;
    p2->MinorLinkerVersion      = 0x00;
    p2->SizeOfCode              = code_size;
    p2->SizeOfInitializedData   = data_size;
    p2->SizeOfUninitializedData = bss_size;
    p2->EntryPoint              = HEADER;        // PC start address (skip MZ header)
    p2->BaseOfCode              = HEADER;        // (.text loaded at ImageBase+BaseOfCode = 0x00401000)
    p2->ImageBase64             = LOAD_ADDRESS;    // (MZ_header loaded at this address)
    p2->SectionAlignment        = PAGE;
    p2->FileAlignment           = PAGE;
    p2->MajorOSVersion          = 6;   // Vista
    p2->MinorOSVersion          = 0;
    p2->MajorImageVersion       = 0;
    p2->MinorImageVersion       = 0;
    p2->MajorSubsystemVersion   = 6;   // 5.0=Win2000, 5.1=Xp, 6.0=Vista, 6.1=Win7
    p2->MinorSubsystemVersion   = 0;
    p2->Win32VersionValue       = 0;
    p2->SizeOfImage             = HEADER + code_size + data_size + bss_size;  // size of image to be loaded in memory, incl. headers (multiple of SectionAlignment)
    p2->SizeOfHeaders           = HEADER;
    p2->CheckSum                = 0;
    p2->Subsystem               = console_app ? 3 : 2;     // 2=WINDOWS_APP, 3=CONSOLE_APP
    p2->DllCharacteristics      = 0;           // 0
    p2->SizeOfStackReserve      = stack_size;  // (1 MB stack : 0x02FFFF to 0x12FFFF)
                                               // (2 MB stack : 0x02FFFF to 0x22FFFF)
                                               // (3 MB stack : 0x02FFFF to 0x32FFFF)
                                               // (4 MB stack : 0x40FFFF to 0x80FFFF)
    p2->SizeOfStackCommit       = PAGE;        // (4 KB pour dmarrer)
    p2->SizeOfHeapReserve       = 0x00100000;  // (1 MB heap)
    p2->SizeOfHeapCommit        = PAGE;        // (4 KB pour dmarrer)
    p2->LoaderFlags             = 0;           // 0
    p2->NumberOfDirectories     = 16;
  }


  // section table

  {
    PE32_PART4 *p4 = (PE32_PART4 *)&image[ofs_pe32_part4];

    // .text (code)
    memcpy (p4->code.name, ".text", 5);
    p4->code.size_in_memory       = code_size;
    p4->code.addr                 = HEADER;
    p4->code.SizeOnDisk           = code_size;   // multiple of FileAlignment
    p4->code.PointerToDisk        = HEADER;      // ptr into section data in file, relative to optional header
    p4->code.PointerToRelocations = 0;
    p4->code.PointerToLinenumbers = 0;
    p4->code.NumberOfRelocations  = 0;
    p4->code.NumberOfLinenumbers  = 0;
    p4->code.Characteristics      = 0x20000020;   // code + exec (most exe's use 60000020 for readable too)

    // .data (constants pool, switch jump tables, dll import tables)
    memcpy (p4->data.name, ".data", 5);
    p4->data.size_in_memory       = data_size;
    p4->data.addr                 = HEADER + code_size;
    p4->data.SizeOnDisk           = data_size;          // multiple of FileAlignment  (0 for .bss section)
    p4->data.PointerToDisk        = HEADER + code_size; // ptr into section data in file, relative to optional header
    p4->data.PointerToRelocations = 0;
    p4->data.PointerToLinenumbers = 0;
    p4->data.NumberOfRelocations  = 0;
    p4->data.NumberOfLinenumbers  = 0;
    p4->data.Characteristics      = 0x40000040;      // init + read (initialized read-only data)

    // .bss (global variables)
    memcpy (p4->bss.name, ".bss", 4);
    p4->bss.size_in_memory       = bss_size;
    p4->bss.addr                 = HEADER + code_size + data_size;
    p4->bss.SizeOnDisk           = 0;  // 0 for .bss section
    p4->bss.PointerToDisk        = 0;
    p4->bss.PointerToRelocations = 0;
    p4->bss.PointerToLinenumbers = 0;
    p4->bss.NumberOfRelocations  = 0;
    p4->bss.NumberOfLinenumbers  = 0;
    p4->bss.Characteristics      = 0xC0000080;       // uninit + read + write
  }
}

/************************************************************************/

void print_crc ()
{
  unsigned long crc;
  int           ofs;

  // compute crc of executable, but skip 4 bytes at offset (ofs_pe32_part1 + 8)
  ofs = ofs_pe32_part1 + 8;

  crc = 0;
  update_crc (&crc, image, ofs);
  update_crc (&crc, &image[ofs + 4], image_index - (ofs + 4));

  printf ("CRC=%08X CODE=%d bytes DATA=%d bytes BSS=%d bytes\n", 
         crc, 
         image_index_end_code - image_index_start_code,
         image_index_end_data - image_index_start_data,
         image_index_end_bss  - image_index_start_bss);
}

/************************************************************************/

void exe_save (char *exec_filename)
{
  int fd;

  fd = open (exec_filename, O_CREAT|O_TRUNC|O_RDWR|O_BINARY, 0660);
  if (fd < 0)
  {
    fatal_compiler_error0 ("cannot save exe file on disk");
    return;
  }

  if (write (fd, image, image_index) != (int)image_index)
  {
    close (fd);
    unlink (exec_filename);
    fatal_compiler_error0 ("cannot write exe file to disk");
    return;
  }

  if (close (fd) < 0)
  {
    unlink (exec_filename);
    fatal_compiler_error0 ("cannot close exe file");
    return;
  }
}

/************************************************************************/

#pragma pack(1)
typedef struct {
  uint    charac;
  uint    datetime;
  uint2   major;
  uint2   minor;
  uint2   nb_names;
  uint2   nb_ids;
} RES_DIR_HEADER;
#pragma pack()

#pragma pack(1)
typedef struct {
  uint    name_or_id;
  uint    next;    // dir has high bit=1, leaf has not.
} RES_DIR_ENTRY;
#pragma pack()

#pragma pack(1)
typedef struct {      // aligned at 2 !!
  uint2   len;
} RES_STRING_ENTRY;   // followed by unicode string
#pragma pack()

#pragma pack(1)
typedef struct {
  uint    res_nva;
  uint    res_size;
  uint    codepage;
  uint    unused;
} RES_NODE;
#pragma pack()



typedef struct {
  string         name;
  uint           id;
  uint           ientry;   // ptr to RES_DIR_ENTRY
} RES_DATA;

// each RES_INFO is sorted by type/id/lang,
// and at each level first the name's then the id's.

typedef struct {
  RES_DATA type;
  RES_DATA id;
  RES_DATA lang;
  string   resource_filename;
  uint     inode;    // ptr to RES_NODE
} RES_INFO;

typedef RES_INFO *PRES_INFO;

static BINARY_TREE res_tree;             // tree of PRES_INFO
static bool        res_tree_created;

/**********************************************************************************/

static int uintcmp (uint a, uint b)
{
  if (a < b)
    return -1;
  if (a > b)
    return +1;
  return 0;
}

/**********************************************************************************/

static int cmp_res_data (RES_DATA *a, RES_DATA *b)
{
  int cmp;

  if (a->name != null && b->name == null)
    return -1;

  if (a->name == null && b->name != null)
    return +1;

  if (a->name != null && b->name != null)
    cmp = stricmp (a->name, b->name);
  else
    cmp = uintcmp (a->id, b->id);

  return cmp;
}

/**********************************************************************************/

static int res_compare (void               *user,
                        const RES_INFO     **a,
                        const unsigned int size1,
                        const RES_INFO     **b,
                        const unsigned int size2)
{
  int cmp;

  cmp = cmp_res_data (&((RES_INFO *)*a)->type, &((RES_INFO *)*b)->type);
  if (cmp != 0)
    return cmp;

  cmp = cmp_res_data (&((RES_INFO *)*a)->id, &((RES_INFO *)*b)->id);
  if (cmp != 0)
    return cmp;

  cmp = cmp_res_data (&((RES_INFO *)*a)->lang, &((RES_INFO *)*b)->lang);
  return cmp;
}

/**********************************************************************************/

static char *new_string (char *s)
{
  char *p;
  p = malloc (strlen(s) + 1);
  if (p == null)
    fatal_out_of_memory_error ("new_string(1)");
  strcpy (p, s);
  return p;
}

/**********************************************************************************/

static void fill_res_data (string value, RES_DATA *p)
{
  if (isdigit(value[0]))
    p->id = atoi (value);
  else
    p->name = new_string (value);
}

/**********************************************************************************/

// returns 0 if OK, -1 if duplicate key

static int insert_resource_in_tree (string namestr, string typstr, string filename)
{
  PRES_INFO p;
  int       rc;

  if (!res_tree_created)
  {
    create_btree (&res_tree, NULL, res_compare);
    res_tree_created = true;
  }

  p = malloc (sizeof(RES_INFO));
  if (p == NULL)
    fatal_out_of_memory_error ("insert_resource_in_tree(1)");
  memset (p, 0x00, sizeof(RES_INFO));

  fill_res_data (typstr,  &p->type);
  fill_res_data (namestr, &p->id);
  fill_res_data ("0",     &p->lang);

  p->resource_filename = new_string (filename);

  rc = insert_btree (&res_tree, &p, sizeof(p));
  if (rc == BT_DUPLICATE_KEY)
    return -1;
  if (rc != 0)
    fatal_out_of_memory_error ("insert_resource_in_tree(2)");

  return 0;
}

/**********************************************************************************/

void parse_resource_file (string current_dir, string res_filename)
{
  FILE *fp;
  int  line_nr, col, len, res_col;
  char line[512], namestr[64+1], typstr[64+1], filename[200+1];
  char absolute_filename[260], msg[512];

  fp = fopen (res_filename, "r");
  if (fp == NULL)                    // no resource file provided
    return;

  line_nr = 0;
  while (fgets (line, sizeof(line), fp) != NULL)
  {
    line_nr++;
    col = 0;

    // skip blanks
    while (line[col] == ' ')
      col++;

    if (line[col] == '\n' || line[col] == 0)
      continue;

    len = 0;
    while (len < sizeof(namestr)-1 && line[col] > ' ')
      namestr[len++] = line[col++];
    namestr[len] = nul;
    if (len == 0)
      resource_error (res_filename, "resource ID expected", line_nr, col);
    if (len == sizeof(namestr))
      resource_error (res_filename, "resource ID is too long", line_nr, col);

    // skip blanks
    while (line[col] == ' ')
      col++;

    len = 0;
    while (len < sizeof(typstr)-1 && line[col] > ' ')
      typstr[len++] = line[col++];
    typstr[len] = nul;
    if (len == 0)
      resource_error (res_filename, "resource TYPE expected", line_nr, col);
    if (len == sizeof(typstr))
      resource_error (res_filename, "resource TYPE is too long", line_nr, col);

    // skip blanks
    while (line[col] == ' ')
      col++;

    res_col = col;

    len = 0;
    while (len < sizeof(filename)-1 && line[col] > ' ')
      filename[len++] = line[col++];
    filename[len] = nul;
    if (len == 0)
      resource_error (res_filename, "resource FILENAME expected", line_nr, col);
    if (len == sizeof(filename))
      resource_error (res_filename, "resource FILENAME is too long", line_nr, col);

    // skip blanks
    while (line[col] == ' ')
      col++;

    if (line[col] != '\n' && line[col] != 0)
      resource_error (res_filename, "end of line expected", line_nr, col);

    if (expand_pathname (current_dir, filename, absolute_filename, sizeof(absolute_filename)-5) < 0)
      resource_error (res_filename, "cannot create resource filename", line_nr, res_col);

    if (access (absolute_filename, 0) < 0)
    {
      sprintf (msg, "cannot open resource %s", absolute_filename);
      resource_error (res_filename, msg, line_nr, res_col);
    }

    if (insert_resource_in_tree (namestr, typstr, filename) < 0)
      resource_error (res_filename, "resource give twice", line_nr, 1);
  }

  fclose (fp);
}

/************************************************************************/

static bool entries_different (RES_DATA *a, RES_DATA *b)
{
  if (a->name != null && b->name != null)
    return stricmp (a->name, b->name) != 0;

  if (a->name == null || b->name == null)
    return a->id != b->id;

  return true;
}

/************************************************************************/

static void write_resources_in_exec_image (string res_filename)
{
  PE32_PART3     *p;
  short          mode;
  RES_INFO       *info, *prev_info;
  int            rc, dummy, size;
  uint           ihead, ient;
  RES_DIR_HEADER *phead;
  RES_DIR_ENTRY  *pent, *tmpent;

  if (!res_tree_created)
    return;

  // 1) write TYPE directory

  align_at (16);

  p = (PE32_PART3 *)&image[ofs_pe32_part3];
  p->resource.addr = image_index;


  // send TYPE directory

  size = sizeof(RES_DIR_HEADER);
  probe (size);
  ihead = image_index;
  phead = (RES_DIR_HEADER *)&image[ihead];
  memset (phead, 0x00, size);
  image_index += size;

  prev_info = null;

  mode = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&res_tree, &info, sizeof(info), sizeof(info), &dummy, mode);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_resources_in_exec_image()");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    if (prev_info == null || entries_different (&prev_info->type, &info->type))
    {
      size = sizeof(RES_DIR_ENTRY);
      probe (size);
      ient = image_index;
      pent = (RES_DIR_ENTRY *)&image[ient];
      memset (pent, 0x00, size);
      image_index += size;

      if (info->type.name == null)
        pent->name_or_id = info->type.id;


      // update counters

      phead = (RES_DIR_HEADER *)&image[ihead];
      if (info->type.name != null)  // it's a name
        phead->nb_names++;
      else     // it's numeric
        phead->nb_ids++;

      prev_info = info;
    }

    info->type.ientry = ient;

    mode = BT_LARGER;
  }


  // send ID directory

  prev_info = null;

  mode = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&res_tree, &info, sizeof(info), sizeof(info), &dummy, mode);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_resources_in_exec_image()");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    if (prev_info == null || entries_different (&prev_info->type, &info->type))
    {
      // start new ID directory table (all the ID's of this type)
      size = sizeof(RES_DIR_HEADER);
      probe (size);
      ihead = image_index;
      phead = (RES_DIR_HEADER *)&image[ihead];
      memset (phead, 0x00, size);
      image_index += size;

      tmpent = (RES_DIR_ENTRY *)&image[info->type.ientry];
      tmpent->next = ((image_index - size) - p->resource.addr) | 0x80000000;

      prev_info = null;
    }

    if (prev_info == null || entries_different (&prev_info->id, &info->id))
    {
      size = sizeof(RES_DIR_ENTRY);
      probe (size);
      ient = image_index;
      pent = (RES_DIR_ENTRY *)&image[ient];
      memset (pent, 0x00, size);
      image_index += size;

      if (info->id.name == null)
        pent->name_or_id = info->id.id;


      // update counters

      phead = (RES_DIR_HEADER *)&image[ihead];
      if (info->id.name != null)  // it's a name
        phead->nb_names++;
      else     // it's numeric
        phead->nb_ids++;

      prev_info = info;
    }

    info->id.ientry = ient;

    mode = BT_LARGER;
  }


  // send LANG directory

  prev_info = null;

  mode = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&res_tree, &info, sizeof(info), sizeof(info), &dummy, mode);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_resources_in_exec_image()");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    if (prev_info == NULL ||
        entries_different (&prev_info->type, &info->type) ||
        entries_different (&prev_info->id,   &info->id))
    {
      // start new LANG directory table (all the LANG's of this type+id)
      size = sizeof(RES_DIR_HEADER);
      probe (size);
      ihead = image_index;
      phead = (RES_DIR_HEADER *)&image[ihead];
      memset (phead, 0x00, size);
      image_index += size;

      tmpent = (RES_DIR_ENTRY *)&image[info->id.ientry];
      tmpent->next = ((image_index - size) - p->resource.addr) | 0x80000000;
    }

    {
      size = sizeof(RES_DIR_ENTRY);
      probe (size);
      ient = image_index;
      pent = (RES_DIR_ENTRY *)&image[ient];
      memset (pent, 0x00, size);
      image_index += size;

      if (info->lang.name == null)
        pent->name_or_id = info->lang.id;


      // update counters

      phead = (RES_DIR_HEADER *)&image[ihead];
      if (info->lang.name != null)  // it's a name
        phead->nb_names++;
      else     // it's numeric
        phead->nb_ids++;

      info->lang.ientry = ient;
    }

    prev_info = info;


    mode = BT_LARGER;
  }


  // send unicode strings

  align_at (2);

  prev_info = null;

  mode = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&res_tree, &info, sizeof(info), sizeof(info), &dummy, mode);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_resources_in_exec_image()");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    if (info->type.name != null)   // a string
    {
      if (prev_info == null || entries_different (&prev_info->type, &info->type))
      {
        uint2* q;
        int    len, i;

        tmpent = (RES_DIR_ENTRY *)&image[info->type.ientry];
        tmpent->name_or_id = (image_index - p->resource.addr) | 0x80000000;

        len = strlen(info->type.name);
        size = 2 * (1 + len);
        probe (size);
        q = (uint2 *)&image[image_index];
        *q++ = (uint2)len;
        for (i=0; i<len; i++)
          *q++ = info->type.name[i];

        image_index += size;
      }
    }

    if (info->id.name != null)   // a string
    {
      if (prev_info == null || entries_different (&prev_info->type, &info->type)
                            || entries_different (&prev_info->id, &info->id))
      {
        uint2* q;
        int    len, i;

        tmpent = (RES_DIR_ENTRY *)&image[info->id.ientry];
        tmpent->name_or_id = (image_index - p->resource.addr) | 0x80000000;

        len = strlen(info->id.name);
        size = 2 * (1 + len);
        probe (size);
        q = (uint2 *)&image[image_index];
        *q++ = (uint2)len;
        for (i=0; i<len; i++)
          *q++ = info->id.name[i];

        image_index += size;
      }
    }

    if (info->lang.name != null)   // a string
    {
      {
        uint2* q;
        int    len, i;

        tmpent = (RES_DIR_ENTRY *)&image[info->lang.ientry];
        tmpent->name_or_id = (image_index - p->resource.addr) | 0x80000000;

        len = strlen(info->lang.name);
        size = 2 * (1 + len);
        probe (size);
        q = (uint2 *)&image[image_index];
        *q++ = (uint2)len;
        for (i=0; i<len; i++)
          *q++ = info->lang.name[i];

        image_index += size;
      }
    }

    prev_info = info;

    mode = BT_LARGER;
  }


  // send resource nodes

  align_at (16);

  mode = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&res_tree, &info, sizeof(info), sizeof(info), &dummy, mode);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_resources_in_exec_image()");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    {
      RES_NODE *r;
      uint     inode;

      size = sizeof(RES_NODE);
      probe (size);
      inode = image_index;
      r = (RES_NODE *)&image[inode];
      memset (r, 0x00, size);
      image_index += size;

      tmpent = (RES_DIR_ENTRY *)&image[info->lang.ientry];
      tmpent->next = inode - p->resource.addr;

      info->inode = inode;
    }

    mode = BT_LARGER;
  }


  // send resources

  mode = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&res_tree, &info, sizeof(info), sizeof(info), &dummy, mode);
    if (rc != 0 && rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error0 ("write_resources_in_exec_image()");

    if (rc == BT_KEY_NOT_FOUND)
      break;

    {
      int fd, size;
      RES_NODE *r;

      fd = open (info->resource_filename, O_RDONLY|O_BINARY);
      if (fd < 0)
      {
        char msg[512];
        sprintf (msg, "cannot open resource %.260s", info->resource_filename);
        resource_error (res_filename, msg, 0, 1);
      }

      size = lseek (fd, 0L, SEEK_END);
      lseek (fd, 0L, SEEK_SET);

      r = (RES_NODE *)&image[info->inode];
      r->res_nva = image_index;
      r->res_size = size;

      probe (size);
      if (read (fd, &image[image_index], size) != size)
      {
        char msg[512];
        sprintf (msg, "cannot read resource %.260s", info->resource_filename);
        resource_error (res_filename, msg, 0, 1);
      }
      image_index += size;

      close (fd);
    }

    mode = BT_LARGER;
  }

  p = (PE32_PART3 *)&image[ofs_pe32_part3];
  p->resource.size = image_index - p->resource.addr;
}

/************************************************************************/
