
// pool.c : constant pool

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "bintree.h"
#include "binsort.h"
#include "tracing.h"

#include "lex.h"
#include "error.h"
#include "goptions.h"
#include "pe.h"
#include "pool.h"

/******************************************************************************************/

#define MAX_REFERENCES  10   // max references per REFERENCE_NODE

typedef struct {
  uint4           offset; // offset within pool constant where to patch the reference
  struct pool_cte *pool;  // pool constant to be referenced
} REFERENCE;

typedef struct REF {
  int        count;           // nb of references in this node (1 .. MAX_REFERENCES)
  REFERENCE  r[MAX_REFERENCES];
  struct REF *next;
} REFERENCE_NODE;

typedef struct backfill_info {
  uint4                backfill_addr;  // 4-byte address to fill
  bool                 absolute;       // absolute or relative address
  uint4                target_offset;  // offset to add to patched value
  struct backfill_info *move_next;     // next backfill to move (used within function)
  struct backfill_info *next;
} BACKFILL_INFO;

typedef struct pool_cte {
  byte            *mem;        // allocated pool constant
  uint4           size;        // size of pool constant
  uint4           align;       // advised alignment
  int8            serial_nr;   // unique serial nr of pool constant, used for references
  REFERENCE_NODE  *first;      // first node of list of references to other pool constants
  REFERENCE_NODE  *last;       // last node
  bool            used;        // referenced by code generation
  BACKFILL_INFO   *backfill;   // list of backfill addresses to fill within code segment

  // after merging:
  struct pool_cte *redirect_pool;   // non-null -> redirect to within another pool constant
  int4            redirect_offset;  //             at specified offset.

  bool            flushed;       // written in .exe
  uint4           rip_address;

} POOL_CTE;

static BINARY_TREE tree;         // contains all referenced pool constants
static int8        serial_nr;    // next unique nr

/******************************************************************************************/

// maximum pool size for all constants together
#define MAX_POOL_SIZE  (256*1024*1024)   // 256 MB

static int8 total_allocated_size;

/******************************************************************************************/

static int compare (void               *user,
                    const POOL_CTE     **data1,
                    const unsigned int size1,
                    const POOL_CTE     **data2,
                    const unsigned int size2)
{
  int8 a, b;

  a = (*data1)->serial_nr;
  b = (*data2)->serial_nr;

  if (a < b)
    return -1;

  if (a > b)
    return +1;

  return 0;
}

/******************************************************************************************/

void init_constant_pool (void)
{
  create_btree (&tree, NULL, compare);
  total_allocated_size = 0;
}

/******************************************************************************************/

POOL *new_pool_constant (uint4 size, uint4 align)
{
  byte     *mem;
  POOL_CTE *p;

  if (size + total_allocated_size > MAX_POOL_SIZE)
  {
    semantic_error ("too many large constants", &token.pos);
    fatal_out_of_memory_error ("out of memory");
  }

  total_allocated_size += size;

  mem = malloc (size);
  if (mem == NULL)
    fatal_out_of_memory_error ("out of memory1 for pool constant");

  memset (mem, 0x00, size);

  p = malloc (sizeof(POOL_CTE));
  if (p == NULL)
    fatal_out_of_memory_error ("out of memory2 for pool constant");

  memset (p, 0x00, sizeof(POOL_CTE));
  p->mem       = mem;
  p->size      = size;
  p->align     = align;
  p->serial_nr = 0;
  p->first     = null;
  p->last      = null;
  p->used      = false;
  p->redirect_pool = null;
  p->redirect_offset = 0;
  p->flushed     = false;
  p->rip_address = 0;
  return (POOL *)p;
}

/******************************************************************************************/

uint4 size_of_pool_cte (POOL *p)
{
  return ((POOL_CTE *)p)->size;
}

/******************************************************************************************/

uint4 align_of_pool_cte (POOL *p)
{
  return ((POOL_CTE *)p)->align;
}

/******************************************************************************************/

// load pool constant integer

void load_integer (POOL *p, uint4 offset, int8 *pvalue, uint size, bool is_signed)
{
  byte *mem, b;
  uint i;

  if (size > 8 ||
      offset > (uint)((POOL_CTE *)p)->size ||
      offset + size > (uint)((POOL_CTE *)p)->size)
    fatal_compiler_error0 ("load_integer");

  mem = &((POOL_CTE *)p)->mem[offset];

  *pvalue = 0;

  for (i=0; i<size; i++)
  {
    if (big_endian)   // Motorola
      b = mem[size-1-i];
    else
      b = mem[i];
    *pvalue += ((int8)b << (i*8));
  }

  if (is_signed && b >= 128)
  {
    for (; i<8; i++)
      *pvalue += ((int8)0xFF << (i*8));
  }
}

/******************************************************************************************/

// store pool constant integer

void store_integer (POOL *p, uint4 offset, int8 value, uint size)
{
  byte *mem;
  uint i;

  if (size > 8 ||
      offset > (uint)((POOL_CTE *)p)->size ||
      offset + size > (uint)((POOL_CTE *)p)->size)
    fatal_compiler_error0 ("store_integer");

  mem = &((POOL_CTE *)p)->mem[offset];

  if (big_endian)   // Motorola
  {
    for (i=0; i<size; i++)
      mem[i] = (byte)((value >> ((size-1-i)*8)) & 255);
  }
  else   // Intel
  {
    for (i=0; i<size; i++)
      mem[i] = (byte)((value >> (i*8)) & 255);
  }
}

/******************************************************************************************/

// load pool constant float

void load_float (POOL *p, uint4 offset, double *pvalue, uint size)
{
  int8 r;

  load_integer (p, offset, &r, size, 0);

  if (size == 4)
  {
    int4 i4 = (int4)r;            // take LSB 4 bytes
    float f = *((float *)&i4);
    *pvalue = f;
  }
  else
  {
    *pvalue = *((double *)&r);
  }
}

/******************************************************************************************/

// store pool constant float

void store_float (POOL *p, uint4 offset, double value, uint size)
{
  int8 r;

  if (size == 4)
  {
    float f = (float)value;
    int4 i4 = *((int4 *)&f);
    r = i4;                    // expand into 8 bytes
  }
  else
  {
    r = *((int8 *)&value);
  }

  store_integer (p, offset, r, size);
}

/******************************************************************************************/

// load/store pool references (references have address_size)
// reference is added in intern linked list
// ! assertion: all references are inserted in increasing offset order !!

void store_reference (POOL *p, uint4 offset, POOL *target)
{
  REFERENCE_NODE *list, *prev;

  if (offset > (uint)((POOL_CTE *)p)->size ||
      offset + address_size > (uint)((POOL_CTE *)p)->size)
    fatal_compiler_error0 ("store_reference(1)");

  list = ((POOL_CTE *)p)->last;
  if (list == null || list->count >= MAX_REFERENCES)
  {
    // allocate a new REF node
    prev = list;
    list = malloc (sizeof(REFERENCE_NODE));
    if (list == null)
      fatal_compiler_error0 ("store_reference(2)");
    memset (list, 0x00, sizeof(REFERENCE_NODE));
    if (prev == null)
    {
      ((POOL_CTE *)p)->first = list;
    }
    else
    {
      prev->next = list;
      if (offset <= prev->r[prev->count-1].offset)        // references must be inserted in order !
        fatal_compiler_error0 ("store_reference(3)");
    }

    ((POOL_CTE *)p)->last = list;
  }
  else
  {
    if (offset <= list->r[list->count-1].offset)        // references must be inserted in order !
      fatal_compiler_error0 ("store_reference(4)");
  }

  list->r[list->count].offset = offset;
  list->r[list->count].pool   = (POOL_CTE *)target;
  list->count++;
}

/******************************************************************************************/

// is retrieved from intern linked list

POOL *load_reference_of (POOL *p, uint4 offset)
{
  REFERENCE_NODE *list;
  int            i;

  if (offset > (uint)((POOL_CTE *)p)->size ||
      offset + address_size > (uint)((POOL_CTE *)p)->size)
    fatal_compiler_error0 ("load_reference_of(1)");

  list = ((POOL_CTE *)p)->first;
  while (list != null)
  {
    for (i=0; i<list->count; i++)
    {
      if (list->r[i].offset == offset)
        return (POOL *)list->r[i].pool;
    }

    list = list->next;
  }

  fatal_compiler_error0 ("load_reference_of(2)");
  return null;
}

/******************************************************************************************/

// copy pool slice, copy and relocate also all references in the slice.

void copy_pool_to_pool (POOL *source_pool, uint4 source_offset,
                        POOL *target_pool, uint4 target_offset,
                        uint size)
{
  REFERENCE_NODE *list;
  int            i;

  if (size > MAX_POOL_SIZE)
    fatal_compiler_error0 ("copy_pool_to_pool(1)");

  if (source_offset > (uint)((POOL_CTE *)source_pool)->size ||
      source_offset + size > (uint)((POOL_CTE *)source_pool)->size)
    fatal_compiler_error0 ("copy_pool_to_pool(2)");

  if (target_offset > (uint)((POOL_CTE *)target_pool)->size ||
      target_offset + size > (uint)((POOL_CTE *)target_pool)->size)
    fatal_compiler_error0 ("copy_pool_to_pool(3)");

  memcpy (&((POOL_CTE *)target_pool)->mem[target_offset],
          &((POOL_CTE *)source_pool)->mem[source_offset],
          size);

  list = ((POOL_CTE *)source_pool)->first;
  while (list != null)
  {
    for (i=0; i<list->count; i++)
    {
      if (list->r[i].offset >= source_offset && list->r[i].offset + address_size <= source_offset + size)
        store_reference (target_pool, list->r[i].offset - source_offset + target_offset, (POOL *)list->r[i].pool);
    }

    list = list->next;
  }
}

/******************************************************************************************/

// incl. recursive test for constants containing references

bool pool_constants_are_identical (POOL *p, POOL *q)
{
  REFERENCE_NODE *lp, *lq;
  int            i;

  if (p == q)       // optimization
    return true;

  if (((POOL_CTE *)p)->size != ((POOL_CTE *)q)->size ||
      memcmp (((POOL_CTE *)p)->mem, ((POOL_CTE *)q)->mem, ((POOL_CTE *)q)->size) != 0)
    return false;

  // check that all references match (we have checked that they were inserted in order)

  lp = ((POOL_CTE *)p)->first;
  lq = ((POOL_CTE *)q)->first;

  for (;;)
  {
    if (lp == null && lq == null)
      return true;
    if (lp == null || lq == null)
      return false;

    if (lp->count != lq->count)
      return false;

    for (i=0; lp->count; i++)
    {
      if (lp->r[i].offset != lq->r[i].offset)
        return false;
      if (!pool_constants_are_identical ((POOL *)lp->r[i].pool, (POOL *)lq->r[i].pool))
        return false;
    }

    lp = lp->next;
    lq = lq->next;
  }

  return true;
}

/******************************************************************************************/

static void set_used_flag (POOL_CTE *p)
{
  REFERENCE_NODE *list;
  int            i;

  if (p->used)     // .used already set for all dependencies
    return;

  p->used = true;

  // assign unique serial nr
  p->serial_nr = serial_nr++;

if (false)
{
  printf ("\n");
  printf ("serial cte %d : ", p->serial_nr);
  for (i=0; i<(int)p->size; i++)
    printf ("%c (%d) ", p->mem[i], p->mem[i]);
}

  // insert in btree, indexed by serial_nr
  if (insert_btree (&tree, &p, sizeof(p)) < 0)
    fatal_out_of_memory_error ("new_pool_constant() failed");

  list = p->first;
  while (list != null)
  {
    for (i=0; i<list->count; i++)
      set_used_flag (list->r[i].pool);
    list = list->next;
  }
}

/******************************************************************************************/

// must be called after the pool constant was fully initialized.

int8 serial_nr_of_pool_cte (POOL *p)
{
  set_used_flag ((POOL_CTE *)p);      // signal that this constant is referenced
  return ((POOL_CTE *)p)->serial_nr;
}

/******************************************************************************************/
/******************************************************************************************/
/******************************************************************************************/

static BACKFILL_INFO *move_chain;

/******************************************************************************************/

// called at start of new function
void reset_pool_backfills ()
{
  move_chain = null;
}

/******************************************************************************************/

void move_all_pool_backfills (uint4 start_addr, int4 offset)
{
  BACKFILL_INFO *n = move_chain;

  while (n != null && n->backfill_addr >= start_addr)
  {
    n->backfill_addr += offset;
    n = n->move_next;
  }
}

/************************************************************************/

static POOL_CTE *pool_cte_of_serial_nr (int8 nr)
{
  POOL_CTE     *p, *q;
  unsigned int dummy;

  p = malloc (sizeof(POOL_CTE));
  if (p == NULL)
    fatal_out_of_memory_error ("out of memory3 for pool constant");

  memset (p, 0x00, sizeof(POOL_CTE));
  p->serial_nr = nr;

  q = p;

  if (retrieve_btree (&tree, &p, sizeof(p), sizeof(p), &dummy, BT_EQUAL) < 0)
    fatal_out_of_memory_error ("retrieve error for pool constant");

  free (q);

  return p;
}

/******************************************************************************************/

// for rip, use current_RIP()-4

void pool_add_backfill_addr4 (int8 serial_nr, uint4 rip, bool absolute, uint4 target_offset)
{
  POOL_CTE      *p;
  BACKFILL_INFO *n;

  p = pool_cte_of_serial_nr (serial_nr);

  n = malloc (sizeof(BACKFILL_INFO));
  if (n == null)
    fatal_out_of_memory_error ("pool_add_backfill_addr4(1)");
  n->backfill_addr = rip;
  n->absolute      = absolute;
  n->target_offset = target_offset;
  n->move_next     = move_chain;
  move_chain = n;

  n->next = p->backfill;
  p->backfill = n;
}

/******************************************************************************************/

void relocate_jumptable_pool_constant (int8 serial_nr, uint4 label_table[], int label_table_count)
{
  POOL *p;
  int  offset, size;
  int8 label_nr;

  p = (POOL *)pool_cte_of_serial_nr (serial_nr);
  size = size_of_pool_cte (p);

  for (offset=0; offset<size; offset+=address_size)
  {
    load_integer (p, offset, &label_nr, address_size, 0);
    if (label_nr < 0 || label_nr >= label_table_count)
      fatal_compiler_error0 ("relocate_jumptable_pool_constant(1)");
    store_integer (p, offset, label_table[label_nr], address_size);
  }
}

/******************************************************************************************/
/******************************************************************************************/
/******************************************************************************************/

static POOL_CTE **merge_table;
static int      merge_table_count;

/******************************************************************************************/

static int operate (void               *user,
                    const POOL_CTE     **data,
                    const unsigned int size)
{
// printf ("operate %d\n", (*data)->serial_nr);
  merge_table[merge_table_count++] = (POOL_CTE *)(*data);
  return 0;
}

/******************************************************************************************/

static int compare_size (const POOL_CTE **p1, const POOL_CTE **p2)
{
// printf ("cmp_size %d  %d\n", (*p1)->serial_nr, (*p2)->serial_nr);
  if ((*p1)->size < (*p2)->size)
    return -1;
  if ((*p1)->size > (*p2)->size)
    return +1;
  return 0;
}

/******************************************************************************************/

// returns -1 if no substring found with interesting alignment

static int substring (byte fragment[], uint4 fragment_size, int fragment_align,
                      byte mem[],      uint4 mem_size,      int mem_align)
{
  int i, last;

  if (fragment_align > mem_align)     // superior alignment cannot be provided
    return -1;

  if (fragment_size == 0)  // empty string always matches
    return 0;

  last = mem_size - fragment_size;
  for (i=0; i<=last; i+=fragment_align)
  {
    if (fragment[0] == mem[i] && memcmp (fragment, &mem[i], fragment_size) == 0)
      return i;
  }

  return -1;
}

/******************************************************************************************/

void merge_pool_subtrings ()
{
  int  i, j, index;
  uint size;

  merge_table = (POOL_CTE **)malloc ((uint4)(sizeof(POOL_CTE *) * serial_nr));
  if (merge_table == null)
    fatal_out_of_memory_error ("merge_pool_subtrings()");

  merge_table_count = 0;

  // initialize table with all used pool constants, ordered by serial_nr
  traverse_btree (&tree, operate, +1);

  // sort table by increasing size
  binary_sort (merge_table, merge_table_count, sizeof(POOL_CTE *), compare_size);

  for (i=merge_table_count-2; i>=0; i--)   // process by decreasing index, from large size to small size
  {
    if (merge_table[i]->first != null)   // don't merge jagged arrays
      continue;

    size = merge_table[i]->size;

    for (j=merge_table_count-1; j>i; j--)   // consider redirecting fragment i into larger chain j
    {
      if (merge_table[j]->redirect_pool != null)  // already redirected
        continue;

      if (merge_table[j]->first != null)   // don't merge jagged arrays
        continue;

      index = substring (merge_table[i]->mem, size,                 merge_table[i]->align,
                         merge_table[j]->mem, merge_table[j]->size, merge_table[j]->align);
      if (index != -1)
      {
        merge_table[i]->redirect_pool = merge_table[j];
        merge_table[i]->redirect_offset = index;

// printf ("merging cte %d into cte %d\n", i, j);

        break;
      }
    }
  }

  free (merge_table);
}

/******************************************************************************************/
/******************************************************************************************/

// returns rip address of constant

static uint4 flush_pool_constant (POOL_CTE *p)
{
  REFERENCE_NODE *r;
  int            i;
  uint4          addr;
  BACKFILL_INFO  *l;

  if (p->flushed)    // was already flushed earlier
    return p->rip_address;

  if (p->redirect_pool != null)
  {
    p->rip_address = flush_pool_constant(p->redirect_pool) + p->redirect_offset;
  }
  else
  {
    // first, flush all dependant constants and fill the jagged references

    r = p->first;
    while (r != null)
    {
      for (i=0; i<r->count; i++)
      {
        addr = flush_pool_constant (r->r[i].pool);     // recursive
        store_integer ((POOL *)p, r->r[i].offset, addr, address_size);
      }

      r = r->next;
    }

    // second flush the constant in exe

    {
      uint4 size, align;

      size = p->size;

      if (size > 8)       // size 9..?
        align = 16;
      else if (size > 4)  // size 5..8
        align = 8;
      else if (size > 2)  // size 3..4
        align = 4;
      else if (size > 1)  // size 2
        align = 2;
      else                // size 1
        align = 1;

      p->rip_address = exe_store_pool_constant (p->mem, size, max(align, p->align));

#if 0
      trace ("pool cte at addr %d size %d\n", p->rip_address, size);
      trace_block (p->mem, size);
#endif
    }
  }


  // backfill all code references to this pool constant

  l = p->backfill;
  while (l != null)
  {
    exe_patch_code (l->backfill_addr, l->absolute, p->rip_address + l->target_offset);
    l = l->next;
  }

  p->flushed = true;

  return p->rip_address;
}

/******************************************************************************************/

static int operate_flush_pool_constant (void               *user,
                                        const POOL_CTE     **data,
                                        const unsigned int size)
{
  (void)flush_pool_constant ((POOL_CTE *)(*data));
  return 0;
}

/******************************************************************************************/

void flush_pool_subtrings ()
{
  traverse_btree (&tree, operate_flush_pool_constant, +1);
  
  if (address_size == 8)  // append 4 byte dummy at the end, to avoid problems when pushing int4 as 8 byte pool constant on stack
  {
    int4 dummy = 0;    
    (void)exe_store_pool_constant ((char *)&dummy, 4, 4);
  }
}

/******************************************************************************************/
