
// statem.c : statements analysis

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "types.h"
#include "token.h"
#include "lex.h"
#include "error.h"
#include "entities.h"
#include "type.h"
#include "typedecl.h"
#include "exp.h"
#include "decl.h"
#include "complete.h"

#include "statem.h"

//=================================================================================

static PENTITY surrounding_statement (PENTITY e)
{
  switch (e->kind)
  {
    case A_BLOCK_STATEMENT  :  return e->info->the_block_statement.outer;
    case AN_IF_STATEMENT    :  return e->info->the_if_statement.outer;
    case A_SWITCH_STATEMENT :  return e->info->the_switch_statement.outer;
    case A_WHILE_STATEMENT  :  return e->info->the_while_statement.outer;
    case A_FOR_STATEMENT    :  return e->info->the_for_statement.outer;
    default:   fatal_compiler_error ("surrounding_statement", &token.pos); return null;
  }
}

//=================================================================================

static void parse_asm_statement ()
{
  int8                    n;
  TEXT_POSITION           pos;
  A_CODE_STATEMENT_ENTITY cs;
  byte                    *buf;
  int                     size, len;
  PENTITY                 e;
  const int CHUNK = 16;

  if (!lexa.within_unsafe)
    semantic_error ("code statements are only allowed within unsafe regions", &token.pos);

  get_token();   // skip _asm

  if (token.kind != LEFT_ACCOLADE)
  {
    syntax_error ("'{' expected", &token.pos);
    skip_until (SEMICOLON, false);
    return;
  }

  get_token();   // skip '{'

  len = 0;
  size = CHUNK;
  buf = malloc (size);
  if (buf == null)
    fatal_out_of_memory_error ("code_statem(1)");

  while (token.kind != RIGHT_ACCOLADE && token.kind != LAST_TOKEN)
  {
    memcpy (&pos, &token.pos, sizeof(pos));
    n = parse_constant_int_or_uint_expression ();
    if (n < 0 || n > 255)
      syntax_error ("constant must be in range 0 .. 255", &pos);

    if (len == size)
    {
      byte *old;
      old = buf;
      buf = malloc (size + CHUNK);
      if (buf == null)
        fatal_out_of_memory_error ("code_statem(2)");
      memcpy (buf, old, size);
      free (old);
      size += CHUNK;
    }

    buf[len++] = (byte)n;
  }

  if (token.kind != RIGHT_ACCOLADE)
  {
    syntax_error ("'}' expected", &token.pos);
    return;
  }

  get_token();   // skip '}'


  // create statement entity

  memset (&cs, 0x00, sizeof(cs));
  store_location (&cs.loc);
  cs.code = buf;
  cs.nb_bytes = len;
  (void)append_new_entity (A_CODE_STATEMENT, &cs, L"", &pos);


  // find surrounding function

  e = surrounding_entity ();
  while (e->kind != A_FUNCTION_BODY)
    e = surrounding_statement (e);
  e->info->the_function_body.contains_code_statements = true;
}

//=================================================================================

static void parse_unused_statement ()
{
  TEXT_POSITION              pos;
  AN_UNUSED_STATEMENT_ENTITY us;
  PENTITY                    e;

  get_token();   // skip _unused

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("variable or parameter expected", &token.pos);
    skip_until (SEMICOLON, false);
    return;
  }

  memcpy (&pos, &token.pos, sizeof(pos));

  e = parse_expanded_name ();
  if (e == null)   // identifier not found (error already given)
    return;

  if (e->kind == A_PARAMETER)
  {
    e->info->the_parameter.is_used = true;
  }
  else if (e->kind == A_LOCAL_VARIABLE)
  {
    e->info->the_local_variable.is_used = true;
  }
  else
  {
    semantic_error ("variable or parameter name expected", &pos);
    skip_until (SEMICOLON, false);
    return;
  }

  memset (&us, 0x00, sizeof(us));
  memcpy (&us.pos, &pos, sizeof(TEXT_POSITION));
  us.item = e;
  (void)append_new_entity (AN_UNUSED_STATEMENT, &us, L"", &pos);
}

//=================================================================================

static void parse_clear_statement ()
{
  get_token();   // skip clear  

  for (;;)
  {
    CONTEXT                  context;
    PEXPRESSION              e;
    TEXT_POSITION            pos;
    A_CLEAR_STATEMENT_ENTITY cl;

    memset (&context, 0x00, sizeof(context));
    memcpy (&pos, &token.pos, sizeof(pos));

    parse_unary_expression (&context, &token.pos, null, &e);


    // check that e denotes a read/write object of a non-jagged type

    if (e->form != AN_OBJECT)
      semantic_error ("name must denote an object", &pos);
    else if (e->access != ACCESS_READWRITE)
      semantic_error ("name must denote a read/write object", &pos);
    else if (e->base_type_or_null != null && is_jagged_type (e->base_type_or_null))
      semantic_error ("name must not have a jagged type", &pos);
    else if (e->base_type_or_null != null &&
             (e->base_type_or_null->kind == AN_INCOMPLETE_TYPE ||
              e->base_type_or_null->kind == A_NULL_POINTER_TYPE ||
              e->base_type_or_null->kind == A_VOID_TYPE))
      semantic_error ("illegal object name", &pos);


    // create statement entity

    memset (&cl, 0x00, sizeof(cl));
    cl.name = e;
    store_location (&cl.loc);
    (void)append_new_entity (A_CLEAR_STATEMENT, &cl, L"", &pos);


    // parse commas

    if (token.kind != COMMA && token.kind != SEMICOLON)
    {
      syntax_error (", or ; expected", &token.pos);
      return;
    }

    if (token.kind == SEMICOLON)
      break;

    get_token ();  // skip comma
  }
}

//=================================================================================

static void parse_statement_starting_with_exp (TEXT_POSITION *pos0,
                                               PENTITY       prefix,
                                               bool          pre_post_statement_allowed,
                                               bool          function_call_allowed)
{
  TEXT_POSITION pos, assign_pos;
  bool          function_call_with_return_value_expected = false;
  PEXPRESSION   e, e2;
  CONTEXT       context;
  TOKEN_KIND    tk;
  ASSIGNMENT_OP op;

  memcpy (&pos, pos0, sizeof(TEXT_POSITION));

  if (function_call_allowed && prefix == null && token.kind == LEFT_PARENTHESIS)    // (void)
  {
    get_look_ahead_token ();
    if (look_ahead_token.kind == TOKEN_void)
    {
      get_token ();  // skip '('
      get_token ();  // skip void

      if (token.kind != RIGHT_PARENTHESIS)
        syntax_error ("')' expected", &token.pos);
      else
        get_token ();  // skip ')'

      function_call_with_return_value_expected = true;
      memcpy (&pos, &token.pos, sizeof(TEXT_POSITION));
    }
  }


  memset (&context, 0x00, sizeof(context));

  parse_unary_expression (&context, &pos, prefix, &e);


  if (function_call_with_return_value_expected)
  {
    if (e->kind != A_FUNCTION_CALL)
      semantic_error ("a function call is expected here", &pos);
    else if (e->base_type_or_null == type_void)
      semantic_error ("(void) is only allowed for a function that returns a value", &pos);
  }
  else    // check either for function call or post/pre inc/dec node
  {
    if (function_call_allowed && e->kind == A_FUNCTION_CALL)
    {
    }
    else if (pre_post_statement_allowed && e->kind == AN_OPERATOR_VALUE)
    {
      KIND_OPERATOR op;

      op = e->info.operator_value_info.op;

      if (op != OP_PRE_DEC && op != OP_PRE_INC && op != OP_POST_DEC && op != OP_POST_INC)
      {
        if (function_call_allowed)
          semantic_error ("assignment, function call or increment/decrement statement expected", &pos);
        else
          semantic_error ("assignment or increment/decrement statement expected", &pos);
      }
    }
    else if (e->form == AN_OBJECT)
    {
      if (e->access != ACCESS_READWRITE)
        semantic_error ("name must denote a read/write object", &pos);
      else if (e->base_type_or_null != null && is_jagged_type (e->base_type_or_null))
        semantic_error ("name must not have a jagged type", &pos);
      else if (e->base_type_or_null != null && is_limited_type (e->base_type_or_null))
        semantic_error ("name must not have a limited type", &pos);
      else if (e->base_type_or_null != null &&
               (e->base_type_or_null->kind == AN_INCOMPLETE_TYPE  ||
                e->base_type_or_null->kind == AN_OPAQUE_TYPE ||
                e->base_type_or_null->kind == A_NULL_POINTER_TYPE ||
                e->base_type_or_null->kind == A_VOID_TYPE))
        semantic_error ("illegal object name", &pos);
    }
    else
    {
      if (function_call_allowed)
        semantic_error ("assignment, function call or increment/decrement statement expected", &pos);
      else if (pre_post_statement_allowed)
        semantic_error ("assignment or increment/decrement statement expected", &pos);
      else
        semantic_error ("assignment statement expected", &pos);
    }
  }


  if (e->form == AN_OBJECT)
  {
    switch (token.kind)
    {
      case ASSIGN:
      case PLUS_EQUAL:
      case MINUS_EQUAL:
      case STAR_EQUAL:
      case SLASH_EQUAL:
      case PERCENT_EQUAL:
      case SHIFT_LEFT_EQUAL:
      case SHIFT_RIGHT_EQUAL:
      case AMPERSAND_EQUAL:
      case VERTICAL_BAR_EQUAL:
      case CARET_EQUAL:
        tk = token.kind;
        break;

      default:
        semantic_error ("'=' or other assignment symbol expected", &token.pos);
        return;
    }

    memcpy (&assign_pos, &token.pos, sizeof(assign_pos));
    get_token ();  // skip assign symbol


    // parse expression

    memset (&context, 0x00, sizeof(context));
    context.base_type_or_null = e->base_type_or_null;
    memcpy (&context.constraint, &e->constraint, sizeof(CONSTRAINT));

    parse_expression (&context, &token.pos, null, &e2);


    // check that expression matches designator

    op = _ASSIGN;   // default

    if (e->base_type_or_null != null && e2->base_type_or_null != null)
    {
      PENTITY t1, t2;

      t1 = e->base_type_or_null;
      t2 = e2->base_type_or_null;

      switch (tk)
      {
        case ASSIGN:
          // returns 0 if OK, -1 if types are not compatible
          // error message is generated in function
          (void) check_assignment_context_compatibility (&context, e2, &assign_pos);

          op = _ASSIGN;
          break;


        case PLUS_EQUAL:
        case MINUS_EQUAL:
          if (t1->kind == AN_ENUMERATION_TYPE && t2->kind == AN_INTEGER_TYPE)
          {
            op = (tk == PLUS_EQUAL) ? _ASSIGN_ADD_INT : _ASSIGN_SUB_INT;
          }
          else if (t1->kind == AN_UNSAFE_POINTER_TYPE && t2->kind == AN_INTEGER_TYPE)
          {
            op = (tk == PLUS_EQUAL) ? _ASSIGN_ADD_PTR_INT : _ASSIGN_SUB_PTR_INT;
            if (t2 == type_long)
              semantic_error ("right operand must not have type long", &assign_pos);
          }
          else if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
          {
            (void) check_assignment_context_compatibility (&context, e2, &assign_pos);
            op = (tk == PLUS_EQUAL) ? _ASSIGN_ADD_INT : _ASSIGN_SUB_INT;
          }
          else if (t1->kind == A_FLOAT_TYPE && t2->kind == A_FLOAT_TYPE)
          {
            (void) check_assignment_context_compatibility (&context, e2, &assign_pos);
            op = (tk == PLUS_EQUAL) ? _ASSIGN_ADD_FLOAT : _ASSIGN_SUB_FLOAT;
          }
          else
          {
            semantic_error ("incompatible expression", &assign_pos);
            op = _ASSIGN_ADD_INT;
          }
          break;


        case STAR_EQUAL:
        case SLASH_EQUAL:
        case PERCENT_EQUAL:

          if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
          {
            bool is_signed;

            if (tk == SLASH_EQUAL || tk == PERCENT_EQUAL)
            {
              if (e2->kind == A_CONST_INTEGER_VALUE && e2->info.const_integer_value_info.value == 0)
                semantic_error ("division by zero", &assign_pos);
            }

            // source type must fit into target type
            (void) check_assignment_context_compatibility (&context, e2, &assign_pos);


            is_signed = integer_info[t1->info->the_integer_type.type].is_signed;
            if (tk == STAR_EQUAL)
              op = is_signed ? _ASSIGN_MULT_INT_SIGNED : _ASSIGN_MULT_INT_UNSIGNED;
            else if (tk == SLASH_EQUAL)
              op = is_signed ? _ASSIGN_DIV_INT_SIGNED  : _ASSIGN_DIV_INT_UNSIGNED;
            else
              op = is_signed ? _ASSIGN_MOD_INT_SIGNED  : _ASSIGN_MOD_INT_UNSIGNED ;
          }
          else if (t1->kind == A_FLOAT_TYPE && t2->kind == A_FLOAT_TYPE)
          {
            if (tk == PERCENT_EQUAL)
              semantic_error ("operator '%=' is not allowed for float types", &assign_pos);
            else if (tk == SLASH_EQUAL)
            {
              if (e2->kind == A_CONST_FLOAT_VALUE && e2->info.const_float_value_info.value == 0.0)
                semantic_error ("division by zero", &pos);
            }

            // source type must fit into target type
            (void) check_assignment_context_compatibility (&context, e2, &assign_pos);

            op = (tk == STAR_EQUAL) ? _ASSIGN_MULT_FLOAT : _ASSIGN_DIV_FLOAT;
          }
          else
          {
            semantic_error ("incompatible expression", &assign_pos);
            op = _ASSIGN_MULT_FLOAT;
          }
          break;


        case SHIFT_LEFT_EQUAL:
        case SHIFT_RIGHT_EQUAL:

          if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
          {
            int8 v2;
            bool is_signed;

            // source type must fit into target type
            if (check_assignment_context_compatibility (&context, e2, &assign_pos) == 0)
            {
              if (e2->kind == A_CONST_INTEGER_VALUE)
              {
                v2 = e2->info.const_integer_value_info.value;

                if (t1 == type_long || t1 == type_int_literal)
                {
                  if (v2 < 0 || v2 > 63)
                    semantic_error ("right operand must be in range 0..63", &assign_pos);
                }
                else
                {
                  if (v2 < 0 || v2 > 31)
                    semantic_error ("right operand must be in range 0..31", &assign_pos);
                }
              }
            }

            is_signed = integer_info[t1->info->the_integer_type.type].is_signed;

            if (tk == SHIFT_LEFT_EQUAL)
              op = is_signed ? _ASSIGN_SHIFT_LEFT_SIGNED  : _ASSIGN_SHIFT_LEFT_UNSIGNED;
            else
              op = is_signed ? _ASSIGN_SHIFT_RIGHT_SIGNED : _ASSIGN_SHIFT_RIGHT_UNSIGNED;
          }
          else
          {
            semantic_error ("incompatible expression", &assign_pos);
            op = _ASSIGN_SHIFT_LEFT_SIGNED;
          }
          break;


        case AMPERSAND_EQUAL:
        case VERTICAL_BAR_EQUAL:
        case CARET_EQUAL:

          if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
          {
            // source type must fit into target type
            (void) check_assignment_context_compatibility (&context, e2, &assign_pos);

            if (tk == AMPERSAND_EQUAL)
              op = _ASSIGN_BITAND;
            else if (tk == VERTICAL_BAR_EQUAL)
              op = _ASSIGN_BITOR;
            else
              op = _ASSIGN_BITXOR;
          }
          else if (t1 == type_bool && t2 == type_bool)
          {
            // source type must fit into target type
            (void) check_assignment_context_compatibility (&context, e2, &assign_pos);

            if (tk == AMPERSAND_EQUAL)
              op = _ASSIGN_AND;
            else if (tk == VERTICAL_BAR_EQUAL)
              op = _ASSIGN_OR;
            else
              op = _ASSIGN_XOR;
          }
          else
          {
            semantic_error ("incompatible expression", &assign_pos);
            op = _ASSIGN_BITAND;
          }
          break;

        default:
          semantic_error ("'=' or other assignment symbol expected", &token.pos);
          return;
      }
    }
  }


  // create statement entity

  if (e->kind == A_FUNCTION_CALL)
  {
    A_FUNCTION_CALL_STATEMENT_ENTITY en;

    memset (&en, 0x00, sizeof(en));
    en.name = e;
    store_location (&en.loc);
    (void)append_new_entity (A_FUNCTION_CALL_STATEMENT, &en, L"", &pos);
  }
  else if (e->form == AN_OBJECT)
  {
    AN_ASSIGNMENT_STATEMENT_ENTITY en;

    memset (&en, 0x00, sizeof(en));
    en.name = e;
    en.op = op;
    en.value = e2;
    store_location (&en.loc);
    (void)append_new_entity (AN_ASSIGNMENT_STATEMENT, &en, L"", &pos);
  }
  else
  {
    A_PRE_OR_POSTFIX_STATEMENT_ENTITY en;

    memset (&en, 0x00, sizeof(en));
    en.name = e;
    store_location (&en.loc);
    (void)append_new_entity (A_PRE_OR_POSTFIX_STATEMENT, &en, L"", &pos);
  }
}

//=================================================================================

static void parse_return_statement ()
{
  PENTITY                   e;
  PEXPRESSION               exp;
  CONTEXT                   context;
  A_RETURN_STATEMENT_ENTITY en;
  TEXT_POSITION             pos0, pos;


  memcpy (&pos0, &token.pos, sizeof(pos0));
  get_token ();  // skip 'return'


  // find surrounding function and its return type

  e = surrounding_entity ();

  while (e->kind != A_FUNCTION_BODY)
    e = surrounding_statement (e);

  e = e->info->the_function_body.to_function_declaration;

  e = e->info->the_function_declaration.to_type;

  e = e->info->the_function_pointer_type.return_type;


  convert_type_into_context (e, &context);


  if (token.kind == SEMICOLON)
  {
    if (e != type_void)
      syntax_error ("return expression expected", &token.pos);
    exp = null;
  }
  else
  {
    if (e == type_void)
      syntax_error (" ; expected", &token.pos);

    memcpy (&pos, &token.pos, sizeof(pos));
    parse_expression (&context, &token.pos, null, &exp);

    if (e != type_void)
      (void)check_assignment_context_compatibility (&context, exp, &pos);
  }


  memset (&en, 0x00, sizeof(en));
  en.value = exp;
  en.type  = e;
  en.outer = surrounding_entity ();
  memcpy (&en.pos, &pos0, sizeof(TEXT_POSITION));
  store_location (&en.loc);
  (void)append_new_entity (A_RETURN_STATEMENT, &en, L"", &token.pos);
}

//=================================================================================

static void parse_break_statement ()
{
  PENTITY                  e;
  A_BREAK_STATEMENT_ENTITY en;


  // check if we're within a "while", "for", or "switch" statement

  e = surrounding_entity ();

  for (;;)
  {
    if (e->kind == A_FUNCTION_BODY || e->kind == A_WHILE_STATEMENT ||
        e->kind == A_FOR_STATEMENT || e->kind == A_SWITCH_STATEMENT)
      break;
    e = surrounding_statement (e);
  }

  if (e->kind == A_FUNCTION_BODY)
    semantic_error ("a break statement is only allowed in while/for/switch statements", &token.pos);

  memset (&en, 0x00, sizeof(en));
  en.outer = surrounding_entity ();
  memcpy (&en.pos, &token.pos, sizeof(TEXT_POSITION));
  (void)append_new_entity (A_BREAK_STATEMENT, &en, L"", &token.pos);

  get_token ();  // skip 'break'
}

//=================================================================================

static void parse_continue_statement ()
{
  PENTITY                     e;
  A_CONTINUE_STATEMENT_ENTITY en;


  // check if we're within a "while" or "for" statement

  e = surrounding_entity ();

  for (;;)
  {
    if (e->kind == A_FUNCTION_BODY || e->kind == A_WHILE_STATEMENT || e->kind == A_FOR_STATEMENT)
      break;
    e = surrounding_statement (e);
  }

  if (e->kind == A_FUNCTION_BODY)
    semantic_error ("a break statement is only allowed in while/for statements", &token.pos);

  memset (&en, 0x00, sizeof(en));
  en.outer = surrounding_entity ();
  memcpy (&en.pos, &token.pos, sizeof(TEXT_POSITION));
  (void)append_new_entity (A_CONTINUE_STATEMENT, &en, L"", &token.pos);

  get_token ();  // skip 'continue'
}

//=================================================================================

static void parse_free_statement ()
{
  PEXPRESSION               exp;
  CONTEXT                   context;
  PENTITY                   t;
  TEXT_POSITION             pos;
  A_FREE_STATEMENT_ENTITY   en;

  get_token ();  // skip 'free'


  memset (&context, 0x00, sizeof(context));

  memcpy (&pos, &token.pos, sizeof(pos));
  parse_expression (&context, &token.pos, null, &exp);

  t = exp->base_type_or_null;
  if (t != null && t->kind != A_POINTER_TYPE && t->kind != A_NULL_POINTER_TYPE)
    semantic_error ("a pointer value is expected here", &pos);

  memset (&en, 0x00, sizeof(en));
  en.value = exp;
  store_location (&en.loc);
  (void)append_new_entity (A_FREE_STATEMENT, &en, L"", &pos);
}

//=================================================================================

static void parse_assert_statement ()
{
  PEXPRESSION                 exp;
  CONTEXT                     context;
  PENTITY                     t;
  TEXT_POSITION               pos;
  AN_ASSERT_STATEMENT_ENTITY  en;

  get_token ();  // skip 'assert'


  memset (&context, 0x00, sizeof(context));

  memcpy (&pos, &token.pos, sizeof(pos));
  parse_expression (&context, &token.pos, null, &exp);

  t = exp->base_type_or_null;
  if (t != null && t != type_bool)
    semantic_error ("a bool expression is expected here", &pos);

  if (exp->kind == A_CONST_ENUMERATION_VALUE && exp->info.const_enumeration_value_info.value == 0)
    semantic_error ("assertion failed", &pos);

  memset (&en, 0x00, sizeof(en));
  en.value = exp;
  store_location (&en.loc);
  (void)append_new_entity (AN_ASSERT_STATEMENT, &en, L"", &pos);
}

//=================================================================================

static void parse_abort_statement ()
{
  AN_ABORT_STATEMENT_ENTITY  en;
  TEXT_POSITION              pos;

  memcpy (&pos, &token.pos, sizeof(pos));

  get_token ();  // skip 'abort'

  memset (&en, 0x00, sizeof(en));
  store_location (&en.loc);
  memcpy (&en.pos, &pos, sizeof(pos));
  (void)append_new_entity (AN_ABORT_STATEMENT, &en, L"", &pos);
}

//=================================================================================

static void parse_sleep_statement ()
{
  PEXPRESSION               exp;
  CONTEXT                   context;
  PENTITY                   t;
  TEXT_POSITION             pos;
  A_SLEEP_STATEMENT_ENTITY  en;

  get_token ();  // skip 'sleep'


  memset (&context, 0x00, sizeof(context));

  memcpy (&pos, &token.pos, sizeof(pos));
  parse_expression (&context, &token.pos, null, &exp);

  t = exp->base_type_or_null;
  if (t != null && t->kind != AN_INTEGER_TYPE && t->kind != A_FLOAT_TYPE)
    semantic_error ("a numeric expression is expected here", &pos);

  if (exp->kind == A_CONST_INTEGER_VALUE)
  {
    if (exp->info.const_integer_value_info.value < -86400 ||
        exp->info.const_integer_value_info.value > 86400)
      semantic_error ("value must be in range -86400 .. +86400", &pos);
  }

  if (exp->kind == A_CONST_FLOAT_VALUE)
  {
    if (exp->info.const_float_value_info.value < -86400.0 ||
        exp->info.const_float_value_info.value > 86400.0)
      semantic_error ("value must be in range -86400 .. +86400", &pos);
  }

  memset (&en, 0x00, sizeof(en));
  en.value = exp;
  store_location (&en.loc);
  (void)append_new_entity (A_SLEEP_STATEMENT, &en, L"", &pos);
}

//=================================================================================

static void parse_block_statement ()
{
  A_BLOCK_STATEMENT_ENTITY en;
  PENTITY                  out_prefix;
  TEXT_POSITION            out_pos;
  PENTITY                  eblock;

  get_token();      // skip '{'

  memset (&en, 0x00, sizeof(en));
  en.inner = new_region ();
  en.outer = surrounding_entity ();

  eblock = append_new_entity (A_BLOCK_STATEMENT, &en, L"", &token.pos);

  create_region_level ();
  append_region (en.inner);


  // 'out_prefix' is an "out" parameter that can either be null,
  // or denote an already parsed expanded name that is not a type_name.
  parse_local_declarations (&out_pos, &out_prefix);

  check_local_completion (eblock->info->the_block_statement.inner->entities.first);

  parse_statements (&out_pos, out_prefix);

  if (token.kind == RIGHT_ACCOLADE)
    get_token();
  else
  {
    syntax_error ("'}' expected", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
  }

  unlink_region_level ();
}

//=================================================================================

static void parse_statement (TEXT_POSITION *pos,
                             PENTITY       prefix,
                             bool          within_if_true_branch);

//=================================================================================

static PEXPRESSION parse_an_expression ()
{
  CONTEXT       context;
  PEXPRESSION   exp;

  memset (&context, 0x00, sizeof(context));
  parse_expression (&context, &token.pos, null, &exp);

  return exp;
}

//=================================================================================

static PEXPRESSION parse_bool_expression ()
{
  PENTITY       t;
  PEXPRESSION   exp;
  TEXT_POSITION pos;

  memcpy (&pos, &token.pos, sizeof(pos));

  exp = parse_an_expression ();

  t = exp->base_type_or_null;
  if (t != null && t != type_bool)
    semantic_error ("a bool expression is expected here", &pos);

  return exp;
}

//=================================================================================

static PEXPRESSION parse_bool_parenthesized_expression ()
{
  PEXPRESSION  exp;

  if (token.kind == LEFT_PARENTHESIS)
    get_token ();  // skip '('
  else
    syntax_error ("'(' is expected here", &token.pos);

  exp = parse_bool_expression ();

  if (token.kind == RIGHT_PARENTHESIS)
    get_token ();  // skip ')'
  else
    syntax_error ("')' is expected here", &token.pos);

  return exp;
}

//=================================================================================

static PEXPRESSION parse_parenthesized_expression ()
{
  PEXPRESSION  exp;

  if (token.kind == LEFT_PARENTHESIS)
    get_token ();  // skip '('
  else
    syntax_error ("'(' is expected here", &token.pos);

  exp = parse_an_expression ();

  if (token.kind == RIGHT_PARENTHESIS)
    get_token ();  // skip ')'
  else
    syntax_error ("')' is expected here", &token.pos);

  return exp;
}

//=================================================================================

static void parse_if_statement (bool within_if_true_branch)
{
  PEXPRESSION            exp;
  AN_IF_STATEMENT_ENTITY en;


  get_token ();  // skip 'if'

  memset (&en, 0x00, sizeof(en));
  store_location (&en.loc);

  exp = parse_bool_parenthesized_expression ();

  en.condition = exp;
  en.true_branch = new_region ();
  en.false_branch = new_region ();
  en.outer = surrounding_entity ();

  (void) append_new_entity (AN_IF_STATEMENT, &en, L"", &token.pos);


  create_region_level ();
  append_region (en.true_branch);

  parse_statement (&token.pos, null, /*within_if_true_branch=*/ true);

  unlink_region_level ();


  if (token.kind == TOKEN_else)
  {
    if (within_if_true_branch)
      syntax_error ("ambiguous 'else', please enclose if-statement in {}", &token.pos);

    get_token();


    create_region_level ();
    append_region (en.false_branch);

    parse_statement (&token.pos, null, /*within_if_true_branch=*/ false);

    unlink_region_level ();
  }
}

//=================================================================================

static bool is_statement_breaking_control_flow (PENTITY e0)
{
  PENTITY e;

  e = e0;

  for (;;)
  {
    if (e == null)
      return false;

    if (e->kind == A_RETURN_STATEMENT   || e->kind == A_BREAK_STATEMENT ||
        e->kind == A_CONTINUE_STATEMENT || e->kind == AN_ABORT_STATEMENT)
      return true;

    if (e->kind != A_BLOCK_STATEMENT)
      return false;

    e = e->info->the_block_statement.inner->entities.last;
  }
}

//=================================================================================

static void parse_case_alternative (PENTITY     sw,       // sw = switch_statement entity
                                    PENTITY     sw_type,  // type of switch expression
                                    BINARY_TREE *btree)
{
  CASE_CONSTANT      *list, *nl;
  TEXT_POSITION      pos, pos_last_stat;
  PEXPRESSION        exp;
  PENTITY            t;
  INTEGER_INFO       *pi;
  int8               case_cte;
  bool               is_valid;
  FLOW_ALTERNATIVE   *fl;
  int                rc;

  list = null;

  while (token.kind == TOKEN_case)
  {
    get_token ();  // skip 'case'

    memcpy (&pos, &token.pos, sizeof(pos));

    exp = parse_an_expression ();
    t = exp->base_type_or_null;

    case_cte = 0;
    is_valid = false;

    if (!is_constant_exp (exp))
    {
      semantic_error ("a constant expression is expected here", &pos);
    }
    else if (sw_type != null && t != null)    // check that types match
    {
      if (sw_type->kind == AN_ENUMERATION_TYPE)
      {
        if (!types_are_equal (sw_type, t))
          semantic_error ("constant has bad type", &pos);
        else
        {
          case_cte = exp->info.const_enumeration_value_info.value;

          if (case_cte < 0 || case_cte > sw_type->info->the_enumeration_type.last)
            semantic_error ("constant is out of range", &pos);
          else
            is_valid = true;
        }
      }
      else if (sw_type->kind == AN_INTEGER_TYPE)
      {
        if (t->kind != AN_INTEGER_TYPE)
          semantic_error ("constant has bad type", &pos);
        else
        {
          // check that case constant is in switch's type range

          case_cte = exp->info.const_integer_value_info.value;
          pi = &integer_info[sw_type->info->the_integer_type.type];

          if (case_cte < pi->min || case_cte > pi->max)
            semantic_error ("constant is out of range", &pos);
          else
            is_valid = true;
        }
      }
    }


    if (is_valid)
    {
      // update count

      sw->info->the_switch_statement.count++;


      // check that all constant values are distinct

      rc = insert_btree (btree, &case_cte, sizeof(case_cte));
      if (rc == 0)
        ;
      else if (rc == BT_DUPLICATE_KEY)
        semantic_error ("constant given twice", &pos);
      else
        fatal_compiler_error ("insert_case_key", &pos);


      // insert case constant in list

      nl = malloc (sizeof(CASE_CONSTANT));
      if (nl == NULL)
        fatal_out_of_memory_error ("case_constant");
      memset (nl, 0x00, sizeof(CASE_CONSTANT));
      nl->value = case_cte;
      nl->next  = list;

      list = nl;
    }


    if (token.kind == COLON)
      get_token ();  // skip ':'
    else
      syntax_error ("':' is expected here", &token.pos);
  }



  fl = malloc (sizeof(FLOW_ALTERNATIVE));
  if (fl == NULL)
    fatal_out_of_memory_error ("flow_alternative");

  memset (fl, 0x00, sizeof(FLOW_ALTERNATIVE));
  fl->is_default = false;
  fl->cte_list   = list;
  fl->inner      = new_region ();
  fl->next       = null;

  create_region_level ();
  append_region (fl->inner);



  if (sw->info->the_switch_statement.first_alt == null)
    sw->info->the_switch_statement.first_alt = fl;
  else  
    sw->info->the_switch_statement.last_alt->next = fl;

  sw->info->the_switch_statement.last_alt = fl;


  // parse statements

  memcpy (&pos_last_stat, &token.pos, sizeof(pos_last_stat));

  while (token.kind != TOKEN_case &&
         token.kind != TOKEN_default &&
         token.kind != RIGHT_ACCOLADE &&
         token.kind != LAST_TOKEN)
  {
    memcpy (&pos_last_stat, &token.pos, sizeof(pos_last_stat));
    parse_statement (&token.pos, null, false);
  }


  // the last statement of a case alternative must be either a
  // break, continue, return statement, abort statement or nested block statement.
  // (it is not allowed to "fall through" into the next case)

  if (!is_statement_breaking_control_flow (fl->inner->entities.last))
    semantic_error ("break, return or continue is expected as last statement in a case", &pos_last_stat);

  unlink_region_level ();
}

//=================================================================================

//  default_case_alternative ::= "default"  ":"
//                                statements

static void parse_default_alternative (PENTITY sw)    // sw = switch_statement entity
{
  TEXT_POSITION     pos_last_stat;
  FLOW_ALTERNATIVE  *fl;

  get_token ();  // skip 'default'

  if (token.kind == COLON)
    get_token ();  // skip ':'
  else
    syntax_error ("':' is expected here", &token.pos);


  fl = malloc (sizeof(FLOW_ALTERNATIVE));
  if (fl == NULL)
    fatal_out_of_memory_error ("flow_alternative");

  memset (fl, 0x00, sizeof(FLOW_ALTERNATIVE));
  fl->is_default = true;
  fl->cte_list   = null;
  fl->inner      = new_region ();
  fl->next       = null;

  create_region_level ();
  append_region (fl->inner);



  if (sw->info->the_switch_statement.first_alt == null)
    sw->info->the_switch_statement.first_alt = fl;
  else  
    sw->info->the_switch_statement.last_alt->next = fl;

  sw->info->the_switch_statement.last_alt = fl;


  // parse statements

  memcpy (&pos_last_stat, &token.pos, sizeof(pos_last_stat));

  while (token.kind != TOKEN_case &&
         token.kind != TOKEN_default &&
         token.kind != RIGHT_ACCOLADE &&
         token.kind != LAST_TOKEN)
  {
    memcpy (&pos_last_stat, &token.pos, sizeof(pos_last_stat));
    parse_statement (&token.pos, null, false);
  }


  // the last statement of a case alternative must be either a
  // break, continue, return statement, abort statement or nested block statement.
  // (it is not allowed to "fall through" into the next case)

  if (!is_statement_breaking_control_flow (fl->inner->entities.last))
    semantic_error ("break, return or continue is expected as last statement in a case", &pos_last_stat);

  unlink_region_level ();
}

//=================================================================================

static int sw_compare (void               *user,
                       const int8         *data1,
                       const unsigned int size1,
                       const int8         *data2,
                       const unsigned int size2)
{
  if (*data1 < *data2)
    return -1;
  if (*data1 > *data2)
    return +1;
  return 0;
}

//=================================================================================

static void parse_switch_statement ()
{
  TEXT_POSITION             pos;
  PEXPRESSION               exp;
  PENTITY                   e, t;
  A_SWITCH_STATEMENT_ENTITY en;
  BINARY_TREE               tree;

  get_token ();  // skip 'switch'

  memcpy (&pos, &token.pos, sizeof(pos));

  memset (&en, 0x00, sizeof(en));
  store_location (&en.loc);

  exp = parse_parenthesized_expression ();

  t = exp->base_type_or_null;
  if (t != null && t->kind != AN_ENUMERATION_TYPE && t->kind != AN_INTEGER_TYPE)
    semantic_error ("an enumeration or integer expression is expected here", &pos);


  en.value     = exp;
  en.count     = 0;           // nb case constants
  en.first_alt = null;
  en.last_alt  = null;
  en.outer     = surrounding_entity ();

  e = append_new_entity (A_SWITCH_STATEMENT, &en, L"", &token.pos);


  if (token.kind == LEFT_ACCOLADE)
    get_token ();  // skip '{'
  else
    syntax_error ("'{' is expected here", &token.pos);


  create_btree (&tree, NULL, sw_compare);


  while (token.kind == TOKEN_case)
  {
    parse_case_alternative (e, t, &tree);
  }


  close_btree (&tree);


  if (token.kind == TOKEN_default)
  {
    parse_default_alternative (e);
  }
  else
  {
    syntax_error ("'default:' is expected here", &token.pos);
  }


  if (token.kind == RIGHT_ACCOLADE)
    get_token ();  // skip '}'
  else
    syntax_error ("'}' is expected here", &token.pos);
}

//=================================================================================

static void parse_while_statement (bool within_if_true_branch)
{
  TEXT_POSITION            pos;
  PEXPRESSION              exp;
  A_WHILE_STATEMENT_ENTITY en;


  memcpy (&pos, &token.pos, sizeof(pos));
  get_token ();  // skip 'while'

  memset (&en, 0x00, sizeof(en));
  store_location (&en.loc);

  exp = parse_bool_parenthesized_expression ();

  en.condition = exp;
  en.inner = new_region ();
  en.outer = surrounding_entity ();
  memcpy (&en.pos, &pos, sizeof(pos));
  
  (void) append_new_entity (A_WHILE_STATEMENT, &en, L"", &token.pos);


  create_region_level ();
  append_region (en.inner);

  parse_statement (&token.pos, null, within_if_true_branch);

  unlink_region_level ();
}

//=================================================================================

static void parse_for_statement (bool within_if_true_branch)
{
  TEXT_POSITION          pos;
  A_FOR_STATEMENT_ENTITY en;
  PENTITY                e;

  memcpy (&pos, &token.pos, sizeof(pos));
  get_token ();  // skip 'for'

  memset (&en, 0x00, sizeof(en));
  en.pre        = new_region ();
  en.exp_region = new_region ();
  en.condition  = null;
  en.post       = new_region ();
  en.inner      = new_region ();
  en.outer      = surrounding_entity ();
  memcpy (&en.pos, &pos, sizeof(pos));

  e = append_new_entity (A_FOR_STATEMENT, &en, L"", &token.pos);

  if (token.kind == LEFT_PARENTHESIS)
    get_token ();  // skip '('
  else
    syntax_error ("'(' is expected here", &token.pos);

  if (token.kind != SEMICOLON && token.kind != LAST_TOKEN)
  {
    create_region_level ();
    append_region (en.pre);

    for (;;)
    {
      // parse assignment statement
      parse_statement_starting_with_exp (&token.pos, null, false, false);

      if (token.kind != COMMA)
        break;

      get_token ();  // skip comma
    }

    unlink_region_level ();
  }


  if (token.kind == SEMICOLON)
    get_token ();  // skip ';'
  else
    syntax_error ("';' is expected here", &token.pos);


  store_location (&e->info->the_for_statement.loc);

  if (token.kind != SEMICOLON)    // there is a condition
  {
    create_region_level ();
    append_region (en.exp_region);

    e->info->the_for_statement.condition = parse_bool_expression ();

    unlink_region_level ();
  }


  if (token.kind == SEMICOLON)
    get_token ();  // skip ';'
  else
    syntax_error ("';' is expected here", &token.pos);


  if (token.kind != RIGHT_PARENTHESIS && token.kind != LAST_TOKEN)
  {
    create_region_level ();
    append_region (en.post);

    for (;;)
    {
      // parse assignment or post/pre statement
      parse_statement_starting_with_exp (&token.pos, null, true, false);

      if (token.kind != COMMA)
        break;

      get_token ();  // skip comma
    }

    unlink_region_level ();
  }


  if (token.kind == RIGHT_PARENTHESIS)
    get_token ();  // skip ')'
  else
    syntax_error ("')' is expected here", &token.pos);



  create_region_level ();
  append_region (en.inner);

  parse_statement (&token.pos, null, within_if_true_branch);

  unlink_region_level ();
}

//=================================================================================

// 'prefix' can either be null,
// or denote an already parsed expanded name that is not a type_name.

static void parse_statement (TEXT_POSITION *pos,
                             PENTITY       prefix,
                             bool          within_if_true_branch)
{
  if (prefix != null)
  {
    parse_statement_starting_with_exp (pos, prefix, true, true);
    parse_semicolon ();
  }
  else
  {
    switch (token.kind)
    {
      case SEMICOLON:     // null statement
        parse_semicolon ();
        break;

      case TOKEN_asm:
        parse_asm_statement ();
        parse_semicolon ();
        break;

      case TOKEN_unused:
        parse_unused_statement ();
        parse_semicolon ();
        break;

      case TOKEN_clear:
        parse_clear_statement ();
        parse_semicolon ();
        break;

      case IDENTIFIER:
        {
          TEXT_POSITION pos0;
          PENTITY       e;

          memcpy (&pos0, &token.pos, sizeof(pos0));

          e = parse_expanded_name ();
          if (e == null)    // could not find entity (error already given)
          {
            skip_until (SEMICOLON, true);
            break;
          }
          parse_statement_starting_with_exp (&pos0, e, true, true);
          parse_semicolon ();
        }
        break;

      case STAR:                 // *p = 0;
      case PLUS_PLUS:            // ++exp;
      case MINUS_MINUS:          // --exp;
      case LEFT_PARENTHESIS:     // (void)function_call   or   (int *)p = 0;
        parse_statement_starting_with_exp (&token.pos, null, true, true);
        parse_semicolon ();
        break;

      case TOKEN_return:
        parse_return_statement ();
        parse_semicolon ();
        break;

      case TOKEN_break:
        parse_break_statement ();
        parse_semicolon ();
        break;

      case TOKEN_continue:
        parse_continue_statement ();
        parse_semicolon ();
        break;

      case TOKEN_free:
        parse_free_statement ();
        parse_semicolon ();
        break;

      case TOKEN_abort:
        parse_abort_statement ();
        parse_semicolon ();
        break;

      case TOKEN_assert:
        parse_assert_statement ();
        parse_semicolon ();
        break;

      case TOKEN_sleep:
        parse_sleep_statement ();
        parse_semicolon ();
        break;

      case LEFT_ACCOLADE:
        parse_block_statement ();
        break;

      case TOKEN_if:
        parse_if_statement (within_if_true_branch);
        break;

      case TOKEN_switch:
        parse_switch_statement ();
        break;

      case TOKEN_while:
        parse_while_statement (within_if_true_branch);
        break;

      case TOKEN_for:
        parse_for_statement (within_if_true_branch);
        break;

      default:
        semantic_error ("statement expected", &token.pos);
        parse_semicolon ();
        break;
    }
  }
}

//=================================================================================

// 'prefix' can either be null,
// or denote an already parsed expanded name that is not a type_name.

void parse_statements (TEXT_POSITION *pos, PENTITY prefix)
{
  if (prefix != null)
    parse_statement (pos, prefix, false);

  while (token.kind != RIGHT_ACCOLADE && token.kind != LAST_TOKEN)
    parse_statement (&token.pos, null, false);
}

//=================================================================================
