
// type.c : type category tests

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "types.h"
#include "goptions.h"
#include "token.h"
#include "entities.h"
#include "lex.h"
#include "error.h"
#include "type.h"

/*****************************************************************************/

bool g_all_full_types_are_visible;   // set to true by code generator

/*****************************************************************************/

// resolves renamed, incomplete, opaque and generic types when their full type is visible.
// returns null if t is null.

PENTITY complete_type_of (PENTITY t)
{
  PENTITY p;

  if (t == null)    // error already given
    return null;

  p = t;

  for (;;)
  {
    if (p->kind == A_RENAMED_TYPE)
    {
      p = p->info->the_renamed_type.actual_type;
    }
    else if (p->kind == AN_INCOMPLETE_TYPE &&
             p->info->the_incomplete_type.full_type != null &&
             (p->info->the_incomplete_type.is_full_type_visible || g_all_full_types_are_visible))
    {
      p = p->info->the_incomplete_type.full_type;
    }
    else if (p->kind == AN_OPAQUE_TYPE &&
             p->info->the_opaque_type.full_type != null &&
             (p->info->the_opaque_type.is_full_type_visible || g_all_full_types_are_visible))
    {
      p = p->info->the_opaque_type.full_type;
    }
    else if (p->kind == A_GENERIC_TYPE &&
             p->info->the_generic_type.actual_type != null)
    {
      p = p->info->the_generic_type.actual_type;
    }
    else
    {
      break;
    }
  }

  return p;
}

//=================================================================================

// same as complete_type_of() but all full types are visible

static PENTITY override_complete_type_of (PENTITY t)
{
  bool    save_override;
  PENTITY t2;

  save_override = g_all_full_types_are_visible;   // save flag

  g_all_full_types_are_visible = true;

  t2 = complete_type_of (t);

  g_all_full_types_are_visible = save_override;   // restore flag

  return t2;
}

//=================================================================================

bool is_open_type (PENTITY t)
{
  PENTITY p;

  p = complete_type_of (t);

  return (p->kind == AN_OPEN_ARRAY_TYPE ||
          (p->kind == A_STRUCT_TYPE && p->info->the_struct_type.is_open_type));
}

//=================================================================================

static PENTITY renamed_type_of (PENTITY t)
{
  PENTITY p;

  p = t;

  while (p->kind == A_RENAMED_TYPE)
    p = p->info->the_renamed_type.actual_type;

  return p;
}

//=================================================================================

// a jagged type is an array or struct type where any inner element/field type is
// an open type.

// recursive types can be introduced by :
// 1) incomplete types
// 2) opaque types
// 3) generic types?
// 4) self-recursive open structs

// note: a jagged type can be open or non-open.

bool is_jagged_type (PENTITY t)
{
  PENTITY p, e, elem_type;


  // handle only renamed types, and no forward reference types !
  // -> leave incomplete/opaque/generic types "as is" (they are never jagged)

  p = renamed_type_of (t);


  if (p->kind == AN_ARRAY_TYPE)
  {
    return is_jagged_type (p->info->the_array_type.open_array);
  }
  else if (p->kind == A_CONSTRAINED_STRUCT_TYPE)
  {
    return is_jagged_type (p->info->the_constrained_struct_type.open_struct);
  }
  else if (p->kind == AN_OPEN_ARRAY_TYPE)
  {
    elem_type = p->info->the_open_array_type.element;
    return is_open_type (elem_type) || is_jagged_type (elem_type);
  }
  else if (p->kind == A_STRUCT_TYPE)
  {
    e = p->info->the_struct_type.fields->entities.first;
    while (e != null)
    {
      if (e->kind == A_FIELD)
      {
        // risk of circular references :
        // handle only renamed types, and no forward reference types !
        // -> leave incomplete/opaque/generic types "as is" (they are never jagged)

        elem_type = renamed_type_of (e->info->the_field.type);
        if (elem_type == p ||          // self-reference : avoid infinite recursion
            is_open_type (elem_type) ||
            is_jagged_type (elem_type))
          return true;
      }
      else if (e->kind == A_VARYING_FIELD)
      {
        // risk of circular references :
        // handle only renamed types, and no forward reference types !
        // -> leave incomplete/opaque/generic types "as is" (they are never jagged)

        elem_type = renamed_type_of (e->info->the_varying_field.type);
        if (elem_type == p ||          // self-reference : avoid infinite recursion
            is_open_type (elem_type) ||
            is_jagged_type (elem_type))
          return true;
      }

      e = e->next;
    }

    return false;
  }
  else
  {
    return false;
  }
}

//=================================================================================

bool is_limited_type (PENTITY t)
{
  PENTITY p, e, elem_type;


  p = complete_type_of (t);


  if (p->kind == AN_ARRAY_TYPE)
  {
    return is_limited_type (p->info->the_array_type.open_array);
  }
  else if (p->kind == A_CONSTRAINED_STRUCT_TYPE)
  {
    return is_limited_type (p->info->the_constrained_struct_type.open_struct);
  }
  else if (p->kind == AN_OPEN_ARRAY_TYPE)
  {
    return is_limited_type (p->info->the_open_array_type.element);
  }
  else if (p->kind == A_STRUCT_TYPE)
  {
    e = p->info->the_struct_type.fields->entities.first;
    while (e != null)
    {
      if (e->kind == A_FIELD)
      {
        elem_type = complete_type_of (e->info->the_field.type);
        if (is_limited_type (elem_type))
          return true;
      }
      else if (e->kind == A_VARYING_FIELD)
      {
        elem_type = complete_type_of (e->info->the_varying_field.type);
        if (is_limited_type (elem_type))
          return true;
      }

      e = e->next;
    }

    return false;
  }
  else if (p->kind == AN_OPAQUE_TYPE)
  {
    return p->info->the_opaque_type.is_limited;
  }
  else
  {
    return false;
  }
}

//=================================================================================

// A packed struct can only contain fields of packed types.
// Enumeration, integer, floating-point and unsafe pointer types are always packed.
// Array types are packed if the element type is packed, non-open and non-jagged.
// Struct types are packed if they feature the keyword "packed";
//  all their fields must then have packed, non-open and non-jagged types.
// Pointer, function pointer, opaque, incomplete, generic and jagged types are never packed.

// note: a packed type can be open or non-open.

bool is_packed_type (PENTITY type)
{
  PENTITY t;

  t = complete_type_of (type);

  if (t == null)
    return false;

  if (t->kind == AN_INTEGER_TYPE || t->kind == A_FLOAT_TYPE || t->kind == AN_ENUMERATION_TYPE ||
      t->kind == AN_UNSAFE_POINTER_TYPE || t->kind == A_UNION_TYPE)
    return true;

  if (t->kind == AN_ARRAY_TYPE)
    return is_packed_type (t->info->the_array_type.open_array);

  if (t->kind == A_CONSTRAINED_STRUCT_TYPE)
    return is_packed_type (t->info->the_constrained_struct_type.open_struct);

  if (t->kind == AN_OPEN_ARRAY_TYPE)
  {
    PENTITY elem_type;
    elem_type = t->info->the_open_array_type.element;
    return is_packed_type (elem_type) && (!is_open_type (elem_type)) && (!is_jagged_type (elem_type));
  }

  if (t->kind == A_STRUCT_TYPE)
    return t->info->the_struct_type.is_packed;

  return false;
}

//=================================================================================

bool is_unsafe_type (PENTITY type)
{
  PENTITY t;

  t = complete_type_of (type);

  if (t == null)
    return false;

  if (t->kind == AN_ARRAY_TYPE)
    return is_unsafe_type (t->info->the_array_type.open_array);

  if (t->kind == A_CONSTRAINED_STRUCT_TYPE)
    return is_unsafe_type (t->info->the_constrained_struct_type.open_struct);

  if (t->kind == AN_UNSAFE_POINTER_TYPE)
    return true;

  if (t->kind == AN_OPEN_ARRAY_TYPE)
    return is_unsafe_type (t->info->the_open_array_type.element);

  if (t->kind == A_STRUCT_TYPE)
    return t->info->the_struct_type.is_unsafe;

  if (t->kind == A_UNION_TYPE)
    return t->info->the_union_type.is_unsafe;

  return false;
}

//=================================================================================

bool is_constant_exp (PEXPRESSION e)
{
  return e->kind == A_CONST_ENUMERATION_VALUE ||
         e->kind == A_CONST_INTEGER_VALUE ||
         e->kind == A_CONST_FLOAT_VALUE ||
         e->kind == A_CONST_NULL_VALUE ||
         e->kind == A_POOL_CONSTANT;
}

/*****************************************************************************/

PENTITY base_type_of (PENTITY type)
{
  PENTITY t;

  t = complete_type_of (type);

  if (t == null)
    return null;

  if (t->kind == AN_ARRAY_TYPE)
  {
    t = complete_type_of (t->info->the_array_type.open_array);
  }
  else if (t->kind == A_CONSTRAINED_STRUCT_TYPE)
  {
    t = complete_type_of (t->info->the_constrained_struct_type.open_struct);
  }

  return t;
}

/*****************************************************************************/

// returns 0 if OK or -1 if overflow

static int add_int4 (int4 a, int4 b, int4 *r)
{
  int4 c;

  c = a + b;

  if (((a < 0) ^ (b >= 0)) && ((c < 0) ^ (a < 0)))
  {
    *r = 0;
    return -1;
  }

  *r = c;
  return 0;
}

/***************************************************************************/

// returns 0 if OK or -1 if overflow

int mul_int4 (int4 a, int4 b, int4 *r)
{
  int4 c;

  c = a * b;

  if (b != 0 && (c / b) != a)
  {
    *r = 0;
    return -1;
  }

  *r = c;
  return 0;
}

/***************************************************************************/

// returns 0 if OK or -1 if overflow

static int align_at (int *psize, int align)
{
  int offset;

  offset = *psize % align;

  if (offset > 0)
  {
    if (add_int4 (*psize, align-offset, psize) < 0)
      return -1;
  }

  return 0;
}

/***************************************************************************/

void begin_struct (FIELD_DATA *data,
                   PENTITY    stru)      // struct or union type entity
{
  PENTITY ot;

  ot = complete_type_of (stru);

  clear (*data);
  data->packed = ot->kind != A_STRUCT_TYPE || ot->info->the_struct_type.is_packed;
  data->struct_alignment = 1;
}

/***************************************************************************/

void skip_to_valid_field (PENTITY *pfield,            // ref
                          PENTITY *ptype,             // out
                          uint4 discriminant_value)   // in
{
  PENTITY e;

  *ptype = type_int;

  e = *pfield;

  while (e != null)
  {
    if (e->kind == A_FIELD)
    {
      *ptype = e->info->the_field.type;
      break;
    }

    if (e->kind == A_VARYING_FIELD && e->info->the_varying_field.discriminant_value == discriminant_value)
    {
      *ptype = e->info->the_varying_field.type;
      break;
    }

    e = e->next;
  }

  *pfield = e;
}

/*****************************************************************************/

void begin_field (FIELD_DATA *data,
                  PENTITY    field)      // field or varying field
{
  PENTITY t;
  int     align, rc;

  if (field->kind == A_FIELD)
  {
    t = field->info->the_field.type;
  }
  else
  {
    t = field->info->the_varying_field.type;
  }

  rc = size_and_alignment_of_type (t, &data->field_size, &align);

  if (rc == -1)
    data->invalid_type = true;
  else if (rc == -2)
    data->size_overflow = true;
  else
  {
    if (!data->packed)
    {
      if (align_at (&data->struct_size, align) < 0)
        data->size_overflow = true;

      data->struct_alignment = max (data->struct_alignment, align);
    }
  }

  data->field_offset = data->struct_size;
}

/***************************************************************************/

void end_field (FIELD_DATA *data)
{
  if (add_int4 (data->struct_size, data->field_size, &data->struct_size) < 0)
    data->size_overflow = true;
}

/***************************************************************************/

void end_struct (FIELD_DATA *data)
{
  if (!data->packed)
  {
    if (align_at (&data->struct_size, data->struct_alignment) < 0)
      data->size_overflow = true;
  }
}

/*****************************************************************************/

// returns 0 if OK, -1 if bad type, -2 if size is too large

int size_and_alignment_of_type (PENTITY type, int *psize, int *alignment)
{
  PENTITY t, field, ot;
  int     fsize, falign, rc;
  uint4   dis, nb_elem;

  if (type == null)
    return -1;

  t = complete_type_of (type);

  switch (t->kind)
  {
    case AN_INTEGER_TYPE:
      *psize = integer_info[t->info->the_integer_type.type].size;
      *alignment = *psize;
      break;

    case A_FLOAT_TYPE:
      *psize = float_info[t->info->the_float_type.type].size;
      *alignment = *psize;
      break;

    case AN_ENUMERATION_TYPE:
      *psize = integer_info[t->info->the_enumeration_type.base].size;
      *alignment = *psize;
      break;

    case AN_OPEN_ARRAY_TYPE:          // open type
      *psize = address_size + 4;      // address(4 or 8) + length(int4)
      *alignment = address_size;
      align_at (psize, *alignment);   // possibly 4 filler bytes
      break;

    case AN_ARRAY_TYPE:
      ot = complete_type_of (t->info->the_array_type.open_array);
      nb_elem = t->info->the_array_type.length;
      rc = size_and_alignment_of_type (ot->info->the_open_array_type.element, &fsize, &falign);
      if (rc < 0)
        return rc;

      if (mul_int4 (fsize, nb_elem, psize) < 0)
        return -2;

      *alignment = falign;
      break;

    case A_STRUCT_TYPE:       // can be open, can be packed
      if (t->info->the_struct_type.is_open_type)   // open (never packed)
      {
        *psize = address_size + 4;      // address(4 or 8) + discriminant(int4)
        *alignment = address_size;
        align_at (psize, *alignment);   // possibly 4 filler bytes
      }
      else
      {
        FIELD_DATA data;

        begin_struct (&data, t);

        field = t->info->the_struct_type.fields->entities.first;
        while (field != null)
        {
          if (field->kind == A_FIELD)
          {
            begin_field (&data, field);
            end_field (&data);
          }

          field = field->next;
        }

        end_struct (&data);

        *psize = data.struct_size;
        *alignment = data.struct_alignment;

        if (data.invalid_type)
          return -1;
        if (data.size_overflow)
          return -2;
      }
      break;

    case A_CONSTRAINED_STRUCT_TYPE:
      ot = complete_type_of (t->info->the_constrained_struct_type.open_struct);
      dis = t->info->the_constrained_struct_type.discriminant_value;

      {
        FIELD_DATA data;

        begin_struct (&data, ot);

        field = ot->info->the_struct_type.fields->entities.first;
        while (field != null)
        {
          if (field->kind == A_FIELD ||
              (field->kind == A_VARYING_FIELD &&
                field->info->the_varying_field.discriminant_value == dis))
          {
            begin_field (&data, field);
            end_field (&data);
          }

          field = field->next;
        }

        end_struct (&data);

        *psize = data.struct_size;
        *alignment = data.struct_alignment;

        if (data.invalid_type)
          return -1;
        if (data.size_overflow)
          return -2;
      }
      break;

    case A_UNION_TYPE:
      *psize = 0;
      *alignment = 1;
      field = t->info->the_union_type.fields->entities.first;
      while (field != null)
      {
        if (field->kind == A_FIELD)
        {
          if (size_and_alignment_of_type (field->info->the_field.type,
                                          &fsize, &falign) < 0)
            return -1;

          if (fsize > *psize)
            *psize = fsize;
          if (falign > *alignment)
            *alignment = falign;
        }

        field = field->next;
      }
      break;

    case A_POINTER_TYPE:
    case A_FUNCTION_POINTER_TYPE:
    case AN_UNSAFE_POINTER_TYPE:
    case A_NULL_POINTER_TYPE:  // only used internally for 'null' literal expressions
      *psize = address_size;
      *alignment = *psize;
      break;

    case A_RENAMED_TYPE:
    case AN_INCOMPLETE_TYPE:
    case AN_OPAQUE_TYPE:
    case A_GENERIC_TYPE:
      return -1;

    default:
      return -1;
  }

  return 0;
}

/*****************************************************************************/

// check if type is not renamed/incomplete/generic/opaque

bool is_type_having_size (PENTITY type)
{
  int rc, size, align;

  // returns 0 if OK, -1 if bad type, -2 if size is too large
  rc = size_and_alignment_of_type (type, &size, &align);

  return (rc == 0);
}

/*****************************************************************************/

// returns 0 if OK, -1 if bad type, -2 if size is too large, -3 if other constraint
// runtime constraints are allowed but they return a jagged pointer + constraint size

int size_and_alignment_of_constant_size_exp (PEXPRESSION exp, int *psize, int *alignment)
{
  PENTITY t, field;
  int     fsize, falign, rc;
  uint4   dis, nb_elem;

  t = exp->base_type_or_null;

  if (t == null)
    return -1;

  t = complete_type_of (t);

  switch (t->kind)
  {
    case AN_OPEN_ARRAY_TYPE:        // open type (address + length)
      if (exp->constraint.kind == CONSTANT_CONSTRAINT)
      {
        nb_elem = exp->constraint.value;
        rc = size_and_alignment_of_type (t->info->the_open_array_type.element, &fsize, &falign);
        if (rc < 0)
          return rc;

        if (mul_int4 (fsize, nb_elem, psize) < 0)
          return -2;

        *alignment = falign;
      }
      else if (exp->constraint.kind == RUNTIME_CONSTRAINT)
      {
        return size_and_alignment_of_type (t, psize, alignment);
      }
      else
      {
        return -3;  // other constraint
      }
      break;

    case A_STRUCT_TYPE:       // can be open, can be packed

      if (!t->info->the_struct_type.is_open_type)   // not open
        return size_and_alignment_of_type (t, psize, alignment);

      if (exp->constraint.kind == CONSTANT_CONSTRAINT)
      {
        dis = exp->constraint.value;

        {
          FIELD_DATA data;

          begin_struct (&data, t);

          field = t->info->the_struct_type.fields->entities.first;
          while (field != null)
          {
            if (field->kind == A_FIELD ||
                (field->kind == A_VARYING_FIELD &&
                  field->info->the_varying_field.discriminant_value == dis))
            {
              begin_field (&data, field);
              end_field (&data);
            }

            field = field->next;
          }

          end_struct (&data);

          *psize = data.struct_size;
          *alignment = data.struct_alignment;

          if (data.invalid_type)
            return -1;
          if (data.size_overflow)
            return -2;
        }
      }
      else if (exp->constraint.kind == RUNTIME_CONSTRAINT)
      {
        return size_and_alignment_of_type (t, psize, alignment);
      }
      else
      {
        return -3;  // other constraint
      }
      break;

    default:
      return size_and_alignment_of_type (t, psize, alignment);
  }

  return 0;
}

/*****************************************************************************/

typedef struct circular {
  PENTITY         e;
  struct circular *next;
} CIRCULAR_NODE;

typedef CIRCULAR_NODE *CIRCULAR_LIST;

static CIRCULAR_LIST new_node (PENTITY e)
{
  CIRCULAR_LIST n;
  n = malloc (sizeof(CIRCULAR_NODE));
  if (n == null)
    fatal_compiler_error ("circular node", &token.pos);
  n->e    = e;
  n->next = null;

  return n;
}


static bool exists_in_list (PENTITY e, CIRCULAR_LIST l)
{
  CIRCULAR_LIST n = l;
  while (n != null)
  {
    if (n->e == e)
      return true;
    n = n->next;
  }

  return false;
}


static bool exists_in_type (PENTITY t, bool in_variant, CIRCULAR_LIST head, CIRCULAR_LIST *ptail)
{
  PENTITY t2, f;

  t2 = override_complete_type_of (t);

  switch (t2->kind)
  {
    case AN_ARRAY_TYPE:
      return exists_in_type (t2->info->the_array_type.open_array, in_variant, head, ptail);

    case AN_OPEN_ARRAY_TYPE:
      return exists_in_type (t2->info->the_open_array_type.element, in_variant, head, ptail);

    case A_STRUCT_TYPE:
      if (in_variant && t2->info->the_struct_type.is_open_type)
        return false;

      if (exists_in_list (t2, head))    // found in list -> circularity
        return true;

      *ptail = new_node (t2);

      f = t2->info->the_struct_type.fields->entities.first;
      while (f != null)
      {
        if (f->kind == A_FIELD)
        {
          if (exists_in_type (f->info->the_field.type, false, head, &(*ptail)->next))
          {
            free (*ptail);
            *ptail = null;
            return true;
          }
        }
        else if (f->kind == A_VARYING_FIELD)
        {
          if (exists_in_type (f->info->the_varying_field.type, true, head, &(*ptail)->next))
          {
            free (*ptail);
            *ptail = null;
            return true;
          }
        }

        f = f->next;
      }

      free (*ptail);
      *ptail = null;

      return false;

    case A_CONSTRAINED_STRUCT_TYPE:    // force in_variant=false
      return exists_in_type (t2->info->the_constrained_struct_type.open_struct, false, head, ptail);

    case A_UNION_TYPE:
      if (exists_in_list (t2, head))    // found in list -> circularity
        return true;

      *ptail = new_node (t2);

      f = t2->info->the_union_type.fields->entities.first;
      while (f != null)
      {
        if (f->kind == A_FIELD)
        {
          if (exists_in_type (f->info->the_field.type, false, head, &(*ptail)->next))
          {
            free (*ptail);
            *ptail = null;
            return true;
          }
        }

        f = f->next;
      }

      free (*ptail);
      *ptail = null;

      return false;

    default:
      return false;   // suppose no circularities for RENAMED, INCOMPLETE, OPAQUE, GENERIC.
  }
}


bool type_has_circular_dependences (PENTITY s,         // struct or union
                                    PENTITY type,      // field's type to be tested
                                    bool    in_variant)
{
  CIRCULAR_LIST list;
  bool          b;

  list = new_node (override_complete_type_of (s));

  b = exists_in_type (type, in_variant, list, &list->next);

  free (list);

  return b;
}

/*****************************************************************************/
