
// typedecl.c : type declarations

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "strings.h"
#include "token.h"
#include "lex.h"
#include "error.h"
#include "entities.h"
#include "goptions.h"
#include "decl.h"
#include "statem.h"
#include "type.h"
#include "exp.h"
#include "complete.h"
#include "undefined.h"
#include "codgen.h"
#include "typedecl.h"

//=================================================================================

void parse_semicolon ()
{
  if (token.kind == SEMICOLON)
    get_token();
  else
  {
    syntax_error ("';' expected", &token.pos);
    skip_until (SEMICOLON, true);
  }
}

//=================================================================================

void check_unsafe_region ()
{
  if (!lexa.within_unsafe)
    semantic_error ("unsafe pointers are only allowed within unsafe regions", &token.pos);
}

//=================================================================================

// The expanded name of a type_name must denote a type.
// returns null in case of error.

PENTITY parse_type_name ()
{
  PENTITY       p;
  TEXT_POSITION pos;

  p = type_entity_of_token (token.kind);
  if (p != null)
  {
    get_token();    // skip type token
    return p;
  }

  if (token.kind == IDENTIFIER)
  {
    memcpy (&pos, &token.pos, sizeof(TEXT_POSITION));

    p = parse_expanded_name ();
    if (p == null)    // could not find entity (error already given)
      return null;

    if (p->kind > LAST_ENTITY_DENOTING_A_TYPE)
    {
      syntax_error ("identifier must denote a type_name", &pos);
      return null;
    }

    if (is_unsafe_type (p))
      check_unsafe_region ();

    return p;
  }

  syntax_error ("type_name expected", &token.pos);
  return null;
}

//=================================================================================

// checks that the type is not incomplete, otherwise set it to type_int.

static void check_not_incomplete_type (PENTITY *t, TEXT_POSITION *pos)
{
  if (complete_type_of(*t)->kind == AN_INCOMPLETE_TYPE)
  {
    syntax_error ("an incomplete type is not allowed here", pos);
    *t = type_int;
  }
}

//=================================================================================

//  type_definition ::= type_name { type_specifier }
//
//  declarator ::=  identifier  { type_specifier }
//
//  type_specifier ::= array_specification
//                   | open_array_specification
//                   | array_constraint
//                   | discriminant_constraint
//                   | pointer_specification
//                   | unsafe_pointer_specification
//
//  array_specification          ::= "[" constant_expression "]"
//  open_array_specification     ::= "[" "]"
//  array_constraint             ::= "(" constant_expression ")"
//  discriminant_constraint      ::= "(" constant_expression ")"
//  pointer_specification        ::= "^"
//  unsafe_pointer_specification ::= "*"

// t0 can denote an already parsed expanded name or null

PENTITY parse_type_definition (TEXT_POSITION *pos0, PENTITY t0)
{
  PENTITY t;

  t = t0;

  if (t == null)
    t = parse_type_name ();

  if (t == null)  // earlier error
    t = type_int;

  for (;;)
  {
    if (token.kind == LEFT_BRACKET)
    {
      check_not_incomplete_type (&t, pos0);

      get_token();    // skip [

      if (token.kind == RIGHT_BRACKET)   // open array
      {
        AN_OPEN_ARRAY_TYPE_ENTITY open_array;

        memset (&open_array, 0x00, sizeof(open_array));
        open_array.element = t;

        t = append_new_entity (AN_OPEN_ARRAY_TYPE, &open_array, L"", &token.pos);

        get_token();    // skip ]
      }
      else
      {
        AN_OPEN_ARRAY_TYPE_ENTITY open_array;
        AN_ARRAY_TYPE_ENTITY      array;
        TEXT_POSITION             pos;
        int8                      value;

        memset (&open_array, 0x00, sizeof(open_array));
        open_array.element = t;

        t = append_new_entity (AN_OPEN_ARRAY_TYPE, &open_array, L"", &token.pos);

        memcpy (&pos, &token.pos, sizeof(pos));

        value = parse_constant_int_or_uint_expression ();

        if (value < 0 || value > 2147483647)
        {
          semantic_error ("expression is out of range", &pos);
          value = 1;
        }

        memset (&array, 0x00, sizeof(array));
        array.open_array = t;
        array.length = (uint4)value;

        t = append_new_entity (AN_ARRAY_TYPE, &array, L"", &token.pos);

        if (token.kind == RIGHT_BRACKET)
        {
          get_token();    // skip ]
        }
        else
        {
          syntax_error ("']' expected here", &token.pos);
          skip_until (IDENTIFIER, false);
          break;
        }
      }
    }
    else if (complete_type_of(t)->kind == AN_OPEN_ARRAY_TYPE && token.kind == LEFT_PARENTHESIS)
    {
      AN_ARRAY_TYPE_ENTITY array;
      TEXT_POSITION        pos;
      int8                 value;

      get_token();    // skip (

      memcpy (&pos, &token.pos, sizeof(pos));

      value = parse_constant_int_or_uint_expression ();

      if (value < 0 || value > 2147483647)
      {
        semantic_error ("expression is out of range", &pos);
        value = 1;
      }

      memset (&array, 0x00, sizeof(array));
      array.open_array = t;
      array.length = (uint4)value;

      t = append_new_entity (AN_ARRAY_TYPE, &array, L"", &token.pos);

      if (token.kind == RIGHT_PARENTHESIS)
      {
        get_token();    // skip )
      }
      else
      {
        syntax_error ("')' expected here", &token.pos);
        skip_until (IDENTIFIER, false);
        break;
      }
    }
    else if (complete_type_of(t)->kind == A_STRUCT_TYPE &&
             complete_type_of(t)->info->the_struct_type.is_open_type &&
             token.kind == LEFT_PARENTHESIS)
    {
      A_CONSTRAINED_STRUCT_TYPE_ENTITY  cons;
      PENTITY                           dis, enumer;
      TEXT_POSITION                     pos;
      PEXPRESSION                       exp;
      int8                              value;

      get_token();    // skip (

      memcpy (&pos, &token.pos, sizeof(pos));

      exp = parse_constant_expression ();


      // check if expression's type and range are valid

      value = 0;   // default

      dis = complete_type_of(t)->info->the_struct_type.discriminant->entities.first;
      if (dis != null && dis->kind == A_FIELD && exp->base_type_or_null != null)
      {
        enumer = dis->info->the_field.type;
        if (enumer->kind == AN_ENUMERATION_TYPE)
        {
          if (!types_are_equal (exp->base_type_or_null, enumer))
          {
            semantic_error ("expression must have discriminant's type", &pos);
          }
          else if (exp->kind != A_CONST_ENUMERATION_VALUE)
          {
            semantic_error ("expression must be constant", &pos);
          }
          else
          {
            value = exp->info.const_enumeration_value_info.value;
            if (value < 0 || value > enumer->info->the_enumeration_type.last)
            {
              semantic_error ("expression is out of range", &pos);
              value = 0;
            }
          }
        }
      }

      memset (&cons, 0x00, sizeof(cons));
      cons.open_struct = t;
      cons.discriminant_value = (uint4)value;

      t = append_new_entity (A_CONSTRAINED_STRUCT_TYPE, &cons, L"", &token.pos);


      if (token.kind == RIGHT_PARENTHESIS)
      {
        get_token();    // skip )
      }
      else
      {
        syntax_error ("')' expected here", &token.pos);
        skip_until (IDENTIFIER, false);
        break;
      }
    }
    else if (token.kind == CARET)
    {
      A_POINTER_TYPE_ENTITY ptr;

      if (is_jagged_type (t))
        semantic_error ("^ not allowed for a jagged type", &token.pos);

      memset (&ptr, 0x00, sizeof(ptr));
      ptr.designated_type = t;

      t = append_new_entity (A_POINTER_TYPE, &ptr, L"", &token.pos);

      get_token();    // skip ^
    }
    else if (token.kind == STAR)
    {
      AN_UNSAFE_POINTER_TYPE_ENTITY ptr;

      if (is_jagged_type (t))
        semantic_error ("* not allowed for a jagged type", &token.pos);

      memset (&ptr, 0x00, sizeof(ptr));
      ptr.designated_type = t;

      t = append_new_entity (AN_UNSAFE_POINTER_TYPE, &ptr, L"", &token.pos);

      check_unsafe_region ();

      get_token();    // skip *
    }
    else
    {
      break;
    }
  }

  return t;
}

//=================================================================================

static PENTITY inner_parse_declarator_suffix (PENTITY       t,
                                              TEXT_POSITION *pos0,
                                              bool          incomplete_type_allowed)
{
  if (token.kind == LEFT_BRACKET)
  {
    get_token();    // skip [

    if (token.kind == RIGHT_BRACKET)   // open array
    {
      AN_OPEN_ARRAY_TYPE_ENTITY open_array;

      get_token();    // skip ]

      memset (&open_array, 0x00, sizeof(open_array));
      open_array.element = inner_parse_declarator_suffix (t, pos0, false);

      return append_new_entity (AN_OPEN_ARRAY_TYPE, &open_array, L"", pos0);
    }
    else
    {
      AN_OPEN_ARRAY_TYPE_ENTITY open_array;
      AN_ARRAY_TYPE_ENTITY      array;
      TEXT_POSITION             pos;
      int8                      value;
      PENTITY                   t2;

      memcpy (&pos, &token.pos, sizeof(pos));

      value = parse_constant_int_or_uint_expression ();

      if (value < 0 || value > 2147483647)
      {
        semantic_error ("expression is out of range", &pos);
        value = 1;
      }

      if (token.kind == RIGHT_BRACKET)
      {
        get_token();    // skip ]
      }
      else
      {
        syntax_error ("']' expected here", &token.pos);
        skip_until (IDENTIFIER, false);
      }

      memset (&open_array, 0x00, sizeof(open_array));
      open_array.element = inner_parse_declarator_suffix (t, pos0, false);

      t2 = append_new_entity (AN_OPEN_ARRAY_TYPE, &open_array, L"", pos0);

      memset (&array, 0x00, sizeof(array));
      array.open_array = t2;
      array.length = (uint4)value;

      return append_new_entity (AN_ARRAY_TYPE, &array, L"", pos0);
    }
  }
  else if (token.kind == CARET)
  {
    A_POINTER_TYPE_ENTITY ptr;
    TEXT_POSITION         pos;

    memcpy (&pos, &token.pos, sizeof(pos));
    get_token();    // skip ^

    memset (&ptr, 0x00, sizeof(ptr));
    ptr.designated_type = inner_parse_declarator_suffix (t, pos0, true);

    if (is_jagged_type (ptr.designated_type))
      semantic_error ("^ not allowed for a jagged type", &pos);

    return append_new_entity (A_POINTER_TYPE, &ptr, L"", &pos);
  }
  else if (token.kind == STAR)
  {
    AN_UNSAFE_POINTER_TYPE_ENTITY ptr;
    TEXT_POSITION                 pos;

    memcpy (&pos, &token.pos, sizeof(pos));
    get_token();    // skip *

    check_unsafe_region ();

    memset (&ptr, 0x00, sizeof(ptr));
    ptr.designated_type = inner_parse_declarator_suffix (t, pos0, true);

    if (is_jagged_type (ptr.designated_type))
      semantic_error ("* not allowed for a jagged type", &pos);

    return append_new_entity (AN_UNSAFE_POINTER_TYPE, &ptr, L"", &pos);
  }
  else if (token.kind == LEFT_PARENTHESIS)
  {
    TEXT_POSITION   pos;
    PEXPRESSION     exp;
    PENTITY         suffix;

    get_token();    // skip (

    memcpy (&pos, &token.pos, sizeof(pos));

    exp = parse_constant_expression ();

    if (token.kind == RIGHT_PARENTHESIS)
    {
      get_token();    // skip )
    }
    else
    {
      syntax_error ("')' expected here", &token.pos);
      skip_until (IDENTIFIER, false);
    }

    suffix = inner_parse_declarator_suffix (t, pos0, false);

    if (complete_type_of(suffix)->kind == AN_OPEN_ARRAY_TYPE)
    {
      AN_ARRAY_TYPE_ENTITY array;
      int8                 value;


      // check that expression has an int or uint type

      if (exp->kind != A_CONST_INTEGER_VALUE)
      {
        semantic_error ("a constant integer is expected here", &pos);
        value = 1;
      }
      else
      {
        value = exp->info.const_integer_value_info.value;

        if (value < 0 || value > 2147483647)
        {
          semantic_error ("expression is out of range", &pos);
          value = 1;
        }
        else if (exp->base_type_or_null == type_long)
        {
          semantic_error ("a constant of type int or uint is expected here", &pos);
        }
      }

      memset (&array, 0x00, sizeof(array));
      array.open_array = suffix;
      array.length = (uint4)value;

      return append_new_entity (AN_ARRAY_TYPE, &array, L"", &pos);
    }
    else if (complete_type_of(suffix)->kind == A_STRUCT_TYPE &&
             complete_type_of(suffix)->info->the_struct_type.is_open_type)
    {
      A_CONSTRAINED_STRUCT_TYPE_ENTITY  cons;
      PENTITY                           dis, enumer;
      int8                              value;


      // check if expression's type and range are valid

      value = 0;   // default

      dis = complete_type_of(suffix)->info->the_struct_type.discriminant->entities.first;
      if (dis != null && dis->kind == A_FIELD && exp->base_type_or_null != null)
      {
        enumer = dis->info->the_field.type;
        if (enumer->kind == AN_ENUMERATION_TYPE)
        {
          if (exp->kind != A_CONST_ENUMERATION_VALUE ||
              !types_are_equal (exp->base_type_or_null, enumer))
          {
            semantic_error ("expression must have discriminant's type", &pos);
          }
          else
          {
            value = exp->info.const_enumeration_value_info.value;
            if (value < 0 || value > enumer->info->the_enumeration_type.last)
            {
              semantic_error ("expression is out of range", &pos);
              value = 0;
            }
          }
        }
      }

      memset (&cons, 0x00, sizeof(cons));
      cons.open_struct = suffix;
      cons.discriminant_value = (uint4)value;

      return append_new_entity (A_CONSTRAINED_STRUCT_TYPE, &cons, L"", &pos);
    }
    else
    {
      semantic_error ("'(' is only allowed for an open type", &pos);
      return t;
    }
  }
  else
  {
    if (!incomplete_type_allowed)
    {
      check_not_incomplete_type (&t, pos0);
    }

    return t;
  }
}

//=================================================================================

static PENTITY parse_declarator_suffix (PENTITY t, TEXT_POSITION *pos0)
{
  return inner_parse_declarator_suffix (t, pos0, false);
}

//=================================================================================

static PENTITY get_incomplete_type (wstring s)
{
  PENTITY incomplete;

  incomplete = search_entity_in_local_scope (token.info._identifier.value);

  if (incomplete != null &&
      incomplete->kind == AN_INCOMPLETE_TYPE &&
      incomplete->info->the_incomplete_type.full_type == null)
  {
    return incomplete;
  }

  return null;
}

//=================================================================================

static PENTITY get_opaque_type (wstring s)
{
  PENTITY opaque;

  opaque = search_entity_in_local_scope (token.info._identifier.value);

  if (opaque != null &&
      opaque->kind == AN_OPAQUE_TYPE &&
      opaque->info->the_opaque_type.full_type == null)
  {
    return opaque;
  }

  return null;
}

//=================================================================================

static void fill_incomplete_type (PENTITY incomplete, PENTITY full_type, TEXT_POSITION *pos)
{
  AN_INCOMPLETE_TYPE_COMPLETITION_ENTITY com;

  if (incomplete != null)
  {
    if (is_jagged_type (full_type))
      semantic_error ("a jagged type is not allowed as the full type of an earlier incomplete type", pos);
    incomplete->info->the_incomplete_type.full_type = full_type;
    incomplete->info->the_incomplete_type.is_full_type_visible = true;

    memset (&com, 0x00, sizeof(com));
    com.incomplete_type = incomplete;
    com.full_type       = full_type;
    (void)append_new_entity (AN_INCOMPLETE_TYPE_COMPLETITION, &com, L"", &token.pos);
  }
}

//=================================================================================

static void fill_opaque_type (PENTITY opaque, PENTITY full_type, TEXT_POSITION *pos)
{
  AN_OPAQUE_TYPE_COMPLETITION_ENTITY com;
  ENTITY_KIND                        surrounding_kind;

  if (opaque != null)
  {
    surrounding_kind = surrounding_entity()->kind;
    if (surrounding_kind != A_UNIT_BODY && surrounding_kind != A_PACKAGE_BODY)
      semantic_error ("the full type of an opaque type declaration is only allowed within an interface body or package body", pos);

    opaque->info->the_opaque_type.full_type = full_type;
    opaque->info->the_opaque_type.is_full_type_visible = true;

    memset (&com, 0x00, sizeof(com));
    com.opaque_type = opaque;
    com.full_type   = full_type;
    (void)append_new_entity (AN_OPAQUE_TYPE_COMPLETITION, &com, L"", &token.pos);
  }
}

//=================================================================================

//  enumeration_declaration ::=
//        "enum" identifier   [ "("  base_type_name  ")" ]
//        "{"  enumeration_literal {"," enumeration_literal}
//        [","]  "}" ";"
//
//  base_type_name ::= type_name
//  enumeration_literal ::= identifier

void parse_enumeration_declaration ()
{
  PENTITY         type_enum, base, incomplete;
  TEXT_POSITION   pos;
  int             size;
  int8            count;
  AN_ENUMERATION_LITERAL_ENTITY lit;

  get_token();    // skip 'enum'

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("identifier expected here", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }


  // check if it's the full type of an earlier incomplete type

  incomplete = get_incomplete_type (token.info._identifier.value);


  {
    AN_ENUMERATION_TYPE_ENTITY enumer;
    memset (&enumer, 0x00, sizeof(enumer));
    enumer.base = a_uint4;
    enumer.last = 0;

    type_enum = append_new_entity (AN_ENUMERATION_TYPE,
                                   &enumer,
                                   incomplete != null ? L"" : token.info._identifier.value,
                                   &token.pos);

    fill_incomplete_type (incomplete, type_enum, &token.pos);
  }


  get_token();      // skip identifier

  if (token.kind == LEFT_PARENTHESIS)   // base type specified
  {
    get_token();      // skip '('

    memcpy (&pos, &token.pos, sizeof(pos));

    base = complete_type_of (parse_type_name());
    if (base != null)         // no earlier error
    {
      if (base->kind != AN_INTEGER_TYPE ||
          (base->info->the_integer_type.type != a_uint1 &&
           base->info->the_integer_type.type != a_uint2 &&
           base->info->the_integer_type.type != a_uint4))
      {
        semantic_error ("type_name must be uint1, uint2 or uint4", &pos);
      }
      else
      {
        type_enum->info->the_enumeration_type.base = base->info->the_integer_type.type;
      }
    }

    if (token.kind != RIGHT_PARENTHESIS)
    {
      syntax_error ("')' expected here", &token.pos);
      skip_until (LEFT_ACCOLADE, false);
    }

    get_token();      // skip ')'
  }

  if (token.kind != LEFT_ACCOLADE)
  {
    syntax_error ("'{' expected here", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }
  get_token();      // skip {

  size = integer_info[type_enum->info->the_enumeration_type.base].size;
  count = 0;

  for (;;)
  {
    if (token.kind != IDENTIFIER)
    {
      syntax_error ("identifier expected here", &token.pos);
      skip_until (SEMICOLON, true);
      return;
    }


    // create the enumeration literal entity

    memset (&lit, 0x00, sizeof(lit));
    lit.type = type_enum;
    lit.value = (uint4)count;
    (void)append_new_entity (AN_ENUMERATION_LITERAL, &lit, token.info._identifier.value, &token.pos);


    type_enum->info->the_enumeration_type.last = (uint4)count;
    count++;

    get_token();      // skip identifier

    if (token.kind == RIGHT_ACCOLADE)
      break;

    if (token.kind != COMMA)
    {
      syntax_error ("'}' or ',' expected here", &token.pos);
      skip_until (SEMICOLON, true);
      return;
    }
    get_token();      // skip comma

    if (token.kind == RIGHT_ACCOLADE)
      break;
  }

  if (type_enum->info->the_enumeration_type.last
      > integer_info[type_enum->info->the_enumeration_type.base].max)
  {
    semantic_error ("too many enumeration literals", &token.pos);
  }

  get_token();      // skip }
  parse_semicolon ();


  create_enumeration_literal_string_table (type_enum);
}

//=================================================================================

static void parse_typedef_declaration_of_type_definition (PENTITY t0, TEXT_POSITION *typ_pos)
{
  wchar                 id[MAX_IDENTIFIER_LENGTH+1];
  TEXT_POSITION         pos;
  PENTITY               t, incomplete, full_type;
  A_RENAMED_TYPE_ENTITY ren;

  wstrcpy (id, token.info._identifier.value);


  // check if it's the full type of an earlier incomplete type

  incomplete = get_incomplete_type (token.info._identifier.value);


  memcpy (&pos, &token.pos, sizeof(pos));

  get_token();      // skip identifier

  t = parse_declarator_suffix (t0, typ_pos);

  memset (&ren, 0x00, sizeof(ren));
  ren.actual_type = t;

  full_type = append_new_entity (A_RENAMED_TYPE,
                                 &ren,
                                 incomplete != null ? L"" : id,
                                 &pos);

  fill_incomplete_type (incomplete, full_type, &pos);

  parse_semicolon ();
}

//=================================================================================

//  formal_parameters ::= [formal_parameter {"," formal_parameter}]
//  formal_parameter ::= [mode] type_definition declarator
//                                       ["=" constant_expression]
//  mode ::=  "ref" | "out"
//
//  If no mode is specified, the mode "in" is implicitly chosen.
//
//  Parameterless functions have an empty set of parentheizes.
//
//  Parameters can have any type except type incomplete, except within typedef declaration.

static void parse_parentheized_formal_parameters (bool inside_typedef_declaration,
                                                  bool is_callback_or_extern_dll)
{
  MODE               mode;
  wchar              id[MAX_IDENTIFIER_LENGTH+1];
  TEXT_POSITION      pos, typ_pos, exp_pos;
  PENTITY            t, typ, par;
  A_PARAMETER_ENTITY param;
  PEXPRESSION        exp;
  CONTEXT            context;

  if (token.kind != LEFT_PARENTHESIS)
  {
    syntax_error ("'(' expected here", &token.pos);
    return;
  }

  get_token();      // skip '('


  if (token.kind != RIGHT_PARENTHESIS)
  {
    for (;;)
    {
      if (token.kind == TOKEN_ref)
      {
        mode = MODE_REF;
        get_token();
      }
      else if (token.kind == TOKEN_out)
      {
        mode = MODE_OUT;
        get_token();
      }
      else
      {
        mode = MODE_IN;
      }

      memcpy (&typ_pos, &token.pos, sizeof(typ_pos));
      t = parse_type_definition (&token.pos, null);

      if (token.kind != IDENTIFIER)
      {
        syntax_error ("identifier expected", &token.pos);
        skip_until (RIGHT_PARENTHESIS, true);
        return;
      }

      wstrcpy (id, token.info._identifier.value);
      memcpy (&pos, &token.pos, sizeof(pos));

      get_token();      // skip identifier

      typ = inner_parse_declarator_suffix (t, &typ_pos, /* incomplete_type_allowed= */ inside_typedef_declaration);


      if (mode != MODE_IN)  // forbid jagged types except object[]
      {
        if (is_jagged_type (typ) && !types_are_equal (typ, type_array_of_object))
          semantic_error ("a jagged type is not allowed for mode out/ref", &typ_pos);
      }

      if (is_callback_or_extern_dll)   // forbid open array/structs
      {
        if (is_open_type (typ))
          semantic_error ("open types are not allowed for callback or dll calls", &typ_pos);
      }


      memset (&param, 0x00, sizeof(param));
      param.type = typ;
      param.mode = mode;
      param.default_value_or_null = null;
      param.is_used = false;
      param.init_error = false;
      memcpy (&param.pos, &pos, sizeof(TEXT_POSITION));

      par = create_new_entity (A_PARAMETER, &param, id, &pos);

      hide_entity (par);

      if (token.kind == ASSIGN)
      {
        if (mode != MODE_IN)
          semantic_error ("a default expression is only allowed for parameters of mode in", &token.pos);
        else if (is_limited_type (typ))
          semantic_error ("a default expression is not allowed for parameters of a limited type", &token.pos);

        get_token();      // skip '='

        memcpy (&exp_pos, &token.pos, sizeof(exp_pos));

        convert_type_into_context (typ, &context);

        exp = parse_constant_expression_with_context (&context);

        check_assignment_context_compatibility (&context, exp, &exp_pos);

        par->info->the_parameter.default_value_or_null = exp;
      }

      append_entity_in_entity_queue (par);  // append late to avoid forward entity references


      if (token.kind != COMMA)
        break;
      get_token();      // skip comma
    }
  }


  if (token.kind != RIGHT_PARENTHESIS)
  {
    syntax_error ("')' expected here", &token.pos);
    return;
  }

  get_token();      // skip ')'
}

//=================================================================================

// assertion: both entities denote function pointer types

// Two functions are compatible if they have the same return type, function
// options ("callback", "extern"), number of parameters, and, for each
// parameter : same mode, identical type.
// The keywords "public" and "inline", the function name and parameter
// names or any default expressions need not match.

static bool are_compatible_function_types (PENTITY func_type1, PENTITY func_type2)
{
  A_FUNCTION_POINTER_TYPE_ENTITY *fa, *fb;
  PENTITY e1, e2;

  fa = &func_type1->info->the_function_pointer_type;
  fb = &func_type2->info->the_function_pointer_type;

  // return type

  if (!types_are_equal (fa->return_type, fb->return_type))
    return false;

  if (fa->is_callback != fb->is_callback)
    return false;

  if ((fa->extern_dll_or_null != null) ^ (fb->extern_dll_or_null != null))
    return false;

  e1 = fa->parameters->entities.first;
  e2 = fb->parameters->entities.first;

  for (;;)
  {
    // skip anonymous declarations
    while (e1 != null && e1->identifier_or_null == null)
      e1 = e1->next;

    while (e2 != null && e2->identifier_or_null == null)
      e2 = e2->next;

    if (e1 == null && e2 == null)   // both null -> done
      break;

    if (e1 == null || e2 == null)  // one null -> incompatible
      return false;

    if (e1->info->the_parameter.mode != e2->info->the_parameter.mode)
      return false;

    if (!types_are_equal (e1->info->the_parameter.type,
                          e2->info->the_parameter.type))
      return false;

    e1 = e1->next;
    e2 = e2->next;
  }

  return true;
}

//=================================================================================

// check if types from function declaration and body match.
// (they can differ due to opaque types becoming complete or redeclaration of arrays/pointers)

bool types_are_equal (PENTITY t1, PENTITY t2)
{
  PENTITY ta, tb;

  ta = t1;
  tb = t2;

  for (;;)
  {
    ta = complete_type_of (ta);
    tb = complete_type_of (tb);

    if (ta->kind != tb->kind)
      return false;

    switch (ta->kind)
    {
      case AN_OPEN_ARRAY_TYPE:
        ta = ta->info->the_open_array_type.element;
        tb = tb->info->the_open_array_type.element;
        break;

      case AN_ARRAY_TYPE:
        if (ta->info->the_array_type.length != tb->info->the_array_type.length)
          return false;
        ta = ta->info->the_array_type.open_array;
        tb = tb->info->the_array_type.open_array;
        break;

      case A_CONSTRAINED_STRUCT_TYPE:
        if (ta->info->the_constrained_struct_type.discriminant_value != tb->info->the_constrained_struct_type.discriminant_value)
          return false;
        ta = ta->info->the_constrained_struct_type.open_struct;
        tb = tb->info->the_constrained_struct_type.open_struct;
        break;

      case A_POINTER_TYPE:
        ta = ta->info->the_pointer_type.designated_type;
        tb = tb->info->the_pointer_type.designated_type;
        break;

      case A_FUNCTION_POINTER_TYPE:
        return are_compatible_function_types (ta, tb);

      case AN_UNSAFE_POINTER_TYPE:
        ta = ta->info->the_unsafe_pointer_type.designated_type;
        tb = tb->info->the_unsafe_pointer_type.designated_type;
        break;

      default:
        return ta == tb;
    }
  }
}

//=================================================================================

// The function specification of a function declaration and body must match.
// In particular : the function name, the function options ("callback"),
// the return type, the number of parameters, and for each parameter :
// its name, type, default value presence and value if any.
// The keywords "inline" and "public" need not match.

static void parse_function_specification_for_body
                   (TEXT_POSITION                  *pos0,
                    PENTITY                        edecl,
                    A_FUNCTION_POINTER_TYPE_ENTITY *decl,
                    A_FUNCTION_POINTER_TYPE_ENTITY *bdy)   // has no parameter info
{
  MODE             mode;
  wchar            id[MAX_IDENTIFIER_LENGTH+1];
  TEXT_POSITION    pos, typ_pos;
  PENTITY          t, typ, e;
  PEXPRESSION      exp;
  CONTEXT          context;
  TEXT_POSITION    exp_pos;
  bool             match;


  match = true;


  // check if return types match

  if (!types_are_equal (decl->return_type, bdy->return_type))
    match = false;


  e = decl->parameters->entities.first;

  // skip anonymous declarations in parameter region
  while (e != null && e->identifier_or_null == null)
    e = e->next;


  get_token();    // skip identifier


  // check if parameters match

  if (token.kind != LEFT_PARENTHESIS)
  {
    syntax_error ("'(' expected here", &token.pos);
    return;
  }

  get_token();      // skip '('


  if (token.kind != RIGHT_PARENTHESIS)
  {
    for (;;)
    {
      if (token.kind == TOKEN_ref)
      {
        mode = MODE_REF;
        get_token();
      }
      else if (token.kind == TOKEN_out)
      {
        mode = MODE_OUT;
        get_token();
      }
      else
      {
        mode = MODE_IN;
      }

      memcpy (&typ_pos, &token.pos, sizeof(typ_pos));
      t = parse_type_definition (&token.pos, null);

      if (token.kind != IDENTIFIER)
      {
        syntax_error ("identifier expected", &token.pos);
        skip_until (RIGHT_PARENTHESIS, true);
        return;
      }

      wstrcpy (id, token.info._identifier.value);
      memcpy (&pos, &token.pos, sizeof(pos));

      get_token();      // skip identifier

      typ = parse_declarator_suffix (t, &typ_pos);

      exp = null;
      if (token.kind == ASSIGN)
      {
        get_token();      // skip '='
        memcpy (&exp_pos, &token.pos, sizeof(exp_pos));

        convert_type_into_context (typ, &context);

        exp = parse_constant_expression_with_context (&context);

        check_assignment_context_compatibility (&context, exp, &exp_pos);
      }

      if (e == null ||
          e->kind != A_PARAMETER ||
          wstrcmp (e->identifier_or_null, id) != 0 ||
          (!types_are_equal (e->info->the_parameter.type, typ)) ||
          e->info->the_parameter.mode != mode ||
          !constant_expressions_match (e->info->the_parameter.default_value_or_null, exp))
      {
        match = false;
        if (e != null && e->kind == A_PARAMETER)
        {
          e->info->the_parameter.is_used = true;   // avoid another error later
        }
      }
      else   // advance e to next parameter
      {
        // set position of parameter in function body
        memcpy (&e->info->the_parameter.pos, &pos, sizeof(TEXT_POSITION));

        e = e->next;

        // skip anonymous declarations in parameter region
        while (e != null && e->identifier_or_null == null)
         e = e->next;
      }

      free_exp (exp);

      if (token.kind != COMMA)
        break;
      get_token();      // skip comma
    }
  }

  if (token.kind != RIGHT_PARENTHESIS)
  {
    syntax_error ("')' expected here", &token.pos);
    return;
  }

  get_token();      // skip ')'

  if (e != null)     // declaration has more parameters
    match = false;

  if (!match)
    semantic_error ("function return type/parameters do not match with declaration", pos0);
}

//=================================================================================

static bool region_contains_break_statement (PREGION r)
{
  PENTITY e;

  e = r->entities.first;

  while (e != null)
  {
    switch (e->kind)
    {
      case A_BREAK_STATEMENT:
        return true;

      case A_BLOCK_STATEMENT:
        if (region_contains_break_statement (e->info->the_block_statement.inner))
          return true;
        break;

      case AN_IF_STATEMENT:
        if (region_contains_break_statement (e->info->the_if_statement.true_branch) ||
            region_contains_break_statement (e->info->the_if_statement.false_branch))
          return true;
        break;

      default:
        break;
    }

    e = e->next;
  }

  return false;
}

//=================================================================================
static bool region_contains_stopper (PREGION r);
//=================================================================================

static bool is_stopper (PENTITY e)
{
  switch (e->kind)
  {
    case A_RETURN_STATEMENT:
      return true;

    case AN_ABORT_STATEMENT:
      return true;

    case A_BLOCK_STATEMENT:
      return region_contains_stopper (e->info->the_block_statement.inner);

    case AN_IF_STATEMENT:
      return region_contains_stopper (e->info->the_if_statement.true_branch)
          && region_contains_stopper (e->info->the_if_statement.false_branch);

    case A_SWITCH_STATEMENT:
      {
        FLOW_ALTERNATIVE *f;
        f = e->info->the_switch_statement.first_alt;
        while (f != null)
        {
          if (!region_contains_stopper (f->inner))
            return false;
          f = f->next;
        }
      }
      return true;

    case A_WHILE_STATEMENT:
      {
        PEXPRESSION exp = e->info->the_while_statement.condition;
        if ((exp == null || (exp->kind == A_CONST_ENUMERATION_VALUE && exp->info.const_enumeration_value_info.value == 1))
            && !region_contains_break_statement (e->info->the_while_statement.inner))
        return true;
      }
      return false;

    case A_FOR_STATEMENT:
      {
        PEXPRESSION exp = e->info->the_for_statement.condition;
        if ((exp == null || (exp->kind == A_CONST_ENUMERATION_VALUE && exp->info.const_enumeration_value_info.value == 1))
            && !region_contains_break_statement (e->info->the_for_statement.inner))
        return true;
      }
      return false;

    default:
      return false;
  }
}

//=================================================================================

static bool region_contains_stopper (PREGION r)
{
  PENTITY e;

  e = r->entities.last;

  if (e == null)    // region contains no entities
    return false;

  if (is_stopper (e))   // return statement is usually last entity
    return true;

  // check all region for any stopper
  e = r->entities.first;
  while (e != null)
  {
    if (is_stopper (e))
      return true;
    e = e->next;
  }

  return false;
}

//=================================================================================

// assertion: current token is LEFT_ACCOLADE.

// function_body ::= function_specification
//                    "{"  declarations  statements "}"

static void parse_function_body (PENTITY edecl)
{
  A_FUNCTION_BODY_ENTITY body;
  PENTITY                ebody, out_prefix;
  TEXT_POSITION          out_pos;

  memset (&body, 0x00, sizeof(body));
  body.to_function_declaration = edecl;
  body.inner = new_region ();
  store_location (&body.begin_loc);

  get_token();      // skip '{'

  ebody = append_new_entity (A_FUNCTION_BODY, &body, L"", &token.pos);

  edecl->info->the_function_declaration.to_function_body_or_null = ebody;

  create_region_level ();
  append_region (edecl->info->the_function_declaration.to_type
                      ->info->the_function_pointer_type.parameters);
  append_region (body.inner);


  // 'out_prefix' is an "out" parameter that can either be null,
  // or denote an already parsed expanded name that is not a type_name.
  parse_local_declarations (&out_pos, &out_prefix);

  check_local_completion (ebody->info->the_function_body.inner->entities.first);

  parse_statements (&out_pos, out_prefix);
  lex_set_in_statement (false);


  if (!ebody->info->the_function_body.contains_code_statements)
  {
    // the last statement of a function body having a non-void return type
    // must be a return statement.

    if (edecl->info->the_function_declaration.to_type->info->the_function_pointer_type.return_type != type_void)
    {
      if (!region_contains_stopper (ebody->info->the_function_body.inner))
        semantic_error ("a return statement is expected here", &token.pos);
    }

    detect_uninitialized_variables
       (edecl->info->the_function_declaration.to_type->info->the_function_pointer_type.parameters,
        ebody->info->the_function_body.inner);
  }

  store_location (&ebody->info->the_function_body.end_loc);

  if (token.kind == RIGHT_ACCOLADE)
    get_token();
  else
  {
    syntax_error ("'}' expected", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
  }

  unlink_region_level ();
}

//=================================================================================

// for function declaration, body, typedef or generic function parameter.

// when t is non-null, a type definition has been parsed earlier and an identifier follows.
// when t is null, current token is : "[", public, inline or void.

void parse_function_specification (TEXT_POSITION *pos0,
                                   PENTITY       t,
                                   bool          body_allowed,
                                   bool          *body_was_parsed,   // ref
                                   bool          in_typedef,
                                   bool          in_generic_part)
{
  A_FUNCTION_POINTER_TYPE_ENTITY func;
  wchar                          id[MAX_IDENTIFIER_LENGTH+1];
  TEXT_POSITION                  id_pos, public_pos;
  PENTITY                        functype, e, func_base_type, incomplete;
  ENTITY_KIND                    surrounding_kind;

  memcpy (&public_pos, pos0, sizeof(public_pos));

  memset (&func, 0x00, sizeof(func));
  func.is_inline           = false;
  func.is_callback         = false;
  func.is_public           = false;
  func.extern_dll_or_null  = null;
  func.return_type         = t;
  func.parameters          = null;

  if (t == null)
  {
    if (token.kind == LEFT_BRACKET)
    {
      if (!lexa.within_unsafe)
        semantic_error ("function options are only allowed within unsafe regions", &token.pos);

      get_token();   // skip '['

      if (token.kind == IDENTIFIER && wstrcmp (token.info._identifier.value, L"callback") == 0)
      {
        func.is_callback = true;
        get_token();   // skip 'callback'
      }
      else if (token.kind == IDENTIFIER && wstrcmp (token.info._identifier.value, L"extern") == 0)
      {
        get_token();   // skip 'extern'

        if (token.kind != STRING_LITERAL)
        {
          syntax_error ("string literal expected", &token.pos);
          skip_until (RIGHT_BRACKET, false);
        }
        else
        {
          func.extern_dll_or_null = malloc (sizeof(wchar)*(1+token.info._string.length));
          if (func.extern_dll_or_null == null)
            fatal_out_of_memory_error ("parse_function_specification()");
          memcpy (func.extern_dll_or_null, token.info._string.value, sizeof(wchar)*token.info._string.length);
          func.extern_dll_or_null[token.info._string.length] = 0;
          get_token();   // skip 'extern'
        }
      }
      else
      {
        syntax_error ("'callback' or 'extern' expected", &token.pos);
        skip_until (RIGHT_BRACKET, false);
      }

      if (token.kind == RIGHT_BRACKET)
      {
        get_token ();
      }
      else
      {
        syntax_error ("']' expected", &token.pos);
        skip_until (RIGHT_BRACKET, true);
      }
    }

    memcpy (&public_pos, &token.pos, sizeof(&public_pos));
    if (token.kind == TOKEN_public)
    {
      func.is_public = true;
      get_token ();
    }

    if (token.kind == TOKEN_inline)
    {
      func.is_inline = true;
      get_token ();
    }

    if (token.kind == TOKEN_void)
    {
      func.return_type = type_void;
      get_token ();
    }
    else
    {
      func.return_type = parse_type_definition (&token.pos, null);
    }


    // In a typedef declaration, the option "callback" is allowed, the option
    // "extern" and the keywords "public" and "inline" are not allowed.

    if (in_typedef)
    {
      if (func.extern_dll_or_null != null || func.is_public || func.is_inline)
      {
        semantic_error ("option not allowed in a typedef declaration", pos0);
        free (func.extern_dll_or_null);
        func.extern_dll_or_null = null;
        func.is_public = false;
        func.is_inline = false;
      }
    }

    // In a generic formal part, the options "callback", "extern" and the keywords
    // "public" and "inline" are not allowed.

    if (in_generic_part)
    {
      if (func.is_callback || func.extern_dll_or_null != null || func.is_public || func.is_inline)
      {
        semantic_error ("option not allowed in generic part", pos0);
        func.is_callback = false;
        free (func.extern_dll_or_null);
        func.extern_dll_or_null = null;
        func.is_public = false;
        func.is_inline = false;
      }
    }
  }


  // A function's return type is defined either by the keyword 'void', or by a
  // type definition which can only denote one of the following types :
  //   integer, enumeration, floating-point, pointer, function pointer,
  //   unsafe pointer.
  // It cannot denote an array/struct/union/opaque/limited/generic type.

  func_base_type = base_type_of (func.return_type);

  if (func_base_type != null)
  {
    if (func_base_type->kind != AN_INTEGER_TYPE &&
        func_base_type->kind != A_FLOAT_TYPE &&
        func_base_type->kind != AN_ENUMERATION_TYPE &&
        func_base_type->kind != A_POINTER_TYPE &&
        func_base_type->kind != A_FUNCTION_POINTER_TYPE &&
        func_base_type->kind != AN_UNSAFE_POINTER_TYPE &&
        func_base_type->kind != A_VOID_TYPE)
    {
      semantic_error ("function has an illegal return type", pos0);
      func.return_type = type_int;
    }
  }


  if (token.kind != IDENTIFIER)
  {
    syntax_error ("identifier expected", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }


  if ((!in_typedef) && (!in_generic_part) && body_allowed)
  {
    // it could be a function body for an earlier function declaration;
    // search for an entity having the same identifier immediately within this region.

    e = search_entity_in_local_scope (token.info._identifier.value);

    if (e != null && e->kind == A_FUNCTION_DECLARATION)
    {
      A_FUNCTION_POINTER_TYPE_ENTITY *ft;

      ft = &e->info->the_function_declaration.to_type->info->the_function_pointer_type;

      if ((ft->extern_dll_or_null == null) &&
          (e->info->the_function_declaration.to_function_body_or_null == null))
      {
        // it is a function body for an earlier declaration

        if (ft->is_public && !func.is_public)
          semantic_error ("keyword 'public' is required when the function was declared earlier in a .h unit or package declaration", &public_pos);
        if (!ft->is_public && func.is_public)
          semantic_error ("keyword 'public' is not allowed", &public_pos);
        if (ft->is_callback != func.is_callback)
          semantic_error ("keyword 'callback' does not match with function declaration", pos0);

        parse_function_specification_for_body (pos0, e, ft, &func);

        if (token.kind == LEFT_ACCOLADE)
          parse_function_body (e);
        else
        {
          syntax_error ("'{' expected", &token.pos);
          skip_until (SEMICOLON, true);
        }

        *body_was_parsed = true;
        return;
      }
    }
  }


  func.parameters = new_region ();

  // save identifier for later function declaration
  wstrcpy (id, token.info._identifier.value);
  memcpy (&id_pos, &token.pos, sizeof(id_pos));



  // The keyword "public" is allowed only for a function body; it must be
  // specified if the function was declared earlier in a .h unit or in a package
  // declaration.

  if ((!in_typedef) && (!in_generic_part))
  {
    surrounding_kind = surrounding_entity()->kind;

    // stand-alone function declaration or function body

    if (surrounding_kind == A_UNIT_INTERFACE || surrounding_kind == A_PACKAGE_DECLARATION)
    {
      if (func.is_public)     // keyword public was specified (but is not allowed)
      {
        if (surrounding_kind == A_PACKAGE_DECLARATION)
          semantic_error ("keyword 'public' is not allowed in a package declaration", &public_pos);
        else
          semantic_error ("keyword 'public' is not allowed in a .h unit", &public_pos);
      }
      func.is_public = true;   // set .is_public to true so that keyword 'public' is required for the body.
    }
    else   // we're in a unit body or package body
    {
      if (func.is_public)
      {
        semantic_error ("keyword 'public' is not allowed", &public_pos);
        func.is_public = false;
      }
    }
  }


  if (in_typedef)
  {
    // check if it's the full type of an earlier incomplete type
    incomplete = get_incomplete_type (id);

    functype = append_new_entity (A_FUNCTION_POINTER_TYPE,
                                  &func,
                                  incomplete != null ? L"" : id,
                                  &id_pos);

    fill_incomplete_type (incomplete, functype, &id_pos);

    hide_entity (functype);

    if (incomplete != null)
      hide_entity (incomplete);
  }
  else  // anonymous function type entity
  {
    functype = append_new_entity (A_FUNCTION_POINTER_TYPE, &func, L"", &token.pos);
  }

  get_token();    // skip identifier

  create_region_level ();
  append_region (func.parameters);

  parse_parentheized_formal_parameters (/* inside_typedef_declaration= */ in_typedef,
                                        (byte)(func.is_callback || (func.extern_dll_or_null != null)));

  unlink_region_level ();

  unhide_all_entities ();


  if (in_generic_part)
  {
    A_GENERIC_FUNCTION_ENTITY ren;

    memset (&ren, 0x00, sizeof(ren));
    ren.to_type = functype;

    e = append_new_entity (A_GENERIC_FUNCTION, &ren, id, &id_pos);
  }
  else if (!in_typedef)      // declare a function declaration with the saved identifier
  {
    A_FUNCTION_DECLARATION_ENTITY ren;

    memset (&ren, 0x00, sizeof(ren));
    ren.to_type                  = functype;
    ren.to_function_body_or_null = null;

    e = append_new_entity (A_FUNCTION_DECLARATION, &ren, id, &id_pos);

    if (wstrcmp (id, L"main") == 0 && (!func.is_callback) && func.extern_dll_or_null == null)
    {
      PENTITY par;
      int     nb_par;

      if (func.return_type != type_void && !types_are_equal (func.return_type, type_int))
        semantic_error ("function main() must have return type 'void' or 'int'", &public_pos);

      par = func.parameters->entities.first;
      nb_par = 0;
      while (par != null)
      {
        if (par->kind == A_PARAMETER)
        {
          nb_par++;
          if (!types_are_equal (par->info->the_parameter.type, type_array_of_string))
          {
            nb_par = 10;   // will trigger error below
            break;
          }
        }
        par = par->next;
      }

      function_main_has_parameter_array_of_string = (bool)(nb_par > 0);

      if (nb_par > 1)
        semantic_error ("function main() must have either no parameters or one parameter of type 'string[]'", &public_pos);

      if (func_main != null)
        semantic_error ("only one function main() is allowed in a program", &public_pos);

      func_main = e;
    }
  }

  if (in_typedef || in_generic_part || (!body_allowed) || func.extern_dll_or_null != null ||
      token.kind != LEFT_ACCOLADE)
  {
    parse_semicolon ();
    return;
  }

  parse_function_body (e);
  *body_was_parsed = true;
}

//=================================================================================

static void parse_typedef_declaration ()
{
  TEXT_POSITION pos0;
  PENTITY       t;

  memcpy (&pos0, &token.pos, sizeof(pos0));

  if (token.kind == LEFT_BRACKET || token.kind == TOKEN_public ||
      token.kind == TOKEN_inline || token.kind == TOKEN_void)
  {
    bool  dummy;
    parse_function_specification (/*pos0            = */ &pos0,
                                  /*t               = */ null,
                                  /*body_allowed    = */ false,
                                  /*body_was_parsed = */ &dummy,
                                  /*in_typedef      = */ true,
                                  /*in_generic_part = */ false);
    return;
  }

  t = parse_type_definition (&token.pos, null);

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("identifier expected", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }

  get_look_ahead_token ();     // fill 'look_ahead_token'

  if (look_ahead_token.kind == LEFT_PARENTHESIS && !is_open_type (t))
  {
    bool dummy;
    parse_function_specification (/*pos0            = */ &pos0,
                                  /*t               = */ t,
                                  /*body_allowed    = */ false,
                                  /*body_was_parsed = */ &dummy,
                                  /*in_typedef      = */ true,
                                  /*in_generic_part = */ false);
  }
  else
  {
    parse_typedef_declaration_of_type_definition (t, &pos0);
  }
}

//=================================================================================

// when t is non-null, a type definition has been parsed earlier and an identifier follows.
// when t is null, current token is : "[", public, inline or void.

void parse_function_declaration_or_function_body (TEXT_POSITION *pos0,
                                                  PENTITY       t,
                                                  bool          body_allowed,
                                                  bool          *body_was_parsed)   // ref
{
  parse_function_specification (/*pos0            = */ pos0,
                                /*t               = */ t,
                                /*body_allowed    = */ body_allowed,
                                /*body_was_parsed = */ body_was_parsed,
                                /*in_typedef      = */ false,
                                /*in_generic_part = */ false);
}

//=================================================================================

static void parse_incomplete_declaration ()
{
  AN_INCOMPLETE_TYPE_ENTITY inc;

  memset (&inc, 0x00, sizeof(inc));

  append_new_entity (AN_INCOMPLETE_TYPE, &inc, token.info._identifier.value, &token.pos);

  get_token();    // skip identifier
  get_token();    // skip semicolon
}

//=================================================================================

void parse_typedef_or_incomplete_declaration ()
{
  get_token();      // skip 'typedef'

  if (token.kind == IDENTIFIER)
  {
    get_look_ahead_token ();     // fill 'look_ahead_token'

    if (look_ahead_token.kind == SEMICOLON)
    {
      parse_incomplete_declaration ();
      return;
    }
  }

  parse_typedef_declaration ();
}

//=================================================================================

static void parse_opaque_type_declaration ()
{
  AN_OPAQUE_TYPE_ENTITY opa;
  ENTITY_KIND           surrounding_kind;

  memset (&opa, 0x00, sizeof(opa));
  opa.full_type = null;
  opa.is_full_type_visible = false;
  opa.is_limited = false;

  append_new_entity (AN_OPAQUE_TYPE, &opa, token.info._identifier.value, &token.pos);

  surrounding_kind = surrounding_entity()->kind;
  if (surrounding_kind != A_UNIT_INTERFACE && surrounding_kind != A_PACKAGE_DECLARATION)
    semantic_error ("an opaque type declaration is only allowed within an interface unit or package declaration", &token.pos);

  get_token();    // skip identifier
  get_token();    // skip semicolon
}

//=================================================================================

// used for struct and union types.

//  field_declaration ::=
//    type_definition  declarator  {"," declarator}  ";"

static void parse_field_declarations (PENTITY estru,   // struct or union
                                      bool    is_packed,
                                      bool    in_variant,
                                      uint4   discriminant_value,
                                      bool    *is_unsafe)
{
  for (;;)
  {
    PENTITY                t, typ;
    wchar                  id[MAX_IDENTIFIER_LENGTH+1];
    TEXT_POSITION          typ_pos, pos;
    A_FIELD_ENTITY         field;
    A_VARYING_FIELD_ENTITY vfield;

    if (token.kind == TOKEN_switch || token.kind == TOKEN_case || token.kind == RIGHT_ACCOLADE)
      break;

    memcpy (&typ_pos, &token.pos, sizeof(typ_pos));
    t = parse_type_definition (&token.pos, null);
    if (t == null)    // some error (not a type)
    {
      skip_until (RIGHT_ACCOLADE, false);
      break;
    }

    for (;;)
    {
      if (token.kind != IDENTIFIER)
      {
        syntax_error ("identifier expected here", &token.pos);
        skip_until (SEMICOLON, true);
        break;
      }

      memcpy (&pos, &token.pos, sizeof(pos));
      wstrcpy (id, token.info._identifier.value);

      get_token();      // skip identifier

      typ = parse_declarator_suffix (t, &typ_pos);

      if (type_has_circular_dependences (estru, typ, in_variant))
      {
        semantic_error ("field type has circular dependencies", &typ_pos);
        typ = type_int;
      }
      else if (is_packed || estru->kind == A_UNION_TYPE)  // must be packed, non-open, non-jagged
      {
        if (!is_packed_type (typ))
        {
          semantic_error ("must be a packed type", &typ_pos);
          typ = type_int;
        }
        else if (is_open_type (typ))
        {
          semantic_error ("type must have a constraint", &typ_pos);
          typ = type_int;
        }
        else if (is_jagged_type (typ))
        {
          semantic_error ("a jagged type is not allowed", &typ_pos);
          typ = type_int;
        }
      }


      if (is_unsafe_type (typ))
        *is_unsafe = true;


      if (in_variant)
      {
        memset (&vfield, 0x00, sizeof(vfield));
        vfield.type = typ;
        vfield.discriminant_value = discriminant_value;
        append_new_entity (A_VARYING_FIELD, &vfield, id, &pos);
      }
      else
      {
        memset (&field, 0x00, sizeof(field));
        field.type = typ;
        append_new_entity (A_FIELD, &field, id, &pos);
      }

      if (token.kind == COMMA)
      {
        get_token();
      }
      else if (token.kind == SEMICOLON)
      {
        get_token();
        break;
      }
      else
      {
        syntax_error ("',' or ';' expected", &token.pos);
        skip_until (SEMICOLON, true);
        break;
      }
    }
  }
}

//=================================================================================

static int sw_compare (void               *user,
                       const uint4        *data1,
                       const unsigned int size1,
                       const uint4        *data2,
                       const unsigned int size2)
{
  if (*data1 < *data2)
    return -1;
  if (*data1 > *data2)
    return +1;
  return 0;
}

//=================================================================================

//  variant_part ::= "switch" "(" identifier ")"
//                   "{" {struct_variant} "}"
//
//  struct_variant ::= "case" constant_expression ":"
//                     variant_declarations
//
//  variant_declarations ::=
//        "null"  ";"
//      | field_declaration { field_declaration }

static void parse_variant_part (PENTITY estru,
                                bool    is_packed,
                                PENTITY discr,
                                bool    *is_unsafe)
{
  TEXT_POSITION pos;
  PEXPRESSION   exp;
  PENTITY       enumer;
  uint4         discriminant_value;
  BINARY_TREE   btree;
  int           rc;

  get_token();    // skip 'switch'

  if (token.kind != LEFT_PARENTHESIS)
  {
    syntax_error ("'(' expected here", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
    return;
  }

  get_token();   // skip '('

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("identifier expected here", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
  }

  // check if identifier matches discriminant identifier

  if (wstrcmp (discr->identifier_or_null, token.info._identifier.value) != 0)
    semantic_error ("identifier must match with the discriminant name", &token.pos);

  get_token();      // skip identifier

  if (token.kind != RIGHT_PARENTHESIS)
  {
    syntax_error ("')' expected here", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
    return;
  }

  get_token();   // skip ')'

  if (token.kind != LEFT_ACCOLADE)
  {
    syntax_error ("'{' expected here", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
    return;
  }

  get_token();   // skip '{'


  create_btree (&btree, NULL, sw_compare);


  for (;;)
  {
    if (token.kind != TOKEN_case)
      break;

    get_token ();   // skip 'case'

    memcpy (&pos, &token.pos, sizeof(pos));

    exp = parse_constant_expression ();


    // check if expression's type and range are valid

    discriminant_value = 0;   // default

    if (exp->base_type_or_null != null)
    {
      enumer = complete_type_of (discr->info->the_field.type);

      if (!types_are_equal (exp->base_type_or_null, enumer))
      {
        semantic_error ("expression must have discriminant's type", &pos);
      }
      else if (exp->kind != A_CONST_ENUMERATION_VALUE)
      {
        semantic_error ("expression must be constant", &pos);
      }
      else
      {
        discriminant_value = exp->info.const_enumeration_value_info.value;
        if (discriminant_value < 0 || discriminant_value > enumer->info->the_enumeration_type.last)
        {
          semantic_error ("expression is out of range", &pos);
          discriminant_value = 0;
        }
        else
        {
          // check that discriminant_value is unique among cases

          rc = insert_btree (&btree, &discriminant_value, sizeof(discriminant_value));
          if (rc == 0)
            ;
          else if (rc == BT_DUPLICATE_KEY)
            semantic_error ("case constant given twice", &pos);
          else
            fatal_compiler_error ("insert_case_key", &pos);
        }
      }
    }

    if (token.kind != COLON)
    {
      syntax_error ("':' expected here", &token.pos);
      skip_until (RIGHT_ACCOLADE, true);
      close_btree (&btree);
      return;
    }

    get_token();   // skip ':'

    if (token.kind == TOKEN_null)
    {
      get_token();         // skip 'null'
      parse_semicolon ();
    }
    else  // field declarations
    {
      if (token.kind == TOKEN_case || token.kind == RIGHT_ACCOLADE)
        syntax_error ("field declaration expected here", &token.pos);
      else
        parse_field_declarations (estru, is_packed, true, discriminant_value, is_unsafe);
    }
  }


  close_btree (&btree);


  if (token.kind != RIGHT_ACCOLADE)
  {
    syntax_error ("token 'case' or '}' expected here", &token.pos);

    skip_until (RIGHT_ACCOLADE, true);
    return;
  }

  get_token();   // skip '}'
}

//=================================================================================

//  opaque_declaration ::=  "struct"  identifier  ";"
//
//  struct_declaration ::=
//    ["packed"] "struct" identifier [ discriminant_part ]
//    "{"  { field_declaration }  [ variant_part ]  "}"
//
//  discriminant_part ::= "(" type_name identifier ")"
//
//  The type name of a discriminant part must denote an enumeration type.
//  A struct variant must be present if and only if there is a discriminant
//  part, the struct type denotes then an open struct type.

void parse_struct_or_opaque_declaration ()
{
  bool                 is_packed;
  TEXT_POSITION        pos0, stru_pos;
  A_STRUCT_TYPE_ENTITY stru;
  PENTITY              incomplete, opaque, estru, discr = null;

  memcpy (&pos0, &token.pos, sizeof(pos0));

  is_packed = false;
  if (token.kind == TOKEN_packed)
  {
    is_packed = true;
    get_token();
  }

  if (token.kind != TOKEN_struct)
  {
    syntax_error ("'struct' expected", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
    return;
  }

  get_token();

  memcpy (&stru_pos, &token.pos, sizeof(stru_pos));

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("identifier expected", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
    return;
  }

  get_look_ahead_token ();     // fill 'look_ahead_token'

  if (look_ahead_token.kind == SEMICOLON)   // opaque declaration
  {
    if (is_packed)
      syntax_error ("keyword 'packed' is not allowed for an opaque type", &pos0);
    parse_opaque_type_declaration ();
    return;
  }


  // check if it's the full type of an earlier incomplete or opaque type

  incomplete = get_incomplete_type (token.info._identifier.value);
  opaque     = get_opaque_type     (token.info._identifier.value);


  memset (&stru, 0x00, sizeof(stru));
  stru.is_packed    = is_packed;
  stru.is_open_type = (look_ahead_token.kind == LEFT_PARENTHESIS);
  stru.discriminant = new_region ();
  stru.fields       = new_region ();

  estru = append_new_entity
      (A_STRUCT_TYPE,
       &stru,
       (incomplete != null || opaque != null) ? L"" : token.info._identifier.value,
       &token.pos);


  fill_incomplete_type (incomplete, estru, &token.pos);
  fill_opaque_type     (opaque,     estru, &token.pos);


  global_locked_struct_entity = estru;


  get_token();    // skip identifier

  create_region_level ();

  append_region (stru.discriminant);

  if (token.kind == LEFT_PARENTHESIS)    // discriminant part
  {
    TEXT_POSITION  pos;
    PENTITY        t;
    A_FIELD_ENTITY field;

    get_token();    // skip '('

    memcpy (&pos, &token.pos, sizeof(pos));

    t = parse_type_name ();

    if (t == null)         // earlier error
      t = type_wchar;

    if (complete_type_of(t)->kind != AN_ENUMERATION_TYPE)
    {
      semantic_error ("an enumeration type is expected here", &pos);
      t = type_wchar;     // force enumeration type
    }

    if (token.kind != IDENTIFIER)
    {
      syntax_error ("an identifier is expected here", &token.pos);
      wstrcpy (token.info._identifier.value, L"$");      // dummy identifier
    }

    memset (&field, 0x00, sizeof(field));
    field.type = t;

    discr = append_new_entity (A_FIELD, &field, token.info._identifier.value, &token.pos);

    if (token.kind == IDENTIFIER)
      get_token();

    if (token.kind == RIGHT_PARENTHESIS)
      get_token();    // skip ')'
    else
    {
      syntax_error ("')' is expected here", &token.pos);
      skip_until (LEFT_ACCOLADE, false);
    }
  }

  append_region (stru.fields);

  if (token.kind == LEFT_ACCOLADE)
  {
    get_token();
  }
  else
  {
    syntax_error ("'{' expected here", &token.pos);
    skip_until (LEFT_ACCOLADE, true);
  }

  parse_field_declarations (estru,
                            is_packed,
                            false,
                            0,
                            &estru->info->the_struct_type.is_unsafe);

  if (token.kind == TOKEN_switch)
  {
    if (!stru.is_open_type)
    {
      syntax_error ("a variant part is only allowed if the type has a discriminant", &token.pos);
      skip_until (RIGHT_ACCOLADE, true);
    }
    else
    {
      parse_variant_part (estru,
                          is_packed,
                          discr,
                          &estru->info->the_struct_type.is_unsafe);
    }
  }
  else
  {
    if (stru.is_open_type)
      syntax_error ("keyword 'switch' is expected here", &token.pos);
  }

  unlink_region_level ();
  global_locked_struct_entity = null;


  // repeat this test now that all fields have been appended to the struct

  if (incomplete != null)
  {
    if (is_jagged_type (estru))
      semantic_error ("a jagged type is not allowed as the full type of an earlier incomplete type", &stru_pos);
  }

  if (opaque != null)
  {
    if (is_jagged_type (estru))
      semantic_error ("a jagged type is not allowed as the full type of an earlier opaque type", &stru_pos);
    else if (is_open_type (estru))
      semantic_error ("an open type is not allowed as the full type of an earlier opaque type", &stru_pos);
  }


  if (token.kind == RIGHT_ACCOLADE)
  {
    get_token();
  }
  else
  {
    syntax_error ("'}' expected here", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
  }
}

//=================================================================================

void parse_union_declaration ()
{
  A_UNION_TYPE_ENTITY  un;
  PENTITY              incomplete, u;

  get_token();   // skip 'union'

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("identifier expected", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
    return;
  }


  // check if it's the full type of an earlier incomplete type

  incomplete = get_incomplete_type (token.info._identifier.value);


  memset (&un, 0x00, sizeof(un));
  un.fields = new_region ();

  u = append_new_entity (A_UNION_TYPE,
                         &un,
                         incomplete != null ? L"" : token.info._identifier.value,
                         &token.pos);

  fill_incomplete_type (incomplete, u, &token.pos);

  hide_entity (u);

  if (incomplete != null)
    hide_entity (incomplete);


  get_token();    // skip identifier


  create_region_level ();

  append_region (un.fields);

  if (token.kind == LEFT_ACCOLADE)
  {
    get_token();
  }
  else
  {
    syntax_error ("'{' expected here", &token.pos);
    skip_until (LEFT_ACCOLADE, true);
  }

  parse_field_declarations (u,
                            /*is_packed  =*/ true,
                            /*in_variant =*/ false,
                            /*discr_value=*/ 0,
                            &u->info->the_union_type.is_unsafe);

  unlink_region_level ();

  unhide_all_entities ();

  if (token.kind == RIGHT_ACCOLADE)
  {
    get_token();
  }
  else
  {
    syntax_error ("'}' expected here", &token.pos);
    skip_until (RIGHT_ACCOLADE, true);
  }
}

//=================================================================================

void parse_ref_declaration ()
{
  PENTITY            t, typ, rf;
  wchar              id[MAX_IDENTIFIER_LENGTH+1];
  TEXT_POSITION      pos, typ_pos, exp_pos;
  A_REFERENCE_ENTITY refer;
  CONTEXT            context;
  PEXPRESSION        exp;

  get_token();

  memcpy (&typ_pos, &token.pos, sizeof(typ_pos));
  t = parse_type_definition (&token.pos, null);

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("identifier expected", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }

  wstrcpy (id, token.info._identifier.value);
  memcpy (&pos, &token.pos, sizeof(pos));

  get_token();      // skip identifier

  typ = parse_declarator_suffix (t, &typ_pos);

  memset (&refer, 0x00, sizeof(refer));
  refer.type = typ;       // can be changed later
  refer.mode = MODE_IN;   // default
  refer.is_used = false;
  refer.references_heap_object = false;
  memcpy (&refer.pos, &pos, sizeof(TEXT_POSITION));
  store_location (&refer.loc);

  rf = create_new_entity (A_REFERENCE, &refer, id, &pos);

  if (token.kind != ASSIGN)
  {
    syntax_error ("'=' expected here", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }

  get_token();      // skip '='

  hide_entity (rf);

  memcpy (&exp_pos, &token.pos, sizeof(exp_pos));

  convert_type_into_context (typ, &context);

  parse_unary_expression (&context, &token.pos, null, &exp);

  unhide_all_entities ();

  if (exp->form != AN_OBJECT)
    semantic_error ("an object is expected here", &exp_pos);


  // if 'typ' denotes an unconstrained array/struct,
  // replace 'typ' by a constrained type built from the expression's constraint.

  if (is_open_type (typ) && exp->constraint.kind == CONSTANT_CONSTRAINT)
  {
    PENTITY tt;

    tt = complete_type_of (typ);

    if (tt->kind == AN_OPEN_ARRAY_TYPE)
    {
      AN_ARRAY_TYPE_ENTITY array;

      memset (&array, 0x00, sizeof(array));
      array.open_array = typ;
      array.length = (uint4)exp->constraint.value;

      typ = append_new_entity (AN_ARRAY_TYPE, &array, L"", &token.pos);
    }
    else    // A STRUCT TYPE
    {
      A_CONSTRAINED_STRUCT_TYPE_ENTITY  cons;

      memset (&cons, 0x00, sizeof(cons));
      cons.open_struct = typ;
      cons.discriminant_value = (uint4)exp->constraint.value;

      typ = append_new_entity (A_CONSTRAINED_STRUCT_TYPE, &cons, L"", &token.pos);
    }
  }


  // check that object is type compatible with type name
  // if type_name has a constraint, it must match !

  (void) check_ref_parameter_context_compatibility (&context, exp, &exp_pos);


  rf->info->the_reference.mode = (exp->access == ACCESS_READWRITE) ? MODE_REF : MODE_IN;
  rf->info->the_reference.type = typ;
  rf->info->the_reference.name = exp;


  append_entity_in_entity_queue (rf);


  parse_semicolon ();
}

//=================================================================================

// when t is non-null, a type definition has been parsed earlier and an identifier follows.
// when t is null, current token is : const or volatile.

void parse_object_declaration (TEXT_POSITION *pos0, PENTITY t0, bool is_global)
{
  PENTITY       t;
  bool          is_const    = false;
  bool          is_volatile = false;
  CONTEXT       context;
  PEXPRESSION   exp;
  TEXT_POSITION typ_pos, exp_pos;
  bool          is_valid;

  memcpy (&typ_pos, pos0, sizeof(typ_pos));
  t = t0;

  if (t == null)
  {
    if (token.kind == TOKEN_const)
    {
      is_const = true;
      get_token();
    }
    else if (token.kind == TOKEN_volatile)
    {
      is_volatile = true;
      get_token();
    }

    memcpy (&typ_pos, &token.pos, sizeof(typ_pos));
    t = parse_type_definition (&token.pos, null);
  }


  // declare object list of type t

  for (;;)
  {
    wchar          id[MAX_IDENTIFIER_LENGTH+1];
    TEXT_POSITION  id_pos;
    PENTITY        typ, obj;

    if (token.kind != IDENTIFIER)
    {
      syntax_error ("identifier expected", &token.pos);
      skip_until (SEMICOLON, true);
      return;
    }

    wstrcpy (id, token.info._identifier.value);
    memcpy (&id_pos, &token.pos, sizeof(id_pos));

    get_token();      // skip identifier

    typ = parse_declarator_suffix (t, &typ_pos);


    if (is_const)
    {
      A_CONSTANT_ENTITY cte;

      memset (&cte, 0x00, sizeof(cte));
      cte.type = typ;       // can later be replaced by a constrained type
      cte.value = null;     // temporary value
      cte.is_used = false;
      memcpy (&cte.pos, &id_pos, sizeof(TEXT_POSITION));

      obj = create_new_entity (A_CONSTANT, &cte, id, &id_pos);
    }
    else if (is_global)
    {
      A_GLOBAL_VARIABLE_ENTITY global;

      memset (&global, 0x00, sizeof(global));
      global.type = typ;
      global.initial_value_or_null = null;
      global.is_volatile = is_volatile;

      obj = create_new_entity (A_GLOBAL_VARIABLE, &global, id, &id_pos);

      if (is_volatile)
      {
        PENTITY tt;

        tt = complete_type_of (typ);

        if (tt->kind == AN_INTEGER_TYPE && integer_info[tt->info->the_integer_type.type].size <= 4)
        {
          // ok
        }
        else if (tt->kind == A_FLOAT_TYPE && float_info[tt->info->the_float_type.type].size <= 4)
        {
          // ok
        }
        else if (tt->kind == AN_ENUMERATION_TYPE     || tt->kind == A_POINTER_TYPE ||
                 tt->kind == A_FUNCTION_POINTER_TYPE || tt->kind == AN_UNSAFE_POINTER_TYPE)
        {
          // ok
        }
        else
        {
          semantic_error ("keyword volatile is not allowed for this type", pos0);
        }
      }
    }
    else   // local
    {
      A_LOCAL_VARIABLE_ENTITY local;

      memset (&local, 0x00, sizeof(local));
      local.type = typ;
      local.initial_value_or_null = null;
      local.is_read    = false;
      local.is_written = false;
      memcpy (&local.pos, &id_pos, sizeof(TEXT_POSITION));
      store_location (&local.loc);

      obj = create_new_entity (A_LOCAL_VARIABLE, &local, id, &id_pos);

      if (is_volatile)
        semantic_error ("keyword volatile is only allowed for global variables", pos0);
    }


    is_valid = true;

    if (token.kind == ASSIGN)
    {
      if (is_limited_type (typ))
        semantic_error ("a default expression is not allowed for a limited type", &token.pos);

      get_token();      // skip '='

      hide_entity (obj);

      memcpy (&exp_pos, &token.pos, sizeof(exp_pos));

      convert_type_into_context (typ, &context);

      parse_expression (&context, &token.pos, null, &exp);

      if (check_assignment_context_compatibility (&context, exp, &exp_pos) < 0)
      {
        typ = type_int;
        free_exp (exp);
        exp = dummy_expression ();
        is_valid = false;
      }

      unhide_all_entities ();
    }
    else
    {
      if (is_const)
      {
        syntax_error ("'=' expected", &token.pos);    // exp mandatory for a constant
        exp = dummy_expression ();
        typ = type_int;
        is_valid = false;
      }
      else
      {
        exp = null;
      }
    }



    // fill entity with type and expression

    if (is_const)
    {
      if (!is_constant_exp (exp))
      {
        semantic_error ("a constant expression is expected here", &exp_pos);
        free_exp (exp);
        exp = dummy_expression ();
        typ = type_int;
        is_valid = false;
      }
      else if (is_valid)
      {
        // if 'typ' denotes an unconstrained array/struct,
        // replace 'typ' by a constrained type built from the expression's constraint.

        if (is_open_type (typ))
        {
          PENTITY tt;

          if (exp->constraint.kind != CONSTANT_CONSTRAINT)
            fatal_compiler_error ("object decl", &token.pos);

          tt = complete_type_of (typ);

          if (tt->kind == AN_OPEN_ARRAY_TYPE)
          {
            AN_ARRAY_TYPE_ENTITY array;

            memset (&array, 0x00, sizeof(array));
            array.open_array = typ;
            array.length = (uint4)exp->constraint.value;

            typ = append_new_entity (AN_ARRAY_TYPE, &array, L"", &token.pos);
          }
          else    // A STRUCT TYPE
          {
            A_CONSTRAINED_STRUCT_TYPE_ENTITY  cons;

            memset (&cons, 0x00, sizeof(cons));
            cons.open_struct = typ;
            cons.discriminant_value = (uint4)exp->constraint.value;

            typ = append_new_entity (A_CONSTRAINED_STRUCT_TYPE, &cons, L"", &token.pos);
          }
        }
      }

      obj->info->the_constant.type  = typ;
      obj->info->the_constant.value = exp;
    }
    else    // global or local variable
    {
      if (is_open_type (typ))
        semantic_error ("type of variable must have a constraint, or must be a constant", &id_pos);
      else if (is_jagged_type (typ))
        semantic_error ("type of variable must not be jagged, or must be a constant", &id_pos);

      if (is_global)
      {
        if (exp != null && !is_constant_exp (exp))
        {
          semantic_error ("a constant initial expression is expected for a global variable", &exp_pos);
          free_exp (exp);
          exp = dummy_expression ();
        }

        obj->info->the_global_variable.initial_value_or_null = exp;
      }
      else   // local
      {
        obj->info->the_local_variable.initial_value_or_null = exp;
      }
    }


    append_entity_in_entity_queue (obj);


    if (token.kind == COMMA)
    {
      get_token();
    }
    else if (token.kind == SEMICOLON)
    {
      get_token();
      break;
    }
    else
    {
      syntax_error ("',' or ';' expected", &token.pos);
      skip_until (SEMICOLON, true);
      break;
    }
  }
}

//=================================================================================

void create_enumeration_literal_string_table (PENTITY t)
{
  int     size, align;
  uint4   i, j, len;
  POOL    *p, *p0;
  uint4   *ofs;
  PENTITY e;

  // create a table of strings, one per enumeration literal

  (void)size_and_alignment_of_type (type_array_of_string, &size, &align);

  p = new_pool_constant (size * (t->info->the_enumeration_type.last + 1), align);
  ofs = 0;

  e = t;
  for (i=0; i<=t->info->the_enumeration_type.last; i++)
  {
    e = e->next;

    if (e->kind != AN_ENUMERATION_LITERAL)
      fatal_compiler_error ("create_enum_lit", &token.pos);

    len = wstrlen(e->identifier_or_null);
    p0 = new_pool_constant (len, 1);

    for (j=0; j<len; j++)
      store_integer (p0, j, (int8)e->identifier_or_null[j], 1);

    store_reference (p, size*i, p0);                 // jagged reference
    store_integer (p, size*i+address_size, len, 4);  // int4 (length)
  }

  t->info->the_enumeration_type.string_table = p;
}

//=================================================================================
