
// undefined.c : detect un-initialized local variables, out parameters, unreachable statements

#include <stdio.h>
#include <malloc.h>

#include "types.h"
#include "token.h"
#include "lex.h"
#include "error.h"
#include "entities.h"
#include "type.h"
#include "typedecl.h"
#include "undefined.h"

//--------------------------------------------------------------------------------

// evaluation mode for expressions

typedef enum {
  EVAL_READ,    // read object
  EVAL_WRITE,   // write to object
  EVAL_OADDR,   // don't read or write object, but evaluate object address (read array indexes)
} EVAL_MODE;

//--------------------------------------------------------------------------------

// list of btrees (there's one btree per scope level)

typedef struct tree_list {
  BINARY_TREE      tree;
  int              level;   // 0 means .next is null
  struct tree_list *next;
} TREE_LIST;

//--------------------------------------------------------------------------------

// content of a btree entry

typedef enum {
  UNDEFINED,
  INITIALIZED,   // has received a value
  READ,          // was read at least once
  TERMINATED,    // an error was generated, so ignore this entity.
} STATE;

typedef struct
{
  PENTITY  e;      // ptr to local variable without initial exp, or to 'out' parameter.
  STATE    state;  // UNDEFINED, INITIALIZED, READ or TERMINATED.
} UNDEFINED_NODE;

//--------------------------------------------------------------------------------

// function for comparing two entries, for sorting

static int sw_compare (void                 *user,
                       const UNDEFINED_NODE *data1,
                       const unsigned int   size1,
                       const UNDEFINED_NODE *data2,
                       const unsigned int   size2)
{
  if (data1->e->nr < data2->e->nr)
    return -1;
  if (data1->e->nr > data2->e->nr)
    return +1;
  return 0;
}

//--------------------------------------------------------------------------------

// create a new node in btree list

static TREE_LIST *new_node (TREE_LIST *previous)
{
  TREE_LIST *p;

  p = malloc (sizeof(TREE_LIST));
  if (p == null)
    fatal_out_of_memory_error ("new_node_undefined");

  create_btree (&p->tree, NULL, sw_compare);
  p->level = (previous == null) ? 0 : (previous->level + 1);
  p->next = previous;

  return p;
}

//--------------------------------------------------------------------------------

// remove one node from bree list

static void remove_node (TREE_LIST **p)
{
  TREE_LIST *t;

  t = *p;
  *p = t->next;

  close_btree (&t->tree);
  free (t);
}

//--------------------------------------------------------------------------------

static void free_list (TREE_LIST *p)
{
  TREE_LIST *t, *q;

  t = p;
  while (t != null)
  {
    q = t->next;
    close_btree (&t->tree);
    free (t);
    t = q;
  }
}

//--------------------------------------------------------------------------------

static TREE_LIST *clone_list (TREE_LIST *p)
{
  TREE_LIST      *t, *prev;
  short          mode;
  int            rc;
  unsigned int   dummy;
  UNDEFINED_NODE n;

  if (p == null)
    return null;

  prev = clone_list (p->next);

  t = new_node (prev);

  mode = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&p->tree, &n, sizeof(n), sizeof(n), &dummy, mode);
    if (rc == BT_KEY_NOT_FOUND)
      break;
    if (rc < 0)
      fatal_compiler_error ("clone_list : retrieve() failed", &token.pos);

    rc = insert_btree (&t->tree, &n, sizeof(n));
    if (rc < 0)
      fatal_compiler_error ("clone_list : insert() failed", &token.pos);

    mode = BT_LARGER;
  }

  return t;
}

//--------------------------------------------------------------------------------

// merge only common levels

static void merge_list (TREE_LIST *master,   // nodes will be updated
                        TREE_LIST *slave)    // nodes will be read
{
  TREE_LIST      *m, *s;
  short          mode;
  int            rc;
  unsigned int   dummy;
  UNDEFINED_NODE nm, ns;

  m = master;
  s = slave;

  if (m == null || s == null)
    return;

  while (m->level > s->level)
    m = m->next;

  while (s->level > m->level)
    s = s->next;

  while (m != null)
  {
    mode = BT_FIRST;
    for (;;)
    {
      rc = retrieve_btree (&s->tree, &ns, sizeof(ns), sizeof(ns), &dummy, mode);
      if (rc == BT_KEY_NOT_FOUND)
        break;
      if (rc < 0)
        fatal_compiler_error ("merge_list : retrieve1() failed", &token.pos);

      nm.e = ns.e;
      rc = retrieve_btree (&m->tree, &nm, sizeof(nm), sizeof(nm), &dummy, BT_EQUAL);
      if (rc == BT_KEY_NOT_FOUND)
        break;
      if (rc < 0)
        fatal_compiler_error ("merge_list : retrieve2() failed", &token.pos);

      if (ns.state == INITIALIZED && nm.state == INITIALIZED)
        nm.state = INITIALIZED;
      else if (ns.state == INITIALIZED && nm.state == READ)
        nm.state = READ;
      else if (ns.state == READ && nm.state == INITIALIZED)
        nm.state = READ;
      else if (ns.state == READ && nm.state == READ)
        nm.state = READ;
      else if (ns.state == TERMINATED || nm.state == TERMINATED)
        nm.state = TERMINATED;
      else
        nm.state = UNDEFINED;

      rc = update_btree (&m->tree, &nm, sizeof(nm));
      if (rc < 0)
        fatal_compiler_error ("merge_list : update() failed", &token.pos);

      mode = BT_LARGER;
    }

    m = m->next;
    s = s->next;
  }
}

//--------------------------------------------------------------------------------

// insert entity in first node of btree list, with state undefined

static void insert_object (TREE_LIST *p, PENTITY e, STATE s)
{
  int            rc;
  UNDEFINED_NODE n;

  n.e = e;
  n.state = s;

  rc = insert_btree (&p->tree, &n, sizeof(n));
  if (rc < 0)
    fatal_compiler_error ("insert_object", &token.pos);
}

//--------------------------------------------------------------------------------

static STATE get_object (TREE_LIST *list, PENTITY e)
{
  TREE_LIST      *p;
  UNDEFINED_NODE n;
  int            rc, dummy;

  p = list;
  while (p != null)
  {
    n.e = e;
    rc = retrieve_btree (&p->tree, &n, sizeof(UNDEFINED_NODE), sizeof(UNDEFINED_NODE), &dummy, BT_EQUAL);
    if (rc == 0)
      return n.state;

    if (rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error ("get_object", &token.pos);

    p = p->next;
  }

  return TERMINATED;
}

//--------------------------------------------------------------------------------

static void set_object (TREE_LIST *list, PENTITY e, EVAL_MODE eval_mode)
{
  TREE_LIST      *p;
  UNDEFINED_NODE n;
  int            rc, dummy;

  p = list;
  while (p != null)
  {
    n.e = e;
    rc = retrieve_btree (&p->tree, &n, sizeof(UNDEFINED_NODE), sizeof(UNDEFINED_NODE), &dummy, BT_EQUAL);
    if (rc == 0)
    {
      n.state = eval_mode;

      rc = update_btree (&p->tree, &n, sizeof(UNDEFINED_NODE));
      if (rc < 0)
        fatal_compiler_error ("set_object1", &token.pos);

      return;
    }

    if (rc != BT_KEY_NOT_FOUND)
      fatal_compiler_error ("set_object2", &token.pos);

    p = p->next;
  }

  fatal_compiler_error ("set_object3", &token.pos);
}

//--------------------------------------------------------------------------------

// perform eval_mode on any local variable or parameter that's in any btree of list

static void detect_in_exp (PEXPRESSION   exp,
                           EVAL_MODE     eval_mode,
                           bool          override_pos,
                           TEXT_POSITION *pos,
                           TREE_LIST     *list)
{
  int       i;
  TREE_LIST *l2;

  if (exp == null)
    return;

  switch (exp->kind)
  {
    case AN_OPERATOR_VALUE:
      switch (exp->info.operator_value_info.op)
      {
        case OP_CONDITIONAL_TEST:
          {
            detect_in_exp (exp->info.operator_value_info.arg[0],
                           EVAL_READ,
                           override_pos,
                           pos,
                           list);

            l2 = clone_list (list);

            detect_in_exp (exp->info.operator_value_info.arg[1],
                           EVAL_READ,
                           override_pos,
                           pos,
                           l2);

            detect_in_exp (exp->info.operator_value_info.arg[2],
                           EVAL_READ,
                           override_pos,
                           pos,
                           list);

            merge_list (list, l2);
            free_list (l2);
          }
          break;

        case OP_SHORT_CIRCUIT_AND:
        case OP_SHORT_CIRCUIT_OR:
          {
            detect_in_exp (exp->info.operator_value_info.arg[0],
                           EVAL_READ,
                           override_pos,
                           pos,
                           list);

            l2 = clone_list (list);

            detect_in_exp (exp->info.operator_value_info.arg[1],
                           EVAL_READ,
                           override_pos,
                           pos,
                           l2);

            merge_list (list, l2);
            free_list (l2);
          }
          break;

        case OP_ADDRESS_OF:         // operator '&' is considered as write+read access
          {
            detect_in_exp (exp->info.operator_value_info.arg[0],
                           EVAL_WRITE,
                           override_pos,
                           pos,
                           list);

            detect_in_exp (exp->info.operator_value_info.arg[0],
                           EVAL_READ,
                           override_pos,
                           pos,
                           list);
          }
          break;

        case OP_LENGTH:
        case OP_SIZE:
          {
            detect_in_exp (exp->info.operator_value_info.arg[0],
                           EVAL_OADDR,       // prefix is used, but no read or write access
                           override_pos,
                           pos,
                           list);
          }
          break;

        default:
          for (i=0; i<3; i++)
          {
            detect_in_exp (exp->info.operator_value_info.arg[i],
                           eval_mode,
                           override_pos,
                           pos,
                           list);
          }
          break;
      }
      break;

    case A_RUN_CALL:
      detect_in_exp (exp->info.run_call_info.function_call,
                     eval_mode,
                     override_pos,
                     pos,
                     list);
      break;

    case A_FUNCTION_VALUE:
      break;

    case A_FUNCTION_CALL:
      {
        LIST_OF_EXPRESSIONS *p;

        detect_in_exp (exp->info.function_call_info.func,
                       EVAL_READ,
                       override_pos,
                       pos,
                       list);

        // before function call, the out parameters are not read (only any indexes are)
        p = exp->info.function_call_info.param;
        while (p != null)
        {
          detect_in_exp (p->exp,
                         p->e->info->the_parameter.mode == MODE_OUT ? EVAL_OADDR : EVAL_READ,
                         override_pos,
                         pos,
                         list);
          p = p->next;
        }

        // after function call, the out parameters are considered initialized
        p = exp->info.function_call_info.param;
        while (p != null)
        {
          if (p->e->info->the_parameter.mode == MODE_OUT)
          {
            detect_in_exp (p->exp,
                           EVAL_WRITE,
                           override_pos,
                           pos,
                           list);
          }
          p = p->next;
        }
      }
      break;

    case A_DISCRIMINANT_VALUE:
      detect_in_exp (exp->info.discriminant_value_info.prefix,
                     EVAL_OADDR,      // reading the discriminant does not read the object
                     override_pos,
                     pos,
                     list);
      break;

    case AN_UNC_ARRAY_AGGREGATE:
      detect_in_exp (exp->info.unc_array_aggregate_info.element,
                     EVAL_READ,
                     override_pos,
                     pos,
                     list);
      break;


    case AN_AGGREGATE_VALUE:
      {
        LIST_OF_EXPRESSIONS *p;

        p = exp->info.aggregate_value_info.list;
        while (p != null)
        {
          detect_in_exp (p->exp,
                         EVAL_READ,
                         override_pos,
                         pos,
                         list);
          p = p->next;
        }
      }
      break;


    case A_QUALIFIED_EXPRESSION:
      detect_in_exp (exp->info.qualified_expression_info.value,
                     EVAL_READ,
                     override_pos,
                     pos,
                     list);
      break;


    case AN_ARRAY_QUALIFIED_EXPRESSION:
      detect_in_exp (exp->info.array_qualified_expression_info.length,
                     EVAL_READ,
                     override_pos,
                     pos,
                     list);
      detect_in_exp (exp->info.array_qualified_expression_info.value,
                     EVAL_READ,
                     override_pos,
                     pos,
                     list);
      break;

    case A_STRUCT_QUALIFIED_EXPRESSION:
      detect_in_exp (exp->info.struct_qualified_expression_info.discriminant,
                     EVAL_READ,
                     override_pos,
                     pos,
                     list);
      detect_in_exp (exp->info.struct_qualified_expression_info.value,
                     EVAL_READ,
                     override_pos,
                     pos,
                     list);
      break;

    case AN_ALLOCATOR:
      detect_in_exp (exp->info.allocator_info.value,
                     EVAL_READ,
                     override_pos,
                     pos,
                     list);
      break;


    case A_GLOBAL_VARIABLE_OBJECT:
      break;


    case A_LOCAL_VARIABLE_OBJECT:

      if (eval_mode == EVAL_READ)
        exp->info.local_variable_object_info.pobject->info->the_local_variable.is_read = true;

      if (eval_mode == EVAL_WRITE)
        exp->info.local_variable_object_info.pobject->info->the_local_variable.is_written = true;

      switch (get_object (list, exp->info.local_variable_object_info.pobject))
      {
        case UNDEFINED:
          if (eval_mode == EVAL_READ)    // read, and the variable is undefined !
          {
            char str[256];

            sprintf (str, "local variable '%.64S' used without having been initialized",
                          exp->info.local_variable_object_info.pobject->identifier_or_null);
            semantic_error (str,
                            override_pos ? pos : &exp->info.local_variable_object_info.pos);

            set_object (list, exp->info.local_variable_object_info.pobject, TERMINATED);
          }
          else if (eval_mode == EVAL_WRITE)    // write, and the variable is undefined.
          {
            set_object (list, exp->info.local_variable_object_info.pobject, INITIALIZED);
          }
          break;

        case INITIALIZED:
          if (eval_mode == EVAL_READ)
          {
            set_object (list, exp->info.local_variable_object_info.pobject, READ);
          }
          break;

        default:
          break;
      }
      break;


    case A_REFERENCE_OBJECT:

      detect_in_exp (exp->info.reference_object_info.pobject->info->the_reference.name,
                     eval_mode,
                     true,
                     override_pos ? pos : &exp->info.reference_object_info.pos,
                     list);
      break;


    case A_PARAMETER_OBJECT:

      switch (get_object (list, exp->info.parameter_object_info.pobject))
      {
        case UNDEFINED:
          if (eval_mode == EVAL_READ)    // read, and the parameter is undefined !
          {
            char str[256];

            sprintf (str, "parameter '%.64S' used without having been initialized",
                          exp->info.parameter_object_info.pobject->identifier_or_null);

            semantic_error (str,
                            override_pos ? pos : &exp->info.parameter_object_info.pos);

            set_object (list, exp->info.parameter_object_info.pobject, TERMINATED);
          }
          else if (eval_mode == EVAL_WRITE)    // write, and the parameter is undefined.
          {
            set_object (list, exp->info.parameter_object_info.pobject, INITIALIZED);
          }
          break;

        case INITIALIZED:
          if (eval_mode == EVAL_READ)
          {
            set_object (list, exp->info.parameter_object_info.pobject, READ);
          }
          break;

        default:
          break;
      }
      break;


    case AN_ARRAY_ELEMENT_OBJECT:
      {
        PENTITY   prefix_type;
        EVAL_MODE mode;

        mode = eval_mode;

        prefix_type = exp->info.array_element_object_info.prefix->base_type_or_null;
        if (prefix_type != null && prefix_type->kind == AN_UNSAFE_POINTER_TYPE)
        {
          // prefix is an unsafe pointer
          mode = EVAL_READ;
        }
        else  // an array
        {
          if (mode == EVAL_WRITE)   // don't propagate write on parts of the full variable
            mode = EVAL_OADDR;
        }

        detect_in_exp (exp->info.array_element_object_info.prefix,
                       mode,
                       override_pos,
                       pos,
                       list);

        detect_in_exp (exp->info.array_element_object_info.index,
                       EVAL_READ,          // always READ the index
                       override_pos,
                       pos,
                       list);
      }
      break;


    case AN_ARRAY_SLICE_OBJECT:
      {
        PENTITY   prefix_type;
        EVAL_MODE mode;

        mode = eval_mode;

        prefix_type = exp->info.array_element_object_info.prefix->base_type_or_null;
        if (prefix_type != null && prefix_type->kind == AN_UNSAFE_POINTER_TYPE)
        {
          // prefix is an unsafe pointer
          mode = EVAL_READ;
        }
        else  // an array
        {
          if (mode == EVAL_WRITE)   // don't propagate write on parts of the full variable
            mode = EVAL_OADDR;
        }

        detect_in_exp (exp->info.array_slice_object_info.prefix,
                       mode,
                       override_pos,
                       pos,
                       list);

        detect_in_exp (exp->info.array_slice_object_info.index,
                       EVAL_READ,          // always READ the index
                       override_pos,
                       pos,
                       list);

        detect_in_exp (exp->info.array_slice_object_info.length,
                       EVAL_READ,          // always READ the length
                       override_pos,
                       pos,
                       list);
      }
      break;


    case A_STRUCT_FIELD_OBJECT:
      {
        EVAL_MODE mode;

        mode = eval_mode;
        if (mode == EVAL_WRITE)   // don't propagate write on parts of the full variable
          mode = EVAL_OADDR;

        detect_in_exp (exp->info.struct_field_object_info.prefix,
                       mode,
                       override_pos,
                       pos,
                       list);
      }
      break;

    case A_DEREFERENCED_OBJECT:
      detect_in_exp (exp->info.dereferenced_object_info.prefix,
                     EVAL_READ,         // evaluate the pointer, always in read
                     override_pos,
                     pos,
                     list);
      break;


    case AN_UNSAFE_DEREFERENCED_OBJECT:
      detect_in_exp (exp->info.unsafe_dereferenced_object_info.unsafe_ptr_value,
                     EVAL_READ,         // evaluate the pointer, always in read
                     override_pos,
                     pos,
                     list);
      break;


    case AN_ATTR_BYTE_OBJECT:
      detect_in_exp (exp->info.attr_byte_object_info.prefix,
                     eval_mode,
                     override_pos,
                     pos,
                     list);
      break;


    case A_BOXED_OBJECT:
      detect_in_exp (exp->info.boxed_object_info.parameter,
                     eval_mode,
                     override_pos,
                     pos,
                     list);
      break;


    case AN_UNBOXED_OBJECT:
      detect_in_exp (exp->info.unboxed_object_info.parameter,
                     eval_mode,
                     override_pos,
                     pos,
                     list);
      break;


    case A_BOXED_ARRAY_OBJECT:
      {
        LIST_OF_EXPRESSIONS *p;
        p = exp->info.boxed_array_object_info.list;
        while (p != null)
        {
          detect_in_exp (p->exp,
                         eval_mode,
                         override_pos,
                         pos,
                         list);
          p = p->next;
        }
      }
      break;

    default:
      break;
  }
}

//--------------------------------------------------------------------------------

// display warnings for never referenced constants/local variables/ref/parameters

static void generate_warning_for_unused_items (PREGION r)
{
  PENTITY e;
  char    str[256];

  e = r->entities.first;

  while (e != null)
  {
    switch (e->kind)
    {
      case A_CONSTANT:
        if (!e->info->the_constant.is_used)
        {
          sprintf (str, "constant '%.64S' is never used", e->identifier_or_null);
          warning (str, &e->info->the_constant.pos);
        }
        break;

      case A_LOCAL_VARIABLE:
        if (!e->info->the_local_variable.is_used)
        {
          sprintf (str, "local variable '%.64S' is never used", e->identifier_or_null);
          warning (str, &e->info->the_local_variable.pos);
        }
        else if (e->info->the_local_variable.is_written && (!e->info->the_local_variable.is_read))
        {
          sprintf (str, "local variable '%.64S' is never read", e->identifier_or_null);
          warning (str, &e->info->the_local_variable.pos);
        }
        break;

      case A_REFERENCE:
        if (!e->info->the_reference.is_used)
        {
          sprintf (str, "reference '%.64S' is never used", e->identifier_or_null);
          warning (str, &e->info->the_reference.pos);
        }
        break;

      case A_PARAMETER:
        if ((!e->info->the_parameter.is_used) &&
            e->info->the_parameter.mode != MODE_OUT)   // out parameter already has its error message
        {                                              // when it's not initialized.
          sprintf (str, "parameter '%.64S' is never used", e->identifier_or_null);
          warning (str, &e->info->the_parameter.pos);
        }
        break;

      default:
        break;
    }

    e = e->next;
  }
}

//--------------------------------------------------------------------------------

static void output_errors_for_undefined_out_parameters (TREE_LIST *list)
{
  TREE_LIST      *l;
  short          mode;
  int            rc;
  unsigned int   dummy;
  UNDEFINED_NODE n;

  l = list;
  while (l->level > 0)       // skip to level 0
    l = l->next;

  mode = BT_FIRST;
  for (;;)
  {
    rc = retrieve_btree (&l->tree, &n, sizeof(n), sizeof(n), &dummy, mode);
    if (rc == BT_KEY_NOT_FOUND)
      break;
    if (rc < 0)
      fatal_compiler_error ("output_out_param : retrieve() failed", &token.pos);

    if (n.e->kind == A_PARAMETER && n.e->info->the_parameter.mode == MODE_OUT &&
        n.state == UNDEFINED && !n.e->info->the_parameter.init_error)
    {
      char str[256];
      sprintf (str, "out parameter '%S' does not receive a value in all flow paths", n.e->identifier_or_null);
      semantic_error (str, &n.e->info->the_parameter.pos);
      n.e->info->the_parameter.init_error = true;    // used to avoid repeating the error message
    }

    mode = BT_LARGER;
  }
}

//--------------------------------------------------------------------------------

static bool is_empty_or_unused_statement_sequence (PENTITY e_list)
{
  PENTITY e = e_list;

  while (e != null)
  {
    if (e->kind != AN_UNUSED_STATEMENT)
      return false;
    e = e->next;
  }

  return true;
}

//--------------------------------------------------------------------------------

static void detect_unused_statement (PENTITY   e,
                                     TREE_LIST **list)
{
  PENTITY n = e->info->the_unused_statement.item;

  if (n->kind == A_LOCAL_VARIABLE)
  {
    n->info->the_local_variable.is_read = true;

    switch (get_object (*list, n))
    {
      case UNDEFINED:   // read, and the variable is undefined !
        {
          char str[256];

          sprintf (str, "local variable '%.64S' used without having been initialized", n->identifier_or_null);
          semantic_error (str, &n->info->the_local_variable.pos);

          set_object (*list, n, TERMINATED);
        }
        break;

      case INITIALIZED:
        set_object (*list, n, UNDEFINED);
        break;

      case READ:     // was read before
        {
          char str[256];

          sprintf (str, "_unused not allowed for local variable '%.64S'", n->identifier_or_null);
          semantic_error (str, &n->info->the_local_variable.pos);

          set_object (*list, n, TERMINATED);
        }
        break;

      default:
        break;
    }
  }
  else
  {
    n->info->the_parameter.is_used = true;

    switch (get_object (*list, n))
    {
      case UNDEFINED:   // read, and the parameter is undefined !
        {
          char str[256];

          sprintf (str, "parameter '%.64S' used without having been initialized", n->identifier_or_null);

          semantic_error (str, &n->info->the_parameter.pos);

          set_object (*list, n, TERMINATED);
        }
        break;

      case INITIALIZED:
        set_object (*list, n, UNDEFINED);
        break;

      case READ:     // was read before
        {
          char str[256];

          sprintf (str, "_unused not allowed for parameter '%.64S'", n->identifier_or_null);
          semantic_error (str, &n->info->the_parameter.pos);

          set_object (*list, n, TERMINATED);
        }
        break;

      default:
        break;
    }
  }
}

//--------------------------------------------------------------------------------

static void process_any_following_unused_statements (PENTITY   e_list,
                                                     TREE_LIST **list)
{
  PENTITY e = e_list;
  while (e != null && e->kind == AN_UNUSED_STATEMENT)
  {
    detect_unused_statement (e, list);
    e = e->next;
  }
}

//--------------------------------------------------------------------------------

// returns true if code after statement is unreachable

static bool detect_in_region (PREGION   r,
                              TREE_LIST **list,
                              TREE_LIST **pbreak_list,
                              TREE_LIST **pcontinue_list,
                              bool      are_active_statements_following_in_outer_block)
{
  PENTITY   e;
  TREE_LIST *l2, *l3;

  e = r->entities.first;

  while (e != null)
  {
    switch (e->kind)
    {
      case A_LOCAL_VARIABLE:
        if (e->info->the_local_variable.initial_value_or_null == null)
        {
          insert_object (*list, e, UNDEFINED);       // add local variable to tree list
        }
        else
        {
          detect_in_exp (e->info->the_local_variable.initial_value_or_null,
                         EVAL_READ,
                         false,
                         &token.pos,
                         *list);

          insert_object (*list, e, INITIALIZED);       // add local variable to tree list
        }
        break;

      case A_REFERENCE:
        detect_in_exp (e->info->the_reference.name,
                       EVAL_OADDR,   // name need not be initialized, but indexes or slices must be
                       false,
                       &token.pos,
                       *list);
        break;

      case A_CLEAR_STATEMENT:
        detect_in_exp (e->info->the_clear_statement.name,
                       EVAL_WRITE,
                       false,
                       &token.pos,
                       *list);
        break;

      case AN_ASSIGNMENT_STATEMENT:
        detect_in_exp (e->info->the_assignment_statement.name,
                       EVAL_OADDR,    // first evaluate name in mode none (for all indexes)
                       false,
                       &token.pos,
                       *list);

        detect_in_exp (e->info->the_assignment_statement.value,
                       EVAL_READ,
                       false,
                       &token.pos,
                       *list);

        if (e->info->the_assignment_statement.op != _ASSIGN)
        {
          detect_in_exp (e->info->the_assignment_statement.name,
                         EVAL_READ,
                         false,
                         &token.pos,
                         *list);
        }
        else
        {
          detect_in_exp (e->info->the_assignment_statement.name,
                         EVAL_WRITE,
                         false,
                         &token.pos,
                         *list);
        }
        break;

      case A_PRE_OR_POSTFIX_STATEMENT:
        detect_in_exp (e->info->the_pre_or_postfix_statement.name,
                       EVAL_READ,
                       false,
                       &token.pos,
                       *list);
        break;

      case A_FUNCTION_CALL_STATEMENT:
        detect_in_exp (e->info->the_function_call_statement.name,
                       EVAL_READ,
                       false,
                       &token.pos,
                       *list);
        break;

      case A_RETURN_STATEMENT:
        detect_in_exp (e->info->the_return_statement.value,
                       EVAL_READ,
                       false,
                       &token.pos,
                       *list);

        // output error messages for undefined 'out' parameters
        output_errors_for_undefined_out_parameters (*list);

        // flag any following statements as unreachable
        if ((!is_empty_or_unused_statement_sequence (e->next)) || are_active_statements_following_in_outer_block)
          warning ("statements following return are unreachable", &e->info->the_return_statement.pos);

        process_any_following_unused_statements (e->next, list);

        return true;  // unreachable


      case A_BREAK_STATEMENT:          // leaves switch, for, while (can break out of 'block' or 'if')
        if (*pbreak_list == null)
          *pbreak_list = clone_list (*list);
        else
          merge_list (*pbreak_list, *list);

        // flag any following statements as unreachable
        if (e->next != null || are_active_statements_following_in_outer_block)
          warning ("statements following break are unreachable", &e->info->the_break_statement.pos);

        return true;  // unreachable


      case A_CONTINUE_STATEMENT:
        if (*pcontinue_list == null)
          *pcontinue_list = clone_list (*list);
        else
          merge_list (*pcontinue_list, *list);

        // flag any following statements as unreachable
        if (e->next != null || are_active_statements_following_in_outer_block)
          warning ("statements following continue are unreachable", &e->info->the_continue_statement.pos);

        return true;  // unreachable


      case A_FREE_STATEMENT:
        detect_in_exp (e->info->the_free_statement.value,
                       EVAL_READ,
                       false,
                       &token.pos,
                       *list);
        break;

      case AN_ASSERT_STATEMENT:
        detect_in_exp (e->info->the_assert_statement.value,
                       EVAL_READ,
                       false,
                       &token.pos,
                       *list);
        break;

      case AN_ABORT_STATEMENT:

        // flag any following statements as unreachable
        if (e->next != null || are_active_statements_following_in_outer_block)
          warning ("statements following abort are unreachable", &e->info->the_abort_statement.pos);

        return true;   // unreachable


      case A_SLEEP_STATEMENT:
        detect_in_exp (e->info->the_sleep_statement.value,
                       EVAL_READ,
                       false,
                       &token.pos,
                       *list);
        break;


      case AN_UNUSED_STATEMENT:
        detect_unused_statement (e, list);
        break;

      case A_BLOCK_STATEMENT:
        {
          bool unreachable;

          // add new node in tree list
          *list = new_node (*list);

          // and call this function recursively
          unreachable = detect_in_region (e->info->the_block_statement.inner, list, pbreak_list, pcontinue_list,
                           /* are_active_statements_following_in_outer_block = */
                                (bool)(!is_empty_or_unused_statement_sequence (e->next)));

          // display warnings for never referenced constants/local variables/ref/parameters (check entities)
          generate_warning_for_unused_items (e->info->the_block_statement.inner);

          // remove last node
          remove_node (list);

          if (unreachable)
          {
            process_any_following_unused_statements (e->next, list);
            return true;
          }
        }
        break;


      case AN_IF_STATEMENT:
        {
          bool unreachable1, unreachable2;

          detect_in_exp (e->info->the_if_statement.condition,
                         EVAL_READ,
                         false,
                         &token.pos,
                         *list);

          l2 = clone_list (*list);

          unreachable1 = detect_in_region (e->info->the_if_statement.true_branch, &l2, pbreak_list, pcontinue_list, false);

          unreachable2 = detect_in_region (e->info->the_if_statement.false_branch, list, pbreak_list, pcontinue_list, false);

          if (unreachable1 == true && unreachable2 == false)
          {
            free_list (l2);     // keep list
          }
          else if (unreachable1 == false && unreachable2 == true)
          {
            free_list (*list);   // keep l2
            *list = l2;
          }
          else     // both or none unreachable
          {
            merge_list (*list, l2);
            free_list (l2);
          }

          if (unreachable1 && unreachable2)
            return true;
        }
        break;


      case A_SWITCH_STATEMENT:
        {
          FLOW_ALTERNATIVE *alt;
          TREE_LIST        *break_list;
          bool             unreachable, all_unreachable;

          detect_in_exp (e->info->the_switch_statement.value,
                         EVAL_READ,
                         false,
                         &token.pos,
                         *list);

          l2 = null;

          all_unreachable = true;

          alt = e->info->the_switch_statement.first_alt;
          while (alt != null)
          {
            l3 = clone_list (*list);

            break_list = null;

            unreachable = detect_in_region (alt->inner, &l3, &break_list, pcontinue_list, false);
            all_unreachable &= (unreachable && break_list == null);

            if (!unreachable)
            {
              merge_list (l3, break_list);
              free_list (break_list);
            }
            else
            {
              free_list (l3);
              l3 = break_list;     // can be null
            }

            if (l3 != null)
            {
              merge_list (l3, l2);
              free_list (l2);
              l2 = l3;
            }

            alt = alt->next;
          }

          if (l2 != null)    // keep list if l2 is null
          {
            free_list (*list);
            *list = l2;
          }

          if (all_unreachable)
            return true;
        }
        break;

      case A_WHILE_STATEMENT:
        {
          PEXPRESSION exp;
          TREE_LIST   *break_list, *continue_list;

          exp = e->info->the_while_statement.condition;

          detect_in_exp (exp,
                         EVAL_READ,
                         false,
                         &token.pos,
                         *list);

          l2 = clone_list (*list);

          break_list = null;
          continue_list = null;
          detect_in_region (e->info->the_while_statement.inner, &l2, &break_list, &continue_list, false);

          free_list (continue_list);   // continue_list is ignored

          if (exp->kind == A_CONST_ENUMERATION_VALUE &&
              exp->info.const_enumeration_value_info.value == 1)  // true (we can only leave with a break)
          {
            if (break_list == null)
            {
              free_list (break_list);
              free_list (*list);
              *list = l2;

              if (e->next != null || are_active_statements_following_in_outer_block)
                warning ("statements following 'while' statement are unreachable", &e->info->the_while_statement.pos);
              return true;   // unreachable
            }
            else
            {
              // remove some nodes of break_list
              while (break_list->level > (*list)->level)
                remove_node (&break_list);

              free_list (l2);
              free_list (*list);

              *list = break_list;
            }
          }
          else   // anything else than true
          {
            free_list (break_list);
            free_list (l2);
          }
        }
        break;


      case A_FOR_STATEMENT:
        {
          PEXPRESSION exp;
          TREE_LIST   *break_list, *continue_list;

          detect_in_region (e->info->the_for_statement.pre, list, pbreak_list, pcontinue_list, false);

          detect_in_region (e->info->the_for_statement.exp_region, list, pbreak_list, pcontinue_list, false);

          exp = e->info->the_for_statement.condition;   // can be null

          detect_in_exp (exp,
                         EVAL_READ,
                         false,
                         &token.pos,
                         *list);

          l2 = clone_list (*list);

          break_list = null;
          continue_list = null;
          detect_in_region (e->info->the_for_statement.inner, &l2, &break_list, &continue_list, false);

          merge_list (l2, continue_list);
          free_list (continue_list);

          detect_in_region (e->info->the_for_statement.post, &l2, pbreak_list, pcontinue_list, false);

          if (exp == null ||
              (exp->kind == A_CONST_ENUMERATION_VALUE &&
               exp->info.const_enumeration_value_info.value == 1))  // true (we can only leave with a break)
          {
            if (break_list == null)
            {
              free_list (break_list);
              free_list (*list);
              *list = l2;

              if (e->next != null || are_active_statements_following_in_outer_block)
                warning ("statements following 'for' statement are unreachable", &e->info->the_for_statement.pos);
              return true;   // unreachable
            }
            else
            {
              // remove some nodes of break_list
              while (break_list->level > (*list)->level)
                remove_node (&break_list);

              free_list (l2);
              free_list (*list);

              *list = break_list;
            }
          }
          else   // anything else than true
          {
            free_list (break_list);
            free_list (l2);
          }
        }
        break;

      default:
        break;
    }

    e = e->next;
  }

  return false;
}

//--------------------------------------------------------------------------------

static void add_parameters_to_list (PREGION parameters, TREE_LIST *list)
{
  PENTITY e;

  e = parameters->entities.first;

  while (e != null)
  {
    switch (e->kind)
    {
      case A_PARAMETER:
        if (e->info->the_parameter.mode != MODE_OUT)
        {
          insert_object (list, e, INITIALIZED);
        }
        else if (!types_are_equal (base_type_of (e->info->the_parameter.type), type_array_of_object))
        {
          insert_object (list, e, UNDEFINED);
        }
        break;

      default:
        break;
    }

    e = e->next;
  }
}

//--------------------------------------------------------------------------------

void detect_uninitialized_variables (PREGION parameters, PREGION inner)
{
  TREE_LIST *list, *break_list, *continue_list;

  // create a tree list node for function's local scope (level 0)
  list = new_node (null);

  // add parameters to list
  add_parameters_to_list (parameters, list);

  break_list = null;
  continue_list = null;

  if (detect_in_region (inner, &list, &break_list, &continue_list, false) == false)  // reachable
  {
    // output error messages for undefined 'out' parameters
    output_errors_for_undefined_out_parameters (list);
  }

  // free list
  remove_node (&list);

  // display warnings for never referenced constants/local variables/ref/parameters (check entities)
  generate_warning_for_unused_items (parameters);
  generate_warning_for_unused_items (inner);
}

//--------------------------------------------------------------------------------
