
// unit.c : unit analysis

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <io.h>

#include "token.h"
#include "source.h"
#include "strings.h"
#include "token.h"
#include "lex.h"
#include "entities.h"
#include "decl.h"
#include "typedecl.h"
#include "error.h"
#include "complete.h"
#include "dbginfo.h"
#include "libunit.h"

#include "unit.h"

//=================================================================================

typedef struct pending_body {
  char                 *library;
  char                 *source;
  bool                 must_exist;
  struct pending_body *next;
} PENDING_BODY;

PENDING_BODY *pending_bodies;

//=================================================================================

static BINARY_TREE unit_tree;    // ordered by library + source file name
static BINARY_TREE unit_tree2;   // ordered by key

typedef struct {
  char    *library;
  char    *source;
  PENTITY unit;     // not filled for unit_tree2
  int     key;      // unique unit key (0=none) (only .c unit receive a key)
} UNIT_DATA;

static int file_unit_unique_key;      // source .c units receive positive values
static int library_unit_unique_key;   // library .c units receive negative values

//=================================================================================

static int sv_compare (void               *user,
                       const UNIT_DATA    *data1,
                       const unsigned int size1,
                       const UNIT_DATA    *data2,
                       const unsigned int size2)
{
  return 2 * strcmp (data1->library, data2->library)
           + strcmp (data1->source, data2->source);
}

//=================================================================================

static int sv_compare2 (void               *user,
                        const UNIT_DATA    *data1,
                        const unsigned int size1,
                        const UNIT_DATA    *data2,
                        const unsigned int size2)
{
  if (data1->key < data2->key)
    return -1;
  if (data1->key > data2->key)
    return +1;
  return 0;
}

//=================================================================================

void init_unit_tree ()
{
  create_btree (&unit_tree,  NULL, sv_compare);
  create_btree (&unit_tree2, NULL, sv_compare2);
}

//=================================================================================

static char *new_string (string str)
{
  char *s;
  s = malloc (strlen(str)+1);
  if (s == null)
  {
    printf ("fatal error: new_string() failed\n");
    exit (-1);
  }
  strcpy (s, str);
  return s;
}

//=================================================================================

// returns unit's unique key

static int store_empty_unit (string library, string source, bool is_body)
{
  UNIT_DATA u;
  int       rc;

  u.library = new_string (library);
  u.source  = new_string (source);
  u.unit    = null;

  if (is_body)
  {
    if (library[0] == '\0')
      u.key = ++file_unit_unique_key;
    else
      u.key = --library_unit_unique_key;
  }
  else
  {
    u.key = 0;
  }

  rc = insert_btree (&unit_tree, &u, sizeof(u));
  if (rc < 0)
    fatal_compiler_error ("store_empty_unit(1)", &token.pos);

  if (u.key != 0)
  {
    rc = insert_btree (&unit_tree2, &u, sizeof(u));
    if (rc < 0)
      fatal_compiler_error ("store_empty_unit(2)", &token.pos);
  }

  return u.key;
}

//=================================================================================

static void update_unit (string library, string source, PENTITY e)
{
  UNIT_DATA u, v;
  int       rc, dummy;

  u.library = new_string (library);
  u.source  = new_string (source);
  u.unit    = e;

  memcpy (&v, &u, sizeof(v));
  rc = retrieve_btree (&unit_tree, &v, sizeof(v), sizeof(v), &dummy, BT_EQUAL);
  if (rc < 0)
    fatal_compiler_error ("update_unit(1)", &token.pos);

  free (u.library);
  free (u.source);

  v.unit = e;

  rc = update_btree (&unit_tree, &v, sizeof(v));
  if (rc < 0)
    fatal_compiler_error ("update_unit(2)", &token.pos);
}

//=================================================================================

static bool unit_exists (string library, string source)
{
  UNIT_DATA u, v;
  int       rc, dummy;

  u.library = new_string (library);
  u.source  = new_string (source);

  memcpy (&v, &u, sizeof(v));
  rc = retrieve_btree (&unit_tree, &v, sizeof(v), sizeof(v), &dummy, BT_EQUAL);

  free (u.library);
  free (u.source);

  if (rc == BT_KEY_NOT_FOUND)
    return false;

  if (rc < 0)
    fatal_compiler_error ("unit_exists", &token.pos);

  return true;
}

//=================================================================================

static PENTITY retrieve_unit (string library, string source)
{
  UNIT_DATA u, v;
  int       rc, dummy;

  u.library = new_string (library);
  u.source  = new_string (source);

  memcpy (&v, &u, sizeof(v));

  rc = retrieve_btree (&unit_tree, &v, sizeof(v), sizeof(v), &dummy, BT_EQUAL);

  free (u.library);
  free (u.source);

  if (rc < 0)
    fatal_compiler_error ("retrieve_unit", &token.pos);

  return v.unit;
}

//=================================================================================

static int unit_operate (void               *user,
                         const UNIT_DATA    *p,
                         const unsigned int size)
{
  dbg_store_unit_name (p->key, p->library, p->source);
  return 0;
}

//=================================================================================

void list_units_for_debug ()
{
  traverse_btree (&unit_tree2, unit_operate, +1);
}

//=================================================================================

void retrieve_library_and_source_names (int key, char library[], char source[])
{
  UNIT_DATA u;
  int       rc, dummy;

  u.key = key;

  rc = retrieve_btree (&unit_tree2, &u, sizeof(u), sizeof(u), &dummy, BT_EQUAL);
  if (rc < 0)
    fatal_compiler_error ("retrieve_library_and_source_names", &token.pos);

  strcpy (library, u.library);
  strcpy (source, u.source);
}

//=================================================================================

static int compute_absolute_filename (char *dir,
                                      char *source,
                                      char *result,
                                      int  result_size)
{
  int d, i, j;

  // dir is a real filename ending with / or \
  // source has the form  {../} {id/} id

  // important thing is : don't cut dir too much, the device
  //   part must not be changed, so first we need to compute
  //   the device part length.

  // compute d so that it points to the first / not to be deleted

  if (dir[0] != '\0' && dir[1] == ':')     // drive letter
  {
    d = 2;   // points at first '/'
  }
  else if (dir[0] == '\\' && dir[1] == '\\')  // computer name
  {
    d = 2;
    while (dir[d] != '\0' && dir[d] != '\\' && dir[d] != '/')
      d++;
  }
  else
  {
    d = 0;
  }

  i = strlen(dir) - 1;   // end of dir
  j = 0;                 // start of source

  while (memcmp (&source[j], "../", 3) == 0)
  {
    j += 3;

    for (;;)
    {
      i--;

      if (i < d)
        return -1;

      if (dir[i] == '/' || dir[i] == '\\')
        break;
    }
  }

  // source[j] starts with name

  if ((i+1) + (int)strlen(&source[j]) + 1 > result_size)  // too long
    return -1;

  sprintf (result, "%.*s%s", i+1, dir, &source[j]);

  return 0;
}

//=================================================================================

typedef struct import_data {
  char               *library;
  char               *source;
  char               *alias;
  TEXT_POSITION      lib_pos, src_pos, alia_pos;
  struct import_data *next;
} IMPORT_DATA;

typedef struct {
  IMPORT_DATA *head;
  IMPORT_DATA *tail;
} IMPORT_QUEUE;

static void append_import_data (IMPORT_QUEUE *queue, char *library, char *source, char *alias,
                                TEXT_POSITION *lib_pos,  TEXT_POSITION *src_pos, TEXT_POSITION *alia_pos)
{
  IMPORT_DATA *p;

  p = malloc (sizeof(IMPORT_DATA));
  if (p == null)
    fatal_out_of_memory_error ("alloc import data");

  p->library = new_string (library);
  p->source = new_string (source);
  p->alias = new_string (alias);

  memcpy (&p->lib_pos,  lib_pos,  sizeof(TEXT_POSITION));
  memcpy (&p->src_pos,  src_pos,  sizeof(TEXT_POSITION));
  memcpy (&p->alia_pos, alia_pos, sizeof(TEXT_POSITION));

  p->next = null;

  if (queue->head == null)
    queue->head = p;
  else
    queue->tail->next = p;
  queue->tail = p;
}


static void free_import_list (IMPORT_QUEUE *q)
{
  IMPORT_DATA *p, *t;

  p = q->head;
  while (p != null)
  {
    t = p;
    p = p->next;

    free (t->library);
    free (t->source);
    free (t->alias);
    free (t);
  }
  q->head = null;
  q->tail = null;
}

//=================================================================================

static void parse_import_clause (string       parent_library,
                                 string       parent_dir,
                                 IMPORT_QUEUE *queue)
{
  char id[MAX_IDENTIFIER_LENGTH+1], library_name[512], source_name[1024], alias_name[MAX_IDENTIFIER_LENGTH+1];
  char full_filename[260+1];
  TEXT_POSITION  lib_pos, src_pos, alia_pos;

  memcpy (&lib_pos, &token.pos, sizeof(TEXT_POSITION));

  if (token.kind == TOKEN_from)
  {
    get_token ();

    memcpy (&lib_pos, &token.pos, sizeof(TEXT_POSITION));

    if (token.kind != IDENTIFIER)
    {
      syntax_error ("library identifier expected", &token.pos);
      skip_until (SEMICOLON, true);
      return;
    }

    if (!is_all_in_lowercase (token.info._identifier.value))
      syntax_error ("library identifier must be in lower case", &token.pos);

    wstring_to_string (token.info._identifier.value, id);
    strcpy (library_name, id);

    get_token ();


    while (token.kind == DOT)
    {
      get_token ();

      if (token.kind != IDENTIFIER)
      {
        syntax_error ("library identifier expected", &token.pos);
        skip_until (SEMICOLON, true);
        return;
      }

      if (!is_all_in_lowercase (token.info._identifier.value))
        syntax_error ("library identifier must be in lower case", &token.pos);

      wstring_to_string (token.info._identifier.value, id);

      if (strlen(library_name) + 1 + strlen(id) + 1 > sizeof(library_name))
      {
        syntax_error ("library name is too long", &token.pos);
        skip_until (SEMICOLON, true);
        return;
      }

      strcat (library_name, ".");
      strcat (library_name, id);

      get_token ();
    }
  }
  else
  {
    library_name[0] = '\0';
  }


  if (token.kind != TOKEN_use)
  {
    syntax_error ("keyword 'use' expected here", &token.pos);
    skip_until (SEMICOLON, true);
    return;
  }

  get_token ();   // skip 'use'

  for (;;)
  {
    memcpy (&src_pos, &token.pos, sizeof(TEXT_POSITION));

    source_name[0] = '\0';

    while (token.kind == DOUBLE_DOT)
    {
      if (library_name[0] != '\0')
        semantic_error ("'..' is not allowed for source names of library units", &token.pos);

      get_token ();   // skip '..'

      if (token.kind != SLASH)
      {
        syntax_error ("'/' expected here", &token.pos);
        skip_until (SEMICOLON, true);
        return;
      }

      get_token ();   // skip /


      if (strlen(source_name) + 3 + 1 > sizeof(source_name))
      {
        syntax_error ("source name is too long", &token.pos);
        skip_until (SEMICOLON, true);
        return;
      }

      strcat (source_name, "../");
    }


    for (;;)
    {
      if (token.kind != IDENTIFIER)
      {
        syntax_error ("source identifier expected", &token.pos);
        skip_until (SEMICOLON, true);
        return;
      }

      if (!is_all_in_lowercase (token.info._identifier.value))
        syntax_error ("source identifier must be in lower case", &token.pos);

      wstring_to_string (token.info._identifier.value, id);

      if (strlen(source_name) + strlen(id) + 1 > sizeof(source_name))
      {
        syntax_error ("source name is too long", &token.pos);
        skip_until (SEMICOLON, true);
        return;
      }

      strcat (source_name, id);

      get_token ();   // skip id


      if (token.kind != SLASH)
        break;

      get_token ();   // skip '/'

      if (strlen(source_name) + 1 + 1 > sizeof(source_name))
      {
        syntax_error ("source name is too long", &token.pos);
        skip_until (SEMICOLON, true);
        return;
      }

      strcat (source_name, "/");
    }


    memcpy (&alia_pos, &src_pos, sizeof(TEXT_POSITION));    // default pos is source name

    if (token.kind == IDENTIFIER && wstrcmp (token.info._identifier.value, L"as") == 0)
    {
      get_token ();   // skip 'as'

      memcpy (&alia_pos, &token.pos, sizeof(TEXT_POSITION));

      if (token.kind != IDENTIFIER)
      {
        syntax_error ("alias identifier expected", &token.pos);
        skip_until (SEMICOLON, true);
        return;
      }

      wstring_to_string (token.info._identifier.value, alias_name);

      get_token ();   // skip id
    }
    else   // no alias name specified : extract alias from source_name's last component
    {
      int i;

      i = strlen(source_name);
      while (i-1 >= 0 && source_name[i-1] != '/')
        i--;

      strcpy (alias_name, &source_name[i]);
    }


    if (library_name[0] == '\0')    // use parent's library/disk
    {
      if (compute_absolute_filename (parent_dir, source_name, full_filename, 255) < 0)
      {
        syntax_error ("illegal filename", &src_pos);
      }
      else
      {
        strcat (full_filename, ".h");

        append_import_data (queue, parent_library, full_filename, alias_name,
                            &lib_pos, &src_pos, &alia_pos);
      }
    }
    else    // absolute source filename from another library than parent
    {
      sprintf (full_filename, "/%s%s", source_name, ".h");

      append_import_data (queue, library_name, full_filename, alias_name,
                          &lib_pos, &src_pos, &alia_pos);
    }

    if (token.kind != COMMA)
      break;

    get_token ();   // skip comma
  }

  parse_semicolon ();
}

//=================================================================================

// returns the created unit entity

static PENTITY parse_unit (string library, string source, bool is_body)
{
  PENTITY e;
  wchar   unit_name[260+1];
  int     i;

  i = strlen(source);
  while (i-1 >= 0 && source[i-1] != '/' && source[i] != '\\')
    i--;

  string_to_wstring (&source[i], unit_name);
  unit_name[wstrlen(unit_name)-2] = 0;


  if (is_body)      // body unit (.c)
  {
    PENTITY            edecl, ebody;
    A_UNIT_BODY_ENTITY body;
    char               source2[260+1];

    edecl = search_entity_in_local_scope (unit_name);

    if (edecl == null)  // no earlier unit interface of this name declared
    {
      A_UNIT_INTERFACE_ENTITY interf;

      memset (&interf, 0x00, sizeof(interf));
      interf.to_unit_body     = null;
      interf.is_body_required = false;
      interf.unit_is_used     = false;
      interf.declarations     = new_region ();

      edecl = append_new_entity (A_UNIT_INTERFACE, &interf, unit_name, &token.pos);

      sprintf (source2, "%.*s%s", strlen(source)-2, source, ".h");
      update_unit (library, source2, edecl);
    }

    memset (&body, 0x00, sizeof(body));
    body.to_unit_interface = edecl;
    body.declarations = new_region ();

    ebody = append_new_entity (A_UNIT_BODY, &body, L"", &token.pos);
    update_unit (library, source, ebody);

    edecl->info->the_unit_interface.to_unit_body = ebody;

    create_region_level ();
    append_region (edecl->info->the_unit_interface.declarations);
    append_region (body.declarations);


    // when entering region (package body. or unit body),
    // scan package decl. or unit interface :
    // - for opaque type :
    //   . set non-limited

    e = edecl->info->the_unit_interface.declarations->entities.first;
    while (e != null)
    {
      if (e->kind == AN_OPAQUE_TYPE)
      {
        e->info->the_opaque_type.is_limited = false;
      }

      e = e->next;
    }



    parse_global_declarations (true);



    check_global_completion (edecl->info->the_unit_interface.declarations->entities.first);
    check_global_completion (ebody->info->the_unit_body.declarations->entities.first);



    // when leaving region (package body. or unit body),
    // scan package decl. or unit interface :
    // - for incomplete type :
    //   . if full type's keynr is past the package body nr,
    //     it means they are in different parts ! -> full type no longer available.
    // - for opaque type :
    //   . full type no longer available.

    e = edecl->info->the_unit_interface.declarations->entities.first;
    while (e != null)
    {
      if (e->kind == AN_INCOMPLETE_TYPE &&
          e->info->the_incomplete_type.full_type != null &&
          e->info->the_incomplete_type.full_type->nr > ebody->nr)
      {
        e->info->the_incomplete_type.is_full_type_visible = false;
      }

      if (e->kind == AN_OPAQUE_TYPE)
      {
        e->info->the_opaque_type.is_full_type_visible = false;
        e->info->the_opaque_type.is_limited = true;
      }

      e = e->next;
    }

    e = ebody;  // created entity
  }
  else    // interface unit (.h)
  {
    A_UNIT_INTERFACE_ENTITY interf;
    PENTITY                 edecl;

    memset (&interf, 0x00, sizeof(interf));
    interf.to_unit_body     = null;
    interf.is_body_required = false;
    interf.unit_is_used     = false;
    interf.declarations     = new_region ();

    edecl = append_new_entity (A_UNIT_INTERFACE, &interf, unit_name, &token.pos);
    update_unit (library, source, edecl);

    create_region_level ();
    append_region (interf.declarations);


    parse_global_declarations (false);


    // make opaque types limited

    e = edecl->info->the_unit_interface.declarations->entities.first;
    while (e != null)
    {
      if (e->kind == AN_OPAQUE_TYPE)
      {
        e->info->the_opaque_type.is_limited = true;
      }

      e = e->next;
    }


    // tests if a unit body is required

    edecl->info->the_unit_interface.is_body_required =
         is_body_required (edecl->info->the_unit_interface.declarations->entities.first);


    e = edecl;   // created entity
  }


  if (token.kind != LAST_TOKEN)
    syntax_error ("declaration expected", &token.pos);

  unlink_region_level ();

  cleanup_main_region_after_compilation ();

  return e;
}

//=================================================================================

static void intern_append_new_unit_entity_for_import
                      (PENTITY       e,
                       string        filename,
                       TEXT_POSITION *pos,
                       bool          add_in_package_list)
{
  wchar wfilename[512];

  string_to_wstring (filename, wfilename);

  append_new_unit_entity_for_import (e, wfilename, pos, add_in_package_list);
}

//=================================================================================

static void append_to_pending_list (string library, string filename, bool must_exist)
{
  PENDING_BODY *p = malloc (sizeof(PENDING_BODY));

  if (p == NULL)
  {
    printf ("fatal error: malloc() failed\n");
    exit (-1);
  }

  p->library = new_string (library);
  p->source = new_string (filename);
  p->must_exist = must_exist;
  p->next = pending_bodies;

  pending_bodies = p;
}

//=================================================================================

static PENTITY DUMMY_unit = (PENTITY)1;   // unique value, not zero

//=================================================================================

// returns 0 if OK
// returns +1 if ret initialized

static int load_in_lex (string library,      // "", or "/lib/xxx"
                        string filename,
                        bool   must_exist,
                        wchar  **utf16_source_text,
                        int    *ret)
{
  byte  *source;
  int   size;
  wchar *source_text;   // Lnul-terminated
  int   max_length;     // source text maximum characters
  int   rc;

  if (library[0] != '\0')  // load file from library
  {
    if (load_library_source (library, filename, &source, &size) < 0)  // was not found
    {
      if (!must_exist)
      {
        // store a special entity signaling a non-existing optional unit
        update_unit (library, filename, DUMMY_unit);
        *ret = 0;     // done
        return +1;
      }

      *ret = +1;    // returns code for file not found
      return +1;
    }
  }
  else      // load from disk
  {
    if (!must_exist)
    {
      if (access (filename, 0) < 0)      // file does not exist
      {
        // store a special entity signaling a non-existing optional unit
        update_unit (library, filename, DUMMY_unit);
        *ret = 0;     // done
        return +1;
      }
    }

    rc = load_source_file (filename, &source, &size);
    if (rc)
    {
      *ret = +1;    // returns code for file not found
      return +1;
    }
  }


  rc = convert_source_file_to_utf16 (source, size,
                                     &source_text, &max_length);
  if (rc)
  {
    printf ("fatal error: converting '%s' to utf-16 failed\n", filename);
    exit (-1);
  }

  free (source);

  lex_start_new_context (source_text, max_length);
  get_token ();

  *utf16_source_text = source_text;

  return 0;
}

//=================================================================================

static int compile_unit_interface (string library,      // "", or "/lib/xxx"
                                   string filename,     // ends with ".h"
                                   bool   must_exist);

//=================================================================================

static void import_and_compile_all_libraries (string       library,      // "", or "/lib/xxx"
                                              string       filename,
                                              int          key,
                                              IMPORT_QUEUE *pqueue)
{
  char        *unit_dir;
  char        *unit_name;
  int         i, len;
  IMPORT_DATA *l;


  // compute unit's directory and name

  i = strlen(filename);
  if (i < 3)         // at least "p.h"
  {
    printf ("intern error unit_name\n");
    printf ("stopping compilation\n");
    exit (-1);
  }

  i--;
  while (i >= 0 && filename[i] != '/' && filename[i] != '\\')
    i--;

  // here filename[i] denotes a directory separator, or i < 0.
  // filename[i+1] denotes the short name.


  // create first the unit name

  len = strlen(&filename[i+1]) - 2;   // length of short name

  unit_name = malloc (len+1);   // on heap to avoid overloading the stack
  if (unit_name == null)
  {
    printf ("fatal error: cannot allocate unit_name\n");
    exit (-1);
  }

  memcpy (unit_name, &filename[i+1], len);
  unit_name[len] = '\0';


  // create now the unit directory

  len = i+1;

  unit_dir = malloc (len+1);   // on heap to avoid overloading the stack
  if (unit_dir == null)
  {
    printf ("fatal error: cannot allocate unit_dir\n");
    exit (-1);
  }

  memcpy (unit_dir, filename, len);
  unit_dir[len] = '\0';


  // collect all import libraries

  set_current_error_source_file (library, filename, key);

  memset (pqueue, 0x00, sizeof(IMPORT_QUEUE));
  while (token.kind == TOKEN_from || token.kind == TOKEN_use)
    parse_import_clause (library, unit_dir, pqueue);

  free (unit_dir);


  // recursively compile all dependant unit interfaces

  l = pqueue->head;
  while (l != null)
  {
    if (strcmp (unit_name, l->alias) == 0)
    {
      char *msg = malloc (90+4*260+1);   // on heap to avoid overloading the stack
      set_current_error_source_file (library, filename, key);  // set again this file
      sprintf (msg, "name collision importing unit '%s' into file '%s' - change filename or try \"use %s as %s2\"",
               l->alias, filename, l->alias, l->alias);
      if (library[0] != '\0')
        sprintf (&msg[strlen(msg)], " [%s]", library);
      semantic_error (msg, &l->alia_pos);
      printf ("stopping compilation\n");
      exit (-1);
    }

    if (compile_unit_interface (l->library, l->source, /*must_exist=*/true) != 0)
    {
      char *msg = malloc (30+260+260+1);   // on heap to avoid overloading the stack
      set_current_error_source_file (library, filename, key);  // set again this file
      sprintf (msg, "cannot open file %s", l->source);
      if (l->library[0] != '\0')
        sprintf (&msg[strlen(msg)], " [%s]", l->library);
      semantic_error (msg, &l->src_pos);
      printf ("stopping compilation\n");
      exit (-1);
    }
    l = l->next;
  }

  free (unit_name);
}

//=================================================================================

static void generate_warnings_for_unused_units (IMPORT_QUEUE *queue)
{
  IMPORT_DATA  *l;
  char         msg[120], *p;
  PENTITY      e;
  int          len;

  if (nb_of_compilation_errors() > 0)
    return;

  l = queue->head;
  while (l != null)
  {
    e = retrieve_unit (l->library, l->source);
    if (e != null && e != DUMMY_unit && !e->info->the_unit_interface.unit_is_used)
    {
      p = strrchr (l->source, '/');
      if (p == NULL)
        p = l->source;
      else
        p++;

      len = strlen(p) - 2;
      if (len > 90)
        len = 90;

      sprintf (msg, "unit '%.*s' is not used", len, p);

      warning (msg, &l->src_pos);
    }
    l = l->next;
  }
}

//=================================================================================

// returns 0 if OK, +1 if file not found

static int compile_unit_interface (string library,      // "", or "/lib/xxx"
                                   string filename,     // ends with ".h"
                                   bool   must_exist)
{
  int          key, rc;
  wchar        *utf16_source_text;
  IMPORT_QUEUE queue;
  IMPORT_DATA  *l;
  char         *import_filename;
  PENTITY      e, edecl;

  // check if this unit already exists in library, if yes -> done.

  if (unit_exists (library, filename))
    return 0;

  // store a first entry with null entity to avoid infinite recursion
  key = store_empty_unit (library, filename, /*is_body=*/ false);

  set_current_error_source_file (library, filename, key);


  if (load_in_lex (library, filename, must_exist, &utf16_source_text, &rc) != 0)
    return rc;


  import_and_compile_all_libraries (library, filename, key, &queue);


  set_current_error_source_file (library, filename, key);  // set again this file



  // prepare scope with all imported libraries being directly visible
  // using aliases if any;
  // if units were imported twice, make them visible only once.

  l = queue.head;
  while (l != null)
  {
    e = retrieve_unit (l->library, l->source);
    if (e == null)
    {
      semantic_error ("imported unit has circular dependencies", &l->src_pos);
    }
    else if (e == DUMMY_unit)   // unit imported non-existing .h of main program
    {
      semantic_error ("unit interface file does not exist", &l->src_pos);
    }
    else
    {
      e->info->the_unit_interface.unit_is_used = false;

      intern_append_new_unit_entity_for_import
                      (e,
                       l->alias,
                       &l->alia_pos,
                       /* add_in_package_list = */ true);
    }

    l = l->next;
  }

  edecl = parse_unit (library, filename, /*is_body=*/ false);


  generate_warnings_for_unused_units (&queue);

  free_import_list (&queue);


  lex_close_context ();
  free (utf16_source_text);



  // put the corresponding .c body on a pending list

  import_filename = malloc (260+1);   // on heap to avoid overloading the stack
  if (import_filename == null)
  {
    printf ("fatal error: cannot allocate import_filename\n");
    exit (-1);
  }

  sprintf (import_filename, "%.*s%s", strlen(filename)-2, filename, ".c");
  append_to_pending_list (library, import_filename, edecl->info->the_unit_interface.is_body_required);

  free (import_filename);

  return 0;
}

//=================================================================================

// returns 0 if OK, +1 if file not found

static int compile_unit_body (string library,      // "", or "/lib/xxx"
                              string filename,     // ends with ".c"
                              bool   must_exist)
{
  int          key, rc;
  wchar        *utf16_source_text;
  IMPORT_QUEUE queue;
  IMPORT_DATA  *l;
  char         *import_filename;
  PENTITY      e;


  // check if this unit already exists in library, if yes -> done.

  if (unit_exists (library, filename))    // could occur for main.c if there's a main.h
    return 0;

  // store a first entry with null entity to avoid infinite recursion
  key = store_empty_unit (library, filename, /*is_body=*/ true);

  set_current_error_source_file (library, filename, key);


  if (load_in_lex (library, filename, must_exist, &utf16_source_text, &rc) != 0)
    return rc;


  import_and_compile_all_libraries (library, filename, key, &queue);


  set_current_error_source_file (library, filename, key);  // set again this file


  import_filename = malloc (260+1);   // on heap to avoid overloading the stack
  if (import_filename == null)
  {
    printf ("fatal error: cannot allocate import_filename\n");
    exit (-1);
  }



  // prepare the corresponding .h unit in scope

  sprintf (import_filename, "%.*s%s", strlen(filename)-2, filename, ".h");

  e = retrieve_unit (library, import_filename);
  if (e == null)
  {
    semantic_error ("interface unit of this file has circular dependencies", &token.pos);
  }
  else if (e == DUMMY_unit)
  {
    // no .h for main .c
  }
  else
  {
    int i;

    i = strlen(filename) - 2;
    while (i-1 >= 0 && filename[i-1] != '/' && filename[i-1] != '\\')
      i--;
    sprintf (import_filename, "%.*s", strlen(&filename[i])-2, &filename[i]);

    intern_append_new_unit_entity_for_import
                    (e,
                     import_filename,
                     &token.pos,
                     /* add_in_package_list = */ false);
  }


  free (import_filename);


  // prepare scope with all imported libraries being directly visible
  // using aliases if any;
  // if units were imported twice, make them visible only once.

  l = queue.head;
  while (l != null)
  {
    e = retrieve_unit (l->library, l->source);
    if (e == null)
    {
      semantic_error ("imported unit has circular dependencies", &l->src_pos);
    }
    else if (e == DUMMY_unit)   // unit imported non-existing .h of main program
    {
      semantic_error ("unit interface file does not exist", &l->src_pos);
    }
    else
    {
      e->info->the_unit_interface.unit_is_used = false;

      intern_append_new_unit_entity_for_import
                      (e,
                       l->alias,
                       &l->alia_pos,
                       /* add_in_package_list = */ true);
    }

    l = l->next;
  }

  parse_unit (library, filename, /*is_body=*/ true);

  generate_warnings_for_unused_units (&queue);

  free_import_list (&queue);


  lex_close_context ();
  free (utf16_source_text);

  return 0;
}

//=================================================================================

// compile main .c
// returns 0 if OK, +1 if file not found

int compile_main_unit (string filename)    // ends with ".c"
{
  char *import_filename;
  int  rc;

  import_filename = malloc (260+1);   // on heap to avoid overloading the stack
  if (import_filename == null)
  {
    printf ("fatal error: cannot allocate import_filename\n");
    exit (-1);
  }

  sprintf (import_filename, "%.*s%s", strlen(filename)-2, filename, ".h");
  compile_unit_interface ("", import_filename, /*must_exist=*/ false);

  free (import_filename);

  rc = compile_unit_body ("", filename, /*must_exist=*/ true);
  if (rc)
  {
    printf ("error: cannot open '%s'\n", filename);
    return rc;
  }

  // compile pending bodies

  while (pending_bodies != NULL)
  {
    PENDING_BODY *item;

    item = pending_bodies;
    pending_bodies = pending_bodies->next;

    rc = compile_unit_body (item->library, item->source, item->must_exist);
    if (rc)
    {
      if (item->library[0] == '\0')
        printf ("error: cannot open '%s'\n", item->source);
      else
        printf ("error: cannot open '%s' [%s]\n", item->source, item->library);
      return rc;
    }

    free (item->library);
    free (item->source);
    free (item);
  }

  return 0;
}

//=================================================================================
