
// backup.c

use ../set, ../db, ../thread, ../files;
use config, dbstruct, transact, dbport;

/**********************************************************************/
#begin unsafe
/**********************************************************************/

void free_backup_info (ref BACKUP_INFO^ bi)
{
  destroy_shared_object (ref bi^.so);

  if (bi^.output_fd != -1)
    __db_close (bi^.output_fd);

  SET_close (ref bi^.interval_to_backup);

  free bi^.page0;
  free bi^.buffers;

  clear bi^;
  free bi;
  bi = null;
}

//---------------------------------------------------------------------------------------------------

public int intern_db_backup_database (int db_handle, string output_filename)
{
  int          fd_in, fd_out;
  int          rc;
  DB_INFO*     p;
  long         nb_blocks;
  BACKUP_INFO^ bi;
  bool         done;

  ENTER ();

  rc = check_db_handle (db_handle, out p);
  if (rc < 0)
  {
    LEAVE ();
    return rc;
  }

  if (p->inside_explicit_transaction ||  // not allowed inside a transaction
      p->pbackup != null)                // backup already busy
  {
    LEAVE ();
    return E_NESTED_TRANS;
  }

  nb_blocks = p->initial_free_space_link;

  fd_out = __db_create_exclusive (output_filename);
  if (fd_out < 0)
  {
    LEAVE ();
    return fd_out;
  }

  LEAVE ();


  // this can take long, especially on slow backup device
  rc = __db_extend_file (fd_out, nb_blocks);
  if (rc < 0)
  {
    __db_close (fd_out);
    return rc;
  }


  ENTER ();

  rc = check_db_handle (db_handle, out p);
  if (rc < 0)
  {
    __db_close (fd_out);
    LEAVE ();
    return rc;
  }

  if (p->inside_explicit_transaction ||  // not allowed inside a transaction
      p->pbackup != null)
  {
    __db_close (fd_out);
    LEAVE ();
    return E_NESTED_TRANS;
  }

  nb_blocks = p->initial_free_space_link;

  rc = __db_extend_file (fd_out, nb_blocks);
  if (rc < 0)
  {
    __db_close (fd_out);
    LEAVE ();
    return rc;
  }

  {
    string(MAX_FILENAME_LENGTH) filename;
    assert get_full_filename (handle => p->db, out filename) == 0;
    fd_in = __db_open_share (filename);
    if (fd_in < 0)
    {
      __db_close (fd_out);
      LEAVE ();
      return fd_in;
    }
  }

  // prepare backup info

  bi = new BACKUP_INFO;
  bi^.output_fd = fd_out;
  SET_create (out bi^.interval_to_backup);
  SET_insert_interval (ref bi^.interval_to_backup, first => 1, last => nb_blocks-1);
  bi^.page0   = new byte[DB_BLOCK_SIZE];
  bi^.buffers = new byte[DB_BLOCK_SIZE];

  rc = __db_read (p->db, 0, out bi^.page0^);
  if (rc < 0)
  {
    __db_close (fd_in);
    free_backup_info (ref bi);
    LEAVE ();
    return rc;
  }


  // we need to copy all table locking blocks using the db descriptor (because a new handle can't read them)

  {
    int i;

    for (i=0; i<MAX_OPEN_TABLES; i++)
    {
      if (p->table[i] == null)
        continue;

      backup_block (ref bi^,
                        p->db,
                        p->table[i]^.locking_info.offset >> DB_BLOCK_SIZE_SHIFTS);
    }
  }


  p->pbackup = bi;

  LEAVE ();


  done = false;
  while (!done && bi^.error == 0)
  {
    long low, high, ofs;

    enter_shared_object (ref bi^.so);

    if (SET_search_any_interval (bi^.interval_to_backup, low => 0, out low, out high))
    {
      ofs = low;

      for (;;)
      {
        // backup [ofs]
        rc = __db_read (fd_in, ofs, out bi^.buffers^);
        if (rc < 0)
        {
          bi^.error = rc;
          break;
        }
        else
        {
          rc = __db_write (fd_out, ofs, bi^.buffers^);
          if (rc < 0)
          {
            bi^.error = rc;
            break;
          }
        }

        if (bi^.waiting)
          break;

        if (ofs == high)
          break;

        ofs++;
      }

      assert SET_delete_interval (ref bi^.interval_to_backup, low, ofs);
    }
    else
    {
      done = true;
    }

    leave_shared_object (ref bi^.so);

    if (bi^.waiting)    // some transaction is waiting to write blocks,
      sleep 0.01;       // give it some undisturbed time slice
  }



  ENTER ();
  p->pbackup = null;   // must be protected in critical section, we cannot free this suddenly during a transaction write
  LEAVE ();


  if (bi^.error < 0)
  {
    __db_close (fd_in);
    rc = bi^.error;
    free_backup_info (ref bi);
    return rc;
  }

  rc = __db_write (fd_out, 0, bi^.page0^);
  if (rc < 0)
  {
    __db_close (fd_in);
    free_backup_info (ref bi);
    return rc;
  }

  rc = __db_flush_file (fd_out);
  if (rc < 0)
  {
    __db_close (fd_in);
    free_backup_info (ref bi);
    return rc;
  }

  rc = __db_close (fd_in);
  if (rc < 0)
  {
    free_backup_info (ref bi);
    return rc;
  }

  rc = __db_close (fd_out);
  if (rc < 0)
  {
    free_backup_info (ref bi);
    return rc;
  }

  bi^.output_fd = -1;
  free_backup_info (ref bi);

  return 0;
}

//---------------------------------------------------------------------------------------------------

// called before writing in database file when backup_info is not null,
// to save the corresponding block before overwriting it.

public void backup_block (ref BACKUP_INFO bi, int db, long block_nr)
{
  int rc;

  if (bi.error < 0)  // previous error occured
    return;

  bi.waiting = true;        // signals the backup we need to write a block NOW !
  enter_shared_object (ref bi.so);
  bi.waiting = false;

  if (SET_item_found (bi.interval_to_backup, item => block_nr))
  {
    rc = __db_read (db, block_nr, out bi.buffers^);
    if (rc < 0)
    {
      bi.error = rc;
    }
    else
    {
      rc = __db_write (bi.output_fd, block_nr, bi.buffers^);
      if (rc < 0)
      {
        bi.error = rc;
      }
    }

    assert SET_delete_item (ref bi.interval_to_backup, block_nr);
  }

  leave_shared_object (ref bi.so);
}

/**********************************************************************/
#end unsafe
/**********************************************************************/
