
// directx.c

use arithm, image, linear_algebra, math, strings, thread, tracing;
use win/windows, win/directx11, win/dxshaders, directxdef;

#begin unsafe

//-----------------------------------------------------------------------------------------------

uint          g_back_buffer_x_res;
uint          g_back_buffer_y_res;
int           g_new_width;
int           g_new_height;
float         g_background_color[4] = {0.0, 0.0, 0.0, 1.0};
const float   g_float_initial_picking[4] = {1.0, 1.0, 1.0, 1.0};
float         g_near_z = 0.0625;  // 6 cm
float         g_far_z  = 1024.0;  // 1024 m
volatile bool g_screen_was_resized;
XMMATRIX      g_default_view_matrix;          // used to compute picking ray
vector3       g_default_camera_eye;
XMMATRIX      g_default_projection_matrix;    // 3D to 2D screen
XMMATRIX      g_default_ortho_matrix;         // 3D to 2D screen
XMMATRIX      g_default_shadow_matrix;
bool          g_shader_supports_indexing;
volatile bool g_projection_matrix_needs_update;
vector4       g_frustum[6];
SIZE_T        g_card_dedicated_ram;           // card dedicated ram memory in bytes
SIZE_T        g_card_max_additional_cpu_ram;  // max additional memory from cpu that card can use
bool          g_shader_constants_stable_dirty;
BLENDING      g_blending;
TEXTURE_ID    g_1_pixel_texture;
bool          g_wait_for_sync;
int           g_depth_stencil_buffer_mode;

int           g_next_frame_shadows_level;
int           g_shadows_level;   // 0 = off, 1 to 4 = number of shadow maps
bool          g_shadow_failure;
int           g_nb_of_instances;  // 1 for normal draw, 1 to 4 for shadows

//-----------------------------------------------------------------------------------------------

const float SHADOW_MAP_SIZE = 1024.0;  // 4 MB x number of maps

const byte INITIAL_STENCIL_VALUE = 0;
const byte COMPARE_STENCIL_VALUE = 0;

//-----------------------------------------------------------------------------------------------

packed struct ShaderConstantsStable  // max allowed size : 16K, must be a multiple of 16 !
{
  XMMATRIX   TransposedView;         // 64 bytes, camera position, rotation (changes if camera moves - once per frame)
  XMMATRIX   TransposedProjection;   // 64 bytes, 3D to 2D screen (changes if screen is resized or z view changed - rare)
  XMMATRIX   ShadowTransposedTransform;   // 64 bytes

  vector3    eyePosition;   // used for specular lightning (changes if camera moves - once per frame)
  float      fogInvRange;   // 1.0/(end - start) [0 if no fog]

  vector3    fogColor;
  float      fogStart;

  vector3    CameraOffset;
  float      FogInvExponent;

  float[2]   filler1;
  int        nb_of_shadow_maps;
  int        shader_supports_indexing;     // 0 = amd slow vertex shader, 1 = indexes supported so fast vertex shader

  vector3    AmbiantLightColor;
  float      filler2;

  vector3    SunLightColor;
  float      filler3;

  vector3    SunLightDirection;
  int        NbLights;

  LIGHT      Light[MAX_LIGHTS];  // 48 x 128 = 6144 bytes
}

//-----------------------------------------------------------------------------------------------

packed struct ShaderConstantsDynamic  // max allowed size : 16K, must be a multiple of 16 !
{
  matrix44   TransposedWorld;     // object position, rotation (changes for each object) (64 bytes)

  float[4]   MaterialColor;       // (16 bytes)

  vector3    EmissiveColor;
  float      specular_cone;       // (0..128) size of specular spot (0 = none, 5 = nice one)

  OBJECT_UV_TRANSFORM uv_transform;  // 24 bytes
  float      tclock;        // current time in secs, used for perlin noise
  float      ShinyFactor;

  int        ambiantlight_on;
  int        sunlight_on;
  bool       ambiant_downwards;  // 1 = ambiant 10% more intense when normal shines from above
  bool       has_bump_texture;
  byte       filler1;
  byte       filler2;
  float      parallax_factor;    // 0 = none, ex: -0.05

  float[4]   objectid;            // unique object id set by user for picking (on 3 bytes only)
}

const ShaderConstantsDynamic DEFAULT_DYNAMIC_CONSTANT =
   {TransposedWorld   => M44_ONE,
    MaterialColor     => {1.0, 1.0, 1.0, 1.0},
    EmissiveColor     => {0.0, 0.0, 0.0},
    specular_cone     => 0.0,
    uv_transform      => {0.0, 0.0, 0.0, 0.0, 0.0, 0.0},
    tclock            => 0.0,
    ShinyFactor       => 0.0,
    ambiantlight_on   => 0,
    sunlight_on       => 0,
    ambiant_downwards => false,
    has_bump_texture  => false,
    filler1           => 0,
    filler2           => 0,
    parallax_factor   => 0.0,
    objectid          => {0.0, 0.0, 0.0, 0.0}};

//-----------------------------------------------------------------------------------------------

ShaderConstantsStable   g_shader_constants_stable;
ShaderConstantsDynamic  g_shader_constants_dynamic;

//-----------------------------------------------------------------------------------------------

struct VERTEX_SHADER
{
  ID3D11VertexShader* m_vertexShader;
  ID3D11InputLayout*  m_layout;
}

struct GEOMETRY_SHADER
{
  ID3D11GeometryShader*  m_geometryShader;
}

struct PIXEL_SHADER
{
  ID3D11PixelShader*  m_pixelShader;
}

//-----------------------------------------------------------------------------------------------

const int SHADER_PASS_RENDERING        = 0;   // this will be the final rendering pass
const int SHADER_PASS_BUILD_SHADOW_MAP = 1;
// more preparation passes could appear here
const int MAX_SHADERS_PASSES_SHIFTS = 1;
const int MAX_SHADER_PASSES = 1 << MAX_SHADERS_PASSES_SHIFTS;    // = 2
const int MAX_SHADER_PASS_MASK = MAX_SHADER_PASSES - 1;

int g_shader_pass = SHADER_PASS_RENDERING;  // one of above

//-----------------------------------------------------------------------------------------------

const int VERTEX_SHADER_NORMAL = 0;
const int VERTEX_SHADER_RIGGED = 1;
const int MAX_VERTEX_SHADERS = 2;
int g_current_vertex_shader;
int g_next_vertex_shader;
VERTEX_SHADER  g_vertex_shader[MAX_SHADER_PASSES * MAX_VERTEX_SHADERS];  // low index = SHADER_PASS_xxx, high index = VERTEX_xxx

//-----------------------------------------------------------------------------------------------

const int GEOMETRY_SHADER_DEFAULT = 0;
const int MAX_GEOMETRY_SHADERS = 1;
int g_current_geometry_shader;
int g_next_geometry_shader;
GEOMETRY_SHADER g_geometry_shader[MAX_SHADER_PASSES * MAX_GEOMETRY_SHADERS];

//-----------------------------------------------------------------------------------------------

const int PIXEL_SHADER_SOLID = 0;
const int MAX_PIXEL_SHADERS = 1;
int g_current_pixel_shader;
int g_next_pixel_shader;
PIXEL_SHADER  g_pixel_shader[MAX_SHADER_PASSES * MAX_PIXEL_SHADERS];

//-----------------------------------------------------------------------------------------------

XMMATRIX g_identityMatrix;

//-----------------------------------------------------------------------------------------------

enum STATE {STOPPED, RUNNING};
STATE g_state;
bool g_stop_rendering;

CALLBACK_FUNCTIONS_3D g_callback_func;

//---------------------------------------------------------------------
void build_view_frustum (    matrix44 m_view,
                             matrix44 m_projection,
                         out vector4  m_frustum[6]);  // 6 plane a, b, c, d
//---------------------------------------------------------------------

// nearest and farthest z-points to draw

public void set_near_far_Z (float near_z = 0.0625,  // 6 cm
                            float far_z = 1024.00)  // 1024 m
{
  g_near_z = near_z;
  g_far_z  = far_z;
  g_projection_matrix_needs_update = true;
}

//---------------------------------------------------------------------

void compute_screen_size ()
{
  RECT rect;

  GetClientRect (main_hWnd, &rect);
  if (rect.bottom == 0)  // avoid division by zero
    rect.bottom = 1;
  if (rect.right == 0)
    rect.right = 1;

  g_back_buffer_x_res = (uint)rect.right;
  g_back_buffer_y_res = (uint)rect.bottom;
}

//---------------------------------------------------------------------

int create_vertex_shader (    ID3D11Device*        device,
                          byte[]                   blob,
                          D3D11_INPUT_ELEMENT_DESC Layout[],
                      out VERTEX_SHADER            shader)
{
  HRESULT result;

  clear shader;

  // Create the vertex shader from the buffer.
  result = device->lpVtbl->CreateVertexShader (device, &blob, blob'size, null, &shader.m_vertexShader);
  if (result < 0)
  {
    log_error ("CreateVertexShader()", result);
    return -1;
  }

  // Create the vertex input layout.
  result = device->lpVtbl->CreateInputLayout (device, &Layout, (uint)Layout'length, &blob, blob'size, &shader.m_layout);
  if (result < 0)
  {
    log_error ("CreateInputLayout()", result);
    return -1;
  }

  return 0;
}

//---------------------------------------------------------------------

int create_geometry_shader (    ID3D11Device*   device,
                                byte[]          blob,
                            out GEOMETRY_SHADER shader)
{
  HRESULT result;

  clear shader;

  // Create the pixel shader from the buffer.
  result = device->lpVtbl->CreateGeometryShader (device, &blob, blob'size, null, &shader.m_geometryShader);
  if (result < 0)
  {
    log_error ("CreateGeometryShader()", result);
    return -1;
  }

  return 0;
}

//---------------------------------------------------------------------

int create_pixel_shader (    ID3D11Device*  device,
                             byte[]         blob,
                         out PIXEL_SHADER   shader)
{
  HRESULT result;

  clear shader;

  // Create the pixel shader from the buffer.
  result = device->lpVtbl->CreatePixelShader (device, &blob, blob'size, null, &shader.m_pixelShader);
  if (result < 0)
  {
    log_error ("CreatePixelShader()", result);
    return -1;
  }

  return 0;
}

//---------------------------------------------------------------------

void select_vs_constant_buffers (ID3D11Buffer* buffers[])
{
  DX.m_deviceContext->lpVtbl->VSSetConstantBuffers (DX.m_deviceContext, StartSlot => 0, NumBuffers => (UINT)buffers'length, (ID3D11Buffer**)&buffers);
}

//---------------------------------------------------------------------

void select_new_vertex_shader (BONE_BUFFER_ID bone_id = {{0,0,0}})
{
  if (g_next_vertex_shader != g_current_vertex_shader)
  {
    ref VERTEX_SHADER vs = g_vertex_shader[g_next_vertex_shader];

    // set vertex input layout & vertex shader.
    DX.m_deviceContext->lpVtbl->IASetInputLayout (DX.m_deviceContext, vs.m_layout);
    DX.m_deviceContext->lpVtbl->VSSetShader      (DX.m_deviceContext, vs.m_vertexShader, null, 0);

    // set constant buffers.
    if ((g_next_vertex_shader >> MAX_SHADERS_PASSES_SHIFTS) == VERTEX_SHADER_NORMAL)
      select_vs_constant_buffers ({DX.m_constantBufferStable, DX.m_constantBufferDynamic});
    else
      select_vs_constant_buffers ({DX.m_constantBufferStable, DX.m_constantBufferDynamic,
                                   *(ID3D11Buffer**)&bone_id.id[0], *(ID3D11Buffer**)&bone_id.id[1], *(ID3D11Buffer**)&bone_id.id[2]});

    g_current_vertex_shader = g_next_vertex_shader;
  }
  else   // same shader as previous call
  {
    if ((g_next_vertex_shader >> MAX_SHADERS_PASSES_SHIFTS) == VERTEX_SHADER_RIGGED)
    {
      // always set buffers because bone_id changes at each call.
      select_vs_constant_buffers ({DX.m_constantBufferStable, DX.m_constantBufferDynamic,
                                   *(ID3D11Buffer**)&bone_id.id[0], *(ID3D11Buffer**)&bone_id.id[1], *(ID3D11Buffer**)&bone_id.id[2]});
    }
  }
}

//---------------------------------------------------------------------

void select_ps_constant_buffers (ID3D11Buffer* buffers[])
{
  DX.m_deviceContext->lpVtbl->PSSetConstantBuffers (DX.m_deviceContext, StartSlot => 0, NumBuffers => (UINT)buffers'length, (ID3D11Buffer**)&buffers);
}

//---------------------------------------------------------------------

void select_new_geometry_shader ()
{
  if (g_next_geometry_shader != g_current_geometry_shader)
  {
    // Set the pixel shader (can be null to disable it).
    DX.m_deviceContext->lpVtbl->GSSetShader (DX.m_deviceContext, g_geometry_shader[g_next_geometry_shader].m_geometryShader, null, 0);

    g_current_geometry_shader = g_next_geometry_shader;
  }
}

//---------------------------------------------------------------------

void select_new_pixel_shader ()
{
  if (g_next_pixel_shader != g_current_pixel_shader)
  {
    // Set the pixel shader (can be null to disable it).
    DX.m_deviceContext->lpVtbl->PSSetShader (DX.m_deviceContext, g_pixel_shader[g_next_pixel_shader].m_pixelShader, null, 0);

    // set the constant buffers in the pixel shader
    select_ps_constant_buffers ({DX.m_constantBufferStable, DX.m_constantBufferDynamic});

    // Set the sampler state in the pixel shader.
    DX.m_deviceContext->lpVtbl->PSSetSamplers (DX.m_deviceContext, 0, (UINT)DX.m_sampleState'length, &DX.m_sampleState);

    g_current_pixel_shader = g_next_pixel_shader;
  }
}

//---------------------------------------------------------------------

void close_vertex_shader (ref VERTEX_SHADER shader)
{
  if (shader.m_layout != null)
    shader.m_layout->lpVtbl->Release((LPVOID)shader.m_layout);

  // Release the vertex shader.
  if (shader.m_vertexShader != null)
    shader.m_vertexShader->lpVtbl->Release((LPVOID)shader.m_vertexShader);

  clear shader;
}

//---------------------------------------------------------------------

void close_geometry_shader (ref GEOMETRY_SHADER shader)
{
  // Release the pixel shader.
  if (shader.m_geometryShader != null)
    shader.m_geometryShader->lpVtbl->Release((LPVOID)shader.m_geometryShader);

  clear shader;
}

//---------------------------------------------------------------------

void close_pixel_shader (ref PIXEL_SHADER shader)
{
  // Release the pixel shader.
  if (shader.m_pixelShader != null)
    shader.m_pixelShader->lpVtbl->Release((LPVOID)shader.m_pixelShader);

  clear shader;
}

//---------------------------------------------------------------------

// Setup the constant buffers that are used in the shaders.

int create_constant_buffers (byte[] constant_data_stable, byte[] constant_data_dynamic)
{
  HRESULT           result;
  D3D11_BUFFER_DESC BufferDesc;

  clear BufferDesc;
  BufferDesc.Usage = D3D11_USAGE_DYNAMIC;    // cpu write, GPU read
  BufferDesc.ByteWidth = constant_data_stable'size;
  BufferDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
  BufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

  result = DX.m_device->lpVtbl->CreateBuffer (DX.m_device, &BufferDesc, pInitialData => null, &DX.m_constantBufferStable);
  if (result < 0)
  {
    log_error ("CreateBuffer(1)", result);
    return -1;
  }

  BufferDesc.ByteWidth = constant_data_dynamic'size;

  result = DX.m_device->lpVtbl->CreateBuffer (DX.m_device, &BufferDesc, pInitialData => null, &DX.m_constantBufferDynamic);
  if (result < 0)
  {
    log_error ("CreateBuffer(2)", result);
    return -1;
  }

  return 0;
}

//---------------------------------------------------------------------

void write_constant_buffer_stable (byte[] constant_data)
{
  HRESULT                  result;
  D3D11_MAPPED_SUBRESOURCE mappedResource;

  // Lock the constant buffer so it can be written to.
  result = DX.m_deviceContext->lpVtbl->Map (DX.m_deviceContext, (ID3D11Resource*)DX.m_constantBufferStable, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
  if (result < 0)
    fatal_abort ("Map()", result);

  // copy the new values into the constant buffer.
  mappedResource.pData[0 : constant_data'size] = constant_data;

  // Unlock the constant buffer.
  DX.m_deviceContext->lpVtbl->Unmap (DX.m_deviceContext, (ID3D11Resource*)DX.m_constantBufferStable, 0);
}

//---------------------------------------------------------------------

void write_constant_buffer_dynamic (byte[] constant_data)
{
  HRESULT                  result;
  D3D11_MAPPED_SUBRESOURCE mappedResource;

  // Lock the constant buffer so it can be written to.
  result = DX.m_deviceContext->lpVtbl->Map (DX.m_deviceContext, (ID3D11Resource*)DX.m_constantBufferDynamic, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
  if (result < 0)
    fatal_abort ("Map()", result);

  // copy the new values into the constant buffer.
  mappedResource.pData[0 : constant_data'size] = constant_data;

  // Unlock the constant buffer.
  DX.m_deviceContext->lpVtbl->Unmap (DX.m_deviceContext, (ID3D11Resource*)DX.m_constantBufferDynamic, 0);
}

//---------------------------------------------------------------------

void get_card_information (out UINT pnumerator, out UINT pdenominator, out SIZE_T pcard_dedicated_ram, out SIZE_T pcard_max_additional_cpu_ram)
{
  UINT           screenWidth, screenHeight;
  HRESULT        result;
  IDXGIFactory*  factory;
  IDXGIAdapter*  adapter;
  IDXGIOutput*   adapterOutput;
  uint           numModes, i;

  // default values
  pnumerator = 60;
  pdenominator = 1;
  pcard_dedicated_ram = 64;
  pcard_max_additional_cpu_ram = 0;

  // get resolution of desktop
  screenWidth  = (UINT)GetSystemMetrics(SM_CXSCREEN);
  screenHeight = (UINT)GetSystemMetrics(SM_CYSCREEN);

  // Create a DirectX graphics interface factory.

  {
    const string       dll_name  = "DXGI.dll\0";
    const string       func_name = "CreateDXGIFactory\0";
    HMODULE            hinstLib;
    LPVOID             func;
    FCreateDXGIFactory CreateDXGIFactory;

    hinstLib  = LoadLibraryA (&dll_name);
    if (hinstLib != 0)
    {
      func = GetProcAddress (hinstLib, &func_name);
      *(LPVOID*)&CreateDXGIFactory = *(LPVOID*)&func;

      if (CreateDXGIFactory != null)
      {
        result = CreateDXGIFactory (IID_IDXGIFactory, &factory);
        if (result >= 0)
        {
          // Use the factory to create an adapter for the primary graphics interface (video card).
          result = factory->lpVtbl->EnumAdapters (factory, 0, &adapter);
          if (result >= 0)
          {
            // Enumerate the primary adapter output (monitor).
            result = adapter->lpVtbl->EnumOutputs (adapter, 0, &adapterOutput);
            if (result >= 0)
            {
              // Get the number of modes that fit the DXGI_FORMAT_R8G8B8A8_UNORM display format for the adapter output (monitor).
              result = adapterOutput->lpVtbl->GetDisplayModeList (adapterOutput, DXGI_FORMAT_R8G8B8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, null);
              if (result >= 0)
              {
                // Create a list to hold all the possible display modes for this monitor/video card combination.
                DXGI_MODE_DESC[]^ displayModeList = new DXGI_MODE_DESC[numModes];

                // Now fill the display mode list structures.
                result = adapterOutput->lpVtbl->GetDisplayModeList(adapterOutput, DXGI_FORMAT_R8G8B8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, &displayModeList^);
                if (result >= 0)
                {
                  // Now go through all the display modes and find the one that matches the screen width and height.
                  // When a match is found store the numerator and denominator of the refresh rate for that monitor.
                  for (i=0; i<numModes; i++)
                  {
                    ref DXGI_MODE_DESC d = displayModeList^[i];
                    if (d.Width == screenWidth)
                    {
                      if (d.Height == screenHeight)
                      {
                        pnumerator   = d.RefreshRate.Numerator;
                        pdenominator = d.RefreshRate.Denominator;
                      }
                    }
                  }

                  // Get the adapter (video card) description.
                  {
                    DXGI_ADAPTER_DESC adapterDesc;

                    result = adapter->lpVtbl->GetDesc (adapter, &adapterDesc);
                    if (result >= 0)
                    {
                      pcard_dedicated_ram = adapterDesc.DedicatedVideoMemory;
                      pcard_max_additional_cpu_ram = adapterDesc.SharedSystemMemory;
                    }
                  }
                }

                // Release the display mode list.
                free displayModeList;
              }

              // Release the adapter output.
              adapterOutput->lpVtbl->Release((LPVOID)adapterOutput);
            }

            // Release the adapter.
            adapter->lpVtbl->Release((LPVOID)adapter);
          }

          // Release the factory.
          factory->lpVtbl->Release((LPVOID)factory);
        }
      }
    }
  }
}

//---------------------------------------------------------------------

void intern_set_depth_stencil_state ()
{
  DX.m_deviceContext->lpVtbl->OMSetDepthStencilState
        (DX.m_deviceContext,
         DX.m_depthStencilStateDepth[g_depth_stencil_buffer_mode],
         StencilRef => COMPARE_STENCIL_VALUE);  // Reference value to perform against when doing a depth-stencil test.
}

//---------------------------------------------------------------------

public void set_depth_buffer (bool testing, bool writing = true)    // default: on/on for render_3D, off/on for render_2D.
{
  g_depth_stencil_buffer_mode = (g_depth_stencil_buffer_mode & (255-1-2)) + (int)testing + 2*(int)writing;
  intern_set_depth_stencil_state ();
}

//---------------------------------------------------------------------

public void set_stencil_buffer (bool testing, bool writing = false)    // default: off/off for render_3D, off/off for render_2D.
{
  g_depth_stencil_buffer_mode = (g_depth_stencil_buffer_mode & (255-4-8)) + 4*(int)testing + 8*(int)writing;
  intern_set_depth_stencil_state ();
}

//---------------------------------------------------------------------

public void set_blending (BLENDING blending)
{
  float blendFactor[4];

  if (g_blending == blending)
    return;

  clear blendFactor; /*  blendFactor[0] = 0.0f; blendFactor[1] = 0.0f; blendFactor[2] = 0.0f; blendFactor[3] = 0.0f; */

  DX.m_deviceContext->lpVtbl->OMSetBlendState (DX.m_deviceContext,
                                               DX.m_BlendingState[(int)blending],
                                               blendFactor,
                                               0xffffffff);
  g_blending = blending;
}

//---------------------------------------------------------------------

void Make2DMatrix (out XMMATRIX m)
{
  clear m;
  m.m[0][0] = 2.0;
  m.m[1][1] = 2.0;
  m.m[3][3] = 1.0;
}

//---------------------------------------------------------------------

int create_all_shaders ()
{
  int p;

  for (p=0; p<MAX_SHADER_PASSES; p++)
  {
    {
      ////////////////////////////////////////////////////////////////////////////////
      // basic vertex shader
      ////////////////////////////////////////////////////////////////////////////////

      // Create the vertex input layout description.
      // This setup needs to match the structure VERTEX and the one in the Shader.

      const string POS = "POSITION\0";
      const string NOR = "NORMAL\0";
      const string TAG = "TANGENT\0";
      const string COL = "COLOR\0";
      const string TEX = "TEXCOORD\0";

      D3D11_INPUT_ELEMENT_DESC Layout[5];

      int rc;

      clear Layout;
      Layout[0].SemanticName = &POS;
      // Layout[0].SemanticIndex = 0;
      Layout[0].Format = DXGI_FORMAT_R32G32B32_FLOAT;
      // Layout[0].InputSlot = 0;
      // Layout[0].AlignedByteOffset = 0;
      Layout[0].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[0].InstanceDataStepRate = 0;

      Layout[1].SemanticName = &NOR;
      // Layout[1].SemanticIndex = 0;
      Layout[1].Format = DXGI_FORMAT_R32G32B32_FLOAT;
      // Layout[1].InputSlot = 0;
      Layout[1].AlignedByteOffset = 12;
      Layout[1].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[1].InstanceDataStepRate = 0;

      Layout[2].SemanticName = &TAG;
      // Layout[1].SemanticIndex = 0;
      Layout[2].Format = DXGI_FORMAT_R32G32B32_FLOAT;
      // Layout[2].InputSlot = 0;
      Layout[2].AlignedByteOffset = 24;
      Layout[2].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[2].InstanceDataStepRate = 0;

      Layout[3].SemanticName = &COL;
      // Layout[3].SemanticIndex = 0;
      Layout[3].Format = DXGI_FORMAT_R8G8B8A8_UNORM;
      // Layout[3].InputSlot = 0;
      Layout[3].AlignedByteOffset = 36;
      Layout[3].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[2].InstanceDataStepRate = 0;

      Layout[4].SemanticName = &TEX;
      // Layout[4].SemanticIndex = 0;
      Layout[4].Format = DXGI_FORMAT_R32G32_FLOAT;
      // Layout[4].InputSlot = 0;
      Layout[4].AlignedByteOffset = 40;
      Layout[4].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[4].InstanceDataStepRate = 0;

      rc = create_vertex_shader (DX.m_device, vertex_shader_basic[p], Layout, out g_vertex_shader[p + MAX_SHADER_PASSES * VERTEX_SHADER_NORMAL]);
      if (rc != 0)
      {
        log_error ("create_vertex_shader(1)", rc);
        return -1;
      }
    }


    {
      ////////////////////////////////////////////////////////////////////////////////
      // vertex shader for rigged mesh
      ////////////////////////////////////////////////////////////////////////////////

      // Create the vertex input layout description.
      // This setup needs to match the structure VERTEX and the one in the Shader.

      const string POS = "POSITION\0";
      const string NOR = "NORMAL\0";
      const string TAG = "TANGENT\0";
      const string TEX = "TEXCOORD\0";
      const string BONE = "BONE\0";
      const string WEIGHT = "WEIGHT\0";

      D3D11_INPUT_ELEMENT_DESC Layout[5+4];

      int  rc;
      uint i, idx, ofs;

      clear Layout;
      Layout[0].SemanticName = &POS;
      // Layout[0].SemanticIndex = 0;
      Layout[0].Format = DXGI_FORMAT_R32G32B32_FLOAT;
      // Layout[0].InputSlot = 0;
      // Layout[0].AlignedByteOffset = 0;
      Layout[0].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[0].InstanceDataStepRate = 0;

      Layout[1].SemanticName = &NOR;
      // Layout[1].SemanticIndex = 0;
      Layout[1].Format = DXGI_FORMAT_R32G32B32_FLOAT;
      // Layout[1].InputSlot = 0;
      Layout[1].AlignedByteOffset = 12;
      Layout[1].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[1].InstanceDataStepRate = 0;

      Layout[2].SemanticName = &TAG;
      // Layout[2].SemanticIndex = 0;
      Layout[2].Format = DXGI_FORMAT_R32G32B32_FLOAT;
      // Layout[2].InputSlot = 0;
      Layout[2].AlignedByteOffset = 24;
      Layout[2].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[2].InstanceDataStepRate = 0;

      Layout[3].SemanticName = &TEX;
      // Layout[3].SemanticIndex = 0;
      Layout[3].Format = DXGI_FORMAT_R32G32_FLOAT;
      // Layout[2].InputSlot = 0;
      Layout[3].AlignedByteOffset = 36;
      Layout[3].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[3].InstanceDataStepRate = 0;

      idx = 4;
      ofs = 36 + 8;
      Layout[idx].SemanticName = &BONE;
      Layout[idx].SemanticIndex = 0;
      Layout[idx].Format = DXGI_FORMAT_R32_UINT;
      // Layout[idx].InputSlot = 0;
      Layout[idx].AlignedByteOffset = ofs;
      Layout[idx].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
      // Layout[idx].InstanceDataStepRate = 0;
      idx++;
      ofs+=4;

      for (i=0; i<4; i++)
      {
        Layout[idx].SemanticName = &WEIGHT;
        Layout[idx].SemanticIndex = i;
        Layout[idx].Format = DXGI_FORMAT_R32_FLOAT;
        // Layout[idx].InputSlot = 0;
        Layout[idx].AlignedByteOffset = ofs;
        Layout[idx].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
        // Layout[idx].InstanceDataStepRate = 0;
        idx++;
        ofs+=4;
      }

      rc = create_vertex_shader (DX.m_device, vertex_shader_rigged[p], Layout, out g_vertex_shader[p + MAX_SHADER_PASSES * VERTEX_SHADER_RIGGED]);
      if (rc != 0)
      {
        log_error ("create_vertex_shader(2)", rc);
        return -1;
      }
    }

    {
      ////////////////////////////////////////////////////////////////////////////////
      // geometry shaders
      ////////////////////////////////////////////////////////////////////////////////

      int rc;

      if (geometry_shader[p]'length > 0)
      {
        rc = create_geometry_shader (DX.m_device, geometry_shader[p], out g_geometry_shader[p + MAX_SHADER_PASSES * GEOMETRY_SHADER_DEFAULT]);
        if (rc != 0)
        {
          log_error ("create_geometry_shader()", rc);
          return -1;
        }
      }
    }

    {
      ////////////////////////////////////////////////////////////////////////////////
      // pixel shaders
      ////////////////////////////////////////////////////////////////////////////////

      int rc;

      if (pixel_shader_solid[p]'length > 0)
      {
        rc = create_pixel_shader (DX.m_device, pixel_shader_solid[p], out g_pixel_shader[p + MAX_SHADER_PASSES * PIXEL_SHADER_SOLID]);
        if (rc != 0)
        {
          log_error ("create_pixel_shader(1)", rc);
          return -1;
        }
      }
    }
  }

  return 0;
}

//---------------------------------------------------------------------

void free_shadow_resources ()
{
  if (DX.m_Shadow_depth_View != null)
  {
    DX.m_Shadow_depth_View->lpVtbl->Release((LPVOID)DX.m_Shadow_depth_View);
    DX.m_Shadow_depth_View = null;
  }

  if (DX.m_shadowTextureView != null)
  {
    DX.m_shadowTextureView->lpVtbl->Release((LPVOID)DX.m_shadowTextureView);
    DX.m_shadowTextureView = null;
  }

  if (DX.m_Shadow_depth_Buffer != null)
  {
    DX.m_Shadow_depth_Buffer->lpVtbl->Release((LPVOID)DX.m_Shadow_depth_Buffer);
    DX.m_Shadow_depth_Buffer = null;
  }

  DX.m_DefaultTextures[2] = *(ID3D11ShaderResourceView**)&g_1_pixel_texture;
}

//---------------------------------------------------------------------

int allocate_shadow_ressources (bool use_float32)
{
  HRESULT result;

  {
    D3D11_TEXTURE2D_DESC  shadowDepthBufferDesc;

    // Set up the description of the depth buffer.
    clear shadowDepthBufferDesc;
    shadowDepthBufferDesc.Width = (UINT)SHADOW_MAP_SIZE;
    shadowDepthBufferDesc.Height = (UINT)SHADOW_MAP_SIZE;
    shadowDepthBufferDesc.MipLevels = 1;
    shadowDepthBufferDesc.ArraySize = (UINT)g_shadows_level;

    // requires 'D' or TYPELESS format
    shadowDepthBufferDesc.Format = use_float32 ? DXGI_FORMAT_R32_TYPELESS : DXGI_FORMAT_R24G8_TYPELESS;

    shadowDepthBufferDesc.SampleDesc.Count = 1;  // multisamples per pixel
    // shadowDepthBufferDesc.SampleDesc.Quality = 0;
    shadowDepthBufferDesc.Usage = D3D11_USAGE_DEFAULT;   // inaccessible to the CPU
    shadowDepthBufferDesc.BindFlags = D3D11_BIND_DEPTH_STENCIL | D3D11_BIND_SHADER_RESOURCE;
    // shadowDepthBufferDesc.CPUAccessFlags = 0;
    // shadowDepthBufferDesc.MiscFlags = 0;

    // Create the texture for the depth buffer using the filled out description.
    result = DX.m_device->lpVtbl->CreateTexture2D (DX.m_device, &shadowDepthBufferDesc, null, &DX.m_Shadow_depth_Buffer);
    if (result < 0)
    {
      log_error ("CreateTexture2D(Shadow)", result);
      free_shadow_resources ();
      return -1;
    }
  }


  {
    D3D11_DEPTH_STENCIL_VIEW_DESC ShadowDepthStencilViewDesc;

    // Set up the depth stencil view description.
    clear ShadowDepthStencilViewDesc;

    // NOT ALLOWED : any format with TYPELESS in the name.
    ShadowDepthStencilViewDesc.Format = use_float32 ? DXGI_FORMAT_D32_FLOAT : DXGI_FORMAT_D24_UNORM_S8_UINT;

    ShadowDepthStencilViewDesc.ViewDimension = D3D11_DSV_DIMENSION_TEXTURE2DARRAY;
    // ShadowDepthStencilViewDesc.d3d11_texunion.Texture2DArray.MipSlice = 0;
    ShadowDepthStencilViewDesc.d3d11_texunion.Texture2DArray.FirstArraySlice = 0;
    ShadowDepthStencilViewDesc.d3d11_texunion.Texture2DArray.ArraySize = (UINT)g_shadows_level;

    // Create the depth stencil view.
    result = DX.m_device->lpVtbl->CreateDepthStencilView (DX.m_device, (ID3D11Resource*)DX.m_Shadow_depth_Buffer, &ShadowDepthStencilViewDesc, &DX.m_Shadow_depth_View);
    if (result < 0)
    {
      log_error ("CreateDepthStencilView(shadow)", result);
      free_shadow_resources ();
      return -1;
    }
  }

  {
    D3D11_SHADER_RESOURCE_VIEW_DESC srvDesc;

    clear srvDesc;
    srvDesc.Format = use_float32 ? DXGI_FORMAT_R32_FLOAT : DXGI_FORMAT_R24_UNORM_X8_TYPELESS;
    srvDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2DARRAY;
    // srvDesc.array.Texture2DArray.MostDetailedMip = 0;
    // srvDesc.array.Texture2DArray.MostDetailedMip = 0;
    srvDesc.array.Texture2DArray.MipLevels = 1;
    // srvDesc.array.Texture2DArray.FirstArraySlice = 0;
    srvDesc.array.Texture2DArray.ArraySize = (UINT)g_shadows_level;

    result = DX.m_device->lpVtbl->CreateShaderResourceView (DX.m_device, *(ID3D11Resource**)&DX.m_Shadow_depth_Buffer, &srvDesc, &DX.m_shadowTextureView);
    if (result < 0)
    {
      log_error ("CreateShaderResourceView(shadow)", result);
      free_shadow_resources ();
      return -1;
    }
  }

  DX.m_DefaultTextures[2] = DX.m_shadowTextureView;

  return 0;
}

//---------------------------------------------------------------------

int create_state ()
{
  HRESULT result;

  {
    ID3D11Resource*  backBufferPtr;

    // Get the pointer to the back buffer.
    result = DX.m_swapChain->lpVtbl->GetBuffer (DX.m_swapChain, 0, IID_ID3D11Texture2D, (LPVOID*)&backBufferPtr);
    if (result < 0)
    {
      log_error ("GetBuffer()", result);
      return -1;
    }

    // Create the render target view with the back buffer pointer.
    result = DX.m_device->lpVtbl->CreateRenderTargetView (DX.m_device, backBufferPtr, null, &DX.m_renderTargetView[0]);
    if (result < 0)
    {
      log_error ("CreateRenderTargetView()", result);
      return -1;
    }

    // Release pointer to the back buffer as we no longer need it.
    backBufferPtr->lpVtbl->Release((LPVOID)backBufferPtr);
  }


  // picking buffer

  {
    D3D11_TEXTURE2D_DESC  pickingBufferDesc;

    // Set up the description of the picking buffer.
    clear pickingBufferDesc;
    pickingBufferDesc.Width = g_back_buffer_x_res;
    pickingBufferDesc.Height = g_back_buffer_y_res;
    pickingBufferDesc.MipLevels = 1;
    pickingBufferDesc.ArraySize = 1;
    pickingBufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;   // a 32-bit format
    pickingBufferDesc.SampleDesc.Count = 1;
    // pickingBufferDesc.SampleDesc.Quality = 0;
    pickingBufferDesc.Usage = D3D11_USAGE_DEFAULT;   // inaccessible to the CPU
    pickingBufferDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_RENDER_TARGET;
    // pickingBufferDesc.CPUAccessFlags = 0;
    // pickingBufferDesc.MiscFlags = 0;

    // create the texture for the picking buffer
    result = DX.m_device->lpVtbl->CreateTexture2D (DX.m_device, &pickingBufferDesc, null, &DX.m_pickingBuffer);
    if (result < 0)
    {
      log_error ("CreateTexture2D(2)", result);
      return -1;
    }

    {
      result = DX.m_device->lpVtbl->CreateRenderTargetView (DX.m_device, (ID3D11Resource *)DX.m_pickingBuffer, null, &DX.m_renderTargetView[1]);
      if (result < 0)
      {
        log_error ("CreateRenderTargetView()", result);
        return -1;
      }
    }


    // picking result output, allocated on CPU

    {
      D3D11_TEXTURE2D_DESC  pickingOutputBufferDesc;

      // Set up the description of the picking buffer.
      clear pickingOutputBufferDesc;
      pickingOutputBufferDesc.Width = g_back_buffer_x_res;
      pickingOutputBufferDesc.Height = g_back_buffer_y_res;
      pickingOutputBufferDesc.MipLevels = 1;
      pickingOutputBufferDesc.ArraySize = 1;
      pickingOutputBufferDesc.Format = pickingBufferDesc.Format;
      pickingOutputBufferDesc.SampleDesc.Count = 1;
      // pickingOutputBufferDesc.SampleDesc.Quality = 0;
      pickingOutputBufferDesc.Usage = D3D11_USAGE_STAGING;   // used only for copy gpu->cpu
      // pickingOutputBufferDesc.BindFlags = 0;
      pickingOutputBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_READ;
      // pickingOutputBufferDesc.MiscFlags = 0;

      // create the texture for the picking buffer output
      result = DX.m_device->lpVtbl->CreateTexture2D (DX.m_device, &pickingOutputBufferDesc, null, &DX.m_pickingOutputBuffer);
      if (result < 0)
      {
        log_error ("CreateTexture2D(3)", result);
        return -1;
      }
    }
  }


  {
    int trial;

    for (trial = 1; trial <= 2; trial++)
    {
      {
        D3D11_TEXTURE2D_DESC  depthBufferDesc;

        // Set up the description of the depth buffer.
        clear depthBufferDesc;
        depthBufferDesc.Width = g_back_buffer_x_res;
        depthBufferDesc.Height = g_back_buffer_y_res;
        depthBufferDesc.MipLevels = 1;
        depthBufferDesc.ArraySize = 1;

        // requires 'D' or TYPELESS format
        depthBufferDesc.Format = (trial == 1) ? DXGI_FORMAT_R32_TYPELESS : DXGI_FORMAT_R24G8_TYPELESS;

        depthBufferDesc.SampleDesc.Count = 1;
        // depthBufferDesc.SampleDesc.Quality = 0;
        depthBufferDesc.Usage = D3D11_USAGE_DEFAULT;   // inaccessible to the CPU
        depthBufferDesc.BindFlags = D3D11_BIND_DEPTH_STENCIL;
        // depthBufferDesc.CPUAccessFlags = 0;
        // depthBufferDesc.MiscFlags = 0;

        // Create the texture for the depth buffer using the filled out description.
        result = DX.m_device->lpVtbl->CreateTexture2D (DX.m_device, &depthBufferDesc, null, &DX.m_depth_Buffer);
        if (result < 0)
        {
          log_error ("CreateTexture2D(A)", result);

          if (trial == 1)
            continue;
          else
            return -1;
        }
      }


      {
        D3D11_DEPTH_STENCIL_VIEW_DESC depthStencilViewDesc;

        // Set up the depth stencil view description.
        clear depthStencilViewDesc;

        // NOT ALLOWED : any format with TYPELESS in the name.
        depthStencilViewDesc.Format = (trial == 1) ? DXGI_FORMAT_D32_FLOAT : DXGI_FORMAT_D24_UNORM_S8_UINT;

        depthStencilViewDesc.ViewDimension = D3D11_DSV_DIMENSION_TEXTURE2D;
        // depthStencilViewDesc.d3d11_texunion.Texture2D.MipSlice = 0;

        // Create the depth stencil view.
        result = DX.m_device->lpVtbl->CreateDepthStencilView (DX.m_device, (ID3D11Resource*)DX.m_depth_Buffer, &depthStencilViewDesc, &DX.m_depth_View);
        if (result < 0)
        {
          log_error ("CreateDepthStencilView(A)", result);

          if (DX.m_depth_Buffer != null)
          {
            DX.m_depth_Buffer->lpVtbl->Release((LPVOID)DX.m_depth_Buffer);
            DX.m_depth_Buffer = null;
          }

          if (trial == 1)
            continue;
          else
            return -1;
        }
      }

      break;
    }
  }


  // create stencil buffer

  {
    D3D11_TEXTURE2D_DESC  stencilBufferDesc;

    // Set up the description of the depth buffer.
    clear stencilBufferDesc;
    stencilBufferDesc.Width = g_back_buffer_x_res;
    stencilBufferDesc.Height = g_back_buffer_y_res;
    stencilBufferDesc.MipLevels = 1;
    stencilBufferDesc.ArraySize = 1;
    stencilBufferDesc.Format = DXGI_FORMAT_R24G8_TYPELESS;    // requires 'D' or TYPELESS format
    stencilBufferDesc.SampleDesc.Count = 1;
    // stencilBufferDesc.SampleDesc.Quality = 0;
    stencilBufferDesc.Usage = D3D11_USAGE_DEFAULT;   // inaccessible to the CPU
    stencilBufferDesc.BindFlags = D3D11_BIND_DEPTH_STENCIL;
    // stencilBufferDesc.CPUAccessFlags = 0;
    // stencilBufferDesc.MiscFlags = 0;

    // Create the texture for the depth buffer using the filled out description.
    result = DX.m_device->lpVtbl->CreateTexture2D (DX.m_device, &stencilBufferDesc, null, &DX.m_stencil_Buffer);
    if (result < 0)
    {
      log_error ("CreateTexture2D(B)", result);
      return -1;
    }
  }


  {
    D3D11_DEPTH_STENCIL_VIEW_DESC depthStencilViewDesc;

    // Set up the depth stencil view description.
    clear depthStencilViewDesc;
    depthStencilViewDesc.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;    // NOT ALLOWED : any format with TYPELESS in the name.
    depthStencilViewDesc.ViewDimension = D3D11_DSV_DIMENSION_TEXTURE2D;
    // depthStencilViewDesc.d3d11_texunion.Texture2D.MipSlice = 0;

    // Create the stencil view.
    result = DX.m_device->lpVtbl->CreateDepthStencilView (DX.m_device, (ID3D11Resource*)DX.m_stencil_Buffer, &depthStencilViewDesc, &DX.m_stencil_View);
    if (result < 0)
    {
      log_error ("CreateDepthStencilView(B)", result);
      return -1;
    }
  }


  {
    int i;

    for (i=0; i<16; i+=4)       // create 16 depthstencilstates
    {
      bool stencil_test  = (i & 4) != 0;
      bool stencil_write = (i & 8) != 0;

      D3D11_DEPTH_STENCIL_DESC depthStencilDesc;

      // Set up the description of the stencil state.
      clear depthStencilDesc;

      // stencil buffer settings

      depthStencilDesc.StencilEnable    = i>=4 ? TRUE : FALSE;     // Enable stencil
      depthStencilDesc.StencilReadMask  = 0x000000FF;   // 8 bits for stencil
      depthStencilDesc.StencilWriteMask = 0x000000FF;   // 8 bits for stencil

      // Stencil operations if pixel is front-facing.

      if (stencil_test)
        depthStencilDesc.FrontFace.StencilFunc      = D3D11_COMPARISON_EQUAL;   // test if equal stencil reference value
      else
        depthStencilDesc.FrontFace.StencilFunc      = D3D11_COMPARISON_ALWAYS;  // stencil always passes

      // operation to perform when stencil testing and depth testing both pass.
      if (stencil_write)
        depthStencilDesc.FrontFace.StencilPassOp    = D3D11_STENCIL_OP_INCR_SAT;  // increment til 255
      else
        depthStencilDesc.FrontFace.StencilPassOp    = D3D11_STENCIL_OP_KEEP;  // Keep the existing stencil data

      // operation to perform when stencil test fails.
      depthStencilDesc.FrontFace.StencilFailOp      = D3D11_STENCIL_OP_KEEP;  // Keep the existing stencil data

      // operation to perform when stencil test passes and depth test fails.
      depthStencilDesc.FrontFace.StencilDepthFailOp = D3D11_STENCIL_OP_KEEP;  // Keep the existing stencil data

      // Stencil operations if pixel is back-facing.
      depthStencilDesc.BackFace.StencilFunc        = D3D11_COMPARISON_NEVER;  // stencil never passes
      depthStencilDesc.BackFace.StencilFailOp      = D3D11_STENCIL_OP_KEEP;
      depthStencilDesc.BackFace.StencilDepthFailOp = D3D11_STENCIL_OP_KEEP;
      depthStencilDesc.BackFace.StencilPassOp      = D3D11_STENCIL_OP_KEEP;


      // depth buffer settings

      // D3D11_COMPARISON_NEVER = writes nothing, D3D11_COMPARISON_ALWAYS = write all
      depthStencilDesc.DepthFunc = D3D11_COMPARISON_GREATER; // If source z > destination z, we write the pixel.

      depthStencilDesc.DepthEnable = FALSE;              // disable depth testing.
      depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;  // Turn off writes to the depth-stencil buffer.

      result = DX.m_device->lpVtbl->CreateDepthStencilState (DX.m_device, &depthStencilDesc, &DX.m_depthStencilStateDepth[i+0]);
      if (result < 0)
      {
        log_error ("CreateDepthStencilState(1)", result);
        return -1;
      }

      depthStencilDesc.DepthEnable = TRUE;              // enable depth testing.
      depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;  // Turn off writes to the depth-stencil buffer.

      result = DX.m_device->lpVtbl->CreateDepthStencilState (DX.m_device, &depthStencilDesc, &DX.m_depthStencilStateDepth[i+1]);
      if (result < 0)
      {
        log_error ("CreateDepthStencilState(2)", result);
        return -1;
      }

      depthStencilDesc.DepthEnable = FALSE;              // disable depth testing.
      depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;  // Turn on writes to the depth-stencil buffer.

      result = DX.m_device->lpVtbl->CreateDepthStencilState (DX.m_device, &depthStencilDesc, &DX.m_depthStencilStateDepth[i+2]);
      if (result < 0)
      {
        log_error ("CreateDepthStencilState(3)", result);
        return -1;
      }

      depthStencilDesc.DepthEnable = TRUE;              // enable depth testing.
      depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;  // Turn on writes to the depth-stencil buffer.

      result = DX.m_device->lpVtbl->CreateDepthStencilState (DX.m_device, &depthStencilDesc, &DX.m_depthStencilStateDepth[i+3]);
      if (result < 0)
      {
        log_error ("CreateDepthStencilState(4)", result);
        return -1;
      }
    }
  }


  // state 16 : for shadow pass
  {
    D3D11_DEPTH_STENCIL_DESC depthStencilDesc;

    clear depthStencilDesc;
    depthStencilDesc.DepthFunc = D3D11_COMPARISON_LESS; // If source z < destination z, we write the pixel.
    depthStencilDesc.DepthEnable = TRUE;                // enable depth testing.
    depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;  // Turn on writes to the depth-stencil buffer.

    result = DX.m_device->lpVtbl->CreateDepthStencilState (DX.m_device, &depthStencilDesc, &DX.m_depthStencilStateDepth[16]);
    if (result < 0)
    {
      log_error ("CreateDepthStencilState(16)", result);
      return -1;
    }
  }

  {
    D3D11_RASTERIZER_DESC rasterDesc;

    // Setup the raster description which will determine how and what polygons will be drawn.
    clear rasterDesc;
    rasterDesc.FillMode = D3D11_FILL_SOLID;  // alternative is wireframe
    rasterDesc.CullMode = D3D11_CULL_BACK;   // Do not draw triangles that are back-facing (alternative: D3D11_CULL_NONE renders all triangles)
    // rasterDesc.FrontCounterClockwise = FALSE;
    // rasterDesc.DepthBias = 0;                // for displaying shadows properly
    // rasterDesc.DepthBiasClamp = 0.0f;
    // rasterDesc.SlopeScaledDepthBias = 0.0f;
    rasterDesc.DepthClipEnable = TRUE;       // clip z larger than back plane (can be disabled for some stencil operations)
    // rasterDesc.ScissorEnable = FALSE;        // can be used for scissor-rectangle culling
    // rasterDesc.MultisampleEnable = FALSE;
    // rasterDesc.AntialiasedLineEnable = FALSE;

    // create a rasterizer state
    result = DX.m_device->lpVtbl->CreateRasterizerState (DX.m_device, &rasterDesc, &DX.m_rasterState[0]);   // normal rendering
    if (result < 0)
    {
      log_error ("CreateRasterizerState()", result);
      return -1;
    }


    rasterDesc.CullMode = D3D11_CULL_FRONT;  // Do not draw triangles that are front-facing (near the camera)
//    rasterDesc.CullMode = D3D11_CULL_NONE;
//    rasterDesc.DepthClipEnable = FALSE;       // don't clip z larger than back plane
    rasterDesc.DepthBias = 260;                 // for displaying shadows properly
    // rasterDesc.DepthBiasClamp = 0.0f;
    rasterDesc.SlopeScaledDepthBias = 2.0;

    result = DX.m_device->lpVtbl->CreateRasterizerState (DX.m_device, &rasterDesc, &DX.m_rasterState[1]);    // for shadows
    if (result < 0)
    {
      log_error ("CreateRasterizerState()", result);
      return -1;
    }
  }


  {
    D3D11_BLEND_DESC  blendStateDescription;

    clear blendStateDescription;

    // Create an alpha enabled blend state description.
    // note: IndependentBlendEnable = TRUE; does not work on DirectX 10 : blending will be used for picking too.
    blendStateDescription.RenderTarget[0].BlendEnable = TRUE;
    blendStateDescription.RenderTarget[0].SrcBlend = D3D11_BLEND_SRC_ALPHA;
    blendStateDescription.RenderTarget[0].DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
    blendStateDescription.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
    blendStateDescription.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND_ONE;
    blendStateDescription.RenderTarget[0].DestBlendAlpha = D3D11_BLEND_ZERO;
    blendStateDescription.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP_ADD;
    blendStateDescription.RenderTarget[0].RenderTargetWriteMask = D3D11_COLOR_WRITE_ENABLE_ALL;

    // Create the blend state using the description.
    result = DX.m_device->lpVtbl->CreateBlendState (DX.m_device, &blendStateDescription, &DX.m_BlendingState[(int)BLENDING_TRANSLUCID]);
    if (result < 0)
    {
      log_error ("CreateBlendState(1)", result);
      return -1;
    }

    blendStateDescription.RenderTarget[0].SrcBlend = D3D11_BLEND_SRC_ALPHA;
    blendStateDescription.RenderTarget[0].DestBlend = D3D11_BLEND_DEST_ALPHA;
    blendStateDescription.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;

    // Create the blend state using the description.
    result = DX.m_device->lpVtbl->CreateBlendState (DX.m_device, &blendStateDescription, &DX.m_BlendingState[(int)BLENDING_ADD]);
    if (result < 0)
    {
      log_error ("CreateBlendState(2)", result);
      return -1;
    }

    blendStateDescription.RenderTarget[0].BlendOp = D3D11_BLEND_OP_MIN;

    // Create the blend state using the description.
    result = DX.m_device->lpVtbl->CreateBlendState (DX.m_device, &blendStateDescription, &DX.m_BlendingState[(int)BLENDING_MIN]);
    if (result < 0)
    {
      log_error ("CreateBlendState(3)", result);
      return -1;
    }

    blendStateDescription.RenderTarget[0].BlendOp = D3D11_BLEND_OP_MAX;

    // Create the blend state using the description.
    result = DX.m_device->lpVtbl->CreateBlendState (DX.m_device, &blendStateDescription, &DX.m_BlendingState[(int)BLENDING_MAX]);
    if (result < 0)
    {
      log_error ("CreateBlendState(4)", result);
      return -1;
    }

    // Modify the description to create an alpha disabled blend state description.
    blendStateDescription.RenderTarget[0].BlendEnable = FALSE;

    // Create the blend state using the description.
    result = DX.m_device->lpVtbl->CreateBlendState (DX.m_device, &blendStateDescription, &DX.m_BlendingState[(int)BLENDING_OFF]);
    if (result < 0)
    {
      log_error ("CreateBlendState(5)", result);
      return -1;
    }
  }

  g_blending = BLENDING_OFF;


  // Create a texture sampler state description.
  {
    D3D11_SAMPLER_DESC samplerDesc;

    // sampler to read from textures
    clear samplerDesc;
    samplerDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
    samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
    samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
    samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
    // samplerDesc.MipLODBias = 0.0f;
    samplerDesc.MaxAnisotropy = 1;
    samplerDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
    // samplerDesc.BorderColor[0] = 0.0;
    // samplerDesc.BorderColor[1] = 0.0;
    // samplerDesc.BorderColor[2] = 0.0;
    // samplerDesc.BorderColor[3] = 0.0;
    samplerDesc.MinLOD = -D3D11_FLOAT32_MAX;
    samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

if (false)
{
  samplerDesc.Filter = D3D11_FILTER_ANISOTROPIC;
	samplerDesc.MaxAnisotropy = 8;
	samplerDesc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
}

    // Create the texture sampler state.
    result = DX.m_device->lpVtbl->CreateSamplerState (DX.m_device, &samplerDesc, &DX.m_sampleState[0]);   // normal rendering
    if (result < 0)
    {
      log_error ("CreateSamplerState(0)", result);
      return -1;
    }


    // sampler to read from shadow maps

    // filter: D3D11_FILTER_COMPARISON_MIN_MAG_LINEAR_MIP_POINT
    //      or D3D11_FILTER_COMPARISON_MIN_MAG_MIP_LINEAR
    //      or D3D11_FILTER_COMPARISON_ANISOTROPIC
    samplerDesc.Filter = D3D11_FILTER_COMPARISON_ANISOTROPIC;

    samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_BORDER;
    samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_BORDER;
    samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_BORDER;
    // samplerDesc.MipLODBias = 0.0f;
    samplerDesc.MaxAnisotropy = 1;
    samplerDesc.ComparisonFunc = D3D11_COMPARISON_LESS;
    // samplerDesc.BorderColor[0] = 0.0;
    // samplerDesc.BorderColor[1] = 0.0;
    // samplerDesc.BorderColor[2] = 0.0;
    // samplerDesc.BorderColor[3] = 0.0;
    samplerDesc.MinLOD = 0.0;
    samplerDesc.MaxLOD = 12.0;

    // Create the texture sampler state.
    result = DX.m_device->lpVtbl->CreateSamplerState (DX.m_device, &samplerDesc, &DX.m_sampleState[1]);
    if (result < 0)
    {
      log_error ("CreateSamplerState(1)", result);
      return -1;
    }
  }

  // Set the type of primitive that should be rendered, in this case always triangles.
  DX.m_deviceContext->lpVtbl->IASetPrimitiveTopology (DX.m_deviceContext, D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);


  // set shaders

  g_current_vertex_shader = -1;
  g_current_geometry_shader = -1;
  g_current_pixel_shader = -1;


  g_projection_matrix_needs_update = true;

  return 0;
}

//---------------------------------------------------------------------

public void select_depth_stencil_buffer (int nr)  // 1 = Depth32, 2 = Depth24Stencil8
{
  // Bind the render target view (back buffer, picking buffer) and depth buffer to the output render pipeline.
  DX.m_deviceContext->lpVtbl->OMSetRenderTargets (DX.m_deviceContext, 2, &DX.m_renderTargetView, nr == 1 ? DX.m_depth_View : DX.m_stencil_View);
}

//---------------------------------------------------------------------

void free_state ()
{
  int i;

  for (i=0; i<DX.m_sampleState'length; i++)
  {
    if (DX.m_sampleState[i] != null)
    {
      DX.m_sampleState[i]->lpVtbl->Release((LPVOID)DX.m_sampleState[i]);
      DX.m_sampleState[i] = null;
    }
  }

  for (i=0; i<DX.m_rasterState'length; i++)
  {
    if (DX.m_rasterState[i] != null)
    {
      DX.m_rasterState[i]->lpVtbl->Release((LPVOID)DX.m_rasterState[i]);
      DX.m_rasterState[i] = null;
    }
  }

  if (DX.m_depth_View != null)
  {
    DX.m_depth_View->lpVtbl->Release((LPVOID)DX.m_depth_View);
    DX.m_depth_View = null;
  }

  if (DX.m_stencil_View != null)
  {
    DX.m_stencil_View->lpVtbl->Release((LPVOID)DX.m_stencil_View);
    DX.m_stencil_View = null;
  }

  for (i=0; i<DX.m_depthStencilStateDepth'length; i++)
  {
    if (DX.m_depthStencilStateDepth[i] != null)
    {
      DX.m_depthStencilStateDepth[i]->lpVtbl->Release((LPVOID)DX.m_depthStencilStateDepth[i]);
      DX.m_depthStencilStateDepth[i] = null;
    }
  }

  if (DX.m_pickingBuffer != null)
  {
    DX.m_pickingBuffer->lpVtbl->Release((LPVOID)DX.m_pickingBuffer);
    DX.m_pickingBuffer = null;
  }

  if (DX.m_pickingOutputBuffer != null)
  {
    DX.m_pickingOutputBuffer->lpVtbl->Release((LPVOID)DX.m_pickingOutputBuffer);
    DX.m_pickingOutputBuffer = null;
  }

  if (DX.m_depth_Buffer != null)
  {
    DX.m_depth_Buffer->lpVtbl->Release((LPVOID)DX.m_depth_Buffer);
    DX.m_depth_Buffer = null;
  }

  if (DX.m_stencil_Buffer != null)
  {
    DX.m_stencil_Buffer->lpVtbl->Release((LPVOID)DX.m_stencil_Buffer);
    DX.m_stencil_Buffer = null;
  }

  if (DX.m_renderTargetView[0] != null)
  {
    DX.m_renderTargetView[0]->lpVtbl->Release((LPVOID)DX.m_renderTargetView[0]);
    DX.m_renderTargetView[0] = null;
  }

  if (DX.m_renderTargetView[1] != null)
  {
    DX.m_renderTargetView[1]->lpVtbl->Release((LPVOID)DX.m_renderTargetView[1]);
    DX.m_renderTargetView[1] = null;
  }

  for (i=0; i<=(int)BLENDING'last; i++)
  {
    if (DX.m_BlendingState[i] != null)
    {
      DX.m_BlendingState[i]->lpVtbl->Release((LPVOID)DX.m_BlendingState[i]);
      DX.m_BlendingState[i] = null;
    }
  }
}

//---------------------------------------------------------------------

// this is the function that closes DirectX

void close_DirectX ()
{
  int i;

  // Before shutting down set to windowed mode or when you release the swap chain it will throw an exception.
  if (DX.m_swapChain != null)
    DX.m_swapChain->lpVtbl->SetFullscreenState (DX.m_swapChain, Fullscreen => FALSE, null);

  free_state ();
  free_shadow_resources ();

  if (DX.m_deviceContext != null)
    DX.m_deviceContext->lpVtbl->Release((LPVOID)DX.m_deviceContext);

  if (DX.m_constantBufferStable != null)
  {
    DX.m_constantBufferStable->lpVtbl->Release((LPVOID)DX.m_constantBufferStable);
    DX.m_constantBufferStable = null;
  }

  if (DX.m_constantBufferDynamic != null)
  {
    DX.m_constantBufferDynamic->lpVtbl->Release((LPVOID)DX.m_constantBufferDynamic);
    DX.m_constantBufferDynamic = null;
  }

  for (i=0; i<g_vertex_shader'length; i++)
    close_vertex_shader (ref g_vertex_shader[i]);
  for (i=0; i<g_geometry_shader'length; i++)
    close_geometry_shader (ref g_geometry_shader[i]);
  for (i=0; i<g_pixel_shader'length; i++)
    close_pixel_shader (ref g_pixel_shader[i]);

  free_texture (g_1_pixel_texture);

  if (DX.m_device != null)
    DX.m_device->lpVtbl->Release((LPVOID)DX.m_device);

  if (DX.m_swapChain != null)
    DX.m_swapChain->lpVtbl->Release((LPVOID)DX.m_swapChain);
}

//---------------------------------------------------------------------

// this function initializes and prepares Direct3D for use on main application window

int init_DirectX (bool debug, bool use_warp)
{
  HRESULT result;

  clear DX;

  compute_screen_size ();

  {
    DXGI_SWAP_CHAIN_DESC  swapChainDesc;
    D3D_FEATURE_LEVEL     featureLevel[7];
    D3D_FEATURE_LEVEL     actualfeaturelevel;
    uint                  fl;

    const string       dll_name  = "d3d11\0";
    const string       func_name = "D3D11CreateDeviceAndSwapChain\0";
    HMODULE            hinstLib;
    LPVOID             func;
    FD3D11CreateDeviceAndSwapChain D3D11CreateDeviceAndSwapChain;

    hinstLib = LoadLibraryA (&dll_name);
    if (hinstLib == 0)
    {
      log_error ("LoadLibraryA(d3d11)", 0);
      return -1;
    }
    
    func = GetProcAddress (hinstLib, &func_name);
    *(LPVOID*)&D3D11CreateDeviceAndSwapChain = *(LPVOID*)&func;
    if (D3D11CreateDeviceAndSwapChain == null)
    {
      log_error ("GetProcAddress(D3D11CreateDeviceAndSwapChain)", 0);
      return -1;
    }
    
    // Initialize the swap chain description.
    clear swapChainDesc;

    // Set to a single back buffer.
    swapChainDesc.BufferCount = 1;

    // Set the width and height of the back buffer.
    swapChainDesc.BufferDesc.Width = g_back_buffer_x_res;
    swapChainDesc.BufferDesc.Height = g_back_buffer_y_res;

    // Set regular 32-bit surface for the back buffer.
    swapChainDesc.BufferDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;

    // Set the refresh rate of the back buffer, and the memory of the card.
    get_card_information (out swapChainDesc.BufferDesc.RefreshRate.Numerator,
                          out swapChainDesc.BufferDesc.RefreshRate.Denominator,
                          out g_card_dedicated_ram,
                          out g_card_max_additional_cpu_ram);

    // Set the usage of the back buffer.
    swapChainDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;

    // Set the handle for the window to render to.
    swapChainDesc.OutputWindow = main_hWnd;

    // Turn multisampling off.
    swapChainDesc.SampleDesc.Count = 1;
    swapChainDesc.SampleDesc.Quality = 0;

    // Set to windowed mode.
    swapChainDesc.Windowed = 1;

    // Set the scan line ordering and scaling to unspecified.
    swapChainDesc.BufferDesc.ScanlineOrdering = DXGI_MODE_SCANLINE_ORDER_UNSPECIFIED;
    swapChainDesc.BufferDesc.Scaling = DXGI_MODE_SCALING_UNSPECIFIED;

    // Discard the back buffer contents after presenting.
    swapChainDesc.SwapEffect = DXGI_SWAP_EFFECT_DISCARD;

    // Don't set the advanced flags.
    swapChainDesc.Flags = 0;

    // Set the feature level
    featureLevel = {D3D_FEATURE_LEVEL_11_1,  // Win8, Win10
                    D3D_FEATURE_LEVEL_11_0,  // some Win7
                    D3D_FEATURE_LEVEL_10_1,  // some Win7
                    D3D_FEATURE_LEVEL_10_0,
                    D3D_FEATURE_LEVEL_9_3,
                    D3D_FEATURE_LEVEL_9_2,
                    D3D_FEATURE_LEVEL_9_1};

    result = -1;
    actualfeaturelevel = 0;
    for (fl=0; fl<(UINT)featureLevel'length; fl++)  // due to a Windows weird behaviour, we need to retry
    {
      // Create the swap chain, Direct3D device, and Direct3D device context.
      result = D3D11CreateDeviceAndSwapChain (null,
                                              use_warp ? D3D_DRIVER_TYPE_WARP : D3D_DRIVER_TYPE_HARDWARE,
                                              0,
                                              debug ? D3D11_CREATE_DEVICE_DEBUG : 0,
                                              &featureLevel[fl], (UINT)featureLevel'length - fl, D3D11_SDK_VERSION,
                                              &swapChainDesc, &DX.m_swapChain, &DX.m_device, &actualfeaturelevel, &DX.m_deviceContext);
      if (result != 0x80070057)  // not invalid arguments (0 or error)
        break;
    }

    if (result < 0)
    {
      log_error ("D3D11CreateDeviceAndSwapChain()", result);
      return -1;
    }

    trace ("info: DirectX feature level = %u.%u\n", actualfeaturelevel >> 12, (actualfeaturelevel >> 8) & 15);

    if (actualfeaturelevel < D3D_FEATURE_LEVEL_10_0)
    {
      log_error ("D3D11CreateDeviceAndSwapChain() - not enough features", -1);
      return -1;
    }
  }

  XMMatrixIdentity (&g_identityMatrix);
  Make2DMatrix (out g_default_ortho_matrix);

  // create constant buffers
  clear g_shader_constants_stable, g_shader_constants_dynamic;
  if (create_constant_buffers (g_shader_constants_stable, g_shader_constants_dynamic) < 0)
    return -1;

  // create shaders
  if (create_all_shaders () < 0)
    return -1;

  // create device context states
  if (create_state () < 0)
    return -1;

  // create g_1_pixel_texture
  {
    IMAGE_INFO img = {pixel => new byte[] ' {255, 255, 255, 255}, width => 1, height => 1};
    g_1_pixel_texture = create_texture (img);
    free img.pixel;
  }

  DX.m_DefaultTextures = {*(ID3D11ShaderResourceView**)&g_1_pixel_texture,
                          *(ID3D11ShaderResourceView**)&g_1_pixel_texture,
                          *(ID3D11ShaderResourceView**)&g_1_pixel_texture};

  assert (ShaderConstantsStable'size & 15) == 0;  // must be M16
  assert (ShaderConstantsDynamic'size & 15) == 0;  // must be M16

  return 0;
}

//---------------------------------------------------------------------

public void signal_DirectX_screen_was_resized (int width, int height)
{
  int w = max (1, width);
  int h = max (1, height);

  if (w == g_new_width && h == g_new_height)
    return;

  g_new_width = w;
  g_new_height = h;

  g_screen_was_resized = true;
}

//---------------------------------------------------------------------

void handle_screen_resize ()
{
  HRESULT result;

  if (DX.m_swapChain == null)
    return;

  result = DX.m_swapChain->lpVtbl->SetFullscreenState (DX.m_swapChain, Fullscreen => FALSE, null);
  if (result < 0)
    fatal_abort ("SetFullscreenState()", result);

  // reset any device context to the default settings : sets to NULL all resource slots, shaders, input layouts, predications,
  // scissor rectangles, depth-stencil state, rasterizer state, blend state, sampler state, and viewports.
  // the primitive topology is set to UNDEFINED.
  DX.m_deviceContext->lpVtbl->ClearState (DX.m_deviceContext);

  // Release all outstanding references to the swap chain's buffers.
  free_state ();

  // Preserve the existing buffer count and format.
  result = DX.m_swapChain->lpVtbl->ResizeBuffers (DX.m_swapChain, 0, Width=> g_back_buffer_x_res, Height=> g_back_buffer_y_res, 0, 0);
  if (result < 0)
    fatal_abort ("ResizeBuffers()", result);

  if (create_state () < 0)
    fatal_abort ("create_state()", -1);
}

//---------------------------------------------------------------------

public uint get_picking_object_id (int x, int y)
{
  HRESULT                  result;
  D3D11_MAPPED_SUBRESOURCE mappedResource;
  uint                     id;

  if ((uint)x >= g_back_buffer_x_res || (uint)y >= g_back_buffer_y_res)
    return 0xFFFFFF;

  // this call blocks until the above resource copy from previous frame was done
  result = DX.m_deviceContext->lpVtbl->Map (DX.m_deviceContext, (ID3D11Resource*)DX.m_pickingOutputBuffer, 0,
                                            D3D11_MAP_READ, 0, &mappedResource);
  if (result < 0)
    fatal_abort ("Map()", result);

  id = 0xFFFFFF & *(uint *)&mappedResource.pData[(x * 4) + (y * (int)mappedResource.RowPitch)];

  DX.m_deviceContext->lpVtbl->Unmap (DX.m_deviceContext, (ID3D11Resource*)DX.m_pickingOutputBuffer, 0);

  return id;
}

//---------------------------------------------------------------------

void user_rendering (FUNC prepare_frame, FUNC user_render_shadow, FUNC user_render_3D, FUNC user_render_2D)
{
  //------------------------------

  if (g_screen_was_resized)
  {
    g_screen_was_resized = false;

    g_back_buffer_x_res = (UINT)g_new_width;
    g_back_buffer_y_res = (UINT)g_new_height;

    handle_screen_resize ();
  }

  //------------------------------

  if (g_projection_matrix_needs_update)  // first run, or screen was resized, or far_z/near_z was changed.
  {
    g_projection_matrix_needs_update = false;

    // Build a left-handed perspective projection matrix based on a field of view.
    XMMatrixPerspectiveFovLH
       (&g_default_projection_matrix,
        45.0 * (float)(PI / 180.0),                                // the horizontal field of view
        (float)g_back_buffer_x_res / (float)g_back_buffer_y_res,   // aspect ratio
        zn => g_far_z,                                             // the far view-plane (REVERSE NEAR FAR !!)
        zf => g_near_z);                                           // the near view-plane (REVERSE NEAR FAR !!)

    g_world_to_screen_factor = (float)g_back_buffer_y_res * g_default_projection_matrix.m[1][1];
    build_view_frustum (*(matrix44*)&g_default_view_matrix, *(matrix44*)&g_default_projection_matrix, out g_frustum);
  }

  //------------------------------

  // clear the back buffers to background color.
  DX.m_deviceContext->lpVtbl->ClearRenderTargetView (DX.m_deviceContext, DX.m_renderTargetView[0], g_background_color);  // screen
  DX.m_deviceContext->lpVtbl->ClearRenderTargetView (DX.m_deviceContext, DX.m_renderTargetView[1], g_float_initial_picking); // picking buffer

  // clear the depth stencil buffers
  DX.m_deviceContext->lpVtbl->ClearDepthStencilView (DX.m_deviceContext, DX.m_depth_View, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 0.0f, INITIAL_STENCIL_VALUE);
  DX.m_deviceContext->lpVtbl->ClearDepthStencilView (DX.m_deviceContext, DX.m_stencil_View, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL, 0.0f, INITIAL_STENCIL_VALUE);

  //------------------------------

  prepare_frame ();   // set camera (generates g_default_view_matrix, g_default_camera_eye), create new & delete old objects, ..

  //------------------------------

  if (g_shadows_level != g_next_frame_shadows_level)
  {
    if (!g_shadow_failure)
    {
      free_shadow_resources ();
      g_shadows_level = g_next_frame_shadows_level;
    }
  }

  //------------------------------

  if (g_shadows_level > 0)
  {
    if (user_render_shadow != null)
    {
      if (DX.m_Shadow_depth_View == null)
      {
        // $ test also shaders with int24 depth buffer !
        if (allocate_shadow_ressources (use_float32 => true) < 0 && allocate_shadow_ressources (use_float32 => false) < 0)
        {
          g_shadows_level = 0;       // turn off shadows !
          g_shadow_failure = true;   // never try again to prevent reallocating shadow ressources at each frame.
        }
      }

      if (DX.m_Shadow_depth_View != null)
      {
        // clear the shadow depth buffer to 1 (NEAR PLANE)
        DX.m_deviceContext->lpVtbl->ClearDepthStencilView (DX.m_deviceContext, DX.m_Shadow_depth_View, D3D11_CLEAR_DEPTH, 1.0f, 0);

        // Bind the render target view (0 output buffers) and depth stencil buffer to the output render pipeline.
        DX.m_deviceContext->lpVtbl->OMSetRenderTargets (DX.m_deviceContext, 0, null, DX.m_Shadow_depth_View);

        clear g_shader_constants_stable;   // clear all constants, including sun, lights, fog, ..
        g_shader_constants_stable.eyePosition = g_default_camera_eye;   // used for specular light
        MXXMatrixTranspose (g_default_view_matrix,       out g_shader_constants_stable.TransposedView);
        MXXMatrixTranspose (g_default_projection_matrix, out g_shader_constants_stable.TransposedProjection);
        MXXMatrixTranspose (g_default_shadow_matrix,     out g_shader_constants_stable.ShadowTransposedTransform);
        g_shader_constants_stable.nb_of_shadow_maps        = g_shadows_level;
        g_shader_constants_stable.shader_supports_indexing = g_shader_supports_indexing ? 1 : 0;
        g_shader_constants_stable_dirty = true;
        g_shader_constants_dynamic = DEFAULT_DYNAMIC_CONSTANT;

        g_depth_stencil_buffer_mode = 16;   // special mode for writing most distant pixel
        intern_set_depth_stencil_state ();

        set_blending (BLENDING_OFF);

        g_shader_pass = SHADER_PASS_BUILD_SHADOW_MAP;
        g_next_geometry_shader = g_shader_pass + MAX_SHADER_PASSES * GEOMETRY_SHADER_DEFAULT;
        select_new_geometry_shader ();

        g_nb_of_instances = g_shadows_level;

        // set the rasterizer state
        DX.m_deviceContext->lpVtbl->RSSetState (DX.m_deviceContext, DX.m_rasterState[1]);  // 0 = normal rendering, 1 = for shadows

        // set white 1x1 texture
        DX.m_deviceContext->lpVtbl->PSSetShaderResources (DX.m_deviceContext, 0, 1, (ID3D11ShaderResourceView **)&g_1_pixel_texture);

        // set viewport
        {
          D3D11_VIEWPORT viewport;

          clear viewport;
          viewport.Width = SHADOW_MAP_SIZE;
          viewport.Height = SHADOW_MAP_SIZE;
          // viewport.MinDepth = 0.0f;
          viewport.MaxDepth = 1.0f;
          // viewport.TopLeftX = 0.0f;
          // viewport.TopLeftY = 0.0f;
          DX.m_deviceContext->lpVtbl->RSSetViewports (DX.m_deviceContext, 1, &viewport);
        }

        user_render_shadow ();
      }
    }
  }

  //------------------------------

  {
    clear g_shader_constants_stable;   // clear all constants, including sun, lights, fog, ..
    g_shader_constants_stable.eyePosition = g_default_camera_eye;   // used for specular light
    MXXMatrixTranspose (g_default_view_matrix, out g_shader_constants_stable.TransposedView);
    MXXMatrixTranspose (g_default_projection_matrix, out g_shader_constants_stable.TransposedProjection);
    if (g_shadows_level > 0)
      MXXMatrixTranspose (g_default_shadow_matrix, out g_shader_constants_stable.ShadowTransposedTransform);
    g_shader_constants_stable.nb_of_shadow_maps        = g_shadows_level;
    g_shader_constants_stable.shader_supports_indexing = g_shader_supports_indexing ? 1 : 0;
    g_shader_constants_stable_dirty = true;
    g_shader_constants_dynamic = DEFAULT_DYNAMIC_CONSTANT;

    g_depth_stencil_buffer_mode = 3;
    intern_set_depth_stencil_state ();

    set_blending (BLENDING_TRANSLUCID);

    g_shader_pass = SHADER_PASS_RENDERING;
    g_next_geometry_shader = g_shader_pass + MAX_SHADER_PASSES * GEOMETRY_SHADER_DEFAULT;
    select_new_geometry_shader ();

    g_nb_of_instances = 1;

    select_depth_stencil_buffer (nr => 1);  // 1 = Depth32, 2 = Depth24Stencil8

    // set the rasterizer state
    DX.m_deviceContext->lpVtbl->RSSetState (DX.m_deviceContext, DX.m_rasterState[0]);  // 0 = normal rendering, 1 = for shadows

    // set viewport
    {
      D3D11_VIEWPORT viewport;

      clear viewport;
      viewport.Width = (float)g_back_buffer_x_res;
      viewport.Height = (float)g_back_buffer_y_res;
      // viewport.MinDepth = 0.0f;
      viewport.MaxDepth = 1.0f;
      // viewport.TopLeftX = 0.0f;
      // viewport.TopLeftY = 0.0f;
      DX.m_deviceContext->lpVtbl->RSSetViewports (DX.m_deviceContext, 1, &viewport);
    }

    user_render_3D ();
  }

  //------------------------------

  if (user_render_2D != null)
  {
    clear g_shader_constants_stable;   // clear all constants, including sun, lights, fog, ..
    g_shader_constants_stable.TransposedView = g_identityMatrix;
    MXXMatrixTranspose (g_default_ortho_matrix, out g_shader_constants_stable.TransposedProjection);
    g_shader_constants_stable.shader_supports_indexing = g_shader_supports_indexing ? 1 : 0;
    g_shader_constants_stable_dirty = true;
    g_shader_constants_dynamic = DEFAULT_DYNAMIC_CONSTANT;

    g_depth_stencil_buffer_mode = 0;
    intern_set_depth_stencil_state ();

    set_blending (BLENDING_TRANSLUCID);

    g_shader_pass = SHADER_PASS_RENDERING;

    user_render_2D ();
  }
}

//---------------------------------------------------------------------

void rendering_loop ()
{
  for (;;)
  {
    g_frame_tick = ticks();

    user_rendering (g_callback_func.prepare_frame,
                    g_callback_func.render_shadow,
                    g_callback_func.render_3D,
                    g_callback_func.render_2D);

    // queue an asynchronous command to copy the picking buffer to the picking output buffer
    DX.m_deviceContext->lpVtbl->CopyResource (This         => DX.m_deviceContext,
                                              pDstResource => (ID3D11Resource *)DX.m_pickingOutputBuffer,
                                              pSrcResource => (ID3D11Resource *)DX.m_pickingBuffer);

    DX.m_swapChain->lpVtbl->Present (DX.m_swapChain, (uint)g_wait_for_sync, 0);

    if (g_stop_rendering)
      break;
  }

  g_callback_func.cleanup_objects();
  g_state = STOPPED;
}

//---------------------------------------------------------------------

public void limit_fps (bool on)   // default: on. (set to off if we don't want to wait for the monitor vertical sync)
{
  g_wait_for_sync = on;
}

//---------------------------------------------------------------------

public int start_DirectX_rendering_thread (CALLBACK_FUNCTIONS_3D func, bool debug = false, bool use_warp = false)
{
  int rc;

  if (g_state == RUNNING)
    return 0;

  g_callback_func = func;

  if (init_DirectX (debug, use_warp) < 0)
    return -1;

  g_stop_rendering = false;
  g_state          = RUNNING;
  g_frame_tick     = ticks();
  g_wait_for_sync  = true;

  g_callback_func.init_objects ();

  // start rendering loop
  rc = run rendering_loop ();
  assert rc == 0;

  return 0;
}

//---------------------------------------------------------------------

public void stop_DirectX_rendering_thread ()
{
  if (g_state == STOPPED)
    return;

  g_stop_rendering = true;
  while (g_state != STOPPED)
    sleep 0.25;

  close_DirectX ();
}

//---------------------------------------------------------------------

public VERTEX_ID create_vertex_buffer (VERTEX[] v, bool allow_update)
{
  D3D11_BUFFER_DESC      vertexBufferDesc;
  D3D11_SUBRESOURCE_DATA vertexData;
  HRESULT                result;
  ID3D11Buffer*          vertexBuffer;

  // Set up the description of the static vertex buffer.
  clear vertexBufferDesc;
  vertexBufferDesc.Usage = allow_update ? D3D11_USAGE_DYNAMIC : D3D11_USAGE_IMMUTABLE;
  vertexBufferDesc.ByteWidth = v'size;
  vertexBufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
  vertexBufferDesc.CPUAccessFlags = allow_update ? D3D11_CPU_ACCESS_WRITE : 0;
  // vertexBufferDesc.MiscFlags = 0;
  // vertexBufferDesc.StructureByteStride = 0;

  // Give the subresource structure a pointer to the vertex data.
  clear vertexData;
  vertexData.pSysMem = (byte*)&v[0];
  // vertexData.SysMemPitch = 0;
  // vertexData.SysMemSlicePitch = 0;

  // Now create the vertex buffer.
  result = DX.m_device->lpVtbl->CreateBuffer (DX.m_device, &vertexBufferDesc, &vertexData, &vertexBuffer);
  if (result < 0)
    fatal_abort ("CreateBuffer()", result);

  return *((VERTEX_ID *)&vertexBuffer);
}

//---------------------------------------------------------------------

public INDEX_ID create_large_index_buffer (uint4[] i, bool allow_update)
{
  D3D11_BUFFER_DESC      indexBufferDesc;
  D3D11_SUBRESOURCE_DATA indexData;
  HRESULT                result;
  ID3D11Buffer*          indexBuffer;

  // Set up the description of the static index buffer.
  clear indexBufferDesc;
  indexBufferDesc.Usage = allow_update ? D3D11_USAGE_DYNAMIC : D3D11_USAGE_IMMUTABLE;
  indexBufferDesc.ByteWidth = i'size;
  indexBufferDesc.BindFlags = D3D11_BIND_INDEX_BUFFER;
  indexBufferDesc.CPUAccessFlags = allow_update ? D3D11_CPU_ACCESS_WRITE : 0;
  // indexBufferDesc.MiscFlags = 0;
  // indexBufferDesc.StructureByteStride = 0;

  // Give the subresource structure a pointer to the index data.
  clear indexData;
  indexData.pSysMem = (byte*)&i[0];
  // indexData.SysMemPitch = 0;
  // indexData.SysMemSlicePitch = 0;

  // Now create the vertex buffer.
  result = DX.m_device->lpVtbl->CreateBuffer (DX.m_device, &indexBufferDesc, &indexData, &indexBuffer);
  if (result < 0)
    fatal_abort ("CreateBuffer()", result);

  return *((INDEX_ID *)&indexBuffer);
}

//---------------------------------------------------------------------

public INDEX_ID create_small_index_buffer (uint2[] i, bool allow_update)
{
  D3D11_BUFFER_DESC      indexBufferDesc;
  D3D11_SUBRESOURCE_DATA indexData;
  HRESULT                result;
  ID3D11Buffer*          indexBuffer;

  // Set up the description of the static index buffer.
  clear indexBufferDesc;
  indexBufferDesc.Usage = allow_update ? D3D11_USAGE_DYNAMIC : D3D11_USAGE_IMMUTABLE;
  indexBufferDesc.ByteWidth = i'size;
  indexBufferDesc.BindFlags = D3D11_BIND_INDEX_BUFFER;
  indexBufferDesc.CPUAccessFlags = allow_update ? D3D11_CPU_ACCESS_WRITE : 0;
  //  indexBufferDesc.MiscFlags = 0;
  //  indexBufferDesc.StructureByteStride = 0;

  // Give the subresource structure a pointer to the index data.
  clear indexData;
  indexData.pSysMem = (byte*)&i[0];
  // indexData.SysMemPitch = 0;
  // indexData.SysMemSlicePitch = 0;

  // Now create the vertex buffer.
  result = DX.m_device->lpVtbl->CreateBuffer (DX.m_device, &indexBufferDesc, &indexData, &indexBuffer);
  if (result < 0)
    fatal_abort ("CreateBuffer()", result);

  return 1 + *((INDEX_ID *)&indexBuffer);    // add 1 to make address odd
}

//---------------------------------------------------------------------

public void update_vertex_buffer (VERTEX_ID v, int offset, VERTEX[] vb, BUFFER_UPDATE_FLAG flag = OVERWRITE_ALL)
{
  D3D11_MAP                f;
  ID3D11Resource*          pResource;
  D3D11_MAPPED_SUBRESOURCE mappedResource;

  switch (flag)
  {
    case OVERWRITE_ALL:
      f = D3D11_MAP_WRITE_DISCARD;
      break;
    case OVERWRITE_PART:
      f = D3D11_MAP_WRITE;
      break;
    case APPEND_DATA:
      f = D3D11_MAP_WRITE_NO_OVERWRITE;
      break;
    default:
      abort;
  }

  pResource'byte = v'byte;

  DX.m_deviceContext->lpVtbl->Map (DX.m_deviceContext, pResource, 0, f, 0, &mappedResource);
  mappedResource.pData[(uint)offset * VERTEX'size : vb'size] = vb'byte;
  DX.m_deviceContext->lpVtbl->Unmap (DX.m_deviceContext, pResource, 0);
}

//---------------------------------------------------------------------

public void update_small_index_buffer (INDEX_ID id, uint offset, uint2[] ib, BUFFER_UPDATE_FLAG flag = OVERWRITE_ALL)
{
  INDEX_ID                 id2 = id;
  D3D11_MAP                f;
  ID3D11Resource*          pResource;
  D3D11_MAPPED_SUBRESOURCE mappedResource;

  switch (flag)
  {
    case OVERWRITE_ALL:
      f = D3D11_MAP_WRITE_DISCARD;
      break;
    case OVERWRITE_PART:
      f = D3D11_MAP_WRITE;
      break;
    case APPEND_DATA:
      f = D3D11_MAP_WRITE_NO_OVERWRITE;
      break;
    default:
      abort;
  }

  assert (id2 & 1) == 1;
  id2--;

  pResource'byte = id2'byte;

  DX.m_deviceContext->lpVtbl->Map (DX.m_deviceContext, pResource, 0, f, 0, &mappedResource);
  mappedResource.pData[(uint)offset * uint2'size : ib'size] = ib'byte;
  DX.m_deviceContext->lpVtbl->Unmap (DX.m_deviceContext, pResource, 0);
}

//---------------------------------------------------------------------

public void update_large_index_buffer (INDEX_ID id, uint offset, uint4[] ib, BUFFER_UPDATE_FLAG flag = OVERWRITE_ALL)
{
  D3D11_MAP                f;
  ID3D11Resource*          pResource;
  D3D11_MAPPED_SUBRESOURCE mappedResource;

  switch (flag)
  {
    case OVERWRITE_ALL:
      f = D3D11_MAP_WRITE_DISCARD;
      break;
    case OVERWRITE_PART:
      f = D3D11_MAP_WRITE;
      break;
    case APPEND_DATA:
      f = D3D11_MAP_WRITE_NO_OVERWRITE;
      break;
    default:
      abort;
  }

  assert (id & 1) == 0;
  pResource'byte = id'byte;

  DX.m_deviceContext->lpVtbl->Map (DX.m_deviceContext, pResource, 0, f, 0, &mappedResource);
  mappedResource.pData[(uint)offset * uint4'size : ib'size] = ib'byte;
  DX.m_deviceContext->lpVtbl->Unmap (DX.m_deviceContext, pResource, 0);
}

//---------------------------------------------------------------------

void flush_constant_buffers ()
{
  if (g_shader_constants_stable_dirty)
  {
    write_constant_buffer_stable (g_shader_constants_stable'byte
           [0 : g_shader_constants_stable'size
                - g_shader_constants_stable.Light'size
                + (uint)g_shader_constants_stable.NbLights * LIGHT'size]);
    g_shader_constants_stable_dirty = false;
  }

  write_constant_buffer_dynamic (g_shader_constants_dynamic);
}

//---------------------------------------------------------------------

void set_pixel_shader_ressources (TEXTURE_ID t_id[2],   // color texture and bump texture
                                  uint       object_id)
{
  if (g_shader_pass == SHADER_PASS_RENDERING)   // actual rendering pass (after shadows)
  {
    // set texture
    {
      ID3D11ShaderResourceView* t[3] = DX.m_DefaultTextures;
      bool                      has_bump;

      if (t_id[0] != 0)
        t[0] = *(ID3D11ShaderResourceView **)&t_id[0];

      has_bump = (t_id[1] != 0);
      g_shader_constants_dynamic.has_bump_texture = has_bump;
      if (has_bump)
        t[1] = *(ID3D11ShaderResourceView **)&t_id[1];

      DX.m_deviceContext->lpVtbl->PSSetShaderResources
         (DX.m_deviceContext,
          0,
          (UINT)DX.m_DefaultTextures'length,   // 3 textures for final rendering  with shadows on
          &t);
    }

    if (object_id >= 0xFFFFFF)   // means no object_id provided
    {
      clear g_shader_constants_dynamic.objectid;   // zero alpha means no object_id provided
    }
    else
    {
      g_shader_constants_dynamic.objectid = {(float)(object_id & 255) * (1.0 / 255.0),
                                             (float)((object_id >> 8) & 255) * (1.0 / 255.0),
                                             (float)((object_id >> 16) & 255) * (1.0 / 255.0),
                                             1.0};    // 1.0 means means valid object_id (will pass blending without change)
    }
  }
}

//---------------------------------------------------------------------

public
void draw_indexed_triangles (VERTEX_ID  v_id,
                             int        v_offset,
                             INDEX_ID   i_id,
                             int        i_offset,
                             int        i_count,
                             TEXTURE_ID t_id[2],  // color and bump textures
                             uint       object_id = 0xFFFFFF)
{
  INDEX_ID i_id2 = i_id;
  uint     stride, offset;

  i_id2 &= -2;    // set even

  g_next_vertex_shader = g_shader_pass + MAX_SHADER_PASSES * VERTEX_SHADER_NORMAL;
  g_next_pixel_shader = g_shader_pass + MAX_SHADER_PASSES * PIXEL_SHADER_SOLID;

  select_new_vertex_shader ();
  select_new_pixel_shader ();

  // set vertex buffer
  stride = VERTEX'size;
  offset = 0;
  DX.m_deviceContext->lpVtbl->IASetVertexBuffers (DX.m_deviceContext, 0, 1, ((ID3D11Buffer **)&v_id), &stride, &offset);

  // set index buffer
  DX.m_deviceContext->lpVtbl->IASetIndexBuffer (DX.m_deviceContext,
                                                *((ID3D11Buffer **)&i_id2),
                                                (i_id & 1) == 0 ? DXGI_FORMAT_R32_UINT : DXGI_FORMAT_R16_UINT,
                                                0);

  set_pixel_shader_ressources (t_id, object_id);
  flush_constant_buffers ();

  // Render the triangles.
  DX.m_deviceContext->lpVtbl->DrawIndexedInstanced (This                  => DX.m_deviceContext,
                                                    IndexCountPerInstance => (UINT)i_count,
                                                    InstanceCount         => (UINT)g_nb_of_instances,
                                                    StartIndexLocation    => (UINT)i_offset,
                                                    BaseVertexLocation    => v_offset,
                                                    StartInstanceLocation => 0);
}

//---------------------------------------------------------------------

public void draw_triangles (VERTEX_ID  v_id,
                            int        v_offset,
                            int        v_count,
                            TEXTURE_ID t_id[2],  // color and bump textures
                            uint       object_id = 0xFFFFFF)
{
  uint stride, offset;

  g_next_vertex_shader = g_shader_pass + MAX_SHADER_PASSES * VERTEX_SHADER_NORMAL;
  g_next_pixel_shader = g_shader_pass + MAX_SHADER_PASSES * PIXEL_SHADER_SOLID;

  select_new_vertex_shader ();
  select_new_pixel_shader ();

  // set vertex buffer
  stride = VERTEX'size;
  offset = 0;
  DX.m_deviceContext->lpVtbl->IASetVertexBuffers (DX.m_deviceContext, 0, 1, ((ID3D11Buffer **)&v_id), &stride, &offset);

  set_pixel_shader_ressources (t_id, object_id);
  flush_constant_buffers ();

  // Render the triangles.
  DX.m_deviceContext->lpVtbl->DrawInstanced (DX.m_deviceContext,
                                             VertexCountPerInstance => (UINT)v_count,
                                             InstanceCount          => (UINT)g_nb_of_instances,
                                             StartVertexLocation    => (UINT)v_offset,
                                             StartInstanceLocation  => 0);
}

//---------------------------------------------------------------------

public TEXTURE_ID create_texture (IMAGE_INFO img, bool allow_update = false, bool use_mipmapping = true)
{
  HRESULT                         hResult;
  D3D11_SUBRESOURCE_DATA          subresource;
  D3D11_TEXTURE2D_DESC            textureDesc;
  ID3D11Texture2D*                m_texture;
  D3D11_SHADER_RESOURCE_VIEW_DESC srvDesc;
  ID3D11ShaderResourceView*       m_textureView;

  clear subresource;
  subresource.pSysMem = &img.pixel^;
  subresource.SysMemPitch = img.width*4;

  // Setup the description of the texture.
  clear textureDesc;
  textureDesc.Height = img.height;
  textureDesc.Width = img.width;
  textureDesc.MipLevels = (UINT)(use_mipmapping ? 0 : 1);
  textureDesc.ArraySize = 1;
  textureDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
  textureDesc.SampleDesc.Count = 1;
//textureDesc.SampleDesc.Quality = 0;

  if (use_mipmapping)
  {
    textureDesc.Usage = D3D11_USAGE_DEFAULT;      // inaccessible to the CPU (the only one that works with mipmapping)
    textureDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_RENDER_TARGET;
    textureDesc.MiscFlags = D3D11_RESOURCE_MISC_GENERATE_MIPS;
  }
  else
  {
    textureDesc.Usage = allow_update ? D3D11_USAGE_DYNAMIC : D3D11_USAGE_IMMUTABLE;
    textureDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
    textureDesc.CPUAccessFlags = allow_update ? D3D11_CPU_ACCESS_WRITE : 0;
  }

  hResult = DX.m_device->lpVtbl->CreateTexture2D (DX.m_device, &textureDesc, (use_mipmapping ? null : &subresource), &m_texture);
  if (hResult < 0)
    fatal_abort ("CreateTexture2D()", hResult);

  clear srvDesc;
  srvDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
  srvDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
  // srvDesc.array.Texture2D.MostDetailedMip = 0;
  srvDesc.array.Texture2D.MipLevels = UINT'max;  // -1

  hResult = DX.m_device->lpVtbl->CreateShaderResourceView (DX.m_device, *(ID3D11Resource**)&m_texture, &srvDesc, &m_textureView);
  if (hResult < 0)
    fatal_abort ("CreateShaderResourceView()", hResult);

  if (use_mipmapping)
  {
    DX.m_deviceContext->lpVtbl->UpdateSubresource(DX.m_deviceContext, *(ID3D11Resource**)&m_texture, 0, null, subresource.pSysMem, subresource.SysMemPitch, 0);
    DX.m_deviceContext->lpVtbl->GenerateMips (DX.m_deviceContext, m_textureView);
  }

  m_texture->lpVtbl->Release((LPVOID)m_texture);

  return *((TEXTURE_ID *)&m_textureView);
}

//---------------------------------------------------------------------

public void update_texture (TEXTURE_ID id, IMAGE_INFO img)
{
  ID3D11ShaderResourceView* texture = *((ID3D11ShaderResourceView**)&id);
  ID3D11Resource*           res;
  LPVOID                    ptr;
  ID3D11Texture2D*          m_texture;
  D3D11_TEXTURE2D_DESC      desc;

  texture->lpVtbl->GetResource (texture, &res);
  res->lpVtbl->QueryInterface ((LPVOID)res, IID_ID3D11Texture2D, out ptr);
  assert ptr != null;
  m_texture'byte = ptr'byte;

  m_texture->lpVtbl->GetDesc (This => m_texture, &desc);

  assert img.width == desc.Width && img.height == desc.Height;

  DX.m_deviceContext->lpVtbl->UpdateSubresource(DX.m_deviceContext, *(ID3D11Resource**)&m_texture, 0, null, &img.pixel^, img.width*4, 0);
  DX.m_deviceContext->lpVtbl->GenerateMips (DX.m_deviceContext, texture);

  m_texture->lpVtbl->Release((LPVOID)m_texture);
  res->lpVtbl->Release((LPVOID)res);
}

//---------------------------------------------------------------------

public void update_texture_part (TEXTURE_ID id, uint[2] offset, IMAGE_INFO img)
{
  ID3D11ShaderResourceView* texture = *((ID3D11ShaderResourceView**)&id);
  ID3D11Resource*           res;
  LPVOID                    ptr;
  ID3D11Texture2D*          m_texture;
  D3D11_TEXTURE2D_DESC      desc;
  D3D11_BOX                 box;

  texture->lpVtbl->GetResource (texture, &res);
  res->lpVtbl->QueryInterface ((LPVOID)res, IID_ID3D11Texture2D, out ptr);
  assert ptr != null;
  m_texture'byte = ptr'byte;

  m_texture->lpVtbl->GetDesc (This => m_texture, &desc);

  assert offset[0] + img.width <= desc.Width && offset[1] + img.height <= desc.Height;

#if 1

  clear box;
  box.left = offset[0];
  box.right = box.left + img.width;
  box.top = offset[1];
  box.bottom = box.top + img.height;
  box.back = 1;

  DX.m_deviceContext->lpVtbl->UpdateSubresource
     (DX.m_deviceContext,
      *(ID3D11Resource**)&m_texture,
      0,
      &box,
      &img.pixel^,
      img.width*4,
      0);

#else
  {
    D3D11_MAPPED_SUBRESOURCE mappedResource;

    DX.m_deviceContext->lpVtbl->Map (DX.m_deviceContext, (ID3D11Resource*)m_texture, 0,
                                     D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);

    {
      byte* psource = &img.pixel^;
      byte* ptarget = mappedResource.pData;
      uint  size = 4 * img.width;
      uint  ln;

      for (ln=0; ln<img.height; ln++)
      {
        ptarget[0:size] = psource[0:size];
        psource += size;
        ptarget += mappedResource.RowPitch;
      }
    }

    DX.m_deviceContext->lpVtbl->Unmap (DX.m_deviceContext, (ID3D11Resource*)m_texture, 0);
  }
#endif

  DX.m_deviceContext->lpVtbl->GenerateMips (DX.m_deviceContext, texture);

  m_texture->lpVtbl->Release((LPVOID)m_texture);
  res->lpVtbl->Release((LPVOID)res);
}

//---------------------------------------------------------------------

public void free_texture (TEXTURE_ID t)
{
  ID3D11ShaderResourceView* m_texture = *((ID3D11ShaderResourceView **)&t);

  if (m_texture != null)
  {
    m_texture->lpVtbl->Release((LPVOID)m_texture);
    m_texture = null;
  }
}

//---------------------------------------------------------------------

public void free_vertex_buffer (VERTEX_ID v)
{
  ID3D11Buffer* m_vertexBuffer = *((ID3D11Buffer **)&v);

  if (m_vertexBuffer != null)
  {
    m_vertexBuffer->lpVtbl->Release((LPVOID)m_vertexBuffer);
    m_vertexBuffer = null;
  }
}

//---------------------------------------------------------------------

public void free_index_buffer (INDEX_ID i)
{
  INDEX_ID      i2 = (i & -2);   // set even in case it's a small index buffer
  ID3D11Buffer* m_indexBuffer = *((ID3D11Buffer **)&i2);

  if (m_indexBuffer != null)
  {
    m_indexBuffer->lpVtbl->Release((LPVOID)m_indexBuffer);
    m_indexBuffer = null;
  }
}

//---------------------------------------------------------------------

public void set_background_color (vector3 color)
{
  g_background_color[0:3] = color;
}

//---------------------------------------------------------------------

public void set_ambiant_light (vector3 color)
{
  if (memcmp (g_shader_constants_stable.AmbiantLightColor, color) != 0)
  {
    g_shader_constants_stable.AmbiantLightColor = color;
    g_shader_constants_stable_dirty = true;
  }
}

//---------------------------------------------------------------------

public void set_sun_light (vector3 color, vector3 direction)
{
  vector3 v;

  D3DXVec3Normalize (out v, {-direction[0], -direction[1], -direction[2]});

  if (memcmp (g_shader_constants_stable.SunLightColor, color) != 0 ||
      memcmp (g_shader_constants_stable.SunLightDirection, v) != 0)
  {
    g_shader_constants_stable.SunLightColor = color;
    g_shader_constants_stable.SunLightDirection = v;
    g_shader_constants_stable_dirty = true;
  }
}

//---------------------------------------------------------------------

public void set_world_transform (matrix44 m)
{
  transpose_44 (m, out g_shader_constants_dynamic.TransposedWorld);
}

//---------------------------------------------------------------------

public void set_object_colors (OBJECT_COLORS colors)
{
  g_shader_constants_dynamic.MaterialColor             = colors.MaterialColor;
  g_shader_constants_dynamic.EmissiveColor             = colors.EmissiveColor;
  g_shader_constants_dynamic.specular_cone             = colors.specular_cone;
  g_shader_constants_dynamic.ShinyFactor               = colors.shininess;
  g_shader_constants_dynamic.ambiantlight_on           = (int)colors.ambiant_enabled;
  g_shader_constants_dynamic.sunlight_on               = (int)colors.sunlight_enabled;
  g_shader_constants_dynamic.ambiant_downwards         = colors.ambiant_downwards;
  g_shader_constants_dynamic.parallax_factor           = colors.parallax_factor;
}

//---------------------------------------------------------------------

public void set_perlin_noise (bool enable, vector3 camera_offset = {0.0, 0.0, 0.0})
{
  if (enable)
  {
    uint POW = 25;
    int tclock = (int)(g_frame_tick & ((1 << POW)-1));  // lower bits
    if ((g_frame_tick & (1 << POW)) != 0)               // overflow
      tclock = (int)(1 << POW) - (int)tclock;           // set a negative value
    g_shader_constants_dynamic.tclock = (1.0 / 1003.0) * (float)(tclock + 1);  // add 1 to make sure it's never zero
    g_shader_constants_stable.CameraOffset = camera_offset;
    g_shader_constants_stable_dirty = true;
  }
  else
    g_shader_constants_dynamic.tclock = 0.0;
}

//---------------------------------------------------------------------

public void set_uv_transform (OBJECT_UV_TRANSFORM uv_transform)
{
  g_shader_constants_dynamic.uv_transform = uv_transform;
}

//---------------------------------------------------------------------

// set lights
public void set_lights (LIGHT[] lights)
{
  if (lights'length == g_shader_constants_stable.NbLights &&
      memcmp (lights, g_shader_constants_stable.Light'byte[0:lights'size]) == 0)
    return;   // already identical

  g_shader_constants_stable.NbLights = lights'length;
  g_shader_constants_stable.Light'byte[0:lights'size] = lights'byte;
  g_shader_constants_stable_dirty = true;
}

//---------------------------------------------------------------------

public void set_fog (
   float   start_z,
   float   end_z,
   vector3 color = {0.62,0.62,0.62},
   float   attenuation = 1.0)
{
  g_shader_constants_stable.fogColor = color;

  if (end_z <= start_z)
    g_shader_constants_stable.fogInvRange = 0.0;
  else
  {
    g_shader_constants_stable.fogStart = start_z;
    g_shader_constants_stable.fogInvRange = 1.0 / (end_z - start_z);
    g_shader_constants_stable.FogInvExponent = attenuation;
  }

  g_shader_constants_stable_dirty = true;
}

//---------------------------------------------------------------------

public void set_camera (vector3 eye, quaternion view)
{
  float x2 = 2.0 * (float)view.x;
  float y2 = 2.0 * (float)view.y;
  float z2 = 2.0 * (float)view.z;

  float xx = x2 * (float)view.x;
  float xy = x2 * (float)view.y;
  float xz = x2 * (float)view.z;
  float xw = x2 * (float)view.w;
  float yy = y2 * (float)view.y;
  float yz = y2 * (float)view.z;
  float yw = y2 * (float)view.w;
  float zz = z2 * (float)view.z;
  float zw = z2 * (float)view.w;

  XMMATRIX vm;
  int      i;

  vm = {m =>
         {{1.0 - (yy + zz),  (xy - zw),        (xz + yw),         0.0},
          {(xy + zw),        1.0 - (xx + zz),  (yz - xw),         0.0},
          {(xz - yw),        (yz + xw),        1.0 - (xx + yy),   0.0},
          {0.0,              0.0,              0.0,               1.0}}};

  for (i=0; i<3; i++)
    vm.m[3][i] = -vm.m[0][i] * eye[0] - vm.m[1][i] * eye[1] - vm.m[2][i] * eye[2];

  g_default_view_matrix = vm;   // used by renderer & by compute_picking_ray()
  g_default_camera_eye = eye;

  build_view_frustum (*(matrix44*)&g_default_view_matrix, *(matrix44*)&g_default_projection_matrix, out g_frustum);
}

//---------------------------------------------------------------------

public void set_shadow_matrix (matrix44 m)
{
  g_default_shadow_matrix = *(XMMATRIX*)&m;
  if (g_shadows_level > 0)
    MXXMatrixTranspose (g_default_shadow_matrix, out g_shader_constants_stable.ShadowTransposedTransform);
  g_shader_constants_stable_dirty = true;
}

//---------------------------------------------------------------------

public void clear_depth_buffer ()
{
  DX.m_deviceContext->lpVtbl->ClearDepthStencilView (DX.m_deviceContext, DX.m_depth_View, D3D11_CLEAR_DEPTH, 0.0f, INITIAL_STENCIL_VALUE);
}

//---------------------------------------------------------------------

public void clear_stencil_buffer ()
{
  DX.m_deviceContext->lpVtbl->ClearDepthStencilView (DX.m_deviceContext, DX.m_depth_View, D3D11_CLEAR_STENCIL, 0.0f, INITIAL_STENCIL_VALUE);
}

//---------------------------------------------------------------------

void intern_compute_picking_ray (    int         mouse_x,
                                     int         mouse_y,
                                     matrix44    view_matrix,
                                 out PICKING_RAY picking_ray)
{
  int       vPickx, vPicky;
  XMFLOAT3  vPickDir, vPickNear, vPickFar;
  float     vPickLen;
  XMMATRIX  m;
  XMFLOAT3  v;
  XMFLOAT3  vPickRayOrig;
  float     d;

  vPickx = mouse_x;
  vPicky = mouse_y;

  // Get the inverse view matrix
  D3DXMatrixInverse (&m, null, (XMMATRIX*)&view_matrix);

#if 0
inverse is :

right.x right.y right.z 0
up.x    up.y    up.z    0
fwd.x   fwd.y   fwd.z   0
eye.x   eye.y   eye.z   1
#endif

  vPickRayOrig = {m.m[3][0], m.m[3][1], m.m[3][2]};     // eye origin

  v = {((float)(2*vPickx) / (float)g_back_buffer_x_res - 1.0) / g_default_projection_matrix.m[0][0],
       - ((float)(2*vPicky) / (float)g_back_buffer_y_res - 1.0) / g_default_projection_matrix.m[1][1],
       1.0f};

  vPickDir = {v[0]*m.m[0][0] + v[1]*m.m[1][0] + v[2]*m.m[2][0],
              v[0]*m.m[0][1] + v[1]*m.m[1][1] + v[2]*m.m[2][1],
              v[0]*m.m[0][2] + v[1]*m.m[1][2] + v[2]*m.m[2][2]};

  D3DXVec3Normalize (out vPickDir, vPickDir);

//  trace ("vPickRayOrig = {%f, %f, %f}\n", vPickRayOrig.x, vPickRayOrig.y, vPickRayOrig.z);
//  trace ("vPickDir     = {%f, %f, %f}\n", vPickDir.x, vPickDir.y, vPickDir.z);

  d = 0.0; // (float)sqrt (near_z*near_z + v.x*v.x + v.y*v.y);

  vPickNear = {vPickRayOrig[0] + d*vPickDir[0],
               vPickRayOrig[1] + d*vPickDir[1],
               vPickRayOrig[2] + d*vPickDir[2]};

//  trace ("vPickNear = {%f, %f, %f}\n", vPickNear.x, vPickNear.y, vPickNear.z);

  d = (float)sqrt (g_far_z*g_far_z + v[0]*v[0] + v[1]*v[1]);

  vPickFar = {vPickRayOrig[0] + d*vPickDir[0],
              vPickRayOrig[1] + d*vPickDir[1],
              vPickRayOrig[2] + d*vPickDir[2]};

//  trace ("vPickFar  = {%f, %f, %f}\n", vPickFar.x, vPickFar.y, vPickFar.z);

  vPickLen = length_3 ({vPickFar[0] - vPickNear[0],
                        vPickFar[1] - vPickNear[1],
                        vPickFar[2] - vPickNear[2]});

  picking_ray = { vPickNear => {vPickNear[0], vPickNear[1], vPickNear[2]},
                  vPickFar  => {vPickFar[0], vPickFar[1], vPickFar[2]},
                  vPickDir  => {vPickDir[0], vPickDir[1], vPickDir[2]},
                  vPickLen  => vPickLen };
}

//---------------------------------------------------------------------

public
void compute_picking_ray (int             mouse_x,
                          int             mouse_y,
                          out PICKING_RAY picking_ray)
{
  intern_compute_picking_ray (mouse_x, mouse_y, *(matrix44*)&g_default_view_matrix, out picking_ray);
}

//---------------------------------------------------------------------

public
void compute_picking_ray_for_hud (int             mouse_x,
                                  int             mouse_y,
                                  out PICKING_RAY picking_ray)
{
  intern_compute_picking_ray (mouse_x, mouse_y, M44_ONE, out picking_ray);
}

//---------------------------------------------------------------------

public
bool ray_intersects_sphere (vector3     center,
                            float       squared_radius,  // radius * radius
                            PICKING_RAY ray)
{
  vector3 vh, vClosest, vToCenter;
  float   a;

  sub_3 (center, ray.vPickNear, out vh);

  a = dot_product_3 (vh, ray.vPickDir);

  // "a" equals the distance from the ray's near plane intersection
  // to the point on the ray nearest the center of the sphere.
  // If this point is outside the line segment contained in the frustum,
  // we constrain it to the near or far end of the segment.

  if (a < 0.0)
    a = 0.0;
  else if (a > ray.vPickLen)
    a = ray.vPickLen;

  mult_3_1 (ray.vPickDir, a, out vClosest);
  add_3 (vClosest, ray.vPickNear, out vClosest);

  sub_3 (center, vClosest, out vToCenter);

  // calculates the square of the distance from the closest point on the line segment
  // to the center of the bounding sphere.
  // If it is less than or equal to the square of the bounding sphere radius,
  // we have a possible hit.

  return (squared_radius >= dot_product_3 (vToCenter, vToCenter));
}

//---------------------------------------------------------------------

public
bool ray_intersects_triangle (    vector3     vert0,
                                  vector3     vert1,
                                  vector3     vert2,
                                  PICKING_RAY ray,
                              out float[2]    texture_info,
                              out float       distance,
                                  bool        test_both_sides = false)  // false is much faster
{
  const float EPSILON = 0.000001;

  vector3 edge1, edge2, tvec, pvec, qvec;
  float   u, v, det, inv_det;

  // find vectors for two edges sharing vert0
  sub_3 (vert1, vert0, out edge1);
  sub_3 (vert2, vert0, out edge2);

  // begin calculating determinant - also used to calculate u parameter
  cross_product_3 (ray.vPickDir, edge2, out pvec);

  // if determinant is near zero, ray lies in plane of triangle
  det = dot_product_3 (edge1, pvec);

  if (!test_both_sides)
  {
    if (det < EPSILON)
    {
      clear texture_info, distance;
      return false;
    }

    // calculate distance from vert0 to ray origin
    sub_3 (ray.vPickNear, vert0, out tvec);

    // calculate U parameter and test bounds
    u = dot_product_3 (tvec, pvec);
    if (u < 0.0 || u > det)
    {
      clear texture_info, distance;
      return false;
    }

    // prepare to test V parameter
    cross_product_3 (tvec, edge1, out qvec);

    // calculate V parameter and test bounds
    v = dot_product_3 (ray.vPickDir, qvec);
    if (v < 0.0 || u+v > det)
    {
      clear texture_info, distance;
      return false;
    }

    // calculate distance, scale uv
    distance = dot_product_3 (edge2, qvec);
    inv_det = 1.0 / det;
    distance *= inv_det;

    if (distance < 0.0 || distance > ray.vPickLen)
    {
      clear texture_info, distance;
      return false;
    }

    texture_info = {u * inv_det, v * inv_det};
  }
  else   // the non-culling branch
  {
    if (det > -EPSILON && det < EPSILON)
    {
      clear texture_info, distance;
      return false;
    }

    inv_det = 1.0 / det;

    // calculate distance from vert0 to ray origin
    sub_3 (ray.vPickNear, vert0, out tvec);

    // calculate U parameter and test bounds
    u = dot_product_3 (tvec, pvec) * inv_det;
    if (u < 0.0 || u > 1.0)
    {
      clear texture_info, distance;
      return false;
    }

    // prepare to test V parameter
    cross_product_3 (tvec, edge1, out qvec);

    // calculate V parameter and test bounds
    v = dot_product_3 (ray.vPickDir, qvec) * inv_det;
    if (v < 0.0 || u+v > 1.0)
    {
      clear texture_info, distance;
      return false;
    }

    // calculate distance
    distance = dot_product_3 (edge2, qvec) * inv_det;

    if (distance < 0.0 || distance > ray.vPickLen)
    {
      clear texture_info, distance;
      return false;
    }

    texture_info = {u, v};
  }

  return true;
}

//---------------------------------------------------------------------

public
void ray_compute_texture_uv (    vector2  tex0,           // .t components of triangle
                                 vector2  tex1,
                                 vector2  tex2,
                                 float[2] texture_info,   // result from ray_intersects_triangle()
                             out vector2  uv)             // texture coordinates
{
  uv = {tex0[0] + texture_info[0] * (tex1[0]-tex0[0]) + texture_info[1] * (tex2[0]-tex0[0]),
        tex0[1] + texture_info[0] * (tex1[1]-tex0[1]) + texture_info[1] * (tex2[1]-tex0[1])};
}

//---------------------------------------------------------------------

public RIGGED_VERTEX_ID create_rigged_vertex_buffer (RIGGED_VERTEX[] v)
{
  D3D11_BUFFER_DESC      vertexBufferDesc;
  D3D11_SUBRESOURCE_DATA vertexData;
  HRESULT                result;
  ID3D11Buffer*          vertexBuffer;

  // Set up the description of the static vertex buffer.
  clear vertexBufferDesc;
  vertexBufferDesc.Usage = D3D11_USAGE_IMMUTABLE;
  vertexBufferDesc.ByteWidth = v'size;
  vertexBufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
//  vertexBufferDesc.CPUAccessFlags = 0;

  // Give the subresource structure a pointer to the vertex data.
  clear vertexData;
  vertexData.pSysMem = (byte*)&v[0];
// vertexData.SysMemPitch = 0;
// vertexData.SysMemSlicePitch = 0;

  // Now create the vertex buffer.
  result = DX.m_device->lpVtbl->CreateBuffer (DX.m_device, &vertexBufferDesc, &vertexData, &vertexBuffer);
  if (result < 0)
    fatal_abort ("CreateBuffer()", result);

  return *((RIGGED_VERTEX_ID *)&vertexBuffer);
}

//---------------------------------------------------------------------

public void free_rigged_vertex_buffer (RIGGED_VERTEX_ID v)
{
  ID3D11Buffer* m_vertexBuffer = *((ID3D11Buffer **)&v);

  if (m_vertexBuffer != null)
  {
    m_vertexBuffer->lpVtbl->Release((LPVOID)m_vertexBuffer);
    m_vertexBuffer = null;
  }
}

//---------------------------------------------------------------------

public void create_bone_buffer (int nb_bones, out BONE_BUFFER_ID id)
{
  HRESULT           result;
  D3D11_BUFFER_DESC constantBufferDesc;
  ID3D11Buffer*     p;
  int               i;
  uint              rest, chunk;

  clear id;
  assert nb_bones <= 192;

  rest = (uint)(nb_bones * 64);
  i = 0;

  while (rest > 0)
  {
    chunk = rest;
    if (chunk > 4096)  // max buffer size
      chunk = 4096;

    clear constantBufferDesc;
    constantBufferDesc.Usage = D3D11_USAGE_DYNAMIC;    // cpu write, GPU read
    constantBufferDesc.ByteWidth = chunk;
    constantBufferDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    constantBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    result = DX.m_device->lpVtbl->CreateBuffer (DX.m_device, &constantBufferDesc, pInitialData => null, &p);
    if (result < 0)
      fatal_abort ("CreateBuffer()", result);

    id.id[i++] = *((RIGGED_VERTEX_ID *)&p);
    rest -= chunk;
  }
}

//---------------------------------------------------------------------

public void write_bone_buffer (BONE_BUFFER_ID id, matrix44[] m)
{
  int  i;
  uint ofs, size, chunk;

  assert m'length <= 192;

  ofs  = 0;
  size = (uint)(m'length * 64);
  i = 0;

  while (ofs < size)
  {
    ID3D11Buffer*            p = *((ID3D11Buffer**)&id.id[i++]);
    HRESULT                  result;
    D3D11_MAPPED_SUBRESOURCE mappedResource;

    chunk = size - ofs;
    if (chunk > 4096)  // max buffer size
      chunk = 4096;

    // Lock the constant buffer so it can be written to.
    result = DX.m_deviceContext->lpVtbl->Map (DX.m_deviceContext, (ID3D11Resource*)p, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedResource);
    if (result < 0)
      fatal_abort ("Map()", result);

    // copy the new values into the constant buffer.
    mappedResource.pData[0 : chunk] = m'byte[ofs : chunk];

    // Unlock the constant buffer.
    DX.m_deviceContext->lpVtbl->Unmap (DX.m_deviceContext, (ID3D11Resource*)p, 0);

    ofs += chunk;
  }
}

//---------------------------------------------------------------------

public void free_bone_buffer (BONE_BUFFER_ID id)
{
  int i;
  for (i=0; i<id.id'length && id.id[i] != 0; i++)
  {
    ID3D11Buffer* p = *((ID3D11Buffer**)&id.id[i]);
    if (p != null)
      p->lpVtbl->Release((LPVOID)p);
  }
}

//---------------------------------------------------------------------

// speeds up, but can cause bad display on AMD cards.

public void set_fast_rigged_vertex_shader (bool fast)
{
  g_shader_supports_indexing = fast;
}

//---------------------------------------------------------------------

public void draw_rigged_triangles (RIGGED_VERTEX_ID  v_id,
                                   int               v_offset,
                                   int               v_count,
                                   BONE_BUFFER_ID    b_id,
                                   TEXTURE_ID        t_id[2],  // color and bump textures
                                   uint              object_id = 0xFFFFFF)
{
  uint stride, offset;

  g_next_vertex_shader = g_shader_pass + MAX_SHADER_PASSES * VERTEX_SHADER_RIGGED;
  g_next_pixel_shader = g_shader_pass + MAX_SHADER_PASSES * PIXEL_SHADER_SOLID;

  select_new_vertex_shader (b_id);
  select_new_pixel_shader ();

  stride = RIGGED_VERTEX'size;
  offset = 0;
  DX.m_deviceContext->lpVtbl->IASetVertexBuffers (DX.m_deviceContext, 0, 1, ((ID3D11Buffer **)&v_id), &stride, &offset);

  set_pixel_shader_ressources (t_id, object_id);
  flush_constant_buffers ();

  // Render the triangles.
  DX.m_deviceContext->lpVtbl->DrawInstanced (DX.m_deviceContext,
                                             VertexCountPerInstance => (UINT)v_count,
                                             InstanceCount          => (UINT)g_nb_of_instances,
                                             StartVertexLocation    => (UINT)v_offset,
                                             StartInstanceLocation  => 0);
}

//---------------------------------------------------------------------

public void draw_indexed_rigged_triangles (RIGGED_VERTEX_ID  v_id,
                                           int               v_offset,
                                           INDEX_ID          i_id,
                                           int               i_offset,
                                           int               i_count,
                                           BONE_BUFFER_ID    b_id,
                                           TEXTURE_ID        t_id[2],  // color and bump textures
                                           uint              object_id = 0xFFFFFF)
{
  INDEX_ID i_id2 = i_id;
  uint     stride, offset;

  i_id2 &= -2;    // set even

  g_next_vertex_shader = g_shader_pass + MAX_SHADER_PASSES * VERTEX_SHADER_RIGGED;
  g_next_pixel_shader = g_shader_pass + MAX_SHADER_PASSES * PIXEL_SHADER_SOLID;

  select_new_vertex_shader (b_id);
  select_new_pixel_shader ();

  // set vertex buffer
  stride = RIGGED_VERTEX'size;
  offset = 0;
  DX.m_deviceContext->lpVtbl->IASetVertexBuffers (DX.m_deviceContext, 0, 1, ((ID3D11Buffer **)&v_id), &stride, &offset);

  // set index buffer
  DX.m_deviceContext->lpVtbl->IASetIndexBuffer (DX.m_deviceContext,
                                                *((ID3D11Buffer **)&i_id2),
                                                (i_id & 1) == 0 ? DXGI_FORMAT_R32_UINT : DXGI_FORMAT_R16_UINT,
                                                0);

  set_pixel_shader_ressources (t_id, object_id);
  flush_constant_buffers ();

  // Render the triangles.
  DX.m_deviceContext->lpVtbl->DrawIndexedInstanced (This                  => DX.m_deviceContext,
                                                    IndexCountPerInstance => (UINT)i_count,
                                                    InstanceCount         => (UINT)g_nb_of_instances,
                                                    StartIndexLocation    => (UINT)i_offset,
                                                    BaseVertexLocation    => v_offset,
                                                    StartInstanceLocation => 0);
}

//---------------------------------------------------------------------
#end unsafe

//---------------------------------------------------------------------

void build_view_frustum (    matrix44 m_view,
                             matrix44 m_projection,
                         out vector4  m_frustum[6])  // 6 plane a, b, c, d
{
  matrix44 viewProjection;
  int i;

  clear m_frustum;
  mult_44_44 (m_view, m_projection, out viewProjection);

  transpose_44 (viewProjection, out viewProjection);

  // Left plane
  add_4 (viewProjection[3], viewProjection[0], out m_frustum[0]);

  // Right plane
  sub_4 (viewProjection[3], viewProjection[0], out m_frustum[1]);

  // Top plane
  sub_4 (viewProjection[3], viewProjection[1], out m_frustum[2]);

  // Bottom plane
  add_4 (viewProjection[3], viewProjection[1], out m_frustum[3]);

  // Near plane
  m_frustum[4] = viewProjection[2];

  // Far plane
  sub_4 (viewProjection[3], viewProjection[2], out m_frustum[5]);

  // Normalize planes
  for (i=0; i<6; i++)
  {
    ref vector4 v = m_frustum[i];
    float f = (float)(1.0 / sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
    mult_4_1 (v, f, out v);
  }
}

//---------------------------------------------------------------------

public bool is_sphere_in_frustum (vector3 point, float radius)
{
	int i;
	for (i=0; i<6; i++)
	{
    if (signed_distance_to_plane (g_frustum[i], point) + radius < 0.0)   // outside plane
      return false;
  }
  return true;
}

//---------------------------------------------------------------------

public bool is_box_in_frustum (vector3 center, vector3 size)
{
	int     i;
  vector3 s;

  mult_3_1 (size, 0.5, out s);

	for (i=0; i<6; i++)
	{
    ref vector4 v = g_frustum[i];

    // check if all 8 points of the rectangle are outside the same plane of the view frustum.

		if (signed_distance_to_plane (v, {(center[0] - s[0]), (center[1] - s[1]), (center[2] - s[2])}) >= 0.0)
			continue;

		if (signed_distance_to_plane (v, {(center[0] + s[0]), (center[1] - s[1]), (center[2] - s[2])}) >= 0.0)
			continue;

		if (signed_distance_to_plane (v, {(center[0] - s[0]), (center[1] + s[1]), (center[2] - s[2])}) >= 0.0)
			continue;

		if (signed_distance_to_plane (v, {(center[0] + s[0]), (center[1] + s[1]), (center[2] - s[2])}) >= 0.0)
			continue;

		if (signed_distance_to_plane (v, {(center[0] - s[0]), (center[1] - s[1]), (center[2] + s[2])}) >= 0.0)
			continue;

		if (signed_distance_to_plane (v, {(center[0] + s[0]), (center[1] - s[1]), (center[2] + s[2])}) >= 0.0)
			continue;

		if (signed_distance_to_plane (v, {(center[0] - s[0]), (center[1] + s[1]), (center[2] + s[2])}) >= 0.0)
			continue;

		if (signed_distance_to_plane (v, {(center[0] + s[0]), (center[1] + s[1]), (center[2] + s[2])}) >= 0.0)
			continue;

		return false;
	}

	return true;
}

//---------------------------------------------------------------------

public bool is_shadow_in_frustum (vector3 position, vector3 shadow_points[4], float radius)
{
  vector3 p[4];
  int     i;

  clear p;
  for (i=0; i<4; i++)
    add_3 (position, shadow_points[i], out p[i]);

  // test if all points are outside same plane frustrum at at least radius distance,
  // then the shadow is not visible in frustrum.

  for (i=0; i<6; i++)
  {
    if (signed_distance_to_plane (g_frustum[i], p[0]) + radius < 0.0 &&   // p[0] outside plane
        signed_distance_to_plane (g_frustum[i], p[1]) + radius < 0.0 &&   // p[1] outside plane
        signed_distance_to_plane (g_frustum[i], p[2]) + radius < 0.0 &&   // p[2] outside plane
        signed_distance_to_plane (g_frustum[i], p[3]) + radius < 0.0)     // p[3] outside same plane
      return false;
  }

  return true;
}

//---------------------------------------------------------------------------------------

public void get_view_matrix (out matrix44 view)
{
#begin unsafe
  view = *((matrix44*)&g_default_view_matrix);
#end unsafe
}

//-------------------------------------------------------------------------------------------------

public void set_shadows_number_of_levels (int level)  // 0 = off
{
  if (g_next_frame_shadows_level != level)
  {
    int l = level;
    if (l < 0) l = 0;
    if (l > 5) l = 5;
    g_next_frame_shadows_level = l;
  }
}

//-------------------------------------------------------------------------------------------------
