
/* zip.c */

use files, memory, pack/pack, pack/unpack;

/***********************************************************************/

#begin unsafe

package MEM

  struct DATA
  {
    byte* Rbuffer;
    int   Rsize;
    int   Rofs;
    byte* Wbuffer;
    int   Wsize;
    int   Wofs;
    bool  manage_Wbuffer;   // true = our layer allocated Wbuffer and will manage it in io_write(), false = it's a user buffer.
  }

  package P1 = new PACK   (USER_INFO => DATA);
  package P2 = new UNPACK (USER_INFO => DATA);

end MEM;

/***********************************************************************/

int io_read (ref DATA   info,
             out byte[] buffer)
{
  int len;
  byte *p = &buffer;
  len = info.Rsize - info.Rofs;
  if (len > buffer'length)
    len = buffer'length;
  p[0:len] = info.Rbuffer[info.Rofs:len];
  info.Rofs += len;
  return len;
}

/***********************************************************************/

int io_write (ref DATA info,
              byte[]   buffer)
{
  int len = buffer'length;

  if (len > info.Wsize - info.Wofs)   // not enough space in buffer
  {
    if (info.manage_Wbuffer)
    {
      while (len > info.Wsize - info.Wofs)  // not enough space in buffer
      {
        info.Wsize <<= 1;   // double buffer size
        info.Wbuffer = memory.realloc (info.Wbuffer, (uint)info.Wsize);
      }
    }
    else   // no buffer management, just write what you can
    {
      len = info.Wsize - info.Wofs;
    }
  }

  info.Wbuffer[info.Wofs:len] = buffer[0:len];
  info.Wofs += len;
  return len;
}

/***********************************************************************/

public void compress_block (    byte[] block_in,
                            out byte[] block_out,
                            out int    block_out_size)
{
  DATA d;

  d = {Rbuffer => &block_in,
       Rsize   => block_in'length,
       Rofs    => 0,

       Wbuffer => &block_out,
       Wsize   => block_out'length,
       Wofs    => 1,

       manage_Wbuffer => false};

  if (pack (ref d, io_read, io_write) < 0 || d.Wofs > 1+block_in'length)
  {
    block_out[0] = 0;   /* no compression */
    block_out[1:block_in'length] = block_in[0:block_in'length];
    d.Wofs = 1+block_in'length;
  }
  else
  {
    block_out[0] = 1;   /* compression */
  }

  block_out_size = d.Wofs;
}

/***********************************************************************/

public void decompress_block (    byte[] block_in,
                              out byte[] block_out,
                              out int block_out_size)
{
  DATA d;
  READ_AHEAD extra;

  byte* p = &block_out;
  _unused p;

  if (block_in[0] == 0)   /* no compression */
  {
    block_out[0:block_in'length-1] = block_in[1:block_in'length-1];
    block_out_size = block_in'length-1;
    return;
  }

  d = {Rbuffer => &block_in,
       Rsize   => block_in'length,
       Rofs    => 1,

       Wbuffer => &block_out,
       Wsize   => block_out'length,
       Wofs    => 0,

       manage_Wbuffer => false};

  if (unpack (ref d, io_read, io_write, out extra) < 0)
    d.Wofs = 0;
  _unused extra;

  block_out_size = d.Wofs;
}

/***********************************************************************/

// returns null if decompression failed

public byte[]^ decompress_and_allocate_block (byte[] block_in)
{
  DATA       d;
  READ_AHEAD extra;
  int        size;
  byte[]^    p;

  if (block_in[0] == 0)   // no compression
    return new byte[] ' (block_in[1 : block_in'length-1]);

  size = (block_in'length + 4095) & (-4096);  // allocate at least compressed size, rounded up by multiple of 4K

  d = {Rbuffer => &block_in,
       Rsize   => block_in'length,
       Rofs    => 1,

       Wbuffer => memory.malloc ((uint)size),
       Wsize   => size,
       Wofs    => 0,

       manage_Wbuffer => true};

  if (unpack (ref d, io_read, io_write, out extra) < 0)
  {
    memory.freem (d.Wbuffer);
    return null;
  }
  _unused extra;

  p = new byte[]'(d.Wbuffer[0 : d.Wofs]);
  memory.freem (d.Wbuffer);
  return p;
}

/***********************************************************************/
#end unsafe
/***********************************************************************/

package P

  struct FDATA
  {
    int fin;
    int fout;
  }

  package P3 = new PACK   (USER_INFO => FDATA);
  package P4 = new UNPACK (USER_INFO => FDATA);

end P;

/***********************************************************************/

int fio_read (ref FDATA  f,
              out byte[] buffer)
{
  return read (f.fin, out buffer);
}

/***********************************************************************/

int fio_write (ref FDATA f,
               byte[]    buffer)
{
  return write (f.fout, buffer);
}

/***********************************************************************/

public int compress_file (string source, string target)
{
  FDATA f;
  int   rc;
  const byte flag = 1;  // compress
  
  clear f;

  f.fin = open (source);
  if (f.fin < 0)
    return -1;

  f.fout = create (target);
  if (f.fout < 0)
  {
    close (f.fin);
    return -1;
  }

  if (write (f.fout, flag) != 1)  // write compression flag
  {
    close (f.fin);
    close (f.fout);
    delete_file (target);
    return -1;
  }
  
  rc = P3.pack (ref f, fio_read, fio_write);

  if (close (f.fin) < 0)
    rc = -1;

  if (close (f.fout) < 0)
    rc = -1;

  if (rc < 0)
    delete_file (target);

  return rc;
}

/***********************************************************************/

public int decompress_file (string source, string target)
{
  FDATA f;
  byte  flag;
  int   rc;
  READ_AHEAD extra;
  
  clear f;

  f.fin = open (source);
  if (f.fin < 0)
    return -1;

  f.fout = create (target);
  if (f.fout < 0)
  {
    close (f.fin);
    return -1;
  }

  if (read (f.fin, out flag) != 1)  // read compression flag
  {
    close (f.fin);
    close (f.fout);
    delete_file (target);
    return -1;
  }

  if (flag == 1)
  {
    rc = P4.unpack (ref f, fio_read, fio_write, out extra);
    _unused extra;
  }
  else   // no compression
  {
    byte buffer[4096];
    
    for (;;)
    {
      rc = read (f.fin, out buffer);
      if (rc < 0)
        break;
      if (write (f.fout, buffer) != rc)
      {
        rc = -1;
        break;
      }
      if (rc < buffer'length)
        break;
    }
  }
  
  if (close (f.fin) < 0)
    rc = -1;

  if (close (f.fout) < 0)
    rc = -1;

  if (rc < 0)
    delete_file (target);

  return rc;
}

/***********************************************************************/

package body PACK

  package PACK1 = new GEN_PACK (USER_INFO => USER_INFO);

  //-------------------------------------------------------------------

  public int pack (ref USER_INFO user_info, IO_READ user_read, IO_WRITE user_write)
  {
    return PACK1.pack (ref user_info, user_read, user_write);
  }

  //-------------------------------------------------------------------

end PACK;



package body UNPACK

  package UNPACK1 = new GEN_UNPACK (USER_INFO => USER_INFO);

  //-------------------------------------------------------------------

  public int unpack (ref USER_INFO user_info, IO_READ user_read, IO_WRITE user_write, out READ_AHEAD extra)
  {
    return UNPACK1.unpack (ref user_info, user_read, user_write, out extra);
  }

  //-------------------------------------------------------------------

end UNPACK;
