
// exp.c : expression syntax & semantic analysis

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "token.h"
#include "lex.h"
#include "strings.h"
#include "entities.h"
#include "error.h"
#include "goptions.h"
#include "typedecl.h"
#include "arithm.h"
#include "type.h"
#include "exp.h"

/*****************************************************************************/

CONTEXT no_context;

/*****************************************************************************/

// attribute

typedef enum {
  ATTR_min, ATTR_max, ATTR_first, ATTR_last, ATTR_length, ATTR_byte, ATTR_string, ATTR_size,
  last_ATTR,
} ATTRIBUTE;

typedef struct {
  ATTRIBUTE attr;
  wchar     *id;
} ATTRIBUTE_INFO;

static ATTRIBUTE_INFO attribute_info[] =
 {{ATTR_min,    L"min"},
  {ATTR_max,    L"max"},
  {ATTR_first,  L"first"},
  {ATTR_last,   L"last"},
  {ATTR_length, L"length"},
  {ATTR_byte,   L"byte"},
  {ATTR_string, L"string"},
  {ATTR_size,   L"size"},
  {last_ATTR,   null},
 };

/******************************************************************************/

static void parse_name (PCONTEXT      context,                 // never null
                        bool          left_parenthesis_parsed,
                        TEXT_POSITION *name_position,          // never null
                        PENTITY       type_name,               // can be null
                        PEXPRESSION   *out);

/*****************************************************************************/

PEXPRESSION new_exp (KIND_EXPRESSION kind)
{
  PEXPRESSION p;

  p = malloc (sizeof(EXPRESSION));
  if (p == NULL)
    fatal_out_of_memory_error ("out_of_memory : new_exp");

  clear (*p);
  p->kind = kind;

  return p;
}

/*****************************************************************************/

static PEXPRESSION duplicate_constant_exp (PEXPRESSION exp)
{
  PEXPRESSION e;

  e = new_exp (exp->kind);

  memcpy (e, exp, sizeof(EXPRESSION));

  switch (e->kind)
  {
    case A_CONST_ENUMERATION_VALUE:
    case A_CONST_INTEGER_VALUE:
    case A_CONST_FLOAT_VALUE:
    case A_CONST_NULL_VALUE:
    case A_POOL_CONSTANT:
      // nothing to do
      break;

    default:
      fatal_compiler_error ("dup_cte_exp", &token.pos);
      break;
  }

  return e;
}

/*****************************************************************************/

// to check if default expressions of function declarations match.

bool constant_expressions_match (PEXPRESSION e1, PEXPRESSION e2)
{
  if (e1 == null && e2 == null)    // both null
    return true;

  if (e1 == null || e2 == null)    // one of both null
    return false;

  if (e1->kind != e2->kind)      // not same kind
    return false;

  if (e1->base_type_or_null == null && e2->base_type_or_null == null)
    ; // ok
  else if (e1->base_type_or_null == null || e2->base_type_or_null == null)
    return false;
  else if (!types_are_equal (e1->base_type_or_null, e2->base_type_or_null))
    return false;

  if (e1->constraint.kind  != e2->constraint.kind ||
      e1->constraint.value != e2->constraint.value ||
      e1->form             != e2->form ||
      e1->access           != e2->access)
    return false;

  switch (e1->kind)
  {
    case A_CONST_ENUMERATION_VALUE:
      if (e1->info.const_enumeration_value_info.value != e2->info.const_enumeration_value_info.value)
        return false;
      break;

    case A_CONST_INTEGER_VALUE:
      if (e1->info.const_integer_value_info.value != e2->info.const_integer_value_info.value)
        return false;
      break;

    case A_CONST_FLOAT_VALUE:
      if (e1->info.const_float_value_info.value != e2->info.const_float_value_info.value)
        return false;
      break;

    case A_CONST_NULL_VALUE:
      break;

    case A_POOL_CONSTANT:
      return pool_constants_are_identical (e1->info.pool_constant_info.pool_cte,
                                           e2->info.pool_constant_info.pool_cte);
    default:
      fatal_compiler_error ("cte_exp_match", &token.pos);
      break;
  }

  return true;
}

//*****************************************************************************

static void free_list_of_exp (LIST_OF_EXPRESSIONS *list)
{
  LIST_OF_EXPRESSIONS *p, *t;

  p = list;
  while (p != null)
  {
    t = p;
    p = p->next;
    free_exp (t->exp);
    free (t);
  }
}

//*****************************************************************************

void free_exp (PEXPRESSION e)
{
  int i;

  if (e == null)
    return;

  switch (e->kind)
  {
    case A_CONST_ENUMERATION_VALUE:
    case A_CONST_INTEGER_VALUE:
    case A_CONST_FLOAT_VALUE:
    case A_CONST_NULL_VALUE:
    case A_POOL_CONSTANT:
      break;

    case AN_OPERATOR_VALUE:
      for (i=0; i<3; i++)
        free_exp (e->info.operator_value_info.arg[i]);
      break;

    case A_RUN_CALL:
      free_exp (e->info.run_call_info.function_call);
      break;

    case A_FUNCTION_VALUE:
      break;

    case A_FUNCTION_CALL:
      free_exp (e->info.function_call_info.func);
      free_list_of_exp (e->info.function_call_info.param);
      break;

    case A_DISCRIMINANT_VALUE:
      free_exp (e->info.discriminant_value_info.prefix);
      break;

    case AN_UNC_ARRAY_AGGREGATE:
      free_exp (e->info.unc_array_aggregate_info.element);
      break;

    case AN_AGGREGATE_VALUE:
      free_list_of_exp (e->info.aggregate_value_info.list);
      break;

    case A_QUALIFIED_EXPRESSION:
      free_exp (e->info.qualified_expression_info.value);
      break;

    case AN_ARRAY_QUALIFIED_EXPRESSION:
      free_exp (e->info.array_qualified_expression_info.length);
      free_exp (e->info.array_qualified_expression_info.value);
      break;

    case A_STRUCT_QUALIFIED_EXPRESSION:
      free_exp (e->info.struct_qualified_expression_info.discriminant);
      free_exp (e->info.struct_qualified_expression_info.value);
      break;

    case AN_ALLOCATOR:
      free_exp (e->info.allocator_info.value);
      break;

    case A_GLOBAL_VARIABLE_OBJECT:
    case A_LOCAL_VARIABLE_OBJECT:
    case A_REFERENCE_OBJECT:
    case A_PARAMETER_OBJECT:
      break;

    case AN_ARRAY_ELEMENT_OBJECT:
      free_exp (e->info.array_element_object_info.prefix);
      free_exp (e->info.array_element_object_info.index);
      break;

    case AN_ARRAY_SLICE_OBJECT:
      free_exp (e->info.array_slice_object_info.prefix);
      free_exp (e->info.array_slice_object_info.index);
      free_exp (e->info.array_slice_object_info.length);
      break;

    case A_STRUCT_FIELD_OBJECT:
      free_exp (e->info.struct_field_object_info.prefix);
      break;

    case A_DEREFERENCED_OBJECT:
      free_exp (e->info.dereferenced_object_info.prefix);
      break;

    case AN_UNSAFE_DEREFERENCED_OBJECT:
      free_exp (e->info.unsafe_dereferenced_object_info.unsafe_ptr_value);
      break;

    case AN_ATTR_BYTE_OBJECT:
      free_exp (e->info.attr_byte_object_info.prefix);
      break;

    case A_BOXED_OBJECT:
      free_exp (e->info.boxed_object_info.parameter);
      break;

    case AN_UNBOXED_OBJECT:
      free_exp (e->info.unboxed_object_info.parameter);
      break;

    case A_BOXED_ARRAY_OBJECT:
      free_list_of_exp (e->info.boxed_array_object_info.list);
      break;

    default:
      fatal_compiler_error ("free_exp", &token.pos);
      break;
  }

  free (e);
}

/*****************************************************************************/

// returns a constant int, value 1, but undefined type.

PEXPRESSION dummy_expression ()
{
  PEXPRESSION e;

  e = new_exp (A_CONST_INTEGER_VALUE);
  e->base_type_or_null = null;
  e->form = A_VALUE;
  e->access = ACCESS_CONSTANT;
  e->info.const_integer_value_info.value = 1;

  return e;
}

/*****************************************************************************/

// assertion: expression has non-null integer type.

static PENTITY promote_one_int (PEXPRESSION e)
{
  PENTITY      t;
  INTEGER_INFO *pi;

  t = e->base_type_or_null;

  if (t == type_int_literal)
    return type_int_literal;

  if (t == type_long)
    return type_long;

  if (e->kind == A_CONST_INTEGER_VALUE &&
       (e->info.const_integer_value_info.value < integer_info[a_int4].min ||
        e->info.const_integer_value_info.value > integer_info[a_uint4].max))
    return type_long;

  pi = &integer_info[t->info->the_integer_type.type];

  if (pi->is_signed)
    return type_int;

  return type_uint;
}

/*****************************************************************************/

// assertion:
// - both expressions have non-null integer types.
// returns null if promotion fails.

static PENTITY promote_two_ints (PEXPRESSION e1, PEXPRESSION e2, bool result_can_be_int_literal)
{
  PENTITY      t1, t2;
  INTEGER_INFO *pi1, *pi2;

  t1 = e1->base_type_or_null;
  t2 = e2->base_type_or_null;

  if (t1 == type_int_literal && t2 == type_int_literal && result_can_be_int_literal)
    return type_int_literal;

  if (t1 == type_long || t2 == type_long)
    return type_long;

  if (e1->kind == A_CONST_INTEGER_VALUE &&
       (e1->info.const_integer_value_info.value < integer_info[a_int4].min ||
        e1->info.const_integer_value_info.value > integer_info[a_uint4].max))
    return type_long;

  if (e2->kind == A_CONST_INTEGER_VALUE &&
       (e2->info.const_integer_value_info.value < integer_info[a_int4].min ||
        e2->info.const_integer_value_info.value > integer_info[a_uint4].max))
    return type_long;

  pi1 = &integer_info[t1->info->the_integer_type.type];
  pi2 = &integer_info[t2->info->the_integer_type.type];

  if (pi1->is_signed)
  {
    if (pi2->is_signed ||
        (t2 == type_int_literal && e2->kind == A_CONST_INTEGER_VALUE &&
         (e2->info.const_integer_value_info.value >= integer_info[a_int4].min ||
          e2->info.const_integer_value_info.value <= integer_info[a_int4].max)))
      return type_int;
  }

  if (pi2->is_signed)
  {
    if (pi1->is_signed ||
        (t1 == type_int_literal && e1->kind == A_CONST_INTEGER_VALUE &&
         (e1->info.const_integer_value_info.value >= integer_info[a_int4].min ||
          e1->info.const_integer_value_info.value <= integer_info[a_int4].max)))
      return type_int;
  }

  if (!pi1->is_signed)
  {
    if ((!pi2->is_signed) ||
        (t2 == type_int_literal && e2->kind == A_CONST_INTEGER_VALUE &&
         (e2->info.const_integer_value_info.value >= integer_info[a_uint4].min ||
          e2->info.const_integer_value_info.value <= integer_info[a_uint4].max)))
      return type_uint;
  }

  if (!pi2->is_signed)
  {
    if ((!pi1->is_signed) ||
        (t1 == type_int_literal && e1->kind == A_CONST_INTEGER_VALUE &&
         (e1->info.const_integer_value_info.value >= integer_info[a_uint4].min ||
          e1->info.const_integer_value_info.value <= integer_info[a_uint4].max)))
      return type_uint;
  }

  return null;
}

/*****************************************************************************/

static bool too_large_for_float4 (double d)
{
  float f = (float)d;
  return fisinfinite (f);
}

/*****************************************************************************/

// assertion:
// - both expressions have non-null float types.
// never fails.

static PENTITY promote_two_floats (PEXPRESSION e1, PEXPRESSION e2, bool result_can_be_float_literal)
{
  PENTITY t1, t2;

  t1 = e1->base_type_or_null;
  t2 = e2->base_type_or_null;

  if (t1 == type_float_literal && t2 == type_float_literal && result_can_be_float_literal)
    return type_float_literal;

  if (t1 == type_double || t2 == type_double)
    return type_double;

  if (e1->kind == A_CONST_FLOAT_VALUE &&
       too_large_for_float4 (e1->info.const_float_value_info.value))
    return type_double;

  if (e2->kind == A_CONST_FLOAT_VALUE &&
       too_large_for_float4 (e2->info.const_float_value_info.value))
    return type_double;

  return type_float;
}

/*****************************************************************************/

static void type_check_bool_operator (PEXPRESSION e, TEXT_POSITION *pos)
{
  PENTITY t;
  t = e->base_type_or_null;
  if (t != null && t != type_bool)
    semantic_error ("expression must have type bool", pos);
}

/*****************************************************************************/

static void convert_type_into_base_type_and_constraint (PENTITY    type,                // in
                                                        bool       always_constrained,  // in
                                                        PENTITY    *base_type_or_null,  // out
                                                        CONSTRAINT *cons)               // out
{
  PENTITY t;

  clear (*base_type_or_null);
  clear (*cons);

  t = complete_type_of (type);

  if (t == null)
    return;

  if (t->kind == AN_OPEN_ARRAY_TYPE)
  {
    cons->kind = always_constrained ? RUNTIME_CONSTRAINT : UNCONSTRAINED;
  }
  else if (t->kind == A_STRUCT_TYPE && t->info->the_struct_type.is_open_type)
  {
    cons->kind = always_constrained ? RUNTIME_CONSTRAINT : UNCONSTRAINED;
  }
  else if (t->kind == AN_ARRAY_TYPE)
  {
    cons->kind = CONSTANT_CONSTRAINT;
    cons->value = t->info->the_array_type.length;

    t = complete_type_of (t->info->the_array_type.open_array);
  }
  else if (t->kind == A_CONSTRAINED_STRUCT_TYPE)
  {
    cons->kind = CONSTANT_CONSTRAINT;
    cons->value = t->info->the_constrained_struct_type.discriminant_value;

    t = complete_type_of (t->info->the_constrained_struct_type.open_struct);
  }

  *base_type_or_null = t;
}

/*****************************************************************************/

void convert_type_into_context (PENTITY  type,      // in
                                PCONTEXT context)   // out
{
  convert_type_into_base_type_and_constraint (type,
                                              /* always_constrained = */ false,
                                              &context->base_type_or_null,
                                              &context->constraint);
  context->address_of = false;
}

/*****************************************************************************/

// stores only simple constants (no jagged types)

static void simple_store_constant_into_pool (PEXPRESSION source, int size, POOL *pool)
{
  int8    r;
  double  rf;

  switch (source->kind)
  {
    case A_CONST_ENUMERATION_VALUE:
      r = source->info.const_enumeration_value_info.value;
      store_integer (pool, 0, r, size);
      break;

    case A_CONST_INTEGER_VALUE:
      r = source->info.const_integer_value_info.value;
      store_integer (pool, 0, r, size);
      break;

    case A_CONST_FLOAT_VALUE:
      rf = source->info.const_float_value_info.value;
      store_float (pool, 0, rf, size);
      break;

    case A_CONST_NULL_VALUE:     // can be unsafe ptr null value
      // nothing to do : pool zone is zeroed by default
      break;

    case A_POOL_CONSTANT:
      copy_pool_to_pool (source->info.pool_constant_info.pool_cte, 0,
                         pool, 0,
                         size);
      break;

    default:
      fatal_compiler_error ("exp#641", &token.pos);
      break;
  }
}

/*****************************************************************************/

static void load_constant_object_from_pool (POOL        *pool_cte,
                                            uint        offset,
                                            uint        size,
                                            PENTITY     type,
                                            PEXPRESSION *pout)
{
  PEXPRESSION out;

  if (type->kind == AN_ENUMERATION_TYPE)
  {
    int8 r;

    load_integer (pool_cte, offset, &r, size, 0);

    out = new_exp (A_CONST_ENUMERATION_VALUE);
    out->base_type_or_null = type;
    out->form = AN_OBJECT;
    out->access = ACCESS_CONSTANT;
    out->info.const_enumeration_value_info.value = (uint4)r;
  }
  else if (type->kind == AN_INTEGER_TYPE)
  {
    INTEGER_INFO  *pi;
    int8 r;

    pi = &integer_info[type->info->the_integer_type.type];
    load_integer (pool_cte, offset, &r, size, pi->is_signed);

    out = new_exp (A_CONST_INTEGER_VALUE);
    out->base_type_or_null = type;
    out->form = AN_OBJECT;
    out->access = ACCESS_CONSTANT;
    out->info.const_integer_value_info.value = r;
  }
  else if (type->kind == A_FLOAT_TYPE)
  {
    double rf;

    load_float (pool_cte, offset, &rf, size);

    out = new_exp (A_CONST_FLOAT_VALUE);
    out->base_type_or_null = type;
    out->form = AN_OBJECT;
    out->access = ACCESS_CONSTANT;
    out->info.const_float_value_info.value = rf;
  }
  else if (type->kind == A_POINTER_TYPE ||
           type->kind == A_FUNCTION_POINTER_TYPE ||
           type->kind == AN_UNSAFE_POINTER_TYPE)
  {
    out = new_exp (A_CONST_NULL_VALUE);
    out->base_type_or_null = type;
    out->form = AN_OBJECT;
    out->access = ACCESS_CONSTANT;
  }
  else if (is_open_type (type))   // open type -> copy jag reference as new pool object
  {
    int8 r;

    out = new_exp (A_POOL_CONSTANT);
    out->form = AN_OBJECT;
    out->access = ACCESS_CONSTANT;
    out->base_type_or_null = complete_type_of (type);
    out->constraint.kind = CONSTANT_CONSTRAINT;
    out->info.pool_constant_info.pool_cte = load_reference_of (pool_cte, offset);
    load_integer (pool_cte, offset+address_size, &r, 4, 0);   // fetch array length/discriminant value
    out->constraint.value = (uint4)r;
  }
  else    // any other constrained type -> copy pool object (array/struct)
  {
    POOL *p;

    out = new_exp (A_POOL_CONSTANT);
    out->form = AN_OBJECT;
    out->access = ACCESS_CONSTANT;

    convert_type_into_base_type_and_constraint (type,
                                                /*always_constrained=*/ true,
                                                &out->base_type_or_null,
                                                &out->constraint);
    p = new_pool_constant (size, align_of_pool_cte (pool_cte));
    copy_pool_to_pool (pool_cte, offset, p, 0, size);
    out->info.pool_constant_info.pool_cte = p;
  }

  *pout = out;
}

/*****************************************************************************/

// promote literal type into actual runtime types

static void promote_parameter (PEXPRESSION e)
{
  if (e->base_type_or_null == null)
    return;

  if (e->base_type_or_null == type_int_literal)
  {
    if (e->kind == A_CONST_INTEGER_VALUE)
    {
      if (e->info.const_integer_value_info.value >= integer_info[a_int4].min &&
          e->info.const_integer_value_info.value <= integer_info[a_int4].max)
      {
        e->base_type_or_null = type_int;
      }
      else
      {
        e->base_type_or_null = type_long;
      }
    }
  }
  else if (e->base_type_or_null == type_float_literal)
  {
    if (e->kind == A_CONST_FLOAT_VALUE)
    {
      if (too_large_for_float4 (e->info.const_float_value_info.value))
      {
        e->base_type_or_null = type_double;
      }
      else
      {
        e->base_type_or_null = type_float;
      }
    }
  }
  else if (e->base_type_or_null == type_null_literal)
  {
    e->base_type_or_null = type_unsafe_ptr_to_byte;
    check_unsafe_region ();
  }
}

/*****************************************************************************/

// error message is generated in function
// used for parameter and ref declaration

int check_ref_parameter_context_compatibility (PCONTEXT target, PEXPRESSION source, TEXT_POSITION *pos)
{
  PENTITY s, t;

  s = complete_type_of (source->base_type_or_null);      // actual
  t = complete_type_of (target->base_type_or_null);      // formal

  if (s == null || t == null)   // previous error
    return -1;

  if (!types_are_equal (s, t))
  {
    semantic_error ("type is not compatible", pos);
    return -1;
  }


  if (s->kind == AN_OPEN_ARRAY_TYPE)
  {
    if (source->constraint.kind == CONSTANT_CONSTRAINT &&
        target->constraint.kind == CONSTANT_CONSTRAINT &&
        source->constraint.value != target->constraint.value)
    {
      semantic_error ("array length mismatch", pos);
      return -1;
    }
  }


  if (s->kind == A_STRUCT_TYPE)
  {
    if (source->constraint.kind == CONSTANT_CONSTRAINT &&
        target->constraint.kind == CONSTANT_CONSTRAINT &&
        source->constraint.value != target->constraint.value)
    {
      semantic_error ("struct discriminant mismatch", pos);
      return -1;
    }
  }


  return 0;
}

/*****************************************************************************/

typedef struct {
  LIST_OF_EXPRESSIONS *head, *tail;
  int                 count;
} EXPRESSIONS;

static void append_expression (EXPRESSIONS *q, EXPRESSION *exp, PENTITY type, PENTITY e)
{
  LIST_OF_EXPRESSIONS *p;

  p = malloc (sizeof(LIST_OF_EXPRESSIONS));
  if (p == null)
    fatal_out_of_memory_error ("append_expression()");

  clear (*p);
  p->exp  = exp;
  p->type = type;
  p->e    = e;

  if (q->tail == null)
    q->head = p;
  else
  {
    q->tail->next = p;
    p->prev = q->tail;
  }
  q->tail = p;

  q->count++;
}

static void free_list (EXPRESSIONS *q)
{
  LIST_OF_EXPRESSIONS *l, *s;
  l = q->head;
  while (l != null)
  {
    s = l;
    l = l->next;
    free (s);
  }
  q->count = 0;
}

/*****************************************************************************/

#define _FMINUS          0x01
#define _FPLUS           0x02
#define _FZERO           0x04
#define _DYN_WIDE        0x08
#define _DYN_PRECISION   0x10
#define _SKIP_PAR        0x20

/*****************************************************************************/

// returns -1 if we're out of arguments, 0 if OK (or minor matching errors)

static int check_arg (char                c,     // '*' means int, d u x e f c C s S
                      int                 flags,
                      int                 *argn,
                      LIST_OF_EXPRESSIONS **plist,
                      TEXT_POSITION       *pos)
{
  PENTITY type;

  if (*plist == null)      // no more actual arguments
  {
    semantic_error ("format string : missing arguments", pos);
    return -1;
  }

  type = (*plist)->type;
  if (type != null)
  {
    switch (c)
    {
      case '*':
        {
          // int
          if (type != type_int)
            semantic_error ("format string : %* must match 'int' argument", pos);
        }
        break;

      case 'd':
        {
          // any signed integer
          if (type->kind != AN_INTEGER_TYPE || !integer_info[type->info->the_integer_type.type].is_signed)
            semantic_error ("format string : %d must match signed integer argument", pos);

          if ((flags & (_FZERO+_FMINUS)) == (_FZERO+_FMINUS))
            semantic_error ("format string : flags '0' and '-' are not allowed together", pos);
        }
        break;

      case 'u':
        {
          // any unsigned integer or enum
          if (type->kind == AN_ENUMERATION_TYPE)
            ;
          else if (type->kind == AN_INTEGER_TYPE && !integer_info[type->info->the_integer_type.type].is_signed)
            ;
          else
            semantic_error ("format string : %u must match enum or unsigned integer argument", pos);

          if ((flags & _FPLUS) > 0)
            semantic_error ("format string : flag '+' is not allowed with type %u", pos);

          if ((flags & (_FZERO+_FMINUS)) == (_FZERO+_FMINUS))
            semantic_error ("format string : flags '0' and '-' are not allowed together", pos);
        }
        break;

      case 'x':
        {
          // any integer or enum
          if (type->kind != AN_ENUMERATION_TYPE && type->kind != AN_INTEGER_TYPE)
            semantic_error ("format string : %x must match enum or integer argument", pos);

          if ((flags & _FPLUS) > 0)
            semantic_error ("format string : flag '+' is not allowed with type %x", pos);

          if ((flags & (_FZERO+_FMINUS)) == (_FZERO+_FMINUS))
            semantic_error ("format string : flags '0' and '-' are not allowed together", pos);
        }
        break;

      case 'e':
      case 'f':
        {
          // any floating-point
          if (type->kind != A_FLOAT_TYPE)
            semantic_error ("format string : %f must match float or double argument", pos);

          if ((flags & (_FZERO+_FMINUS)) == (_FZERO+_FMINUS))
            semantic_error ("format string : flags '0' and '-' are not allowed together", pos);
        }
        break;

      case 'c':
        {
          // char or string
          if (type == type_char || types_are_equal (type, type_string))
            ;
          else
            semantic_error ("format string : %c must match char or string argument", pos);

          if ((flags & (_FZERO+_FPLUS)) != 0)
            semantic_error ("format string : illegal flags for type %c", pos);
        }
        break;

      case 'C':
        {
          // wchar or wstring
          if (type == type_wchar || types_are_equal (type, type_wstring))
            ;
          else
            semantic_error ("format string : %C must match wchar or wstring argument", pos);

          if ((flags & (_FZERO+_FPLUS)) != 0)
            semantic_error ("format string : illegal flags for type %C", pos);
        }
        break;

      case 's':
        {
          // string
          if (!types_are_equal (type, type_string))
            semantic_error ("format string : %s must match string argument", pos);

          if ((flags & (_FZERO+_FPLUS)) != 0)
            semantic_error ("format string : illegal flags for type %s", pos);
        }
        break;

      case 'S':
        {
          // wstring
          if (!types_are_equal (type, type_wstring))
            semantic_error ("format string : %S must match wstring argument", pos);

          if ((flags & (_FZERO+_FPLUS)) != 0)
            semantic_error ("format string : illegal flags for type %S", pos);
        }
        break;

      default:
         fatal_compiler_error ("check_arg", pos);
         break;
    }
  }

  (*argn)++;
  (*plist) = (*plist)->next;
  return 0;
}

/*****************************************************************************/

static int next_wchar (POOL *p, uint4 size, bool is_wchar, uint4 *pi)
{
  int8 n;

  if (*pi >= size)
    return -1;

  load_integer (p, *pi, &n, is_wchar ? 2 : 1, 0);
  (*pi) += is_wchar ? 2 : 1;

  return (int)n;
}

/*****************************************************************************/

static void check_format_string (POOL          *p,
                                 uint4         size,
                                 bool          is_wchar,
                                 EXPRESSIONS   *varying_list,
                                 ACCESS        access,       // ACCESS_READONLY, ACCESS_READWRITE
                                 TEXT_POSITION *pos)
{
  uint4 i;      // string pos
  int   argn;   // count arg
  int   c;
  int   flags;
  LIST_OF_EXPRESSIONS *list;

#define NEXT_CHAR  next_wchar(p,size,is_wchar,&i)

  i = 0;
  argn = 0;
  list = varying_list->head;

  for (;;)
  {
    c = NEXT_CHAR;
    if (c < 0)
      break;

    if (c != (int)'%')     // skip anything else than %
      continue;

    c = NEXT_CHAR;
    if (c == (int)'%')                  // means %%
      continue;

    flags = 0;

    if (access == ACCESS_READONLY)
    {
      // flags

      while (c == (int)'-' || c == (int)'+' || c == (int)'0')
      {
        if (c == (int)'-')
        {
          flags |= _FMINUS;
          c = NEXT_CHAR;
        }
        else if (c == (int)'+')
        {
          flags |= _FPLUS;
          c = NEXT_CHAR;
        }
        else if (c == (int)'0')
        {
          flags |= _FZERO;
          c = NEXT_CHAR;
        }
      }


      // width

      if (c == (int)'*')
      {
        flags |= _DYN_WIDE;
        c = NEXT_CHAR;
      }
      else if (c >= (int)'0' && c <= (int)'9')
      {
        while (c >= (int)'0' && c <= (int)'9')
          c = NEXT_CHAR;
      }


      // precision

      if (c == (int)'.')
      {
        c = NEXT_CHAR;

        if (c == (int)'*')
        {
          flags |= _DYN_PRECISION;
          c = NEXT_CHAR;
        }
        else if (c >= (int)'0' && c <= (int)'9')
        {
          while (c >= (int)'0' && c <= (int)'9')
            c = NEXT_CHAR;
        }
      }



      if (flags & _DYN_WIDE)    // swallow extra int parameter
      {
        if (check_arg ('*', 0, &argn, &list, pos) < 0)
          return;
      }


      if (flags & _DYN_PRECISION)    // swallow extra int parameter
      {
        if (check_arg ('*', 0, &argn, &list, pos) < 0)
          return;
      }


      // type

      if (c == (int)'d' || c == (int)'u' || c == (int)'x' || c == (int)'f' ||
          c == (int)'e' ||
          c == (int)'c' || c == (int)'C' || c == (int)'s' || c == (int)'S')
      {
        if (check_arg ((char)c, flags, &argn, &list, pos) < 0)
          return;
      }
      else
      {
        semantic_error ("illegal format string : bad type", pos);
        continue;
      }
    }
    else   // out / ref parameter
    {

      // skip par

      if (c == (int)'*')
      {
        flags |= _SKIP_PAR;
        c = NEXT_CHAR;
      }


      // width

      if (c >= (int)'0' && c <= (int)'9')
      {
        while (c >= (int)'0' && c <= (int)'9')
          c = NEXT_CHAR;
      }


      // type

      if (c == (int)'d' || c == (int)'u' || c == (int)'x' || c == (int)'f' ||
          c == (int)'c' || c == (int)'C' || c == (int)'s' || c == (int)'S')
      {
        if ((flags & _SKIP_PAR) == 0)
        {
          if (check_arg ((char)c, 0, &argn, &list, pos) < 0)
            return;
        }
      }
      else
      {
        char msg[128];
        sprintf (msg, "illegal format string : bad type %%%c", c);
        semantic_error (msg, pos);
        continue;
      }
    }
  }


  if (argn < varying_list->count)
    semantic_error ("format string : too many arguments", pos);
}

/*****************************************************************************/

static PEXPRESSION box_into_object (PEXPRESSION e, MODE par_mode, TEXT_POSITION *pos)
{
  PEXPRESSION n;


  // case 1 : it's a constant expression

  if (is_constant_exp (e))
  {
    int rc, size, align;

    // returns 0 if OK, -1 if bad type, -2 if size is too large, -3 if other constraint
    rc = size_and_alignment_of_constant_size_exp (e, &size, &align);
    if (rc == 0)
    {
      // create a new pool object of type byte[]

      POOL *p;

      n = new_exp (A_POOL_CONSTANT);
      n->base_type_or_null = type_object;   // byte[]
      n->form = AN_OBJECT;
      n->access = ACCESS_CONSTANT;

      n->constraint.kind = CONSTANT_CONSTRAINT;
      n->constraint.value = size;

      p = new_pool_constant (size, align);

      n->info.pool_constant_info.pool_cte = p;

      simple_store_constant_into_pool (e, size, p);

      free_exp (e);

      return n;
    }
    else if (rc == -1)
      semantic_error ("type has no size", pos);
    else if (rc == -2)
      semantic_error ("size is too large", pos);
    else
      fatal_compiler_error ("parameter: cannot compute size", pos);

    return e;   // default after error
  }


  // case 2 : it has already type object

  if (types_are_equal (e->base_type_or_null, type_object))
    return e;


  // case 3 : it has a constant size

  if (e->constraint.kind == DOES_NOT_APPLY ||
      e->constraint.kind == CONSTANT_CONSTRAINT)
  {
    int rc, size, align;

    // returns 0 if OK, -1 if bad type, -2 if size is too large, -3 if other constraint
    rc = size_and_alignment_of_constant_size_exp (e, &size, &align);
    if (rc == 0)
    {
      // compute actual parameter size and fill constraint

      n = new_exp (A_BOXED_OBJECT);
      n->base_type_or_null = type_object;
      n->form = AN_OBJECT;
      n->access = e->access;
      n->info.boxed_object_info.parameter = e;

      n->constraint.kind = CONSTANT_CONSTRAINT;
      n->constraint.value = size;

      return n;
    }
    else if (rc == -2)
      semantic_error ("size is too large", pos);
    else
      fatal_compiler_error ("parameter: cannot compute size", pos);

    return e;   // default after error
  }


  // case 4 : it has a runtime size

  n = new_exp (A_BOXED_OBJECT);
  n->base_type_or_null = type_object;
  n->form = AN_OBJECT;
  n->access = e->access;
  n->info.boxed_object_info.parameter = e;

  n->constraint.kind = RUNTIME_CONSTRAINT;

  return n;
}

/*****************************************************************************/

//  actual_parameter ::=       [identifier "=>"] expression
//                     | "ref" [identifier "=>"] object_designator
//                     | "out" [identifier "=>"] object_designator

static void parse_actual_parameter (PENTITY     param,        // formal parameter (A_PARAMETER)
                                    bool        param_valid,
                                    bool        in_varying_list,
                                    EXPRESSIONS *list,        // actual parameter must be appended to list
                                    int         varying_list_count)
{
  MODE          mode, par_mode;
  PEXPRESSION   e;
  CONTEXT       context;
  TEXT_POSITION mode_pos, pos;
  PENTITY       formal_parameter_type, basetype_before_boxing;

  // mode

  memcpy (&mode_pos, &token.pos, sizeof(mode_pos));

  if (token.kind == TOKEN_ref)
  {
    get_token();   // skip
    mode = MODE_REF;
  }
  else if (token.kind == TOKEN_out)
  {
    get_token();   // skip
    mode = MODE_OUT;
  }
  else
  {
    mode = MODE_IN;
  }

  if (param_valid)
  {
    par_mode = param->info->the_parameter.mode;
    if (mode != par_mode)
    {
      if (par_mode == MODE_OUT)
        semantic_error ("keyword 'out' expected", &mode_pos);
      else if (par_mode == MODE_REF)
        semantic_error ("keyword 'ref' expected", &mode_pos);
      else
        semantic_error ("no parameter mode allowed here", &mode_pos);
    }
  }


  // identifier

  if (token.kind == IDENTIFIER)
  {
    get_look_ahead_token ();
    if (look_ahead_token.kind == ARROW)
    {
      if (param_valid && wstrcmp (token.info._identifier.value, param->identifier_or_null) != 0)
      {
        char id[MAX_IDENTIFIER_LENGTH + 1];
        char msg[MAX_IDENTIFIER_LENGTH + 128];

        wstring_to_string (param->identifier_or_null, id);
        sprintf (msg, "parameter '%s' expected", id);

        semantic_error (msg, &token.pos);
      }

      get_token();   // skip identifier
      get_token();   // skip "=>"
    }
  }


  memset (&context, 0x00, sizeof(context));

  if (param_valid)
  {
    if (in_varying_list)   // type object required
    {
      formal_parameter_type = type_object;  // arbitrary choice - maybe changed later into type_array_of_object
    }
    else
    {
      formal_parameter_type = param->info->the_parameter.type;
    }

    convert_type_into_context (formal_parameter_type, &context);
  }


  memcpy (&pos, &token.pos, sizeof(pos));

  if (mode == MODE_IN)
  {
    parse_expression (&context,
                      &token.pos,
                      null,
                      &e);
  }
  else
  {
    parse_unary_expression (&context,
                            &token.pos,
                            null,
                            &e);
  }


  // box or unbox actual parameter, if necessary

  if (param_valid)
  {
    if (in_varying_list)   // VARYING ARGUMENT LIST : type object[] or packed type required
    {
      promote_parameter (e);

      basetype_before_boxing = e->base_type_or_null;

      if (e->base_type_or_null != null)
      {
        // check actual type and set formal type

        if (varying_list_count == 0)  // first varying parameter
        {
          // make sure actual parameter has either type array_of_object or is packed

          if (types_are_equal (e->base_type_or_null, type_array_of_object))
          {
            formal_parameter_type = type_array_of_object;
            convert_type_into_context (formal_parameter_type, &context);  // fix context
          }
          else if (is_packed_type (e->base_type_or_null))
          {
            /* ok */
          }
          else
          {
            semantic_error ("parameter must have type object[] or have a packed type", &pos);
          }
        }
        else   // not first one
        {
          // make sure actual parameter is packed

          if (!is_packed_type (e->base_type_or_null))
            semantic_error ("parameter must have packed type", &pos);
        }


        if (is_packed_type (e->base_type_or_null))
        {
          e = box_into_object (e, par_mode, &pos);
        }
      }
    }
    else if (types_are_equal (context.base_type_or_null, type_object))   // BOXING (packed type -> type object)
    {
      // formal parameter has type byte[] or byte[N]

      promote_parameter (e);

      basetype_before_boxing = e->base_type_or_null;

      if (e->base_type_or_null != null)
      {
        if (!is_packed_type (e->base_type_or_null))
          semantic_error ("parameter must have packed type", &pos);
        else
          e = box_into_object (e, par_mode, &pos);
      }
    }
    else if (e->base_type_or_null != null && types_are_equal (e->base_type_or_null, type_object))  // UNBOXING
    {
      // actual parameter has type object (= byte[]), formal parameter has packed type

      basetype_before_boxing = e->base_type_or_null;


      // make sure formal parameter is packed and is not open struct

      if (context.base_type_or_null != null)
      {
        if (!is_packed_type (context.base_type_or_null))
        {
          semantic_error ("parameter has incompatible type", &pos);
        }
        else
        {
          // make sure formal parameter is not an open struct

          if (context.base_type_or_null->kind == A_STRUCT_TYPE &&
              context.base_type_or_null->info->the_struct_type.is_open_type)
          {
            semantic_error ("a byte[] parameter is not compatible with an open struct type", &pos);
          }
        }
      }



      if (context.constraint.kind == UNCONSTRAINED)  // formal parameter has unconstrained type
      {
        PENTITY array_type, element_type;
        int     rc, element_size, align;


        // formal param is open array (or struct but not allowed)

        if (context.base_type_or_null->kind == AN_OPEN_ARRAY_TYPE &&
            is_packed_type (context.base_type_or_null))
        {
          array_type = context.base_type_or_null;
          element_type = complete_type_of (array_type->info->the_open_array_type.element);

          // returns 0 if OK, -1 if bad type, -2 if size is too large
          rc = size_and_alignment_of_type (element_type, &element_size, &align);
          if (rc < 0)
            fatal_compiler_error ("parameter element_size", &token.pos);
        }
        else
        {
          element_size = 1;
        }


        if (element_size == 0)
        {
          semantic_error ("formal array element size cannot be zero", &pos);
          element_size = 1;
        }

        if (e->constraint.kind == CONSTANT_CONSTRAINT &&
            e->constraint.value % element_size != 0)
        {
          semantic_error ("parameter size is truncated", &pos);
        }

        if (e->constraint.kind == RUNTIME_CONSTRAINT && element_size > 1)
        {
          warning ("risk of parameter size truncation", &pos);
        }


        if (e->kind == A_POOL_CONSTANT)    // actual parameter is pool cte of type byte[]
        {
          int         length, size;
          PEXPRESSION n;
          POOL        *p;

          length = e->constraint.value / element_size;
          size = length * element_size;

          n = new_exp (A_POOL_CONSTANT);
          n->form = AN_OBJECT;
          n->access = ACCESS_CONSTANT;
          n->base_type_or_null = context.base_type_or_null;
          n->constraint.kind = CONSTANT_CONSTRAINT;
          n->constraint.value = length;

          p = new_pool_constant (size, align_of_pool_cte(e->info.pool_constant_info.pool_cte));
          copy_pool_to_pool (e->info.pool_constant_info.pool_cte, 0, p, 0, size);

          n->info.pool_constant_info.pool_cte = p;

          e = n;
        }
        else       // actual parameter is NOT a pool constant
        {
          PEXPRESSION n;

          n = new_exp (AN_UNBOXED_OBJECT);
          n->base_type_or_null = context.base_type_or_null;
          n->form = AN_OBJECT;
          n->access = e->access;
          n->info.unboxed_object_info.parameter = e;
          n->info.unboxed_object_info.to_open_array = true;

          if (e->constraint.kind == CONSTANT_CONSTRAINT)
          {
            n->constraint.kind = CONSTANT_CONSTRAINT;
            n->constraint.value = e->constraint.value / element_size;
          }
          else
          {
            n->constraint.kind = RUNTIME_CONSTRAINT;
          }

          e = n;
        }
      }
      else      // formal parameter has no constraint or constant constraint
      {
        int rc, size, align;

        if (is_packed_type (formal_parameter_type))
        {
          // returns 0 if OK, -1 if bad type, -2 if size is too large
          rc = size_and_alignment_of_type (formal_parameter_type, &size, &align);
          if (rc < 0)
            fatal_compiler_error ("parameter conversion2", &token.pos);

          if (e->constraint.kind == CONSTANT_CONSTRAINT &&
              e->constraint.value != (uint4)size)
          {
            semantic_error ("parameter size does not match", &pos);
            size = 0;
          }
        }
        else   // earlier error
        {
          size = 0;
        }


        if (e->kind == A_POOL_CONSTANT)    // actual parameter is pool cte of type byte[]
        {
          PEXPRESSION n;

          // convert pool constant into packed type

          load_constant_object_from_pool (e->info.pool_constant_info.pool_cte,
                                          0,
                                          size,
                                          formal_parameter_type,
                                          &n);
          free_exp (e);

          e = n;
        }
        else    // convert runtime byte[] actual parameter into constant-size packed type
        {
          PEXPRESSION n;

          n = new_exp (AN_UNBOXED_OBJECT);
          n->base_type_or_null = context.base_type_or_null;
          memcpy (&n->constraint, &context.constraint, sizeof(CONSTRAINT));
          n->form = AN_OBJECT;
          n->access = e->access;
          n->info.unboxed_object_info.parameter = e;
          n->info.unboxed_object_info.to_open_array = false;

          e = n;
        }
      }
    }
    else
    {
      // normal case : no boxing/unboxing

      basetype_before_boxing = e->base_type_or_null;
    }



    // check that parameters are compatible

    if (par_mode == MODE_IN)
    {
      (void)check_assignment_context_compatibility (&context, e, &pos);
    }
    else
    {
      (void)check_ref_parameter_context_compatibility (&context, e, &pos);
    }


    // check mode of expression (constants are not allowed for modes out and ref)

    if (par_mode != MODE_IN)      // out or ref
    {
      if (e->access != ACCESS_READWRITE)
        semantic_error ("a read/write object is expected here", &pos);
    }


    // append node to list

    append_expression (list, e, basetype_before_boxing, param);
  }
}

/*****************************************************************************/

static void skip_to_next_param (PENTITY *pparam)
{
  while (*pparam != null && (*pparam)->kind != A_PARAMETER)
    *pparam = (*pparam)->next;
}

/*****************************************************************************/

// function_call ::= name  "("  actual_parameters  ")"
//
// actual_parameters ::= [actual_parameter {"," actual_parameter}]

static void parse_function_call (TEXT_POSITION *prefix_pos,
                                 PEXPRESSION   prefix,
                                 PEXPRESSION   *pout)
{
  PENTITY       ftype, param, next_param;
  bool          param_valid, in_varying_list;
  EXPRESSIONS   list, varying_list;
  TEXT_POSITION pos;
  PEXPRESSION   e;

  param = null;

  param_valid = false;
  in_varying_list = false;
  clear (list);
  clear (varying_list);


  ftype = prefix->base_type_or_null;

  if (ftype != null)
  {
    if (ftype->kind == A_FUNCTION_POINTER_TYPE)
    {
      if (ftype->info->the_function_pointer_type.is_callback)
      {
        if (!lexa.within_unsafe)
          semantic_error ("a callback function can only be called within an unsafe region", &token.pos);
      }

      if (ftype->info->the_function_pointer_type.extern_dll_or_null != null)
      {
        if (!lexa.within_unsafe)
          semantic_error ("an extern function can only be called within an unsafe region", &token.pos);
      }

      param = ftype->info->the_function_pointer_type.parameters->entities.first;
      skip_to_next_param (&param);
      param_valid = true;
    }
    else
    {
      semantic_error ("prefix must denote a function", &token.pos);
      ftype = null;
    }
  }

  if (prefix->kind == A_CONST_NULL_VALUE)
  {
    semantic_error ("calling null pointer function", &token.pos);
    ftype = null;
  }

  get_token();   // skip '('

  memcpy (&pos, &token.pos, sizeof(pos));

  if (token.kind != RIGHT_PARENTHESIS)
  {
    for (;;)
    {
      if (param_valid)
      {
        if (param == null)
        {
          semantic_error ("too many parameters", &token.pos);
          param_valid = false;
        }
        else if (!in_varying_list)
        {
          // compute next_param
          next_param = param->next;
          skip_to_next_param (&next_param);

          // check if param denotes a varying parameter list
          if (next_param == null &&
              types_are_equal (base_type_of (param->info->the_parameter.type), type_array_of_object))
          {
            in_varying_list = true;
          }
        }
      }

      parse_actual_parameter (param,
                              param_valid,
                              in_varying_list,
                              in_varying_list ? &varying_list : &list,
                              varying_list.count);

      if (param_valid && (!in_varying_list))
        param = next_param;


      if (token.kind == RIGHT_PARENTHESIS)
        break;

      if (token.kind != COMMA)
      {
        syntax_error ("',' or ')' expected", &token.pos);
        break;
      }

      get_token();   // skip comma
    }
  }

  if (token.kind == RIGHT_PARENTHESIS)
    get_token();   // skip ')'


  // check that all formal parameters were processed

  if (param_valid)
  {
    if (param != null)  // some formal parameters were not processed, or varying-list must be closed
    {
      if (in_varying_list)     // a varying-list with at least 1 parameter is being processed and must be closed
      {
        PEXPRESSION e;
        PENTITY     formal_parameter_type;
        CONTEXT     context;

        formal_parameter_type = param->info->the_parameter.type;
        convert_type_into_context (formal_parameter_type, &context);

        if (varying_list.count == 1 &&
            varying_list.head->exp->base_type_or_null != null &&
            types_are_equal (varying_list.head->exp->base_type_or_null, type_array_of_object))
        {
          append_expression (&list, varying_list.head->exp, type_array_of_object, varying_list.head->e);
        }
        else   // a varying list of packed types
        {
          // otherwise append an extra parameter to list : a A_BOXED_ARRAY_OBJECT to 'varying_list'
          e = new_exp (A_BOXED_ARRAY_OBJECT);
          e->base_type_or_null = type_array_of_object;
          e->constraint.kind = CONSTANT_CONSTRAINT;
          e->constraint.value = varying_list.count;

          if (context.constraint.kind == CONSTANT_CONSTRAINT &&
              context.constraint.value != (uint4)varying_list.count)
            semantic_error ("number of parameters does not match object[]", &pos);

          e->form = AN_OBJECT;
          e->access = (param->info->the_parameter.mode == MODE_IN) ? ACCESS_READONLY : ACCESS_READWRITE;
          e->info.boxed_array_object_info.list  = varying_list.head;

          append_expression (&list, e, type_array_of_object, param);
        }
      }
      else   // check if all remaining formal parameters have default expressions
      {      // the last parameter possibly being a varying parameter list

        while (param != null && param->info->the_parameter.default_value_or_null != null)
        {
          // append a copy of the default expression to the actual parameter list.

          append_expression (&list,
                             duplicate_constant_exp (param->info->the_parameter.default_value_or_null),
                             param->info->the_parameter.type,
                             param);

          param = param->next;
          skip_to_next_param (&param);
        }


        // check if there's an additional last parameter of type object[] -> if yes, add an empty table

        if (param != null)
        {
          PENTITY     formal_parameter_type;
          CONTEXT     context;

          formal_parameter_type = param->info->the_parameter.type;
          convert_type_into_context (formal_parameter_type, &context);

          next_param = param->next;
          skip_to_next_param (&next_param);

          // check if param denotes a varying parameter list
          if (next_param == null &&
              types_are_equal (base_type_of (formal_parameter_type), type_array_of_object))
          {
            PEXPRESSION e;

            // otherwise append an extra parameter to list : a A_BOXED_ARRAY_OBJECT to an empty 'varying_list'
            e = new_exp (A_BOXED_ARRAY_OBJECT);
            e->base_type_or_null = type_array_of_object;
            e->constraint.kind = CONSTANT_CONSTRAINT;
            e->constraint.value = 0;

            if (context.constraint.kind == CONSTANT_CONSTRAINT &&
                context.constraint.value != 0)
              semantic_error ("number of parameters does not match object[]", &pos);

            e->form = AN_OBJECT;
            e->access = (param->info->the_parameter.mode == MODE_IN) ? ACCESS_READONLY : ACCESS_READWRITE;
            e->info.boxed_array_object_info.list  = null;

            append_expression (&list, e, type_array_of_object, param);


            param = param->next;
            skip_to_next_param (&param);
          }
        }


        if (param != null)
          semantic_error ("missing parameters", &token.pos);
      }
    }
  }


  // check matching of format string

  if (list.tail != null &&
      list.tail->exp->kind == A_BOXED_ARRAY_OBJECT &&   // varying list
      list.tail->prev != null)
  {
    LIST_OF_EXPRESSIONS *format;
    PENTITY             type;
    POOL                *p;
    uint4               size;

    format = list.tail->prev;
    type = format->exp->base_type_or_null;

    if (format->exp->kind == A_POOL_CONSTANT &&
        type != null &&
        (types_are_equal (type, type_string) ||
         types_are_equal (type, type_wstring)))
    {
      // check if format string matches with varying_list

      p    = format->exp->info.pool_constant_info.pool_cte;
      size = size_of_pool_cte (p);

      check_format_string (p,
                           size,
                           types_are_equal (type, type_wstring),
                           &varying_list,
                           list.tail->exp->access,
                           &pos);
    }
  }


  // create function call

  e = new_exp (A_FUNCTION_CALL);

  if (ftype != null)
    e->base_type_or_null = base_type_of (ftype->info->the_function_pointer_type.return_type);

  e->form = A_VALUE;
  e->access = ACCESS_READONLY;
  e->info.function_call_info.func = prefix;
  e->info.function_call_info.param = list.head;

  *pout = e;
}

/*****************************************************************************/

static void count_parameters (PENTITY param, int *pcount, PENTITY *pparam)
{
  PENTITY e = param;
  int     count = 0;

  while (e != null)
  {
    if (e->kind == A_PARAMETER)
    {
      *pparam = e;
      count++;
    }
    e = e->next;
  }

  *pcount = count;
}

/*****************************************************************************/

static bool is_fixed_object (LIST_OF_EXPRESSIONS *list)
{
  PEXPRESSION e;

  if (list == null)
    return true;

  e = list->exp;

  while (e != null && (e->kind == AN_ARRAY_ELEMENT_OBJECT ||
                       e->kind == AN_ARRAY_SLICE_OBJECT ||
                       e->kind == A_STRUCT_FIELD_OBJECT))
  {
    if (e->kind == AN_ARRAY_ELEMENT_OBJECT)
      e = e->info.array_element_object_info.prefix;

    if (e->kind == AN_ARRAY_SLICE_OBJECT)
      e = e->info.array_slice_object_info.prefix;

    if (e->kind == A_STRUCT_FIELD_OBJECT)
        e = e->info.struct_field_object_info.prefix;
  }

  if (e != null)
    return (e->kind == A_POOL_CONSTANT || e->kind == A_GLOBAL_VARIABLE_OBJECT);

  return true;  // default in case of earlier error
}

/*****************************************************************************/

// run_call ::= "run"  function_call

static void parse_run_call (PEXPRESSION *pout)
{
  TEXT_POSITION                  pos;
  PEXPRESSION                    prefix, exp, exp2;
  PENTITY                        ftype, param1, t;
  A_FUNCTION_POINTER_TYPE_ENTITY *pfunc;
  int                            count;

  get_token();    // skip token 'run'

  memcpy (&pos, &token.pos, sizeof(pos));
  parse_name (&no_context, false, &token.pos, null, &exp);

  if (exp->kind != A_FUNCTION_CALL)
    semantic_error ("a function call is expected here", &pos);
  else
  {
    // get prefix (a value of function pointer)
    prefix = exp->info.function_call_info.func;
    if (prefix != null)
    {
      ftype = prefix->base_type_or_null;
      if (ftype != null && ftype->kind == A_FUNCTION_POINTER_TYPE)
      {
        pfunc = &ftype->info->the_function_pointer_type;
        if (pfunc->return_type != type_void)
          semantic_error ("a function returning void is expected here", &pos);

        if (pfunc->extern_dll_or_null != null)
          semantic_error ("a function with option extern is not allowed here", &pos);

        if (pfunc->is_callback)
          semantic_error ("a callback function is not allowed here", &pos);

        count_parameters (pfunc->parameters->entities.first, &count, &param1);

        if (count > 1)
          semantic_error ("only 1 parameter can be passed to a thread", &pos);

        if (count == 1)
        {
          t = complete_type_of (param1->info->the_parameter.type);

          if (t != null)
          {
            if (is_open_type (t))
              semantic_error ("an open-type formal parameter is not allowed for a thread function", &pos);
            else
            {
              t = base_type_of (t);

              if (param1->info->the_parameter.mode == MODE_IN)
              {
                if ((t->kind == AN_INTEGER_TYPE && t != type_long) ||
                    t->kind == AN_ENUMERATION_TYPE ||
                    t->kind == A_POINTER_TYPE ||
                    t->kind == A_FUNCTION_POINTER_TYPE ||
                    t->kind == AN_UNSAFE_POINTER_TYPE)
                {
                  // ok
                }
                else if (t->kind == AN_OPEN_ARRAY_TYPE ||
                         t->kind == A_STRUCT_TYPE ||
                         t->kind == A_UNION_TYPE)
                {
                  if (!is_fixed_object (exp->info.function_call_info.param))
                    semantic_error ("actual parameter of thread function must be a constant or a global variable", &pos);
                }
                else
                {
                  semantic_error ("formal parameter of thread function has bad type", &pos);
                }
              }
              else     // mode out / ref
              {
                if (!is_fixed_object (exp->info.function_call_info.param))
                  semantic_error ("actual parameter of thread function must be a global variable", &pos);
              }
            }
          }
        }
      }
    }
  }

  exp2 = new_exp (A_RUN_CALL);
  exp2->base_type_or_null = type_int;
  exp2->form = A_VALUE;
  exp2->access = ACCESS_READONLY;
  exp2->info.run_call_info.function_call = exp;

  *pout = exp2;
}

/*****************************************************************************/

//  attribute_id ::= identifier | "string" | "byte"

static ATTRIBUTE parse_attr_id ()
{
  int       i;
  ATTRIBUTE attr;

  if (token.kind == IDENTIFIER)
  {
    for (i=0; attribute_info[i].attr!=last_ATTR; i++)
    {
      if (wstrcmp (attribute_info[i].id, token.info._identifier.value) == 0)
        break;
    }

    attr = attribute_info[i].attr;

    if (attr == last_ATTR)
      syntax_error ("unknown attribute", &token.pos);
  }
  else if (token.kind == TOKEN_byte)
  {
    attr = ATTR_byte;
  }
  else if (token.kind == TOKEN_string)
  {
    attr = ATTR_string;
  }
  else
  {
    syntax_error ("attribute identifier expected", &token.pos);
    return last_ATTR;
  }

  get_token();

  return attr;
}

/*****************************************************************************/

//  attribute    ::= "'" attribute_id ["(" constant_expression ")"]

static int parse_array_dimension (void)
{
  PEXPRESSION   e;
  TEXT_POSITION pos;
  int8          dim;

  if (token.kind != LEFT_PARENTHESIS)
    return 1;

  get_token();   // skip '('
  memcpy (&pos, &token.pos, sizeof(pos));

  parse_expression (&no_context, &token.pos, null, &e);

  if (e->kind != A_CONST_INTEGER_VALUE)
  {
    semantic_error ("an integer constant is expected here", &pos);
    dim = 1;
  }
  else
  {
    dim = e->info.const_integer_value_info.value;

    if (dim < 1 || dim > (1<<30))
    {
      semantic_error ("array dimension is out of valid range", &pos);
      dim = 1;
    }

    if (e->base_type_or_null == type_long)
    {
      semantic_error ("an int or uint constant is expected here", &pos);
    }
  }

  if (token.kind == RIGHT_PARENTHESIS)   // ')'
    get_token();   // skip ')'
  else
    syntax_error ("')' expected", &token.pos);

  return (int)dim;
}

/*****************************************************************************/

static void parse_type_attribute (PENTITY type_name, PEXPRESSION *pout)
{
  PENTITY       type;
  ATTRIBUTE     attr;
  TEXT_POSITION pos;
  PEXPRESSION   n;

  if (token.kind != APOSTROPHE)
  {
    syntax_error ("apostrophe (attribute) expected", &token.pos);
    *pout = dummy_expression ();
    return;
  }

  memcpy (&pos, &token.pos, sizeof(pos));
  get_token();    // skip apostrophe

  attr = parse_attr_id ();
  if (attr == last_ATTR)
  {
    *pout = dummy_expression ();
    return;
  }

  type = complete_type_of (type_name);

  switch (attr)
  {
    case ATTR_min:
    case ATTR_max:
    {
      INTEGER_TYPE it;

      if (type->kind != AN_INTEGER_TYPE)
      {
        semantic_error ("prefix must denote an integer type", &pos);
        *pout = dummy_expression ();
        return;
      }

      it = type->info->the_integer_type.type;

      n = new_exp (A_CONST_INTEGER_VALUE);
      n->base_type_or_null = type;
      n->form = A_VALUE;
      n->access = ACCESS_CONSTANT;
      n->info.const_integer_value_info.value =
          (attr == ATTR_min) ? integer_info[it].min : integer_info[it].max;
    }
    break;

    case ATTR_first:
    case ATTR_last:
    {
      if (type->kind != AN_ENUMERATION_TYPE)
      {
        semantic_error ("prefix must denote an enumeration type", &pos);
        *pout = dummy_expression ();
        return;
      }

      n = new_exp (A_CONST_ENUMERATION_VALUE);
      n->base_type_or_null = type;
      n->form = A_VALUE;
      n->access = ACCESS_CONSTANT;
      n->info.const_enumeration_value_info.value =
           (attr == ATTR_first) ? 0 : type->info->the_enumeration_type.last;
    }
    break;

    case ATTR_length:
    {
      PENTITY t;
      int     dim, i;

      if (type->kind != AN_ARRAY_TYPE)
      {
        semantic_error ("prefix must denote an array", &pos);
        *pout = dummy_expression ();
        return;
      }

      dim = parse_array_dimension();
      t = type;

      for (i=2; i<=dim; i++)
      {
        t = complete_type_of (t->info->the_array_type.open_array);

        if (t == null || t->kind != AN_OPEN_ARRAY_TYPE)
        {
          t = null;
          break;
        }

        t = complete_type_of (t->info->the_open_array_type.element);

        if (t == null || t->kind != AN_ARRAY_TYPE)
        {
          t = null;
          break;
        }
      }

      if (t == null)
      {
        semantic_error ("prefix must denote an N-dimensional array", &pos);
        *pout = dummy_expression ();
        return;
      }

      n = new_exp (A_CONST_INTEGER_VALUE);
      n->base_type_or_null = type_int;
      n->form = A_VALUE;
      n->access = ACCESS_CONSTANT;
      n->info.const_integer_value_info.value = t->info->the_array_type.length;
    }
    break;

    case ATTR_byte:
    {
      semantic_error ("attribute 'byte requires an object as prefix, not a type", &pos);
      *pout = dummy_expression ();
      return;
    }
    break;

    case ATTR_string:
    {
      semantic_error ("attribute 'string requires a value as prefix, not a type", &pos);
      *pout = dummy_expression ();
      return;
    }
    break;

    case ATTR_size:
    {
      int rc, size, align;

      if (lexa.within_unsafe)
      {
        if (!is_type_having_size (type_name))
        {
          semantic_error ("prefix must not denote a renamed/incomplete/generic/opaque type", &pos);
          *pout = dummy_expression ();
          return;
        }
      }
      else
      {
        if (!is_packed_type (type_name))
        {
          semantic_error ("prefix must denote a packed type", &pos);
          *pout = dummy_expression ();
          return;
        }
      }

      if (is_open_type (type_name))
      {
        semantic_error ("prefix must not denote an open type", &pos);
        *pout = dummy_expression ();
        return;
      }

      // returns 0 if OK, -1 if bad type, -2 if size is too large
      rc = size_and_alignment_of_type (type_name, &size, &align);
      if (rc == -1)
      {
        semantic_error ("size is not constant", &pos);
        *pout = dummy_expression ();
        return;
      }
      else if (rc == -2)
      {
        semantic_error ("size is too large", &pos);
        *pout = dummy_expression ();
        return;
      }

      n = new_exp (A_CONST_INTEGER_VALUE);
      n->base_type_or_null = type_uint;
      n->form = A_VALUE;
      n->access = ACCESS_CONSTANT;
      n->info.const_integer_value_info.value = size;
    }
    break;

    default:
      fatal_compiler_error ("parse_type_attribute", &token.pos);
  }

  *pout = n;
}

/*****************************************************************************/

static void parse_attribute (PEXPRESSION prefix, PEXPRESSION *pout)
{
  PENTITY       type;
  ATTRIBUTE     attr;
  TEXT_POSITION pos;
  PEXPRESSION   n;

  if (token.kind != APOSTROPHE)
  {
    syntax_error ("apostrophe (attribute) expected", &token.pos);
    *pout = dummy_expression ();
    return;
  }

  memcpy (&pos, &token.pos, sizeof(pos));
  get_token();    // skip apostrophe

  attr = parse_attr_id ();
  if (attr == last_ATTR)
  {
    *pout = dummy_expression ();
    return;
  }

  type = prefix->base_type_or_null;

  if (type == null)    // earlier error occured
  {
    *pout = dummy_expression ();
    return;
  }

  switch (attr)
  {
    case ATTR_min:
    case ATTR_max:
    {
      INTEGER_TYPE it;

      if (type->kind != AN_INTEGER_TYPE || prefix->form != AN_OBJECT)
      {
        semantic_error ("prefix of attribute 'min/max must denote an integer variable", &pos);
        free_exp (prefix);
        *pout = dummy_expression ();
        return;
      }

      if (prefix->kind == A_CONST_INTEGER_VALUE)
        semantic_error ("only variables are allowed as prefix for this attribute", &pos);

      it = type->info->the_integer_type.type;

      n = new_exp (A_CONST_INTEGER_VALUE);
      n->base_type_or_null = type;
      n->form = A_VALUE;
      n->access = ACCESS_CONSTANT;
      n->info.const_integer_value_info.value =
          (attr == ATTR_min) ? integer_info[it].min : integer_info[it].max;

      free_exp (prefix);
    }
    break;

    case ATTR_first:
    case ATTR_last:
    {
      if (type->kind != AN_ENUMERATION_TYPE || prefix->form != AN_OBJECT)
      {
        semantic_error ("prefix of attribute first/last must denote an enumeration variable", &pos);
        free_exp (prefix);
        *pout = dummy_expression ();
        return;
      }

      if (prefix->kind == A_CONST_ENUMERATION_VALUE)
        semantic_error ("only variables are allowed as prefix for this attribute", &pos);

      n = new_exp (A_CONST_ENUMERATION_VALUE);
      n->base_type_or_null = type;
      n->form = A_VALUE;
      n->access = ACCESS_CONSTANT;
      n->info.const_enumeration_value_info.value =
           (attr == ATTR_first) ? 0 : type->info->the_enumeration_type.last;

      free_exp (prefix);
    }
    break;

    case ATTR_length:
    {
      int dim;

      if (type->kind != AN_OPEN_ARRAY_TYPE || prefix->form != AN_OBJECT)
      {
        semantic_error ("prefix of attribute 'length must denote an array object", &pos);
        free_exp (prefix);
        *pout = dummy_expression ();
        return;
      }

      dim = parse_array_dimension();

      if (dim == 1)
      {
        if (prefix->constraint.kind == CONSTANT_CONSTRAINT)
        {
          n = new_exp (A_CONST_INTEGER_VALUE);
          n->base_type_or_null = type_int;
          n->form = A_VALUE;
          n->access = ACCESS_CONSTANT;
          n->info.const_integer_value_info.value = prefix->constraint.value;

          free_exp (prefix);
        }
        else     // runtime constraint
        {
          n = new_exp (AN_OPERATOR_VALUE);
          n->base_type_or_null = type_int;
          n->form = A_VALUE;
          n->access = ACCESS_READONLY;
          n->info.operator_value_info.op = OP_LENGTH;
          n->info.operator_value_info.arg[0] = prefix;
        }
      }
      else   // dim > 1
      {
        PENTITY t;
        int     i;

        t = type;     // an open array type

        t = complete_type_of (t->info->the_open_array_type.element);
        if (t == null || t->kind != AN_ARRAY_TYPE)
          t = null;
        else
        {
          for (i=3; i<=dim; i++)
          {
            t = complete_type_of (t->info->the_array_type.open_array);

            if (t == null || t->kind != AN_OPEN_ARRAY_TYPE)
            {
              t = null;
              break;
            }

            t = complete_type_of (t->info->the_open_array_type.element);

            if (t == null || t->kind != AN_ARRAY_TYPE)
            {
              t = null;
              break;
            }
          }
        }

        if (t == null)
        {
          semantic_error ("prefix must denote an N-dimensional array", &pos);
          free_exp (prefix);
          *pout = dummy_expression ();
          return;
        }

        n = new_exp (A_CONST_INTEGER_VALUE);
        n->base_type_or_null = type_int;
        n->form = A_VALUE;
        n->access = ACCESS_CONSTANT;
        n->info.const_integer_value_info.value = t->info->the_array_type.length;

        free_exp (prefix);
      }
    }
    break;

    case ATTR_byte:
    {
      if (prefix->form != AN_OBJECT)
      {
        semantic_error ("prefix of attribute 'byte must denote an object", &pos);
        free_exp (prefix);
        *pout = dummy_expression ();
        return;
      }

      if (!is_packed_type (type))
      {
        semantic_error ("prefix of attribute 'byte must denote a packed object", &pos);
        free_exp (prefix);
        *pout = dummy_expression ();
        return;
      }


      if (is_constant_exp (prefix))
      {
        int  rc, size, align;
        POOL *p;

        // create a new pool object of type byte[]

        n = new_exp (A_POOL_CONSTANT);
        n->base_type_or_null = type_object;   // byte[]
        n->form = AN_OBJECT;
        n->access = ACCESS_CONSTANT;

        // returns 0 if OK, -1 if bad type, -2 if size is too large, -3 if missing/non-constant constraint
        rc = size_and_alignment_of_constant_size_exp (prefix, &size, &align);
        if (rc)
          fatal_compiler_error ("attr_byte : cannot compute size", &pos);

        n->constraint.kind = CONSTANT_CONSTRAINT;
        n->constraint.value = size;

        p = new_pool_constant (size, align);

        n->info.pool_constant_info.pool_cte = p;

        simple_store_constant_into_pool (prefix, size, p);   // simple type or pool constant

        free_exp (prefix);
      }
      else if (prefix->constraint.kind == DOES_NOT_APPLY ||
               prefix->constraint.kind == CONSTANT_CONSTRAINT)
      {
        int rc, size, align;

        // runtime with constant size

        n = new_exp (AN_ATTR_BYTE_OBJECT);
        n->base_type_or_null = type_object;   // byte[]
        n->form = AN_OBJECT;
        n->access = prefix->access;

        // returns 0 if OK, -1 if bad type, -2 if size is too large, -3 if missing/non-constant constraint
        rc = size_and_alignment_of_constant_size_exp (prefix, &size, &align);
        if (rc)
          fatal_compiler_error ("attr_byte : cannot compute size", &pos);

        n->constraint.kind = CONSTANT_CONSTRAINT;
        n->constraint.value = size;
        n->info.attr_byte_object_info.prefix = prefix;
      }
      else    // runtime with runtime size
      {
        n = new_exp (AN_ATTR_BYTE_OBJECT);
        n->base_type_or_null = type_object;   // byte[]
        n->form = AN_OBJECT;
        n->access = prefix->access;
        n->constraint.kind = RUNTIME_CONSTRAINT;
        n->info.attr_byte_object_info.prefix = prefix;
      }
    }
    break;

    case ATTR_string:
    {
      if (type->kind != AN_ENUMERATION_TYPE)
      {
        semantic_error ("prefix of attribute 'string must have an enumeration type", &pos);
        free_exp (prefix);
        *pout = dummy_expression ();
        return;
      }

      if (type == type_char || type == type_wchar)
      {
        semantic_error ("prefix of attribute 'string cannot have types char/wchar", &pos);
        free_exp (prefix);
        *pout = dummy_expression ();
        return;
      }

      if (prefix->kind == A_CONST_ENUMERATION_VALUE)
      {
        int   rc, size, align;
        uint4 r;

        r = prefix->info.const_enumeration_value_info.value;

        if (r > type->info->the_enumeration_type.last)
        {
          semantic_error ("prefix of attribute 'string is too large", &pos);
          r = 0;
        }

        // returns 0 if OK, -1 if bad type, -2 if size is too large
        rc = size_and_alignment_of_type (type_string, &size, &align);
        if (rc < 0)
          fatal_compiler_error ("attr_string", &token.pos);

        load_constant_object_from_pool (type->info->the_enumeration_type.string_table,
                                        /* offset = */ r * size,
                                        /* size   = */ size,
                                        type_string,
                                        &n);
      }
      else   // runtime value
      {
        PEXPRESSION  ex, idx;

        ex = new_exp (A_POOL_CONSTANT);
        ex->form   = AN_OBJECT;
        ex->access = ACCESS_CONSTANT;
        ex->base_type_or_null = type_array_of_string;
        ex->constraint.kind = CONSTANT_CONSTRAINT;
        ex->constraint.value = type->info->the_enumeration_type.last + 1;
        ex->info.pool_constant_info.pool_cte = type->info->the_enumeration_type.string_table;

        idx = new_exp (AN_OPERATOR_VALUE);
        idx->form   = A_VALUE;
        idx->access = ACCESS_READONLY;
        idx->base_type_or_null = type_int;
        idx->info.operator_value_info.op = OP_CONVERT_INT_INT;  // convert enum to int
        idx->info.operator_value_info.arg[0] = prefix;

        n = new_exp (AN_ARRAY_ELEMENT_OBJECT);
        n->form   = AN_OBJECT;
        n->access = ACCESS_READONLY;
        n->base_type_or_null = type_string;
        n->constraint.kind = RUNTIME_CONSTRAINT;
        n->info.array_element_object_info.prefix = ex;
        n->info.array_element_object_info.index  = idx;
      }
    }
    break;

    case ATTR_size:
    {
      if (prefix->form != AN_OBJECT)
      {
        semantic_error ("prefix of attribute 'size must denote an object", &pos);
        free_exp (prefix);
        *pout = dummy_expression ();
        return;
      }

      if (lexa.within_unsafe)
      {
        if (!is_type_having_size (type))
        {
          semantic_error ("prefix of attribute 'size must not denote a renamed/incomplete/generic/opaque type", &pos);
          free_exp (prefix);
          *pout = dummy_expression ();
          return;
        }
      }
      else
      {
        if (!is_packed_type (type))
        {
          semantic_error ("prefix of attribute 'size must denote a packed object", &pos);
          free_exp (prefix);
          *pout = dummy_expression ();
          return;
        }
      }


      // note: 'type' can be an open type.

      if (prefix->constraint.kind == DOES_NOT_APPLY ||
          prefix->constraint.kind == CONSTANT_CONSTRAINT)      // constant size
      {
        int rc, size, align;

        // returns 0 if OK, -1 if bad type, -2 if size is too large, -3 if missing/non-constant constraint
        rc = size_and_alignment_of_constant_size_exp (prefix, &size, &align);
        if (rc)
          fatal_compiler_error ("attr_size : cannot compute size", &pos);

        n = new_exp (A_CONST_INTEGER_VALUE);
        n->base_type_or_null = type_uint;
        n->form = A_VALUE;
        n->access = ACCESS_CONSTANT;
        n->info.const_integer_value_info.value = size;

        free_exp (prefix);
      }
      else    // runtime with runtime size
      {
        n = new_exp (AN_OPERATOR_VALUE);
        n->base_type_or_null = type_uint;
        n->form = A_VALUE;
        n->access = ACCESS_READONLY;
        n->info.operator_value_info.op = OP_SIZE;
        n->info.operator_value_info.arg[0] = prefix;
      }
    }
    break;

    default:
      fatal_compiler_error ("parse_attribute", &token.pos);
  }

  *pout = n;
}

/*****************************************************************************/

static void parse_struct_field_selection (PEXPRESSION e, bool address_of, PEXPRESSION *out)
{
  PENTITY     discr, field, f, f_type;
  PEXPRESSION n;

  if (e->base_type_or_null != null)
  {
    if (e->base_type_or_null->kind != A_STRUCT_TYPE &&
        e->base_type_or_null->kind != A_UNION_TYPE)
    {
      semantic_error ("prefix must denote a struct or union", &token.pos);
    }
    else if (e->form != AN_OBJECT)
    {
      semantic_error ("prefix must denote an object", &token.pos);
    }
  }


  //  discriminant_value  ::= "."  discriminant_simple_name
  //  struct_field        ::= "."  field_simple_name

  get_token();    // skip dot or single arrow

  if (token.kind != IDENTIFIER)
  {
    syntax_error ("field identifier expected", &token.pos);
    free_exp (e);
    *out = dummy_expression ();
    return;
  }


  // search for identifier in field region

  discr = null;
  field = null;

  if (e->base_type_or_null != null)
  {
    if (e->base_type_or_null->kind == A_STRUCT_TYPE)
    {
      discr = find_tree
                (e->base_type_or_null->info->the_struct_type.discriminant->identifier_btree,
                 token.info._identifier.value);

      if (discr == null)
      {
        field = find_tree
                 (e->base_type_or_null->info->the_struct_type.fields->identifier_btree,
                  token.info._identifier.value);
      }
    }
    else if (e->base_type_or_null->kind == A_UNION_TYPE)
    {
      field = find_tree
                (e->base_type_or_null->info->the_union_type.fields->identifier_btree,
                 token.info._identifier.value);
    }
  }

  if (discr == null && field == null)
    semantic_error ("identifier must denote a struct or union field", &token.pos);



  // check constraint in case of varying field (if it's constant)

  if (field != null && field->kind == A_VARYING_FIELD)
  {
    if (e->constraint.kind == CONSTANT_CONSTRAINT)
    {
      if (e->constraint.value != field->info->the_varying_field.discriminant_value)
        semantic_error ("field does not exist in this struct variant", &token.pos);
    }
  }


  // build result node

  if (discr != null)       // it's a discriminant
  {
    if (e->constraint.kind == CONSTANT_CONSTRAINT)
    {
      n = new_exp (A_CONST_ENUMERATION_VALUE);
      n->base_type_or_null = complete_type_of (discr->info->the_field.type);
      n->form = A_VALUE;
      n->access = ACCESS_CONSTANT;
      n->info.const_enumeration_value_info.value = e->constraint.value;

      free_exp (e);
    }
    else     // it's a runtime constraint
    {
      n = new_exp (A_DISCRIMINANT_VALUE);
      n->base_type_or_null = complete_type_of (discr->info->the_field.type);
      n->form = A_VALUE;
      n->access = ACCESS_READONLY;
      n->info.discriminant_value_info.prefix = e;
      n->info.discriminant_value_info.field = field;
    }
  }
  else if (field != null)   // it's a field
  {
    if (e->kind == A_POOL_CONSTANT &&   // can only be struct, or open-struct with constant-constraint.
        !address_of)                    // makes sure the whole pool constant is loaded into the executable
    {
      FIELD_DATA  data;

      begin_struct (&data, e->base_type_or_null);

      f = e->base_type_or_null->info->the_struct_type.fields->entities.first;

      for (;;)
      {
        skip_to_valid_field (&f, &f_type, e->constraint.value);

        begin_field (&data, f);

        if (f == field)      // found
          break;

        end_field (&data);

        f = f->next;
      }

      load_constant_object_from_pool (e->info.pool_constant_info.pool_cte,
                                      data.field_offset,
                                      data.field_size,
                                      f_type,
                                      &n);
    }
    else            // it's a runtime node
    {
      PENTITY  field_type;

      n = new_exp (A_STRUCT_FIELD_OBJECT);
      n->form   = AN_OBJECT;
      n->access = e->access;   // constant, readonly, readwrite

      if (field->kind == A_FIELD)
        field_type = field->info->the_field.type;
      else
        field_type = field->info->the_varying_field.type;

      convert_type_into_base_type_and_constraint (field_type,
                                                  /*always_constrained=*/ true, // (can be jagged)
                                                  &n->base_type_or_null,
                                                  &n->constraint);

      n->info.struct_field_object_info.prefix = e;
      n->info.struct_field_object_info.field = field;
    }
  }
  else      // some error occured (message already generated) : return dummy expression
  {
    free_exp (e);
    n = dummy_expression ();
  }

  get_token();  // skip identifier

  *out = n;
}

/*****************************************************************************/

//  name ::=
//     | "(" expression ")"
//     | object_name
//     | type_name                    (only used as attribute prefix)
//     | function_name                (returns a function pointer value)
//     | function_call
//     | name  array_element
//     | name  array_slice
//     | name  discriminant_value
//     | name  struct_field
//     | name  dereferenced_object
//     | name  postfixed_object
//     | name  deref_unsafe_field
//     | name  attribute

static void parse_name (PCONTEXT      context,                 // never null
                        bool          left_parenthesis_parsed,
                        TEXT_POSITION *name_position,          // never null
                        PENTITY       type_name,               // can be null
                        PEXPRESSION   *out)
{
  PENTITY      name;
  PEXPRESSION  e;

  if (left_parenthesis_parsed)
  {
    parse_expression (context, name_position, type_name, &e);

    if (token.kind == RIGHT_PARENTHESIS)   // ')'
      get_token();   // skip ')'
    else
      syntax_error ("')' expected", &token.pos);
  }
  else if (type_name == null && token.kind == LEFT_PARENTHESIS)
  {
    get_token();   // skip '('

    parse_expression (context, &token.pos, null, &e);

    if (token.kind == RIGHT_PARENTHESIS)   // ')'
      get_token();   // skip ')'
    else
      syntax_error ("')' expected", &token.pos);
  }
  else  // an object, type or function name must follow.
  {
    if (type_name != null)   // name already parsed
    {
      name = type_name;
    }
    else if (token.kind == TOKEN_false)
    {
      name = enum_false;
      get_token();
    }
    else if (token.kind == TOKEN_true)
    {
      name = enum_true;
      get_token();
    }
    else
    {
      name = type_entity_of_token (token.kind);
      if (name != null)
      {
        get_token();    // skip type token
      }
      else if (token.kind == IDENTIFIER)
      {
        name = parse_expanded_name ();
        if (name == null)   // identifier not found (error already given)
        {
          *out = dummy_expression ();
          return;
        }
      }
      else
      {
        syntax_error ("name is expected here", &token.pos);
        *out = dummy_expression ();
        return;
      }
    }


    // create expression from name (object_name, type_name, function_name)

    if (name->kind < LAST_ENTITY_DENOTING_A_TYPE)   // a type
    {
      if (complete_type_of(name) == global_locked_struct_entity)
        semantic_error ("incomplete struct identifier must not be used in expression", name_position);

      parse_type_attribute (name, &e);
    }
    else if (name->kind == AN_ENUMERATION_LITERAL)
    {
      e = new_exp (A_CONST_ENUMERATION_VALUE);
      e->base_type_or_null = name->info->the_enumeration_literal.type;
      e->form = A_VALUE;
      e->access = ACCESS_CONSTANT;
      e->info.const_enumeration_value_info.value = name->info->the_enumeration_literal.value;
    }
    else if (name->kind == A_CONSTANT)
    {
      name->info->the_constant.is_used = true;     // is referenced

      if (name->info->the_constant.value != null &&
          is_constant_exp (name->info->the_constant.value))
      {
        e = duplicate_constant_exp (name->info->the_constant.value);

        e->form   = AN_OBJECT;           // convert into constant object
        e->access = ACCESS_CONSTANT;

        convert_type_into_base_type_and_constraint (name->info->the_constant.type,
                                                    /*always_constrained=*/ true, // can be jagged
                                                    &e->base_type_or_null,
                                                    &e->constraint);
      }
      else
      {
        e = dummy_expression ();
      }
    }
    else if (name->kind == A_GLOBAL_VARIABLE)
    {
      e = new_exp (A_GLOBAL_VARIABLE_OBJECT);
      e->form   = AN_OBJECT;
      e->access = ACCESS_READWRITE;
      e->info.global_variable_object_info.pobject = name;

      convert_type_into_base_type_and_constraint (name->info->the_global_variable.type,
                                                  /*always_constrained=*/ true,
                                                  &e->base_type_or_null,
                                                  &e->constraint);
    }
    else if (name->kind == A_LOCAL_VARIABLE)
    {
      name->info->the_local_variable.is_used = true;     // is referenced

      e = new_exp (A_LOCAL_VARIABLE_OBJECT);
      e->form   = AN_OBJECT;
      e->access = ACCESS_READWRITE;
      e->info.local_variable_object_info.pobject = name;
      memcpy (&e->info.local_variable_object_info.pos, name_position, sizeof(TEXT_POSITION));

      convert_type_into_base_type_and_constraint (name->info->the_local_variable.type,
                                                  /*always_constrained=*/ true,
                                                  &e->base_type_or_null,
                                                  &e->constraint);
    }
    else if (name->kind == A_REFERENCE)
    {
      name->info->the_reference.is_used = true;     // is referenced

      e = new_exp (A_REFERENCE_OBJECT);
      e->form   = AN_OBJECT;
      e->access = name->info->the_reference.mode == MODE_IN ? ACCESS_READONLY : ACCESS_READWRITE;
      e->info.reference_object_info.pobject = name;
      memcpy (&e->info.reference_object_info.pos, name_position, sizeof(TEXT_POSITION));

      convert_type_into_base_type_and_constraint (name->info->the_reference.type,
                                                  /*always_constrained=*/ true,
                                                  &e->base_type_or_null,
                                                  &e->constraint);
    }
    else if (name->kind == A_PARAMETER)
    {
      name->info->the_parameter.is_used = true;     // is referenced

      e = new_exp (A_PARAMETER_OBJECT);
      e->form   = AN_OBJECT;
      e->access = name->info->the_parameter.mode == MODE_IN ? ACCESS_READONLY : ACCESS_READWRITE;
      e->info.parameter_object_info.pobject = name;
      memcpy (&e->info.parameter_object_info.pos, name_position, sizeof(TEXT_POSITION));

      convert_type_into_base_type_and_constraint (name->info->the_parameter.type,
                                                  /*always_constrained=*/ true,
                                                  &e->base_type_or_null,
                                                  &e->constraint);
    }
    else if (name->kind == A_GENERIC_FUNCTION)
    {
      e = new_exp (A_FUNCTION_VALUE);
      e->base_type_or_null = name->info->the_generic_function.to_type;
      e->form   = A_VALUE;
      e->access = ACCESS_READONLY;
      e->info.function_value_info.to_function_declaration_or_generic_function = name;
    }
    else if (name->kind == A_FUNCTION_DECLARATION)
    {
      e = new_exp (A_FUNCTION_VALUE);
      e->base_type_or_null = name->info->the_function_declaration.to_type;
      e->form   = A_VALUE;
      e->access = ACCESS_READONLY;
      e->info.function_value_info.to_function_declaration_or_generic_function = name;
    }
    else
    {
      semantic_error ("a name is expected here", name_position);
      e = dummy_expression ();
    }
  }


  // suffix list

  for (;;)
  {
    switch (token.kind)
    {
      case LEFT_PARENTHESIS:   // function_call
        {
          PEXPRESSION n;
          parse_function_call (name_position, e, &n);
          e = n;
        }
        break;


      case LEFT_BRACKET:   // array_element or array_slice
      {
        PEXPRESSION   e1, e2;
        TEXT_POSITION pos1, pos2;
        bool          valid_range;
        PENTITY       ebt;

        //  array_element ::= "["  expression  "]"
        //  array_slice   ::= "["  expression  ":"  expression "]"


        ebt = e->base_type_or_null;
        if (ebt != null)
        {
          if (ebt->kind == AN_OPEN_ARRAY_TYPE)
          {
            if (e->form != AN_OBJECT)
            {
              semantic_error ("prefix must denote an object", &token.pos);
            }
          }
          else if (ebt->kind == AN_UNSAFE_POINTER_TYPE)
          {
            PENTITY element_type;

            element_type = ebt->info->the_unsafe_pointer_type.designated_type;
            if (is_open_type (element_type))
              semantic_error ("prefix must not be an unsafe pointer to an open type", &token.pos);

            // ok (accept object or value)
            check_unsafe_region ();
          }
          else
          {
            semantic_error ("prefix must denote an array", &token.pos);
          }
        }


        get_token();    // skip [

        memcpy (&pos1, &token.pos, sizeof(pos1));
        parse_expression (&no_context, &token.pos, null, &e1);

        valid_range = true;


        // check type (int or uint)

        if (e1->base_type_or_null != null)
        {
          if (e1->base_type_or_null->kind != AN_INTEGER_TYPE ||
              e1->base_type_or_null == type_long)
          {
            semantic_error ("index must have type int or uint", &pos1);
          }
        }


        if (token.kind != COLON)   // an array element
        {
          // if index is constant, check that it's in int or uint range.

          if (ebt != null && ebt->kind == AN_OPEN_ARRAY_TYPE && e1->kind == A_CONST_INTEGER_VALUE)
          {
            int8 r;

            r = e1->info.const_integer_value_info.value;

            if (r < 0 || r >= 2147483647)
            {
              semantic_error ("index is out of range", &pos1);
              valid_range = false;
            }
            else
            {
              // if the array has constant bounds, check them.
              if (e->constraint.kind == CONSTANT_CONSTRAINT && r >= e->constraint.value)
              {
                semantic_error ("index is too large", &pos1);
                valid_range = false;
              }
            }
          }


          // build result node : either a pool constant, a simple constant, or a runtime.

          if (ebt != null && ebt->kind == AN_OPEN_ARRAY_TYPE && valid_range &&
              e->kind == A_POOL_CONSTANT && e1->kind == A_CONST_INTEGER_VALUE &&
              !context->address_of)     // makes sure the whole pool constant is loaded into the executable
          {
            PENTITY     element_type;
            int         elem_size, align, rc, offset;
            int8        r;
            PEXPRESSION n;

            // it's a constant node (result is either a pool constant or a simple constant)

            element_type = complete_type_of (e->base_type_or_null->info->the_open_array_type.element);

            // returns 0 if OK, -1 if bad type, -2 if size is too large
            rc = size_and_alignment_of_type (element_type, &elem_size, &align);
            if (rc < 0)
              fatal_compiler_error ("array_index(elem_type)", &token.pos);

            r = e1->info.const_integer_value_info.value;

            offset = (int)r * elem_size;

            load_constant_object_from_pool (e->info.pool_constant_info.pool_cte,
                                            offset,
                                            elem_size,
                                            element_type,
                                            &n);
            free_exp (e);
            free_exp (e1);

            e = n;
          }
          else            // it's a runtime node
          {
            PEXPRESSION n;
            PENTITY     element_type;

            n = new_exp (AN_ARRAY_ELEMENT_OBJECT);
            n->form   = AN_OBJECT;

            if (ebt != null && ebt->kind == AN_OPEN_ARRAY_TYPE)
            {
              element_type = ebt->info->the_open_array_type.element;
              n->access = e->access;   // constant, readonly, readwrite
            }
            else if (ebt != null && ebt->kind == AN_UNSAFE_POINTER_TYPE)
            {
              element_type = ebt->info->the_unsafe_pointer_type.designated_type;
              n->access = ACCESS_READWRITE;
            }
            else
            {
              element_type = null;
              n->access = e->access;   // constant, readonly, readwrite
            }

            convert_type_into_base_type_and_constraint (element_type,
                                                        /*always_constrained=*/ true, // (can be jagged)
                                                        &n->base_type_or_null,
                                                        &n->constraint);

            n->info.array_element_object_info.prefix = e;
            n->info.array_element_object_info.index  = e1;

            e = n;
          }
        }
        else             // an array slice
        {
          get_token();   // skip ':'
          memcpy (&pos2, &token.pos, sizeof(pos2));
          parse_expression (&no_context, &token.pos, null, &e2);


          // check type (int or uint)

          if (e2->base_type_or_null != null)
          {
            if (e2->base_type_or_null->kind != AN_INTEGER_TYPE ||
                e2->base_type_or_null == type_long)
            {
              semantic_error ("index must have type int or uint", &pos2);
            }
          }


          // if slice offset is constant, check that it's in int or uint range.

          if (ebt != null && ebt->kind == AN_OPEN_ARRAY_TYPE && e1->kind == A_CONST_INTEGER_VALUE)
          {
            int8 r;

            r = e1->info.const_integer_value_info.value;

            if (r < 0 || r > 2147483647)
            {
              semantic_error ("slice offset is out of range", &pos1);
              valid_range = false;
            }
            else
            {
              // if the array has constant bounds, check them.
              if (e->constraint.kind == CONSTANT_CONSTRAINT && r > e->constraint.value)
              {
                semantic_error ("slice offset is too large", &pos1);
                valid_range = false;
              }
            }
          }


          // if slice length is constant, check that it's in int or uint range.

          if (ebt != null && ebt->kind == AN_OPEN_ARRAY_TYPE && e2->kind == A_CONST_INTEGER_VALUE)
          {
            int8 r;

            r = e2->info.const_integer_value_info.value;

            if (r < 0 || r > 2147483647)
            {
              semantic_error ("slice length is out of range", &pos2);
              valid_range = false;
            }
            else
            {
              // if the slice length has constant bounds, check them.
              if (e->constraint.kind == CONSTANT_CONSTRAINT && r > e->constraint.value)
              {
                semantic_error ("slice length is too large", &pos2);
                valid_range = false;
              }
            }
          }


          // if slice offset and length are constant, check that both are <= array'length

          if (ebt != null && ebt->kind == AN_OPEN_ARRAY_TYPE &&
              valid_range &&
              e1->kind == A_CONST_INTEGER_VALUE &&
              e2->kind == A_CONST_INTEGER_VALUE &&
              e->constraint.kind == CONSTANT_CONSTRAINT &&
              e1->info.const_integer_value_info.value
               + e2->info.const_integer_value_info.value
                 > e->constraint.value)
          {
            semantic_error ("slice offset+length is too large", &pos1);
            valid_range = false;
          }


          // build result node : either a pool constant, or a runtime slice.

          if (ebt != null && ebt->kind == AN_OPEN_ARRAY_TYPE &&
              valid_range &&
              e->kind == A_POOL_CONSTANT &&
              e1->kind == A_CONST_INTEGER_VALUE &&
              e2->kind == A_CONST_INTEGER_VALUE &&
              !context->address_of)     // makes sure the whole pool constant is loaded into the executable
          {
            PENTITY     element_type;
            int         elem_size, align, rc;
            POOL        *p;
            int8        ofs, len;
            PEXPRESSION n;

            // it's a constant pool constant

            element_type = complete_type_of (e->base_type_or_null->info->the_open_array_type.element);

            // returns 0 if OK, -1 if bad type, -2 if size is too large
            rc = size_and_alignment_of_type (element_type, &elem_size, &align);
            if (rc < 0)
              fatal_compiler_error ("array_index(elem_type)", &token.pos);

            ofs = e1->info.const_integer_value_info.value;
            len = e2->info.const_integer_value_info.value;

            n = new_exp (A_POOL_CONSTANT);
            n->form = AN_OBJECT;
            n->access = ACCESS_CONSTANT;
            n->base_type_or_null = e->base_type_or_null;
            n->constraint.kind = CONSTANT_CONSTRAINT;
            n->constraint.value = (uint4)len;

            p = new_pool_constant ((uint4)len * elem_size, align);

            copy_pool_to_pool (e->info.pool_constant_info.pool_cte, (uint4)ofs * elem_size,
                               p, 0,
                               (uint4)len * elem_size);

            n->info.pool_constant_info.pool_cte = p;

            free_exp (e);
            free_exp (e1);
            free_exp (e2);

            e = n;
          }
          else            // it's a runtime node
          {
            PEXPRESSION n;

            n = new_exp (AN_ARRAY_SLICE_OBJECT);
            n->form   = AN_OBJECT;

            if (ebt != null && ebt->kind == AN_OPEN_ARRAY_TYPE)
            {
              n->base_type_or_null = ebt;
              n->access = e->access;   // constant, readonly, readwrite
            }
            else if (ebt != null && ebt->kind == AN_UNSAFE_POINTER_TYPE)
            {
              PENTITY element_type;
              AN_OPEN_ARRAY_TYPE_ENTITY open_array;

              element_type = ebt->info->the_unsafe_pointer_type.designated_type;

              memset (&open_array, 0x00, sizeof(open_array));
              open_array.element = element_type;

              n->base_type_or_null =
                  append_new_entity (AN_OPEN_ARRAY_TYPE, &open_array, L"", &token.pos);

              n->access = ACCESS_READWRITE;
            }
            else
            {
              n->base_type_or_null = null;
              n->access = e->access;   // constant, readonly, readwrite
            }

            if (valid_range && e2->kind == A_CONST_INTEGER_VALUE)
            {
              n->constraint.kind = CONSTANT_CONSTRAINT;
              n->constraint.value = (uint4)e2->info.const_integer_value_info.value;
            }
            else
            {
              n->constraint.kind = RUNTIME_CONSTRAINT;
            }

            n->info.array_slice_object_info.prefix = e;
            n->info.array_slice_object_info.index  = e1;
            n->info.array_slice_object_info.length = e2;

            e = n;
          }
        }

        if (token.kind == RIGHT_BRACKET)   // ']'
          get_token();   // skip ']'
        else
          syntax_error ("']' expected", &token.pos);

        break;
      }


      case DOT:             // discriminant_value or struct_field
      {
        PEXPRESSION n;

        parse_struct_field_selection (e, context->address_of, &n);

        e = n;
      }
      break;


      case MINUS_MINUS:        // postfixed_object
      case PLUS_PLUS:
      {
        PEXPRESSION n;
        PENTITY     type;

        //  postfixed_object    ::= "--" | "++"

        type = e->base_type_or_null;

        if (e->access != ACCESS_READWRITE)
        {
          semantic_error ("prefix must denote a read/write object", &token.pos);
        }
        else if (e->form != AN_OBJECT)
        {
          semantic_error ("prefix must denote an object", &token.pos);
        }
        else if (type != null)
        {
          if (type->kind != AN_INTEGER_TYPE &&
              type->kind != AN_ENUMERATION_TYPE &&
              type->kind != AN_UNSAFE_POINTER_TYPE)
          {
            semantic_error ("prefix must denote enum, integer or unsafe pointer object", &token.pos);
          }
        }

        n = new_exp (AN_OPERATOR_VALUE);
        n->base_type_or_null = type;
        n->form = A_VALUE;
        n->access = ACCESS_READONLY;
        n->info.operator_value_info.op = (token.kind == MINUS_MINUS) ? OP_POST_DEC : OP_POST_INC;
        n->info.operator_value_info.arg[0] = e;

        e = n;

        get_token();
      }
      break;


      case SINGLE_ARROW:       // deref_unsafe_field
      {
        PEXPRESSION n;
        PENTITY     type;

        //  deref_unsafe_field  ::= "->" field_simple_name

        type = e->base_type_or_null;
        if (type != null && type->kind != AN_UNSAFE_POINTER_TYPE)
          semantic_error ("prefix must be unsafe pointer", &token.pos);

        if (e->kind == A_CONST_NULL_VALUE)
          semantic_error ("dereferencing null pointer is not allowed", &token.pos);

        n = new_exp (AN_UNSAFE_DEREFERENCED_OBJECT);
        n->form = AN_OBJECT;
        n->access = ACCESS_READWRITE;
        n->info.unsafe_dereferenced_object_info.unsafe_ptr_value = e;

        if (type != null && type->kind == AN_UNSAFE_POINTER_TYPE)
        {
          PENTITY dt;

          check_unsafe_region ();

          dt = type->info->the_unsafe_pointer_type.designated_type;
          if (is_open_type (dt))
            semantic_error ("left operand of '->' must not be an unsafe pointer to an open type", &token.pos);

          convert_type_into_base_type_and_constraint
                  (type->info->the_unsafe_pointer_type.designated_type,
                   /*always_constrained=*/ true,
                   &n->base_type_or_null,
                   &n->constraint);
        }

        e = n;

        parse_struct_field_selection (e, false, &n);
        e = n;
      }
      break;

      case APOSTROPHE:         // attribute
      {
        PEXPRESSION n;
        parse_attribute (e, &n);
        e = n;
      }
      break;

      default:
      {
        if (token.kind == CARET &&                 //  dereferenced_object ::= "^"
            e->base_type_or_null != null &&
            e->base_type_or_null->kind == A_POINTER_TYPE)
        {
          PEXPRESSION n;

          if (e->kind == A_CONST_NULL_VALUE)
            semantic_error ("dereferencing a null pointer is not allowed", &token.pos);

          n = new_exp (A_DEREFERENCED_OBJECT);
          n->form   = AN_OBJECT;
          n->access = ACCESS_READWRITE;

          convert_type_into_base_type_and_constraint (e->base_type_or_null->info->the_pointer_type.designated_type,
                                                      /*always_constrained=*/ true,
                                                      &n->base_type_or_null,
                                                      &n->constraint);
          n->info.dereferenced_object_info.prefix = e;

          e = n;

          get_token();   // skip caret

          break;
        }

        if (is_unsafe_type (e->base_type_or_null))
          check_unsafe_region ();

        *out = e;
        return;
      }
    }
  }
}

/*****************************************************************************/

// returns 0 if OK, -1 if overflow error

static int check_float_overflow_underflow (double *rf, PENTITY t, TEXT_POSITION *pos)
{
  bool overflow  = false;
  bool underflow = false;

  if (t == type_float)
  {
    float f = (float)(*rf);
    if (fisinfinite (f))
      overflow = true;
    else if (f == 0.0 && (*rf) != 0.0)
      underflow = true;
    (*rf) = f;      // reduce precision (compiler note: be careful that this is not optimized away)
  }
  else  // double
  {
    if (isinfinite (*rf))
    {
      overflow = true;
    }
  }

  if (overflow)
  {
    semantic_error ("overflow", pos);
    *rf = 0.0;
    return -1;
  }

  if (underflow)
    warning ("underflow", pos);

  return 0;
}

/*****************************************************************************/

// returns 0 if OK, -1 if types are not compatible
// error message is generated in function
// note: this is also used for checking compatibility of in-parameters,
//       opaque types are allowed.

int check_assignment_context_compatibility (PCONTEXT target, PEXPRESSION source, TEXT_POSITION *pos)
{
  PENTITY s, t;

  s = complete_type_of (source->base_type_or_null);
  t = complete_type_of (target->base_type_or_null);

  if (s == null || t == null)   // previous error
    return -1;

  switch (t->kind)
  {
    case AN_INTEGER_TYPE:
      if (s->kind != AN_INTEGER_TYPE)
      {
        semantic_error ("type is not compatible with integer", pos);
        return -1;
      }

      if (s == type_int_literal)   // a constant
      {
        int8         r;
        INTEGER_INFO *pt;

        if (source->kind != A_CONST_INTEGER_VALUE)
          fatal_compiler_error ("int-literal#1", pos);

        r = source->info.const_integer_value_info.value;

        pt = &integer_info[t->info->the_integer_type.type];
        if (r < pt->min || r > pt->max)
        {
          semantic_error ("overflow", pos);
          return -1;
        }

        return 0;
      }


      // check that source type fits into target type

      {
        INTEGER_INFO *ps, *pt;

        ps = &integer_info[s->info->the_integer_type.type];
        pt = &integer_info[t->info->the_integer_type.type];

        if (ps->min < pt->min || ps->max > pt->max)
        {
          semantic_error ("type does not fit into target type - use conversion", pos);
          return -1;
        }

        return 0;
      }

      break;

    case A_FLOAT_TYPE:
      if (s->kind != A_FLOAT_TYPE)
      {
        semantic_error ("type is not compatible with floating-point", pos);
        return -1;
      }

      if (s == type_float_literal)   // a constant
      {
        double f;

        if (source->kind != A_CONST_FLOAT_VALUE)
          fatal_compiler_error ("float-literal#1", pos);

        f = source->info.const_float_value_info.value;

        if (check_float_overflow_underflow (&f, t, pos) < 0)
          return -1;

        return 0;
      }


      // check that source type fits into target type

      if (s == type_double && t == type_float)
      {
        semantic_error ("type does not fit into target type - use conversion", pos);
        return -1;
      }

      return 0;

      break;

    case AN_OPEN_ARRAY_TYPE:
      if (s->kind != AN_OPEN_ARRAY_TYPE || !types_are_equal (s, t))
      {
        semantic_error ("type is not compatible with array", pos);
        return -1;
      }

      if (source->constraint.kind == CONSTANT_CONSTRAINT &&
          target->constraint.kind == CONSTANT_CONSTRAINT &&
          source->constraint.value != target->constraint.value)
      {
        semantic_error ("array length does not match", pos);
        return -1;
      }

      return 0;

      break;

    case A_STRUCT_TYPE:       // can be open
      if (s->kind != A_STRUCT_TYPE || !types_are_equal (s, t))
      {
        semantic_error ("type is not compatible with struct", pos);
        return -1;
      }

      if (source->constraint.kind == CONSTANT_CONSTRAINT &&
          target->constraint.kind == CONSTANT_CONSTRAINT &&
          source->constraint.value != target->constraint.value)
      {
        semantic_error ("struct discriminant does not match", pos);
        return -1;
      }

      return 0;

      break;

    case AN_ENUMERATION_TYPE:
    case A_UNION_TYPE:
    case AN_INCOMPLETE_TYPE:
    case A_GENERIC_TYPE:
    case AN_OPAQUE_TYPE:
      if (!types_are_equal (s, t))
      {
        semantic_error ("type is not compatible", pos);
        return -1;
      }

      return 0;

      break;

    case A_POINTER_TYPE:
    case A_FUNCTION_POINTER_TYPE:
    case AN_UNSAFE_POINTER_TYPE:
      if (s->kind == A_NULL_POINTER_TYPE)
      {
        // null is compatible with any pointer kind
        return 0;
      }
      else if (!types_are_equal (s, t))
      {
        semantic_error ("type is not compatible", pos);
        return -1;
      }

      return 0;

      break;

    default:
      semantic_error ("type is not compatible", pos);
      return -1;
      break;
  }
}

/*****************************************************************************/

// all checks were done before, no error should occur

static void store_constant_into_aggregate_pool (PCONTEXT target, PEXPRESSION source, POOL *pool, uint offset)
{
  PENTITY      s, t;
  int8         r;
  double       rf;
  INTEGER_INFO *pi;
  FLOAT_INFO   *pf;
  POOL         *ps;

  s = complete_type_of (source->base_type_or_null);
  t = complete_type_of (target->base_type_or_null);

  switch (t->kind)
  {
    case AN_ENUMERATION_TYPE:
      if (s->kind != AN_ENUMERATION_TYPE || source->kind != A_CONST_ENUMERATION_VALUE)
        fatal_compiler_error ("exp#791", &token.pos);

      r = source->info.const_enumeration_value_info.value;
      pi = &integer_info[t->info->the_enumeration_type.base];

      store_integer (pool, offset, r, pi->size);

      return;


    case AN_INTEGER_TYPE:
      if (s->kind != AN_INTEGER_TYPE || source->kind != A_CONST_INTEGER_VALUE)
        fatal_compiler_error ("exp#804", &token.pos);

      r = source->info.const_integer_value_info.value;
      pi = &integer_info[t->info->the_integer_type.type];

      store_integer (pool, offset, r, pi->size);

      return;


    case A_FLOAT_TYPE:
      if (s->kind != A_FLOAT_TYPE || source->kind != A_CONST_FLOAT_VALUE)
        fatal_compiler_error ("exp#817", &token.pos);

      rf = source->info.const_float_value_info.value;
      pf = &float_info[t->info->the_float_type.type];

      store_float (pool, offset, rf, pf->size);

      return;


    case A_POINTER_TYPE:
    case A_FUNCTION_POINTER_TYPE:
    case AN_UNSAFE_POINTER_TYPE:
      if (source->kind != A_CONST_NULL_VALUE)
        fatal_compiler_error ("exp#834", &token.pos);

      // nothing to do : pool zone is zeroed by default

      return;

    case AN_OPEN_ARRAY_TYPE:
    case A_STRUCT_TYPE:       // can be open
      if (source->kind != A_POOL_CONSTANT)
        fatal_compiler_error ("exp#867", &token.pos);

      ps = source->info.pool_constant_info.pool_cte;

      if (target->constraint.kind == CONSTANT_CONSTRAINT ||  // array with/struct length constraint
          target->constraint.kind == DOES_NOT_APPLY)         // non-open struct
      {
        copy_pool_to_pool (ps, 0, pool, offset, size_of_pool_cte (ps));  // copy of pool zone
      }
      else    // open array/struct : jagged reference (store address & length/discriminant)
      {
        store_reference (pool, offset, ps);         // jagged reference
        store_integer (pool, offset+address_size, source->constraint.value, 4);
      }
      return;


    default:
      fatal_compiler_error ("exp#895", &token.pos);
      break;
  }
}

/*****************************************************************************/

//  positional_array_aggregate ::=
//              "{" [ expression {"," expression} [","] ] "}"
//
//  positional_struct_aggregate ::=
//              "{" [  expression {"," expression} [","]  ] "}"

static void parse_non_named_aggregate (PCONTEXT      context,        // never null
                                       bool          context_ok,
                                       TEXT_POSITION *paggregate_pos,
                                       PEXPRESSION   *out)
{
  bool ok = context_ok;

  if (ok && context->base_type_or_null->kind == AN_OPEN_ARRAY_TYPE)
  {
    PENTITY     element_type;
    CONTEXT     element_context;
    int         length = 0;
    int         elem_size, size, align, rc;
    PEXPRESSION e;
    EXPRESSIONS list;
    TEXT_POSITION pos;
    bool        is_constant = true;

    clear (list);

    // compute element type
    element_type = complete_type_of (context->base_type_or_null->info->the_open_array_type.element);
    convert_type_into_context (element_type, &element_context);

    while (token.kind != RIGHT_ACCOLADE)
    {
      memcpy (&pos, &token.pos, sizeof(pos));
      parse_expression (&element_context, &pos, null, &e);
      if (length == 2147483647)
      {
        semantic_error ("too many array elements", &token.pos);
        length = 0;
        ok = false;
      }
      length++;

      if (check_assignment_context_compatibility (&element_context, e, &pos) < 0)
        ok = false;
      is_constant &= is_constant_exp (e);
      append_expression (&list, e, element_type, null);

      if (token.kind == RIGHT_ACCOLADE)
        break;

      if (token.kind != COMMA)
      {
        syntax_error ("',' or '}' expected", &token.pos);
        break;
      }

      get_token();
    }

    if (!is_constant)
    {
      if (context->base_type_or_null != null && is_jagged_type (context->base_type_or_null))
        semantic_error ("jagged aggregates must be constant", paggregate_pos);
    }

    if (is_constant && ok)
    {
      // returns 0 if OK, -1 if bad type, -2 if size is too large
      rc = size_and_alignment_of_type (element_type, &elem_size, &align);
      if (rc == -1)
      {
        semantic_error ("array type is not constant", paggregate_pos);
        ok = false;
      }
      else if (rc == -2 || mul_int4 (elem_size, length, &size) < 0)
      {
        semantic_error ("aggregate is too large", paggregate_pos);
        ok = false;
      }
    }

    if (is_constant && ok)
    {
      POOL  *p;
      LIST_OF_EXPRESSIONS *l;
      uint4 offset;

      e = new_exp (A_POOL_CONSTANT);
      e->base_type_or_null = context->base_type_or_null;
      e->constraint.kind = CONSTANT_CONSTRAINT;
      e->constraint.value = length;
      e->form = A_VALUE;
      e->access = ACCESS_CONSTANT;

      p = new_pool_constant (size, align);

      e->info.pool_constant_info.pool_cte = p;

      l = list.head;
      offset = 0;
      while (l != null)
      {
        store_constant_into_aggregate_pool (&element_context, l->exp, p, offset);
        offset += elem_size;
        l = l->next;
      }

      free_list (&list);
    }
    else     // runtime aggregate
    {
      e = new_exp (AN_AGGREGATE_VALUE);
      e->base_type_or_null = ok ? context->base_type_or_null : null;
      e->constraint.kind = CONSTANT_CONSTRAINT;
      e->constraint.value = length;
      e->form = A_VALUE;
      e->access = ACCESS_READONLY;
      e->info.aggregate_value_info.list = list.head;
    }

    *out = e;
  }
  else   // struct, or open-struct with constant constraint, or error.
  {
    PENTITY       field = null;
    PENTITY       field_type = null;
    CONTEXT       field_context;
    PEXPRESSION   e;
    EXPRESSIONS   list;
    TEXT_POSITION pos;
    bool          is_constant = true;
    FIELD_DATA    data;

    clear (list);

    if (ok)
    {
      field = context->base_type_or_null->info->the_struct_type.fields->entities.first;
      skip_to_valid_field (&field, &field_type, context->constraint.value);
    }

    while (token.kind != RIGHT_ACCOLADE)
    {
      if (ok && field == null)
      {
        semantic_error ("too many expressions in aggregate", &token.pos);
        ok = false;
      }

      convert_type_into_context (field_type, &field_context);

      memcpy (&pos, &token.pos, sizeof(pos));
      parse_expression (&field_context, &pos, null, &e);

      if (ok)
      {
        if (check_assignment_context_compatibility (&field_context, e, &pos) < 0)
          ok = false;
        is_constant &= is_constant_exp (e);
        append_expression (&list, e, field_type, field);

        if (field != null)
          field = field->next;
        skip_to_valid_field (&field, &field_type, context->constraint.value);
      }
      else
      {
        free_exp (e);
      }

      if (token.kind == RIGHT_ACCOLADE)
        break;

      if (token.kind != COMMA)
      {
        syntax_error ("',' or '}' expected", &token.pos);
        break;
      }

      get_token();
    }


    if (ok && field != null)
    {
      semantic_error ("missing fields in aggregate", &token.pos);
      ok = false;
    }


    if (!is_constant)
    {
      if (context->base_type_or_null != null && is_jagged_type (context->base_type_or_null))
        semantic_error ("jagged aggregates must be constant", paggregate_pos);
    }


    if (is_constant && ok)
    {
      LIST_OF_EXPRESSIONS *l;

      begin_struct (&data, context->base_type_or_null);

      l = list.head;
      while (l != null)
      {
        begin_field (&data, l->e);
        end_field (&data);
        l = l->next;
      }

      end_struct (&data);

      if (data.invalid_type)
      {
        semantic_error ("struct type is not constant", paggregate_pos);
        ok = false;
      }
      else if (data.size_overflow)
      {
        semantic_error ("aggregate is too large", paggregate_pos);
        ok = false;
      }
    }

    if (is_constant && ok)
    {
      POOL *p;
      LIST_OF_EXPRESSIONS *l;

      e = new_exp (A_POOL_CONSTANT);
      e->base_type_or_null = context->base_type_or_null;

      if (context->base_type_or_null->kind == A_STRUCT_TYPE &&
          context->base_type_or_null->info->the_struct_type.is_open_type)
      {
        e->constraint.kind = CONSTANT_CONSTRAINT;
        e->constraint.value = context->constraint.value;
      }

      e->form = A_VALUE;
      e->access = ACCESS_CONSTANT;

      p = new_pool_constant (data.struct_size, data.struct_alignment);

      e->info.pool_constant_info.pool_cte = p;

      begin_struct (&data, context->base_type_or_null);

      l = list.head;
      while (l != null)
      {
        begin_field (&data, l->e);

        convert_type_into_context (l->type, &field_context);
        store_constant_into_aggregate_pool (&field_context, l->exp, p, data.field_offset);

        end_field (&data);
        l = l->next;
      }

      end_struct (&data);

      free_list (&list);
    }
    else     // runtime aggregate
    {
      e = new_exp (AN_AGGREGATE_VALUE);
      e->base_type_or_null = ok ? context->base_type_or_null : null;

      if (context->base_type_or_null != null &&
          context->base_type_or_null->kind == A_STRUCT_TYPE &&
          context->base_type_or_null->info->the_struct_type.is_open_type)
      {
        e->constraint.kind = CONSTANT_CONSTRAINT;
        e->constraint.value = context->constraint.value;
      }

      e->form   = A_VALUE;
      e->access = ACCESS_READONLY;
      e->info.aggregate_value_info.list = list.head;
    }

    *out = e;
  }

  if (token.kind == RIGHT_ACCOLADE)
    get_token();
}

/*****************************************************************************/

//  open_array_aggregate ::= "{"  "all"  "=>"  expression  "}"
//
//  named_struct_aggregate ::=
//         "{"  [ identifier "=>" expression
//                 {"," identifier "=>" expression} [","]  ] "}"

static void parse_named_aggregate (PCONTEXT      context,        // never null
                                   bool          context_ok,
                                   TEXT_POSITION *paggregate_pos,
                                   PEXPRESSION   *out)
{
  bool ok = context_ok;

  if (ok && context->base_type_or_null->kind == AN_OPEN_ARRAY_TYPE)
  {
    PEXPRESSION   e, e2;
    PENTITY       element_type;
    CONTEXT       element_context;
    int           length = 0;
    TEXT_POSITION pos;
    bool          is_constant = true;
    int           elem_size, size;
    int           align, rc;

    if (token.kind != IDENTIFIER || wstrcmp (token.info._identifier.value, L"all") != 0)
    {
      syntax_error ("'all' is expected for an array aggregate", &token.pos);
      *out = dummy_expression ();
      return;
    }

    if (context->constraint.kind != CONSTANT_CONSTRAINT &&
        context->constraint.kind != RUNTIME_CONSTRAINT)
    {
      syntax_error ("an array type with constraint is required as context", paggregate_pos);
      ok = false;
    }

    get_token();

    if (token.kind != ARROW)
    {
      syntax_error ("'=>' expected", &token.pos);
      *out = dummy_expression ();
      return;
    }

    get_token();

    // compute element type
    element_type = complete_type_of (context->base_type_or_null->info->the_open_array_type.element);

    convert_type_into_context (element_type, &element_context);

    memcpy (&pos, &token.pos, sizeof(pos));
    parse_expression (&element_context, &pos, null, &e);
    is_constant &= is_constant_exp (e);
    if (check_assignment_context_compatibility (&element_context, e, &pos) < 0)
      ok = false;

    if (token.kind != RIGHT_ACCOLADE)
      syntax_error ("'}' expected", &token.pos);


    if (!is_constant)
    {
      if (context->base_type_or_null != null && is_jagged_type (context->base_type_or_null))
        semantic_error ("jagged aggregates must be constant", paggregate_pos);
    }


    if (context->constraint.kind == CONSTANT_CONSTRAINT && is_constant && ok)
    {
      length = context->constraint.value;

      // returns 0 if OK, -1 if bad type, -2 if size is too large
      rc = size_and_alignment_of_type (element_type, &elem_size, &align);
      if (rc == -1)
      {
        semantic_error ("array type is not constant", paggregate_pos);
        ok = false;
      }
      else if (rc == -2 || mul_int4 (elem_size, length, &size) < 0)
      {
        semantic_error ("aggregate is too large", paggregate_pos);
        ok = false;
      }
    }


    // if constant context length and constant expression, it's a constant array pool object.

    if (context->constraint.kind == CONSTANT_CONSTRAINT && is_constant && ok)
    {
      POOL  *p;
      uint4 offset, i;

      e2 = new_exp (A_POOL_CONSTANT);
      e2->base_type_or_null = context->base_type_or_null;
      e2->constraint.kind = CONSTANT_CONSTRAINT;
      e2->constraint.value = length;
      e2->form = A_VALUE;
      e2->access = ACCESS_CONSTANT;

      p = new_pool_constant (size, align);

      e2->info.pool_constant_info.pool_cte = p;

      offset = 0;
      for (i=0; i<(uint4)length; i++)
      {
        store_constant_into_aggregate_pool (&element_context, e, p, offset);
        offset += elem_size;
      }

      e = e2;
    }
    else     // runtime aggregate
    {
      e2 = new_exp (AN_UNC_ARRAY_AGGREGATE);
      e2->base_type_or_null = ok ? context->base_type_or_null : null;
      memcpy (&e2->constraint, &context->constraint, sizeof(CONSTRAINT));
      e2->form = A_VALUE;
      e2->access = ACCESS_READONLY;
      e2->info.unc_array_aggregate_info.element = e;

      e = e2;
    }

    *out = e;
  }
  else   // struct, or open-struct with constant constraint, or error.
  {
    PENTITY       field = null;
    PENTITY       field_type = null;
    CONTEXT       field_context;
    PEXPRESSION   e;
    EXPRESSIONS   list;
    TEXT_POSITION pos;
    bool          is_constant = true;
    FIELD_DATA    data;

    clear (list);

    if (ok)
    {
      field = context->base_type_or_null->info->the_struct_type.fields->entities.first;
      skip_to_valid_field (&field, &field_type, context->constraint.value);
    }

    while (token.kind != RIGHT_ACCOLADE)
    {
      if (ok && field == null)
      {
        semantic_error ("too many expressions in aggregate", &token.pos);
        ok = false;
      }

      if (token.kind != IDENTIFIER)
      {
        syntax_error ("'identifier =>' expected", &token.pos);
        break;
      }

      if (field != null && field->identifier_or_null != null && wstrcmp (token.info._identifier.value, field->identifier_or_null) != 0)
      {
        syntax_error ("identifier is not a struct field", &token.pos);
        ok = false;
        break;
      }

      get_token();

      if (token.kind != ARROW)
      {
        syntax_error ("'=>' expected", &token.pos);
        break;
      }

      get_token();


      convert_type_into_context (field_type, &field_context);  // works even if field_type is null

      memcpy (&pos, &token.pos, sizeof(pos));
      parse_expression (&field_context, &pos, null, &e);


      if (ok)
      {
        if (check_assignment_context_compatibility (&field_context, e, &pos) < 0)
          ok = false;
        is_constant &= is_constant_exp (e);
        append_expression (&list, e, field_type, field);

        if (field != null)
          field = field->next;
        skip_to_valid_field (&field, &field_type, context->constraint.value);
      }
      else
      {
        free_exp (e);
      }

      if (token.kind == RIGHT_ACCOLADE)
        break;

      if (token.kind != COMMA)
      {
        syntax_error ("',' or '}' expected", &token.pos);
        break;
      }

      get_token();
    }


    if (ok && field != null)
    {
      semantic_error ("missing fields in aggregate", &token.pos);
      ok = false;
    }


    if (!is_constant)
    {
      if (context->base_type_or_null != null && is_jagged_type (context->base_type_or_null))
        semantic_error ("jagged aggregates must be constant", paggregate_pos);
    }


    if (is_constant && ok)
    {
      LIST_OF_EXPRESSIONS *l;

      begin_struct (&data, context->base_type_or_null);

      l = list.head;
      while (l != null)
      {
        begin_field (&data, l->e);
        end_field (&data);
        l = l->next;
      }

      end_struct (&data);

      if (data.invalid_type)
      {
        semantic_error ("struct type is not constant", paggregate_pos);
        ok = false;
      }
      else if (data.size_overflow)
      {
        semantic_error ("aggregate is too large", paggregate_pos);
        ok = false;
      }
    }

    if (is_constant && ok)
    {
      POOL *p;
      LIST_OF_EXPRESSIONS *l;

      e = new_exp (A_POOL_CONSTANT);
      e->base_type_or_null = context->base_type_or_null;

      if (context->base_type_or_null->kind == A_STRUCT_TYPE &&
          context->base_type_or_null->info->the_struct_type.is_open_type)
      {
        e->constraint.kind = CONSTANT_CONSTRAINT;
        e->constraint.value = context->constraint.value;
      }

      e->form = A_VALUE;
      e->access = ACCESS_CONSTANT;

      p = new_pool_constant (data.struct_size, data.struct_alignment);

      e->info.pool_constant_info.pool_cte = p;

      begin_struct (&data, context->base_type_or_null);

      l = list.head;
      while (l != null)
      {
        begin_field (&data, l->e);

        convert_type_into_context (l->type, &field_context);
        store_constant_into_aggregate_pool (&field_context, l->exp, p, data.field_offset);

        end_field (&data);
        l = l->next;
      }

      end_struct (&data);

      free_list (&list);
    }
    else     // runtime aggregate
    {
      e = new_exp (AN_AGGREGATE_VALUE);
      e->base_type_or_null = ok ? context->base_type_or_null : null;

      if (context->base_type_or_null != null &&
          context->base_type_or_null->kind == A_STRUCT_TYPE &&
          context->base_type_or_null->info->the_struct_type.is_open_type)
      {
        e->constraint.kind = CONSTANT_CONSTRAINT;
        e->constraint.value = context->constraint.value;
      }

      e->form   = A_VALUE;
      e->access = ACCESS_READONLY;
      e->info.aggregate_value_info.list = list.head;
    }

    *out = e;
  }

  if (token.kind == RIGHT_ACCOLADE)
    get_token();
}

/*****************************************************************************/

// aggregate ::= positional_array_aggregate
//             | open_array_aggregate
//             | positional_struct_aggregate
//             | named_struct_aggregate

static void parse_aggregate (PCONTEXT    context,        // never null
                             PEXPRESSION *out)
{
  bool          context_ok;
  TEXT_POSITION aggregate_pos;

  if (token.kind != LEFT_ACCOLADE)
  {
    syntax_error ("'{' expected", &token.pos);
    *out = dummy_expression ();
    return;
  }


  // check that the context defines either:
  // - an array
  // - an open array
  // - a non-open struct context
  // (an open struct context is not allowed)

  context_ok = true;   // default

  if (context->base_type_or_null == null)
  {
    syntax_error ("an array or struct context is expected here", &token.pos);
    context_ok = false;
  }
  else if (context->base_type_or_null->kind == AN_OPEN_ARRAY_TYPE)
  {
    // array context
  }
  else if (context->base_type_or_null->kind == A_STRUCT_TYPE)    // struct context
  {
    if (context->base_type_or_null->info->the_struct_type.is_open_type)
    {
      if (context->constraint.kind != CONSTANT_CONSTRAINT)
      {
        syntax_error ("an open struct context without constant constraint is not allowed here", &token.pos);
        context_ok = false;
      }
    }
  }
  else
  {
    syntax_error ("an array or struct context is expected here", &token.pos);
    context_ok = false;
  }


  memcpy (&aggregate_pos, &token.pos, sizeof(aggregate_pos));

  get_token();  // skip '{'

  if (token.kind != IDENTIFIER)
  {
    parse_non_named_aggregate (context, context_ok, &aggregate_pos, out);
    return;
  }

  get_look_ahead_token ();
  if (look_ahead_token.kind == ARROW)
  {
    parse_named_aggregate (context, context_ok, &aggregate_pos, out);
  }
  else
  {
    parse_non_named_aggregate (context, context_ok, &aggregate_pos, out);
  }
}

/*****************************************************************************/

//  runtime_type ::= type_name
//                 | type_name "["  "]"
//                 | type_name "[" expression "]"
//                 | type_name "(" expression ")"

// returns one of the following, with field .value being null :
//
//  typedef struct {
//    PEXPRESSION  value;        // can be null if used inside an allocator
//  } QUALIFIED_EXPRESSION_INFO;
//
//  typedef struct {
//    PEXPRESSION  length;       // runtime array length
//    PEXPRESSION  value;        // can be null if used inside an allocator
//  } ARRAY_QUALIFIED_EXPRESSION_INFO;
//
//  typedef struct {
//    PEXPRESSION  discriminant; // runtime discriminant value
//    PEXPRESSION  value;        // can be null if used inside an allocator
//  } STRUCT_QUALIFIED_EXPRESSION_INFO;
//

static void parse_runtime_type (PENTITY       type_name,  // is NEVER null !
                                TEXT_POSITION *pruntime_exp_pos,
                                PEXPRESSION   *pout)
{
  PEXPRESSION out;

  clear (*pruntime_exp_pos);

  if (token.kind == LEFT_BRACKET)   // array
  {
    AN_OPEN_ARRAY_TYPE_ENTITY open_array;
    PENTITY t;


    // create open array
    memset (&open_array, 0x00, sizeof(open_array));
    open_array.element = type_name;

    t = append_new_entity (AN_OPEN_ARRAY_TYPE, &open_array, L"", &token.pos);


    get_token();   // skip '['

    if (token.kind == RIGHT_BRACKET)   // ']'
    {
      get_token();   // skip ']'

      out = new_exp (A_QUALIFIED_EXPRESSION);
      out->base_type_or_null = t;
      out->constraint.kind = UNCONSTRAINED;
      out->constraint.value = 0;
      out->form = A_VALUE;
      out->access = ACCESS_READONLY;

      *pout = out;
    }
    else    // an array
    {
      PEXPRESSION   e;
      TEXT_POSITION pos;
      int8          value;

      memcpy (&pos, &token.pos, sizeof(pos));
      parse_expression (&no_context, &pos, null, &e);

      if (e->base_type_or_null != null)
      {
        if (e->base_type_or_null->kind != AN_INTEGER_TYPE ||
            e->base_type_or_null == type_long)
        {
          semantic_error ("an int or uint expression is expected here", &pos);
        }
      }


      // 2 cases : constant or runtime length

      if (e->kind == A_CONST_INTEGER_VALUE)    // constant length
      {
        value = e->info.const_integer_value_info.value;

        if (value < 0 || value > 2147483647)
          semantic_error ("expression is out of range", &pos);

        out = new_exp (A_QUALIFIED_EXPRESSION);
        out->base_type_or_null = t;
        out->constraint.kind = CONSTANT_CONSTRAINT;
        out->constraint.value = (uint4)value;
        out->form = A_VALUE;
        out->access = ACCESS_READONLY;

        free_exp (e);

        *pout = out;
      }
      else   // runtime length
      {
        out = new_exp (AN_ARRAY_QUALIFIED_EXPRESSION);
        out->base_type_or_null = t;
        out->constraint.kind = RUNTIME_CONSTRAINT;
        out->constraint.value = 0;
        out->form = A_VALUE;
        out->access = ACCESS_READONLY;
        out->info.array_qualified_expression_info.length = e;

        memcpy (pruntime_exp_pos, &pos, sizeof(pos));

        *pout = out;
      }

      if (token.kind == RIGHT_BRACKET)   // ']'
        get_token();   // skip ']'
      else
        syntax_error ("']' expected", &token.pos);
    }
  }
  else if (token.kind == LEFT_PARENTHESIS)   // open array or open struct
  {
    PEXPRESSION   e;
    TEXT_POSITION pos0, pos;
    int8          value;
    PENTITY       t;

    memcpy (&pos0, &token.pos, sizeof(pos0));

    get_token();

    memcpy (&pos, &token.pos, sizeof(pos));
    parse_expression (&no_context, &pos, null, &e);

    t = complete_type_of (type_name);

    if (t->kind == AN_OPEN_ARRAY_TYPE)
    {
      if (e->base_type_or_null != null)
      {
        if (e->base_type_or_null->kind != AN_INTEGER_TYPE ||
            e->base_type_or_null == type_long)
        {
          semantic_error ("an int or uint expression is expected here", &pos);
        }
      }


      // 2 cases : constant or runtime length

      if (e->kind == A_CONST_INTEGER_VALUE)    // constant length
      {
        value = e->info.const_integer_value_info.value;

        if (value < 0 || value > 2147483647)
          semantic_error ("expression is out of range", &pos);

        out = new_exp (A_QUALIFIED_EXPRESSION);
        out->base_type_or_null = t;
        out->constraint.kind = CONSTANT_CONSTRAINT;
        out->constraint.value = (uint4)value;
        out->form = A_VALUE;
        out->access = ACCESS_READONLY;

        free_exp (e);

        *pout = out;
      }
      else   // runtime length
      {
        out = new_exp (AN_ARRAY_QUALIFIED_EXPRESSION);
        out->base_type_or_null = t;
        out->constraint.kind = RUNTIME_CONSTRAINT;
        out->constraint.value = 0;
        out->form = A_VALUE;
        out->access = ACCESS_READONLY;
        out->info.array_qualified_expression_info.length = e;

        memcpy (pruntime_exp_pos, &pos, sizeof(pos));

        *pout = out;
      }
    }
    else if (t->kind == A_STRUCT_TYPE && t->info->the_struct_type.is_open_type)
    {
      PENTITY discrim, discrim_type;

      // get the discriminant
      discrim = t->info->the_struct_type.discriminant->entities.first;

      // get the discriminant's type (it is garanteed to be an enumeration type)
      discrim_type = discrim->info->the_field.type;

      if (e->base_type_or_null != null)
      {
        if (e->base_type_or_null != discrim_type)
        {
          semantic_error ("this expression must be compatible with the discriminant", &pos);
        }
      }


      // 2 cases : constant or runtime discriminant

      if (e->kind == A_CONST_ENUMERATION_VALUE)    // constant discriminant
      {
        value = e->info.const_enumeration_value_info.value;

        if (value < 0 || value > discrim_type->info->the_enumeration_type.last)
          semantic_error ("expression is out of range", &pos);

        out = new_exp (A_QUALIFIED_EXPRESSION);
        out->base_type_or_null = t;
        out->constraint.kind = CONSTANT_CONSTRAINT;
        out->constraint.value = (uint4)value;
        out->form = A_VALUE;
        out->access = ACCESS_READONLY;

        free_exp (e);

        *pout = out;
      }
      else   // runtime length
      {
        out = new_exp (A_STRUCT_QUALIFIED_EXPRESSION);
        out->base_type_or_null = t;
        out->constraint.kind = RUNTIME_CONSTRAINT;
        out->constraint.value = 0;
        out->form = A_VALUE;
        out->access = ACCESS_READONLY;
        out->info.struct_qualified_expression_info.discriminant = e;

        memcpy (pruntime_exp_pos, &pos, sizeof(pos));

        *pout = out;
      }
    }
    else
    {
      CONTEXT context;

      semantic_error ("this constraint is only allowed for an open array or struct type", &pos0);

      convert_type_into_context (type_name, &context);

      out = new_exp (A_QUALIFIED_EXPRESSION);
      out->base_type_or_null = context.base_type_or_null;
      memcpy (&out->constraint, &context.constraint, sizeof(CONSTRAINT));
      out->form = A_VALUE;
      out->access = ACCESS_READONLY;

      *pout = out;

      free_exp (e);
    }

    if (token.kind == RIGHT_PARENTHESIS)   // ')'
      get_token();   // skip ')'
    else
      syntax_error ("')' expected", &token.pos);
  }
  else    // return type itself
  {
    CONTEXT context;

    convert_type_into_context (type_name, &context);

    out = new_exp (A_QUALIFIED_EXPRESSION);
    out->base_type_or_null = context.base_type_or_null;
    memcpy (&out->constraint, &context.constraint, sizeof(CONSTRAINT));
    out->form = A_VALUE;
    out->access = ACCESS_READONLY;

    *pout = out;
  }
}

/*****************************************************************************/

// merge constraints of (context and runtime_type) into a new context.

// possible values are : DOES_NOT_APPLY, UNCONSTRAINED,
//                       CONSTANT_CONSTRAINT, RUNTIME_CONSTRAINT.

// note: the context only helps if the runtime type is unconstrained.

static void merge_constraints (PCONTEXT    context,     // in
                               PEXPRESSION qual,        // in (runtime type)
                               PCONTEXT    context2)    // out
{
  memset (context2, 0x00, sizeof(CONTEXT));

  context2->base_type_or_null = qual->base_type_or_null;

  if (qual->constraint.kind == DOES_NOT_APPLY)   // not array/struct type
  {
    memcpy (&context2->constraint, &qual->constraint, sizeof(CONSTRAINT));
  }
  else if (qual->constraint.kind == UNCONSTRAINED)   // s = string ' ( exp ) + ".";
  {
    if (context->constraint.kind == CONSTANT_CONSTRAINT ||
        context->constraint.kind == RUNTIME_CONSTRAINT)
    {
      // only if base types are compatible, otherwise byte[] conversion of parameters
      // might use wrong context constraint.

      if (context->base_type_or_null != null && qual->base_type_or_null != null &&
          types_are_equal (context->base_type_or_null, qual->base_type_or_null))
      {
        memcpy (&context2->constraint, &context->constraint, sizeof(CONSTRAINT));
      }
      else
      {
        memcpy (&context2->constraint, &qual->constraint, sizeof(CONSTRAINT));
      }
    }
    else   // context is DOES_NOT_APPLY or UNCONSTRAINED.
    {
      memcpy (&context2->constraint, &qual->constraint, sizeof(CONSTRAINT));
    }
  }
  else   // CONSTANT_CONSTRAINT or RUNTIME_CONSTRAINT
  {
    memcpy (&context2->constraint, &qual->constraint, sizeof(CONSTRAINT));
  }
}

/*****************************************************************************/

//  qualified_expression ::= runtime_type  "'"  "(" expression ")"
//                         | runtime_type  "'"  aggregate

static void parse_qualified_expression (PCONTEXT    context,        // never null
                                        PENTITY     type_name,      // is NEVER null !
                                        PEXPRESSION *out)
{
  PEXPRESSION   qual, e;
  CONTEXT       inner_context, qual_context;
  TEXT_POSITION runtime_exp_pos, pos;

  parse_runtime_type (type_name, &runtime_exp_pos, &qual);
  if (qual->kind != A_QUALIFIED_EXPRESSION)
    semantic_error ("expression must be constant or be removed", &runtime_exp_pos);

  lex_force_into_apostrophe ();
  if (token.kind != APOSTROPHE)
  {
    syntax_error ("apostrophe ' expected", &token.pos);
    *out = qual;
    return;
  }

  get_token();   // skip apostrophe


  if (is_limited_type (type_name))
    semantic_error ("an expression of a limited type is not allowed here", &token.pos);


  merge_constraints (context, qual, &inner_context);


  if (token.kind == LEFT_PARENTHESIS)
  {
    memcpy (&pos, &token.pos, sizeof(pos));

    get_token();   // skip '('

    parse_expression (&inner_context, &token.pos, null, &e);

    if (token.kind == RIGHT_PARENTHESIS)   // ')'
      get_token();   // skip ')'
    else
      syntax_error ("')' expected", &token.pos);
  }
  else   // aggregate
  {
    memcpy (&pos, &token.pos, sizeof(pos));
    parse_aggregate (&inner_context, &e);
  }


  if (qual->kind == A_QUALIFIED_EXPRESSION)
    qual->info.qualified_expression_info.value = e;
  else if (qual->kind == AN_ARRAY_QUALIFIED_EXPRESSION)
    qual->info.array_qualified_expression_info.value = e;
  else if (qual->kind == A_STRUCT_QUALIFIED_EXPRESSION)
    qual->info.struct_qualified_expression_info.value = e;


  memset (&qual_context, 0x00, sizeof(qual_context));
  qual_context.base_type_or_null = qual->base_type_or_null;
  memcpy (&qual_context.constraint, &qual->constraint, sizeof(CONSTRAINT));


  // check if expression is assignment-compatible with runtime type.

  if (check_assignment_context_compatibility (&qual_context, e, &pos) == 0)  // ok
  {
    // in case the expression is a constant and there is no runtime constraint,
    // remove the node to keep just the constant;
    // this applies for any simple or array/struct type.

    if (is_constant_exp (e) && qual->constraint.kind != RUNTIME_CONSTRAINT)
    {
      e->base_type_or_null = qual->base_type_or_null;  // keep exact int/float/enum type
      free (qual);
      qual = e;
    }
    else
    {
      // if runtime type is unconstrained, use constraint from expression which is more precise
      if (qual->constraint.kind == UNCONSTRAINED)
        memcpy (&qual->constraint, &e->constraint, sizeof(CONSTRAINT));

      // if runtime type is runtime, use constraint from expression which is maybe constant
      if (qual->constraint.kind == RUNTIME_CONSTRAINT)
        memcpy (&qual->constraint, &e->constraint, sizeof(CONSTRAINT));
    }
  }

  *out = qual;
}

/*****************************************************************************/

//  new_allocator ::= "new"  qualified_expression
//                  | "new"  runtime_type

static void parse_new_allocator (PCONTEXT    context,        // never null
                                 PEXPRESSION *pout)
{
  PENTITY       type_name, designated_type;
  PEXPRESSION   qual, alloc, e;
  CONTEXT       designated_type_context;
  CONTEXT       inner_context, qual_context;
  TEXT_POSITION pos0, pos, posq, typ_pos, runtime_exp_pos;

  if (context->base_type_or_null == null ||
      complete_type_of (context->base_type_or_null)->kind != A_POINTER_TYPE)
  {
    syntax_error ("a pointer type context is required here", &token.pos);
    *pout = dummy_expression ();
    return;
  }

  designated_type =
      complete_type_of (
             complete_type_of (context->base_type_or_null)
              ->info->the_pointer_type.designated_type
                       );


  memcpy (&pos0, &token.pos, sizeof(pos0));
  get_token();   // skip 'new'

  memcpy (&typ_pos, &token.pos, sizeof(typ_pos));
  type_name = parse_type_name ();
  if (type_name == null)
  {
    *pout = dummy_expression ();
    return;
  }

  if (is_jagged_type (type_name))
    semantic_error ("a jagged type is not allowed here", &typ_pos);


  memcpy (&posq, &token.pos, sizeof(posq));
  parse_runtime_type (type_name, &runtime_exp_pos, &qual);


  // note: the designated_type can be an open type, but cannot be jagged.

  convert_type_into_context (designated_type, &designated_type_context);


  merge_constraints (&designated_type_context, qual, &inner_context);


  lex_force_into_apostrophe ();
  if (token.kind == APOSTROPHE)  // a qualified expression
  {
    get_token();   // skip apostrophe

    if (is_limited_type (type_name))
      semantic_error ("an expression of a limited type is not allowed here", &token.pos);

    if (token.kind == LEFT_PARENTHESIS)
    {
      if (qual->kind != A_QUALIFIED_EXPRESSION)
        semantic_error ("expression must be constant or be removed", &runtime_exp_pos);

      memcpy (&pos, &token.pos, sizeof(pos));

      get_token();   // skip '('

      parse_expression (&inner_context, &token.pos, null, &e);

      if (token.kind == RIGHT_PARENTHESIS)   // ')'
        get_token();   // skip ')'
      else
        syntax_error ("')' expected", &token.pos);
    }
    else   // aggregate
    {
      memcpy (&pos, &token.pos, sizeof(pos));
      parse_aggregate (&inner_context, &e);

      if (qual->kind != A_QUALIFIED_EXPRESSION)   // runtime constraint
      {
        if (e->kind != AN_UNC_ARRAY_AGGREGATE || (!is_open_type (designated_type)))
          semantic_error ("expression must be constant or be removed", &runtime_exp_pos);
      }
    }


    if (qual->kind == A_QUALIFIED_EXPRESSION)
      qual->info.qualified_expression_info.value = e;
    else if (qual->kind == AN_ARRAY_QUALIFIED_EXPRESSION)
      qual->info.array_qualified_expression_info.value = e;
    else if (qual->kind == A_STRUCT_QUALIFIED_EXPRESSION)
      qual->info.struct_qualified_expression_info.value = e;


    memset (&qual_context, 0x00, sizeof(qual_context));
    qual_context.base_type_or_null = qual->base_type_or_null;
    memcpy (&qual_context.constraint, &qual->constraint, sizeof(CONSTRAINT));

    if (check_assignment_context_compatibility (&qual_context, e, &pos) == 0)  // ok
    {
      // if runtime type is unconstrained, use constraint from expression which is more precise
      if (qual->constraint.kind == UNCONSTRAINED)
        memcpy (&qual->constraint, &e->constraint, sizeof(CONSTRAINT));

      // if runtime type is runtime, use constraint from expression which is maybe constant
      if (qual->constraint.kind == RUNTIME_CONSTRAINT)
        memcpy (&qual->constraint, &e->constraint, sizeof(CONSTRAINT));


      // check that designated_type base-type matches with qual base-type.

      if (!types_are_equal (base_type_of (designated_type), qual->base_type_or_null))
        semantic_error ("pointer types do not match", &pos0);
      else
      {
        // if designated_type is constrained, constraints must match.
        if (designated_type_context.constraint.kind == CONSTANT_CONSTRAINT &&
            qual->constraint.kind == CONSTANT_CONSTRAINT &&
            designated_type_context.constraint.value != qual->constraint.value)
        {
          semantic_error ("designated arrays/structs have different length/discriminant", &pos0);
        }
      }
    }


    // add an extra allocator node

    alloc = new_exp (AN_ALLOCATOR);
    alloc->base_type_or_null = complete_type_of (context->base_type_or_null);
    alloc->form = A_VALUE;
    alloc->access = ACCESS_READONLY;
    alloc->info.allocator_info.value = qual;
    alloc->info.allocator_info.has_header = is_open_type (designated_type);

    *pout = alloc;
  }
  else   // just a runtime_type (allocates a zero-filled heap object)
  {
    // check that a runtime constraint is only allowed for an open designated type

    if (qual->kind != A_QUALIFIED_EXPRESSION &&   // runtime constraint
        (!is_open_type (designated_type)))        // designated type is not open
      semantic_error ("expression must be constant", &runtime_exp_pos);


    // check that the runtime type is not an open type (because it has no size)

    if (qual->constraint.kind == UNCONSTRAINED)
      semantic_error ("a length/discriminant constraint is required here", &posq);


    // check that designated_type base-type matches with qual base-type.

    if (!types_are_equal (base_type_of (designated_type), qual->base_type_or_null))
      semantic_error ("pointer types do not match", &pos0);
    else
    {
      // if designated_type is constrained, constraints must match.
      if (designated_type_context.constraint.kind == CONSTANT_CONSTRAINT &&
          qual->constraint.kind == CONSTANT_CONSTRAINT &&
          designated_type_context.constraint.value != qual->constraint.value)
      {
        semantic_error ("designated arrays/structs have different length/discriminant", &pos0);
      }
    }


    // add an extra allocator node

    alloc = new_exp (AN_ALLOCATOR);
    alloc->base_type_or_null = complete_type_of (context->base_type_or_null);
    alloc->form = A_VALUE;
    alloc->access = ACCESS_READONLY;
    alloc->info.allocator_info.value = qual;
    alloc->info.allocator_info.has_header = is_open_type (designated_type);

    *pout = alloc;
  }
}

/*****************************************************************************/

static void parse_primary_starting_with_a_name (PCONTEXT      context,        // never null
                                                TEXT_POSITION *prefix_pos,
                                                PENTITY       prefix_name,    // is NEVER null !
                                                PEXPRESSION   *pout)
{
  if (prefix_name->kind > LAST_ENTITY_DENOTING_A_TYPE)    // it's not a type
  {
    parse_name (context, false, prefix_pos, prefix_name, pout);
    return;
  }


  // either a name or a qualified expression

  if (token.kind == LEFT_BRACKET || token.kind == LEFT_PARENTHESIS)
  {
    parse_qualified_expression (context, prefix_name, pout);
    return;
  }

  lex_force_into_apostrophe ();
  if (token.kind == APOSTROPHE)
  {
    get_look_ahead_token ();
    if (look_ahead_token.kind == LEFT_PARENTHESIS || look_ahead_token.kind == LEFT_ACCOLADE)
    {
      parse_qualified_expression (context, prefix_name, pout);
      return;
    }
  }

  parse_name (context, false, prefix_pos, prefix_name, pout);
}

/*****************************************************************************/

//  primary ::= integer_literal           (integer)
//            | floating_point_literal    (floating-point)
//            | character_literal         (char, wchar)
//            | string_literal            (string, wstring)
//            | "nul"                     (char)
//            | "Lnul"                    (wchar)
//            | "null"                    (pointer, function/unsafe pointer)
//            | name                      (value or object)
//            | run_call                  (int)
//            | aggregate                 (array, struct)
//            | qualified_expression      (value)
//            | new_allocator             (pointer)

static void parse_primary (PCONTEXT      context,        // never null
                           TEXT_POSITION *prefix_pos,
                           PENTITY       prefix_name,    // can be null
                           PEXPRESSION   *pout)
{
  PEXPRESSION out;
  bool        wide;
  uint        len, charsize, i;
  POOL        *p;

  if (prefix_name != null)   // it's a name
  {
    parse_primary_starting_with_a_name (context, prefix_pos, prefix_name, pout);
    return;
  }

  switch (token.kind)
  {
    case INTEGER_LITERAL:
      out = new_exp (A_CONST_INTEGER_VALUE);
      out->base_type_or_null = token.info._integer.has_suffix_L ? type_long : type_int_literal;
      out->form = A_VALUE;
      out->access = ACCESS_CONSTANT;
      out->info.const_integer_value_info.value = token.info._integer.value;

      get_token();
      break;

    case FLOAT_LITERAL:
      out = new_exp (A_CONST_FLOAT_VALUE);

      if (token.info._float.suffix == 'F')
        out->base_type_or_null = type_float;
      else if (token.info._float.suffix == 'D')
        out->base_type_or_null = type_double;
      else
        out->base_type_or_null = type_float_literal;     // compatible with any float type

      out->form = A_VALUE;
      out->access = ACCESS_CONSTANT;
      out->info.const_float_value_info.value = token.info._float.value;

      get_token();
      break;

    case CHAR_LITERAL:
      out = new_exp (A_CONST_ENUMERATION_VALUE);
      out->base_type_or_null = token.info._char.has_prefix_L ? type_wchar : type_char;
      out->form = A_VALUE;
      out->access = ACCESS_CONSTANT;
      out->info.const_enumeration_value_info.value = token.info._char.value;

      get_token();
      break;

    case STRING_LITERAL:
      wide = token.info._string.has_prefix_L;
      len = (uint4)token.info._string.length;
      charsize = wide ? 2 : 1;

      out = new_exp (A_POOL_CONSTANT);
      out->base_type_or_null = wide ? type_wstring : type_string;
      out->constraint.kind = CONSTANT_CONSTRAINT;
      out->constraint.value = len;
      out->form = A_VALUE;
      out->access = ACCESS_CONSTANT;
      out->info.pool_constant_info.pool_cte = new_pool_constant (charsize * len, charsize);

      p = out->info.pool_constant_info.pool_cte;
      for (i=0; i<len; i++)
        store_integer (p, i*charsize, token.info._string.value[i], charsize);

      get_token();
      break;

    case TOKEN_nul:      // (char)
      out = new_exp (A_CONST_ENUMERATION_VALUE);
      out->base_type_or_null = type_char;
      out->form = A_VALUE;
      out->access = ACCESS_CONSTANT;
      out->info.const_enumeration_value_info.value = 0;

      get_token();
      break;

    case TOKEN_Lnul:     // (wchar)
      out = new_exp (A_CONST_ENUMERATION_VALUE);
      out->base_type_or_null = type_wchar;
      out->form = A_VALUE;
      out->access = ACCESS_CONSTANT;
      out->info.const_enumeration_value_info.value = 0;

      get_token();
      break;

    case TOKEN_null:     // (pointer, function/unsafe pointer)
      out = new_exp (A_CONST_NULL_VALUE);
      out->base_type_or_null = type_null_literal;     // compatible with any pointer, function pointer or unsafe pointer.
      out->form = A_VALUE;
      out->access = ACCESS_CONSTANT;

      get_token();
      break;

    case TOKEN_run:      // run_call
      parse_run_call (&out);
      break;

    case TOKEN_new:      // new_allocator (pointer)
      parse_new_allocator (context, &out);
      break;

    case LEFT_ACCOLADE:   // aggregate (array, struct)
      parse_aggregate (context, &out);
      break;

    case LEFT_PARENTHESIS:  // parentheized expression name
      parse_name (context, false, &token.pos, null, &out);
      break;

    default:    // maybe it's a name or a qualified expression
      {
        PENTITY       e2;
        TEXT_POSITION name_pos;

        memcpy (&name_pos, &token.pos, sizeof(name_pos));
        e2 = type_entity_of_token (token.kind);
        if (e2 != null)
        {
          get_token();    // skip type token
        }
        else if (token.kind == IDENTIFIER)
        {
          e2 = parse_expanded_name ();
          if (e2 == null)  // could not find entity (error already given)
          {
            parse_name (context, false, &name_pos, e2, pout);
            return;
          }
        }
        else
        {
          parse_name (context, false, &name_pos, null, pout);
          return;
        }

        // e2 denotes a type

        parse_primary_starting_with_a_name (context, &name_pos, e2, pout);
        return;
      }
  }

  *pout = out;
}

/*****************************************************************************/

typedef struct {
  ENTITY_KIND   source;
  ENTITY_KIND   target;
  KIND_OPERATOR op;
} CONVERSION;

static CONVERSION conversion_table[] =
{
 {AN_INTEGER_TYPE,             AN_INTEGER_TYPE,             OP_CONVERT_INT_INT},
 {AN_ENUMERATION_TYPE,         AN_INTEGER_TYPE,             OP_CONVERT_INT_INT},
 {A_FLOAT_TYPE,                AN_INTEGER_TYPE,             OP_CONVERT_FLOAT_INT},
 {AN_INTEGER_TYPE,             AN_ENUMERATION_TYPE,         OP_CONVERT_INT_INT},
 {A_FLOAT_TYPE,                A_FLOAT_TYPE,                OP_CONVERT_FLOAT_FLOAT},
 {AN_INTEGER_TYPE,             A_FLOAT_TYPE,                OP_CONVERT_INT_FLOAT},
 {AN_UNSAFE_POINTER_TYPE,      AN_UNSAFE_POINTER_TYPE,      OP_CONVERT_PTR_PTR},
 {A_NULL_POINTER_TYPE,         AN_UNSAFE_POINTER_TYPE,      OP_NONE},  // null to unsafe ptr
 {LAST_ENTITY_DENOTING_A_TYPE, LAST_ENTITY_DENOTING_A_TYPE, OP_NONE},
};

/*****************************************************************************/

// unary_expression  ::= { unary_operator | conversion }  primary
// unary_operator  ::=  "+" | "-" | "!" | "~" | "&" | "*" | "--" | "++"
// conversion      ::= "("  restricted_type_definition  ")"
// restricted_type_definition ::= type_name {"*"}

void parse_unary_expression (PCONTEXT      context,        // never null
                             TEXT_POSITION *prefix_position,
                             PENTITY       prefix_name,    // can be null
                             PEXPRESSION   *out)
{
  KIND_OPERATOR op;
  PEXPRESSION   e, e2;
  PENTITY       t, t2;
  TEXT_POSITION pos, posc, pos_name;
  int8          r, r2;
  double        rf;
  int           rc;
  INTEGER_INFO  *pi;
  int           i;

  if (prefix_name != null)
  {
    parse_primary_starting_with_a_name (context, prefix_position, prefix_name, out);
    return;
  }

  if (token.kind == LEFT_PARENTHESIS)    // either conversion or name
  {
    memcpy (&pos, &token.pos, sizeof(pos));
    get_token();    // skip (


    // check if now comes a type_name followed either by '*' or ')'

    memcpy (&pos_name, &token.pos, sizeof(pos_name));
    t2 = type_entity_of_token (token.kind);
    if (t2 != null)
    {
      get_token();    // skip type token
    }
    else if (token.kind == IDENTIFIER)
    {
      t2 = parse_expanded_name ();

      if (t2 == null ||  // could not find entity (error already given)
          t2->kind > LAST_ENTITY_DENOTING_A_TYPE)
      {
        parse_name (context, true, &pos_name, t2, out);
        return;
      }
    }
    else
    {
      parse_name (context, true, &pos_name, null, out);
      return;
    }

    // we have parsed a type_name

    if (token.kind == STAR || token.kind == RIGHT_PARENTHESIS)
    {
      // it's a type conversion

      t2 = complete_type_of (t2);

      while (token.kind == STAR)   // produce unsafe pointer type
      {
        AN_UNSAFE_POINTER_TYPE_ENTITY ptr;

        if (is_jagged_type (t2))
          semantic_error ("a jagged type is not allowed here", &pos_name);

        clear (ptr);
        ptr.designated_type = t2;

        t2 = append_new_entity (AN_UNSAFE_POINTER_TYPE, &ptr, L"", &token.pos);

        get_token();    // skip *
      }

      if (token.kind == RIGHT_PARENTHESIS)
        get_token();    // skip )
      else
        syntax_error ("')' expected", &token.pos);

      parse_unary_expression (&no_context,  // no context for a conversion
                              &token.pos,
                              null,  // no prefix
                              &e);

      t = e->base_type_or_null;


      if (is_unsafe_type (t2))
        check_unsafe_region ();


      // check that the conversion from 't' to 't2' is allowed

      op = OP_NONE;    // default

      if (t != null)
      {
        for (i=0; conversion_table[i].source != LAST_ENTITY_DENOTING_A_TYPE; i++)
        {
          if (conversion_table[i].source == t->kind &&
              conversion_table[i].target == t2->kind)
            break;
        }

        if (conversion_table[i].source == LAST_ENTITY_DENOTING_A_TYPE)
        {
          semantic_error ("conversion is not allowed", &pos_name);
          t = null;     // indicates forbidden conversion
        }

        op = conversion_table[i].op;
      }


      if (t != null &&
          e->kind == A_CONST_NULL_VALUE &&    // null to unsafe-ptr
          t->kind == A_NULL_POINTER_TYPE)
      {
        e2 = new_exp (A_CONST_NULL_VALUE);
        e2->base_type_or_null = t2;
        e2->form = A_VALUE;
        e2->access = ACCESS_CONSTANT;

        free_exp (e);
      }
      else if (t != null &&
                   (e->kind == A_CONST_ENUMERATION_VALUE ||
                    e->kind == A_CONST_INTEGER_VALUE ||
                    e->kind == A_CONST_FLOAT_VALUE) &&       // conversion of a constant
                   (t->kind == AN_ENUMERATION_TYPE ||
                    t->kind == AN_INTEGER_TYPE ||
                    t->kind == A_FLOAT_TYPE))
      {
        r = 0;
        rf = 0.0;


        // retrieve value

        if (e->kind == A_CONST_ENUMERATION_VALUE)
          r = e->info.const_enumeration_value_info.value;

        if (e->kind == A_CONST_INTEGER_VALUE)
          r = e->info.const_integer_value_info.value;

        if (e->kind == A_CONST_FLOAT_VALUE)
          rf = e->info.const_float_value_info.value;


        // convert between float and int

        if (op == OP_CONVERT_FLOAT_INT)   // float to int
        {
          if (rf < integer_info[a_int8].min || rf > integer_info[a_int8].max)
            semantic_error ("overflow", &pos);
          else
            r = (int8)rf;
        }


        if (op == OP_CONVERT_INT_FLOAT)   // int to float
        {
          rf = (double)r;   // cannot overflow
        }


        // check that value fits into target type

        if (t2->kind == AN_ENUMERATION_TYPE)
        {
          if (r < 0 || r > t2->info->the_enumeration_type.last)
          {
            semantic_error ("overflow", &pos);
            r = 0;
          }

          e2 = new_exp (A_CONST_ENUMERATION_VALUE);
          e2->base_type_or_null = t2;
          e2->form = A_VALUE;
          e2->access = ACCESS_CONSTANT;
          e2->info.const_enumeration_value_info.value = (uint4)r;
        }
        else if (t2->kind == AN_INTEGER_TYPE)
        {
          // check if r outside range of t2

          pi = &integer_info[t2->info->the_integer_type.type];

          if (r < pi->min || r > pi->max)
          {
            semantic_error ("overflow", &pos);
            r = 0;
          }

          e2 = new_exp (A_CONST_INTEGER_VALUE);
          e2->base_type_or_null = t2;
          e2->form = A_VALUE;
          e2->access = ACCESS_CONSTANT;
          e2->info.const_integer_value_info.value = r;
        }
        else if (t2->kind == A_FLOAT_TYPE)
        {
          check_float_overflow_underflow (&rf, t2, &pos);

          e2 = new_exp (A_CONST_FLOAT_VALUE);
          e2->base_type_or_null = t2;
          e2->form = A_VALUE;
          e2->access = ACCESS_CONSTANT;
          e2->info.const_float_value_info.value = rf;
        }
        else
          fatal_compiler_error ("conversion", &token.pos);

        free_exp (e);
      }
      else   // runtime
      {
        e2 = new_exp (AN_OPERATOR_VALUE);
        e2->base_type_or_null = t2;
        e2->form = A_VALUE;
        e2->access = ACCESS_READONLY;
        e2->info.operator_value_info.op = op;
        e2->info.operator_value_info.arg[0] = e;
      }

      *out = e2;
    }
    else
    {
      // it's a parentheized name
      parse_name (context, true, &pos_name, t2, out);
    }

    return;
  }


  if (token.kind == STAR)  // --> AN_UNSAFE_DEREFERENCED_OBJECT
  {
    check_unsafe_region ();

    get_token();
    memcpy (&pos, &token.pos, sizeof(pos));
    parse_unary_expression (context, &pos, prefix_name, &e);

    t = e->base_type_or_null;
    if (t != null && t->kind != AN_UNSAFE_POINTER_TYPE)
      semantic_error ("unsafe pointer value expected", &pos);

    if (e->kind == A_CONST_NULL_VALUE)
      semantic_error ("dereferencing null pointer is not allowed", &token.pos);

    e2 = new_exp (AN_UNSAFE_DEREFERENCED_OBJECT);
    e2->form = AN_OBJECT;
    e2->access = ACCESS_READWRITE;
    e2->info.unsafe_dereferenced_object_info.unsafe_ptr_value = e;

    if (t != null && t->kind == AN_UNSAFE_POINTER_TYPE)
    {
      PENTITY dt;

      dt = t->info->the_unsafe_pointer_type.designated_type;
      if (is_open_type (dt))
        semantic_error ("operand of '*' must not be an unsafe pointer to an open type", &token.pos);

      convert_type_into_base_type_and_constraint
              (dt,
               /*always_constrained=*/ true,
               &e2->base_type_or_null,
               &e2->constraint);
    }

    *out = e2;

    return;
  }


  switch (token.kind)
  {
    case PLUS:      // +
      op = OP_UNARY_PLUS;
      break;

    case MINUS:     // -
      op = OP_UNARY_MINUS;
      break;

    case NOT:       // !
      op = OP_BOOL_NOT;
      break;

    case TILDE:     // ~
      op = OP_INT_NOT;
      break;

    case AMPERSAND: // &
      op = OP_ADDRESS_OF;
      break;

    case MINUS_MINUS:  // --
      op = OP_PRE_DEC;
      break;

    case PLUS_PLUS:    // ++
      op = OP_PRE_INC;
      break;

    default:
      op = OP_NONE;
      break;
  }

  if (op == OP_NONE)
  {
    parse_primary (context, prefix_position, prefix_name, out);
    return;
  }

  memcpy (&posc, &token.pos, sizeof(pos));
  get_token();
  memcpy (&pos, &token.pos, sizeof(pos));

  {
    CONTEXT context2;
    memcpy (&context2, context, sizeof(CONTEXT));
    if (op == OP_ADDRESS_OF)
      context2.address_of = true;
    parse_unary_expression (&context2, &pos, prefix_name, &e);
  }

  t = e->base_type_or_null;
  if (t == null)
  {
    *out = e;
    return;
  }


  switch (op)
  {
    case OP_UNARY_PLUS:
    case OP_UNARY_MINUS:

      if (t->kind == AN_INTEGER_TYPE)
      {
        t2 = promote_one_int (e);

        if (e->kind == A_CONST_INTEGER_VALUE)
        {
          r = e->info.const_integer_value_info.value;

          if (op == OP_UNARY_MINUS)
          {
            rc = neg_int (r, &r2);
            if (rc < 0)
            {
              semantic_error ("overflow", &posc);
              r = 0;
            }
            else
            {
              r = r2;
            }
          }


          // check if r outside range of t2

          pi = &integer_info[t2->info->the_integer_type.type];
          if (r < pi->min || r > pi->max)
          {
            semantic_error ("overflow", &posc);
            r = 0;
          }

          e2 = new_exp (A_CONST_INTEGER_VALUE);
          e2->base_type_or_null = t2;
          e2->form = A_VALUE;
          e2->access = ACCESS_CONSTANT;
          e2->info.const_integer_value_info.value = r;

          free_exp (e);

          *out = e2;
        }
        else  // runtime
        {
          e2 = new_exp (AN_OPERATOR_VALUE);
          e2->base_type_or_null = t2;
          e2->form = A_VALUE;
          e2->access = ACCESS_READONLY;
          e2->info.operator_value_info.op = op;
          e2->info.operator_value_info.arg[0] = e;

          *out = e2;
        }
      }
      else if (t->kind == A_FLOAT_TYPE)
      {
        if (e->kind == A_CONST_FLOAT_VALUE)
        {
          rf = e->info.const_float_value_info.value;

          if (op == OP_UNARY_MINUS)
            rf = -rf;

          e2 = new_exp (A_CONST_FLOAT_VALUE);
          e2->base_type_or_null = t;
          e2->form = A_VALUE;
          e2->access = ACCESS_CONSTANT;
          e2->info.const_float_value_info.value = rf;

          free_exp (e);

          *out = e2;
        }
        else  // runtime
        {
          e2 = new_exp (AN_OPERATOR_VALUE);
          e2->base_type_or_null = t;
          e2->form = A_VALUE;
          e2->access = ACCESS_READONLY;
          e2->info.operator_value_info.op = op;
          e2->info.operator_value_info.arg[0] = e;

          *out = e2;
        }
      }
      else
      {
        semantic_error ("illegal operand", &pos);
        *out = e;
      }
      break;


    case OP_BOOL_NOT:
      if (t == type_bool)
      {
        if (e->kind == A_CONST_ENUMERATION_VALUE)
        {
          r = e->info.const_enumeration_value_info.value;

          r = !r;

          e2 = new_exp (A_CONST_ENUMERATION_VALUE);
          e2->base_type_or_null = type_bool;
          e2->form = A_VALUE;
          e2->access = ACCESS_CONSTANT;
          e2->info.const_enumeration_value_info.value = (uint4)r;

          free_exp (e);

          *out = e2;
        }
        else  // runtime
        {
          e2 = new_exp (AN_OPERATOR_VALUE);
          e2->base_type_or_null = type_bool;
          e2->form = A_VALUE;
          e2->access = ACCESS_READONLY;
          e2->info.operator_value_info.op = op;
          e2->info.operator_value_info.arg[0] = e;

          *out = e2;
        }
      }
      else
      {
        semantic_error ("bool operand expected", &pos);
        *out = e;
      }
      break;


    case OP_INT_NOT:
      if (t->kind == AN_INTEGER_TYPE)
      {
        t2 = promote_one_int (e);

        if (e->kind == A_CONST_INTEGER_VALUE)
        {
          r = e->info.const_integer_value_info.value;

          r = ~r;

          // mask with range of t2

          pi = &integer_info[t2->info->the_integer_type.type];
          if (!pi->is_signed)
            r &= pi->mask;        // reduce r to the length of t2

          e2 = new_exp (A_CONST_INTEGER_VALUE);
          e2->base_type_or_null = t2;
          e2->form = A_VALUE;
          e2->access = ACCESS_CONSTANT;
          e2->info.const_integer_value_info.value = r;

          free_exp (e);

          *out = e2;
        }
        else  // runtime
        {
          e2 = new_exp (AN_OPERATOR_VALUE);
          e2->base_type_or_null = t2;
          e2->form = A_VALUE;
          e2->access = ACCESS_READONLY;
          e2->info.operator_value_info.op = op;
          e2->info.operator_value_info.arg[0] = e;

          *out = e2;
        }
      }
      else
      {
        semantic_error ("operand must have integer type", &pos);
        *out = e;
      }
      break;


    case OP_ADDRESS_OF:
      if (e->form != AN_OBJECT)
      {
        semantic_error ("an object is expected here", &pos);
        *out = e;
        return;
      }

      check_unsafe_region ();

      if (is_jagged_type (t))
        semantic_error ("a jagged object is not allowed here", &pos);

      {
        AN_UNSAFE_POINTER_TYPE_ENTITY ptr;

        clear (ptr);

        if (base_type_of (t)->kind == AN_OPEN_ARRAY_TYPE)
          t = base_type_of (t)->info->the_open_array_type.element;

        ptr.designated_type = t;

        t = append_new_entity (AN_UNSAFE_POINTER_TYPE, &ptr, L"", &posc);
      }

      e2 = new_exp (AN_OPERATOR_VALUE);
      e2->base_type_or_null = t;
      e2->form = A_VALUE;
      e2->access = ACCESS_READONLY;
      e2->info.operator_value_info.op = op;
      e2->info.operator_value_info.arg[0] = e;

      *out = e2;

      break;


    case OP_PRE_DEC:
    case OP_PRE_INC:
      if (e->access != ACCESS_READWRITE)
      {
        semantic_error ("a read/write object is expected here", &pos);
      }
      else if (e->form != AN_OBJECT)
      {
        semantic_error ("an object is expected here", &pos);
      }
      else if (t->kind != AN_INTEGER_TYPE &&
               t->kind != AN_ENUMERATION_TYPE &&
               t->kind != AN_UNSAFE_POINTER_TYPE)
      {
        semantic_error ("enum, integer or unsafe pointer expected here", &pos);
      }

      e2 = new_exp (AN_OPERATOR_VALUE);
      e2->base_type_or_null = t;
      e2->form = A_VALUE;
      e2->access = ACCESS_READONLY;
      e2->info.operator_value_info.op = op;
      e2->info.operator_value_info.arg[0] = e;

      *out = e2;

      break;


    default:
      fatal_compiler_error ("bad operand", &token.pos);
      *out = e;
      break;
  }
}

/*****************************************************************************/

static void parse_term (PCONTEXT      context,        // never null
                        TEXT_POSITION *prefix_position,
                        PENTITY       prefix_name,    // can be null
                        PEXPRESSION   *out)
{
  PEXPRESSION   e1, e2, e3;
  TOKEN_KIND    tk;
  KIND_OPERATOR op;
  PENTITY       t1, t2, t3;
  TEXT_POSITION pos;
  int8          v1, v2, r;
  double        vf1, vf2, rf;
  int           rc;
  INTEGER_INFO  *pi;
  bool          is_signed;

  parse_unary_expression (context, prefix_position, prefix_name, &e1);

  while (token.kind == STAR || token.kind == SLASH || token.kind == PERCENT)
  {
    tk = token.kind;
    memcpy (&pos, &token.pos, sizeof(pos));
    get_token();     // skip * / %

    parse_unary_expression (&no_context, &token.pos, null, &e2);

    // e1 = e1 op e2

    // check type

    t1 = e1->base_type_or_null;
    t2 = e2->base_type_or_null;

    if (t1 == null)   // error during earlier evaluation
    {
      free_exp (e2);
      e3 = e1;
    }
    else if (t2 == null)   // error during earlier evaluation
    {
      free_exp (e1);
      e3 = e2;
    }
    else
    {
      if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
      {
        t3 = promote_two_ints (e1, e2, true);
        if (t3 == null)
          semantic_error ("incompatible integer operands", &pos);

        if (tk == SLASH || tk == PERCENT)
        {
          if (e2->kind == A_CONST_INTEGER_VALUE && e2->info.const_integer_value_info.value == 0)
            semantic_error ("division by zero", &pos);
        }

        if (e1->kind == A_CONST_INTEGER_VALUE && e2->kind == A_CONST_INTEGER_VALUE)
        {
          v1 = e1->info.const_integer_value_info.value;
          v2 = e2->info.const_integer_value_info.value;

          if (tk == STAR)
          {
            rc = mul_int (v1, v2, &r);
          }
          else if (tk == SLASH)
          {
            rc = div_int (v1, v2, &r);
          }
          else
          {
            rc = mod_int (v1, v2, &r);
          }

          if (rc < 0)
          {
            semantic_error ("overflow", &pos);
            r = 0;
          }
          else if (t3 != null)
          {
            // check if r outside range of t3

            pi = &integer_info[t3->info->the_integer_type.type];
            if (r < pi->min || r > pi->max)
            {
              semantic_error ("overflow", &pos);
              r = 0;
            }
          }

          e3 = new_exp (A_CONST_INTEGER_VALUE);
          e3->base_type_or_null = t3;
          e3->form = A_VALUE;
          e3->access = ACCESS_CONSTANT;
          e3->info.const_integer_value_info.value = r;

          free_exp (e1);
          free_exp (e2);
        }
        else  // runtime
        {
          e3 = new_exp (AN_OPERATOR_VALUE);
          e3->base_type_or_null = t3;
          e3->form = A_VALUE;
          e3->access = ACCESS_READONLY;

          is_signed = false;
          if (t3 != null)
          {
            pi = &integer_info[t3->info->the_integer_type.type];
            if (pi->is_signed)
              is_signed = true;
          }

          if (tk == STAR)
            op = is_signed ? OP_MUL_INT_SIGNED : OP_MUL_INT_UNSIGNED;
          else if (tk == SLASH)
            op = is_signed ? OP_DIV_INT_SIGNED : OP_DIV_INT_UNSIGNED;
          else
            op = is_signed ? OP_MOD_INT_SIGNED : OP_MOD_INT_UNSIGNED;

          e3->info.operator_value_info.op = op;
          e3->info.operator_value_info.arg[0] = e1;
          e3->info.operator_value_info.arg[1] = e2;
        }
      }
      else if (t1->kind == A_FLOAT_TYPE && t2->kind == A_FLOAT_TYPE)
      {
        t3 = promote_two_floats (e1, e2, true);

        if (tk == PERCENT)
          semantic_error ("operator '%' is not allowed for float types", &pos);
        else if (tk == SLASH)
        {
          if (e2->kind == A_CONST_FLOAT_VALUE && e2->info.const_float_value_info.value == 0.0)
            semantic_error ("division by zero", &pos);
        }

        if (e1->kind == A_CONST_FLOAT_VALUE && e2->kind == A_CONST_FLOAT_VALUE)
        {
          vf1 = e1->info.const_float_value_info.value;
          vf2 = e2->info.const_float_value_info.value;

          if (tk == STAR)
          {
            rf = vf1 * vf2;
          }
          else
          {
            rf = vf1 / vf2;
          }

          check_float_overflow_underflow (&rf, t3, &pos);

          e3 = new_exp (A_CONST_FLOAT_VALUE);
          e3->base_type_or_null = t3;
          e3->form = A_VALUE;
          e3->access = ACCESS_CONSTANT;
          e3->info.const_float_value_info.value = rf;

          free_exp (e1);
          free_exp (e2);
        }
        else  // runtime
        {
          e3 = new_exp (AN_OPERATOR_VALUE);
          e3->base_type_or_null = t3;
          e3->form = A_VALUE;
          e3->access = ACCESS_READONLY;
          e3->info.operator_value_info.op = (tk == STAR) ? OP_MUL_FLOAT : OP_DIV_FLOAT;
          e3->info.operator_value_info.arg[0] = e1;
          e3->info.operator_value_info.arg[1] = e2;
        }
      }
      else
      {
        semantic_error ("illegal operands", &pos);
        free_exp (e2);
        e3 = e1;
      }
    }

    e1 = e3;
  }

  *out = e1;
}

/*****************************************************************************/

static void parse_sum (PCONTEXT      context,        // never null
                       TEXT_POSITION *prefix_position,
                       PENTITY       prefix_name,    // can be null
                       PEXPRESSION   *out)
{
  PEXPRESSION   e1, e2, e3;
  TOKEN_KIND    tk;
  PENTITY       t1, t2, t3;
  TEXT_POSITION pos;
  int8          v1, v2, r;
  double        vf1, vf2, rf;
  int           rc;

  parse_term (context, prefix_position, prefix_name, &e1);

  while (token.kind == PLUS || token.kind == MINUS)
  {
    tk = token.kind;
    memcpy (&pos, &token.pos, sizeof(pos));
    get_token();     // skip + or -

    parse_term (&no_context, &token.pos, null, &e2);

    // e3 = e1 op e2

    // check type

    t1 = e1->base_type_or_null;
    t2 = e2->base_type_or_null;

    if (t1 == null)   // error during earlier evaluation
    {
      free_exp (e2);
      e3 = e1;
    }
    else if (t2 == null)   // error during earlier evaluation
    {
      free_exp (e1);
      e3 = e2;
    }
    else
    {
      if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
      {
        t3 = promote_two_ints (e1, e2, true);
        if (t3 == null)
          semantic_error ("incompatible integer operands", &pos);

        if (e1->kind == A_CONST_INTEGER_VALUE && e2->kind == A_CONST_INTEGER_VALUE)
        {
          v1 = e1->info.const_integer_value_info.value;
          v2 = e2->info.const_integer_value_info.value;

          if (tk == PLUS)
          {
            rc = add_int (v1, v2, &r);
          }
          else
          {
            rc = sub_int (v1, v2, &r);
          }

          if (rc < 0)
          {
            semantic_error ("overflow", &pos);
            r = 0;
          }
          else if (t3 != null)
          {
            INTEGER_INFO *pi;

            // check if r outside range of t3

            pi = &integer_info[t3->info->the_integer_type.type];
            if (r < pi->min || r > pi->max)
            {
              semantic_error ("overflow", &pos);
              r = 0;
            }
          }

          e3 = new_exp (A_CONST_INTEGER_VALUE);
          e3->base_type_or_null = t3;
          e3->form = A_VALUE;
          e3->access = ACCESS_CONSTANT;
          e3->info.const_integer_value_info.value = r;

          free_exp (e1);
          free_exp (e2);
        }
        else  // runtime
        {
          e3 = new_exp (AN_OPERATOR_VALUE);
          e3->base_type_or_null = t3;
          e3->form = A_VALUE;
          e3->access = ACCESS_READONLY;
          e3->info.operator_value_info.op = (tk == PLUS) ? OP_ADD_INT : OP_SUB_INT;
          e3->info.operator_value_info.arg[0] = e1;
          e3->info.operator_value_info.arg[1] = e2;
        }
      }
      else if (t1->kind == A_FLOAT_TYPE && t2->kind == A_FLOAT_TYPE)
      {
        t3 = promote_two_floats (e1, e2, true);

        if (e1->kind == A_CONST_FLOAT_VALUE && e2->kind == A_CONST_FLOAT_VALUE)
        {
          vf1 = e1->info.const_float_value_info.value;
          vf2 = e2->info.const_float_value_info.value;

          if (tk == PLUS)
          {
            rf = vf1 + vf2;
          }
          else
          {
            rf = vf1 - vf2;
          }

          check_float_overflow_underflow (&rf, t3, &pos);

          e3 = new_exp (A_CONST_FLOAT_VALUE);
          e3->base_type_or_null = t3;
          e3->form = A_VALUE;
          e3->access = ACCESS_CONSTANT;
          e3->info.const_float_value_info.value = rf;

          free_exp (e1);
          free_exp (e2);
        }
        else  // runtime
        {
          e3 = new_exp (AN_OPERATOR_VALUE);
          e3->base_type_or_null = t3;
          e3->form = A_VALUE;
          e3->access = ACCESS_READONLY;
          e3->info.operator_value_info.op = (tk == PLUS) ? OP_ADD_FLOAT : OP_SUB_FLOAT;
          e3->info.operator_value_info.arg[0] = e1;
          e3->info.operator_value_info.arg[1] = e2;
        }
      }
      else if (t1->kind == AN_ENUMERATION_TYPE && t2->kind == AN_INTEGER_TYPE)
      {
        if (e1->kind == A_CONST_ENUMERATION_VALUE && e2->kind == A_CONST_INTEGER_VALUE)
        {
          v1 = e1->info.const_enumeration_value_info.value;
          v2 = e2->info.const_enumeration_value_info.value;

          if (tk == PLUS)
          {
            r = v1 + v2;
          }
          else
          {
            r = v1 - v2;
          }

          if (r < 0 || r > t1->info->the_enumeration_type.last)
          {
            semantic_error ("overflow", &pos);
            r = 0;
          }

          e3 = new_exp (A_CONST_ENUMERATION_VALUE);
          e3->base_type_or_null = t1;
          e3->form = A_VALUE;
          e3->access = ACCESS_CONSTANT;
          e3->info.const_enumeration_value_info.value = (uint4)r;

          free_exp (e1);
          free_exp (e2);
        }
        else  // runtime
        {
          e3 = new_exp (AN_OPERATOR_VALUE);
          e3->base_type_or_null = t1;
          e3->form = A_VALUE;
          e3->access = ACCESS_READONLY;
          e3->info.operator_value_info.op = (tk == PLUS) ? OP_ADD_INT : OP_SUB_INT;
          e3->info.operator_value_info.arg[0] = e1;
          e3->info.operator_value_info.arg[1] = e2;
        }
      }
      else if (t1->kind == AN_UNSAFE_POINTER_TYPE && t2->kind == AN_INTEGER_TYPE)
      {
        PENTITY element_type;

        element_type = t1->info->the_unsafe_pointer_type.designated_type;
        if (is_open_type (element_type))
          semantic_error ("left operand must not be an unsafe pointer to an open type", &pos);

        if (t2 == type_long)
          semantic_error ("right operand must not have type long", &pos);

        e3 = new_exp (AN_OPERATOR_VALUE);
        e3->base_type_or_null = t1;
        e3->form = A_VALUE;
        e3->access = ACCESS_READONLY;
        e3->info.operator_value_info.op = (tk == PLUS) ? OP_ADD_PTR_INT : OP_SUB_PTR_INT;
        e3->info.operator_value_info.arg[0] = e1;
        e3->info.operator_value_info.arg[1] = e2;
      }
      else if (tk == MINUS && t1->kind == AN_UNSAFE_POINTER_TYPE && t2->kind == AN_UNSAFE_POINTER_TYPE)
      {
        PENTITY element_type;

        if (!types_are_equal (t1, t2))
          semantic_error ("incompatible operands", &pos);
        else
        {
          element_type = t1->info->the_unsafe_pointer_type.designated_type;
          if (is_open_type (element_type))
            semantic_error ("operands must not be unsafe pointers to an open type", &pos);
        }

        e3 = new_exp (AN_OPERATOR_VALUE);
        e3->base_type_or_null = type_uint;
        e3->form = A_VALUE;
        e3->access = ACCESS_READONLY;
        e3->info.operator_value_info.op = OP_SUB_PTR_PTR;
        e3->info.operator_value_info.arg[0] = e1;
        e3->info.operator_value_info.arg[1] = e2;
      }
      else if (tk == PLUS &&
               (t1 == type_char || t1 == type_wchar || types_are_equal (t1, type_string) || types_are_equal (t1, type_wstring)) &&
               (t2 == type_char || t2 == type_wchar || types_are_equal (t2, type_string) || types_are_equal (t2, type_wstring)))
      {
        if ((e1->kind != A_POOL_CONSTANT && e1->kind != A_CONST_ENUMERATION_VALUE) ||
            (e2->kind != A_POOL_CONSTANT && e2->kind != A_CONST_ENUMERATION_VALUE))
        {
          semantic_error ("operands must be constant", &pos);
          free_exp (e2);
          e3 = e1;
        }
        else
        {
          uint len, target_charsize, i, offset;
          bool wide;
          POOL *p, *src;
          int8 v;

          // build pool constant with concatenated string constants

          len = 0;

          if (t1 == type_char || t1 == type_wchar)
            len++;
          else
            len += e1->constraint.value;

          if (t2 == type_char || t2 == type_wchar)
            len++;
          else
            len += e2->constraint.value;

          // if any argument is wide, then the result will have type wstring,
          // otherwise the result has type string.
          wide = (t1 == type_wchar || types_are_equal (t1, type_wstring) ||
                  t2 == type_wchar || types_are_equal (t2, type_wstring));

          e3 = new_exp (A_POOL_CONSTANT);
          e3->base_type_or_null = wide ? type_wstring : type_string;
          e3->constraint.kind = CONSTANT_CONSTRAINT;
          e3->constraint.value = len;
          e3->form = A_VALUE;
          e3->access = ACCESS_CONSTANT;

          target_charsize = wide ? 2 : 1;

          p = new_pool_constant (target_charsize * len, target_charsize);
          offset = 0;

          e3->info.pool_constant_info.pool_cte = p;

          if (t1 == type_char || t1 == type_wchar)
          {
            store_integer (p, offset, e1->info.const_enumeration_value_info.value, target_charsize);
            offset += target_charsize;
          }
          else
          {
            int source_charsize = types_are_equal (t1, type_string) ? 1 : 2;
            src = e1->info.pool_constant_info.pool_cte;
            for (i=0; i<e1->constraint.value; i++)
            {
              load_integer (src, i*source_charsize, &v, source_charsize, 0);
              store_integer (p, offset, v, target_charsize);
              offset += target_charsize;
            }
          }

          if (t2 == type_char || t2 == type_wchar)
          {
            store_integer (p, offset, e2->info.const_enumeration_value_info.value, target_charsize);
            offset += target_charsize;
          }
          else
          {
            int source_charsize = types_are_equal (t2, type_string) ? 1 : 2;
            src = e2->info.pool_constant_info.pool_cte;
            for (i=0; i<e2->constraint.value; i++)
            {
              load_integer (src, i*source_charsize, &v, source_charsize, 0);
              store_integer (p, offset, v, target_charsize);
              offset += target_charsize;
            }
          }

          free_exp (e1);
          free_exp (e2);
        }
      }
      else
      {
        semantic_error ("incompatible operands", &pos);
        free_exp (e2);
        e3 = e1;
      }
    }

    e1 = e3;
  }

  *out = e1;
}

/*****************************************************************************/

static void parse_simple_expression (PCONTEXT      context,        // never null
                                     TEXT_POSITION *prefix_position,
                                     PENTITY       prefix_name,    // can be null
                                     PEXPRESSION   *out)
{
  PEXPRESSION   e1, e2, e3;
  PENTITY       t1, t2, t3;
  KIND_OPERATOR op;
  TOKEN_KIND    tk;
  TEXT_POSITION pos;
  int8          v1, v2, r;

  parse_sum (context, prefix_position, prefix_name, &e1);

  tk = token.kind;

  if (tk != SHIFT_LEFT && tk != SHIFT_RIGHT)
  {
    *out = e1;
    return;
  }

  memcpy (&pos, &token.pos, sizeof(pos));
  get_token();     // skip << or >>
  parse_sum (&no_context, &token.pos, null, &e2);


  // check types

  t1 = e1->base_type_or_null;
  t2 = e2->base_type_or_null;

  if (t1 == null || t2 == null)   // error during earlier evaluation
  {
    free_exp (e2);
    *out = e1;
    return;
  }


  if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
  {
    t3 = promote_two_ints (e1, e2, true);
    if (t3 == null)
      semantic_error ("incompatible integer operands", &pos);

    if (e2->kind == A_CONST_INTEGER_VALUE && t3 != null)
    {
      v2 = e2->info.const_integer_value_info.value;

      if (t3 == type_long || t3 == type_int_literal)
      {
        if (v2 < 0 || v2 > 63)
        {
          semantic_error ("right operand must be in range 0..63", &pos);
          v2 = 0;
        }
      }
      else
      {
        if (v2 < 0 || v2 > 31)
        {
          semantic_error ("right operand must be in range 0..31", &pos);
          v2 = 0;
        }
      }
    }

    if (e1->kind == A_CONST_INTEGER_VALUE && e2->kind == A_CONST_INTEGER_VALUE)
    {
      v1 = e1->info.const_integer_value_info.value;
      v2 = e2->info.const_integer_value_info.value;

      if (tk == SHIFT_LEFT)
      {
        r = v1 << v2;
        if (r >> v2 != v1)
        {
          semantic_error ("overflow", &pos);
          r = 0;
        }
      }
      else
      {
        r = v1 >> v2;
      }

      if (t3 != null)
      {
        INTEGER_INFO *pi;

        // check if r outside range of t3

        pi = &integer_info[t3->info->the_integer_type.type];

        if (r < pi->min || r > pi->max)
        {
          semantic_error ("overflow", &pos);
          r = 0;
        }
      }

      e3 = new_exp (A_CONST_INTEGER_VALUE);
      e3->base_type_or_null = t3;
      e3->form = A_VALUE;
      e3->access = ACCESS_CONSTANT;
      e3->info.const_integer_value_info.value = r;

      free_exp (e1);
      free_exp (e2);
    }
    else  // runtime
    {
      if (t3 != null && integer_info[t3->info->the_integer_type.type].is_signed)
        op = (tk == SHIFT_LEFT) ? OP_SHIFT_LEFT_SIGNED : OP_SHIFT_RIGHT_SIGNED;
      else
        op = (tk == SHIFT_LEFT) ? OP_SHIFT_LEFT_UNSIGNED : OP_SHIFT_RIGHT_UNSIGNED;

      e3 = new_exp (AN_OPERATOR_VALUE);
      e3->base_type_or_null = t3;
      e3->form = A_VALUE;
      e3->access = ACCESS_READONLY;
      e3->info.operator_value_info.op = op;
      e3->info.operator_value_info.arg[0] = e1;
      e3->info.operator_value_info.arg[1] = e2;
    }
  }
  else
  {
    semantic_error ("operands must have type integer", &pos);
    free_exp (e2);
    e3 = e1;
  }

  *out = e3;
}

/*****************************************************************************/

static bool compare (int8 x, COMPARISON_FLAG cmp, int8 y)
{
  bool b;

  switch (cmp)
  {
    case CMP_SMALLER:
      b = x < y;
      break;

    case CMP_EQUAL:
      b = x == y;
      break;

    case CMP_SMALLER_OR_EQUAL:
      b = x <= y;
      break;

    case CMP_LARGER:
      b = x > y;
      break;

    case CMP_NOT_EQUAL:
      b = x != y;
      break;

    case CMP_LARGER_OR_EQUAL:
      b = x >= y;
      break;

    default:
      fatal_compiler_error ("compare()", &token.pos);
      break;
  }

  return b;
}

/*****************************************************************************/

static bool fcompare (double x, COMPARISON_FLAG cmp, double y)
{
  bool b;

  switch (cmp)
  {
    case CMP_SMALLER:
      b = x < y;
      break;

    case CMP_EQUAL:
      b = x == y;
      break;

    case CMP_SMALLER_OR_EQUAL:
      b = x <= y;
      break;

    case CMP_LARGER:
      b = x > y;
      break;

    case CMP_NOT_EQUAL:
      b = x != y;
      break;

    case CMP_LARGER_OR_EQUAL:
      b = x >= y;
      break;

    default:
      fatal_compiler_error ("fcompare()", &token.pos);
      break;
  }

  return b;
}

/*****************************************************************************/

static void parse_relation (PCONTEXT      context,        // never null
                            TEXT_POSITION *prefix_position,
                            PENTITY       prefix_name,    // can be null
                            PEXPRESSION   *pout)
{
  TEXT_POSITION   pos1, pos;
  PEXPRESSION     e1, e2, e3;
  KIND_OPERATOR   op;
  PENTITY         t1, t2, t3;
  bool            b;
  COMPARISON_FLAG cmp;

  memcpy (&pos1, &token.pos, sizeof(TEXT_POSITION));
  parse_simple_expression (context, prefix_position, prefix_name, &e1);

  memcpy (&pos, &token.pos, sizeof(TEXT_POSITION));
  switch (token.kind)
  {
    case SMALLER:
      cmp = CMP_SMALLER;
      break;

    case EQUAL:
      cmp = CMP_EQUAL;
      break;

    case SMALLER_OR_EQUAL:
      cmp = CMP_SMALLER_OR_EQUAL;
      break;

    case LARGER:
      cmp = CMP_LARGER;
      break;

    case NOT_EQUAL:
      cmp = CMP_NOT_EQUAL;
      break;

    case LARGER_OR_EQUAL:
      cmp = CMP_LARGER_OR_EQUAL;
      break;

    default:
      cmp = CMP_FILLER;
      break;
  }

  if (cmp == CMP_FILLER)
  {
    *pout = e1;
    return;
  }


  get_token();
  parse_simple_expression (&no_context, &token.pos, null, &e2);


  // check types

  t1 = e1->base_type_or_null;
  t2 = e2->base_type_or_null;

  if (t1 == null || t2 == null)   // error during earlier evaluation
  {
    e3 = new_exp (A_CONST_ENUMERATION_VALUE);
    e3->base_type_or_null = type_bool;
    e3->form = A_VALUE;
    e3->access = ACCESS_CONSTANT;
    e3->info.const_enumeration_value_info.value = 0;

    free_exp (e1);
    free_exp (e2);

    *pout = e3;
    return;
  }


  if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
  {
    t3 = promote_two_ints (e1, e2, true);
    if (t3 == null)
      semantic_error ("incompatible integer operands", &pos);

    if (e1->kind == A_CONST_INTEGER_VALUE && e2->kind == A_CONST_INTEGER_VALUE)
    {
      b = compare (e1->info.const_integer_value_info.value,
                   cmp,
                   e2->info.const_integer_value_info.value);

      e3 = new_exp (A_CONST_ENUMERATION_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_CONSTANT;
      e3->info.const_enumeration_value_info.value = (int)b;

      free_exp (e1);
      free_exp (e2);
    }
    else  // runtime
    {
      if (e1->kind == A_CONST_INTEGER_VALUE)
      {
        if (e1->info.const_integer_value_info.value == integer_info[t2->info->the_integer_type.type].min)
        {
          if (cmp == CMP_LARGER)   // t'min > e -> false
            warning ("always false", &pos);
          else if (cmp == CMP_SMALLER_OR_EQUAL)  // t'min <= e -> true
            warning ("always true", &pos);
        }

        if (e1->info.const_integer_value_info.value == integer_info[t2->info->the_integer_type.type].max)
        {
          if (cmp == CMP_SMALLER)   // t'max < e -> false
            warning ("always false", &pos);
          else if (cmp == CMP_LARGER_OR_EQUAL)  // t'max >= e -> true
            warning ("always true", &pos);
        }
      }
      else if (e2->kind == A_CONST_INTEGER_VALUE)
      {
        if (e2->info.const_integer_value_info.value == integer_info[t1->info->the_integer_type.type].min)
        {
          if (cmp == CMP_SMALLER)   // e < t'min -> false
            warning ("always false", &pos);
          else if (cmp == CMP_LARGER_OR_EQUAL)  // e >= t'min -> true
            warning ("always true", &pos);
        }

        if (e2->info.const_integer_value_info.value == integer_info[t1->info->the_integer_type.type].max)
        {
          if (cmp == CMP_LARGER)   // e > t'max -> false
            warning ("always false", &pos);
          else if (cmp == CMP_SMALLER_OR_EQUAL)  // e <= t'max -> true
            warning ("always true", &pos);
        }
      }

      if (t3 != null && integer_info[t3->info->the_integer_type.type].is_signed)
        op = OP_COMPARE_SIGNED;
      else
        op = OP_COMPARE_UNSIGNED;

      e3 = new_exp (AN_OPERATOR_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_READONLY;
      e3->info.operator_value_info.op = op;
      e3->info.operator_value_info.cmp = cmp;
      e3->info.operator_value_info.arg[0] = e1;
      e3->info.operator_value_info.arg[1] = e2;
      e3->info.operator_value_info.type = t3;
    }
  }
  else if (t1->kind == A_FLOAT_TYPE && t2->kind == A_FLOAT_TYPE)
  {
    t3 = promote_two_floats (e1, e2, true);

    if (e1->kind == A_CONST_FLOAT_VALUE && e2->kind == A_CONST_FLOAT_VALUE)
    {
      b = fcompare (e1->info.const_float_value_info.value,
                    cmp,
                    e2->info.const_float_value_info.value);

      e3 = new_exp (A_CONST_ENUMERATION_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_CONSTANT;
      e3->info.const_enumeration_value_info.value = (int)b;

      free_exp (e1);
      free_exp (e2);
    }
    else  // runtime
    {
      e3 = new_exp (AN_OPERATOR_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_READONLY;
      e3->info.operator_value_info.op = OP_COMPARE_FLOAT;
      e3->info.operator_value_info.cmp = cmp;
      e3->info.operator_value_info.arg[0] = e1;
      e3->info.operator_value_info.arg[1] = e2;
      e3->info.operator_value_info.type = t3;
    }
  }
  else if (t1->kind == AN_ENUMERATION_TYPE && t2->kind == AN_ENUMERATION_TYPE)
  {
    if (!types_are_equal (t1, t2))
      semantic_error ("incompatible enum operands", &pos);

    if (e1->kind == A_CONST_ENUMERATION_VALUE && e2->kind == A_CONST_ENUMERATION_VALUE)
    {
      b = compare (e1->info.const_enumeration_value_info.value,
                   cmp,
                   e2->info.const_enumeration_value_info.value);

      e3 = new_exp (A_CONST_ENUMERATION_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_CONSTANT;
      e3->info.const_enumeration_value_info.value = (int)b;

      free_exp (e1);
      free_exp (e2);
    }
    else  // runtime
    {
      if (e1->kind == A_CONST_ENUMERATION_VALUE)
      {
        if (e1->info.const_enumeration_value_info.value == 0)
        {
          if (cmp == CMP_LARGER)   // 0 > e -> false
            warning ("always false", &pos);
          else if (cmp == CMP_SMALLER_OR_EQUAL)  // 0 <= e -> true
            warning ("always true", &pos);
        }
      }
      else if (e2->kind == A_CONST_ENUMERATION_VALUE)
      {
        if (e2->info.const_enumeration_value_info.value == 0)
        {
          if (cmp == CMP_SMALLER)   // e < 0 -> false
            warning ("always false", &pos);
          else if (cmp == CMP_LARGER_OR_EQUAL)  // e >= 0 -> true
            warning ("always true", &pos);
        }
      }

      e3 = new_exp (AN_OPERATOR_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_READONLY;
      e3->info.operator_value_info.op = OP_COMPARE_UNSIGNED;
      e3->info.operator_value_info.cmp = cmp;
      e3->info.operator_value_info.arg[0] = e1;
      e3->info.operator_value_info.arg[1] = e2;
      e3->info.operator_value_info.type = t1;
    }
  }
  else if (t1->kind == A_NULL_POINTER_TYPE && t2->kind == A_NULL_POINTER_TYPE)
  {
    if (cmp != CMP_EQUAL && cmp != CMP_NOT_EQUAL)
      semantic_error ("order comparison of null pointer is not allowed", &pos);

    b = compare (0, cmp, 0);

    e3 = new_exp (A_CONST_ENUMERATION_VALUE);
    e3->base_type_or_null = type_bool;
    e3->form = A_VALUE;
    e3->access = ACCESS_CONSTANT;
    e3->info.const_enumeration_value_info.value = (int)b;

    free_exp (e1);
    free_exp (e2);
  }
  else if ((t1->kind == A_POINTER_TYPE || t1->kind == A_NULL_POINTER_TYPE) &&
           (t2->kind == A_POINTER_TYPE || t2->kind == A_NULL_POINTER_TYPE))
  {
    if (t1->kind == A_POINTER_TYPE && t2->kind == A_POINTER_TYPE && (!types_are_equal (t1, t2)))
      semantic_error ("incompatible pointers", &pos);

    if (cmp != CMP_EQUAL && cmp != CMP_NOT_EQUAL)
      semantic_error ("order comparison of pointers is not allowed", &pos);

    if (e1->kind == A_CONST_NULL_VALUE && e2->kind == A_CONST_NULL_VALUE)
    {
      b = compare (0, cmp, 0);

      e3 = new_exp (A_CONST_ENUMERATION_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_CONSTANT;
      e3->info.const_enumeration_value_info.value = (int)b;

      free_exp (e1);
      free_exp (e2);
    }
    else  // runtime
    {
      e3 = new_exp (AN_OPERATOR_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_READONLY;
      e3->info.operator_value_info.op = OP_COMPARE_PTR;
      e3->info.operator_value_info.cmp = cmp;
      e3->info.operator_value_info.arg[0] = e1;
      e3->info.operator_value_info.arg[1] = e2;
    }
  }
  else if ((t1->kind == A_FUNCTION_POINTER_TYPE || t1->kind == A_NULL_POINTER_TYPE) &&
           (t2->kind == A_FUNCTION_POINTER_TYPE || t2->kind == A_NULL_POINTER_TYPE))
  {
    if (t1->kind == A_FUNCTION_POINTER_TYPE && t2->kind == A_FUNCTION_POINTER_TYPE && (!types_are_equal (t1, t2)))
      semantic_error ("incompatible function pointers", &pos);

    if (cmp != CMP_EQUAL && cmp != CMP_NOT_EQUAL)
      semantic_error ("order comparison of function pointers is not allowed", &pos);

    if (e1->kind == A_CONST_NULL_VALUE && e2->kind == A_CONST_NULL_VALUE)
    {
      b = compare (0, cmp, 0);

      e3 = new_exp (A_CONST_ENUMERATION_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_CONSTANT;
      e3->info.const_enumeration_value_info.value = (int)b;

      free_exp (e1);
      free_exp (e2);
    }
    else  // runtime
    {
      e3 = new_exp (AN_OPERATOR_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_READONLY;
      e3->info.operator_value_info.op = OP_COMPARE_PTR;
      e3->info.operator_value_info.cmp = cmp;
      e3->info.operator_value_info.arg[0] = e1;
      e3->info.operator_value_info.arg[1] = e2;
    }
  }
  else if ((t1->kind == AN_UNSAFE_POINTER_TYPE || t1->kind == A_NULL_POINTER_TYPE) &&
           (t2->kind == AN_UNSAFE_POINTER_TYPE || t2->kind == A_NULL_POINTER_TYPE))
  {
    if (t1->kind == AN_UNSAFE_POINTER_TYPE && t2->kind == AN_UNSAFE_POINTER_TYPE && (!types_are_equal (t1, t2)))
      semantic_error ("incompatible unsafe pointers", &pos);

    if (e1->kind == A_CONST_NULL_VALUE && e2->kind == A_CONST_NULL_VALUE)
    {
      b = compare (0, cmp, 0);

      e3 = new_exp (A_CONST_ENUMERATION_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_CONSTANT;
      e3->info.const_enumeration_value_info.value = (int)b;

      free_exp (e1);
      free_exp (e2);
    }
    else  // runtime
    {
      e3 = new_exp (AN_OPERATOR_VALUE);
      e3->base_type_or_null = type_bool;
      e3->form = A_VALUE;
      e3->access = ACCESS_READONLY;
      e3->info.operator_value_info.op = OP_COMPARE_PTR;
      e3->info.operator_value_info.cmp = cmp;
      e3->info.operator_value_info.arg[0] = e1;
      e3->info.operator_value_info.arg[1] = e2;
    }
  }
  else
  {
    semantic_error ("illegal operands for comparison", &pos);

    e3 = new_exp (A_CONST_ENUMERATION_VALUE);
    e3->base_type_or_null = type_bool;
    e3->form = A_VALUE;
    e3->access = ACCESS_CONSTANT;
    e3->info.const_enumeration_value_info.value = 0;

    free_exp (e1);
    free_exp (e2);
  }

  *pout = e3;
}

/*****************************************************************************/

static void parse_sub_expression (PCONTEXT      context,        // never null
                                  TEXT_POSITION *prefix_position,
                                  PENTITY       prefix_name,    // can be null
                                  PEXPRESSION   *pout)
{
  TEXT_POSITION pos1, pos2, pos;
  PEXPRESSION   e1, e2, e3;
  PENTITY       t1, t2, t3;
  int8          v1, v2, v3;
  TOKEN_KIND    tk;
  KIND_OPERATOR op;

  memcpy (&pos1, &token.pos, sizeof(TEXT_POSITION));

  parse_relation (context, prefix_position, prefix_name, &e1);

  switch (token.kind)
  {
    case DOUBLE_AMPERSAND:     // &&
    case DOUBLE_VERTICAL_BAR:  // ||

      tk = token.kind;

      while (token.kind == tk)
      {
        get_token();
        memcpy (&pos2, &token.pos, sizeof(TEXT_POSITION));
        parse_relation (&no_context, &token.pos, null, &e2);

        // e3 = e1 op e2;

        type_check_bool_operator (e1, &pos1);
        type_check_bool_operator (e2, &pos2);

        // compute result

        if (e1->kind == A_CONST_ENUMERATION_VALUE && e2->kind == A_CONST_ENUMERATION_VALUE)
        {
          v1 = e1->info.const_enumeration_value_info.value;
          v2 = e2->info.const_enumeration_value_info.value;

          e3 = new_exp (A_CONST_ENUMERATION_VALUE);
          e3->base_type_or_null = type_bool;
          e3->form = A_VALUE;
          e3->access = ACCESS_CONSTANT;
          e3->info.const_enumeration_value_info.value =
             (uint4)((tk==DOUBLE_AMPERSAND) ? (v1 && v2) : (v1 || v2));

          free_exp (e1);
          free_exp (e2);
        }
        else   // runtime
        {
          e3 = new_exp (AN_OPERATOR_VALUE);
          e3->base_type_or_null = type_bool;
          e3->form = A_VALUE;
          e3->access = ACCESS_READONLY;
          e3->info.operator_value_info.op =
              (tk==DOUBLE_AMPERSAND) ? OP_SHORT_CIRCUIT_AND : OP_SHORT_CIRCUIT_OR;
          e3->info.operator_value_info.arg[0] = e1;
          e3->info.operator_value_info.arg[1] = e2;
        }

        e1 = e3;
      }
      break;


    case AMPERSAND:     // &
    case VERTICAL_BAR:  // |
    case CARET:         // ^

      tk = token.kind;

      while (token.kind == tk)
      {
        memcpy (&pos, &token.pos, sizeof(TEXT_POSITION));
        get_token();
        memcpy (&pos2, &token.pos, sizeof(TEXT_POSITION));
        parse_relation (&no_context, &token.pos, null, &e2);

        // e3 = e1 op e2;

        t1 = e1->base_type_or_null;
        t2 = e2->base_type_or_null;

        if (t1 == null)   // error during earlier evaluation
        {
          free_exp (e2);
          e3 = e1;
        }
        else if (t2 == null)   // error during earlier evaluation
        {
          free_exp (e1);
          e3 = e2;
        }
        else
        {
          if (t1->kind == AN_INTEGER_TYPE && t2->kind == AN_INTEGER_TYPE)
          {
            t3 = promote_two_ints (e1, e2, true);
            if (t3 == null)
              semantic_error ("incompatible integer operands", &pos);

            if (e1->kind == A_CONST_INTEGER_VALUE && e2->kind == A_CONST_INTEGER_VALUE)
            {
              v1 = e1->info.const_integer_value_info.value;
              v2 = e2->info.const_integer_value_info.value;

              if (tk == AMPERSAND)
                v3 = v1 & v2;
              else if (tk == VERTICAL_BAR)
                v3 = v1 | v2;
              else
                v3 = v1 ^ v2;

              e3 = new_exp (A_CONST_INTEGER_VALUE);
              e3->base_type_or_null = t3;
              e3->form = A_VALUE;
              e3->access = ACCESS_CONSTANT;
              e3->info.const_integer_value_info.value = v3;

              free_exp (e1);
              free_exp (e2);
            }
            else  // runtime
            {
              e3 = new_exp (AN_OPERATOR_VALUE);
              e3->base_type_or_null = t3;
              e3->form = A_VALUE;
              e3->access = ACCESS_READONLY;

              if (tk == AMPERSAND)
                op = OP_BITAND;
              else if (tk == VERTICAL_BAR)
                op = OP_BITOR;
              else
                op = OP_BITXOR;

              e3->info.operator_value_info.op = op;

              e3->info.operator_value_info.arg[0] = e1;
              e3->info.operator_value_info.arg[1] = e2;
            }
          }
          else if (t1 == type_bool && t2 == type_bool)
          {
            if (e1->kind == A_CONST_ENUMERATION_VALUE && e2->kind == A_CONST_ENUMERATION_VALUE)
            {
              v1 = e1->info.const_enumeration_value_info.value;
              v2 = e2->info.const_enumeration_value_info.value;

              if (tk == AMPERSAND)
                v3 = v1 && v2;
              else if (tk == VERTICAL_BAR)
                v3 = v1 || v2;
              else
                v3 = (v1 ^ v2) != 0;

              e3 = new_exp (A_CONST_ENUMERATION_VALUE);
              e3->base_type_or_null = type_bool;
              e3->form = A_VALUE;
              e3->access = ACCESS_CONSTANT;
              e3->info.const_enumeration_value_info.value = (uint4)v3;

              free_exp (e1);
              free_exp (e2);
            }
            else  // runtime
            {
              e3 = new_exp (AN_OPERATOR_VALUE);
              e3->base_type_or_null = type_bool;
              e3->form = A_VALUE;
              e3->access = ACCESS_READONLY;

              if (tk == AMPERSAND)
                op = OP_AND;
              else if (tk == VERTICAL_BAR)
                op = OP_OR;
              else
                op = OP_XOR;

              e3->info.operator_value_info.op = op;
              e3->info.operator_value_info.arg[0] = e1;
              e3->info.operator_value_info.arg[1] = e2;
            }
          }
          else
          {
            semantic_error ("operator has bad operands", &pos);
            e3 = e1;
            free_exp (e2);
          }
        }

        e1 = e3;
      }
      break;

    default:
      break;
  }

  *pout = e1;
}

/*****************************************************************************/

void parse_expression (PCONTEXT      context,           // never null
                       TEXT_POSITION *prefix_position,  // never null
                       PENTITY       prefix_name,       // can be null
                       PEXPRESSION   *pout)
{
  TEXT_POSITION pos, posc;
  PEXPRESSION   e1, e2, e3, e4;
  PENTITY       t1, t2, t3, t4;
  CONTEXT       inner_context;

  parse_sub_expression (context, prefix_position, prefix_name, &e1);

  if (token.kind != QUESTION_MARK)
  {
    *pout = e1;
    return;
  }

  memcpy (&pos, &token.pos, sizeof(TEXT_POSITION));
  get_token();     // skip '?'

  memcpy (&inner_context, context, sizeof(inner_context));
  if (inner_context.constraint.kind == RUNTIME_CONSTRAINT)
    inner_context.constraint.kind = UNCONSTRAINED;

  parse_sub_expression (&inner_context, &token.pos, null, &e2);

  if (token.kind != COLON)
  {
    syntax_error ("':' expected", &token.pos);
    free_exp (e1);
    *pout = e2;
    return;
  }

  memcpy (&posc, &token.pos, sizeof(TEXT_POSITION));
  get_token();     // skip ':'

  parse_sub_expression (&inner_context, &token.pos, null, &e3);


  // type check

  t1 = e1->base_type_or_null;
  t2 = e2->base_type_or_null;
  t3 = e3->base_type_or_null;

  if (t1 != null && t1 != type_bool)
    semantic_error ("prefix must have type bool", &pos);

  if (t2 == null)   // error during earlier evaluation
  {
    free_exp (e1);
    free_exp (e3);
    e4 = e2;
  }
  else if (t3 == null)   // error during earlier evaluation
  {
    free_exp (e1);
    free_exp (e2);
    e4 = e3;
  }
  else
  {
    if (t2->kind == AN_INTEGER_TYPE && t3->kind == AN_INTEGER_TYPE)
    {
      t4 = promote_two_ints (e2, e3, true);
      if (t4 == null)
        semantic_error ("incompatible integer types", &pos);

      if (e1->kind == A_CONST_ENUMERATION_VALUE &&
          e2->kind == A_CONST_INTEGER_VALUE &&
          e3->kind == A_CONST_INTEGER_VALUE)
      {
        e4 = new_exp (A_CONST_INTEGER_VALUE);
        e4->base_type_or_null = t4;
        e4->form = A_VALUE;
        e4->access = ACCESS_CONSTANT;
        e4->info.const_integer_value_info.value =
             (e1->info.const_enumeration_value_info.value == 1)
             ? e2->info.const_integer_value_info.value
             : e3->info.const_integer_value_info.value;

        free_exp (e1);
        free_exp (e2);
        free_exp (e3);
      }
      else
      {
        e4 = new_exp (AN_OPERATOR_VALUE);
        e4->base_type_or_null = promote_two_ints (e2, e3, false);
        e4->form = A_VALUE;
        e4->access = ACCESS_READONLY;
        e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
        e4->info.operator_value_info.arg[0] = e1;
        e4->info.operator_value_info.arg[1] = e2;
        e4->info.operator_value_info.arg[2] = e3;
      }
    }
    else if (t2->kind == A_FLOAT_TYPE && t3->kind == A_FLOAT_TYPE)
    {
      t4 = promote_two_floats (e2, e3, true);

      if (e1->kind == A_CONST_ENUMERATION_VALUE &&
          e2->kind == A_CONST_FLOAT_VALUE &&
          e3->kind == A_CONST_FLOAT_VALUE)
      {
        e4 = new_exp (A_CONST_FLOAT_VALUE);
        e4->base_type_or_null = t4;
        e4->form = A_VALUE;
        e4->access = ACCESS_CONSTANT;
        e4->info.const_float_value_info.value =
             (e1->info.const_enumeration_value_info.value == 1)
             ? e2->info.const_float_value_info.value
             : e3->info.const_float_value_info.value;

        free_exp (e1);
        free_exp (e2);
        free_exp (e3);
      }
      else
      {
        e4 = new_exp (AN_OPERATOR_VALUE);
        e4->base_type_or_null = promote_two_floats (e2, e3, false);
        e4->form = A_VALUE;
        e4->access = ACCESS_READONLY;
        e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
        e4->info.operator_value_info.arg[0] = e1;
        e4->info.operator_value_info.arg[1] = e2;
        e4->info.operator_value_info.arg[2] = e3;
      }
    }
    else if (t2->kind == AN_ENUMERATION_TYPE && t3->kind == AN_ENUMERATION_TYPE)
    {
      if (!types_are_equal (t2, t3))
        semantic_error ("incompatible enum types", &posc);

      t4 = t2;

      if (e1->kind == A_CONST_ENUMERATION_VALUE &&
          e2->kind == A_CONST_ENUMERATION_VALUE &&
          e3->kind == A_CONST_ENUMERATION_VALUE)
      {
        e4 = new_exp (A_CONST_ENUMERATION_VALUE);
        e4->base_type_or_null = t4;
        e4->form = A_VALUE;
        e4->access = ACCESS_CONSTANT;
        e4->info.const_enumeration_value_info.value =
             (e1->info.const_enumeration_value_info.value == 1)
             ? e2->info.const_enumeration_value_info.value
             : e3->info.const_enumeration_value_info.value;

        free_exp (e1);
        free_exp (e2);
        free_exp (e3);
      }
      else
      {
        e4 = new_exp (AN_OPERATOR_VALUE);
        e4->base_type_or_null = t4;
        e4->form = A_VALUE;
        e4->access = ACCESS_READONLY;
        e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
        e4->info.operator_value_info.arg[0] = e1;
        e4->info.operator_value_info.arg[1] = e2;
        e4->info.operator_value_info.arg[2] = e3;
      }
    }
    else if (t2->kind == A_NULL_POINTER_TYPE && t3->kind == A_NULL_POINTER_TYPE)
    {
      e4 = new_exp (A_CONST_NULL_VALUE);
      e4->base_type_or_null = type_null_literal;
      e4->form = A_VALUE;
      e4->access = ACCESS_CONSTANT;

      free_exp (e1);
      free_exp (e2);
      free_exp (e3);
    }
    else if ((t2->kind == A_POINTER_TYPE || t2->kind == A_NULL_POINTER_TYPE) &&
             (t3->kind == A_POINTER_TYPE || t3->kind == A_NULL_POINTER_TYPE))
    {
      t4 = (t2->kind == A_POINTER_TYPE) ? t2 : t3;

      if (t2->kind == A_POINTER_TYPE && t3->kind == A_POINTER_TYPE && (!types_are_equal (t2, t3)))
        semantic_error ("incompatible pointer types", &posc);

      e4 = new_exp (AN_OPERATOR_VALUE);
      e4->base_type_or_null = t4;
      e4->form = A_VALUE;
      e4->access = ACCESS_READONLY;
      e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
      e4->info.operator_value_info.arg[0] = e1;
      e4->info.operator_value_info.arg[1] = e2;
      e4->info.operator_value_info.arg[2] = e3;
    }
    else if ((t2->kind == A_FUNCTION_POINTER_TYPE || t2->kind == A_NULL_POINTER_TYPE) &&
             (t3->kind == A_FUNCTION_POINTER_TYPE || t3->kind == A_NULL_POINTER_TYPE))
    {
      t4 = (t2->kind == A_FUNCTION_POINTER_TYPE) ? t2 : t3;

      if (t2->kind == A_FUNCTION_POINTER_TYPE && t3->kind == A_FUNCTION_POINTER_TYPE && (!types_are_equal (t2, t3)))
        semantic_error ("incompatible function pointer types", &posc);

      e4 = new_exp (AN_OPERATOR_VALUE);
      e4->base_type_or_null = t4;
      e4->form = A_VALUE;
      e4->access = ACCESS_READONLY;
      e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
      e4->info.operator_value_info.arg[0] = e1;
      e4->info.operator_value_info.arg[1] = e2;
      e4->info.operator_value_info.arg[2] = e3;
    }
    else if ((t2->kind == AN_UNSAFE_POINTER_TYPE || t2->kind == A_NULL_POINTER_TYPE) &&
             (t3->kind == AN_UNSAFE_POINTER_TYPE || t3->kind == A_NULL_POINTER_TYPE))
    {
      t4 = (t2->kind == AN_UNSAFE_POINTER_TYPE) ? t2 : t3;

      if (t2->kind == AN_UNSAFE_POINTER_TYPE && t3->kind == AN_UNSAFE_POINTER_TYPE && (!types_are_equal (t2, t3)))
        semantic_error ("incompatible unsafe pointer types", &posc);

      e4 = new_exp (AN_OPERATOR_VALUE);
      e4->base_type_or_null = t4;
      e4->form = A_VALUE;
      e4->access = ACCESS_READONLY;
      e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
      e4->info.operator_value_info.arg[0] = e1;
      e4->info.operator_value_info.arg[1] = e2;
      e4->info.operator_value_info.arg[2] = e3;
    }
    else if (t2->kind == AN_OPEN_ARRAY_TYPE && t3->kind == AN_OPEN_ARRAY_TYPE)
    {
      if (!types_are_equal (t2, t3))
        semantic_error ("incompatible operands", &posc);

      e4 = new_exp (AN_OPERATOR_VALUE);
      e4->base_type_or_null = t2;

      // check array constraints
      // (e2 const or runtime, e3 const or runtime, context unc., const or runtime)

      if (context->constraint.kind == UNCONSTRAINED)
      {
        // in an unconstrained context, the lengths can (but need not) be different.

        if (e2->constraint.kind == CONSTANT_CONSTRAINT &&
            e3->constraint.kind == CONSTANT_CONSTRAINT &&
            e2->constraint.value == e3->constraint.value)
        {
          e4->constraint.kind = CONSTANT_CONSTRAINT;
          e4->constraint.value = e2->constraint.value;
        }
        else   // different or runtime
        {
          e4->constraint.kind = RUNTIME_CONSTRAINT;
          e4->constraint.value = 0;
        }
      }
      else  // the context is constrained  (or does not apply)
      {
        // both lengths must be equal

        if (e2->constraint.kind == CONSTANT_CONSTRAINT &&
            e3->constraint.kind == CONSTANT_CONSTRAINT &&
            e2->constraint.value != e3->constraint.value)
        {
          semantic_error ("array lengths do not match", &posc);
        }

        if (e2->constraint.kind == CONSTANT_CONSTRAINT ||
            e3->constraint.kind == CONSTANT_CONSTRAINT)
        {
          e4->constraint.kind = CONSTANT_CONSTRAINT;

          if (e2->constraint.kind == CONSTANT_CONSTRAINT)
            e4->constraint.value = e2->constraint.value;
          else
            e4->constraint.value = e3->constraint.value;
        }
        else   // both runtime
        {
          e4->constraint.kind = RUNTIME_CONSTRAINT;
          e4->constraint.value = 0;
        }
      }

      e4->form = A_VALUE;
      e4->access = ACCESS_READONLY;
      e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
      e4->info.operator_value_info.arg[0] = e1;
      e4->info.operator_value_info.arg[1] = e2;
      e4->info.operator_value_info.arg[2] = e3;
    }
    else if (t2->kind == A_STRUCT_TYPE && t2->info->the_struct_type.is_open_type)
    {
      if (!types_are_equal (t2, t3))
        semantic_error ("incompatible operands", &posc);

      e4 = new_exp (AN_OPERATOR_VALUE);
      e4->base_type_or_null = t2;

      // check struct constraints
      // (e2 const or runtime, e3 const or runtime, context unc., const or runtime)

      if (context->constraint.kind == UNCONSTRAINED)
      {
        // in an unconstrained context, the discriminants can (but need not) be different.

        if (e2->constraint.kind == CONSTANT_CONSTRAINT &&
            e3->constraint.kind == CONSTANT_CONSTRAINT &&
            e2->constraint.value == e3->constraint.value)
        {
          e4->constraint.kind = CONSTANT_CONSTRAINT;
          e4->constraint.value = e2->constraint.value;
        }
        else   // different or runtime
        {
          e4->constraint.kind = RUNTIME_CONSTRAINT;
          e4->constraint.value = 0;
        }
      }
      else  // the context is constrained.
      {
        // both discriminants must be equal

        if (e2->constraint.kind == CONSTANT_CONSTRAINT &&
            e3->constraint.kind == CONSTANT_CONSTRAINT &&
            e2->constraint.value != e3->constraint.value)
        {
          semantic_error ("struct discriminants do not match", &posc);
        }

        if (e2->constraint.kind == CONSTANT_CONSTRAINT ||
            e3->constraint.kind == CONSTANT_CONSTRAINT)
        {
          e4->constraint.kind = CONSTANT_CONSTRAINT;

          if (e2->constraint.kind == CONSTANT_CONSTRAINT)
            e4->constraint.value = e2->constraint.value;
          else
            e4->constraint.value = e3->constraint.value;
        }
        else   // both runtime
        {
          e4->constraint.kind = RUNTIME_CONSTRAINT;
          e4->constraint.value = 0;
        }
      }

      e4->form = A_VALUE;
      e4->access = ACCESS_READONLY;
      e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
      e4->info.operator_value_info.arg[0] = e1;
      e4->info.operator_value_info.arg[1] = e2;
      e4->info.operator_value_info.arg[2] = e3;
    }
    else if ((t2->kind == A_STRUCT_TYPE && !t2->info->the_struct_type.is_open_type) ||
             t2->kind == A_UNION_TYPE ||
             t2->kind == AN_OPAQUE_TYPE ||
             t2->kind == A_GENERIC_TYPE)
    {
      if (!types_are_equal (t2, t3))
        semantic_error ("incompatible operands", &posc);

      e4 = new_exp (AN_OPERATOR_VALUE);
      e4->base_type_or_null = t2;
      e4->form = A_VALUE;
      e4->access = ACCESS_READONLY;
      e4->info.operator_value_info.op = OP_CONDITIONAL_TEST;
      e4->info.operator_value_info.arg[0] = e1;
      e4->info.operator_value_info.arg[1] = e2;
      e4->info.operator_value_info.arg[2] = e3;
    }
    else
    {
      semantic_error ("illegal operands", &posc);
      free_exp (e1);
      free_exp (e3);
      e4 = e2;
    }
  }

  *pout = e4;
}

/*****************************************************************************/

int8 parse_constant_int_or_uint_expression (void)
{
  TEXT_POSITION pos;
  PEXPRESSION   e;
  int8          value;

  memcpy (&pos, &token.pos, sizeof(pos));

  parse_expression (&no_context, &token.pos, null, &e);

  if (e->kind != A_CONST_INTEGER_VALUE)
  {
    semantic_error ("must be integer constant expression", &pos);
    value = 1;
  }
  else
  {
    value = e->info.const_integer_value_info.value;
    if (e->base_type_or_null == type_long)
      semantic_error ("must be int/uint constant expression", &pos);
  }

  return value;
}

/*****************************************************************************/

PEXPRESSION parse_constant_expression_with_context (PCONTEXT context)
{
  TEXT_POSITION pos;
  PEXPRESSION   e;

  memcpy (&pos, &token.pos, sizeof(pos));

  parse_expression (context, &token.pos, null, &e);

  if (!is_constant_exp (e))
  {
    semantic_error ("must be constant expression", &pos);
    return dummy_expression();
  }

  return e;
}

/*****************************************************************************/

PEXPRESSION parse_constant_expression (void)
{
  return parse_constant_expression_with_context (&no_context);
}

/*****************************************************************************/
